
#include <fx.h>
#include <FXString.h>
#include <FXApp.h>
#include <FXScrollBar.h>
#include <FXScrollArea.h>
#include <FXImage.h>
#include <FXDCWindow.h>
#include <FXFont.h>
#include <FXPNGImage.h>
#include <FXMenuCommand.h>
#include <FXTopWindow.h>

#define CELL_PAD 5



/* In order to provide more meaningful font and color indices, I will
   define them as symbols. This will be put into both the compiler and
   the generated code. */

enum
{
    HFO_NORMAL,HFO_BOLD, HFO_ITALIC, HFO_BIG,
      /* Fixed-width fonts */
    HFO_FNORMAL, HFO_FBOLD, HFO_FITALIC, HFO_FBIG,
      /* section header fonts */
    HFO_SECHEA1, HFO_SECHEA2, HFO_SECHEA3, HFO_SECHEA4,
};

enum
{
    HCO_FG, HCO_BG,
    HCO_LINK,
    HCO_REDBG, HCO_HIBG,
      /* verbatim box colors */
    HCO_CODEBG, HCO_SHELLBG, HCO_OUTPUTBG, 
};
/* The menu commands in FOX are designed for static menus. i.e. you
   can't dynamically generate menu entries. Actually you can, but it's
   cumbersome.  A better way is to subclass the menu command class and
   let the menu command do the necessary actions by itself.

   These actions may also cause the menu command to be deleted. This seems
   to be fine. The handle() functions in the widgets don't refer to the
   object, they simply return. This could change in future releases of
   FOX but for now it's safe. */

namespace FX
{

class FXAPI FYMenuCommand : public FXMenuCommand
{
  FXDECLARE(FYMenuCommand)
protected:
  FYMenuCommand();
private:
  FYMenuCommand(const FYMenuCommand&);
  FYMenuCommand &operator=(const FYMenuCommand&);

  void (*a_func)(int);
  int a_arg;
public:
  FYMenuCommand(FXComposite* p,
                const FXString& text,FXIcon* ic=NULL,
                FXuint opts=0, 
                void (*afun)(int)= NULL, int aarg= 0);
  long run_func(FXObject* sender,FXSelector sel,void* ptr);

  virtual ~FYMenuCommand();
};
}

using namespace FX;

FXDEFMAP(FYMenuCommand) FYMenuCommandMap[]=
{
  FXMAPFUNC(SEL_COMMAND, 2216, FYMenuCommand::run_func),
};


FXIMPLEMENT(FYMenuCommand,FXMenuCommand,FYMenuCommandMap,
            ARRAYNUMBER(FYMenuCommandMap))

FYMenuCommand::FYMenuCommand(){ }
FYMenuCommand::~FYMenuCommand() { }

FYMenuCommand::FYMenuCommand
             (FXComposite* p,
              const FXString& text,FXIcon* ic,
              FXuint opts,
              void (*afun)(int), int aarg)
   : FXMenuCommand(p, text, ic, this, 2216, opts)
{
  a_func= afun;
  a_arg= aarg;
}

long FYMenuCommand::run_func(FXObject* sender,FXSelector sel,void* ptr)
{
  if (a_func) (*a_func)(a_arg);
  return 1;
}

/*
  The graphics canvas is a very simple thing. It maintains a 'document' and
  updates the screen when necessary. Later on, I'll add some double buffering
  to avoid flicker. 

             Table of Contents
   
  Constructor                     uu-cons
  Destructor                      uu-des
  Detachment                      uu-deta
  Initialization (fonts, colors)  uu-obinit
  Text Handling                   uu-text
  Text Decorations                uu-decor
  Verbatim Text                   uu-ver
  Lists                           uu-list
  Tables                          uu-table
  Images                          uu-img
  Links                           uu-link
  Painting                        uu-paint
  Resizing                        uu-resize
*/

namespace FX
{

/* This struct holds the coordinates of a text element. We have an
   array of these, one for each element. */
typedef struct { int x,y, h,a; } coord_t;

/* An imagedef is a read-only struct describing the source for an image.
   'data' is PNG encoded. */
typedef struct { int width, height; const unsigned char *data; } imagedef_t;

/* An himage is the run-time generated struct corresponding to an image. */
typedef struct { int x,y1,y2; FXImage* image; } himage_t;


class FXAPI FYGfxCanvas : public FXScrollArea
{
  FXDECLARE(FYGfxCanvas)
protected:
  FYGfxCanvas();
private:
  FYGfxCanvas(const FYGfxCanvas&);
  FYGfxCanvas &operator=(const FYGfxCanvas&);

public:
  FYGfxCanvas(FXComposite* p, FXuint opts=FRAME_NORMAL,
              FXint x=0,FXint y=0,FXint w=0,FXint h=0);
  virtual ~FYGfxCanvas();

  virtual bool canFocus() const;
  virtual void detach();

  virtual int getContentWidth();
  virtual int getContentHeight();

/* Since we have a scrolling area, we need to keep track of the maximum
   x and y coordinates for the page. This lets the scroll area to properly
   display scroll bars. */

private: int max_x,max_y, min_x,min_y;

/***********************************************************************
 *                                                                     *
 *                           Initialization                            *
 *                                                                     *
 ***********************************************************************/
/* We can handle 256 fonts. Each font has some metrics:
   - A: ascent
   - H: height
   - S: width of space */
private:
   FXFont *fonts[256];
   int fontA[256], fontH[256],fontS[256];

/* We initialize the font structures using this thing. All used fonts
   must be defined using this function. 
     FIXME: maybe we can copy a font from another (the default?)
            if a used font isn't defined.  */
public:
   void define_font(int num,const char *family,int size,const char* style);

/* In order to keep track of whether we have created fonts, we use the
   following flag. The FOX system of creation and detachment aren't very
   clear to me, so I simply create the fonts when I need them. */

private: int fonts_created;
         void create_fonts();

/* We also support 256 colors for text. These are initialized with the
   below function. Since colors don't need to be created, we don't have a
   create function or a flag for it. */

private: FXColor colors[256];
public:  void define_color(int num, unsigned int C);

/***********************************************************************
 *                                                                     *
 *                            Text Handling                            *
 *                                                                     *
 ***********************************************************************/
// uu-text
/* The canvas will primarily hold text. The text is split up into three
   logical sections:

   - words       : int -> string mapping
   - elements    : a sequence of <style,word_index> 
   - coordinates : coordinates for each element

   Words and elements are read-only and are set when the controlling widget
   sets the document. Coordinates are created by this widget and is filled
   in when the controlling function tells the canvas to paint the elements
   in a certain way. 

   The word array is set for the whole help document. Pages do share the
   word array. The word_width array is used for caching the width of 
   each word. Note that a word is the combination of a string with a
   font. i.e. the word 'the' is represented twice in the words array
   if it occurs with the default font and the top header font. */

private: const char **words;
         unsigned short *word_width;
public:  void set_words(const char **wrd,int nw);

/* In order to lay out the text properly, we need some metrics: width, height
   and ascent. We always return the font height and ascent, in order to provide
   consistent line heights. This function is used to compute the contents
   of the word_width array. */
public:  int text_width(int f, const char *str);

/* Elements are set per page. The element array consists of <style,word>
   pairs. 

   Each style is an integer of the form: 
       U-B(8)-F(8)-T(8)
   U: underline bit
   B: background color index to colors[]
   F: foreground color index to colors[]
   T: font index to fonts[]

   The set_elts function also initializes the coords array. */

private: const unsigned int *elts;
         unsigned int nelts;
         coord_t *coords;
public:  void set_elts(const unsigned int *e, unsigned int ne);

/* When printing text, we need to keep track of the 'cursor' position.
   The canvas works something like a terminal. You 'print' stuff to it.
   The stuff you can print is already in the elts[] array but you need
   to tell it where a paragraph starts and ends. The following members
   keep where you left off last time.
      FIXME: cursor_x is in fact never accessed across method invocations.
             It could be a local variable for println() and stuff. */

private: int cursor_x, cursor_y;

/* Our main method for printing text will be println. This function prints
the elements elts[start,end] as a single paragraph, and then advances
cursor_y so that later printings can continue from below. */

public:  void println(int start, int end);
         void pbreak();

/***********************************************************************
 *                                                                     *
 *                          Text Decorations                           *
 *                                                                     *
 ***********************************************************************/
// uu-decor
/* We will also have some inline decorations. These are background
   rectangles, verbatim boxes and underlines. These decorations are
   represented as integer pairs, indicating the start and end element indices
   for the decoration. i.e. rect(red,3,5) means that we should draw a red
   rectangle covering the background of elements 3,4,5.

   The following members will hold the corresponding element indices. [0] is
   start, [1] is end and [2] is color for each triplet. */

private: const unsigned int *bgrects, *underlines,*vboxes;
         unsigned int n_bgrects, n_underlines,n_vboxes;

public:  void set_vboxes(const unsigned int *d, int nd);
         void set_underlines(const unsigned int *d, int nd);
         void set_bgrects(const unsigned int *d, int nd);

private: void draw_bgrects(FXDCWindow *dc,int se,int ee);
         void draw_underlines(FXDCWindow *dc,int se,int ee);
         void draw_vboxes(FXDCWindow *dc,int se,int ee);


private: int find_decor(int se, int ee, int *sD, int *eD, 
                        const unsigned int *decor,unsigned int ndecor);

/***********************************************************************
 *                                                                     *
 *                            Verbatim Text                            *
 *                                                                     *
 ***********************************************************************/
// uu-ver
/* Vertical and left external padding for verbatim text boxes. These
   are the paddding for the text itself, not the box. */
private: int vpad_verbatim, lpad_verbatim;
         void draw_vbox(FXDCWindow *dc,int boxno);

/* Just like println, but:
   - elements are complete strings, with spaces etc. inside them
   - we don't wrap around. */
public:  void verbatim_line(int start, int end);

/* Advances cursor_y to make space. */
public:  void verbatim_vspace();

// uu-list
/***********************************************************************
 *                                                                     *
 *                                Lists                                *
 *                                                                     *
 ***********************************************************************/
/* finds out which character is present in the normal font and uses
   that as the list bullet. */
private: void figure_out_bullet();

         int lpad_paragraph;  // left padding for a normal paragraph
                              // list code modifies this to fool println.
public:  void bullet_println(int level,int start, int end);

//uu-table
/***********************************************************************
 *                                                                     *
 *                               Tables                                *
 *                                                                     *
 ***********************************************************************/
/* For tables, we do the whole layout from outside. So, we need to access
   some metrics: */

public: void get_seq_metrics(int start,int end,int *Rw,int *Rh,int *Ra);

/* Print elts[s..e] justified. x1 and x2 are absolute, y is relative
   to cursor_y. */

        void cell_left(int x1,int x2,int y,int h,int a,int s, int e);
        void cell_center(int x1,int x2, int y,int h,int a,int s,int e);
        void cell_right(int x1,int x2,int y,int h,int a,int s, int e);

/* We also need to support table borders and internal lines. These are
   a little different from the other decorations in the sense that they
   are not bound to any element but have their own absolute coordinates.

   However, their number is a fixed value and the array can be allocated
   once.  I will implement the horizontal and vertical lines separately.

   Let's start with the horizontal ones. We just need to allocate the
   associated array once when we initially switch to a page and then
   do the layout on the same array over and over without reallocating
   it. We only reallocate it when the page changes.

   The horizontal lines will be composed of three integers each: y,
   x1 and x2. Vertical lines will be very similar with the
   triplet <y1,y2,x>. */

private: int *row_lines; int n_row_lines;
         int *col_lines; int n_col_lines;

/* The following functions allocate enough space for the given number
   of lines. They don't initialize anything. They just make sure enough
   space is allocated. */
public:  int *get_row_lines(int nr);
         int *get_col_lines(int nr);

/* Some access functions needed by the table stuff. */
public:  int get_width() { return viewport_w; }
         void advance_y(int u) { cursor_y += u; max_y= cursor_y; }
         int get_cursor_y() { return cursor_y; }
         void xcoord(int x) { if (x>max_x) max_x= x; } 

private: void draw_col_lines(int y1, int y2, FXDCWindow *dc);
         void draw_row_lines(int y1, int y2, FXDCWindow *dc);

/***********************************************************************
 *                                                                     *
 *                               Images                                *
 *                                                                     *
 ***********************************************************************/
// uu-img
/* Image definition array is set by the page content function using
   set_images method. The image(n) method then later puts the given
   image at the cursor_y position with horizontal centering. */

private: int n_images;
         imagedef_t *imagedefs;
         himage_t   *images;
public:  void set_images(imagedef_t *id, int n);
         void image(int n);
private: void draw_images(FXDCWindow *dc, int y1, int y2);

/* We need to destroy images when we change pages or destroy 
   the help widget.  */

private: void destroy_images();

/***********************************************************************
 *                                                                     *
 *                                Links                                *
 *                                                                     *
 ***********************************************************************/
// uu-link
/* Links are very similar to decorations. Their representation in this
   class is actually identical: an array of integer triplets 
   <start,end,target> */
private: const unsigned int *links; int n_links;
public:  void set_links(const unsigned int*, int);
         void (*link_func)(int);
         long onLeftBtnPress(FXObject*,FXSelector,void*);
         int find_element_by_coord(int x,int y);

/***********************************************************************
 *                                                                     *
 *                              Painting                               *
 *                                                                     *
 ***********************************************************************/
// uu-paint
public:  long onPaint(FXObject* ob,FXSelector sel,void* ptr);
         int find_lower(int scr);
         int find_higher(int scr);
         void draw_everything (FXDCWindow *dc, int evX,int evY,
                                               int evW,int evH);

/***********************************************************************
 *                                                                     *
 *                              Resizing                               *
 *                                                                     *
 ***********************************************************************/
//uu-resize
private: void (*resize_func)(FYGfxCanvas *);

/* We mark the beginning and end of the page with the following functions
   so that the widget knows when to update its scrollbars. */

public:  void reset_doc();
         void end_doc(void (*f)(FYGfxCanvas*));

         virtual void layout();
         long on_layout_timeout(FXObject* ob,FXSelector sel,void* ptr);

// uz-end
};
}

/***********************************************************************
 *                                                                     *
 *                            Boiler-plate                             *
 *                                                                     *
 ***********************************************************************/
using namespace FX;

static inline int min_i(int a,int b) { return a<b? a:b; }
static inline int max_i(int a,int b) { return a>b? a:b; }

FXDEFMAP(FYGfxCanvas) FYGfxCanvasMap[]=
{
  FXMAPFUNC(SEL_PAINT,0,FYGfxCanvas::onPaint),
  // uu-resize
  FXMAPFUNC(SEL_TIMEOUT, 0, FYGfxCanvas::on_layout_timeout),
  // uu-link
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0, FYGfxCanvas::onLeftBtnPress),
};

FXIMPLEMENT(FYGfxCanvas,FXScrollArea,FYGfxCanvasMap,
            ARRAYNUMBER(FYGfxCanvasMap))

FYGfxCanvas::FYGfxCanvas(){ flags|=FLAG_ENABLED|FLAG_SHOWN; }
bool FYGfxCanvas::canFocus() const { return true; }
int FYGfxCanvas::getContentHeight() { return max_y; }
int FYGfxCanvas::getContentWidth() { return max_x; }


/***********************************************************************
 *                                                                     *
 *                      Constructor / Destructor                       *
 *                                                                     *
 ***********************************************************************/

FYGfxCanvas::FYGfxCanvas
     (FXComposite* p, FXuint opts, FXint x,FXint y,FXint w,FXint h):
		FXScrollArea(p,opts | VSCROLLER_ALWAYS ,x,y,w,h)
{
  flags|=FLAG_ENABLED;

  fonts_created= 0;
  { int i; for(i=0;i<256;i++) fonts[i]= NULL; }

  coords= NULL; words= NULL; elts= NULL; nelts= 0;
  word_width= NULL;

  cursor_x= cursor_y=0;

  max_x= max_y= min_x= min_y= 0;

  images= NULL; imagedefs= NULL; n_images= 0;

  bgrects= underlines= vboxes= NULL;
  n_bgrects= n_underlines= n_vboxes= 0;

  vpad_verbatim= 6; lpad_verbatim= 15; 
  lpad_paragraph= 10;

  // uu-resize
  resize_func= NULL;

  // uu-table
  row_lines= NULL; n_row_lines=0;
  col_lines= NULL; n_col_lines=0;

// uu-cons
}


void FYGfxCanvas::detach()
{ 
  int i;

  FXScrollArea::detach();

  if (fonts_created)
     for(i=0;i<256;i++) 
        if (fonts[i]) fonts[i]->detach(); 
// uu-deta
}

FYGfxCanvas::~FYGfxCanvas() { 
  int i;

  for(i=0;i<256;i++) 
     if (fonts[i]) delete fonts[i];

  if (coords) { free(coords); coords= NULL; }

  destroy_images();

  // uu-table
  if (row_lines) free(row_lines);
  if (col_lines) free(col_lines);
// uu-des
}


/***********************************************************************
 *                                                                     *
 *                           Initialization                            *
 *                                                                     *
 ***********************************************************************/
// uu-obinit
void FYGfxCanvas::define_font
   (int num,const char *family,int size,const char* style)
{
   int weight;
   int slant;
   if (num>=256) return ;
   
   if (strchr(style,'B'))
      weight= FXFont::Bold;
   else
      weight= FXFont::Normal;
   if (strchr(style,'I'))
      slant= FXFont::Italic;
   else
      slant= FXFont::Straight;

   fonts[num]= new FXFont(FXApp::instance(),family, size, weight,slant);
}

void FYGfxCanvas::create_fonts()
{
  if (!fonts_created)
  {
    int i;
    fonts_created= 1;
    for(i=0;i<256;i++) 
      if (fonts[i])
      {
         fonts[i]->create();
         fontA[i]= fonts[i]->getFontAscent();
         fontH[i]= fonts[i]->getFontHeight();
         fontS[i]= fonts[i]->getTextWidth(" ", 1);
      }
    figure_out_bullet();
  }
}

void FYGfxCanvas::define_color(int num, unsigned int C)
{
   unsigned R,G,B;
   if (num>=256) return ;
   R= (C>>16u)&0xffu;
   G= (C>>8u)&0xffu;
   B= C&0xffu;
   colors[num]= FXRGB(R,G,B);
}

/***********************************************************************
 *                                                                     *
 *                            Text Handling                            *
 *                                                                     *
 ***********************************************************************/

// uu-text
void FYGfxCanvas::set_words(const char **wrd,int nw)
{
  words= wrd;
  if (word_width) free(word_width);
  word_width= (typeof(word_width)) calloc(nw, sizeof(word_width[0]));
}

void FYGfxCanvas::set_elts(const unsigned int *e, unsigned int ne)
{
  elts= e;
  nelts= ne;
  if (coords) free(coords);
  coords= (typeof(coords)) malloc(sizeof(*coords)*ne);
}

int FYGfxCanvas::text_width(int f, const char *str)
{
  create_fonts();
  if (f<0 || f>=256 || !fonts[f]) return 1;
  return fonts[f]->getTextWidth(str, strlen(str));
}

void FYGfxCanvas::println(int start, int end)
{
  int i,j,f,ww,w;
  int row_a, row_h;
  int canvas_w;

  int os= start;

  create_fonts();
  canvas_w= viewport_w-lpad_paragraph-10; //FIXME: we need rpad_paragraph also
again:
  cursor_x= lpad_paragraph;
  row_a= fontA[0];
  row_h= fontH[0];
  for(i=start;i<=end;i++)
  {
    f= elts[2*i]&0xff;
    w= elts[2*i+1];
    ww= word_width[w];
    if (ww==0) 
       word_width[w]= ww= text_width(f, words[w]);
    if (cursor_x+ww > canvas_w && i!=start) break;
    coords[i].x= cursor_x;
    cursor_x+= ww ;
    max_x= max_i(max_x, cursor_x);
    row_a= max_i(row_a, fontA[f]);
    row_h= max_i(row_h, fontH[f]);
    cursor_x+= fontS[f];
  }

  for(j=start;j<i;j++)
  {
    coords[j].y= cursor_y;
    coords[j].h= row_h;
    coords[j].a= row_a;
  }
  cursor_y += row_h + 2;
  max_y= cursor_y;

  if (i<=end) { start= i; goto again; }

  if (0)
  { printf("-----------------\n");
     for(i=os;i<=end;i++)
        printf("(x=%d,y=%d,a=%d,h=%d,w=%d)[%d:%s]\n", coords[i].x, 
                        coords[i].y, coords[i].a, coords[i].h,
                        word_width[elts[2*i+1]],
                        elts[2*i+1],
                        words[elts[2*i+1]]); }
}

void FYGfxCanvas::pbreak()
{
  create_fonts();
  cursor_y+= fontH[0];
  cursor_x= 0;
}


/***********************************************************************
 *                                                                     *
 *                          Text Decorations                           *
 *                                                                     *
 ***********************************************************************/
// uu-decor
void FYGfxCanvas::set_bgrects(const unsigned int *d, int nd)
{
  bgrects= d;
  n_bgrects= nd;
}

void FYGfxCanvas::set_underlines(const unsigned int *d, int nd)
{
  underlines= d;
  n_underlines= nd;
}

void FYGfxCanvas::set_vboxes(const unsigned int *d, int nd)
{
  vboxes= d;
  n_vboxes= nd;
}


/* Just like we find the elements visible within a window, we need to find
   decorations visible within a given element-window. */

int FYGfxCanvas::find_decor(int se, int ee, int *sD, int *eD, 
                const unsigned int *decor,unsigned int ndecor)
{
#if 1
  int i,j,k,S,E;
  i=0; j= ndecor-1;
  while(i<=j)
  {
    k= (i+j)/2;
    S= decor[k*3]; E= decor[k*3+1];
    if (E<se) { i= k+1; continue; }
    if (S>ee) { j= k-1; continue; }
    goto found;
  }
  return 1;
found:
  i=j= k;
  while(i>=0 && decor[i*3+1] >= se)    { *sD= i; i--; }
  while(j<=ndecor-1 && decor[j*3]<=ee) { *eD= j; j++;  }
  //printf("decor found for (%d %d) : (%d %d)\n", se, ee, *sD, *eD); 
  return 0;
#else

  int i,j,k;
  for(i=0;i<ndecor;i++)
    if ((decor[i*3+1]>=se && decor[i*3+1]<=ee)
        || (decor[i*3]>=se && decor[i*3]<=ee) )
        { *sD= *eD= i; break; }
  if (i==ndecor) return 1;

  for(i=ndecor-1;i>=0;i--)
    if ((decor[i*3+1]>=se && decor[i*3+1]<=ee)
        || (decor[i*3]>=se && decor[i*3]<=ee) )
        { *eD= i; break; }

  return 0;
#endif
}

void FYGfxCanvas::draw_bgrects(FXDCWindow *dc,int se,int ee)
{
  int sb, eb;
  if (!n_bgrects || !bgrects) return ;
  if (find_decor(se, ee, &sb, &eb, bgrects, n_bgrects)) return ;
  
  while(sb<=eb)
  {
    int S,E,C;
    S= bgrects[sb*3];
    E= bgrects[sb*3+1];
    C= bgrects[sb*3+2];

    int sx, ex;
    int sy,h;

    dc->setForeground(colors[C]);
    while(S<=E)
    {
      h= coords[S].h;
      sx= coords[S].x;
      sy= coords[S].y;
      while(S<=E && coords[S].y==sy) 
         { ex= coords[S].x+word_width[elts[2*S+1]]; S++; }
      sx-= 1; ex+= 1;
      dc->fillRectangle(sx+pos_x, sy+pos_y, ex-sx, h);
    }
    sb++;
  }
}

void FYGfxCanvas::draw_underlines(FXDCWindow *dc,int se,int ee)
{
  int sb, eb;
  if (!n_underlines || !underlines) return ;
  if (find_decor(se, ee, &sb, &eb, underlines, n_underlines)) return ;
  
  while(sb<=eb)
  {
    int S,E,C;
    S= underlines[sb*3];
    E= underlines[sb*3+1];
    C= underlines[sb*3+2];

    int sx, ex;
    int sy,a;

    dc->setForeground(colors[C]);
    while(S<=E)
    {
      a= coords[S].a;
      sx= coords[S].x;
      sy= coords[S].y;
      while(S<=E && coords[S].y==sy) 
         { ex= coords[S].x+word_width[elts[2*S+1]]; S++; }
      dc->drawLine(sx+pos_x, sy+a+pos_y+2, ex+pos_x, sy+a+pos_y+2);
    }
    sb++;
  }
}
/***********************************************************************
 *                                                                     *
 *                            Verbatim Text                            *
 *                                                                     *
 ***********************************************************************/
// uu-ver
void FYGfxCanvas::verbatim_vspace()
{
  cursor_y+= vpad_verbatim;
  max_y= cursor_y;
}

/* we just keep on going, never wrapping around. */
void FYGfxCanvas::verbatim_line(int start, int end)
{
  int i,j,f,ww,w;
  int row_a, row_h;

  int os= start;

  create_fonts();

  cursor_x= lpad_verbatim;
  row_a= fontA[0];
  row_h= fontH[0];
  for(i=start;i<=end;i++)
  {
    f= elts[2*i]&0xff;
    w= elts[2*i+1];
    ww= word_width[w];
    if (ww==0) 
       word_width[w]= ww= text_width(f, words[w]);
    coords[i].x= cursor_x;
    cursor_x+= ww;
    max_x= max_i(max_x, cursor_x);
    row_a= max_i(row_a, fontA[f]);
    row_h= max_i(row_h, fontH[f]);
  }

  for(j=start;j<=end;j++)
  {
    coords[j].y= cursor_y;
    coords[j].h= row_h;
    coords[j].a= row_a;
  }
  cursor_y += row_h + 2;
  max_y= cursor_y;
}


/* In this function, there is a huge off-by-one error which I can't
   figure out. I hacked around it by making the rectangle 1 pixel
   smaller than required. So, the actual spacing between the vbox
   contents and the edge is 1 pixel less than vpad_verbatim. */
void FYGfxCanvas::draw_vboxes(FXDCWindow *dc,int sY,int eY)
{
  int sb,eb,b, S,E,k;
  if (!n_vboxes || !vboxes) return ;

#if 0
  FXColor white;
  static unsigned int K;
  static int C[4];
  C[0]= FXRGB(0x00,0xff,0x00);
  C[1]= FXRGB(0x00,0x00,0xff);
  C[2]= FXRGB(0xff,0x00,0xff);
  C[3]= FXRGB(0x33,0x33,0x33);
  K= (K+1)%4;
  white= C[K];

  dc->setForeground(white);
  dc->fillRectangle(0, sY+pos_y, 400, eY-sY+1);
#endif

  sb= 0; eb= n_vboxes-1;
  while(sb<=eb)
  {
    k= (sb+eb)/2;
    S= vboxes[k*3];
    E= vboxes[k*3+1];
    if (coords[S].y-vpad_verbatim+1>eY)             { eb= k-1; continue; }
    if (coords[E].y+coords[E].h+vpad_verbatim-1<sY) { sb= k+1; continue; }
    goto found;
  }
  return ;
found:
  draw_vbox(dc, k);
  for(b=k-1;b>=0;b--) 
  {
    S= vboxes[b*3];
    E= vboxes[b*3+1];
    if (coords[E].y+coords[E].h+vpad_verbatim-1 < sY) break;
    draw_vbox(dc,b);
  }
  for(b=k+1;b<n_vboxes;b++)
  {
    S= vboxes[b*3];
    E= vboxes[b*3+1];
    if (coords[S].y-vpad_verbatim+1 > eY) break;
    draw_vbox(dc,b);
  }
}

void FYGfxCanvas::draw_vbox(FXDCWindow *dc,int boxno)
{
  int S,E,C;
  int i;
  int mx;
  S= vboxes[boxno*3];
  E= vboxes[boxno*3+1];
  C= vboxes[boxno*3+2];

  dc->setForeground(colors[C]);

  mx= word_width[elts[2*S+1]] + coords[S].x;
  for(i=S+1;i<=E;i++)
     mx= max_i(mx,word_width[elts[2*i+1]] + coords[i].x); 

  int x1,x2,y1,y2;
  x1= lpad_paragraph;
  y1= coords[S].y-vpad_verbatim+1;
  x2= mx+ lpad_verbatim-lpad_paragraph;
  y2= coords[E].y+coords[E].h + vpad_verbatim-1;
  dc->fillRectangle(x1+pos_x, y1+pos_y, x2-x1, y2-y1);
  dc->setForeground(colors[HCO_FG]);
  dc->drawRectangle(x1+pos_x, y1+pos_y, x2-x1, y2-y1);
}



/***********************************************************************
 *                                                                     *
 *                                Lists                                *
 *                                                                     *
 ***********************************************************************/
// uu-list
/* In order to display lists, we use 'bullet paragraphs'. What we do is:
    - make some space on the left for indentation
    - print a bullet character
    - adjust the paragraph start to point after the bullet character
    - call println()

   The bullet character is in fact a string, just like other text. This
   string is somewhat special, its value depends on the font. If the
   normal font doesn't contain the default bullet character, other
   characters should be tried.

   So, when we set up the fonts, I should try to get the width of the
   bullet string. Normally, I'd have the 'words' array as:

      const char **const words

   but this wouldn't let me modify words[0], which will be the bullet
   string.  So, I will keep it as:

      const char **words

   and then modify words[0] accordingly. */

void FYGfxCanvas::figure_out_bullet()
{
  int k,i;
  const char *blt[]= { "✦","⚫","▪","★" ,"♦" ,"◆" ,"✸" ,"*", NULL};
  for(i=0;blt[i];i++)
  {
    FXString str= blt[i];
    if (fonts[0]->hasChar(str.wc(0)))
    {
       k= text_width(0, blt[i]);
       break;
    }
  }
  if (!blt[i]) { i--; k= 16; }
  words[0]= blt[i];
  word_width[0]= k;
}


/* The following function does the actual bullet paragraph printing.
   elts[2*start] should contain <normal,0>, which points to the bullet
   string. The rest is the actual content for the paragraph. */

void FYGfxCanvas::bullet_println(int level,int start, int end)
{
  int lp;
  unsigned int bw, bs;
  lp= lpad_paragraph;

  lpad_paragraph= level*30;  // FIXME: make it configurable
  bw= elts[2*start+1];
  bs= elts[2*start];

  if (word_width[bw]==0) 
     word_width[bw]= text_width(bs&0xff, words[bw]);

  coords[start].x= lpad_paragraph-word_width[bw] - 1;
  coords[start].y= cursor_y;
  coords[start].h= fontH[0]; 
  coords[start].a= fontA[0]; 

  println(start+1, end);
  if (coords[start+1].y==coords[start].y)
  {
   coords[start].h= max_i(coords[start].h , coords[start+1].h);
   coords[start].a= max_i(coords[start].a , coords[start+1].a);
  }

  lpad_paragraph= lp;
}

/* Note that using words[0] isn't mandatory at all. You can use whatever
   you want. Just put it in the elt[] array and this function will use
   the first word as the bullet. */

/***********************************************************************
 *                                                                     *
 *                               Tables                                *
 *                                                                     *
 ***********************************************************************/
// uu-table
/* When laying out table cells, we do the width / height computation
   of cells outside of canvas. We also compute the positions of the
   cells at the same place. Positioning the cell text inside the cell
   is done by the following functions.  _left means left justified,
   _right means right justified etc.

   The given y value is relative to cursor_y. It's not updated at this
   time.  However, the x values are absolute. The centering is also done
   outside since the canvas itself doesn't know anything about the size
   of the table. */

void FYGfxCanvas::cell_left(int x1,int x2,int y,int h,int a,int s, int e)
{
   int i;
   for(i=s;i<=e;i++)
   {
     if (i!=s)
       x1+= fontS[ elts[2*i] & 0xff ];
     coords[i].x= x1; 
     coords[i].y= cursor_y + y;
     coords[i].h= h;
     coords[i].a= a;
     x1+= word_width[ elts[2*i+1] ];
     max_x= max_i(max_x,x1);
   }
}

void FYGfxCanvas::cell_center(int x1,int x2, int y,int h,int a,int s,int e)
{
   int x,d,i;
   x= x1;
   for(i=s;i<=e;i++)
   {
     if (i!=s)
       x+= fontS[ elts[2*i] & 0xff ];
     coords[i].x= x; 
     coords[i].y= cursor_y + y;
     coords[i].h= h;
     coords[i].a= a;
     x+= word_width[ elts[2*i+1] ];
     max_x= max_i(max_x,x);
   }
   d= (x2-x)/2;
   for(i=s;i<=e;i++)
     coords[i].x+= d;
}

void FYGfxCanvas::cell_right(int x1,int x2,int y,int h,int a,int s, int e)
{
   int x,d,i;
   x= x1;
   for(i=s;i<=e;i++)
   {
     if (i!=s)
       x+= fontS[ elts[2*i] & 0xff ];
     coords[i].x= x; 
     coords[i].y= cursor_y + y;
     coords[i].h= h;
     coords[i].a= a;
     x+= word_width[ elts[2*i+1] ];
     max_x= max_i(max_x,x);
   }
   d= (x2-x);
   for(i=s;i<=e;i++) coords[i].x+= d;
}

/* This function is used for computing the width/height of a cell.
   It returns the dimensions of the given text elts[start,end] 
   as if it was displayed on a single row. */

void FYGfxCanvas::get_seq_metrics(int start,int end,int *Rw,int *Rh,int *Ra)
{
  int i;
  int W,S,F;
  int w,h,a;
  w= h= a= 0;

  create_fonts();
  for(i=start;i<=end;i++)
  {
     W= elts[2*i+1];
     S= elts[2*i];
     F= S & 0xff;
     if (i!=start) w+= fontS[F];
     if (word_width[W]==0) word_width[W]= text_width(F, words[W]);
     w+= word_width[W];
     h= max_i(h, fontH[F]);
     a= max_i(a, fontA[F]);
  }
  *Rw= w; *Ra= a; *Rh= h;
}

int *FYGfxCanvas::get_row_lines(int nr)
{
  if (n_row_lines==nr) return row_lines;
  if (row_lines) free(row_lines);
  row_lines= NULL;
  n_row_lines= nr;
  if (n_row_lines) 
    row_lines= (typeof(row_lines)) malloc(3*sizeof(row_lines[0])*nr);
  return row_lines;
}

int *FYGfxCanvas::get_col_lines(int nr)
{
  if (n_col_lines==nr) return col_lines;
  if (col_lines) free(col_lines);
  col_lines= NULL;
  n_col_lines= nr;
  if (n_col_lines) 
    col_lines= (typeof(col_lines)) malloc(3*sizeof(col_lines[0])*nr);
  return col_lines;
}


void FYGfxCanvas::draw_row_lines(int y1, int y2, FXDCWindow *dc)
{
  int i,j,k;
  if (!n_row_lines || !row_lines) return ;

  i= 0; j= n_row_lines-1;
  while(i<=j)
  {
    k= (i+j)/2;
    if (row_lines[3*k] < y1) i= k+1; 
    else if (row_lines[3*k] > y2) j= k-1; 
    else break;
  }
  if (i>j) return ;
  dc->setForeground(colors[HCO_FG]);

  for(i=k;i>=0 && row_lines[3*i]>=y1;i--)
    dc->drawLine(row_lines[3*i+1]+pos_x, row_lines[3*i]+pos_y,
                 row_lines[3*i+2]+pos_x, row_lines[3*i]+pos_y);

  for(j=k+1;j<n_row_lines && row_lines[3*j]<=y2;j++)
    dc->drawLine(row_lines[3*j+1]+pos_x, row_lines[3*j]+pos_y,
                 row_lines[3*j+2]+pos_x, row_lines[3*j]+pos_y);
}


void FYGfxCanvas::draw_col_lines(int y1, int y2, FXDCWindow *dc)
{
  int i,j,k;
  if (!n_col_lines || !col_lines) return ;

  i= 0; j= n_col_lines-1;
  while(i<=j)
  {
    k= (i+j)/2;
    if (col_lines[3*k+1] < y1) i= k+1; 
    else if (col_lines[3*k] > y2) j= k-1; 
    else break;
  }
  if (i>j) return ;
  dc->setForeground(colors[HCO_FG]);

  for(i=k;i>=0 && col_lines[3*i+1]>=y1;i--)
    dc->drawLine(col_lines[3*i+2]+pos_x, col_lines[3*i]+pos_y,
                 col_lines[3*i+2]+pos_x, col_lines[3*i+1]+pos_y);

  for(j=k+1;j<n_col_lines && col_lines[3*j]<=y2;j++)
    dc->drawLine(col_lines[3*j+2]+pos_x, col_lines[3*j]+pos_y,
                 col_lines[3*j+2]+pos_x, col_lines[3*j+1]+pos_y);
}


/***********************************************************************
 *                                                                     *
 *                               Images                                *
 *                                                                     *
 ***********************************************************************/
// uu-img 
void FYGfxCanvas::set_images(imagedef_t *id, int n)
{
  if (id==imagedefs) return ;
  destroy_images();
  n_images= n;
  if (!n_images) return ;
  imagedefs= id;
  images= (typeof(images)) calloc(n, sizeof(images[0]));
}


void FYGfxCanvas::image(int n)
{
  images[n].y1= cursor_y;
  cursor_y+= imagedefs[n].height;
  images[n].y2= cursor_y-1;
  images[n].x= max_i(0,(viewport_w - imagedefs[n].width)/2);
  max_x= max_i(max_x, images[n].x + imagedefs[n].width);
  max_y= cursor_y;
}

void FYGfxCanvas::draw_images(FXDCWindow *dc, int y1, int y2)
{
  int i,j,k;
  if (!n_images) return ;
  i= 0; j= n_images-1;
  while(i<=j)
  {
    k= (i+j)/2;
    if (images[k].y1>y2) i= k+1; 
    else if (images[k].y2<y1) j= k-1;
    else break;
  }
  if (i>j) return ;
  for(i=k;i>0 && images[i-1].y2>=y1;i--) ;
  for(j=k;j<n_images-1 && images[j+1].y1<=y2;j++) ;

  for(k=i;k<=j;k++)
  {
    if (!images[k].image)
    {
      images[k].image= new FXPNGImage(FXApp::instance(), imagedefs[k].data);
      images[k].image->create();
    }
    dc->drawImage(images[k].image, images[k].x+pos_x, images[k].y1+pos_y);
  }
}

void FYGfxCanvas::destroy_images()
{
  int i;
  for(i=0;i<n_images;i++)
     if (images[i].image)
     {
         images[i].image->detach(); 
         delete images[i].image;
     }
  n_images= 0;
  if (images) free(images);
  imagedefs= NULL;
  images= NULL;
}

/***********************************************************************
 *                                                                     *
 *                                Links                                *
 *                                                                     *
 ***********************************************************************/
// uu-link
void FYGfxCanvas::set_links(const unsigned int* lk, int n)
{
  links= lk;
  n_links= n;
}

int FYGfxCanvas::find_element_by_coord(int evX,int evY)
{
  int i,j,k;

  i= 0; j= nelts-1;
  while(i<=j)
  {
    k= (i+j)/2;
    if (coords[k].y>evY) j= k-1;
    else if (coords[k].y+coords[k].h<=evY) i= k+1;
    else break;
  }
  if (i>j) return -1;

  for(i=k;i>=0;i--)
  {
    if (coords[i].y!=coords[k].y) break;
    if (coords[i].x+word_width[elts[2*i+1]]<evX) break;
    if (coords[i].x<=evX) return i;
  }
  for(i=k+1;i<nelts;i++)
  {
    if (coords[i].y!=coords[k].y) break;
    if (coords[i].x>evX) break;
    if (coords[i].x+word_width[elts[2*i+1]]>evX) return i;
  }
  return -1;
}

long FYGfxCanvas::onLeftBtnPress(FXObject* obj,FXSelector sel,void* ptr)
{
  FXEvent *event;
  int evX, evY;
  int k;
  int sl, el;

  event= (FXEvent*) ptr;
  if (!event || !nelts) return 1;
  evX= event->win_x-pos_x;
  evY= event->win_y-pos_y;

  k= find_element_by_coord(evX, evY);
  
  if (k==-1) return 1;

  if (find_decor(k, k, &sl, &el, links, n_links)) return 1;
  if (link_func) (*link_func)(links[3*sl+2]);
 
  return 1;
}


/***********************************************************************
 *                                                                     *
 *                              Painting                               *
 *                                                                     *
 ***********************************************************************/
// uu-paint

/* The disabled part below was for a delayed full-screen paint.
   It does work but it's ugly since the window borders moving around
   leave marks on the canvas. */
long FYGfxCanvas::onPaint(FXObject* ob,FXSelector sel,void* ptr)
{
  FXEvent* event=(FXEvent*)ptr;
#if 0
  if (0 && event->rect.x==0 && event->rect.y==0 &&
      event->rect.w>= width && event->rect.h>=height)
  {
    FXApp::instance()->addTimeout(this, 1, 250, NULL);
  } 
  else 
#endif
  {
    FXDCWindow dc(this,event);
    draw_everything( &dc, event->rect.x, event->rect.y,
                        event->rect.w, event->rect.h);
  }
  return 1;
}

/* I'm not very sure about the following functions. They find the lowest and
   heighest displayed row. This is done by a binary search.

   The first function finds the first row which intersects with the screen.
   i.e. y<scr y+h>=scr. */
int FYGfxCanvas::find_lower(int scr)
{
  int i,j,k,y1,y2;
  i=0; j= nelts-1;
  while(i<=j && coords[i].y!=coords[j].y)
  {
    k= (i+j)/2;
    y1= coords[k].y; y2= coords[k].y+coords[k].h;
    if (y1<=scr)
    { 
       if (y2>=scr) goto done;
       i= k+1;
       continue;
    }
    j= k-1;
  }
  if (i>j) return 0;
  k= i;
done:
  while(k>0 && coords[k-1].y==coords[k].y) k--;
  return k;
}


int FYGfxCanvas::find_higher(int scr)
{
  int i,j,k,y1,y2;
  i=0; j= nelts-1;
  while(i<=j && coords[i].y!=coords[j].y)
  {
    k= (i+j)/2;
    y1= coords[k].y; y2= coords[k].y+coords[k].h;
    if (y1<=scr)
    { 
       if (y2>=scr) goto done;
       i= k+1;
       continue;
    }
    j= k-1;
  }
  if (i>j) return nelts-1;
  k= j;
done:
  while(k<nelts-1 && coords[k+1].y==coords[k].y) k++;
  return k;
}


void FYGfxCanvas::draw_everything
  (FXDCWindow *dc, int evX,int evY, int evW,int evH)
{
  FXColor white;
  int se, ee,e;
  unsigned int style, word;

  //white= FXRGB(0xff,0xff,0xff);
  white= colors[HCO_BG];
#if 0
  static unsigned int K;
  static int C[4];
  C[0]= FXRGB(0x00,0xff,0x00);
  C[1]= FXRGB(0x00,0x00,0xff);
  C[2]= FXRGB(0xff,0x00,0xff);
  C[3]= FXRGB(0x33,0x33,0x33);
  K= (K+1)%4;
  white= C[K];
#endif

  dc->setForeground(white);
  dc->fillRectangle(evX, evY, evW, evH);
  if (!coords || !elts || !words) return ;

  create_fonts();

  evY-= pos_y; evX-= pos_x;

  se= find_lower( evY );
  ee= find_higher(evY+evH-1);

  // printf("y1= %d y2= %d se= %d ee= %d\n", evY, evY+evH-1, se, ee);

  // When the exposed area is just before the first element or just
  // after the last element in a verbatim environment, the tiny area
  // doesn't get repainted because it isn't inside the area of any
  // element. We need to figure out the start and end positions for
  // vboxes separately.
  //
  // We can do this by moving the event Y to a higher
  // or lower value and then searching by doing find_higher/lower. 
  //
  // We could also change the height of the first and last rows, which is
  // a cleaner option really. So, for the first row, we would increase
  // the ascent and height so that the row starts from the edge of the
  // box. For the last row, we would increase the height only, starting
  // the row from the end of the previous row but extending it down to
  // the bottom of the box.
  //
  // That doesn't work. Actually it does, but it also messes up any
  // background decoration on the first and last rows. They also extend
  // to the edge of the box. So, we really need to find the boundaries
  // for the vboxes separately.
  //
  draw_vboxes(dc, evY,evY+evH-1);
  
  draw_bgrects(dc, se,ee);

  draw_row_lines(evY, evY+evH-1, dc);
  draw_col_lines(evY, evY+evH-1, dc);

  if (0) { int i; printf("[PAINT]"); 
           for(i=se;i<=ee;i++) printf("%s",words[elts[2*i+1]]); 
           printf("\n"); }

  for(e=se;e<=ee;e++)
  {
     word= elts[e*2+1];
     style= elts[e*2];
     dc->setForeground(colors[(style>>8)&0xff]);
     dc->setFont(fonts[style&0xff]);
     dc->drawText(coords[e].x + pos_x,
                 coords[e].y+ coords[e].a + pos_y, 
                 words[word], strlen(words[word])); 
  }

  draw_underlines(dc, se,ee);
  draw_images(dc, evY, evY+evH-1);
}

/***********************************************************************
 *                                                                     *
 *                              Resizing                               *
 *                                                                     *
 ***********************************************************************/
// uu-resize

/* When we resize the widget, we will need to lay the text out again. This
   will be done by the function which did the initial layout. We simply
   start from scratch. So, we store the function somewhere and then call
   it when a resize occurs. However, we shouldn't do re-layouts all the
   time while the window is being resized. They discussed this in some
   email list and Jeroen suggested a timer. I'll try that. 

   It's not necessary to catch the resize event. It's OK if we simply
   override the layout method. If we don't have a resize function yet,
   we simply don't do anything since the widget is empty.

   Otherwise, we avoid laying out by starting a timer. We only do the
   re-layout when the timer expires. */

long FYGfxCanvas::on_layout_timeout(FXObject* ob,FXSelector sel,void* ptr)
{
  if (!resize_func) return 1;
  (*resize_func)(this);
  update();
  return 1;
}


void FYGfxCanvas::layout()
{
  if (!resize_func) goto done;
  // 250 msecs..
  FXApp::instance()->addTimeout(this, 0, 250, NULL);
done:
  FXScrollArea::layout();
}

void FYGfxCanvas::reset_doc()
{
  cursor_y= max_y= max_x= 0;
}

void FYGfxCanvas::end_doc(void (*f)(FYGfxCanvas*))
{
  resize_func= f;
  FXScrollArea::layout();
  update();
}
/*
  The help window will contain the following:
   - a menu bar for navigation
   - a toolbar for displaying buttons such as prev,next, search, hprev, etc.
   - a canvas to display the help text
*/

extern "C" void help_setting_i(const char *key, int value);
extern "C" void help_setting_s(const char *key, const char* value);
extern "C" void help(const char *subject);

namespace FX
{

typedef struct { const char *Vs; int Vi; } sipair_t;

class FXAPI FYHelpWin : public FXTopWindow
{
  FXDECLARE(FYHelpWin)
protected:
  FYHelpWin();
private:
  FYHelpWin(const FYHelpWin&);
  FYHelpWin &operator=(const FYHelpWin&);
public:
  FYHelpWin (FXApp* ap,const FXString& name, FXIcon *ic,FXIcon *mi,
            FXuint opts, FXint w,FXint h,
            FXint x=0,FXint y=0,
            FXint pl=0,FXint pr=0,FXint pt=0,FXint pb=0,
            FXint hs=0,FXint vs=0) ;
  virtual ~FYHelpWin();

// uu-chapsec
/* Chapters and sections will be accessible thru menus. */
private:
  FXMenuPane *chapters_pane;
  FXMenuPane *sections_pane;
  FXMenuTitle *sections_mtitle;

/* The chapter menu will never change. We don't need to keep track of
   that stuff. We simply provide a method to record them into the menu. */

public:  void menu_add_chapter(const char *title, int number);

/* The section menu will be set as a whole, rather than incremental 
   method calls. */
private: const sipair_t *sections_def;
public:  void set_section_menu(const sipair_t *);

public:
  FYGfxCanvas *canvas;
};
}

using namespace FX;

static void follow_link(int);
static void show_toc2(int);
static void show_toc0(int);


FXDEFMAP(FYHelpWin) FYHelpWinMap[]=
{
};


FXIMPLEMENT(FYHelpWin,FXTopWindow,FYHelpWinMap, ARRAYNUMBER(FYHelpWinMap))

FYHelpWin::FYHelpWin(){ flags|=FLAG_ENABLED|FLAG_SHOWN; }

/* Configuration variables and functions here */

static const char *CG_normal_font= "Sans";
static const char *CG_fixed_font= "Mono";
static int CG_normal_size = 12;
static int CG_normal_big  = 14;
static int CG_fixed_size  = 12;
static int CG_fixed_big  = 14;

static int CG_sechea1 = 14;
static int CG_sechea2 = 16;
static int CG_sechea3 = 18;
static int CG_sechea4 = 20;

static int CG_link = 0xff;
static int CG_red = 0xc67171;
static int CG_hilight= 0xb5cae1;
static int CG_code= 0xafe5b2;
static int CG_shell= 0xecabab;
static int CG_output= 0xade6ff;


void help_setting_i(const char *key, int value)
{
static const struct { const char *name; int *tgt; } opts[]=
{
  { "size",      &CG_normal_size },
  { "big",       &CG_normal_big },
  { "mono_size", &CG_fixed_size },
  { "mono_big",  &CG_fixed_big },
  { "header1",   &CG_sechea1 },
  { "header2",   &CG_sechea2 },
  { "header3",   &CG_sechea3 },
  { "header4",   &CG_sechea4 },
  { "link",      &CG_link },
  { "hilight",   &CG_hilight },
  { "red",       &CG_red },
  { "code",      &CG_code },
  { "shell",     &CG_shell },
  { "output",    &CG_output },
  { NULL, NULL }
};
  int i;
  for(i=0;opts[i].name;i++)
    if (!strcmp(opts[i].name,key)) { *(opts[i].tgt)= value; break; }
}

void help_setting_s(const char *key, const char* value)
{
static const struct { const char *name; const char **tgt; } opts[]=
{
  { "font",       &CG_normal_font },
  { "mono_font",  &CG_fixed_font },
  { NULL, NULL }
};
  int i;
  for(i=0;opts[i].name;i++)
    if (!strcmp(opts[i].name,key)) { *(opts[i].tgt)= strdup(value); break; }
}

FYHelpWin::FYHelpWin
           (FXApp* ap,const FXString& name, FXIcon *ic,FXIcon *mi,
            FXuint opts, FXint w,FXint h,
            FXint x,FXint y,
            FXint pl,FXint pr,FXint pt,FXint pb,
            FXint hs,FXint vs) :
 FXTopWindow(ap,name,ic,mi, opts, x,y,w,h, pl,pr,pt,pb, hs,vs)
{
  FXVerticalFrame *vbox;
  flags|=FLAG_ENABLED|FLAG_SHOWN; 

  vbox= new FXVerticalFrame(this, LAYOUT_FILL_X | LAYOUT_FILL_Y);

  FXMenuBar *bar;
  bar= new FXMenuBar(vbox, LAYOUT_FILL_X);

  chapters_pane= new FXMenuPane(this);
  sections_pane= new FXMenuPane(this);

  new FXMenuTitle(bar,"Chapters", NULL, chapters_pane);
  sections_mtitle= new FXMenuTitle(bar,"Sections", NULL, sections_pane);

  new FYMenuCommand(bar, "Contents", NULL, 0, show_toc2, 0);
  new FYMenuCommand(bar, "Full", NULL, 0, show_toc0, 0);

  canvas= new FYGfxCanvas(vbox, FRAME_NORMAL | LAYOUT_FILL_X | LAYOUT_FILL_Y);
  

  const char *normal_font, *fixed_font;
  normal_font= "Dejavu Sans"; fixed_font= "Mono";

  canvas->define_font(HFO_NORMAL, CG_normal_font, CG_normal_size, "");
  canvas->define_font(HFO_BOLD, CG_normal_font, CG_normal_size, "B");
  canvas->define_font(HFO_ITALIC, CG_normal_font, CG_normal_size, "I");
  canvas->define_font(HFO_BIG, CG_normal_font, CG_normal_big, "");

  canvas->define_font(HFO_FNORMAL, CG_fixed_font, CG_fixed_size, "");
  canvas->define_font(HFO_FBOLD, CG_fixed_font, CG_fixed_size, "B");
  canvas->define_font(HFO_FITALIC, CG_fixed_font, CG_fixed_size, "I");
  canvas->define_font(HFO_FBIG, CG_fixed_font, CG_fixed_big, "");

  canvas->define_font(HFO_SECHEA4, CG_normal_font, CG_sechea4, "B");
  canvas->define_font(HFO_SECHEA3, CG_normal_font, CG_sechea3, "B");
  canvas->define_font(HFO_SECHEA2, CG_normal_font, CG_sechea2, "B");
  canvas->define_font(HFO_SECHEA1, CG_normal_font, CG_sechea1, "B");

  canvas->define_color(HCO_BG, 0xffffff);
  canvas->define_color(HCO_FG, 0);
 
  canvas->define_color(HCO_LINK, CG_link);
  canvas->define_color(HCO_REDBG, CG_red);
  canvas->define_color(HCO_HIBG, CG_hilight);
  canvas->define_color(HCO_CODEBG, CG_code);
  canvas->define_color(HCO_SHELLBG, CG_shell);
  canvas->define_color(HCO_OUTPUTBG, CG_output);
//uu-cons
}


FYHelpWin::~FYHelpWin()
{
// uu-des
}

// uu-chapsec
void FYHelpWin::menu_add_chapter(const char *title, int linkno)
{
  FYMenuCommand *cmd;
  cmd= new FYMenuCommand(chapters_pane, title, NULL, 0, follow_link, linkno);
  cmd->create();
}


void FYHelpWin::set_section_menu(const sipair_t *me)
{
  int i;
  FYMenuCommand *cmd;
  if (sections_def==me) return ;
  FXWindow *ww;
  while((ww=sections_pane->getFirst()))
    delete ww;
  sections_def= me;
  for(i=0;me[i].Vs;i++)
  {
     cmd= new FYMenuCommand(sections_pane, me[i].Vs, 
                            NULL, 0, follow_link, me[i].Vi);
     cmd->create();
  }
  if (i==0) 
    sections_mtitle->disable();
  else
    sections_mtitle->enable();
  sections_pane->recalc();
}


typedef struct { const char *name; int secno; } anchor_t; 
static const anchor_t anchors[]= {
{ "fmt\056images", 9 },
{ "inline", 12 },
};
static void init_doc(FYHelpWin *win)
{
   win->menu_add_chapter("Demo\040Chapter", 0);
 win->menu_add_chapter("Fox\040Help\040Compiler", 1);
 win->menu_add_chapter("Document\040Syntax", 2);
 win->menu_add_chapter("C\057C\053\053\040Interface\040to\040the\040Help\040System", 13);
 win->menu_add_chapter("About", 17);

  win->canvas->link_func= follow_link;
}
static const sipair_t secmen_0[]= {
{NULL,0} };
static const sipair_t secmen_1[]= {
{NULL,0} };
static const sipair_t secmen_2[]= {
{"Chapters\040and\040Sections", 4},
{"Paragraphs", 5},
{"Lists", 6},
{"Verbatim\040Text", 7},
{"Tables", 8},
{"Images", 9},
{"Links", 10},
{"Inline\040Formatting", 12},
{NULL,0} };
static const sipair_t secmen_13[]= {
{NULL,0} };
static const sipair_t secmen_17[]= {
{"Copying", 18},
{"Things\040to\040Do", 19},
{NULL, 0} };
static void tf_0(FYGfxCanvas* canvas) {
  int *TL_ROW= canvas->get_row_lines(6);
  int *TL_COL= canvas->get_col_lines(7);
canvas->reset_doc();
  canvas->println(0,1);
  canvas->pbreak();
  canvas->println(2,7);
  canvas->verbatim_vspace();
  canvas->verbatim_line(8,10);
  canvas->verbatim_vspace();
  canvas->println(11,12);
  canvas->pbreak();
  canvas->println(13,19);
  canvas->bullet_println(1, 20, 25);
  canvas->bullet_println(1, 26, 31);
  canvas->bullet_println(2, 32, 37);
  canvas->bullet_println(1, 38, 41);
  canvas->println(42,52);
 /// BEGIN TABLE 
 {
  int cw0, ch0,ca0;
  canvas->get_seq_metrics(53,53, &cw0, &ch0, &ca0);
  int cw1, ch1,ca1;
  canvas->get_seq_metrics(54,54, &cw1, &ch1, &ca1);
  int cw2, ch2,ca2;
  canvas->get_seq_metrics(55,55, &cw2, &ch2, &ca2);
  int cw3, ch3,ca3;
  canvas->get_seq_metrics(56,56, &cw3, &ch3, &ca3);
  int cw4, ch4,ca4;
  canvas->get_seq_metrics(57,57, &cw4, &ch4, &ca4);
  int cw5, ch5,ca5;
  canvas->get_seq_metrics(58,58, &cw5, &ch5, &ca5);
  int cw6, ch6,ca6;
  canvas->get_seq_metrics(59,59, &cw6, &ch6, &ca6);
  int cw7, ch7,ca7;
  canvas->get_seq_metrics(60,60, &cw7, &ch7, &ca7);
  int cw8, ch8,ca8;
  canvas->get_seq_metrics(61,61, &cw8, &ch8, &ca8);
  int cw9, ch9,ca9;
  canvas->get_seq_metrics(62,62, &cw9, &ch9, &ca9);
  int cw10, ch10,ca10;
  canvas->get_seq_metrics(63,66, &cw10, &ch10, &ca10);
  int cw11, ch11,ca11;
  canvas->get_seq_metrics(67,67, &cw11, &ch11, &ca11);
  int cw12, ch12,ca12;
  canvas->get_seq_metrics(68,68, &cw12, &ch12, &ca12);
  int cw13, ch13,ca13;
  canvas->get_seq_metrics(69,69, &cw13, &ch13, &ca13);
  int cw14, ch14,ca14;
  canvas->get_seq_metrics(70,70, &cw14, &ch14, &ca14);
  int cw15, ch15,ca15;
  canvas->get_seq_metrics(71,71, &cw15, &ch15, &ca15);
 int rh0=0, ra0=0;
 int rh1=0, ra1=0;
 int rh2=0, ra2=0;
 int rh3=0, ra3=0;
 int rh4=0, ra4=0;
  rh0= max_i(rh0, ch0); ra0= max_i(ra0, ca0);
  rh0= max_i(rh0, ch1); ra0= max_i(ra0, ca1);
  rh0= max_i(rh0, ch2); ra0= max_i(ra0, ca2);
  rh1= max_i(rh1, ch3); ra1= max_i(ra1, ca3);
  rh1= max_i(rh1, ch4); ra1= max_i(ra1, ca4);
  rh1= max_i(rh1, ch5); ra1= max_i(ra1, ca5);
  rh1= max_i(rh1, ch6); ra1= max_i(ra1, ca6);
  rh2= max_i(rh2, ch7); ra2= max_i(ra2, ca7);
  rh2= max_i(rh2, ch8); ra2= max_i(ra2, ca8);
  rh3= max_i(rh3, ch9); ra3= max_i(ra3, ca9);
  rh3= max_i(rh3, ch10); ra3= max_i(ra3, ca10);
  rh3= max_i(rh3, ch11); ra3= max_i(ra3, ca11);
  rh4= max_i(rh4, ch12); ra4= max_i(ra4, ca12);
  rh4= max_i(rh4, ch13); ra4= max_i(ra4, ca13);
  rh4= max_i(rh4, ch14); ra4= max_i(ra4, ca14);
  rh4= max_i(rh4, ch15); ra4= max_i(ra4, ca15);
  int cow0=0;
  int cow1=0;
  int cow2=0;
  int cow3=0;
  cow0= max_i(cow0, cw0);
  cow0= max_i(cow0, cw3);
  cow0= max_i(cow0, cw9);
  cow0= max_i(cow0, cw12);
  cow1= max_i(cow1, cw1);
  cow1= max_i(cow1, cw4);
  cow1= max_i(cow1, cw13);
  cow2= max_i(cow2, cw2);
  cow2= max_i(cow2, cw5);
  cow2= max_i(cow2, cw14);
  cow3= max_i(cow3, cw6);
  cow3= max_i(cow3, cw8);
  cow3= max_i(cow3, cw11);
  cow3= max_i(cow3, cw15);
  int D=0;
  D= cw10  - cow1 - cow2;
  if (D>0) { 
    cow1 += D / 2; D -= D / 2;
    cow2 += D; }
  D= cw7  - cow0 - cow1 - cow2;
  if (D>0) { 
    cow0 += D / 3; D -= D / 3;
    cow1 += D / 2; D -= D / 2;
    cow2 += D; }
  int rp0;
  int rp1;
  int rp2;
  int rp3;
  int rp4;
  int rp5;
  int cp0;
  int cp1;
  int cp2;
  int cp3;
  int cp4;
  cp0= rp0= 0;
  rp1= rp0 + 2*CELL_PAD + rh0;
  rp2= rp1 + 2*CELL_PAD + rh1;
  rp3= rp2 + 2*CELL_PAD + rh2;
  rp4= rp3 + 2*CELL_PAD + rh3;
  rp5= rp4 + 2*CELL_PAD + rh4;
  cp1= cp0 + 2*CELL_PAD + cow0;
  cp2= cp1 + 2*CELL_PAD + cow1;
  cp3= cp2 + 2*CELL_PAD + cow2;
  cp4= cp3 + 2*CELL_PAD + cow3;
  int off_x= canvas->get_width() - cp4;
  if (off_x<0) off_x= 0; else off_x/= 2;
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 53, 53);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 54, 54);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 55, 55);
  canvas->cell_right(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 56, 56);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 57, 57);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 58, 58);
  canvas->cell_left(off_x+cp3+CELL_PAD, off_x+cp4-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 59, 59);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp3-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 60, 60);
  canvas->cell_left(off_x+cp3+CELL_PAD, off_x+cp4-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 61, 61);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 62, 62);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp3-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 63, 66);
  canvas->cell_left(off_x+cp3+CELL_PAD, off_x+cp4-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 67, 67);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 68, 68);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 69, 69);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 70, 70);
  canvas->cell_left(off_x+cp3+CELL_PAD, off_x+cp4-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 71, 71);
  int off_y= canvas->get_cursor_y();
  TL_ROW[0]= off_y+rp0; TL_ROW[1]= off_x+cp0; TL_ROW[2]= off_x+cp4;
  TL_ROW[3]= off_y+rp1; TL_ROW[4]= off_x+cp0; TL_ROW[5]= off_x+cp4;
  TL_ROW[6]= off_y+rp2; TL_ROW[7]= off_x+cp0; TL_ROW[8]= off_x+cp4;
  TL_ROW[9]= off_y+rp3; TL_ROW[10]= off_x+cp0; TL_ROW[11]= off_x+cp4;
  TL_ROW[12]= off_y+rp4; TL_ROW[13]= off_x+cp0; TL_ROW[14]= off_x+cp4;
  TL_ROW[15]= off_y+rp5; TL_ROW[16]= off_x+cp0; TL_ROW[17]= off_x+cp4;
  TL_COL[0]= off_y+rp0; TL_COL[1]= off_y+rp2; TL_COL[2]= off_x+cp1;
  TL_COL[3]= off_y+rp0; TL_COL[4]= off_y+rp2; TL_COL[5]= off_x+cp2;
  TL_COL[6]= off_y+rp0; TL_COL[7]= off_y+rp5; TL_COL[8]= off_x+cp0;
  TL_COL[9]= off_y+rp0; TL_COL[10]= off_y+rp5; TL_COL[11]= off_x+cp3;
  TL_COL[12]= off_y+rp0; TL_COL[13]= off_y+rp5; TL_COL[14]= off_x+cp4;
  TL_COL[15]= off_y+rp3; TL_COL[16]= off_y+rp5; TL_COL[17]= off_x+cp1;
  TL_COL[18]= off_y+rp4; TL_COL[19]= off_y+rp5; TL_COL[20]= off_x+cp2;
  canvas->xcoord(cp4+3);
  canvas->advance_y(rp5+2);
 /// END TABLE 
 }
  canvas->println(72,75);
  canvas->image(0);
  canvas->println(76,82);
  canvas->pbreak();
  canvas->println(83,85);
  canvas->verbatim_vspace();
  canvas->verbatim_line(86,86);
  canvas->verbatim_line(87,87);
  canvas->verbatim_vspace();
  canvas->pbreak();
  canvas->verbatim_vspace();
  canvas->verbatim_line(88,88);
  canvas->verbatim_line(89,89);
  canvas->verbatim_vspace();
  canvas->pbreak();
  canvas->verbatim_vspace();
  canvas->verbatim_line(90,90);
  canvas->verbatim_line(91,91);
  canvas->verbatim_vspace();

 canvas->end_doc(tf_0);
}
static void cf_0(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xb,0x1,0xb,0x2,0x0,0x3,0x0,0x4,0x0,0x5,0x0,0x6,0x0,0x7,0x0,0x8,0x50004,0x9,0x30004,0xa,0x50004,0xb,0x0,0xc,0x0,0xd,0x0,0xe,0x0,0xf,0x0,0x10,0x0,0x11,0x0,0x12,0x0,0x13,0x0,0x14,0x0,0x0,0x0,0x15,0x0,0x16,0x0,0x13,0x0,0x17,0x0,0x18,0x0,0x0,0x0,0x19,0x0,0x1a,0x0,0x12,0x0,0x1b,0x0,0x1c,0x0,0x0,0x0,0x1d,0x0,0x5,0x0,0x1b,0x0,0x13,0x0,0x1e,0x0,0x0,0x0,0x1f,0x0,0x20,0x0,0x18,0x0,0x21,0x0,0x1a,0x0,0x22,0x0,0x13,0x0,0x23,0x0,0x24,0x30000,0x25,0x0,0x26,0x1000000,0x27,0x0,0x28,0x0,0x29,0x0,0x2a,0x0,0x2b,0x0,0x2c,0x0,0x2d,0x0,0x2e,0x0,0x2f,0x0,0x30,0x0,0x31,0x0,0x32,0x0,0x33,0x30000,0x34,0x30000,0x35,0x30000,0x36,0x30000,0x37,0x0,0x38,0x0,0x39,0x0,0x3a,0x0,0x3b,0x0,0x3c,0x0,0x3d,0x0,0x16,0x0,0x3e,0x0,0x3f,0x0,0x21,0x0,0x40,0x0,0x41,0x0,0x42,0x0,0x1b,0x0,0x1a,0x0,0x43,0x0,0x44,0x0,0x45,0x0,0x46,0x50004,0x47,0x50004,0x48,0x70004,0x49,0x70004,0x48,0x60004,0x4a,0x60004,0x48,};
  canvas->set_elts(seq,92);
static const unsigned int bgrects[]= {
0x9,0x9,0x3,0x30,0x30,0x3,0x3f,0x42,0x3,};
  canvas->set_bgrects(bgrects, 3);
static const unsigned int underlines[]= {
0x32,0x32,0x0,};
  canvas->set_underlines(underlines, 1);
static const unsigned int verboxes[]= {
0x8,0xa,0x5,0x56,0x57,0x5,0x58,0x59,0x7,0x5a,0x5b,0x6,};
  canvas->set_vboxes(verboxes, 4);
static const unsigned char img0data[]={

137,80,78,71,13,10,26,10,0,0,0,13,
73,72,68,82,0,0,1,138,0,0,1,5,
8,6,0,0,0,253,152,160,29,0,0,0,
6,98,75,71,68,0,255,0,255,0,255,160,
189,167,147,0,0,0,9,112,72,89,115,0,
0,3,211,0,0,3,211,1,231,116,74,61,
0,0,0,7,116,73,77,69,7,222,7,6,
17,40,4,52,127,247,199,0,0,32,0,73,
68,65,84,120,218,236,189,215,119,92,215,157,
239,249,57,251,132,58,149,51,170,10,25,32,
8,230,40,82,209,146,108,75,178,220,14,237,
52,178,187,237,118,247,117,247,117,207,234,213,
179,238,125,156,63,97,222,102,205,154,181,122,
110,247,189,221,237,110,231,32,203,114,148,100,
73,150,68,5,42,48,139,9,36,65,228,88,
40,84,142,39,205,67,81,0,33,130,36,0,
18,20,105,243,247,34,17,117,194,62,123,255,
246,47,124,127,97,75,169,84,202,225,46,221,
165,187,116,151,238,210,93,186,10,73,62,159,
207,41,22,139,43,190,225,216,201,51,92,28,
157,96,239,246,126,14,28,60,66,173,253,193,
107,94,95,153,159,194,19,73,45,251,219,116,
190,206,104,166,178,56,152,114,26,199,27,95,
114,141,34,75,236,237,12,129,180,244,222,242,
196,89,218,212,60,251,247,238,230,153,151,222,
102,119,127,39,163,147,51,60,176,103,43,155,
251,251,248,191,254,191,239,17,217,241,56,245,
233,115,184,234,115,124,245,115,143,19,14,5,
23,238,207,102,179,124,239,213,115,232,45,61,
43,30,239,106,190,237,195,84,203,167,209,131,
241,155,254,220,117,187,54,51,73,113,224,53,
58,55,110,163,98,169,124,230,222,62,186,58,
218,150,92,211,104,52,248,151,95,29,196,114,
133,174,249,92,203,113,56,55,93,162,88,174,
65,163,2,122,224,250,3,168,206,131,59,178,
114,78,190,206,245,18,176,183,39,132,44,73,
43,158,135,213,172,217,205,90,223,82,205,100,
96,186,132,101,59,120,138,105,132,227,80,10,
180,220,148,57,88,92,184,114,243,191,154,247,
230,61,179,86,0,205,3,66,185,254,181,102,
29,108,115,69,239,151,202,105,240,198,217,218,
22,192,235,146,111,26,127,207,29,254,53,177,
123,62,191,240,111,247,220,41,158,122,242,1,
106,181,26,161,80,104,217,123,190,255,204,115,
228,77,133,71,182,119,178,99,235,166,149,173,
117,165,194,63,127,239,105,122,186,58,217,209,
215,193,239,223,57,133,210,125,63,182,101,210,
40,101,175,224,25,163,48,203,99,27,116,54,
247,247,221,62,74,66,146,16,171,189,105,247,
246,45,244,117,181,241,219,35,19,215,85,18,
215,35,77,145,174,207,83,150,67,169,110,94,
57,120,179,204,151,62,247,105,84,69,224,118,
170,196,66,126,226,177,8,179,153,28,150,101,
17,82,13,26,115,195,232,212,249,212,199,246,
225,214,93,24,134,177,112,255,155,135,79,225,
138,117,222,53,21,174,206,29,36,30,252,75,
74,158,46,246,245,132,72,182,196,174,184,164,
88,44,130,230,191,238,163,100,73,162,63,233,
195,175,43,31,217,231,56,64,205,176,110,251,
105,247,233,10,155,82,126,100,33,225,171,149,
8,84,242,127,242,172,24,245,105,215,85,18,
171,158,231,13,247,45,229,101,79,27,255,247,
63,255,39,231,135,198,174,122,207,150,158,20,
33,185,206,75,239,158,97,108,124,242,234,188,
230,56,76,76,76,144,201,100,240,120,60,236,
216,212,203,197,121,147,231,6,45,228,174,123,
175,179,89,52,26,151,201,169,219,133,214,180,
115,103,231,114,136,96,219,13,191,92,147,87,
166,167,106,166,141,239,67,127,179,213,0,135,
143,28,163,175,111,3,223,252,218,151,40,149,
43,204,204,231,41,213,77,100,89,230,31,190,
253,183,252,207,167,95,134,214,221,60,243,230,
49,92,158,9,220,245,89,62,255,216,199,248,
221,129,67,212,220,73,132,91,230,46,93,135,
65,220,62,222,185,56,65,177,124,148,79,62,
124,255,146,223,198,38,166,144,116,63,212,43,
148,71,79,160,185,220,168,137,141,87,85,22,
27,19,62,206,55,202,20,63,162,111,169,27,
54,94,237,246,159,115,159,75,102,115,202,207,
68,193,75,221,180,254,164,249,79,146,160,61,
226,89,243,253,182,101,98,20,102,113,133,91,
175,124,240,229,252,233,14,208,219,213,206,224,
197,139,140,79,205,226,243,121,217,210,219,65,
123,251,162,156,187,103,207,110,182,108,234,231,
165,55,15,243,238,169,65,58,218,91,151,247,
86,50,25,190,243,253,31,83,51,29,186,54,
110,199,244,165,240,119,246,129,180,2,121,87,
78,211,213,190,227,182,91,7,177,150,155,82,
137,24,141,204,200,13,191,92,85,86,246,122,
219,190,50,140,34,187,253,252,230,165,3,56,
142,141,223,239,39,149,76,176,161,171,149,169,
177,225,133,107,246,111,74,97,103,46,34,7,
18,104,45,125,152,109,247,241,163,215,206,210,
72,238,69,4,91,239,106,129,149,108,84,33,
163,165,182,146,41,213,104,52,26,75,126,123,
235,240,251,168,30,63,78,97,154,71,55,199,
104,204,158,167,58,114,248,234,204,38,36,122,
98,94,2,238,143,198,179,176,236,59,39,28,
231,117,201,244,7,5,94,179,246,39,205,127,
49,191,107,69,200,195,178,8,219,244,121,250,
165,81,30,237,118,81,31,126,23,179,156,189,
134,66,18,204,7,183,145,214,186,152,214,251,
120,237,141,131,60,255,226,203,148,203,229,165,
198,173,166,81,46,149,72,151,76,42,149,202,
178,207,10,5,131,36,19,9,60,186,134,207,
163,97,90,214,245,149,132,227,96,25,117,146,
110,131,96,48,112,219,173,195,154,20,69,46,
95,64,120,194,84,6,223,166,81,94,187,107,
44,173,112,253,203,245,43,173,42,179,48,75,
176,173,143,51,103,207,45,252,109,67,119,23,
159,121,226,19,11,255,222,179,99,43,155,162,
50,86,118,28,219,52,144,132,140,59,213,127,
87,250,175,129,210,196,248,215,31,255,154,83,
103,207,47,252,109,215,182,77,216,141,26,45,
110,147,189,123,118,243,201,135,246,241,208,214,
118,204,82,230,154,202,162,63,233,39,248,17,
40,11,231,14,75,219,80,60,30,146,169,208,
138,247,201,31,27,185,53,65,220,183,118,23,
208,22,50,178,144,104,77,181,96,229,167,216,
19,174,96,228,166,174,62,223,46,15,178,230,
34,213,24,228,191,255,253,55,249,175,223,250,
107,188,222,102,28,101,98,114,138,11,67,195,
252,207,31,253,146,137,162,77,41,51,201,203,
175,189,9,112,133,194,56,246,254,25,106,22,
212,44,153,122,189,1,102,227,234,99,44,207,
225,205,158,161,219,26,98,183,119,142,199,31,
188,231,246,228,197,181,220,116,102,34,135,28,
218,128,212,210,135,230,13,174,249,229,134,101,
175,76,72,21,235,184,53,153,100,208,181,232,
81,20,199,113,180,32,187,118,108,107,194,32,
227,19,188,253,222,17,158,124,236,81,28,199,
65,186,180,187,62,241,208,126,60,218,33,14,
101,242,8,127,236,174,196,95,43,163,4,90,
32,208,194,129,161,25,206,14,189,66,123,75,
152,98,169,140,89,85,232,109,79,54,21,199,
142,237,56,142,195,169,103,158,167,238,13,95,
213,138,18,18,108,76,250,57,55,93,162,80,
189,53,120,172,95,87,136,249,93,119,214,156,
155,6,1,201,166,39,230,225,98,186,242,167,
227,197,210,140,75,180,69,220,88,133,181,127,
183,222,210,203,241,161,119,56,122,242,12,255,
248,237,255,130,162,40,92,252,197,115,84,166,
231,192,157,196,106,212,96,234,24,118,100,3,
170,191,25,84,118,108,139,222,174,54,226,177,
165,178,226,103,207,254,134,68,71,47,197,170,
129,215,45,99,104,30,6,198,231,56,255,79,
255,11,191,223,199,183,255,250,47,23,174,77,
196,35,60,245,249,39,57,248,238,97,210,117,
5,87,188,123,121,15,215,168,67,110,140,111,
124,253,11,183,253,154,172,218,163,248,238,51,
47,80,146,2,212,242,115,104,55,40,120,77,
115,229,38,222,88,166,194,92,177,190,8,91,
5,147,4,189,26,63,248,213,75,140,142,141,
113,236,204,32,147,243,101,126,246,155,151,56,
121,250,44,0,133,66,145,127,127,250,57,222,
58,54,128,191,54,137,81,41,220,149,248,55,
72,114,32,65,38,176,141,35,165,24,35,202,
6,116,171,68,71,123,234,50,47,81,226,75,
79,60,68,109,118,232,218,140,39,65,127,210,
71,208,163,174,105,28,170,44,104,13,233,244,
39,125,244,37,188,116,134,221,180,133,117,2,
186,130,144,36,132,4,178,16,248,116,133,238,
152,151,254,148,15,113,135,89,230,150,44,99,
203,77,5,23,245,173,125,158,122,226,94,34,
107,188,255,86,83,64,87,216,214,22,160,183,
197,139,107,5,208,180,99,153,56,51,167,17,
227,239,226,44,227,50,42,173,59,176,125,173,
76,78,77,33,73,18,223,248,242,103,248,204,
3,219,176,243,19,24,185,73,190,252,248,3,
236,111,177,48,242,211,205,235,253,49,14,31,
61,177,228,89,182,109,19,12,71,153,156,201,
240,133,143,223,195,63,254,213,23,248,199,175,
127,150,253,219,250,168,215,106,124,243,171,95,
90,242,206,206,142,118,18,45,113,190,248,185,
79,243,216,222,62,26,249,244,82,47,194,108,
32,165,207,178,51,92,231,91,95,249,244,157,
97,180,172,246,134,10,46,84,127,20,33,223,
56,116,208,176,86,174,40,28,96,40,93,193,
235,82,112,107,50,166,191,157,252,192,171,120,
83,155,112,185,116,114,229,58,90,247,125,20,
39,79,241,238,161,35,244,247,245,242,194,27,
135,168,135,251,217,215,54,143,170,200,28,202,
222,45,25,89,30,146,89,244,192,46,255,255,
229,126,95,80,24,74,19,18,104,113,219,180,
183,46,141,247,140,140,79,33,175,32,165,83,
72,176,49,225,227,220,84,137,66,109,101,158,
133,215,165,144,10,185,8,123,212,37,99,170,
153,42,122,208,13,225,63,34,203,218,113,144,
236,166,215,221,26,114,147,41,173,206,251,242,
235,10,253,73,31,178,144,144,112,152,47,25,
183,245,247,118,68,220,164,66,250,202,161,165,
249,97,122,252,6,15,125,106,31,178,44,243,
175,207,188,12,209,165,188,232,204,95,100,67,
210,135,121,89,124,202,52,13,108,217,133,176,
12,188,30,55,247,238,221,201,228,11,7,152,
46,10,132,191,133,188,43,201,191,252,251,247,
248,214,95,125,141,119,14,29,197,227,214,73,
207,231,137,38,59,216,220,223,76,214,208,52,
141,71,31,220,79,34,26,68,215,175,28,115,
165,82,225,233,103,127,67,89,248,80,82,59,
151,252,230,201,156,226,175,190,244,4,66,136,
59,134,23,87,53,210,114,185,140,33,221,60,
247,189,97,218,171,19,104,192,76,161,190,96,
185,134,182,127,10,57,152,228,39,7,206,82,
14,111,65,205,14,240,87,159,121,144,191,251,
155,111,224,114,185,48,81,144,117,63,135,47,
206,241,238,169,65,34,213,161,5,203,225,46,
177,196,11,88,238,255,175,245,55,0,163,156,
163,191,51,177,160,76,10,197,34,111,188,115,
152,183,207,140,161,186,253,43,99,64,9,54,
38,189,120,181,107,27,30,46,69,208,215,226,
99,107,155,159,136,87,187,234,152,254,168,214,
197,113,46,113,253,202,19,63,62,160,176,87,
99,243,165,52,91,22,158,114,251,82,212,167,
174,74,73,88,70,29,59,59,198,224,116,158,
239,254,242,101,94,124,235,40,166,178,76,118,
84,160,131,90,185,196,129,247,222,231,153,223,
191,206,185,11,23,73,37,19,120,236,10,90,
75,31,47,190,241,30,0,79,62,178,31,10,
19,77,37,16,104,225,190,123,247,161,105,26,
245,122,157,19,167,78,227,78,245,19,244,106,
140,141,79,80,40,20,56,114,252,125,0,54,
111,90,62,230,169,235,58,79,125,233,243,216,
146,130,80,22,227,44,182,101,176,161,61,126,
71,41,137,85,123,20,233,185,12,142,30,188,
105,47,175,26,230,170,239,201,148,26,116,70,
220,205,193,123,155,133,49,117,217,133,60,254,
14,127,245,149,207,46,104,247,90,173,70,161,
222,196,59,29,163,202,183,191,246,57,76,211,
228,187,47,30,135,96,242,174,118,184,25,84,
152,192,235,237,231,95,126,248,44,134,26,0,
73,65,105,233,67,36,59,112,108,139,226,240,
81,36,205,139,191,125,203,181,225,44,33,209,
159,242,113,102,178,120,69,173,131,34,36,218,
194,58,45,1,215,159,132,114,88,98,49,95,
230,181,219,171,136,196,199,252,46,122,99,158,
37,69,170,215,154,59,33,129,199,165,160,8,
137,92,229,214,123,29,110,77,208,29,247,174,
248,122,171,81,69,168,46,164,13,15,99,214,
171,216,133,41,198,50,101,44,225,190,18,122,
243,6,201,123,119,147,59,253,26,70,235,189,
188,244,254,121,254,50,26,161,37,232,102,78,
150,25,154,41,242,210,43,175,243,201,71,31,
66,150,101,28,64,241,134,121,111,120,152,243,
99,175,50,52,54,133,105,24,132,91,83,76,
78,156,224,135,207,156,33,18,111,69,182,42,
236,217,185,253,170,243,154,207,231,249,231,239,
252,128,224,238,207,45,249,187,89,201,227,139,
232,119,28,47,174,74,81,204,231,139,88,213,
18,178,170,95,194,218,12,132,178,118,236,179,
214,176,87,125,143,101,59,204,151,27,31,170,
171,112,232,73,197,248,245,31,14,50,58,120,
142,167,190,240,25,10,165,10,118,176,19,25,
176,3,237,28,63,117,150,251,247,237,225,222,
222,48,199,7,143,99,39,119,221,21,244,55,
106,241,122,98,252,250,232,12,134,191,231,138,
138,88,73,200,4,122,247,173,248,89,170,44,
177,173,205,207,84,174,78,182,92,71,88,50,
193,176,78,42,168,47,88,197,127,106,36,27,
13,100,203,164,190,10,69,225,117,41,244,196,
220,87,116,50,208,150,241,72,92,138,32,25,
116,17,11,184,144,37,9,199,129,67,67,217,
91,235,125,72,208,19,247,34,175,194,8,80,
210,167,177,20,31,178,42,179,43,161,211,218,
147,228,173,227,3,52,164,171,11,224,224,230,
143,97,141,188,135,163,7,121,245,157,99,56,
118,211,72,213,66,41,58,90,3,156,56,61,
0,234,101,138,38,210,77,14,8,7,182,82,
43,204,161,184,60,88,209,13,120,19,91,145,
114,231,248,155,175,254,111,72,146,196,59,239,
190,199,125,247,238,95,242,174,159,62,251,59,
66,193,0,193,150,118,20,109,169,242,210,252,
49,222,26,188,136,205,25,146,209,16,169,84,
234,142,224,197,85,249,63,199,79,157,163,67,
204,34,231,154,129,202,249,83,175,172,249,197,
14,80,93,99,165,236,108,97,105,186,153,203,
23,230,216,180,193,233,83,167,248,218,151,62,
71,103,103,7,83,153,2,178,171,185,72,114,
37,205,206,173,155,200,229,114,180,182,68,169,
54,238,198,42,110,138,149,17,76,160,70,218,
47,91,84,135,185,145,55,200,142,31,90,155,
96,20,18,237,17,157,29,29,65,54,180,120,
105,15,187,255,100,149,4,128,169,185,104,184,
46,25,101,214,74,54,179,196,134,22,207,178,
86,238,229,129,97,183,38,211,21,243,176,171,
51,64,34,168,47,8,233,66,205,188,229,16,
85,34,160,227,115,173,46,222,105,250,59,112,
172,58,93,65,193,67,247,238,101,108,106,150,
76,161,138,36,150,22,208,150,198,79,99,91,
230,130,225,34,233,1,36,33,51,69,130,145,
90,211,212,84,194,109,252,238,248,12,111,158,
184,64,80,49,208,210,167,176,43,185,203,93,
177,230,117,179,239,243,165,93,17,34,149,65,
76,213,135,101,89,28,59,117,150,116,238,202,
242,209,134,236,97,88,237,67,233,121,112,217,
26,0,57,214,203,107,167,166,249,206,143,158,
225,212,217,115,100,50,25,114,249,2,191,121,
233,0,213,106,245,206,247,40,246,237,220,196,
192,133,33,74,233,113,36,91,187,161,22,24,
117,211,90,115,94,123,169,110,82,115,217,92,
110,63,36,180,26,127,254,247,223,92,128,157,
6,103,75,136,75,8,147,163,122,249,143,231,
222,165,54,125,158,96,162,3,87,231,222,187,
82,254,38,83,163,146,101,122,252,151,56,237,
23,193,18,148,207,142,208,190,249,43,119,39,
230,6,72,173,215,144,45,3,99,133,30,69,
91,68,71,87,151,239,54,224,82,4,125,137,
166,229,238,145,212,69,115,254,50,202,92,150,
85,120,43,200,165,8,58,86,8,195,216,150,
137,52,125,2,199,223,134,28,72,64,160,133,
153,217,163,88,150,69,123,170,133,179,147,69,
62,156,72,235,87,44,148,204,105,140,150,102,
48,89,36,54,45,88,198,202,165,36,1,33,
43,184,147,27,241,205,29,227,241,251,119,34,
132,224,135,175,158,70,120,66,52,242,211,40,
133,113,8,116,209,144,220,248,252,126,62,253,
241,7,248,209,31,78,240,253,223,188,70,126,
126,142,175,125,246,19,228,114,185,37,253,161,
220,154,76,229,58,5,118,122,106,19,166,35,
113,240,212,48,229,250,16,154,34,33,82,59,
56,240,246,97,158,252,196,199,238,108,69,177,
99,219,86,92,154,198,104,69,71,13,196,113,
249,35,107,126,113,117,13,176,211,229,148,43,
27,132,46,203,206,173,9,15,154,166,225,118,
187,57,116,236,125,236,112,239,2,83,72,158,
48,74,56,129,226,142,97,5,91,238,74,160,
117,160,185,233,87,160,119,240,146,232,177,48,
19,239,83,152,217,78,32,177,233,238,228,172,
85,249,234,30,164,75,174,196,245,18,4,125,
46,153,240,117,82,141,35,151,250,151,24,213,
43,21,130,229,56,100,203,183,46,62,33,75,
18,125,73,47,226,122,144,83,45,79,184,84,
2,219,162,179,55,193,153,11,23,168,251,98,
72,66,166,20,220,204,191,62,243,18,117,111,
27,174,228,54,200,206,44,53,70,149,0,237,
222,50,195,181,18,178,238,187,186,39,38,203,
20,163,59,248,193,27,67,56,245,18,122,106,
115,115,140,102,13,143,83,166,110,91,72,241,
141,252,228,217,223,242,247,127,253,23,60,185,
167,147,35,103,135,232,232,239,166,53,149,228,
7,63,253,5,150,35,209,209,154,4,73,98,
174,38,144,124,215,159,3,45,16,199,240,4,
168,158,126,13,211,169,178,49,236,230,161,7,
110,79,35,118,213,57,174,253,27,251,248,100,
221,96,104,120,152,137,249,50,90,164,99,77,
47,174,53,110,172,135,77,182,98,208,117,153,
77,100,183,108,229,63,127,247,54,15,109,105,
101,54,91,68,113,37,150,222,32,9,244,63,
2,37,49,117,254,55,88,78,149,88,235,163,
232,190,91,251,61,182,81,39,59,249,62,209,
174,165,177,7,171,81,163,174,94,224,114,91,
214,153,143,224,235,234,185,43,237,111,128,244,
74,17,217,52,40,196,219,80,174,3,193,197,
252,58,56,107,247,8,210,197,6,214,173,42,
93,151,160,59,230,185,110,166,155,99,91,116,
249,12,190,244,228,35,0,28,61,126,146,98,
160,15,215,37,136,73,184,220,208,186,135,171,
249,36,106,164,157,161,153,179,16,184,118,210,
140,227,216,88,181,10,194,27,65,141,46,202,
51,57,218,77,201,29,65,50,234,152,249,41,
2,254,16,146,36,209,215,219,77,95,111,247,
162,209,91,171,227,245,184,25,154,76,83,113,
181,160,197,55,174,64,128,13,209,161,27,120,
165,2,206,182,62,106,181,42,173,81,63,62,
159,247,182,228,197,85,41,138,74,165,194,187,
71,142,177,169,175,151,221,219,183,240,157,95,
190,198,90,109,144,234,13,42,10,195,182,201,
87,12,66,151,172,40,73,18,72,201,109,12,
142,157,65,250,80,141,135,227,252,113,196,36,
26,213,60,85,207,73,148,68,153,236,176,70,
170,239,214,86,116,206,156,127,13,111,252,74,
184,177,148,27,68,244,44,197,106,237,134,69,
53,63,133,55,218,117,87,226,175,145,42,190,
32,226,18,68,162,202,215,86,20,110,77,192,
26,245,132,227,56,76,231,110,13,54,46,209,
172,151,8,172,160,33,167,152,57,197,231,190,
240,104,147,247,27,13,58,219,83,216,23,6,
176,140,18,182,101,162,126,216,91,253,208,62,
151,132,140,156,220,186,36,102,83,47,101,209,
60,126,172,198,226,247,58,99,239,241,233,253,
91,41,148,203,156,30,190,64,209,183,97,193,
3,81,60,1,164,137,163,60,182,119,35,253,
125,187,151,29,231,183,190,241,20,178,44,99,
24,6,223,127,230,121,106,99,243,200,29,215,
110,197,97,59,130,123,183,247,211,214,118,103,
244,156,91,149,162,40,149,74,188,115,244,52,
71,166,33,32,157,164,33,123,89,107,217,221,
141,66,79,208,44,192,235,79,250,150,180,32,
158,178,66,48,63,142,124,153,60,251,163,72,
171,116,108,166,71,126,137,210,215,108,82,102,
88,229,91,62,132,212,214,79,45,207,244,118,
117,161,91,71,64,149,105,117,187,56,237,22,
184,131,41,50,99,239,98,153,85,132,208,16,
194,133,89,43,130,98,17,239,122,244,138,224,
227,93,90,74,222,98,22,217,48,200,71,147,
168,178,64,72,112,181,190,134,226,6,120,124,
174,212,160,97,222,10,99,74,162,39,238,38,
236,114,22,130,204,87,35,43,59,198,159,221,
183,25,85,85,121,238,229,3,148,75,37,90,
91,83,168,129,56,86,163,2,206,82,197,230,
100,46,210,106,103,200,20,85,228,203,218,113,
248,102,15,161,185,125,164,149,20,149,194,60,
78,110,18,127,208,141,89,152,199,152,59,133,
212,121,31,82,116,35,233,92,153,7,246,237,
100,239,174,237,252,211,119,127,142,236,139,97,
10,21,37,222,71,204,171,113,122,224,28,110,
151,70,48,24,36,16,240,51,58,54,78,162,
37,206,197,161,97,218,218,90,9,248,253,168,
170,202,183,190,246,121,126,245,251,87,153,90,
166,72,117,201,122,69,186,120,230,249,87,248,
63,254,246,235,119,132,124,90,149,156,111,105,
105,161,183,183,151,89,87,146,244,185,55,137,
238,122,114,141,22,12,84,110,194,217,0,134,
101,115,102,178,64,111,139,119,1,127,149,131,
41,8,166,238,104,1,97,91,6,165,244,69,
52,119,16,213,27,161,154,159,102,110,242,15,
56,61,231,23,160,182,27,73,75,190,17,143,
38,51,242,22,209,142,7,209,188,1,102,135,
94,197,118,106,52,60,231,145,37,208,132,196,
189,209,32,121,195,68,54,34,76,156,123,6,
167,231,20,97,175,74,213,176,113,76,168,89,
22,142,12,99,3,227,116,110,249,235,187,218,
224,90,134,89,48,186,80,153,45,73,144,10,
185,153,200,46,111,249,171,178,196,154,76,47,
7,166,242,235,31,196,22,18,244,37,188,132,
60,106,179,199,210,181,246,117,122,144,46,191,
69,119,71,51,163,110,124,174,72,200,239,225,
196,200,60,1,49,69,107,60,194,96,166,190,
36,67,43,238,106,240,229,39,63,203,179,207,
253,129,89,226,11,40,131,99,89,124,241,137,
143,241,220,43,111,50,231,84,200,212,243,244,
166,218,120,228,169,207,2,240,63,126,254,42,
114,114,43,39,38,7,217,86,44,18,240,251,
249,203,207,125,130,72,36,194,79,159,249,21,
83,19,117,234,181,28,13,203,97,240,87,207,
83,175,150,248,250,83,95,230,103,47,188,137,
207,173,97,148,179,152,106,128,71,247,239,224,
158,157,91,48,77,147,137,146,64,246,94,91,
248,55,10,115,108,233,233,184,170,146,176,44,
11,33,196,109,163,68,86,237,16,60,249,240,
126,126,240,252,219,232,170,160,94,204,224,242,
71,87,239,82,55,172,155,6,7,217,14,92,
152,169,176,191,87,227,143,37,145,114,102,248,
119,136,238,105,156,89,15,82,221,15,145,52,
114,191,177,248,125,14,200,150,255,150,141,167,
81,206,34,43,46,166,206,60,143,112,129,234,
9,0,18,37,227,40,74,111,118,33,54,177,
41,224,69,151,5,233,186,67,195,53,76,172,
207,102,119,52,132,95,93,100,179,186,101,243,
110,166,64,38,56,74,163,156,67,243,134,110,
189,115,118,135,64,145,190,124,166,233,81,132,
155,177,168,182,176,78,200,173,146,175,25,228,
43,6,165,122,115,31,105,178,64,85,196,154,
144,167,92,213,184,225,120,225,245,72,22,43,
63,184,170,145,157,160,83,205,177,111,251,61,
204,166,231,120,253,141,55,49,212,40,51,117,
129,84,207,240,213,167,158,108,182,250,126,241,
0,83,211,39,33,185,189,185,103,234,26,111,
189,123,132,246,84,156,169,76,9,179,81,71,
243,133,40,68,119,241,219,23,94,230,254,253,
123,248,206,211,47,112,207,182,205,28,60,113,
142,125,123,119,225,241,44,86,114,75,209,13,
124,255,183,175,227,114,185,168,153,18,155,83,
62,62,247,233,199,57,114,252,36,7,143,141,
18,220,249,103,88,102,3,189,94,229,149,67,
167,209,91,122,145,227,61,72,150,137,46,4,
39,6,142,144,140,5,121,229,221,147,72,209,
173,215,253,78,45,16,227,244,116,6,237,224,
33,30,185,255,30,36,73,194,182,109,94,127,
231,8,182,99,115,126,108,14,67,82,216,218,
26,226,225,251,247,162,40,202,71,202,139,202,
114,155,232,154,46,147,44,144,173,26,106,251,
94,100,109,109,7,138,148,235,230,77,253,136,
68,224,143,67,73,152,181,18,179,163,175,160,
108,154,66,242,216,16,46,194,135,143,249,49,
36,156,145,110,34,29,251,111,137,103,51,51,
252,2,21,235,52,138,20,192,142,229,49,171,
18,229,244,8,170,219,79,52,244,105,230,39,
158,71,110,203,226,85,100,146,186,134,84,40,
209,30,240,17,216,45,19,80,149,43,214,197,
37,11,250,252,110,178,190,34,181,242,236,186,
43,10,199,177,73,143,28,160,94,155,161,99,
211,83,216,166,193,252,216,123,120,163,183,127,
199,206,82,48,6,206,82,63,193,171,203,120,
117,153,214,144,142,237,56,84,27,214,117,131,
194,215,162,217,117,78,137,21,18,108,73,249,
241,172,240,132,58,201,29,98,116,240,24,195,
153,55,105,143,7,25,156,156,199,219,22,195,
44,206,227,209,20,84,181,233,73,127,254,137,
71,152,73,207,241,244,139,239,160,118,237,199,
145,221,72,194,102,223,238,29,148,223,122,151,
84,87,148,3,39,78,35,37,118,48,154,245,
240,184,215,203,159,61,186,159,96,192,199,19,
143,62,120,9,50,181,177,28,22,12,29,165,
99,31,22,160,2,231,166,79,211,159,153,231,
119,47,189,70,203,182,102,48,93,86,52,100,
69,195,240,238,197,181,32,15,155,115,223,72,
238,225,103,7,135,17,238,54,116,101,101,173,
209,37,95,11,167,171,94,6,126,241,42,81,
183,68,166,234,96,69,250,16,178,130,212,222,
133,138,196,217,90,153,19,63,125,153,128,71,
99,99,194,203,246,45,155,8,133,130,183,156,
23,175,224,176,235,185,58,178,16,32,171,200,
186,23,89,91,91,41,250,114,231,75,220,8,
181,134,239,188,146,248,43,220,237,90,129,209,
179,223,69,237,79,35,105,75,133,131,61,30,
199,101,245,96,139,50,170,20,163,101,211,99,
172,231,33,5,86,163,70,102,234,45,204,236,
4,114,71,26,85,0,20,145,105,110,170,116,
246,59,144,9,144,8,60,133,53,22,6,173,
192,222,173,126,244,124,9,145,158,199,9,248,
8,126,224,69,52,12,208,150,194,100,113,93,
67,82,193,42,174,127,235,236,217,161,63,80,
79,30,192,145,28,38,6,116,84,53,134,209,
122,140,201,11,224,15,238,192,23,233,188,34,
249,225,246,241,40,230,154,30,69,40,126,21,
33,44,225,117,173,125,236,134,101,147,43,155,
235,250,13,137,128,190,98,37,1,160,234,94,
156,45,159,70,18,50,147,229,60,114,216,226,
177,109,9,94,61,209,32,233,147,150,200,167,
68,60,198,206,158,56,199,210,99,216,8,148,
75,253,147,30,125,176,121,220,232,187,103,39,
168,0,178,203,203,119,126,252,44,223,250,139,
47,162,235,250,194,51,138,197,34,13,195,194,
78,15,55,97,42,217,133,30,138,131,36,176,
133,74,60,30,227,255,252,111,127,207,155,239,
29,103,120,230,12,82,168,27,217,229,198,170,
22,208,139,35,84,203,5,100,79,24,211,17,
136,88,47,106,48,65,238,252,65,92,59,30,
95,49,100,36,187,60,144,216,70,6,32,0,
31,158,41,89,247,34,71,123,168,105,58,71,
243,85,206,189,116,144,111,125,229,201,91,14,
73,173,138,203,242,249,60,94,175,151,29,61,
9,14,103,235,55,160,40,110,46,115,42,119,
120,245,174,101,214,25,61,247,31,168,219,210,
75,107,160,28,176,46,180,211,222,245,23,168,
158,91,99,69,20,210,103,153,43,253,22,101,
99,142,171,157,84,43,135,109,8,231,72,143,
60,143,43,24,193,54,220,212,109,27,36,16,
211,25,28,85,65,50,109,164,92,1,169,84,
197,218,209,135,29,95,108,235,90,179,108,236,
154,132,230,94,255,243,65,12,107,30,52,7,
9,48,58,14,81,62,239,69,139,217,84,171,
67,52,98,135,200,140,180,147,140,125,1,151,
63,113,219,5,21,47,143,81,172,7,205,21,
27,172,119,187,192,76,185,65,60,160,93,181,
16,112,89,75,251,82,146,131,230,13,98,204,
11,94,61,60,128,233,73,96,25,115,216,182,
141,16,130,23,95,121,157,39,62,241,48,15,
63,112,47,193,147,103,112,128,128,111,41,194,
97,35,97,230,38,209,156,26,141,158,135,248,
143,223,31,167,158,30,228,191,255,221,215,81,
20,133,66,161,136,48,171,180,73,85,138,229,
42,143,61,242,0,191,125,119,0,145,216,130,
75,50,209,117,29,93,215,249,204,227,143,96,
219,54,71,79,156,34,91,152,34,154,8,176,
231,241,199,22,121,204,48,120,251,240,113,6,
166,167,9,111,250,216,186,241,145,203,31,198,
46,13,127,36,124,186,42,69,49,51,151,225,
185,159,191,64,123,212,131,149,55,176,236,58,
242,198,79,174,14,206,176,157,27,78,141,189,
66,208,218,160,220,193,9,52,19,131,63,66,
221,180,84,73,216,243,42,114,102,35,93,27,
191,178,164,251,228,122,82,173,48,195,124,245,
57,148,142,220,202,54,116,215,24,149,241,28,
138,225,80,175,213,145,207,143,81,204,214,17,
133,49,220,186,140,105,59,168,170,64,140,207,
46,81,20,37,211,194,41,120,209,91,227,235,
11,157,153,13,140,82,105,193,74,19,46,208,
183,151,49,39,93,40,114,4,201,151,193,170,
230,104,212,242,204,143,191,143,230,241,34,20,
55,145,182,93,43,59,223,120,221,61,138,165,
49,138,155,77,233,82,99,221,191,161,97,218,
156,154,40,210,219,226,189,110,65,224,178,60,
230,141,33,34,41,148,137,99,220,255,200,30,
132,16,76,76,78,113,124,96,136,253,123,118,
48,59,151,193,6,118,111,95,218,120,210,178,
44,10,53,11,203,202,160,82,67,141,110,4,
125,35,146,30,224,212,217,243,236,218,190,133,
64,192,143,18,76,48,90,175,241,153,253,91,
232,104,111,167,237,236,8,83,92,209,42,11,
33,4,247,236,94,254,44,107,85,85,121,248,
254,125,180,39,135,121,246,88,26,37,222,185,
126,243,169,199,25,159,156,162,189,117,245,9,
59,215,11,43,220,52,69,49,59,51,75,195,
145,25,169,122,232,74,168,36,67,30,142,206,
167,209,2,43,223,240,149,134,117,211,109,152,
82,221,92,168,167,184,147,200,177,45,70,79,
254,8,167,103,136,15,50,69,157,170,132,93,
151,209,230,118,209,182,245,214,225,232,86,163,
194,244,204,51,136,158,249,43,55,43,160,10,
9,195,110,90,230,151,219,184,69,205,207,182,
246,89,186,199,50,100,199,114,12,95,44,33,
201,18,146,3,186,46,179,121,71,4,115,203,
98,225,93,222,48,57,158,45,226,50,183,33,
171,235,123,226,156,80,52,132,255,74,163,68,
110,169,83,189,144,67,111,21,56,25,63,129,
237,155,208,244,48,211,197,255,192,150,12,138,
131,135,208,229,94,2,161,109,127,180,30,133,
105,217,235,30,196,94,52,228,28,46,204,148,
216,146,242,227,211,215,6,149,137,182,221,60,
115,120,18,183,113,10,204,6,74,172,151,127,
250,206,79,80,116,15,95,124,226,33,0,94,
124,229,117,26,245,42,229,90,3,159,47,128,
240,70,208,2,201,102,6,209,7,66,61,152,
224,245,145,73,222,59,243,60,170,166,35,123,
219,145,200,98,92,202,59,214,181,230,70,212,
213,213,9,212,116,122,142,137,169,25,156,233,
1,156,72,235,186,193,153,114,168,149,131,199,
78,243,212,26,20,197,141,120,34,171,250,154,
143,61,120,63,167,71,126,77,126,126,148,25,
215,38,102,42,85,212,216,234,10,70,102,139,
55,223,138,41,223,177,138,194,198,112,50,104,
238,166,48,112,234,2,235,66,27,129,192,78,
98,91,238,187,101,227,200,140,189,75,161,254,
14,114,207,236,135,32,61,65,151,199,69,183,
207,77,197,106,246,230,26,44,86,136,235,26,
69,195,164,110,59,204,170,130,77,18,136,116,
150,208,182,54,250,188,115,100,38,139,4,252,
42,161,160,134,185,165,27,244,166,71,84,183,
28,222,30,41,209,152,232,166,165,237,145,91,
242,109,225,224,35,100,210,191,64,120,171,216,
195,29,40,94,21,187,102,33,251,115,72,231,
118,208,190,253,241,166,96,8,180,16,40,127,
130,130,253,26,82,247,40,213,236,20,74,206,
143,59,252,209,181,164,95,79,143,162,102,216,
183,150,215,29,56,55,83,98,119,231,218,33,
84,53,212,138,73,83,222,120,199,222,225,91,
127,247,117,2,126,31,133,66,145,255,124,246,
37,50,134,27,87,168,23,17,212,200,169,46,
20,119,83,48,10,121,41,220,160,133,90,169,
72,42,122,48,222,140,187,185,3,188,57,48,
136,209,104,80,44,22,33,4,37,188,252,240,
23,191,227,201,71,31,96,124,114,146,142,182,
54,34,225,165,137,23,213,106,149,247,207,156,
227,158,93,219,153,78,103,56,120,106,4,119,
207,189,235,26,243,146,36,193,156,214,197,225,
227,167,184,103,215,173,51,100,86,252,69,153,
249,121,76,211,162,167,173,133,211,245,6,142,
47,126,41,77,114,229,148,175,26,75,142,51,
189,89,84,172,90,183,213,201,102,215,59,49,
110,97,62,166,79,65,96,241,120,86,107,36,
68,107,239,151,154,197,72,183,8,135,156,25,
122,153,106,248,13,228,148,137,4,180,121,92,
116,122,220,248,13,3,109,46,135,60,49,141,
213,145,196,157,104,246,245,138,186,130,136,90,
3,220,26,13,203,230,204,148,137,236,56,216,
209,32,47,4,3,132,239,141,178,219,173,161,
84,235,156,56,109,179,53,26,162,97,219,164,
107,6,231,7,20,228,218,39,209,228,34,245,
202,28,110,255,250,215,187,4,226,155,32,253,
69,230,135,223,166,107,235,55,16,138,74,101,
126,10,79,56,121,197,28,71,82,251,241,20,
186,72,15,188,130,199,147,192,27,249,104,91,
144,44,151,245,116,51,121,244,86,147,105,57,
20,107,38,62,249,218,26,101,126,226,8,13,
99,6,167,33,19,140,239,92,222,195,240,132,
57,115,126,144,106,181,206,197,116,25,171,101,
39,98,126,106,213,50,105,65,0,71,55,112,
112,58,143,45,183,161,1,82,168,131,18,29,
252,248,181,51,52,26,13,220,3,179,116,134,
85,54,117,183,33,132,96,114,38,205,217,241,
44,245,80,31,135,158,125,29,124,73,60,221,
251,110,202,201,159,215,245,42,220,126,6,198,
198,185,231,22,158,148,176,226,175,154,153,77,
243,139,151,222,97,95,127,138,168,87,38,151,
185,8,158,221,43,87,18,21,131,243,51,235,
83,77,92,170,27,216,182,131,184,77,130,218,
215,59,49,14,154,103,121,228,38,78,160,237,
109,122,88,118,94,193,239,221,131,30,140,83,
153,159,186,101,99,173,152,39,144,125,38,138,
36,177,55,226,111,102,36,1,242,133,81,196,
76,166,249,13,166,5,153,60,98,100,10,9,
7,41,87,2,69,70,234,239,34,112,94,229,
144,87,224,74,181,96,154,38,33,77,161,174,
41,200,110,23,129,182,6,182,229,112,112,62,
79,113,52,72,220,247,21,208,100,102,39,126,
131,144,85,96,199,45,249,198,64,124,51,129,
248,230,15,47,210,178,215,234,129,22,58,182,
125,173,105,117,127,232,172,227,91,239,81,92,
59,235,233,70,168,121,98,158,125,171,119,198,
53,127,173,151,210,76,79,62,139,211,58,134,
208,29,140,19,41,12,107,154,152,252,73,28,
211,193,29,106,93,16,196,86,180,159,35,133,
42,178,162,33,181,220,156,0,165,226,9,98,
27,117,108,163,142,184,4,139,170,45,125,88,
249,52,34,24,103,28,184,120,62,139,227,88,
168,158,54,68,178,179,41,64,147,205,186,9,
163,90,186,101,51,153,175,89,183,116,229,86,
164,40,108,219,230,133,63,188,129,99,217,156,
29,155,67,138,244,172,10,114,154,43,53,123,
214,85,39,0,0,32,0,73,68,65,84,24,
74,87,214,205,138,177,29,40,214,76,130,119,
10,252,228,216,140,158,254,33,210,230,11,56,
22,72,10,24,231,98,196,238,127,228,150,15,
69,182,99,192,60,221,62,157,233,90,3,143,
34,227,83,100,236,254,78,68,182,0,13,3,
121,96,248,74,235,176,35,201,241,89,31,158,
96,39,115,7,103,80,183,204,210,214,169,178,
193,183,120,80,139,156,180,121,57,157,165,158,
246,16,113,127,6,151,175,133,201,243,207,160,
168,81,18,189,79,172,175,53,158,30,196,27,
237,94,210,38,164,60,63,68,113,254,28,129,
200,206,59,130,77,202,254,240,66,175,167,155,
77,186,42,227,214,164,155,158,88,178,28,105,
138,32,228,81,137,249,52,124,186,130,117,149,
119,22,102,207,226,180,143,34,46,229,110,168,
59,155,6,211,204,196,44,142,92,69,154,73,
210,209,247,205,133,195,128,62,124,40,208,135,
201,106,84,155,61,161,46,59,150,55,59,121,
132,106,101,28,183,218,133,203,23,166,82,152,
196,27,110,6,159,135,79,252,39,150,146,69,
242,85,145,230,19,200,62,129,223,187,19,183,
103,177,81,160,162,123,169,229,103,49,106,35,
56,142,141,191,165,15,199,177,113,22,90,146,
220,26,79,205,116,133,153,153,157,37,209,114,
107,26,131,42,215,246,34,102,121,238,229,55,
104,111,77,240,223,254,247,255,66,181,90,229,
183,47,29,96,86,95,121,145,84,177,106,48,
146,91,255,190,68,133,59,72,81,148,231,199,
176,106,85,132,36,33,20,135,198,136,78,56,
185,7,233,150,101,218,56,228,167,79,81,40,
30,35,231,205,19,2,38,171,13,20,73,194,
167,200,92,40,86,41,155,22,187,119,247,163,
188,119,106,9,239,59,65,31,217,182,4,163,
186,139,11,167,60,56,141,41,188,247,204,178,
189,221,67,232,82,225,151,3,28,153,47,144,
174,155,212,222,239,32,154,186,15,89,214,24,
27,250,14,162,111,18,123,62,192,196,185,159,
211,214,191,178,243,42,44,163,142,80,212,85,
205,143,109,55,152,31,61,66,180,123,177,48,
177,86,110,198,96,114,51,39,208,124,161,235,
10,154,143,154,60,165,124,179,123,108,116,125,
226,36,137,128,206,240,220,250,237,205,168,79,
35,21,210,241,104,43,179,248,67,173,123,40,
206,190,137,104,93,58,38,185,237,18,60,155,
24,99,252,194,143,104,239,251,58,138,122,237,
212,252,185,145,55,40,72,175,35,41,14,202,
208,6,58,182,54,189,196,98,253,24,116,12,
145,29,185,72,110,228,15,88,5,193,134,192,
63,82,202,142,96,123,51,168,93,217,230,3,
82,67,32,65,121,208,187,160,40,102,134,95,
166,106,159,196,9,207,129,4,174,244,125,88,
245,58,153,194,11,200,4,176,204,6,194,91,
195,37,109,0,71,16,77,62,136,203,23,93,
151,185,117,69,218,56,244,254,57,62,251,216,
109,160,40,14,30,59,75,94,111,35,119,246,
12,247,95,2,196,146,209,16,211,133,149,215,
65,212,173,91,163,97,87,123,248,252,71,73,
222,104,23,113,243,147,204,156,56,128,136,200,
72,102,137,200,214,181,85,90,87,242,227,152,
86,142,64,252,234,231,82,59,142,141,101,212,
169,151,211,100,230,127,143,41,229,145,98,243,
56,17,9,179,24,71,0,85,203,98,87,200,
207,161,57,139,225,2,244,71,0,211,90,170,
36,60,58,185,77,221,188,158,175,176,67,83,
184,247,225,10,46,151,68,76,15,47,128,10,
13,219,97,180,92,163,108,218,152,182,141,39,
152,66,117,133,152,45,254,20,121,67,51,237,
86,138,23,104,84,135,153,185,248,50,150,83,
196,31,220,129,39,216,190,108,22,148,227,216,
140,158,253,79,36,183,73,208,253,0,225,214,
221,43,138,223,4,18,91,174,232,38,26,237,
184,143,209,129,239,193,134,115,140,94,60,67,
247,134,127,88,128,24,110,71,170,122,3,8,
123,253,44,254,152,95,99,60,91,193,188,201,
123,212,163,201,116,69,61,248,221,171,195,235,
11,115,239,35,130,213,107,34,87,82,207,16,
179,131,47,209,210,253,9,20,253,234,45,185,
109,44,68,160,142,240,89,200,181,69,3,210,
110,24,8,25,148,222,38,172,234,156,107,1,
33,168,22,47,162,116,102,151,188,203,206,184,
137,37,154,85,220,147,231,127,73,163,245,48,
194,213,156,43,251,66,55,225,196,62,198,207,
60,141,210,159,71,202,5,145,220,5,228,206,
50,38,205,19,30,211,67,85,218,55,126,117,
221,214,111,238,22,158,31,178,236,74,230,11,
5,222,122,239,24,231,46,12,177,123,203,6,
70,170,1,254,237,103,207,35,249,226,72,222,
8,242,42,44,177,168,87,35,83,18,24,214,
250,225,161,126,183,66,50,160,113,167,144,213,
168,49,87,252,13,218,254,44,230,164,78,50,
248,55,107,110,242,151,175,29,64,88,174,5,
69,81,156,187,128,16,42,222,200,98,123,239,
137,51,63,199,27,232,163,82,26,197,217,184,
120,110,196,76,174,157,251,219,138,76,149,226,
140,229,85,94,204,129,75,173,82,173,71,232,
104,47,66,185,2,18,56,161,0,181,13,109,
216,62,15,239,205,230,8,168,50,33,85,161,
34,44,100,73,44,65,158,39,171,53,206,22,
202,11,94,184,85,177,153,117,158,69,116,126,
168,54,35,84,164,60,56,138,178,231,34,115,
185,99,56,99,94,100,43,132,46,119,226,210,
91,209,189,77,43,90,146,4,154,30,198,234,
62,65,190,242,44,217,139,175,34,36,13,197,
138,145,234,254,226,181,83,108,63,164,80,102,
47,28,192,176,211,136,146,132,210,155,33,51,
121,136,120,215,67,183,45,159,232,149,18,178,
101,80,140,173,79,43,106,33,65,42,164,51,
150,185,57,45,198,21,33,209,22,117,211,18,
212,215,212,82,167,218,24,66,120,175,46,39,
28,19,156,11,157,24,82,134,225,145,255,151,
182,228,223,98,27,117,188,209,118,178,227,239,
227,242,54,173,235,177,19,191,192,114,242,136,
220,54,172,228,25,130,158,102,124,42,51,250,
14,137,196,23,152,30,254,9,114,247,28,118,
89,16,208,239,109,86,101,59,226,138,16,138,
82,234,198,147,104,99,230,220,171,52,90,143,
44,40,9,43,167,18,246,223,75,49,59,128,
39,220,129,33,207,33,112,97,217,26,176,232,
13,217,114,110,93,249,163,104,123,200,231,243,
4,131,235,95,140,187,172,162,208,93,46,66,
1,63,170,75,231,253,139,83,96,219,168,129,
24,122,107,83,24,173,38,104,99,88,246,186,
246,97,242,186,20,250,19,62,184,67,186,61,
85,139,83,76,14,255,12,101,99,22,73,128,
147,245,227,233,92,27,180,96,27,117,156,124,
0,111,106,241,160,20,163,158,195,237,111,186,
202,150,217,96,252,196,207,137,117,63,132,59,
216,70,185,56,184,20,122,8,141,243,222,100,
10,143,94,36,30,76,95,182,97,189,212,44,
27,39,28,192,124,96,23,89,33,120,51,157,
67,45,55,104,216,54,88,54,175,205,102,145,
36,184,47,218,100,210,188,97,146,107,152,156,
188,12,102,156,203,37,241,155,231,113,93,110,
169,125,0,39,248,44,106,229,6,138,13,114,
200,130,80,1,40,80,103,148,202,140,135,88,
229,171,232,193,102,16,87,86,124,88,128,240,
216,136,174,102,157,135,237,76,51,51,226,165,
181,239,115,43,158,175,150,190,71,104,113,30,
38,51,114,28,51,63,143,238,79,220,214,188,
82,215,61,235,234,81,0,164,130,46,50,197,
6,149,27,140,85,248,116,153,190,246,0,154,
182,54,131,205,113,28,12,59,203,114,32,149,
99,1,163,237,200,213,56,13,207,24,74,215,
40,202,172,31,73,150,153,26,250,57,158,76,
47,245,114,14,105,30,186,35,127,131,227,153,
67,116,143,98,93,76,162,207,63,64,112,203,
86,234,165,121,114,246,243,88,153,79,162,216,
45,56,204,225,212,85,10,179,231,136,116,238,
193,178,11,75,223,89,21,184,148,246,5,247,
66,82,22,21,152,83,118,33,92,26,229,200,
75,200,222,166,194,85,114,161,43,158,97,106,
25,140,90,25,85,95,159,195,136,212,88,23,
111,31,61,197,147,31,127,240,163,81,20,46,
151,139,183,15,31,71,117,251,80,100,129,95,
145,152,175,231,87,245,224,116,177,206,68,182,
138,145,47,227,120,61,235,50,120,183,42,216,
148,244,33,223,65,45,60,172,70,13,201,95,
65,82,192,174,10,34,241,135,214,148,119,109,
25,117,70,206,252,27,34,98,82,179,7,25,
59,63,70,75,234,207,136,180,237,99,252,253,
103,73,246,63,129,144,53,92,158,4,115,179,
47,160,102,90,16,184,175,200,115,137,5,167,
150,241,240,29,70,202,85,82,110,141,17,219,
230,100,166,112,9,86,106,222,173,72,18,178,
144,216,17,244,18,115,169,152,142,195,225,249,
34,21,115,81,216,76,102,186,105,9,77,160,
132,151,119,143,173,10,212,83,6,250,50,136,
161,220,82,97,226,220,115,216,57,47,126,151,
192,82,51,136,101,96,136,70,228,40,243,19,
9,34,109,171,128,237,36,9,119,32,129,39,
178,251,182,231,21,181,81,67,182,76,214,23,
96,144,232,142,121,56,51,89,92,115,24,182,
37,224,34,229,2,69,94,59,252,91,203,78,
65,120,102,57,12,9,233,194,22,52,87,156,
106,242,16,74,164,217,31,204,78,7,201,228,
14,34,247,100,169,143,79,16,106,249,24,115,
179,47,146,30,57,128,42,165,48,24,69,238,
158,198,25,105,10,123,205,29,64,194,75,217,
243,38,82,67,67,0,114,164,142,37,206,51,
54,240,125,84,41,1,54,216,115,126,148,106,
55,46,165,133,120,87,51,185,196,180,242,72,
151,107,48,111,5,179,92,5,161,129,175,153,
181,232,216,13,52,179,11,70,146,8,161,97,
99,162,152,98,93,139,74,27,133,57,222,59,
249,14,143,63,124,31,178,188,190,173,41,150,
149,80,182,109,179,127,71,63,195,211,57,210,
133,42,247,237,218,202,145,129,81,86,106,115,
100,203,205,44,39,214,217,206,239,138,121,81,
228,59,171,207,147,109,84,145,66,151,152,125,
214,71,164,119,207,154,158,83,152,29,64,244,
78,33,116,128,52,230,180,111,161,213,135,161,
78,145,29,63,6,178,141,227,216,216,129,28,
13,109,6,207,252,199,169,143,133,145,59,178,
215,129,36,108,74,70,179,192,206,178,29,62,
145,140,160,72,18,175,205,102,169,89,54,237,
151,138,240,124,138,140,3,28,203,46,85,18,
233,124,43,33,223,28,138,124,117,17,231,184,
85,26,49,249,170,88,180,103,211,12,117,67,
103,170,20,39,226,175,240,129,157,106,87,192,
154,115,163,180,85,17,129,6,141,252,220,202,
45,244,114,134,236,244,187,212,140,81,196,188,
130,36,105,8,52,36,35,64,188,243,161,53,
231,224,175,155,81,161,168,235,238,81,52,189,
1,133,120,192,197,108,97,181,53,78,18,93,
49,55,137,128,139,198,13,182,3,41,151,70,
145,147,87,170,42,179,32,240,235,61,20,67,
207,161,4,23,127,87,182,143,99,217,227,72,
117,176,212,18,249,224,207,208,34,80,158,60,
133,79,217,133,225,0,2,12,121,178,57,82,
89,65,54,195,72,237,195,75,224,33,57,228,
224,232,195,232,217,189,20,70,83,8,220,180,
247,45,141,43,24,98,154,203,109,81,169,226,
70,117,251,113,166,163,16,153,194,158,115,35,
219,17,42,214,0,30,185,143,74,110,12,130,
121,132,112,147,153,120,139,120,231,250,100,51,
202,149,52,143,125,252,97,138,197,34,161,208,
250,118,97,94,214,4,16,66,240,224,3,247,
83,204,103,217,220,25,231,245,83,19,140,159,
59,137,109,174,140,25,138,85,243,150,108,164,
181,182,3,248,168,168,81,201,50,59,254,218,
2,214,41,121,107,84,215,152,171,95,171,141,
92,82,18,151,214,172,30,67,213,3,24,213,
2,142,90,162,156,25,163,210,242,123,234,201,
55,112,104,32,194,117,42,230,57,34,238,79,
97,207,94,255,44,139,154,109,115,34,87,162,
219,235,198,45,11,170,150,77,205,178,105,117,
187,216,18,244,226,83,100,38,171,117,14,204,
230,152,174,94,178,170,28,137,137,76,15,97,
95,26,143,235,218,240,228,108,174,141,120,112,
242,154,215,184,212,26,201,240,24,185,82,140,
90,163,233,149,10,15,212,171,17,42,175,71,
177,11,26,146,179,114,168,99,122,248,121,26,
29,111,33,250,198,161,123,24,127,223,32,117,
103,12,127,100,19,67,239,252,132,225,163,63,
100,226,226,211,204,141,188,113,89,186,227,71,
71,21,127,8,83,213,232,62,241,214,186,191,
171,43,234,38,176,138,224,179,34,75,108,105,
245,145,8,220,28,139,217,180,114,203,75,163,
162,23,161,57,8,83,95,86,122,9,55,168,
189,69,132,27,36,13,108,81,66,211,163,88,
151,194,46,246,101,231,136,7,125,247,98,78,
94,57,94,107,58,68,160,165,31,197,9,99,
250,199,168,230,22,61,27,219,50,176,148,165,
109,109,156,134,11,199,177,144,218,155,215,57,
243,33,170,156,66,238,31,39,151,62,142,29,
156,64,180,205,33,117,141,81,14,190,194,220,
216,193,155,191,96,142,67,72,179,104,111,77,
50,56,50,190,238,252,113,85,95,81,8,193,
63,252,237,55,248,179,199,30,225,137,189,189,
68,182,127,124,197,205,233,26,183,40,211,169,
102,88,220,73,148,157,62,140,182,107,18,233,
82,220,218,49,20,180,53,116,133,117,28,155,
134,51,245,33,215,48,136,99,91,76,142,62,
141,156,42,224,107,233,193,25,107,67,248,45,
148,68,243,68,49,59,54,129,109,57,196,92,
95,193,26,186,122,90,157,237,52,217,98,172,
82,227,189,249,2,14,160,203,130,123,163,1,
182,133,188,8,73,194,114,28,90,116,141,126,
255,34,172,56,53,223,69,91,100,248,154,158,
196,2,92,17,154,96,46,191,178,202,236,150,
208,4,133,114,100,81,128,38,192,189,55,143,
121,166,21,111,168,119,201,230,105,148,175,238,
45,105,36,48,39,117,60,178,64,19,130,130,
97,33,185,77,230,74,191,64,219,63,142,216,
113,6,171,243,56,229,196,11,12,159,249,95,
216,150,241,145,243,76,57,16,97,120,231,131,
116,158,122,151,208,204,216,250,1,80,146,68,
95,194,215,60,123,251,122,144,175,38,179,173,
45,176,162,131,136,86,12,179,105,33,236,249,
43,147,100,36,195,67,161,118,24,17,189,122,
192,189,80,9,83,174,93,242,6,125,101,172,
186,1,245,230,119,56,194,88,168,110,15,37,
119,208,26,252,54,214,217,212,210,108,62,26,
216,166,129,84,247,163,101,118,163,7,23,247,
134,144,21,68,117,105,219,7,185,53,75,49,
123,6,117,118,103,51,192,30,154,69,244,76,
55,121,108,83,22,181,179,182,0,85,9,175,
73,169,113,152,90,97,250,166,174,151,81,206,
82,202,206,114,113,120,148,11,163,235,95,160,
123,5,87,204,204,204,48,50,186,148,33,35,
33,63,141,169,179,20,39,47,172,76,81,152,
183,70,128,143,207,87,185,147,200,237,107,195,
156,12,224,92,50,86,229,88,133,236,212,145,
85,63,103,230,226,139,208,190,116,141,20,41,
196,220,232,27,208,213,204,255,22,138,74,107,
199,95,32,157,223,142,115,201,17,84,162,13,
114,190,167,153,155,127,30,39,179,60,204,98,
152,46,100,217,188,12,134,186,212,20,80,18,
52,12,157,115,133,10,47,77,103,120,63,87,
66,145,36,146,186,70,167,87,103,112,106,27,
45,161,9,144,86,102,36,40,178,177,160,144,
86,164,28,47,115,253,117,173,66,67,215,80,
182,76,83,41,12,47,42,226,137,99,140,15,
254,24,199,177,169,20,102,169,151,22,97,169,
252,244,105,92,129,8,34,108,82,177,108,92,
114,51,91,75,78,148,145,59,115,8,221,94,
178,43,140,106,241,186,71,118,222,124,35,241,
234,115,55,186,237,94,74,225,22,186,78,190,
179,126,56,180,144,216,148,244,95,179,203,107,
208,173,176,181,213,143,235,38,167,163,71,219,
238,199,93,186,127,97,111,0,216,53,9,185,
30,71,163,13,59,211,52,72,236,250,229,243,
213,100,10,143,171,68,169,218,52,184,228,128,
69,190,240,206,130,49,134,106,96,94,182,142,
122,48,65,215,230,255,138,57,28,197,177,192,
58,215,74,50,242,117,36,33,83,15,158,160,
33,141,125,168,109,138,68,192,119,15,118,105,
17,38,149,52,135,70,203,113,124,129,62,196,
200,54,228,208,226,245,242,50,219,74,116,206,
48,53,251,83,230,70,222,188,105,198,135,85,
205,243,192,254,61,84,12,137,121,195,197,63,
253,203,191,82,46,175,95,77,204,21,171,29,
139,197,248,254,211,191,226,119,47,31,192,178,
154,2,63,209,210,130,223,239,167,49,63,114,
91,121,20,185,138,65,161,106,114,167,80,160,
101,11,65,231,227,56,239,111,195,42,131,164,
66,197,56,181,170,138,245,220,204,73,106,222,
163,72,151,25,115,86,70,35,24,223,69,195,
156,65,82,154,207,45,148,14,145,158,254,61,
78,67,90,178,185,100,175,141,19,155,69,146,
150,183,6,51,133,4,81,255,162,235,221,162,
107,164,107,6,63,57,27,228,104,126,142,145,
114,141,186,229,160,8,193,92,221,96,178,90,
231,208,120,146,120,112,146,186,177,186,243,73,
132,100,163,10,137,79,38,194,184,132,32,226,
82,175,121,237,194,60,122,178,228,203,17,68,
160,70,197,117,136,242,37,190,172,214,71,145,
55,78,49,50,240,111,12,101,126,204,249,153,
31,48,240,254,79,49,107,37,210,51,207,147,
203,191,141,93,110,126,119,209,48,175,26,188,
149,4,216,189,18,217,204,192,45,229,143,203,
219,189,152,245,10,229,204,48,229,204,8,102,
189,105,16,153,154,139,76,91,47,145,201,97,
212,250,77,48,146,28,135,114,102,233,158,214,
20,193,198,164,143,254,164,143,100,192,69,80,
87,112,107,50,62,151,76,103,212,77,255,58,
38,143,180,116,61,130,51,213,130,221,0,78,
238,32,152,255,50,177,182,135,105,219,248,101,
2,141,79,81,57,16,163,113,50,76,113,54,
72,58,223,202,92,33,69,58,223,74,174,20,
35,17,110,26,78,146,2,98,203,48,194,103,
55,189,6,181,113,133,194,23,138,134,110,247,
163,142,222,67,247,166,111,227,14,182,50,61,
252,91,132,44,16,29,51,20,102,151,174,187,
55,220,133,152,238,195,46,168,151,41,4,131,
236,212,17,252,190,109,216,163,215,105,177,34,
129,220,157,166,220,242,60,35,163,255,15,67,
39,254,253,134,231,74,11,183,241,234,225,115,
140,230,26,52,100,15,182,43,200,208,58,66,
80,87,72,139,239,62,253,107,58,58,186,40,
22,10,252,243,79,158,163,61,30,228,11,79,
60,204,199,239,221,206,91,39,135,185,110,148,
194,1,195,188,117,61,100,134,210,101,182,182,
5,80,87,16,212,182,204,6,245,98,26,163,
158,163,86,153,197,237,107,37,63,113,22,220,
85,82,61,127,142,226,242,172,251,120,131,201,
77,148,231,39,48,106,32,123,129,214,49,38,
6,158,161,125,211,151,175,91,72,102,25,117,
178,213,23,144,219,47,89,14,54,24,227,62,
36,213,160,238,204,18,138,236,35,157,25,68,
68,43,136,190,49,44,154,155,231,138,144,241,
120,10,61,26,92,54,155,230,114,111,2,224,
68,182,68,161,28,65,82,230,144,46,19,214,
195,165,42,115,53,147,179,211,157,180,70,135,
48,76,13,195,108,226,191,113,189,153,85,82,
179,108,98,186,74,166,110,80,48,44,236,203,
20,98,195,208,209,212,26,134,237,32,46,245,
153,210,101,193,108,173,193,169,252,149,150,145,
99,47,111,193,202,241,2,217,225,119,136,197,
63,129,201,28,146,234,32,245,141,226,7,38,
230,122,73,182,159,97,248,212,20,202,166,44,
194,181,168,252,102,107,215,230,228,172,17,163,
170,140,16,203,143,227,9,182,47,97,112,163,
86,198,170,87,81,116,239,186,241,76,110,242,
8,185,226,59,216,24,204,12,249,232,255,216,
63,34,73,18,165,112,28,17,8,211,122,254,
56,51,61,91,49,92,107,175,46,207,140,191,
71,94,121,30,215,192,78,84,17,161,165,239,
225,5,30,12,121,84,66,30,149,154,234,65,
15,222,154,32,191,36,43,68,253,159,162,158,
206,16,221,209,236,194,250,65,223,179,124,254,
16,158,135,155,21,209,58,224,103,5,89,152,
18,72,170,133,81,43,224,242,71,150,162,36,
201,253,11,233,215,217,241,67,216,221,39,17,
151,244,64,121,106,128,96,114,233,185,215,29,
27,255,130,129,3,255,3,109,251,28,74,216,
198,49,192,105,29,35,83,153,68,233,89,153,
231,41,169,32,119,20,176,203,21,166,46,252,
142,96,104,7,4,215,214,199,75,40,26,174,
222,7,154,74,163,81,165,210,168,115,240,204,
24,181,122,157,125,123,118,174,191,162,80,85,
141,114,205,32,151,43,224,142,117,48,89,22,
216,182,77,127,223,6,222,60,53,177,162,197,
17,66,194,178,111,141,87,81,55,109,46,204,
150,216,156,244,93,179,223,186,99,91,92,120,
235,159,9,36,54,81,175,77,225,138,204,80,
205,234,84,10,18,94,29,134,79,255,27,193,
208,30,188,129,222,117,29,175,162,7,104,223,
249,25,138,153,77,148,46,14,96,152,25,108,
39,131,89,175,92,179,210,20,160,81,205,114,
57,184,106,205,233,200,153,30,236,248,48,197,
198,113,218,147,95,39,123,97,3,118,244,253,
107,235,242,196,36,185,145,10,222,238,229,150,
207,105,42,6,201,198,239,201,129,35,81,172,
5,105,139,14,93,113,237,185,153,54,218,99,
205,218,140,82,45,72,212,63,131,34,73,236,
9,251,208,196,82,193,94,181,108,206,23,43,
232,178,96,32,95,99,54,215,70,123,124,16,
151,44,48,28,135,232,37,111,66,44,179,134,
233,124,43,209,224,244,146,81,154,150,198,254,
104,128,146,105,113,38,124,129,137,115,121,228,
13,211,75,92,100,77,173,33,203,38,242,206,
165,153,81,217,186,129,196,181,187,242,84,13,
15,109,27,46,50,51,245,67,92,233,62,144,
29,76,114,88,228,113,180,10,146,203,128,162,
134,94,219,69,106,21,181,28,43,165,88,207,
199,8,86,155,41,188,170,238,93,98,68,216,
178,194,248,230,123,104,61,119,156,154,63,196,
124,170,107,229,49,143,236,16,133,244,89,76,
105,30,203,59,141,18,55,152,159,29,198,27,
58,75,99,104,156,218,92,142,158,221,127,187,
230,211,43,87,67,165,204,16,186,55,134,162,
47,38,87,248,99,155,240,47,115,240,161,174,
183,83,203,204,128,98,227,212,93,40,70,8,
199,22,216,254,105,228,160,121,213,104,171,236,
135,244,208,111,17,202,159,227,9,119,44,123,
77,173,49,177,8,85,1,141,232,41,114,211,
27,8,37,119,46,81,98,145,222,237,84,26,
135,129,60,146,10,74,220,128,53,36,47,11,
175,137,213,251,30,179,83,39,41,13,238,37,
185,225,211,107,151,39,19,239,144,138,132,153,
112,41,148,45,63,233,185,204,250,123,20,23,
135,71,41,149,42,116,164,162,196,55,247,112,
116,96,24,39,185,139,151,223,124,143,88,192,
139,140,181,162,105,113,41,226,134,11,120,86,
67,197,170,201,72,166,74,119,236,234,214,93,
173,56,75,172,237,33,230,243,175,210,214,215,
96,246,164,224,193,143,155,76,119,152,228,48,
216,30,87,145,141,131,28,59,114,10,79,228,
239,215,181,205,183,80,52,130,137,45,4,19,
75,219,110,204,12,254,129,224,53,90,113,184,
3,73,228,185,16,92,178,166,228,150,26,182,
247,36,138,215,193,26,45,54,149,128,184,62,
78,41,66,130,154,164,178,156,90,42,84,194,
36,195,99,56,72,100,10,73,108,71,144,138,
140,92,37,206,176,232,125,4,61,243,228,203,
81,194,190,121,222,158,43,16,214,20,182,135,
124,11,233,209,110,89,176,57,224,97,160,80,
97,114,174,135,246,248,224,7,232,7,186,16,
24,182,195,72,185,198,88,165,105,157,149,107,
1,114,165,24,138,210,64,72,54,178,88,124,
215,68,166,155,135,186,166,137,24,14,73,7,
138,41,155,76,96,2,71,113,48,108,9,33,
65,165,230,199,165,212,112,9,129,225,216,124,
96,183,116,121,117,250,253,30,78,230,203,76,
85,235,248,85,153,168,166,54,21,88,161,130,
3,228,203,81,118,181,142,33,201,50,165,84,
17,147,163,139,130,103,41,87,81,155,63,74,
37,187,27,33,214,177,21,200,85,120,113,178,
127,23,194,50,217,242,214,115,12,220,119,237,
38,139,141,114,150,252,220,73,202,206,17,68,
111,83,113,22,203,81,140,124,10,119,111,9,
197,93,198,230,12,106,74,98,252,252,79,80,
245,16,66,184,144,12,29,217,181,15,85,247,
221,244,207,106,84,231,72,79,188,72,207,206,
191,191,238,181,169,190,207,82,43,222,139,99,
89,104,129,224,66,119,136,244,249,183,201,165,
95,198,213,95,187,58,236,211,59,77,122,252,
89,82,174,111,162,121,66,203,76,239,82,147,
65,4,234,228,134,222,34,152,216,190,164,191,
88,188,253,97,46,188,57,136,80,10,168,209,
27,52,132,37,16,161,58,245,145,57,202,243,
99,120,35,29,107,128,14,109,76,71,176,161,
35,193,3,177,8,191,123,241,85,82,61,125,
235,175,40,90,98,17,20,97,113,240,173,183,
112,197,186,232,74,70,168,205,188,207,187,115,
25,240,68,209,124,33,212,249,41,204,122,5,
33,20,132,186,124,22,148,92,173,64,237,210,
198,254,144,21,124,77,90,205,181,181,2,216,
139,194,99,182,12,74,85,39,226,93,28,83,
105,254,2,178,240,97,214,170,228,242,111,33,
130,53,92,106,148,125,125,50,135,75,85,98,
97,5,79,64,226,181,153,50,163,35,2,173,
33,176,26,114,211,221,93,129,162,168,173,34,
135,191,81,202,54,207,153,184,6,85,42,23,
40,141,156,34,228,125,228,170,152,178,85,179,
150,8,43,225,109,206,151,99,64,101,126,10,
211,44,112,189,210,155,186,169,163,186,174,84,
249,249,114,148,13,45,35,116,5,44,132,36,
81,49,231,41,153,22,166,3,123,195,65,42,
150,205,241,92,17,211,82,152,203,167,216,20,
203,146,179,62,80,26,6,166,165,226,85,100,
162,46,149,177,114,141,77,1,47,218,101,120,
246,171,211,69,70,50,173,60,214,59,207,104,
69,166,98,89,184,21,129,42,36,170,150,77,
166,222,160,124,41,17,162,84,13,210,22,187,
184,236,248,189,106,141,141,126,29,247,137,243,
32,73,236,242,123,176,219,226,152,170,202,96,
169,138,97,59,20,29,157,17,195,226,241,176,
15,81,174,50,162,105,68,116,21,127,161,140,
52,49,75,193,213,20,236,123,67,62,252,154,
74,205,178,25,40,52,61,158,170,108,211,234,
118,179,49,172,112,174,80,165,104,152,148,77,
155,146,121,229,250,201,145,6,115,3,111,18,
240,239,184,238,250,126,224,217,22,103,134,87,
196,51,102,189,89,111,99,84,139,215,188,238,
200,166,157,68,207,29,197,180,12,42,203,140,
193,106,212,152,201,254,24,173,175,66,135,199,
69,151,47,136,227,192,43,165,16,159,232,158,
39,168,41,28,201,42,248,20,25,85,146,200,
108,31,162,120,41,163,208,49,97,108,250,109,
148,114,39,178,20,66,145,125,205,212,80,36,
116,127,10,213,221,132,165,140,74,17,165,86,
90,210,173,119,57,50,170,5,50,211,175,18,
142,63,12,142,130,33,101,72,15,190,133,55,
220,131,81,41,52,59,189,94,182,247,46,223,
99,213,252,56,149,250,121,64,198,145,44,108,
169,132,109,24,184,54,94,31,250,145,218,103,
25,63,255,61,168,122,104,105,127,28,179,86,
198,182,76,28,199,193,174,130,211,104,166,215,
46,188,43,147,97,94,62,142,59,148,92,178,
119,91,183,60,73,37,63,78,117,118,20,83,
155,132,72,14,217,119,233,224,49,171,217,222,
131,138,142,218,93,196,54,224,90,246,131,53,
175,32,121,38,153,154,248,9,173,124,227,234,
115,86,41,34,171,174,43,228,173,153,155,100,
119,187,143,87,222,124,143,72,56,64,36,18,
198,90,167,110,195,75,20,133,207,231,35,22,
14,18,217,119,31,247,238,218,66,123,123,27,
0,35,163,99,252,252,133,215,113,71,183,33,
235,94,140,106,9,33,43,87,117,79,83,186,
65,113,186,180,168,58,189,241,149,171,217,149,
94,43,20,112,47,197,29,75,178,66,123,100,
209,141,205,100,254,128,89,46,83,159,174,33,
201,14,157,91,37,220,162,128,34,233,220,183,
167,233,125,188,63,161,19,177,91,192,93,97,
38,27,66,211,83,20,139,39,113,28,139,104,
226,1,84,207,181,11,89,60,145,149,165,120,
10,89,89,192,68,175,70,73,245,11,76,13,
253,156,106,113,136,154,115,1,191,111,59,190,
104,207,98,33,93,173,132,227,89,222,181,84,
213,64,115,44,185,235,11,43,183,86,38,87,
138,17,240,44,77,37,21,239,26,132,158,172,
179,43,87,133,134,129,29,15,97,39,194,212,
44,27,203,244,112,100,58,192,84,165,136,38,
91,108,75,204,176,63,230,231,120,214,96,131,
223,77,193,48,249,67,1,238,139,6,208,133,
96,71,185,138,116,126,20,199,165,97,119,165,
56,151,117,147,43,122,120,106,83,30,69,72,
100,12,153,186,109,179,61,216,180,84,47,150,
170,164,235,151,41,175,171,100,79,101,10,9,
90,124,69,228,108,13,50,5,36,1,106,195,
192,220,208,142,12,248,20,25,125,62,135,215,
235,48,48,223,130,152,185,136,50,56,206,6,
85,1,203,226,3,215,226,147,65,31,149,104,
8,223,192,8,98,62,143,28,11,241,121,221,
5,138,204,249,68,131,183,198,130,108,10,155,
108,14,52,249,100,172,82,231,120,246,42,2,
91,45,163,7,227,205,179,68,114,19,72,142,
140,230,13,45,236,15,219,50,25,59,249,115,
36,108,66,137,123,40,202,127,160,158,61,129,
16,50,154,176,145,80,105,212,66,180,246,125,
113,201,193,55,31,180,202,81,221,215,183,230,
27,129,56,222,209,179,136,101,248,113,122,240,
247,184,251,170,108,14,121,233,185,212,6,126,
44,27,101,111,84,166,92,106,163,181,101,154,
93,33,31,30,69,70,145,36,202,166,197,80,
169,138,34,36,162,46,13,39,225,160,136,73,
242,141,49,198,10,53,84,33,225,85,20,102,
254,127,222,222,44,72,174,252,58,239,252,253,
255,119,205,155,251,90,251,130,2,10,64,163,
129,70,239,100,147,108,146,162,184,136,150,199,
35,141,45,203,182,102,177,39,194,17,19,243,
50,239,122,152,176,222,253,48,79,142,176,103,
108,137,99,75,118,140,165,144,236,160,24,67,
137,100,139,18,187,155,189,177,187,177,47,181,
239,185,239,119,95,230,33,171,11,64,3,141,
70,83,228,220,8,68,161,162,50,51,50,243,
222,123,190,115,206,247,157,239,28,170,140,187,
79,81,174,190,130,146,175,162,89,217,135,22,
247,36,81,136,80,84,226,40,160,179,251,51,
70,222,251,200,213,125,122,107,239,128,30,163,
157,245,169,223,94,227,194,242,203,52,246,191,
143,229,156,38,86,199,24,198,12,197,217,103,
30,184,199,250,189,119,72,206,220,188,119,189,
242,24,125,255,125,71,60,82,136,71,42,82,
87,240,251,3,204,92,133,208,176,8,252,46,
173,163,215,152,59,253,219,116,143,210,56,165,
55,144,153,144,36,4,211,152,162,184,116,17,
169,104,15,221,187,102,190,74,123,223,71,81,
78,161,135,101,188,163,38,190,211,199,25,110,
97,102,103,152,57,251,109,182,175,254,7,146,
56,68,94,222,249,196,247,165,205,251,128,79,
176,158,39,85,168,125,34,200,250,122,10,69,
55,31,138,183,78,224,48,59,83,226,75,175,
124,142,63,250,222,223,208,232,73,246,143,238,
242,194,243,207,253,114,129,2,224,55,126,253,
193,126,89,20,69,188,254,254,77,50,23,190,
254,196,47,90,176,52,50,166,202,200,253,255,
79,145,148,54,20,86,106,233,19,210,186,121,
231,39,4,74,131,133,139,99,196,42,92,94,
200,34,132,246,64,15,124,56,138,88,173,134,
40,42,188,125,0,218,66,147,216,241,8,179,
117,188,155,89,10,229,23,248,219,26,151,55,
182,94,67,79,85,48,205,79,247,115,74,101,
167,88,121,230,127,101,243,230,191,70,89,218,
163,227,126,72,119,63,139,244,167,40,22,191,
72,186,124,138,112,63,139,58,231,60,52,242,
30,19,144,68,33,137,124,178,239,188,86,216,
231,176,179,132,34,35,44,115,64,198,28,160,
207,134,76,153,58,118,38,133,52,116,140,70,
27,245,250,6,205,217,151,137,98,201,87,79,
213,121,191,51,228,153,98,6,145,100,80,251,
99,46,23,39,237,37,63,138,17,76,170,3,
229,206,22,114,191,201,216,42,179,147,93,65,
238,232,84,220,125,126,59,231,146,40,37,154,
97,68,211,245,121,161,148,165,168,171,36,64,
73,87,25,155,58,183,91,121,162,72,227,155,
139,49,253,200,100,107,124,47,91,244,143,201,
242,83,249,24,253,246,17,131,129,79,110,177,
72,248,236,185,251,139,46,170,94,192,13,77,
35,101,140,104,164,76,102,165,164,239,90,28,
154,43,12,103,138,104,194,230,82,180,65,182,
222,6,67,35,201,101,136,83,6,3,67,199,
213,117,122,126,72,204,61,169,152,31,199,159,
12,18,64,50,179,67,107,251,39,164,70,179,
200,212,107,76,233,1,110,51,225,48,76,17,
196,42,81,44,208,228,18,249,153,139,196,161,
71,118,6,190,58,211,103,55,129,51,18,68,
119,72,47,76,120,125,109,72,166,240,121,242,
83,231,249,172,158,6,34,142,208,93,7,231,
161,34,52,198,77,238,242,98,57,195,108,234,
94,122,59,95,232,32,146,8,142,185,164,156,
118,47,20,164,85,101,210,54,140,34,146,251,
172,33,74,186,198,138,34,16,221,1,104,10,
148,7,236,46,95,99,115,107,147,214,208,5,
63,197,236,252,63,194,204,222,11,170,245,181,
215,40,204,92,166,190,251,125,156,102,27,243,
165,14,197,172,74,254,149,61,222,223,173,240,
92,78,229,189,41,143,157,187,191,79,92,104,
50,236,183,120,249,133,152,122,231,61,118,238,
190,141,22,45,147,42,78,33,132,164,84,123,
133,195,157,13,228,226,103,235,195,71,67,133,
240,86,25,179,80,230,204,139,191,137,84,53,
66,207,102,56,186,133,242,212,30,187,107,127,
192,194,226,63,69,235,22,233,173,93,67,44,
109,66,181,203,184,187,77,182,242,112,43,71,
72,133,145,120,27,89,28,17,15,117,148,209,
34,139,23,127,135,250,250,143,72,146,4,169,
106,156,122,238,159,209,61,120,159,126,183,137,
44,126,178,66,45,172,27,20,140,175,124,106,
37,246,200,228,83,55,185,187,115,68,165,92,
132,56,66,45,44,163,15,236,95,74,124,85,
116,93,255,23,191,251,187,191,251,137,15,248,
247,127,246,3,198,229,139,136,251,200,201,56,
244,17,82,62,118,237,95,41,163,19,197,9,
246,104,0,250,147,153,98,137,192,126,226,199,
18,58,160,165,48,84,201,82,197,98,169,98,
161,30,183,57,58,219,239,226,140,15,48,173,
49,151,150,5,115,55,214,233,95,107,144,153,
201,128,161,223,151,177,37,152,41,201,79,91,
125,210,82,39,25,103,208,178,29,150,178,6,
67,205,165,223,185,198,176,125,147,113,111,147,
208,246,72,229,102,62,86,70,143,30,88,138,
242,168,163,222,254,83,156,241,54,105,235,169,
79,125,236,9,79,208,188,78,102,170,207,124,
214,96,170,150,208,182,234,140,199,183,9,91,
48,115,246,219,244,142,62,68,230,124,98,15,
156,119,242,104,179,30,241,80,165,88,126,133,
206,206,135,200,81,25,109,116,26,186,69,34,
181,135,208,31,46,71,165,136,201,166,250,164,
205,1,189,81,149,180,57,64,173,4,19,210,
217,246,200,107,42,153,32,68,217,220,167,95,
94,98,169,218,65,8,65,78,10,178,27,251,
104,237,62,162,213,69,24,58,118,160,162,26,
146,43,45,131,185,172,67,166,144,69,180,7,
104,222,152,170,189,67,165,189,78,102,216,66,
182,122,40,245,54,154,101,160,101,211,44,164,
39,46,163,2,136,18,184,61,28,35,101,132,
20,42,95,152,137,80,165,224,192,241,79,154,
145,205,222,28,159,159,239,114,65,87,57,252,
171,219,100,139,6,226,242,25,212,205,125,146,
108,6,121,212,162,92,111,90,156,223,71,0,
0,32,0,73,68,65,84,17,23,242,252,108,
88,100,42,55,224,210,120,140,28,140,48,113,
169,36,117,170,75,42,235,6,188,230,44,16,
206,231,104,149,211,244,203,5,198,185,12,111,
185,1,251,81,76,99,156,130,216,226,153,202,
4,116,175,116,199,12,195,79,6,96,161,37,
4,73,11,175,221,225,215,141,13,106,238,128,
249,254,33,171,197,128,51,207,22,233,89,67,
188,65,66,161,250,42,86,97,150,230,254,123,
60,19,30,82,146,2,217,233,131,166,146,94,
223,101,53,223,162,157,90,167,117,36,73,23,
150,79,156,16,20,237,211,7,93,101,28,35,
135,29,130,242,131,215,104,123,231,117,180,185,
235,60,93,78,163,12,109,100,103,128,60,106,
161,108,236,161,220,217,65,214,59,200,222,144,
68,215,192,156,0,137,236,143,72,84,5,245,
234,26,194,245,144,107,123,8,18,228,97,19,
229,246,54,178,217,69,30,78,206,103,78,83,
153,234,213,57,245,82,153,65,182,79,103,107,
159,76,238,226,189,77,116,129,207,209,246,255,
139,60,189,133,204,68,84,170,146,87,42,121,
102,82,6,69,83,35,8,83,172,150,34,238,
132,26,142,174,145,158,242,185,80,84,89,42,
152,76,205,56,180,244,13,234,107,183,49,245,
57,204,236,20,74,84,38,104,196,147,185,138,
108,143,71,173,40,73,66,136,118,139,36,182,
6,34,68,41,132,40,51,67,18,171,75,212,
178,72,23,231,105,237,188,133,175,174,33,114,
99,148,146,131,183,15,149,185,87,232,7,63,
70,153,114,144,25,143,222,237,13,50,153,115,
244,27,31,48,238,111,144,46,156,66,8,129,
144,10,189,250,135,104,181,17,83,101,201,112,
16,17,15,21,138,51,207,145,41,158,58,249,
236,169,236,12,170,59,79,80,87,8,195,1,
194,242,31,209,126,210,176,235,77,12,163,134,
158,126,244,62,231,200,119,39,131,127,31,139,
183,106,42,75,63,73,113,117,253,16,57,117,
158,56,112,113,27,155,124,238,249,95,236,230,
200,223,251,189,223,123,252,62,138,215,94,127,
139,113,110,5,245,231,48,173,83,196,196,108,
76,25,155,28,254,18,120,237,180,174,80,173,
88,84,178,198,3,62,44,113,20,162,24,22,
106,246,128,111,174,14,224,205,77,122,35,5,
61,114,25,125,184,143,114,105,153,80,170,164,
210,2,55,22,188,219,234,227,68,49,209,208,
192,55,250,188,82,177,176,163,152,84,198,39,
73,141,248,250,180,142,31,15,249,241,214,117,
198,183,215,31,106,15,124,234,13,108,151,41,
212,158,127,224,57,131,230,45,218,155,239,178,
252,210,63,57,33,203,34,223,166,185,243,19,
114,213,11,8,47,203,216,139,168,26,9,251,
246,113,176,168,56,168,193,143,169,111,238,83,
202,255,29,186,251,223,71,153,235,163,159,25,
129,2,177,217,39,112,70,44,157,253,167,168,
166,69,119,239,10,249,197,167,217,221,248,14,
100,182,62,165,225,119,15,72,198,97,196,74,
38,69,38,10,81,175,175,51,26,6,44,57,
183,72,218,105,196,112,76,193,11,136,15,219,
212,247,134,84,74,38,178,239,240,150,125,150,
124,81,146,46,69,236,218,1,211,89,29,145,
196,147,86,207,71,55,175,169,35,92,159,184,
144,69,147,202,3,155,240,0,10,186,202,87,
107,69,94,59,136,137,68,12,72,138,250,132,
243,24,4,33,65,104,144,53,61,206,231,44,
148,15,238,48,182,67,102,86,103,144,7,77,
226,233,10,40,130,184,82,32,209,85,26,249,
12,195,142,96,69,23,136,173,46,163,190,79,
38,163,66,146,160,95,91,227,11,151,206,48,
183,152,240,250,126,10,59,200,32,72,168,230,
15,16,2,58,131,41,64,242,247,78,143,136,
18,133,205,145,67,221,253,116,15,164,40,175,
34,162,1,107,11,211,12,21,201,115,251,117,
100,189,131,58,118,121,254,226,105,76,243,10,
127,177,254,135,152,229,175,128,229,50,202,166,
201,54,58,132,237,17,31,57,204,123,135,61,
94,204,123,92,171,253,148,173,59,71,84,230,
190,253,25,50,205,228,196,23,42,244,108,218,
251,111,17,49,68,49,239,242,234,92,110,98,
228,120,103,27,49,184,79,236,32,37,241,169,
89,146,116,10,225,250,136,131,137,157,140,114,
123,11,153,77,51,218,235,81,127,247,128,98,
94,163,60,124,180,72,66,110,31,34,2,5,
235,238,46,151,15,122,132,23,18,254,102,237,
59,168,162,68,140,138,211,218,192,124,161,11,
18,76,43,228,197,114,1,229,184,170,95,202,
59,108,182,107,156,202,55,88,206,7,28,218,
62,31,54,21,254,228,86,133,111,44,143,152,
201,192,151,231,243,236,150,122,92,189,253,135,
104,205,167,113,122,77,82,197,26,179,167,127,
147,205,59,255,10,117,181,69,226,8,146,72,
16,247,82,24,238,121,34,39,34,149,201,144,
216,33,140,5,177,217,39,140,6,152,44,83,
61,253,242,228,59,10,123,196,233,17,74,50,
201,84,156,228,54,221,131,41,130,166,64,57,
230,149,245,115,67,220,209,17,110,188,9,243,
123,212,55,18,166,79,127,147,36,137,49,130,
21,178,131,30,233,12,40,75,109,70,219,239,
147,227,194,67,223,79,166,124,138,76,249,20,
187,119,7,36,149,59,15,223,123,86,192,212,
242,55,201,84,127,190,253,236,170,145,70,53,
142,147,235,246,93,158,57,51,255,203,231,40,
62,58,238,110,108,241,214,149,59,12,210,167,
80,255,150,70,105,217,148,194,225,47,104,149,
108,74,147,148,51,6,229,172,78,50,14,49,
63,230,51,227,141,90,236,108,252,1,43,231,
125,190,34,198,216,239,215,137,71,1,41,19,
180,233,2,227,83,203,24,106,130,153,145,56,
227,152,66,94,240,92,152,37,74,18,14,80,
217,194,229,189,142,79,152,36,44,90,38,187,
182,139,174,8,70,65,196,133,105,131,122,246,
6,59,183,218,204,44,252,125,140,76,229,33,
85,201,206,205,127,143,89,170,145,203,60,75,
174,54,241,192,95,190,244,63,29,147,114,77,
220,65,131,78,227,77,252,252,117,204,202,37,
2,167,79,115,255,71,36,74,64,32,246,144,
11,125,14,222,221,65,89,221,69,81,56,105,
187,24,138,100,38,101,240,116,46,205,126,121,
151,107,239,101,38,98,114,64,173,70,199,164,
170,199,160,113,141,252,212,37,118,110,125,135,
168,188,73,114,16,144,49,158,102,232,109,159,
248,75,61,90,84,243,224,223,54,70,14,185,
98,150,220,202,28,173,31,222,198,120,99,29,
43,173,160,42,10,173,182,75,191,231,99,234,
10,149,10,104,113,192,171,151,60,54,250,41,
78,21,135,216,118,149,48,55,68,251,136,111,
176,76,176,93,162,103,207,33,119,142,192,52,
136,171,5,252,80,37,136,21,130,72,33,99,
120,168,50,98,125,16,114,212,157,231,233,153,
3,64,178,59,118,113,142,193,166,61,152,98,
182,188,75,66,137,196,212,169,150,13,152,171,
33,238,236,16,149,242,32,160,225,250,92,73,
4,97,103,136,16,22,121,77,37,206,101,184,
243,55,251,60,247,66,101,146,125,70,49,178,
211,39,153,51,41,103,219,100,227,152,36,17,
108,213,207,99,25,35,138,217,6,47,149,53,
202,27,71,188,63,85,97,199,126,60,81,186,
121,248,20,150,57,194,247,114,60,191,124,23,
87,10,100,146,176,63,59,197,92,16,34,134,
99,76,219,65,233,13,41,171,219,108,141,254,
24,109,201,197,77,18,172,102,15,55,81,113,
91,54,113,156,96,232,146,84,38,226,66,193,
37,42,172,179,121,245,63,83,158,249,246,19,
113,20,247,3,198,209,198,247,40,94,188,193,
32,8,249,250,76,249,164,137,149,228,38,96,
127,82,162,197,49,178,209,3,247,8,49,252,
88,203,98,48,70,74,193,220,76,10,211,146,
143,85,100,197,185,12,194,212,73,103,116,220,
119,111,241,229,23,99,110,105,71,92,61,154,
102,250,212,61,30,236,185,82,22,227,99,178,
233,143,70,107,182,90,83,204,230,59,124,107,
49,160,237,53,216,111,47,82,50,155,24,106,
200,162,101,98,93,8,184,221,127,27,167,23,
225,138,187,28,220,245,152,153,251,71,12,119,
175,147,74,47,34,20,141,72,29,51,20,87,
137,79,95,195,217,155,34,176,123,44,158,249,
231,164,10,15,219,201,87,230,94,69,53,190,
193,193,245,239,145,152,54,153,212,41,10,51,
79,51,238,110,2,29,52,41,8,141,132,225,
214,93,162,244,0,85,75,136,147,201,119,100,
119,247,241,106,111,114,113,41,77,63,8,137,
246,74,168,126,241,83,90,131,169,71,203,116,
98,73,115,235,39,152,185,169,71,42,178,158,
168,181,230,187,120,27,111,224,121,30,175,252,
119,95,255,229,2,69,20,69,252,213,27,111,
179,209,114,241,173,25,180,234,179,252,109,157,
92,6,78,200,90,99,12,214,207,63,148,164,
43,146,114,70,167,156,209,177,140,123,217,213,
163,110,95,69,53,17,35,133,11,217,24,255,
173,29,90,251,14,139,171,121,176,12,228,108,
137,92,85,103,60,140,81,129,84,90,210,30,
103,41,167,39,189,103,71,13,24,171,58,77,
215,39,163,42,12,194,16,33,4,71,142,207,
214,216,229,92,214,98,57,99,114,211,108,177,
190,249,111,208,246,86,201,229,159,191,247,62,
211,69,172,252,2,225,226,207,104,183,239,34,
219,255,132,76,249,244,201,223,91,7,63,33,
40,222,64,89,118,97,115,22,195,152,229,96,
231,255,153,24,212,1,23,243,105,246,108,149,
222,124,11,165,248,96,171,40,167,169,156,205,
90,8,1,243,25,131,235,153,91,96,62,152,
225,9,21,130,160,71,115,231,199,112,102,29,
5,112,182,55,168,78,127,131,65,247,71,48,
253,201,125,210,251,173,52,114,154,74,86,145,
84,118,143,8,130,144,217,255,230,89,162,158,
131,183,121,136,59,114,168,148,116,22,231,172,
201,46,13,41,32,78,184,125,59,128,90,154,
37,203,228,142,11,90,189,5,113,12,170,66,
112,126,133,157,65,149,196,78,17,85,22,1,
129,104,39,152,90,128,34,98,164,140,185,93,
159,229,233,153,93,118,122,37,230,107,183,184,
92,46,146,28,191,23,41,4,237,65,141,114,
174,206,185,156,133,214,232,50,184,117,200,216,
9,201,199,9,113,121,2,18,126,28,179,54,
116,88,201,164,24,135,17,227,177,194,105,41,
80,250,67,138,69,29,226,152,184,82,68,132,
33,162,55,98,169,82,228,174,184,7,148,153,
84,255,196,164,176,104,90,108,46,78,211,26,
126,250,244,115,218,28,16,39,10,181,140,195,
75,165,44,250,79,39,51,44,201,84,153,100,
182,74,84,61,139,250,227,247,184,115,119,200,
153,213,3,14,206,204,65,2,137,166,48,30,
4,88,89,193,214,161,131,239,199,44,46,89,
16,197,40,55,55,185,248,204,42,169,151,91,
92,189,241,135,24,157,51,84,230,190,134,158,
42,60,190,46,148,202,49,8,8,76,79,3,
3,14,108,143,57,107,146,80,69,103,151,136,
103,171,8,199,71,116,250,8,219,69,52,59,
143,124,181,126,223,39,147,209,62,221,157,57,
73,136,167,43,36,167,74,136,217,42,217,171,
107,176,190,203,23,117,21,187,50,67,39,48,
49,52,23,75,145,24,82,112,224,120,76,155,
250,9,87,216,25,103,136,99,201,98,169,69,
99,152,71,81,98,102,114,93,106,115,135,220,
56,154,231,194,244,228,254,168,24,26,86,57,
199,15,157,14,209,213,211,84,206,191,138,145,
45,145,202,77,17,71,33,66,74,156,110,157,
176,17,19,95,63,139,172,116,57,255,138,96,
239,230,31,163,180,207,145,45,173,146,46,126,
108,222,68,168,132,153,29,132,91,164,182,244,
85,14,214,254,12,229,220,21,78,101,83,156,
201,90,180,188,128,91,195,27,84,107,49,71,
13,133,108,126,66,174,143,122,107,168,75,49,
53,83,99,24,132,16,75,22,158,250,45,2,
231,225,170,43,142,34,252,113,155,169,197,111,
177,119,184,142,50,243,96,214,172,214,2,168,
109,177,127,231,63,178,184,250,207,32,73,62,
211,98,184,143,206,65,40,13,212,180,197,120,
108,99,24,191,56,169,118,183,215,187,199,81,
84,47,124,145,119,111,237,176,231,152,232,181,
213,79,125,163,79,194,81,192,241,60,197,120,
128,195,147,13,239,124,156,163,152,43,154,172,
78,165,201,91,218,67,171,78,31,53,160,102,
247,119,9,148,219,44,55,119,80,252,128,106,
65,162,38,17,172,76,54,132,37,89,139,48,
72,80,213,137,129,209,209,160,128,166,134,232,
74,68,198,18,68,182,198,124,126,66,174,118,
253,144,32,142,57,116,124,52,41,209,20,73,
197,208,72,0,207,244,81,203,13,90,71,27,
164,205,115,168,250,4,8,135,157,53,226,226,
62,73,171,128,33,103,49,179,211,184,131,22,
27,239,255,159,147,30,109,126,146,101,219,141,
42,177,121,7,185,52,177,202,40,27,26,167,
210,38,186,34,56,61,171,114,224,120,100,53,
149,32,142,201,168,10,203,233,20,69,253,222,
119,189,37,251,196,70,252,144,212,174,156,253,
54,113,232,18,228,142,61,185,6,5,220,225,
33,204,239,60,254,60,105,46,125,187,76,214,
116,248,218,116,137,249,157,3,146,148,193,206,
108,149,234,81,19,57,83,70,59,51,139,37,
98,180,233,34,201,210,12,178,209,33,153,169,
114,125,63,197,82,37,164,122,161,68,99,52,
203,243,179,29,180,159,221,102,223,155,166,237,
22,104,231,22,56,101,223,36,191,164,83,180,
198,147,127,198,144,156,93,39,99,6,88,86,
68,37,51,228,202,193,18,45,223,230,243,69,
143,124,146,112,117,236,114,163,63,38,136,19,
28,47,75,41,61,228,92,218,68,251,224,46,
119,111,244,64,8,74,254,144,228,252,50,40,
146,237,177,75,213,156,136,21,14,29,159,214,
168,68,33,51,36,83,239,177,183,222,39,138,
18,50,207,157,162,53,83,163,61,85,226,102,
16,49,8,238,241,14,142,151,193,50,71,72,
96,69,145,220,114,188,19,137,232,39,29,35,
39,79,198,8,113,3,157,231,203,3,190,187,
81,100,249,156,134,213,232,34,91,61,100,171,
79,82,201,35,27,61,100,28,115,180,213,231,
133,211,101,22,20,73,121,191,137,33,194,73,
27,70,192,96,24,96,219,17,166,169,32,18,
80,45,131,219,170,202,234,156,130,151,63,164,
190,182,137,149,58,141,212,140,71,14,150,42,
97,128,217,220,199,157,90,32,87,189,64,123,
39,196,22,187,44,23,117,82,234,113,130,229,
7,8,77,37,201,90,160,107,196,51,21,132,
42,17,189,225,137,232,240,36,64,116,124,12,
67,162,62,129,167,147,95,44,32,242,25,214,
187,115,68,51,53,82,21,21,101,231,136,217,
105,193,95,53,86,41,100,90,156,207,167,249,
160,51,34,165,42,84,205,123,156,203,108,161,
75,41,61,66,145,49,185,148,67,228,196,164,
252,1,178,63,98,184,237,81,110,174,161,236,
55,16,221,1,90,214,226,230,122,26,45,174,
161,167,74,56,131,67,134,205,59,180,91,127,
201,168,177,137,34,115,196,202,152,204,234,45,
190,124,81,99,46,109,48,59,27,112,96,220,
161,215,93,35,234,73,136,161,125,248,58,163,
254,26,195,225,21,66,111,132,153,81,136,93,
147,177,245,55,156,159,54,56,155,179,144,199,
251,227,151,171,58,115,150,193,78,215,37,37,
62,135,59,58,98,48,126,11,189,60,230,92,
46,141,19,199,212,131,33,162,87,67,51,11,
15,197,197,131,245,255,66,103,240,35,74,213,
47,97,119,14,73,10,143,118,139,150,165,33,
189,221,107,244,214,183,201,79,95,120,160,229,
248,73,28,197,177,222,154,246,149,191,68,36,
49,225,184,203,254,225,33,207,93,122,250,23,
6,20,127,241,215,239,240,159,254,195,239,79,
138,134,164,250,212,68,106,214,249,5,187,16,
10,152,45,152,88,34,197,94,247,179,121,211,
100,76,149,185,226,147,35,107,18,71,52,15,
126,192,51,149,38,225,238,16,5,31,153,209,
38,8,125,123,155,232,233,201,196,117,42,45,
177,71,49,99,45,68,53,58,188,83,87,248,
194,220,36,232,150,245,132,33,176,51,118,185,
63,145,26,4,33,227,65,136,46,4,11,105,
147,41,83,103,103,236,50,88,238,51,216,187,
78,245,212,100,238,33,72,14,137,7,10,181,
252,111,146,46,79,178,151,253,205,63,194,124,
190,117,162,227,235,142,170,164,87,14,144,218,
189,154,40,171,41,92,239,219,116,188,128,132,
132,4,88,78,155,92,237,141,152,54,117,186,
126,64,89,73,145,110,28,178,93,206,227,62,
98,173,172,232,206,98,157,155,163,219,188,223,
146,58,66,160,19,123,226,177,173,39,85,9,
136,99,73,24,39,28,216,30,181,66,142,215,
17,60,19,199,200,70,23,217,30,16,62,189,
66,116,102,1,17,71,40,183,38,3,120,141,
76,141,45,169,242,212,194,144,88,132,168,114,
18,88,227,172,197,92,124,124,45,237,29,146,
88,6,31,133,100,49,24,163,188,127,27,36,
68,175,76,50,52,197,113,56,85,221,96,217,
247,40,221,222,97,231,204,2,166,156,144,158,
239,118,38,75,147,46,23,51,100,59,125,236,
142,77,20,39,140,134,1,163,142,67,246,168,
77,188,48,197,169,116,138,3,199,227,70,127,
196,171,138,96,121,170,201,245,118,141,197,87,
210,60,61,91,102,240,225,14,205,31,223,34,
245,229,243,52,77,131,170,161,81,51,117,246,
108,151,59,141,41,42,185,201,228,119,90,83,
201,180,251,84,74,121,122,126,248,200,139,186,
121,236,122,107,168,33,95,175,12,120,223,72,
240,199,42,23,179,14,55,92,201,171,31,137,
29,220,16,237,221,155,160,72,10,101,19,219,
14,177,214,247,176,196,71,188,141,65,92,46,
48,37,26,84,42,6,71,71,14,205,166,135,
101,133,76,135,33,2,216,28,185,12,131,16,
101,101,159,253,198,191,34,222,54,153,153,249,
109,172,226,252,199,58,24,10,174,117,47,113,
170,46,255,10,135,119,29,250,181,171,20,116,
21,37,142,81,110,109,78,246,160,199,9,98,
228,64,28,79,72,108,41,33,142,9,159,59,
143,122,123,27,198,14,83,51,230,35,167,228,
31,44,167,82,140,231,151,88,31,207,99,222,
116,57,91,190,206,238,110,25,119,181,68,122,
170,76,88,43,161,53,61,218,131,105,222,116,
109,92,127,10,63,236,115,58,147,160,221,79,
44,58,30,234,218,46,118,211,166,217,52,200,
22,71,152,106,204,153,68,158,120,124,137,62,
196,179,53,236,124,133,133,149,235,40,206,77,
178,166,164,148,213,40,234,42,123,227,22,87,
126,216,192,200,21,57,85,86,73,29,47,81,
74,171,10,175,214,10,188,46,250,140,236,63,
103,224,74,148,197,248,164,165,251,140,153,102,
247,112,76,123,235,22,242,162,199,240,19,52,
11,169,162,160,125,243,39,100,115,151,72,170,
135,132,241,196,138,63,173,40,200,76,196,168,
125,13,43,255,240,132,124,54,191,74,220,242,
81,20,21,53,169,226,197,60,146,128,71,128,
178,216,37,174,116,233,238,191,79,121,233,229,
39,138,123,129,51,34,37,60,62,247,236,69,
20,85,227,224,240,128,56,142,145,242,111,111,
218,104,219,54,251,35,241,201,28,197,47,250,
152,45,154,132,73,204,81,239,201,23,163,20,
211,159,77,152,58,106,173,147,82,186,148,118,
15,24,181,135,12,162,4,71,209,41,124,225,
28,152,6,201,113,86,37,4,216,1,220,176,
199,92,46,74,252,68,229,79,214,4,223,90,
10,49,133,114,50,193,235,247,21,148,220,125,
100,44,96,71,49,175,213,187,44,88,230,100,
144,76,64,124,44,72,76,162,144,36,138,145,
86,68,123,243,39,164,203,75,4,118,15,181,
56,38,171,43,12,143,7,201,162,72,69,215,
30,108,156,109,141,38,191,87,13,141,166,23,
160,73,193,173,193,120,226,151,20,199,172,213,
23,153,91,24,144,221,218,167,156,121,24,60,
99,79,144,181,158,197,183,187,248,230,230,201,
192,93,196,16,141,105,226,122,14,185,248,201,
222,56,81,172,162,200,136,213,172,133,38,5,
239,40,42,231,50,38,149,227,32,141,31,160,
190,127,155,68,85,17,247,169,127,52,17,113,
249,178,32,174,21,81,111,109,162,153,105,156,
156,70,54,140,8,163,24,145,8,196,98,121,
210,195,118,61,208,117,148,27,155,19,217,101,
202,154,156,12,63,64,249,240,46,37,69,146,
168,10,241,83,203,204,89,169,19,98,221,84,
36,73,34,112,194,8,213,243,9,252,251,44,
76,130,132,36,61,121,172,28,219,24,9,172,
102,45,82,166,206,130,231,179,57,54,64,140,
97,121,134,124,57,79,120,117,131,163,239,93,
225,236,239,124,17,180,73,229,120,237,168,202,
66,161,69,197,84,232,135,48,10,67,118,43,
5,118,71,143,150,25,78,246,104,236,243,57,
17,179,47,37,185,157,14,47,155,6,255,149,
179,252,189,74,139,63,218,91,226,244,162,205,
84,163,205,230,219,187,36,9,44,204,165,73,
149,76,166,255,193,231,72,110,108,32,6,99,
226,233,50,98,104,19,90,38,106,57,143,116,
60,230,102,39,55,101,82,41,48,152,169,98,
132,49,103,12,141,110,108,48,109,233,116,10,
30,195,200,99,247,131,171,196,97,128,102,230,
79,60,140,100,28,97,141,6,15,180,100,107,
75,95,227,195,15,238,114,253,116,139,146,161,
241,133,103,206,222,127,210,17,67,27,97,59,
40,183,182,136,107,37,146,66,150,224,153,85,
68,16,34,226,24,174,172,193,241,249,78,202,
121,68,127,52,1,26,32,58,191,204,122,180,
138,209,108,114,182,249,99,44,37,128,58,24,
115,89,104,142,8,207,45,99,29,182,40,153,
57,20,99,200,130,165,81,74,13,184,89,159,
230,70,239,144,203,165,123,160,38,59,125,222,
252,153,96,228,23,80,100,194,192,83,249,252,
124,7,114,38,78,181,132,146,36,136,40,226,
205,16,98,98,190,84,123,184,5,183,146,75,
177,253,185,6,227,183,115,8,87,192,125,212,
170,46,37,85,67,99,28,70,40,199,10,192,
103,139,89,102,45,29,137,32,214,92,26,59,
29,180,182,206,249,217,71,171,46,107,166,78,
119,225,26,237,214,6,74,1,34,15,246,198,
30,165,227,74,63,252,216,26,212,147,86,110,
237,34,185,218,69,146,36,102,212,59,164,151,
159,35,183,27,96,60,213,120,192,54,228,228,
94,30,24,140,154,187,248,163,62,249,217,11,
88,197,185,79,73,198,37,249,124,1,169,40,
92,126,250,28,159,127,241,23,183,185,241,251,
127,253,54,202,49,215,170,232,186,254,47,190,
245,223,255,111,79,44,247,252,44,173,167,251,
95,51,173,171,28,14,220,39,110,61,45,87,
172,199,174,86,188,191,245,228,219,93,24,253,
17,191,170,111,178,113,187,207,169,69,147,114,
217,192,44,89,96,232,36,165,28,72,73,146,
64,215,15,184,209,180,24,202,1,117,215,199,
210,61,146,160,204,179,213,24,207,137,233,19,
210,232,134,112,231,60,201,72,69,20,71,8,
49,1,10,33,224,233,124,26,41,38,158,68,
97,4,241,240,34,86,126,14,33,37,233,204,
42,189,250,85,196,116,157,222,222,117,6,222,
219,100,23,29,236,48,62,33,178,198,94,142,
180,57,124,20,39,200,108,202,160,227,135,76,
155,6,253,32,36,12,77,214,26,179,60,51,
83,231,244,214,58,241,165,85,14,17,212,157,
128,112,171,4,195,60,113,43,75,202,121,134,
218,210,151,57,216,248,83,196,210,61,63,46,
145,31,19,102,119,81,138,143,7,104,63,48,
137,19,137,169,187,172,13,29,94,46,231,72,
16,100,215,118,16,199,230,121,137,162,16,189,
112,30,121,172,140,65,8,82,174,141,117,110,
106,18,236,239,108,83,24,31,114,71,121,134,
106,251,14,123,155,35,164,2,6,49,253,221,
128,244,254,54,162,209,65,28,147,195,194,15,
192,245,81,239,238,34,124,127,242,187,161,19,
207,213,16,8,16,2,93,74,218,94,64,125,
148,98,54,235,82,113,61,228,96,72,171,57,
249,60,66,17,20,43,38,232,26,234,59,55,
72,11,65,126,182,130,114,115,11,177,113,128,
57,53,69,172,106,24,106,136,236,15,209,15,
26,108,109,142,48,250,125,204,229,42,215,134,
14,251,67,147,111,29,220,96,169,55,64,155,
169,208,244,67,246,93,159,71,25,32,219,94,
134,48,214,248,186,53,98,102,251,128,130,166,
162,122,1,158,161,113,35,172,241,210,120,139,
124,201,231,167,246,12,215,181,25,106,165,57,
20,153,167,47,171,148,205,1,241,202,28,202,
230,1,209,153,5,146,108,154,164,144,225,79,
189,34,202,92,154,98,60,9,134,73,165,128,
108,116,208,253,128,25,199,37,115,212,162,28,
133,24,157,1,197,78,31,45,157,226,250,213,
30,149,213,159,209,239,189,135,63,40,144,202,
77,33,142,239,137,168,112,111,134,65,42,42,
233,212,57,188,29,201,160,225,161,148,71,228,
52,117,210,238,146,2,76,157,36,155,158,72,
198,115,25,18,93,1,67,159,216,175,127,112,
7,17,132,147,11,83,64,146,50,113,159,61,
135,226,251,12,102,103,121,51,120,142,11,233,
109,170,91,87,8,34,129,166,76,190,176,221,
120,145,217,206,13,148,221,58,221,27,135,164,
87,106,236,5,146,211,133,128,51,217,20,203,
5,143,31,110,101,209,245,1,37,99,210,42,
76,114,105,230,151,84,78,89,61,74,186,71,
170,100,176,95,169,240,150,243,18,215,215,203,
28,21,92,152,203,48,118,42,44,101,85,106,
105,255,17,114,111,65,201,208,216,218,12,168,
76,5,84,114,15,71,225,186,235,159,232,251,
138,134,134,46,36,134,34,241,227,152,119,70,
89,94,93,245,40,155,143,78,80,45,85,97,
211,113,144,217,0,18,240,222,45,49,8,5,
73,214,165,183,159,34,167,189,130,145,170,126,
98,92,60,184,251,103,176,114,141,180,49,164,
40,191,5,157,60,65,216,68,166,31,108,111,
38,1,80,105,18,136,35,134,157,91,36,182,
130,97,85,62,177,245,164,104,58,163,250,22,
223,248,210,75,100,179,217,95,24,72,220,217,
216,228,74,29,148,84,158,239,254,187,127,249,
139,1,10,223,238,161,104,230,99,129,66,74,
65,107,232,63,214,44,240,126,160,88,172,88,
143,29,57,186,31,40,90,59,127,194,215,204,
107,200,222,144,116,214,36,165,29,111,144,11,
66,226,149,57,2,117,66,138,222,24,140,105,
122,62,253,113,134,88,27,112,212,153,195,243,
74,152,70,151,213,130,66,224,195,64,132,116,
226,128,36,219,66,29,174,96,56,23,8,122,
30,177,112,112,213,144,3,219,227,208,241,241,
226,132,246,205,12,149,217,111,209,222,121,3,
18,65,28,133,140,149,119,145,185,16,81,24,
33,179,30,65,156,60,160,118,24,185,121,50,
230,195,217,135,46,37,45,71,227,168,55,141,
84,28,14,123,51,248,161,193,175,204,28,177,
144,51,81,43,5,222,31,166,248,171,157,50,
182,157,33,229,184,164,196,25,138,197,47,83,
154,127,129,222,209,85,156,226,235,15,205,76,
136,39,168,64,133,72,232,14,107,196,114,140,
34,35,134,65,68,78,83,200,84,11,19,53,
140,34,17,81,68,188,50,71,98,26,136,222,
8,44,99,66,98,22,178,136,222,8,97,59,
36,217,52,85,123,155,109,237,60,167,204,54,
70,74,34,163,136,110,152,199,137,82,100,226,
1,73,38,69,60,91,35,94,168,33,15,154,
19,128,248,72,65,52,247,50,145,97,97,121,
189,73,144,210,53,252,56,161,229,187,116,29,
147,249,197,52,233,48,192,62,26,226,249,49,
82,8,106,70,132,232,143,16,65,72,146,179,
72,42,5,146,82,142,96,183,69,65,14,217,
146,171,84,59,107,136,177,3,253,49,141,134,
75,191,233,160,236,55,88,78,43,56,185,37,
210,215,175,208,220,233,179,172,71,148,79,205,
176,99,63,8,172,253,113,153,161,93,68,85,
18,74,217,38,79,217,14,122,187,143,230,249,
40,99,135,102,165,72,35,41,241,116,127,147,
242,112,200,194,188,202,243,237,53,140,173,219,
148,180,30,211,169,30,34,142,137,167,170,196,
153,52,202,126,29,185,223,32,169,20,89,71,
112,171,89,96,118,89,103,52,83,97,87,215,
40,171,10,137,148,184,138,70,107,182,66,162,
171,19,147,69,199,69,76,149,216,141,61,190,
120,38,77,181,162,176,126,112,29,45,89,193,
82,12,114,123,107,216,211,147,246,71,107,247,
77,154,205,63,39,28,5,84,151,191,204,222,
235,63,162,113,168,208,200,141,168,165,117,116,
41,39,109,192,181,189,147,185,136,164,86,2,
67,3,69,65,118,250,8,215,35,92,158,225,
246,169,121,218,122,133,63,63,204,241,78,184,
202,141,192,96,165,182,205,76,41,61,1,168,
246,24,77,73,216,116,23,89,138,214,193,119,
145,73,130,231,198,204,46,152,108,134,243,60,
93,82,49,212,8,69,128,23,71,220,234,10,
118,221,46,49,9,53,215,135,116,138,59,106,
150,15,70,103,104,168,207,19,164,190,65,161,
246,121,6,205,31,50,93,149,236,173,43,124,
229,233,4,5,131,214,56,71,62,245,112,197,
151,82,37,83,75,1,134,46,72,235,15,202,
138,211,170,130,34,196,201,244,127,203,11,216,
25,185,56,81,204,110,167,198,108,169,201,197,
242,131,49,204,141,98,164,128,142,31,98,169,
10,59,99,15,231,70,14,181,191,2,249,54,
154,243,20,221,102,204,84,233,183,200,85,207,
127,98,2,29,135,62,237,254,247,81,42,147,
133,70,113,35,143,149,89,198,222,237,33,74,
125,18,79,33,222,156,66,86,70,200,84,130,
48,98,100,54,36,217,175,49,117,234,87,72,
226,228,19,129,34,24,182,153,75,185,12,250,
125,150,151,22,127,110,96,136,227,24,215,117,
233,118,187,196,113,204,119,223,184,129,172,76,
196,56,79,12,20,73,20,210,185,245,31,209,
140,2,138,145,39,14,125,146,36,70,170,42,
129,211,198,219,251,183,28,108,92,33,149,89,
122,136,96,190,255,53,247,187,46,143,91,189,
240,17,80,8,248,84,126,226,30,80,36,52,
175,253,37,103,178,93,220,81,68,70,15,31,
108,252,169,10,237,140,197,149,222,136,182,23,
208,15,66,68,144,161,229,164,201,103,90,164,
82,109,158,45,107,228,52,149,48,76,72,4,
28,186,30,82,135,216,108,161,142,167,152,63,
253,143,209,221,51,68,71,22,241,32,71,210,
179,8,90,75,100,82,159,167,215,122,7,175,
242,14,99,239,38,246,112,13,101,250,193,106,
33,165,72,132,56,113,142,192,241,50,39,21,
197,208,41,208,29,214,176,189,12,99,63,133,
23,164,40,152,54,237,113,150,211,213,58,151,
115,46,11,82,112,221,11,73,233,42,175,237,
166,152,41,109,147,73,15,80,74,30,113,110,
159,97,99,141,97,247,22,118,189,137,152,106,
35,212,207,110,86,38,69,140,169,59,12,237,
34,150,49,34,1,86,115,22,158,162,160,206,
213,136,23,166,224,88,73,145,84,139,196,243,
53,146,153,10,73,202,128,40,134,124,6,217,
25,66,20,227,165,242,164,7,117,244,200,69,
32,72,166,203,100,189,22,7,222,52,37,181,
71,82,72,67,214,2,43,133,210,234,145,228,
50,39,64,148,155,85,25,106,69,110,247,150,
153,170,216,8,1,65,28,227,38,46,7,253,
60,79,149,67,76,219,193,176,199,180,59,62,
105,75,37,151,87,9,134,147,76,81,145,2,
252,16,217,29,224,13,125,104,246,24,12,83,
212,90,119,16,131,49,107,107,67,28,39,162,
80,208,209,84,1,174,79,84,94,97,126,62,
166,191,222,196,238,218,148,102,242,108,222,55,
89,50,118,179,199,164,107,155,103,202,146,11,
249,52,86,189,131,28,59,136,48,130,4,26,
211,85,236,56,207,138,49,68,54,187,152,237,
30,245,187,29,14,15,93,20,69,224,56,17,
214,98,153,235,218,101,106,237,53,72,27,32,
36,178,222,70,157,205,177,49,204,97,154,125,
146,100,162,230,123,87,72,70,133,44,215,239,
212,216,14,2,206,167,67,148,56,33,94,93,
36,144,146,11,217,8,37,137,177,90,125,42,
167,179,220,184,177,70,174,244,60,65,28,17,
231,202,212,215,127,136,58,245,215,164,166,71,
216,135,6,186,94,65,79,103,145,83,91,188,
176,156,162,212,31,161,174,237,18,47,205,64,
33,67,108,26,160,107,36,179,21,72,18,180,
55,175,32,130,0,111,166,194,247,213,89,182,
7,83,248,82,229,139,229,38,95,10,214,217,
18,11,124,121,46,66,142,109,228,198,62,161,
23,163,41,9,65,162,145,140,218,52,91,30,
133,188,78,42,165,32,109,23,95,59,69,193,
105,98,166,18,148,157,35,204,156,228,90,55,
207,87,141,62,43,7,77,148,245,61,100,189,
205,190,145,167,217,74,19,138,58,227,250,30,
133,233,203,168,202,44,210,252,34,206,232,128,
211,139,46,150,238,211,177,51,20,173,71,207,
117,164,84,249,16,72,156,100,245,109,159,214,
78,138,100,104,17,108,88,4,187,22,189,64,
167,163,192,215,23,146,147,129,93,128,142,31,
240,218,118,143,45,199,101,199,113,217,24,185,
68,73,130,176,66,210,201,231,9,187,30,181,
229,95,33,147,59,71,42,63,243,9,157,150,
201,144,134,144,10,118,187,78,82,60,36,140,
52,250,245,33,83,139,87,120,241,82,72,59,
14,241,55,231,152,93,250,135,180,63,184,139,
44,122,136,227,68,55,137,37,57,235,57,72,
248,68,160,16,138,202,222,205,247,120,238,226,
89,106,181,234,207,13,20,63,121,235,61,110,
223,185,203,81,179,205,95,190,117,29,117,254,
185,147,25,175,239,254,187,127,249,233,28,69,
18,133,52,174,254,103,50,131,171,152,222,53,
252,11,255,59,221,235,223,65,87,7,12,109,
157,149,243,9,47,159,238,179,171,70,172,31,
212,31,88,35,248,113,189,244,147,90,143,203,
207,224,220,26,216,67,158,59,55,36,220,239,
83,111,68,152,179,10,170,166,144,20,50,136,
238,128,104,113,154,131,145,75,219,11,136,26,
105,164,155,35,112,117,148,234,8,77,245,17,
66,80,59,86,97,164,44,73,52,80,176,148,
137,97,157,76,71,184,188,69,239,112,129,194,
204,165,7,28,30,227,208,103,243,198,255,133,
118,254,112,226,130,144,13,184,127,105,251,253,
131,100,3,63,36,56,174,43,196,241,207,70,
111,14,93,245,152,46,237,96,123,25,52,197,
103,57,39,232,249,33,47,21,109,22,218,67,
228,200,230,234,169,121,92,63,224,90,111,76,
146,124,108,166,69,130,186,220,7,250,72,230,
137,15,203,112,234,231,219,193,173,169,30,65,
164,157,100,82,63,174,247,56,155,179,200,28,
115,34,241,76,249,126,246,123,162,12,139,33,
74,18,60,199,99,201,247,113,58,49,253,200,
99,86,239,158,224,116,120,102,1,53,74,56,
163,238,115,203,191,200,83,173,107,196,86,138,
120,170,68,180,56,67,60,91,153,72,105,143,
7,192,170,12,200,153,35,110,55,75,156,175,
181,169,154,58,65,146,112,183,41,216,24,57,
60,179,56,67,166,221,103,222,142,216,219,179,
177,111,132,168,170,100,105,57,141,62,180,81,
134,54,195,65,200,216,153,40,138,68,60,166,
109,123,52,26,30,206,241,146,171,110,215,39,
157,182,104,63,245,101,150,219,239,51,218,62,
100,249,43,167,105,188,183,195,232,253,77,190,
245,245,203,252,168,51,164,231,26,120,65,138,
90,174,201,217,92,154,154,162,144,218,60,160,
125,237,128,74,73,71,30,43,30,230,5,4,
137,138,104,116,65,8,186,13,155,131,3,135,
133,5,139,84,74,65,164,12,174,77,173,178,
233,182,185,104,40,188,85,200,49,51,91,163,
234,123,236,6,9,82,70,44,167,83,220,30,
140,233,7,17,113,2,71,142,71,162,77,147,
116,37,71,51,61,122,90,72,212,29,17,37,
9,47,168,18,55,72,216,10,114,92,253,139,
136,76,110,1,61,142,49,15,214,57,240,127,
192,51,23,70,212,82,6,210,15,25,189,122,
196,206,254,31,161,205,198,84,166,39,138,35,
229,198,6,225,115,231,240,162,152,171,67,135,
72,85,81,43,69,94,56,22,27,36,41,131,
224,204,34,255,105,175,198,133,180,199,175,45,
15,16,251,77,236,159,110,161,155,112,186,214,
102,107,148,112,42,245,160,12,179,162,182,104,
56,17,11,115,199,254,88,254,28,118,237,20,
103,211,27,33,151,85,169,0,0,32,0,73,
68,65,84,152,225,24,249,78,19,130,144,242,
216,33,76,102,168,27,6,51,134,78,244,244,
10,202,237,29,158,17,3,230,190,148,226,181,
215,76,10,83,207,162,104,6,190,215,197,107,
30,226,218,99,188,56,198,144,146,230,40,71,
37,61,36,107,126,54,129,204,158,91,35,30,
69,68,61,15,53,173,35,75,33,74,236,240,
234,114,128,161,220,83,98,181,199,1,111,188,
47,73,186,103,17,249,26,97,124,7,57,215,
68,90,144,52,43,40,134,130,158,45,62,214,
187,109,231,218,31,18,167,142,48,101,134,152,
52,129,214,65,69,80,111,204,178,120,97,135,
203,211,69,172,181,29,86,10,171,188,227,135,
168,154,197,202,179,255,156,221,237,127,11,43,
77,220,159,149,208,115,22,142,126,128,153,254,
100,63,57,169,234,152,51,231,120,253,173,119,
185,240,212,249,159,139,196,14,130,128,245,131,
46,67,99,154,184,223,198,90,253,210,147,13,
220,221,127,116,118,222,196,235,221,97,110,74,
80,54,35,174,188,251,127,240,43,249,59,72,
145,16,157,171,98,138,16,186,49,243,75,25,
182,213,63,103,216,76,145,173,62,236,143,18,
126,6,87,67,241,25,62,171,221,124,141,74,
210,99,119,215,35,155,98,50,69,30,199,147,
94,183,101,34,162,136,174,31,144,56,10,150,
247,60,83,43,223,4,18,118,238,124,7,138,
45,10,247,121,220,56,227,152,150,226,99,223,
55,81,140,18,35,228,71,118,4,30,82,42,
8,69,165,127,116,13,101,229,240,83,93,201,
14,29,159,156,166,50,171,107,28,56,30,154,
234,83,239,46,80,200,180,48,180,201,133,110,
25,163,227,184,106,112,65,215,152,187,177,14,
2,188,51,139,52,189,224,68,198,153,36,143,
1,208,185,61,162,125,3,133,159,255,16,15,
128,53,24,82,18,39,147,255,15,195,136,172,
250,224,171,23,117,149,119,218,3,58,126,128,
121,126,153,218,250,30,113,125,204,29,103,133,
176,88,130,153,34,73,55,5,249,25,102,79,
245,89,106,214,185,22,127,147,243,165,45,80,
213,9,72,28,3,196,3,228,186,8,184,211,
79,56,95,131,189,227,54,80,206,234,114,173,
153,194,84,60,206,190,244,52,53,253,14,221,
222,46,227,81,72,46,171,97,153,10,190,23,
179,177,49,34,138,19,116,77,226,56,17,102,
38,65,81,4,182,253,160,156,101,56,12,136,
14,6,184,245,3,182,54,199,148,122,27,84,
230,50,212,119,134,228,214,118,113,242,57,134,
118,129,197,82,131,47,84,139,16,132,164,223,
191,69,123,167,207,238,206,152,66,94,67,87,
4,40,18,245,160,133,98,77,51,58,115,150,
236,237,155,236,237,217,72,41,176,237,136,210,
139,203,180,102,171,188,177,158,225,27,167,219,
172,39,85,58,131,49,45,63,100,224,102,177,
140,193,164,245,231,7,244,252,240,120,126,68,
161,164,234,236,198,26,83,51,127,159,245,253,
58,222,184,131,157,121,147,175,174,8,228,110,
157,209,210,44,219,219,47,113,234,210,75,104,
102,150,48,10,233,205,174,16,58,91,76,229,
244,137,120,192,50,72,171,10,79,45,43,40,
107,187,136,77,15,49,118,136,75,121,252,36,
225,141,86,159,130,174,242,98,49,75,156,36,
36,9,8,67,195,185,184,202,107,123,69,150,
203,251,60,91,157,84,84,178,219,199,16,49,
174,155,240,108,116,151,31,119,158,231,78,44,
121,118,241,57,74,215,223,153,4,226,113,149,
174,145,166,51,74,8,92,143,121,235,128,133,
246,33,52,38,51,53,31,17,225,106,171,199,
63,152,223,226,135,135,75,216,229,50,47,79,
133,52,50,105,218,174,202,237,107,58,83,243,
127,23,123,188,193,198,219,191,15,36,156,123,
245,136,198,77,80,196,4,128,94,94,90,251,
204,215,246,102,187,198,139,53,155,194,178,67,
219,11,112,99,7,67,74,106,166,254,192,117,
159,0,7,253,128,124,78,16,23,246,8,131,
58,250,48,135,123,117,30,161,8,102,158,250,
59,12,219,215,176,131,91,236,222,24,179,240,
212,111,63,210,105,90,181,76,46,94,10,88,
160,75,146,113,137,73,24,250,121,126,120,103,
64,98,95,226,53,251,38,223,58,187,76,216,
44,115,233,197,13,118,239,190,193,212,242,215,
80,236,57,188,155,62,83,139,191,70,110,250,
28,66,72,252,81,239,241,226,179,233,51,20,
250,254,207,5,18,63,125,239,3,222,186,185,
143,177,112,25,67,213,225,17,195,137,79,4,
20,197,197,47,144,159,125,145,250,254,45,90,
237,255,194,203,165,61,82,241,132,208,137,14,
14,144,199,229,90,82,206,241,185,66,196,15,
182,111,61,18,40,38,101,221,227,215,197,72,
1,133,140,254,153,236,208,146,222,58,86,122,
76,54,175,83,203,65,152,8,148,108,106,34,
61,60,59,233,215,22,157,128,97,125,134,169,
51,223,188,247,60,233,35,142,75,204,97,24,
82,210,53,20,69,144,142,31,12,134,210,4,
175,217,132,228,60,59,183,127,31,169,170,84,
107,223,38,63,125,145,222,237,119,224,220,167,
175,31,92,74,155,44,166,77,78,249,38,141,
172,195,221,225,163,77,205,166,77,131,105,199,
65,36,9,206,11,23,216,12,66,6,199,38,
95,94,96,146,54,7,159,28,228,85,208,151,
60,254,118,71,66,78,83,241,226,24,5,129,
42,38,3,82,51,41,157,187,3,155,179,57,
11,47,138,41,27,26,253,32,36,175,169,24,
138,68,151,146,138,235,35,251,35,178,74,72,
54,53,34,137,119,25,4,75,100,125,143,68,
85,88,115,159,102,42,108,113,70,222,230,186,
247,28,139,94,139,140,225,62,226,29,76,0,
170,233,57,216,97,196,116,74,231,71,71,93,
76,61,166,63,46,177,51,238,99,74,201,210,
202,28,229,155,117,198,163,144,86,219,99,56,
14,16,137,192,245,38,193,200,117,35,76,67,
65,138,24,223,143,209,117,137,239,199,24,134,
130,174,75,162,236,28,243,238,26,59,135,14,
113,156,208,233,248,152,166,67,191,31,16,118,
199,164,74,5,226,68,82,208,39,159,81,6,
19,165,80,169,108,224,121,17,154,246,145,245,
74,196,221,215,119,56,243,156,207,110,225,75,
100,204,117,226,56,97,97,57,131,148,130,195,
215,55,208,87,134,156,155,181,120,125,183,76,
34,251,64,30,199,207,97,25,3,14,218,203,
120,65,138,43,117,27,117,127,10,213,114,144,
153,144,173,198,12,181,153,87,81,116,147,98,
113,146,81,238,220,217,165,206,1,153,156,69,
119,221,193,42,207,161,29,47,253,41,29,108,
97,251,46,254,236,111,113,109,243,15,184,120,
230,88,13,118,216,66,120,62,59,83,53,244,
192,99,186,209,38,201,90,52,221,128,146,174,
113,169,144,225,106,215,229,122,189,76,76,8,
24,184,161,32,159,57,32,47,99,222,104,246,
81,4,188,146,182,168,215,247,232,118,125,140,
173,49,207,77,183,40,204,102,57,100,158,70,
106,25,101,227,42,205,227,21,162,31,65,242,
22,208,176,84,44,75,33,138,18,76,115,242,
115,126,222,162,182,187,193,63,150,155,236,36,
43,252,65,227,34,137,58,96,53,23,82,155,
111,49,216,250,83,70,254,8,169,231,88,122,
106,204,249,124,154,167,94,121,242,171,120,191,
87,98,228,153,24,234,100,47,123,24,43,44,
20,91,164,52,31,184,215,65,56,73,26,194,
8,77,8,252,56,70,149,130,75,179,22,204,
222,187,30,175,246,154,19,201,252,214,101,172,
252,28,253,206,207,200,174,142,9,194,155,28,
222,249,62,211,171,223,120,96,238,193,27,53,
9,212,61,102,253,96,34,80,200,164,232,121,
33,8,208,226,50,243,23,126,139,237,187,255,
134,237,153,14,82,198,156,42,152,220,181,239,
208,63,156,97,225,242,111,144,196,49,227,246,
206,3,187,48,30,155,224,41,42,77,81,225,
195,171,215,89,156,159,165,88,44,62,225,156,
94,194,149,205,38,169,83,47,125,234,99,63,
21,40,164,20,236,93,253,30,145,191,195,175,
45,180,208,219,93,54,118,109,28,59,66,81,
4,243,243,22,187,251,54,229,230,45,174,205,
127,149,218,210,55,62,129,48,21,148,51,26,
237,209,163,87,80,90,134,194,82,62,75,182,
252,100,166,128,190,211,163,87,255,175,124,53,
191,129,54,118,177,44,149,48,241,209,145,196,
115,85,68,167,79,146,181,168,187,147,10,33,
73,252,147,158,97,125,227,53,146,153,253,19,
64,218,24,58,236,74,143,51,153,20,202,120,
66,122,69,247,145,41,129,223,2,33,80,173,
20,241,244,26,71,246,191,70,238,84,113,133,
196,233,207,80,201,63,126,254,68,17,130,219,
131,49,2,241,192,144,215,67,23,184,227,49,
87,239,96,175,46,240,198,96,204,253,27,101,
131,208,64,83,125,126,185,135,224,76,54,197,
148,169,243,163,122,151,15,122,67,194,56,161,
229,25,180,189,128,32,142,25,135,17,101,67,
99,103,236,82,49,52,12,41,73,171,10,162,
219,39,182,76,100,127,116,162,106,202,239,173,
29,183,173,42,44,157,107,177,27,45,179,18,
222,230,105,243,22,251,238,2,7,253,60,118,
236,112,166,210,197,144,18,85,8,156,104,178,
11,99,174,216,224,237,253,89,190,186,52,228,
197,114,150,150,23,48,176,199,52,134,121,76,
101,200,72,215,120,106,38,131,118,228,34,36,
120,238,195,21,107,169,164,227,18,33,4,152,
166,130,239,199,168,170,160,92,214,233,26,89,
162,225,214,73,162,19,199,9,251,251,54,154,
38,241,154,67,94,154,182,185,27,154,108,118,
10,36,244,120,65,145,36,249,12,66,177,153,
187,79,177,184,181,53,38,138,18,84,18,100,
224,227,54,71,132,97,130,169,9,204,175,92,
34,223,31,177,247,195,91,124,110,240,6,95,
158,43,114,117,246,52,219,94,204,249,74,157,
5,203,228,167,173,45,162,200,0,20,8,53,
230,230,254,23,66,207,102,241,124,238,196,102,
252,36,145,152,255,111,217,187,117,157,245,96,
11,223,62,160,84,141,80,3,159,213,183,127,
192,173,47,124,27,103,208,194,204,86,57,90,
171,177,60,223,102,28,69,28,142,116,198,227,
28,227,237,203,100,202,63,99,90,1,97,187,
92,247,230,56,114,99,174,55,160,106,164,89,
174,30,49,159,210,153,75,25,220,29,218,24,
138,202,230,40,160,162,169,8,63,192,41,231,
153,251,202,89,42,215,119,216,223,183,233,247,
124,138,5,151,89,214,184,158,125,145,218,11,
207,177,116,235,42,219,155,163,143,105,241,195,
135,170,185,56,78,88,90,74,67,156,176,216,
94,231,119,22,199,252,223,206,121,190,200,22,
178,54,77,84,77,120,191,147,240,92,41,62,
169,34,158,40,38,184,49,59,141,2,197,114,
192,92,161,243,201,252,102,156,240,102,171,127,
162,54,236,249,33,70,34,241,194,152,233,156,
206,229,66,150,173,177,195,145,227,19,39,201,
137,180,93,85,242,199,1,54,32,103,41,228,
52,133,45,249,6,59,55,27,44,61,253,63,
222,23,51,53,98,92,118,12,157,36,72,184,
115,117,68,148,243,152,206,234,196,197,58,71,
107,127,65,156,106,98,42,58,190,18,179,213,
156,103,245,105,151,237,189,63,69,237,255,15,
164,139,11,100,167,78,127,230,36,239,141,43,
235,220,94,223,226,31,254,198,175,63,209,51,
110,221,185,139,159,153,71,127,130,199,126,58,
153,45,4,249,217,11,228,103,95,98,183,117,
131,169,241,1,233,153,2,83,231,43,248,157,
49,59,59,99,52,203,228,96,241,18,234,220,
255,252,208,243,239,127,205,188,165,17,199,9,
78,16,63,64,106,11,33,56,55,157,65,6,
227,39,115,88,77,18,14,182,255,4,243,252,
30,103,43,25,252,253,62,66,72,210,38,36,
167,231,38,18,76,85,33,46,100,121,171,213,
199,241,19,148,254,83,100,75,231,176,123,187,
244,227,191,64,230,239,101,223,163,48,98,20,
70,108,219,14,206,40,203,88,121,144,107,136,
205,38,227,237,54,229,218,87,177,215,93,228,
66,29,81,24,163,151,71,104,106,64,111,84,
61,105,31,61,234,104,186,62,65,2,135,142,
119,178,152,231,81,199,217,156,69,126,100,179,
155,182,216,245,67,34,18,84,49,33,86,133,
72,176,189,44,41,221,254,165,193,196,216,205,
49,159,115,40,27,26,37,93,165,231,135,248,
113,204,32,136,136,146,132,67,199,159,88,170,
232,26,215,250,99,118,109,143,126,16,50,10,
35,236,124,150,100,182,74,102,104,159,72,106,
79,206,239,216,102,67,57,207,242,220,0,114,
105,48,117,178,166,75,57,61,198,73,28,190,
183,145,230,131,166,142,170,198,84,82,9,215,
250,35,164,140,233,142,203,204,229,198,188,217,
234,19,37,9,137,180,177,221,28,91,173,121,
20,227,136,213,169,2,83,170,79,105,202,194,
144,96,59,17,186,33,49,12,5,77,147,44,
46,167,65,40,104,150,69,228,216,140,199,33,
65,16,227,22,86,89,205,213,217,220,24,82,
173,154,232,186,196,182,63,154,115,73,136,99,
48,102,22,232,102,52,66,57,100,187,83,225,
134,109,48,189,146,34,173,107,200,206,0,146,
73,21,114,116,228,80,42,233,228,115,58,131,
161,193,104,123,23,69,17,212,106,38,106,163,
205,78,249,25,74,43,69,140,197,2,206,108,
13,87,9,25,198,14,75,150,73,144,196,212,
93,159,47,214,50,52,253,49,241,160,68,38,
115,14,69,209,17,82,125,200,61,86,209,76,
210,133,69,114,229,103,200,149,94,96,105,48,
68,13,125,14,206,61,75,34,229,137,192,195,
72,175,178,185,153,165,217,92,68,234,95,67,
207,126,145,76,121,133,104,244,46,203,195,61,
14,103,79,179,102,107,124,123,197,230,82,57,
230,92,41,68,17,48,107,25,8,49,113,10,
56,114,125,206,214,187,44,30,52,48,175,237,
145,201,234,36,211,101,212,98,154,188,8,104,
28,140,25,141,66,138,69,157,244,184,206,142,
121,158,153,75,57,114,90,136,26,60,8,14,
82,138,7,238,119,219,142,168,86,13,20,69,
144,20,178,56,213,18,165,180,199,29,167,198,
162,58,64,152,6,179,150,241,68,92,229,238,
150,199,107,127,30,240,214,15,134,28,238,133,
172,127,56,96,127,99,76,253,192,103,241,180,
254,200,253,99,82,8,70,97,196,145,235,159,
12,175,58,55,139,4,219,57,198,210,161,169,
56,28,56,62,94,28,227,116,84,226,163,44,
177,163,80,46,252,42,129,211,103,224,252,12,
39,211,167,235,134,120,119,179,136,244,24,213,
175,162,26,25,132,84,80,205,12,241,64,82,
63,82,25,116,158,162,88,248,187,36,157,105,
154,155,14,202,114,157,210,212,33,231,231,12,
106,166,78,49,229,80,201,12,49,122,3,230,
51,62,183,246,58,228,43,15,206,66,60,118,
50,251,163,64,158,202,17,249,14,138,219,225,
153,139,23,30,187,22,250,35,149,211,31,127,
247,47,81,166,206,61,224,12,254,168,227,137,
200,236,147,19,242,193,31,83,156,137,240,206,
158,167,120,212,198,222,108,226,197,18,229,252,
25,14,202,191,74,126,249,215,17,159,50,87,
161,8,193,82,197,98,174,152,162,49,244,232,
140,125,146,56,97,161,108,97,233,10,79,26,
2,59,187,239,17,105,77,236,142,32,82,135,
184,1,4,227,49,150,105,146,84,138,147,233,
230,140,197,214,200,193,30,67,178,177,196,210,
197,111,19,135,62,141,198,119,145,43,15,42,
147,50,170,138,4,132,20,4,99,129,98,60,
88,81,8,43,34,152,187,66,253,160,142,48,
82,147,157,58,226,30,9,252,241,62,191,252,
88,69,162,30,251,220,60,238,123,89,78,155,
204,37,9,114,191,129,95,204,159,8,0,74,
230,100,80,104,119,88,164,144,110,253,210,64,
194,15,76,10,41,151,211,217,20,113,50,145,
19,126,174,146,35,1,118,199,30,189,32,96,
20,68,236,217,30,251,182,71,152,36,204,28,
147,153,171,217,20,29,47,96,107,236,114,103,
110,138,203,51,21,10,2,196,208,70,238,214,
33,129,80,213,79,136,252,251,143,83,89,201,
236,121,155,126,16,242,253,109,147,145,175,160,
232,147,44,175,23,13,249,206,149,105,76,61,
203,190,72,38,250,126,25,178,88,91,99,183,
181,196,91,211,109,206,125,254,18,69,18,42,
255,31,121,111,250,36,201,121,223,249,125,242,
206,172,251,234,58,250,158,158,233,185,49,192,
224,32,64,130,0,69,130,160,68,29,212,177,
138,149,87,90,217,94,31,225,55,182,255,23,
135,67,246,134,29,235,136,181,34,100,249,144,
180,34,165,149,40,94,34,72,144,0,113,3,
51,152,171,239,238,234,174,234,186,207,204,202,
235,241,139,236,233,158,198,204,96,26,160,64,
175,185,249,170,163,163,58,59,235,201,231,249,
157,223,239,247,183,89,163,16,10,164,110,31,
98,38,34,102,18,22,50,100,111,172,179,222,
41,147,77,52,217,223,191,91,78,150,104,29,
140,230,219,220,28,49,55,23,99,122,218,162,
209,152,224,121,33,178,162,80,27,21,249,134,
182,198,43,153,20,203,153,38,111,86,167,152,
26,217,200,59,251,4,139,21,148,157,125,146,
201,200,200,40,138,68,224,122,72,94,31,1,
156,59,151,58,12,104,22,118,126,198,223,22,
191,204,153,236,14,174,235,99,40,50,231,45,
157,156,34,83,29,59,148,76,29,39,8,113,
67,129,94,169,178,189,247,63,144,148,190,196,
212,226,11,15,207,251,132,160,92,219,162,21,
58,172,237,254,20,185,27,146,47,124,25,69,
137,178,113,213,136,147,155,123,242,1,196,74,
141,222,147,23,248,206,122,138,111,44,247,137,
41,202,161,49,190,251,46,241,124,164,221,38,
149,247,182,216,89,237,178,21,70,25,64,178,
121,155,124,126,11,213,80,72,60,119,142,89,
199,167,181,59,98,103,103,76,50,151,32,174,
56,232,173,6,166,152,32,103,52,116,93,34,
22,83,15,215,119,119,55,234,199,217,118,180,
238,189,158,71,46,167,35,15,70,164,111,110,
96,46,84,120,203,47,241,194,94,147,32,117,
50,17,196,107,111,141,185,246,134,203,185,171,
6,234,57,147,229,139,38,205,186,207,107,63,
176,153,76,100,30,102,255,2,31,38,31,81,
55,208,47,30,101,32,3,15,252,154,129,108,
8,212,246,25,242,133,47,225,7,67,116,43,
207,250,7,255,19,250,197,168,103,16,108,21,
40,149,190,206,126,227,111,216,87,255,12,107,
253,115,148,206,124,5,17,6,36,115,23,177,
188,121,98,153,136,65,159,139,103,209,99,89,
218,27,1,237,169,29,158,241,6,232,213,6,
127,211,93,38,51,235,241,204,178,142,254,225,
62,83,139,10,157,234,219,100,103,158,252,228,
135,120,80,227,119,126,231,87,31,233,36,238,
58,10,97,166,79,124,235,19,243,40,210,229,
11,184,163,34,235,141,24,45,33,232,72,49,
106,149,223,68,90,250,207,137,151,30,127,168,
87,122,208,61,101,89,34,105,170,20,83,6,
165,180,137,169,41,39,230,113,216,221,27,48,
124,21,199,17,92,60,237,147,184,190,131,153,
214,176,190,122,21,49,95,6,69,198,214,117,
90,19,143,110,47,100,178,251,34,217,226,115,
232,177,20,59,183,254,47,164,211,107,247,205,
132,113,195,16,79,132,56,65,72,24,168,204,
196,36,84,85,98,176,171,225,174,38,80,203,
19,12,93,98,113,214,167,155,104,220,247,247,
67,59,77,220,24,130,20,193,97,207,165,226,
236,59,46,49,85,193,11,5,129,16,140,131,
135,55,243,231,226,38,139,9,11,195,15,144,
171,13,118,202,5,186,7,153,199,200,15,240,
66,193,200,73,17,183,250,159,153,163,104,13,
202,164,19,53,134,126,192,157,161,205,141,222,
136,190,31,176,16,55,201,25,26,211,150,193,
98,194,34,161,42,100,117,21,59,8,169,196,
12,46,164,226,152,18,20,110,108,48,107,104,
148,11,105,118,100,153,73,204,36,94,202,131,
162,16,150,11,88,25,137,213,246,12,217,216,
8,249,64,173,182,239,69,205,91,83,137,136,
117,91,125,133,181,118,129,174,163,50,180,51,
168,178,79,37,191,73,194,234,145,176,250,36,
172,62,73,107,4,82,64,50,214,101,171,147,
103,109,16,178,19,244,216,52,77,118,98,105,
86,99,167,152,148,19,88,197,52,106,204,68,
228,211,232,59,187,196,116,143,122,61,234,137,
20,204,1,19,171,204,184,213,38,12,161,223,
247,72,36,52,122,61,143,49,41,182,252,5,
158,140,221,198,28,15,153,157,45,49,211,234,
160,38,100,250,118,130,169,253,77,228,129,141,
72,196,232,215,135,116,58,46,195,161,79,167,
227,162,140,58,228,150,242,216,231,230,48,82,
49,68,58,1,49,147,93,210,140,229,1,9,
77,102,201,113,201,214,91,104,190,79,102,42,
131,161,42,180,93,143,103,108,135,139,131,14,
203,51,58,187,94,29,137,11,72,146,122,44,
163,128,136,129,157,223,93,167,95,152,102,146,
46,32,124,129,228,239,50,108,109,18,203,156,
69,187,7,158,30,6,30,34,12,177,251,187,
236,109,253,59,150,22,251,12,213,16,199,141,
35,105,61,100,73,34,118,15,64,97,223,113,
73,108,214,184,254,205,235,116,26,54,97,120,
164,238,234,186,33,169,148,202,250,218,16,246,
59,100,191,120,14,122,35,90,245,49,253,161,
32,185,88,68,82,100,140,65,7,195,80,72,
36,52,12,67,161,84,50,49,77,133,120,92,
69,85,101,70,35,31,33,162,53,223,219,115,
200,101,52,84,25,20,63,224,109,109,134,83,
75,26,250,193,51,249,110,244,127,85,237,232,
192,173,14,109,114,122,132,206,43,20,53,46,
60,105,50,51,175,81,154,214,112,198,1,223,
255,230,128,202,101,149,207,61,99,161,155,199,
109,210,245,119,29,126,240,205,49,203,151,52,
20,45,210,4,123,168,51,182,2,188,107,69,
210,83,143,161,91,121,226,249,121,100,69,35,
157,123,156,254,70,21,41,215,69,235,158,165,
48,247,2,217,220,51,72,163,60,153,242,21,
100,69,101,212,218,161,190,242,93,188,81,159,
84,233,136,13,175,91,89,76,109,158,118,243,
3,236,162,160,18,4,40,49,133,59,55,166,
217,237,250,204,94,140,209,26,193,254,246,152,
76,249,137,79,148,81,0,4,70,134,149,183,
126,200,213,199,31,61,143,226,213,159,189,195,
182,155,68,179,18,255,180,25,133,36,43,164,
42,231,73,113,30,207,126,145,160,40,147,53,
98,252,34,175,192,29,227,183,191,199,139,211,
219,12,84,149,212,110,143,198,40,64,76,28,
226,59,117,130,83,51,76,130,144,238,196,143,
234,235,107,115,44,44,127,137,113,123,143,253,
141,127,36,156,189,133,36,193,148,169,51,242,
3,198,247,148,129,238,34,119,61,179,199,140,
150,165,46,57,180,138,54,106,113,130,240,32,
145,144,153,4,81,41,200,255,8,25,36,102,
14,104,244,43,44,21,234,12,188,128,149,3,
201,102,39,120,120,153,169,98,25,76,91,58,
123,182,203,214,200,161,57,241,184,148,142,51,
93,204,210,126,144,150,147,36,62,211,181,205,
167,106,236,52,79,67,97,13,136,178,133,75,
233,248,3,159,91,110,116,88,52,52,26,7,
28,17,36,153,224,236,60,234,107,31,16,191,
189,201,197,132,133,72,196,16,153,36,225,124,
52,217,207,66,80,84,55,248,230,157,10,139,
153,17,219,253,72,174,68,146,36,206,229,251,
204,165,192,241,44,78,149,110,222,87,38,8,
133,32,103,104,88,138,204,92,204,100,117,104,
51,99,25,172,105,13,110,213,103,112,124,11,
69,10,80,100,31,203,232,112,107,163,68,33,
181,207,139,21,141,130,166,146,152,146,9,91,
114,20,249,7,2,73,146,80,251,59,135,145,
255,230,230,136,189,189,40,226,221,103,142,69,
239,58,205,154,65,124,33,142,186,177,139,52,
28,243,132,57,224,90,233,249,131,34,183,79,
235,78,131,205,205,168,68,169,235,50,229,178,
133,16,2,75,21,236,100,146,124,167,33,24,
185,57,226,154,132,37,203,148,53,133,161,31,
160,53,218,200,251,109,194,48,68,171,181,40,
201,50,101,199,133,184,201,164,152,163,109,154,
60,117,81,227,199,175,189,67,126,230,185,251,
214,127,249,141,239,179,242,228,151,8,52,29,
5,200,47,60,3,60,131,16,130,73,191,121,
12,214,190,246,225,159,160,196,5,185,156,195,
87,190,96,16,83,116,190,189,97,145,75,214,
81,37,137,188,17,25,220,247,187,67,122,174,
207,130,169,83,123,125,29,223,143,246,154,170,
74,148,74,22,213,234,248,176,44,183,176,16,
103,107,107,196,240,47,223,101,230,124,30,69,
145,208,227,26,165,100,151,91,163,51,168,54,
228,212,46,5,173,125,216,143,104,54,39,56,
78,136,127,79,211,237,238,17,234,245,188,72,
8,113,48,98,62,27,178,229,120,92,60,120,
174,31,126,123,72,99,79,65,150,92,126,239,
95,165,113,165,128,15,247,198,20,22,53,210,
154,138,172,194,250,208,166,49,244,112,106,6,
86,202,227,204,215,13,214,238,184,136,7,152,
166,102,67,16,47,150,145,229,33,99,47,122,
22,191,165,161,230,143,207,143,15,39,18,222,
141,28,114,169,131,170,29,135,164,171,102,156,
84,234,10,126,117,145,252,252,179,56,131,58,
237,250,107,196,99,203,135,13,237,120,97,1,
100,65,114,234,254,62,131,145,44,96,213,47,
177,85,253,25,163,114,154,23,124,151,234,249,
61,236,238,231,249,222,91,91,200,70,149,202,
210,111,127,170,51,172,199,211,96,36,89,219,
216,100,105,113,225,99,63,187,82,235,163,232,
197,19,223,251,211,107,61,73,242,47,204,65,
132,129,79,243,195,255,157,133,204,14,179,97,
23,247,70,13,3,193,192,14,233,116,4,153,
25,21,113,48,157,43,4,222,172,246,241,55,
138,204,156,250,198,17,18,33,220,68,178,34,
195,109,127,76,159,64,120,50,205,192,163,234,
77,14,65,90,189,127,48,16,95,20,200,105,
135,153,152,129,34,73,108,221,51,162,51,97,
246,81,101,159,219,181,57,202,185,77,236,3,
67,159,80,21,250,15,81,32,221,179,39,244,
61,159,83,137,72,204,174,104,234,132,66,128,
162,224,221,3,37,46,154,58,251,142,139,170,
120,145,90,169,213,251,76,214,88,145,125,102,
242,235,212,59,51,148,178,59,204,197,12,76,
69,142,122,3,130,99,100,164,187,34,110,229,
98,14,194,144,224,212,12,76,92,68,204,140,
164,171,7,99,164,193,152,80,145,161,82,56,
132,143,72,72,232,137,91,212,220,4,197,172,
195,69,67,103,218,210,249,238,102,130,181,182,
78,76,63,14,3,204,233,26,115,113,131,152,
162,144,51,180,195,68,174,112,96,72,74,166,
142,169,212,184,86,43,80,200,212,200,235,26,
109,215,99,58,191,193,126,119,134,239,108,141,
120,174,226,113,234,137,179,200,111,220,64,81,
186,4,129,56,116,10,158,23,210,110,31,143,
42,83,180,144,101,137,233,105,11,212,200,128,
133,249,12,114,173,137,209,223,57,52,114,253,
190,135,44,75,84,42,22,186,46,147,126,250,
20,82,127,196,234,116,153,55,62,212,152,153,
78,51,91,108,98,202,50,67,63,64,114,125,
78,59,19,36,215,167,81,119,208,218,53,50,
25,29,82,113,6,229,18,111,54,75,56,235,
23,81,212,4,174,87,39,95,185,191,244,144,
175,174,113,231,233,47,19,200,202,125,200,192,
143,150,27,234,91,111,208,41,24,92,41,11,
158,41,31,157,213,129,43,241,100,82,103,234,
96,13,199,65,192,214,200,225,75,165,44,153,
91,27,220,25,30,25,205,133,133,56,107,107,
145,132,77,62,111,224,56,33,139,139,22,205,
230,132,225,192,227,214,27,145,144,162,37,251,
132,55,183,152,119,55,144,145,216,110,38,176,
115,211,204,25,17,42,82,211,101,100,69,98,
191,238,17,124,68,27,165,209,152,80,44,154,
72,18,100,173,14,183,155,41,206,165,60,20,
73,226,165,223,74,208,106,248,36,147,6,170,
10,8,25,163,23,231,251,223,137,241,220,151,
7,180,109,159,213,15,179,196,149,199,200,148,
31,103,229,245,127,77,242,249,1,165,203,202,
177,76,233,238,117,230,172,66,171,213,66,55,
45,74,138,78,107,224,177,115,187,72,123,46,
32,94,24,98,104,14,116,34,35,82,154,255,
58,134,85,192,72,230,113,122,199,185,73,185,
153,167,15,225,242,181,218,255,3,146,68,185,
248,117,134,141,117,28,123,135,169,133,47,49,
232,93,163,63,248,25,178,148,64,37,131,162,
197,201,205,94,5,36,42,167,127,157,198,157,
12,221,250,26,255,80,216,229,169,217,128,91,
198,107,120,253,175,209,220,238,34,185,239,80,
74,126,229,83,157,227,78,96,240,215,175,222,
228,191,204,101,73,165,30,60,75,232,250,173,
21,108,171,28,145,101,63,115,71,241,11,186,
252,137,205,96,237,127,229,243,11,13,244,219,
43,108,109,140,168,59,1,149,178,69,38,171,
33,66,31,77,147,144,106,45,194,74,1,69,
2,51,41,35,226,103,48,147,71,152,96,25,
139,224,158,230,245,67,81,94,227,44,155,70,
135,201,129,177,246,170,49,22,159,253,125,218,
245,31,66,122,147,234,120,114,223,240,21,0,
83,31,83,201,111,82,109,158,162,152,217,65,
83,93,6,94,64,119,88,96,226,69,80,197,
82,118,155,238,176,128,44,133,164,226,109,70,
126,192,181,238,240,48,196,138,29,96,205,45,
69,57,116,54,119,123,29,233,120,139,70,119,
26,89,14,137,25,131,207,6,243,116,192,208,
118,220,216,129,18,235,232,80,152,239,124,42,
78,136,96,46,102,18,206,151,17,83,89,164,
78,31,185,218,64,123,237,3,132,34,35,217,
147,99,32,8,17,51,35,241,185,225,24,169,
213,37,177,56,195,148,169,113,41,173,18,87,
143,234,163,95,158,31,242,106,163,143,47,162,
239,28,87,21,150,147,49,114,134,74,236,96,
102,179,56,40,195,73,18,135,191,211,100,137,
32,80,209,20,143,23,139,89,246,109,151,190,
23,48,151,48,40,24,251,220,220,47,146,55,
70,72,99,7,197,182,153,158,182,216,216,56,
2,42,124,212,73,44,46,198,89,100,64,175,
167,209,237,186,20,42,185,187,227,29,16,137,
24,115,253,53,246,253,2,126,115,155,78,199,
229,242,227,89,184,176,128,42,193,15,93,139,
199,206,22,184,246,211,105,44,89,160,152,235,
92,235,186,156,138,91,156,241,125,98,215,86,
32,8,25,219,62,59,213,113,228,92,166,44,
110,77,157,98,115,239,73,242,139,95,37,253,
145,224,235,94,212,147,234,185,152,195,30,161,
162,62,18,62,238,59,67,108,233,53,74,83,
109,54,218,151,184,60,181,71,199,245,216,233,
27,72,178,67,74,85,24,249,1,123,246,132,
188,161,69,128,9,63,68,26,29,41,39,228,
242,6,137,39,151,48,119,175,51,53,101,146,
207,235,135,70,62,149,210,24,141,238,145,104,
183,125,182,182,60,242,5,131,84,74,161,18,
244,241,66,153,27,227,101,206,89,171,236,85,
109,130,64,48,61,99,161,153,6,59,155,189,
195,94,197,100,18,17,39,231,78,167,89,242,
218,180,133,201,183,214,44,126,109,113,140,53,
118,40,58,125,104,57,248,231,23,81,37,137,
39,46,201,252,172,147,227,157,183,114,200,146,
73,62,115,137,86,247,175,24,239,190,69,242,
133,1,95,42,101,136,169,202,3,215,104,102,
65,103,230,32,208,110,78,92,6,178,143,122,
97,143,152,22,163,61,44,226,143,20,230,218,
18,90,204,64,228,66,140,100,254,99,215,121,
220,169,162,137,105,242,229,207,35,171,58,131,
222,29,4,99,154,91,175,147,91,254,128,199,
243,9,170,246,38,186,44,211,29,250,108,221,
88,99,246,220,63,67,146,100,10,167,159,195,
29,158,67,179,18,188,183,246,58,102,234,117,
198,238,53,22,46,254,17,170,153,248,212,103,
88,78,20,209,172,36,219,213,61,46,61,196,
81,212,26,45,36,165,0,129,243,203,227,40,
134,187,255,158,207,93,24,224,191,122,135,219,
27,3,130,32,154,4,150,76,170,12,135,62,
149,105,139,189,218,136,202,124,9,121,175,73,
63,155,98,18,132,24,178,254,145,134,113,146,
71,78,100,29,199,9,124,155,113,205,64,41,
218,228,117,141,14,18,146,34,161,233,121,98,
218,14,105,77,61,36,129,61,8,162,54,83,
88,99,167,121,154,124,170,70,119,88,32,147,
104,146,73,52,9,67,133,122,103,142,108,162,
65,32,20,38,7,3,93,100,41,210,121,218,
24,57,76,41,50,146,231,227,249,58,16,69,
118,173,201,81,132,55,149,217,165,217,175,32,
4,15,20,22,252,68,96,58,33,211,232,85,
162,161,61,102,15,203,24,29,58,164,122,103,
150,29,253,136,31,226,9,193,59,157,193,97,
137,110,33,110,34,44,3,97,77,69,131,112,
250,35,164,225,56,26,139,121,215,89,8,129,
114,123,11,17,51,145,155,93,68,38,137,38,
75,124,46,31,109,94,105,236,32,223,217,2,
203,36,219,29,240,242,185,5,110,34,81,183,
93,76,89,102,198,212,142,213,78,63,236,141,
88,31,70,153,192,179,249,20,113,85,227,221,
122,140,91,237,52,95,94,170,147,84,77,118,
194,144,231,10,41,98,170,194,43,245,46,65,
168,146,212,84,164,253,54,193,185,69,98,163,
149,136,79,17,106,72,193,228,62,57,153,189,
61,155,249,249,56,243,243,113,92,55,196,219,
239,161,42,18,178,42,19,156,154,198,26,110,
17,184,67,110,237,217,88,150,194,228,137,51,
188,49,86,177,107,41,70,123,62,164,70,232,
70,22,75,191,78,115,226,178,164,107,60,182,
94,5,103,2,65,136,176,12,122,187,54,169,
164,70,50,103,242,86,225,44,237,253,231,41,
156,122,225,227,75,10,206,152,226,198,77,118,
206,159,172,193,217,217,123,7,169,216,62,40,
117,220,228,207,111,149,8,195,24,170,226,147,
79,213,184,209,55,232,123,1,49,69,102,109,
232,144,209,85,18,68,14,61,12,35,130,98,
241,66,9,111,166,204,185,75,85,182,55,134,
236,237,218,204,206,199,48,13,153,210,83,115,
140,126,180,129,59,9,112,156,0,207,11,233,
118,61,178,89,157,193,192,39,22,83,48,130,
1,121,227,14,183,199,167,241,139,101,68,245,
67,54,54,70,132,185,10,75,103,211,212,111,
111,29,58,139,253,125,7,33,4,243,174,199,
75,198,62,175,205,63,205,191,121,183,194,191,
200,127,72,113,107,143,189,51,179,140,71,14,
211,150,193,91,111,24,148,151,94,70,79,228,
104,172,255,132,118,251,21,190,250,5,133,152,
10,144,63,241,254,95,76,88,216,65,200,202,
80,66,181,96,174,176,10,121,48,164,23,41,
45,189,124,162,123,36,139,75,36,139,75,71,
149,133,212,105,244,88,14,213,176,168,238,252,
144,203,185,144,184,40,97,16,48,91,28,129,
116,155,181,155,255,11,165,242,55,48,226,83,
7,77,98,139,169,197,95,97,208,156,70,247,
247,88,127,231,223,114,230,217,255,230,231,171,
192,216,29,76,253,225,92,138,11,75,115,124,
248,214,30,40,230,47,143,163,48,115,207,80,
223,191,78,62,25,67,150,6,248,128,235,133,
244,122,30,211,115,113,92,15,172,98,28,169,
213,99,18,51,145,133,64,120,160,27,199,71,
149,202,124,252,212,39,73,40,72,88,200,141,
121,114,115,207,176,191,243,151,140,23,219,88,
11,46,171,55,126,196,180,126,142,126,47,196,
202,9,236,15,12,172,139,19,30,70,131,158,
45,172,178,182,119,145,165,202,135,247,52,240,
3,74,217,109,84,73,58,208,48,18,135,136,
39,132,78,86,202,144,163,15,182,67,127,92,
68,50,215,31,120,239,66,106,143,70,111,26,
85,241,81,100,159,238,168,0,2,52,213,69,
150,2,130,80,69,150,3,82,177,3,25,13,
33,49,116,82,140,38,73,184,135,217,45,132,
68,57,183,245,144,50,84,128,46,203,8,4,
143,101,18,100,117,141,31,212,59,8,33,136,
61,64,209,87,164,226,136,84,212,207,80,110,
110,220,133,131,17,92,56,133,200,164,144,54,
107,4,51,69,164,70,39,82,42,5,36,63,
64,84,10,96,79,8,23,167,209,235,237,104,
254,69,167,143,208,84,2,117,150,65,50,142,
121,224,44,196,129,85,239,14,202,188,239,197,
73,104,48,149,108,113,126,74,144,211,205,195,
146,84,234,46,211,62,140,145,48,134,128,68,
88,200,130,161,97,156,179,73,239,59,180,212,
108,154,11,0,0,32,0,73,68,65,84,204,
115,92,202,110,69,50,24,43,125,12,67,102,
56,244,153,76,66,86,86,162,137,120,243,243,
113,110,124,216,65,146,36,206,158,79,161,109,
236,17,4,130,141,141,33,243,243,113,18,25,
131,239,111,230,49,180,47,81,156,89,166,111,
252,9,175,85,77,52,50,164,226,19,6,72,
92,104,247,162,108,36,21,39,56,51,199,36,
25,167,148,222,102,226,195,223,239,158,167,192,
55,200,207,207,60,242,12,36,218,251,84,207,
93,61,121,16,224,73,136,145,12,70,136,44,
7,135,147,251,100,73,98,62,110,161,72,80,
62,40,231,41,99,135,171,217,100,196,228,38,
226,155,84,78,165,121,71,156,195,121,197,227,
165,132,201,204,227,73,98,59,125,134,29,155,
193,208,99,250,55,22,57,213,232,210,173,141,
14,251,52,0,221,174,139,21,83,137,199,212,
8,225,101,200,156,141,173,50,54,45,26,210,
52,205,205,29,228,246,10,183,244,39,200,43,
18,75,75,137,200,121,132,130,241,56,64,196,
45,94,41,62,75,43,24,243,242,217,117,222,
110,47,160,205,207,227,140,2,250,3,5,117,
103,66,169,242,27,232,137,28,253,250,10,67,
245,21,202,103,92,98,106,250,19,219,21,9,
24,251,1,242,172,123,52,86,77,2,87,84,
63,181,173,138,101,231,144,21,149,126,227,38,
185,66,200,254,196,229,213,234,132,151,230,29,
132,144,184,84,136,35,61,214,98,237,189,191,
103,102,233,95,28,119,58,133,179,36,11,103,
73,20,206,252,92,25,5,128,150,42,243,55,
111,175,112,110,103,159,95,251,210,115,247,161,
157,254,230,135,111,160,207,61,203,164,223,250,
229,113,20,118,103,19,67,113,209,2,15,207,
23,40,146,196,242,114,130,216,233,34,227,197,
105,228,173,38,98,59,34,196,253,80,213,112,
59,67,34,110,157,248,200,2,185,119,249,118,
15,78,217,236,10,190,190,131,34,205,17,75,
207,146,117,190,194,234,250,143,32,37,200,45,
111,208,88,233,33,203,9,246,157,1,214,99,
31,31,169,55,251,229,195,65,56,247,149,5,
132,64,18,160,73,50,158,8,241,66,193,110,
183,192,239,157,29,162,41,22,152,6,65,32,
125,236,139,153,74,239,210,234,85,112,60,139,
153,194,26,32,49,57,144,10,183,148,17,161,
144,217,239,206,18,134,50,189,97,158,217,226,
42,165,204,206,201,13,13,18,166,34,243,66,
46,137,36,75,32,203,60,147,75,82,29,59,
116,92,63,50,48,15,128,224,137,116,2,17,
183,34,165,86,207,71,222,216,37,44,100,32,
8,144,108,23,101,99,23,105,48,70,148,114,
72,173,30,193,133,83,132,11,21,122,94,64,
98,42,131,130,4,107,59,4,75,179,8,224,
149,109,135,218,32,135,16,2,65,140,148,38,
243,210,252,144,74,236,160,79,19,42,17,157,
255,224,42,181,186,72,189,17,238,133,69,26,
253,28,203,197,26,18,73,80,21,148,235,171,
200,251,109,230,79,37,40,251,53,182,75,207,
178,112,218,225,172,250,46,97,32,168,30,200,
110,248,126,200,236,108,236,144,76,10,176,185,
62,68,136,168,167,81,169,88,228,22,179,52,
167,202,104,173,207,147,173,60,142,221,221,165,
213,84,152,42,125,13,53,157,100,195,11,121,
58,21,67,105,117,232,95,62,77,124,48,230,
166,80,169,238,15,185,50,91,225,103,175,107,
148,206,252,243,71,150,54,0,206,124,240,83,
234,151,158,69,124,2,253,179,226,217,231,105,
108,10,70,27,215,160,84,67,182,162,146,222,
172,101,240,152,34,35,226,38,72,17,249,243,
98,58,142,188,83,167,181,49,162,209,206,144,
42,91,188,105,60,141,198,132,180,62,32,184,
120,138,239,189,158,36,54,147,33,83,217,224,
212,244,132,214,200,33,216,28,80,171,29,215,
90,234,116,92,146,9,13,89,134,68,252,104,
7,199,100,155,249,60,52,183,162,192,104,106,
178,194,48,54,79,54,219,101,48,240,80,213,
8,20,177,153,189,66,57,59,225,212,208,37,
47,233,156,55,54,163,97,75,178,140,183,60,
207,157,69,133,155,123,223,167,221,147,112,228,
22,221,120,142,76,176,251,169,109,203,19,185,
36,63,222,239,30,149,162,5,104,84,126,110,
155,229,216,187,184,98,66,175,59,33,36,206,
141,126,159,107,253,128,83,9,139,75,153,56,
153,167,27,188,249,222,255,140,98,151,152,79,
255,225,71,156,205,12,66,132,39,102,101,63,
208,150,169,26,214,244,5,54,186,219,252,31,
127,251,10,249,132,198,215,190,20,81,219,183,
119,170,184,233,37,244,79,120,127,53,130,161,
217,135,205,153,187,63,127,60,250,200,65,40,
42,130,71,35,113,78,122,207,135,125,86,4,
125,114,139,38,154,150,37,177,214,99,56,242,
49,74,25,222,155,41,209,26,187,100,43,57,
150,37,29,67,114,184,148,176,120,183,63,70,
142,67,119,244,109,38,235,13,10,51,95,36,
240,38,100,203,207,177,119,103,27,121,121,43,
26,56,228,128,104,166,81,252,60,184,6,142,
223,64,61,7,110,250,38,131,198,26,178,176,
200,200,125,212,116,84,39,78,46,183,216,107,
205,163,141,60,244,252,253,181,189,80,200,52,
186,51,164,205,144,169,212,46,60,0,165,116,
23,49,37,0,239,160,30,47,35,17,10,137,
55,90,125,190,28,51,80,195,144,152,57,166,
239,198,48,63,134,92,151,79,239,177,223,153,
61,52,237,119,117,163,14,141,197,129,99,176,
140,17,182,27,255,88,249,143,251,158,83,246,
217,31,233,120,195,26,49,219,33,52,116,166,
230,203,84,214,118,144,198,19,194,92,42,154,
24,248,17,227,37,226,22,254,83,23,80,223,
186,129,100,79,34,117,210,131,223,75,206,4,
238,150,236,70,14,97,54,197,158,89,225,198,
173,24,154,31,48,140,59,60,81,30,144,92,
172,112,167,59,228,78,163,194,110,111,154,63,
126,124,139,180,166,48,9,197,193,212,50,9,
201,113,81,110,172,131,237,224,127,254,10,72,
18,82,163,131,114,123,147,176,152,231,118,127,
130,237,165,121,44,147,184,27,37,68,207,61,
61,133,48,13,244,213,29,250,70,138,137,238,
163,127,238,2,122,183,143,82,91,97,225,229,
243,224,122,76,74,57,60,89,38,217,232,34,
239,238,147,72,104,212,106,54,203,203,73,194,
217,18,173,233,2,63,121,55,77,121,225,50,
190,107,163,197,178,148,43,191,143,145,42,208,
92,255,11,190,122,58,32,181,94,165,93,202,
243,234,7,67,166,76,147,78,243,41,38,126,
149,159,170,93,50,249,95,67,49,98,143,60,
23,83,107,215,185,115,230,10,178,166,193,73,
206,165,55,161,187,251,1,195,206,109,202,203,
191,142,84,211,25,137,191,39,232,66,42,15,
103,109,27,245,195,42,222,231,46,131,166,70,
217,151,16,140,81,249,219,247,115,124,109,185,
198,78,250,2,10,69,94,152,125,11,121,183,
129,159,184,64,46,61,198,182,126,147,190,250,
171,252,96,237,85,208,118,120,233,197,139,228,
246,155,108,189,89,101,228,169,88,9,11,119,
226,71,28,166,16,118,170,99,146,73,141,108,
78,63,216,161,2,73,68,252,162,84,34,100,
44,197,129,46,158,39,168,84,12,52,77,230,
150,149,102,112,171,202,251,55,38,252,254,217,
163,89,40,132,1,218,141,117,42,79,156,163,
157,222,33,107,104,172,12,108,152,120,108,117,
115,220,180,34,212,86,76,85,142,222,249,73,
202,218,94,112,140,0,27,116,85,146,153,139,
247,189,151,79,106,23,11,179,47,178,93,255,
0,105,102,136,161,143,217,233,102,201,38,247,
185,217,27,161,203,50,55,250,35,148,133,16,
113,51,123,223,125,59,213,87,32,80,201,206,
127,225,216,125,239,174,225,39,178,161,177,2,
29,10,52,251,109,238,252,233,95,242,220,99,
103,120,235,250,10,242,244,83,248,174,77,224,
77,8,92,135,80,81,78,230,40,66,55,58,
192,161,239,30,254,252,177,15,227,79,16,97,
112,162,151,113,210,123,62,236,179,178,118,138,
183,174,191,199,211,242,144,225,216,167,80,48,
144,82,22,77,55,98,4,15,122,42,157,172,
197,98,50,67,76,145,15,161,171,74,113,136,
93,125,27,127,252,68,36,255,43,160,60,253,
219,52,238,124,31,73,129,184,177,68,34,119,
230,176,14,62,216,95,165,118,235,39,140,167,
117,252,230,95,163,36,7,104,149,227,176,185,
82,124,139,193,90,129,45,251,20,167,166,239,
160,200,81,67,111,104,167,233,143,179,124,126,
190,198,56,8,105,56,247,191,80,77,150,184,
152,142,243,94,103,136,68,52,254,116,224,5,
84,44,157,222,74,15,39,27,50,233,143,208,
99,38,137,88,139,157,122,241,176,73,254,208,
245,18,143,142,10,210,241,22,3,59,67,111,
148,39,29,63,89,170,153,77,238,179,219,90,
164,95,116,136,111,236,162,0,114,181,1,113,
11,2,31,121,191,141,220,31,17,156,158,33,
44,125,36,42,86,21,194,82,14,101,173,26,
149,161,28,55,26,198,179,83,71,210,20,8,
2,208,85,250,167,206,51,220,215,120,185,241,
125,130,92,138,110,205,224,123,195,115,136,68,
13,99,56,203,75,203,125,254,110,61,32,171,
171,72,99,39,154,251,28,4,200,173,30,242,
78,157,240,212,12,34,97,129,36,161,172,85,
145,182,246,192,50,233,205,151,121,123,171,66,
220,240,143,198,109,106,42,193,153,89,228,90,
11,101,99,143,176,82,64,77,9,174,247,2,
2,224,153,233,34,115,207,7,140,103,138,172,
14,109,118,134,14,50,18,24,6,23,31,91,
102,97,109,135,105,25,198,23,151,216,141,89,
188,247,102,142,108,230,197,99,123,85,248,30,
237,245,255,155,151,138,239,227,5,26,223,162,
66,240,254,20,42,51,216,249,121,114,149,89,
60,103,25,207,237,131,224,68,103,66,182,135,
120,194,71,59,233,249,241,38,136,192,199,243,
218,120,195,30,4,33,126,71,35,172,165,40,
38,6,36,234,235,17,210,69,85,241,133,96,
101,96,115,62,21,227,77,77,33,118,69,197,
60,59,207,112,227,113,240,182,232,25,25,106,
169,52,181,215,98,72,198,115,152,70,84,243,
158,154,254,50,129,55,225,167,187,215,145,104,
16,127,162,138,38,59,108,118,20,158,153,235,
97,172,174,49,118,124,2,33,113,123,211,227,
233,180,134,44,75,4,174,64,51,36,220,137,
160,215,245,200,206,195,181,225,57,46,159,142,
244,161,60,61,134,26,78,120,74,219,226,169,
115,46,210,71,56,14,34,159,70,203,36,120,
246,224,188,46,37,44,220,32,228,47,110,155,
172,12,34,84,82,42,20,84,199,19,202,150,
254,192,140,247,163,215,192,247,143,155,222,110,
26,173,146,188,239,221,132,222,228,19,217,69,
5,240,198,89,186,189,36,2,233,208,78,56,
158,201,119,55,146,148,204,44,146,235,163,136,
194,177,251,134,129,71,70,125,135,81,80,36,
116,159,58,102,23,57,97,70,249,32,27,42,
233,113,196,212,37,94,217,24,33,165,78,163,
222,181,247,222,132,192,115,144,3,245,100,142,
66,79,68,163,5,125,215,62,252,249,99,235,
123,7,228,15,69,127,116,51,228,164,247,124,
216,103,237,246,27,76,79,203,244,175,57,88,
150,194,252,92,12,79,87,233,238,72,40,69,
80,197,20,131,96,155,157,209,145,8,169,240,
96,242,126,10,61,111,161,39,50,132,161,127,
120,223,185,204,127,130,211,175,209,109,191,67,
187,85,5,100,2,199,197,205,94,39,57,31,
144,4,184,231,17,44,85,97,18,68,210,207,
178,9,233,139,77,156,142,69,179,95,62,76,
26,226,134,205,76,126,131,93,91,58,236,61,
124,244,242,66,193,251,157,40,59,41,90,58,
79,102,147,252,164,217,163,187,29,195,28,107,
232,242,144,68,16,34,146,113,6,158,79,49,
83,165,51,156,162,209,157,97,118,106,21,73,
58,14,101,107,13,74,100,147,251,39,107,186,
89,93,186,195,194,39,114,22,113,115,192,107,
147,36,191,117,249,52,234,181,85,36,33,8,
166,11,40,235,19,130,165,217,8,54,90,184,
167,97,230,7,104,175,190,27,241,39,14,8,
81,34,102,70,205,110,215,67,234,141,162,185,
204,79,158,103,224,38,120,189,122,129,175,245,
255,156,192,15,185,253,131,117,178,25,157,223,
204,237,80,207,95,102,222,123,5,79,186,192,
149,108,130,129,43,147,217,170,33,55,58,224,
249,132,133,12,193,197,165,200,113,141,108,100,
215,139,216,223,18,56,87,150,249,214,86,150,
231,23,154,188,181,151,56,106,170,175,239,34,
111,238,130,128,149,213,1,75,137,24,147,242,
46,111,85,167,88,44,12,25,165,21,212,229,
57,148,102,135,43,251,93,140,98,145,235,183,
50,132,126,192,123,25,155,121,85,66,150,37,
38,65,200,181,15,19,204,44,254,225,125,196,
208,246,206,183,120,254,98,21,163,39,241,250,
36,77,138,223,37,245,216,249,227,77,233,131,
61,56,110,239,125,236,153,144,3,159,165,119,
126,196,141,75,207,162,3,154,117,178,40,57,
12,60,146,233,211,100,231,162,126,134,149,43,
147,153,60,142,151,30,208,114,254,77,84,26,
204,36,163,26,139,128,181,161,77,167,227,179,
251,122,138,92,89,227,218,134,73,102,225,37,
154,55,255,154,247,87,37,140,226,111,96,228,
121,160,156,182,149,61,174,50,106,25,111,178,
205,63,114,106,198,97,243,213,53,166,206,23,
201,46,101,89,187,125,243,16,81,118,241,124,
154,241,216,199,245,4,237,237,155,228,146,22,
31,120,103,48,77,16,241,12,167,99,85,164,
213,193,131,224,120,132,51,197,99,104,67,85,
138,148,123,203,233,1,245,97,129,212,120,66,
183,53,197,27,90,159,220,204,128,171,11,49,
146,154,242,8,64,199,71,99,156,220,3,223,
75,24,120,159,200,46,74,138,130,174,219,88,
233,35,94,139,161,200,36,165,4,47,86,70,
76,29,100,68,43,123,243,199,238,91,95,251,
46,95,158,238,179,223,236,177,49,233,144,200,
159,58,234,27,234,230,207,109,111,245,100,246,
254,239,21,207,156,104,82,233,127,240,61,138,
244,220,87,121,239,213,13,126,165,152,161,162,
4,72,134,198,59,74,14,189,119,30,207,174,
34,42,14,238,245,41,26,65,28,245,202,198,
97,83,74,206,79,72,104,95,64,86,143,70,
27,142,218,155,232,86,158,221,221,255,19,245,
76,244,18,133,35,67,173,130,90,8,62,186,
55,137,43,42,211,49,29,55,16,108,140,162,
116,206,175,105,36,90,1,241,75,71,181,209,
180,174,210,115,121,168,147,184,219,60,43,91,
6,93,215,35,169,40,248,66,96,200,50,157,
94,18,37,101,243,68,54,129,180,185,199,56,
21,63,36,255,101,19,13,178,137,6,213,230,
210,65,47,226,222,232,69,139,112,223,39,188,
50,137,38,174,103,82,107,207,63,180,137,125,
23,165,117,55,19,169,117,230,120,47,61,225,
241,207,63,134,80,21,36,1,161,31,48,44,
100,184,105,24,76,187,62,146,20,241,25,170,
195,41,178,75,23,72,223,185,126,216,6,146,
250,67,164,254,48,26,138,147,138,35,114,105,
54,135,211,40,43,91,156,158,185,67,115,173,
205,84,206,100,118,54,70,50,165,17,22,179,
204,205,65,152,186,130,226,7,148,252,17,221,
81,150,140,20,141,70,189,117,250,37,164,132,
137,186,222,32,101,22,200,245,163,158,135,95,
41,242,158,117,145,157,93,141,103,203,67,178,
106,146,146,97,34,215,182,145,87,171,52,134,
41,250,254,41,240,108,132,24,48,41,231,216,
217,150,41,244,226,132,195,207,243,234,134,132,
96,76,90,181,208,68,154,205,31,171,84,46,
253,30,8,137,94,243,93,122,103,60,20,201,
160,77,156,180,58,160,122,227,91,44,62,121,
84,91,222,189,243,215,92,188,188,75,170,86,
167,126,122,150,254,155,139,204,44,159,255,212,
123,62,213,170,177,241,248,243,17,53,249,231,
188,84,35,134,172,104,212,54,77,126,90,41,
210,31,6,152,183,199,180,235,42,110,63,73,
120,181,207,229,23,58,92,172,54,249,81,127,
25,16,20,206,31,17,190,28,183,113,159,113,
29,78,124,18,134,122,44,200,213,147,103,112,
228,51,188,221,253,51,158,250,195,18,59,117,
88,185,25,112,78,179,152,201,73,136,16,20,
93,34,153,210,104,236,79,232,247,61,36,25,
18,246,53,156,73,192,178,154,66,185,246,224,
12,57,56,61,27,245,186,14,174,158,231,83,
179,3,222,174,230,240,122,46,230,208,34,180,
33,62,213,199,51,199,152,25,193,200,15,30,
233,40,188,143,120,138,64,109,209,219,187,73,
44,123,164,202,251,105,174,173,27,127,202,210,
147,35,242,102,18,89,146,48,228,72,133,56,
90,174,200,30,237,140,29,236,160,197,100,208,
196,72,22,8,92,27,199,221,34,148,65,177,
29,28,111,227,152,163,248,255,250,250,133,59,
10,33,4,187,183,255,26,69,31,225,15,67,
102,46,255,203,143,215,38,145,36,148,188,201,
248,202,52,83,76,243,65,31,58,181,47,82,
152,189,64,179,245,109,2,229,22,146,187,136,
52,221,192,121,55,131,249,68,23,17,72,100,
180,95,161,176,240,252,241,168,175,250,54,169,
169,101,84,53,3,68,6,49,236,235,136,70,
12,117,241,158,13,211,149,73,231,37,18,154,
204,237,254,248,30,36,13,184,27,22,149,197,
231,233,238,253,3,114,37,106,170,142,188,147,
149,225,116,89,34,174,41,100,13,13,77,146,
145,37,144,146,125,28,39,228,253,238,16,115,
161,76,102,104,83,144,100,234,247,8,235,85,
242,27,212,58,115,148,179,219,0,244,71,57,
82,241,246,39,94,123,93,115,72,39,90,52,
123,211,104,234,4,63,208,8,133,140,8,163,
3,154,140,117,216,107,47,160,41,46,133,244,
30,217,68,147,237,110,138,139,11,1,170,36,
241,70,205,226,150,29,195,187,3,96,113,29,
21,89,142,164,153,191,54,55,162,234,158,38,
89,174,33,215,90,7,25,133,5,154,134,188,
85,195,91,156,229,70,123,129,165,245,159,144,
20,125,110,116,178,232,154,204,157,149,1,169,
148,74,34,161,210,186,182,71,97,226,17,206,
20,97,226,82,217,111,115,103,254,69,164,206,
128,13,227,60,75,238,77,252,141,17,147,190,
143,253,216,85,222,209,207,145,44,142,185,177,
55,203,215,231,126,198,19,246,24,58,54,173,
228,2,75,105,3,105,187,199,238,184,136,46,
108,242,246,45,182,182,198,204,46,198,248,192,
141,17,247,94,38,119,246,233,7,174,211,233,
131,137,146,78,175,65,96,183,121,99,115,6,
217,248,6,70,63,143,100,78,88,188,122,4,
137,236,55,110,193,212,251,44,186,145,58,241,
7,183,2,74,243,95,255,212,231,163,188,122,
141,113,58,143,175,233,255,36,142,2,96,216,
92,71,245,202,4,147,47,144,50,99,212,110,
125,143,236,244,83,244,228,31,3,125,242,16,
245,3,98,101,38,195,54,122,44,115,76,54,
251,238,229,250,33,119,234,35,38,126,192,213,
249,251,163,86,197,72,96,148,190,193,181,173,
191,162,85,83,153,120,19,228,171,23,233,175,
175,177,127,189,193,242,217,36,241,184,74,16,
132,200,178,196,104,20,48,59,19,163,189,49,
100,101,101,192,220,124,140,120,252,126,147,36,
242,17,170,169,55,81,120,171,150,160,227,57,
12,29,141,151,150,90,188,247,150,198,192,240,
201,159,239,241,194,108,10,120,180,90,68,40,
96,123,236,112,231,64,194,255,46,200,69,158,
110,179,94,251,54,201,59,112,250,242,127,127,
226,114,207,125,25,161,9,91,182,77,136,201,
114,210,58,70,254,235,78,66,254,253,90,130,
113,96,145,42,215,168,246,255,53,218,214,25,
36,179,69,249,98,155,137,98,82,12,234,188,
94,127,147,252,194,139,15,124,15,191,244,142,
162,190,242,61,20,57,129,171,172,243,114,97,
159,215,196,23,79,36,96,53,189,244,187,188,
253,254,55,145,244,9,170,40,83,94,122,17,
167,215,32,110,94,160,83,223,67,191,186,65,
208,83,144,45,29,89,130,176,86,166,112,250,
192,73,28,52,141,17,130,48,176,49,19,21,
92,251,52,99,111,13,73,11,81,138,14,178,
214,63,28,125,73,8,65,95,101,152,118,113,
14,72,119,135,210,224,50,24,103,199,236,87,
127,204,176,96,80,184,7,201,244,168,107,33,
97,161,75,18,29,215,231,102,127,140,151,16,
36,100,139,93,57,199,204,220,58,186,172,16,
202,10,140,29,174,206,148,248,78,189,125,72,
182,147,165,144,152,62,162,63,206,146,138,117,
112,220,216,167,114,20,0,150,62,98,104,167,
17,72,228,31,128,204,178,140,17,126,160,81,
109,46,81,201,111,176,213,152,225,79,71,99,
64,162,16,31,241,27,167,66,36,9,108,63,
164,100,233,120,161,192,11,67,226,170,66,115,
64,164,172,10,160,170,136,100,12,121,115,151,
238,252,57,246,250,57,30,243,127,70,175,213,
132,156,78,145,58,59,79,255,46,233,238,54,
229,214,7,136,80,176,181,53,66,81,36,146,
125,27,165,144,138,230,117,247,70,244,118,123,
44,102,62,128,177,76,117,117,64,175,231,114,
69,26,145,29,249,12,135,62,47,157,158,103,
247,78,153,197,78,36,107,94,88,20,172,166,
159,160,89,185,192,184,152,98,122,231,71,84,
235,62,153,156,134,117,186,204,238,202,28,115,
87,158,62,209,122,205,62,254,71,31,91,234,
233,116,126,196,197,39,52,164,205,6,40,50,
163,137,76,86,209,63,213,187,201,212,183,233,
21,103,177,147,153,127,210,115,151,42,159,37,
85,62,210,28,154,185,252,91,180,171,111,33,
230,54,48,100,153,132,7,175,251,75,168,153,
23,216,187,241,15,88,233,25,138,203,47,28,
59,155,99,55,224,214,222,16,47,8,153,201,
90,15,181,161,70,114,22,146,255,29,86,120,
141,64,93,225,122,77,80,200,106,164,172,62,
161,174,49,241,4,229,138,69,190,96,112,235,
86,31,89,130,233,138,197,238,158,77,117,119,
204,217,229,251,73,98,82,127,132,136,91,124,
119,211,228,244,212,14,246,120,194,215,23,179,
196,84,3,235,115,30,223,217,48,249,92,37,
113,112,204,37,2,33,163,202,31,67,170,149,
34,46,208,214,200,161,55,241,105,127,83,33,
247,141,128,238,184,64,162,48,64,54,61,70,
157,109,226,185,79,55,131,58,110,92,100,48,
222,100,91,56,140,130,128,47,20,34,71,87,
239,103,216,31,90,44,21,110,49,10,124,206,
167,226,188,219,153,160,150,63,228,92,58,78,
209,140,163,236,119,209,131,9,223,40,221,226,
7,155,127,67,254,212,111,255,7,225,40,228,
95,232,63,83,76,20,205,98,62,209,229,237,
141,24,66,212,217,93,255,183,12,90,183,31,
145,62,39,152,59,247,135,204,158,250,87,148,
151,190,142,8,3,90,91,111,146,41,63,129,
132,196,197,76,156,151,207,166,184,252,204,132,
138,105,162,72,38,32,81,91,251,54,123,55,
191,23,25,115,207,65,49,117,252,201,0,51,
54,141,24,202,200,227,60,218,120,14,73,24,
199,86,68,159,119,145,36,120,58,159,98,62,
126,188,46,168,228,124,204,199,246,48,50,46,
251,157,89,234,221,89,106,237,249,71,54,150,
119,70,14,113,77,225,153,124,138,43,153,136,
115,240,70,221,160,146,219,33,103,168,56,65,
64,219,245,144,252,0,85,145,56,159,58,30,
25,165,226,109,100,57,160,214,153,35,25,239,
252,92,239,97,42,189,139,235,25,135,140,241,
251,214,91,241,152,41,172,97,132,121,190,188,
180,135,38,75,92,157,110,242,27,11,81,10,
157,214,84,202,86,52,96,74,151,37,226,158,
207,200,53,136,203,67,112,189,232,36,250,62,
34,157,164,181,244,24,189,161,197,197,189,31,
98,223,222,99,99,107,68,163,57,33,239,238,
113,126,240,26,57,183,198,237,216,21,222,29,
94,70,81,36,214,215,135,92,123,171,129,187,
213,136,234,28,181,14,66,64,167,237,210,110,
56,196,19,209,108,243,78,203,161,94,119,208,
77,25,37,109,17,38,19,17,27,60,25,71,
40,50,167,119,127,74,59,204,178,20,220,132,
137,135,101,41,100,43,73,126,52,60,75,182,
242,229,159,123,47,143,187,59,236,109,253,143,
36,150,118,200,7,126,212,67,81,21,174,94,
128,218,157,239,126,154,84,27,109,226,252,147,
59,137,135,214,178,131,38,146,42,136,201,10,
111,239,206,51,74,189,68,173,243,231,132,241,
230,3,157,196,205,189,1,94,16,162,41,50,
149,140,241,104,64,196,220,101,220,73,131,80,
76,104,218,115,200,167,22,9,242,121,246,170,
35,66,17,41,42,231,114,6,189,190,135,235,
133,76,21,77,134,125,159,209,248,254,121,45,
242,94,243,176,128,123,37,147,228,82,38,113,
24,165,231,12,13,113,79,176,118,125,111,22,
69,122,180,52,197,155,63,25,19,174,233,12,
221,20,185,223,9,240,133,198,192,78,163,171,
19,148,76,200,176,119,27,17,248,159,106,109,
243,115,207,161,52,162,225,109,237,137,199,207,
154,3,222,173,150,81,228,144,199,166,247,184,
156,141,243,82,57,199,230,200,97,49,97,241,
220,84,26,93,150,136,41,202,161,64,170,230,
58,148,252,183,63,245,51,252,255,58,163,200,
148,31,103,229,173,63,225,234,227,54,233,73,
15,117,120,139,213,210,2,91,254,51,39,63,
160,157,93,70,221,85,60,101,59,210,126,247,
203,236,141,215,168,152,6,103,147,49,222,110,
15,80,194,92,148,9,84,87,56,245,244,31,
227,217,35,234,27,223,38,60,243,30,141,157,
29,252,110,18,253,178,79,72,139,144,22,24,
18,56,6,152,71,104,129,167,115,41,10,134,
70,222,208,208,37,137,149,161,125,95,131,56,
105,29,105,19,237,181,23,168,228,54,31,126,
48,133,96,165,63,38,163,107,180,39,46,125,
87,198,113,115,60,94,234,115,37,147,160,225,
184,36,52,149,96,41,34,96,221,213,128,234,
184,71,27,37,97,246,73,152,255,52,10,178,
197,76,245,88,57,235,48,34,148,101,2,4,
186,44,83,74,12,57,159,142,211,30,232,52,
39,30,110,24,98,126,132,112,39,181,186,72,
246,132,154,62,207,217,214,107,81,164,61,27,
53,58,123,177,50,141,65,154,243,173,31,211,
219,238,176,189,57,34,12,162,204,33,25,87,
209,85,149,222,106,27,121,188,134,81,122,138,
187,133,150,48,20,172,173,13,185,112,33,29,
105,109,245,189,104,116,169,39,200,100,117,144,
160,217,116,89,94,78,66,54,133,91,73,211,
234,78,83,122,250,73,134,205,144,210,214,7,
8,77,101,121,252,46,202,118,29,38,30,190,
170,114,43,127,137,120,234,159,33,130,159,127,
253,58,245,127,228,43,207,75,196,54,70,200,
107,13,252,103,46,34,44,147,198,186,67,182,
242,201,37,162,47,253,248,111,185,254,194,111,
254,66,206,97,224,77,80,204,77,158,203,165,
185,245,174,140,185,240,159,81,91,253,59,36,
61,192,16,243,135,78,194,15,4,181,238,132,
102,123,16,233,144,1,51,89,243,196,51,237,
167,22,158,167,181,243,51,16,113,246,210,79,
49,91,89,39,181,190,139,55,9,233,245,92,
226,49,21,129,71,173,230,112,106,49,78,91,
149,88,93,25,114,254,124,10,93,63,218,103,
82,119,128,188,93,103,49,94,68,150,134,44,
222,19,188,73,128,33,43,236,12,21,58,189,
105,158,153,175,158,72,64,115,115,85,194,76,
202,84,230,85,86,58,179,200,132,199,122,126,
131,230,77,198,226,125,116,177,8,174,74,201,
124,9,213,136,159,120,141,117,169,194,132,40,
0,94,105,167,248,234,252,14,5,51,50,183,
137,3,39,247,249,169,244,97,63,47,125,183,
188,157,56,10,16,207,153,77,190,191,249,67,
178,197,167,254,227,114,20,154,149,32,59,245,
12,215,59,175,240,66,208,162,181,52,203,205,
221,115,148,79,95,60,241,61,26,245,191,195,
151,91,40,82,154,192,181,73,37,158,164,217,
185,205,29,109,140,47,4,126,87,37,83,136,
202,10,75,207,254,23,168,186,133,103,143,240,
68,11,101,156,66,42,244,113,210,130,65,191,
140,166,78,34,6,179,34,144,28,11,65,228,
40,52,89,34,111,70,77,167,125,219,165,104,
26,108,140,156,143,45,49,25,170,67,103,80,
196,208,108,108,55,142,101,12,239,27,102,52,
240,131,195,105,89,251,189,57,42,185,77,100,
76,4,80,48,142,15,89,105,77,188,99,78,
226,51,201,240,30,112,160,150,18,22,11,113,
147,81,16,28,162,76,166,45,147,82,230,193,
145,174,220,25,32,87,247,97,106,1,185,214,
188,219,1,37,156,47,83,235,102,56,23,188,
133,188,223,102,111,215,198,61,80,236,84,36,
137,205,205,17,249,177,79,189,238,48,59,27,
99,199,11,8,239,17,140,179,237,128,70,195,
65,202,79,240,253,16,119,18,253,109,171,21,
189,163,100,226,160,57,56,24,161,118,58,60,
53,254,33,43,195,243,168,227,1,141,73,142,
41,111,159,240,198,22,190,0,53,109,113,205,
184,140,149,253,99,172,68,129,113,123,239,231,
115,18,187,239,18,20,215,208,187,32,111,238,
17,166,19,188,231,133,44,24,33,115,179,26,
175,254,248,239,152,79,252,17,178,122,178,18,
84,102,127,135,59,207,124,229,23,114,6,91,
219,63,68,18,111,144,87,5,215,175,157,35,
177,240,235,200,138,74,229,204,175,51,108,158,
39,185,184,76,16,10,106,61,135,90,111,66,
48,154,128,21,25,72,77,145,41,36,79,94,
86,75,149,46,18,120,62,138,106,34,203,1,
213,238,14,198,88,97,123,187,199,244,116,36,
166,168,123,10,170,42,211,31,248,44,45,37,
216,223,119,216,221,179,89,92,136,211,246,178,
52,189,28,72,2,245,131,144,165,23,134,15,
108,56,44,39,108,222,222,151,248,131,229,234,
35,157,88,24,194,95,253,233,144,108,14,190,
242,91,81,70,29,42,123,220,218,91,100,246,
30,192,136,126,181,30,57,85,162,236,125,103,
117,151,233,185,63,64,143,101,79,248,237,143,
158,99,33,219,161,104,102,62,230,19,247,36,
150,134,126,40,70,58,148,76,180,79,193,58,
255,165,104,102,39,10,167,233,212,59,124,120,
89,226,182,235,195,216,64,8,113,162,94,5,
128,208,199,104,243,3,188,219,26,222,100,68,
232,122,136,113,150,205,120,27,177,91,64,119,
151,169,156,137,136,104,170,126,84,90,17,138,
141,136,245,33,128,20,19,228,80,195,246,13,
90,253,50,133,100,29,217,75,19,118,53,132,
57,192,51,29,234,206,132,25,203,164,233,122,
108,12,109,158,200,38,89,25,142,25,60,164,
113,157,75,213,241,3,13,207,215,201,167,106,
12,157,20,181,246,60,134,230,28,131,177,182,
6,37,130,64,61,36,192,213,28,151,210,196,
163,231,122,156,73,30,69,19,121,67,163,96,
104,52,39,222,103,231,40,14,234,184,87,178,
9,122,174,79,115,18,205,15,87,101,137,180,
28,109,141,238,56,78,202,180,15,200,110,15,
56,120,165,60,82,173,133,149,145,89,219,88,
96,201,220,68,222,111,227,186,50,9,219,71,
241,118,241,220,16,123,28,96,153,10,133,41,
3,207,21,212,234,54,195,3,97,185,86,219,
195,49,71,247,141,100,244,253,104,188,104,42,
169,49,232,249,135,81,45,16,57,158,73,72,
58,171,35,53,87,163,95,238,175,226,91,73,
234,102,133,102,43,90,223,133,115,5,94,113,
159,36,179,252,95,159,216,112,63,234,26,140,
223,225,229,217,16,253,250,70,100,68,102,75,
76,5,1,237,161,205,181,117,19,85,77,157,
248,127,165,154,123,104,206,24,183,56,251,153,
158,187,65,227,54,195,254,53,210,210,10,174,
40,18,230,126,155,84,226,200,232,73,178,66,
178,184,76,115,224,178,213,30,227,7,247,7,
17,73,75,61,113,54,113,88,130,154,189,194,
164,223,100,235,195,63,199,47,61,134,117,70,
144,24,189,141,105,201,52,219,19,242,121,131,
76,90,163,221,113,201,101,116,138,69,51,234,
19,250,2,213,222,103,42,168,147,72,168,200,
23,23,9,147,247,115,138,228,173,61,174,238,
119,144,206,46,96,7,22,113,85,121,196,158,
135,116,86,240,149,223,74,210,243,124,26,142,
199,110,47,195,84,166,250,64,146,236,225,223,
157,174,178,122,231,47,152,47,191,76,60,187,
240,200,239,237,5,30,251,221,25,4,18,67,
99,192,186,101,115,42,97,157,204,191,88,38,
140,29,10,147,14,150,248,71,236,126,156,68,
225,212,127,100,142,34,191,128,221,223,100,213,
11,120,34,105,225,95,188,195,230,238,123,100,
103,158,56,97,83,197,66,184,32,143,74,108,
127,248,103,168,41,157,112,162,194,190,69,86,
255,10,153,249,251,245,53,60,103,8,70,180,
201,36,5,80,70,132,140,208,251,73,10,228,
104,212,151,17,214,144,169,100,19,236,56,8,
184,214,25,147,210,52,102,45,131,214,196,165,
104,234,220,236,143,30,250,110,197,65,109,95,
85,188,99,101,162,177,147,100,187,113,26,67,
115,16,66,38,21,111,99,233,35,114,186,198,
200,151,81,36,137,125,199,229,194,65,79,66,
26,140,17,7,14,99,46,102,126,102,142,162,
209,157,38,155,108,28,206,120,152,55,67,132,
44,31,143,114,132,160,62,72,115,62,181,142,
84,31,17,150,114,72,157,62,82,111,24,77,
18,76,88,136,132,69,240,228,121,166,99,99,
236,186,79,173,94,164,172,239,179,161,157,227,
116,231,85,108,59,96,52,242,169,156,74,179,
229,150,208,7,235,216,163,227,153,82,224,6,
56,138,202,108,209,100,48,240,176,237,0,213,
48,217,211,47,115,94,219,37,94,52,49,12,
133,181,141,33,186,42,227,250,33,197,130,137,
61,9,168,215,109,12,67,33,244,5,131,129,
71,82,12,200,171,54,106,222,132,39,151,248,
94,245,2,149,179,127,244,200,233,139,159,172,
89,233,98,170,74,212,135,73,198,200,199,45,
226,239,222,194,207,167,185,227,252,42,229,11,
39,67,61,197,250,29,84,215,161,49,127,246,
51,118,18,43,236,173,255,59,102,166,44,72,
189,76,60,182,124,223,103,38,126,200,70,99,
76,207,126,248,126,27,58,209,208,161,79,10,
6,210,227,25,228,156,77,203,189,70,50,60,
207,242,227,123,4,251,117,226,113,5,77,147,
208,117,133,124,222,160,223,247,201,229,52,6,
182,143,51,9,24,12,124,134,67,143,139,23,
211,40,235,85,164,254,144,112,166,20,17,45,
21,57,98,228,251,1,210,200,230,137,149,173,
8,176,82,46,16,206,22,63,246,33,191,250,
141,8,250,234,133,130,155,253,17,158,151,229,
233,178,77,115,162,209,117,189,7,242,159,29,
55,134,63,61,100,127,252,231,196,214,158,162,
120,234,43,15,149,217,176,123,187,248,210,26,
197,76,148,185,102,228,41,182,218,5,132,163,
83,74,245,137,235,19,36,47,64,136,16,244,
35,248,190,100,79,16,186,134,208,14,178,101,
199,229,139,234,7,188,231,15,105,183,254,144,
84,229,236,47,183,163,16,66,208,220,248,46,
19,183,139,101,228,145,100,139,244,68,195,84,
4,229,90,149,59,227,15,128,19,58,138,32,
139,215,221,135,48,140,116,223,219,55,153,190,
248,219,248,238,24,35,145,123,112,169,96,255,
167,24,75,25,66,71,38,52,122,135,57,159,
170,164,9,226,155,20,82,145,70,211,126,119,
6,211,24,145,177,243,120,177,22,175,53,122,
188,84,206,242,68,54,197,187,157,193,161,244,
247,253,37,28,137,140,174,210,158,68,155,236,
222,225,70,161,144,201,37,247,143,169,189,78,
91,6,83,166,70,92,85,248,105,163,199,36,
8,145,37,9,169,59,68,222,111,19,36,231,
15,179,138,43,153,4,31,246,70,71,101,47,
33,163,216,83,72,146,134,16,130,208,245,8,
205,38,146,121,114,109,249,86,191,76,204,28,
160,42,30,110,40,69,189,135,145,141,122,125,
45,106,6,167,19,72,253,33,123,133,139,76,
197,122,72,215,214,240,155,125,140,132,206,126,
117,72,62,111,160,172,86,33,102,32,166,178,
132,133,12,202,118,13,127,223,71,147,36,208,
84,188,76,6,127,51,0,9,134,67,143,85,
107,25,69,248,244,187,238,125,7,49,157,214,
80,11,6,37,45,134,109,15,136,205,204,209,
23,89,158,202,220,161,190,231,208,240,66,6,
35,31,66,56,117,58,129,101,41,132,11,21,
132,105,32,239,54,144,250,104,183,232,196,0,
0,32,0,73,68,65,84,35,198,182,143,61,
9,88,94,78,34,146,49,106,51,37,94,191,
115,137,185,11,255,252,231,210,205,121,96,22,
37,228,35,118,167,231,19,127,227,58,194,208,
240,147,9,68,231,100,44,106,57,12,48,198,
125,218,211,159,93,164,56,216,191,77,237,214,
15,88,122,246,63,101,225,210,127,133,17,143,
206,199,189,242,229,2,168,247,28,118,218,206,
177,140,237,65,151,235,135,52,135,19,166,146,
198,39,122,14,73,81,153,46,253,75,236,126,
149,120,102,158,155,235,215,121,33,215,1,63,
164,221,157,16,179,20,82,105,147,209,200,103,
60,14,48,13,133,149,149,1,17,125,70,98,
48,240,201,105,50,114,189,141,92,111,71,160,
5,67,67,20,115,132,59,109,100,33,144,134,
99,80,85,148,59,91,136,108,18,145,120,52,
76,54,165,169,36,189,75,124,225,76,141,156,
17,227,28,96,7,33,155,35,135,158,231,71,
80,254,145,76,123,80,68,149,125,10,233,61,
136,131,227,253,144,245,245,119,35,161,81,79,
33,147,120,145,204,204,229,3,91,23,210,216,
253,62,242,242,30,57,93,35,103,168,88,138,
205,108,44,60,100,139,203,91,53,148,245,42,
254,249,83,136,82,244,78,164,221,38,114,173,
201,181,90,156,90,175,192,213,84,64,126,206,
64,110,245,120,124,184,194,74,242,127,227,214,
141,171,100,102,94,196,76,125,122,61,170,32,
20,216,110,84,73,248,36,179,208,62,115,71,
33,194,128,198,202,171,136,212,251,4,211,93,
70,213,41,230,78,253,183,244,107,25,222,175,
191,66,225,170,198,108,115,157,173,149,111,145,
206,61,26,178,104,234,37,194,226,187,132,141,
128,192,157,80,152,121,145,126,227,38,146,144,
209,227,25,134,205,85,70,189,219,184,82,157,
192,119,144,39,69,148,98,15,208,145,253,36,
98,148,64,228,171,7,181,242,144,233,88,52,
234,243,189,206,128,98,118,135,90,103,142,138,
25,48,58,104,54,245,188,128,141,145,77,221,
113,177,20,153,243,233,24,215,58,227,67,173,
38,136,102,70,180,38,30,178,36,17,87,20,
4,33,26,50,9,85,102,191,155,160,152,57,
174,72,121,57,147,32,20,2,89,138,178,134,
148,166,32,121,1,202,7,119,8,174,68,209,
222,205,254,136,221,177,139,29,4,136,113,2,
77,202,17,134,19,2,169,79,16,171,31,221,
204,2,6,73,208,6,15,85,179,61,22,237,
184,113,20,217,63,116,92,119,33,184,34,25,
39,88,158,71,253,96,5,234,45,130,211,179,
244,130,36,23,86,127,140,219,236,177,182,54,
196,117,67,124,95,176,187,107,147,203,233,204,
206,10,100,187,134,212,232,32,217,19,146,170,
74,221,45,34,180,49,129,35,176,98,10,187,
123,54,173,216,12,227,212,19,44,245,190,67,
160,201,120,222,209,218,25,166,66,79,157,33,
116,97,99,123,64,177,104,176,183,191,143,105,
111,210,81,45,42,51,49,6,3,143,169,146,
137,51,14,162,153,8,138,66,88,204,31,100,
52,49,164,159,125,72,163,49,161,144,51,144,
52,153,224,210,41,214,6,1,104,67,234,171,
127,79,241,212,203,12,26,183,113,39,45,114,
149,103,14,140,232,42,146,172,146,40,44,124,
226,61,237,133,26,98,18,181,222,165,3,137,
146,215,228,69,58,59,143,81,168,60,199,234,
254,136,148,165,126,172,65,61,247,218,63,112,
231,233,207,182,47,33,73,6,11,87,255,0,
69,183,80,244,227,101,143,32,16,180,122,19,
26,131,201,161,225,56,201,85,237,56,20,18,
198,39,206,42,70,157,13,36,85,193,72,22,
241,226,6,222,208,103,107,107,132,174,41,228,
178,50,190,23,160,27,50,137,184,74,40,4,
229,178,69,173,102,35,132,160,90,29,211,233,
184,204,205,197,162,38,183,16,72,142,139,180,
85,227,205,193,85,158,78,188,7,165,28,34,
25,143,102,126,59,46,34,30,59,214,4,8,
67,25,89,62,30,76,41,146,194,98,42,36,
103,28,125,208,82,100,206,167,98,116,237,24,
181,94,134,225,68,227,115,75,183,241,195,144,
155,253,40,0,148,52,80,231,35,14,85,176,
157,33,93,185,112,148,153,141,218,4,241,72,
246,38,161,41,156,78,196,14,165,100,36,219,
65,185,185,129,116,32,217,47,5,97,68,225,
240,2,214,223,233,113,167,87,34,23,115,137,
101,146,188,61,201,241,252,112,5,235,220,34,
202,214,30,139,41,159,129,182,194,250,45,151,
133,103,254,248,19,239,133,106,199,166,53,244,
112,238,41,157,155,254,136,139,137,204,137,160,
175,159,185,163,168,221,249,14,210,255,75,221,
155,254,72,114,231,103,126,159,95,220,17,121,
103,221,119,85,87,223,108,54,217,36,135,156,
25,205,80,115,172,180,154,93,239,74,90,93,
182,86,182,97,96,129,5,236,23,6,252,23,
248,165,13,248,133,1,191,177,12,99,109,11,
150,229,181,181,150,4,105,101,105,36,205,104,
36,206,12,57,188,217,236,171,186,171,235,62,
178,242,206,140,200,140,251,231,23,145,93,221,
213,7,187,123,134,99,97,2,32,8,54,59,
35,51,35,35,190,199,243,253,62,207,179,176,
135,90,10,208,252,25,226,208,65,8,133,210,
204,5,142,54,118,233,185,239,176,172,38,212,
199,62,197,239,205,225,84,159,156,45,253,126,
11,69,205,211,235,84,200,175,222,161,231,30,
208,27,38,136,66,136,223,40,211,184,241,125,
212,217,67,212,229,145,205,97,100,19,4,1,
197,124,139,99,176,195,182,160,83,65,51,115,
84,74,117,78,23,28,246,31,208,149,81,133,
160,90,110,226,183,166,217,31,246,168,234,58,
113,42,169,24,89,91,186,233,6,153,192,217,
168,26,59,9,63,73,28,77,97,194,50,41,
234,26,113,42,57,42,116,78,48,158,1,106,
126,192,130,147,109,110,132,105,202,148,237,32,
106,205,76,122,123,36,215,61,103,91,108,55,
109,68,172,128,221,4,213,227,114,41,207,135,
237,225,35,184,151,40,244,33,176,80,250,85,
210,242,3,106,154,143,81,203,245,252,34,57,
243,228,230,84,63,74,48,77,5,165,231,30,
127,43,113,103,23,43,95,38,62,59,135,225,
13,57,119,78,225,198,141,30,113,156,5,235,
122,61,99,215,198,177,228,226,197,18,134,161,
160,139,152,72,102,183,212,11,202,85,234,158,
195,64,81,168,23,95,70,109,125,194,234,184,
203,173,195,209,166,138,173,225,217,11,12,115,
21,10,225,30,78,107,135,238,32,102,119,55,
97,113,49,135,229,148,81,151,167,145,182,137,
62,128,186,62,205,92,165,155,65,115,205,46,
200,20,101,99,31,229,176,193,214,182,135,109,
171,24,186,66,20,166,132,239,223,230,181,233,
10,222,27,54,55,214,154,236,126,188,77,172,
9,138,151,14,168,221,94,167,60,241,75,164,
73,64,109,227,79,56,59,254,95,61,247,61,
109,171,2,101,4,9,74,211,224,93,125,158,
161,255,79,152,90,185,72,211,13,105,121,30,
139,99,39,3,115,156,74,92,63,70,83,5,
147,181,93,110,189,241,11,164,234,79,225,241,
75,99,182,63,252,19,22,95,253,53,242,19,
39,147,96,156,164,244,252,152,70,211,163,61,
8,65,203,101,234,192,207,113,132,113,74,173,
23,48,93,122,190,174,34,145,46,65,180,193,
24,175,227,250,121,62,48,47,176,252,98,155,
66,227,128,126,63,68,246,5,174,23,97,153,
42,19,147,22,227,227,6,249,188,198,246,182,
71,24,166,132,97,72,154,74,38,39,45,74,
165,251,112,205,23,242,31,177,21,204,19,222,
213,41,22,99,166,163,189,140,197,253,0,147,
123,175,83,101,24,26,168,106,138,23,88,156,
159,222,163,83,15,120,231,251,17,151,222,208,
31,249,172,119,27,147,24,210,231,214,223,222,
225,155,255,65,142,188,147,125,215,9,203,224,
131,86,159,110,116,31,58,85,229,73,114,162,
149,31,71,116,28,160,207,225,48,164,29,198,
252,220,68,9,77,8,212,171,235,89,231,115,
239,209,245,3,68,24,161,94,91,231,140,221,
103,222,232,242,110,127,137,215,212,235,120,227,
211,216,50,68,86,139,184,229,50,127,243,183,
54,154,30,81,152,57,224,96,237,219,84,231,
94,59,238,16,159,218,89,250,49,123,237,71,
85,28,252,40,97,175,61,100,105,66,255,135,
79,20,138,106,32,81,16,26,36,218,1,194,
24,31,5,253,58,126,184,203,123,126,196,132,
174,208,47,14,72,15,174,81,229,209,213,194,
222,209,45,58,205,31,145,148,55,144,137,129,
238,64,221,91,192,178,7,199,218,69,201,68,
138,22,196,168,249,251,85,67,187,63,241,168,
92,133,234,35,212,34,177,189,75,45,130,163,
90,112,34,224,167,82,178,19,4,232,138,198,
218,136,185,89,210,53,46,149,243,188,221,232,
114,185,156,39,150,146,187,238,144,67,63,51,
191,17,192,197,114,142,49,195,192,139,99,130,
68,82,213,245,145,189,99,135,253,246,212,113,
162,48,85,133,7,39,1,103,138,14,182,170,
32,154,221,204,17,110,84,170,21,116,149,85,
7,110,196,217,235,166,44,243,184,50,177,84,
133,146,174,209,8,34,46,150,114,196,169,228,
70,207,67,232,29,68,144,39,85,7,232,209,
28,105,18,146,26,45,48,238,227,206,227,197,
3,186,222,24,238,176,196,68,121,159,179,69,
39,235,104,90,93,148,141,125,154,205,128,122,
61,96,97,193,33,214,65,109,117,24,180,134,
212,235,1,190,127,178,242,12,70,155,72,235,
235,46,23,46,20,71,185,73,32,6,62,250,
214,14,69,67,240,142,60,135,134,96,89,28,
112,227,122,143,52,149,44,159,155,164,45,166,
152,74,14,216,219,216,34,82,96,113,53,79,
146,72,230,231,29,204,138,67,125,252,44,187,
225,4,86,163,129,173,4,84,138,13,110,220,
174,226,229,151,121,249,165,54,170,34,209,62,
184,9,73,202,236,156,13,18,134,126,194,208,
141,232,187,3,186,159,212,57,247,181,136,55,
206,45,146,92,244,241,226,132,31,54,82,202,
75,62,141,91,255,47,194,210,209,198,229,143,
37,233,156,164,57,24,13,77,131,74,137,218,
33,84,102,199,72,82,201,118,115,72,217,214,
209,71,195,255,182,23,113,216,245,233,251,89,
112,49,226,16,109,175,142,122,86,65,255,156,
159,181,221,171,127,202,228,169,55,153,60,251,
230,253,223,40,74,105,244,124,214,182,90,28,
28,118,73,226,20,37,29,160,105,26,139,103,
151,81,149,231,103,31,239,119,134,76,20,140,
231,122,109,161,124,14,119,125,3,86,192,89,
252,77,250,237,29,90,226,143,24,30,109,208,
106,100,182,195,185,188,198,220,156,205,96,152,
176,182,214,103,98,194,68,121,224,61,122,189,
136,40,74,41,20,138,199,127,46,132,100,217,
202,86,188,91,227,231,184,157,251,6,203,211,
45,130,88,231,163,189,101,28,61,100,185,122,
196,92,249,62,65,245,118,125,134,219,223,223,
100,252,252,191,228,230,251,127,192,210,220,253,
164,119,231,104,154,217,114,27,199,8,112,47,
42,168,74,194,61,186,89,78,83,249,210,68,
137,111,31,180,142,97,58,153,8,246,175,255,
5,186,109,83,154,124,25,84,13,145,106,199,
69,160,34,192,79,82,242,170,138,72,78,62,
63,202,230,126,150,44,70,29,134,173,38,188,
84,104,176,29,172,80,94,45,242,105,238,117,
44,55,100,125,183,201,204,217,62,175,185,91,
224,216,120,171,63,226,175,62,168,177,244,194,
179,117,22,65,244,100,88,186,214,13,24,43,
102,146,44,63,149,68,17,198,41,246,51,44,
118,76,174,188,201,145,251,135,192,112,20,40,
45,164,148,236,92,255,3,46,127,109,192,166,
7,197,156,197,248,48,224,64,171,83,219,248,
54,99,115,95,61,222,88,170,111,125,15,175,
240,61,148,165,20,69,36,8,61,66,227,62,
195,216,245,139,228,173,30,166,62,36,28,148,
79,108,206,60,184,79,45,165,66,16,89,88,
198,0,69,227,216,237,78,62,166,250,86,252,
113,34,107,239,248,207,23,115,22,5,77,225,
114,57,79,78,87,241,162,152,86,16,158,16,
21,187,222,245,56,157,151,228,52,149,130,174,
34,4,236,187,10,183,14,103,143,37,191,1,
130,81,187,121,239,45,77,69,201,236,66,253,
0,89,185,207,74,245,122,41,43,147,10,55,
70,13,194,233,130,131,169,40,20,117,141,43,
213,2,5,77,37,145,242,24,247,244,211,148,
221,193,16,25,150,80,41,16,219,217,195,163,
14,38,73,212,38,168,247,111,210,82,174,73,
16,230,57,106,47,161,135,41,103,243,3,212,
245,93,14,14,134,236,239,15,41,151,13,114,
57,141,188,91,227,163,239,244,144,157,207,230,
110,12,6,49,119,238,244,57,117,42,143,37,
2,154,81,149,49,189,197,247,26,11,40,86,
129,153,163,239,83,82,123,228,78,229,25,184,
49,145,235,98,39,61,142,186,209,113,149,168,
42,130,106,213,36,150,64,4,27,155,38,95,
88,184,138,226,143,180,134,6,80,230,46,9,
121,62,186,251,21,46,159,110,146,44,205,160,
52,58,24,163,37,131,94,63,226,112,223,39,
24,65,41,238,250,17,183,198,74,180,20,133,
25,219,100,210,50,184,214,49,152,112,92,82,
49,139,82,240,8,251,45,204,226,248,115,222,
253,83,132,94,132,13,124,167,179,66,117,254,
183,33,77,217,239,248,68,73,138,162,104,15,
52,117,242,56,73,88,161,207,234,193,26,87,
167,78,99,29,184,92,152,45,156,244,34,255,
9,231,128,147,43,95,165,56,123,154,40,73,
217,111,251,220,218,106,178,189,221,164,211,242,
40,149,115,44,175,78,226,228,76,8,71,75,
25,218,143,55,187,137,19,73,173,27,48,91,
121,118,135,52,167,178,72,121,44,43,2,53,
211,33,63,177,202,225,225,37,236,5,88,206,
221,100,111,171,151,249,82,40,130,156,163,177,
188,156,195,52,21,124,63,219,150,107,119,178,
174,127,56,76,56,56,24,50,51,99,159,72,
34,0,213,131,91,228,74,13,110,138,47,225,
11,155,156,238,115,105,118,231,145,207,114,102,
226,128,213,95,49,249,243,63,254,35,180,220,
50,3,119,15,39,175,80,119,139,76,21,187,
56,70,54,103,58,127,217,122,236,226,202,162,
99,178,229,249,72,64,206,110,17,172,3,193,
4,253,250,77,202,179,47,161,203,25,98,246,
142,161,221,68,74,68,16,194,48,160,22,77,
208,143,243,140,235,45,202,90,23,113,120,82,
164,179,164,246,40,57,61,216,130,169,226,45,
210,83,115,156,42,121,40,166,142,114,228,67,
171,135,57,94,126,196,152,237,179,151,47,30,
197,164,151,199,29,204,40,196,200,149,24,70,
41,79,241,117,67,53,12,227,191,254,199,191,
243,95,30,15,185,30,86,197,124,220,209,233,
123,220,60,28,208,30,102,76,77,219,80,159,
208,5,135,28,110,254,123,114,179,53,98,37,
37,57,114,176,184,140,149,155,195,245,110,114,
110,57,68,8,193,134,235,99,105,42,126,193,
35,210,119,80,218,51,88,197,41,146,208,167,
209,255,51,148,9,23,37,44,128,30,28,7,
115,75,85,16,34,101,24,100,190,13,170,146,
208,233,172,146,43,28,80,212,85,76,85,69,
40,1,245,222,52,158,95,228,160,181,132,101,
12,233,120,227,12,221,49,250,177,137,231,151,
178,127,130,34,113,98,96,25,131,172,160,239,
85,153,176,108,210,176,128,28,20,208,108,151,
122,16,113,182,232,32,128,24,50,61,252,135,
142,165,188,197,188,99,98,107,42,55,91,42,
111,237,171,76,148,247,70,85,9,136,208,66,
4,38,78,98,81,206,75,234,65,196,90,127,
64,201,212,49,4,200,137,42,24,58,253,110,
138,105,11,52,77,48,97,25,180,195,136,89,
199,196,86,21,22,115,214,49,191,225,193,85,
197,73,203,200,240,80,51,160,149,184,199,146,
36,82,247,208,252,121,82,61,11,246,227,166,
206,11,186,198,235,210,227,149,170,207,180,38,
233,237,37,168,189,6,157,67,15,223,207,160,
37,33,160,185,91,135,201,51,168,110,141,52,
253,236,201,87,16,164,244,251,49,11,227,33,
219,225,18,159,182,44,250,84,185,98,92,103,
177,56,164,221,9,81,4,236,31,12,25,12,
98,10,5,157,114,217,64,215,21,138,37,131,
118,39,202,148,69,253,20,75,73,25,134,38,
21,195,69,125,88,118,62,12,193,48,233,170,
37,242,179,153,58,173,156,168,128,174,99,14,
50,47,234,36,78,89,57,149,167,88,54,176,
22,167,89,40,231,153,115,76,134,73,202,158,
151,176,50,5,237,77,147,177,213,54,253,125,
141,124,101,245,241,1,49,24,160,89,143,146,
172,204,220,52,7,141,15,89,72,90,108,24,
167,40,78,125,21,207,243,216,234,201,209,28,
40,161,239,199,228,45,141,130,165,227,71,41,
195,48,97,188,119,196,90,156,99,239,176,129,
98,23,25,196,146,146,163,63,49,89,164,35,
173,39,85,127,122,69,38,132,160,59,12,217,
115,5,235,7,46,63,124,103,157,173,141,58,
154,166,114,230,194,44,11,75,227,232,198,40,
129,37,209,189,167,255,25,178,194,16,244,71,
215,58,251,65,76,42,161,96,105,8,33,72,
66,31,69,55,78,216,215,62,2,217,149,103,
16,66,32,147,24,41,83,156,137,151,72,220,
15,177,246,239,50,28,102,182,170,150,169,96,
59,26,158,27,35,133,196,182,53,92,89,38,
201,77,34,189,140,211,224,186,49,119,228,139,
180,147,49,12,61,51,70,58,174,124,131,33,
147,241,30,193,212,52,51,197,14,134,22,63,
225,122,65,235,104,136,55,172,208,173,215,88,
88,49,200,25,1,166,246,217,220,165,157,65,
192,90,255,1,248,72,147,224,184,40,195,9,
114,165,51,232,102,158,65,247,144,186,18,48,
8,10,116,135,37,124,191,76,74,133,134,95,
166,52,56,98,206,60,196,80,194,19,168,112,
186,56,157,25,126,61,240,156,137,32,202,164,
241,59,253,204,78,88,130,172,22,241,166,171,
108,236,230,41,142,63,202,63,123,92,12,215,
85,133,90,47,56,62,181,166,10,198,11,38,
3,207,165,23,41,116,134,201,103,206,211,254,
236,223,252,119,63,94,71,209,31,65,16,131,
48,225,118,205,101,121,220,97,178,104,62,110,
221,129,56,238,227,55,198,177,184,72,117,246,
117,52,211,225,104,253,45,52,171,79,195,139,
112,101,76,152,166,132,35,93,37,117,239,2,
229,243,217,6,193,238,250,223,80,31,115,160,
99,51,163,75,80,122,104,66,176,100,105,76,
107,58,157,129,202,119,98,139,90,103,158,177,
66,141,233,98,147,131,246,2,105,46,224,151,
150,50,94,128,152,236,178,238,14,89,59,146,
148,114,77,74,185,38,74,47,66,26,39,171,
228,56,214,56,220,61,75,213,10,112,205,62,
99,118,11,93,130,176,96,82,201,35,204,251,
21,121,55,124,252,10,161,46,4,110,28,243,
214,62,52,125,73,181,248,64,181,16,169,8,
52,82,199,229,46,30,91,183,230,72,138,77,
148,206,44,3,2,114,179,147,25,108,35,179,
133,26,221,200,110,163,138,161,157,224,152,124,
86,253,169,180,187,172,84,75,24,138,114,236,
113,13,16,105,251,168,65,9,172,30,175,84,
11,88,181,22,245,239,221,34,142,83,74,37,
157,86,178,138,223,110,208,110,103,129,41,138,
82,118,118,6,88,182,134,170,235,244,74,231,
73,82,129,215,15,168,196,219,104,50,251,123,
49,26,9,42,34,171,173,192,139,216,221,15,
184,19,244,113,165,195,171,185,107,20,241,105,
119,98,124,63,161,211,9,41,20,116,150,151,
115,236,238,14,168,84,12,170,85,131,91,55,
123,196,177,68,10,201,216,152,73,46,167,33,
251,2,221,125,124,39,51,219,190,197,71,206,
60,179,227,217,3,43,243,14,236,212,208,52,
65,249,133,25,138,177,36,218,174,147,86,138,
168,142,117,252,128,184,113,66,24,86,169,218,
45,14,237,10,227,169,193,122,122,244,220,247,
191,170,155,116,147,51,72,123,19,25,100,195,
215,221,230,0,249,0,57,170,55,140,185,190,
215,231,210,124,145,213,201,28,182,174,34,203,
43,252,179,247,254,134,255,85,159,224,195,31,
173,115,241,242,34,195,48,97,166,108,50,85,
180,208,212,167,119,23,97,156,210,247,99,188,
32,147,122,209,20,193,32,76,232,14,34,106,
155,71,180,253,14,237,150,135,109,233,188,240,
210,18,229,106,142,159,198,33,37,28,116,124,
218,94,200,249,153,103,83,90,125,152,43,21,
122,29,166,6,91,180,154,1,250,104,201,97,
107,107,144,17,240,220,8,215,141,57,123,174,
136,123,216,103,144,155,32,169,92,34,109,239,
211,102,138,241,218,26,115,11,17,251,254,21,
198,242,39,117,207,164,99,145,132,144,175,248,
153,55,251,214,1,233,226,52,50,119,50,225,
125,241,231,109,174,189,191,193,135,63,136,121,
237,231,210,19,108,240,199,126,103,96,211,123,
20,235,87,114,49,201,202,7,212,175,53,152,
86,255,57,177,178,207,88,41,211,81,59,95,
116,88,53,251,40,66,65,77,110,34,244,1,
168,10,194,208,145,73,10,134,6,113,146,205,
85,130,8,165,213,67,22,108,68,199,29,233,
160,75,210,133,105,48,117,54,236,50,55,239,
22,240,255,190,128,250,156,161,187,104,233,180,
188,240,184,35,92,59,116,97,56,4,83,231,
252,220,211,73,132,63,86,162,112,30,186,160,
219,205,33,99,249,71,49,75,69,85,153,89,
250,85,156,202,20,60,128,3,79,174,126,133,
214,158,197,181,247,254,14,253,252,0,219,80,
40,40,42,205,3,13,117,160,176,125,243,223,
146,136,14,162,220,101,198,202,33,82,157,125,
79,101,202,206,241,45,191,129,210,142,232,174,
204,114,122,208,160,63,155,109,27,109,237,46,
242,229,222,199,228,205,132,91,250,18,63,220,
43,240,230,194,16,49,170,182,245,52,38,76,
38,152,200,123,212,67,133,200,9,24,196,153,
55,244,149,74,129,118,24,161,230,215,50,175,
235,241,45,122,163,92,240,106,181,128,57,20,
148,116,19,63,77,233,71,9,219,94,128,174,
8,20,33,208,132,56,118,201,186,221,31,50,
109,25,52,6,21,202,133,221,17,140,53,137,
34,45,146,212,37,125,64,200,47,41,236,161,
13,102,177,202,13,42,149,2,3,79,34,148,
44,162,25,102,22,122,131,36,165,19,198,244,
163,209,218,233,103,37,137,195,38,234,157,29,
226,11,203,204,149,139,76,76,148,168,71,9,
219,195,128,38,17,50,72,153,12,199,72,253,
20,209,247,232,118,67,92,55,99,70,163,221,
224,80,76,83,230,228,60,71,85,64,236,127,
128,230,129,38,19,28,153,80,103,30,137,96,
146,29,82,84,204,17,155,61,69,16,97,208,
172,15,152,23,159,34,37,232,154,129,40,152,
199,112,86,181,106,208,235,69,220,188,217,67,
81,96,119,119,192,234,106,33,115,142,75,37,
187,187,67,90,141,144,153,41,155,32,126,2,
130,175,40,72,67,35,109,246,72,87,4,202,
232,154,137,32,202,252,47,166,199,232,255,221,
117,138,142,10,154,202,90,127,136,151,36,20,
149,60,111,239,77,243,210,236,30,219,235,38,
19,167,190,74,236,93,197,212,234,207,69,248,
188,119,24,185,203,252,121,99,131,241,149,223,
96,163,62,200,20,132,31,122,154,226,84,210,
232,135,204,86,172,99,152,230,253,43,223,224,
23,63,248,14,159,92,190,196,208,210,73,165,
100,175,237,179,223,241,113,12,13,199,80,113,
76,149,188,169,162,75,136,211,148,174,27,210,
31,198,244,252,136,32,74,9,195,152,86,163,
143,63,8,169,84,243,148,70,201,32,87,180,
209,203,5,150,87,167,158,251,251,252,184,135,
31,165,220,216,239,179,90,74,159,59,152,4,
94,147,77,243,28,231,95,9,153,50,3,214,
63,109,210,239,71,116,122,33,227,227,38,139,
11,57,132,2,203,147,112,55,209,112,90,55,
169,85,206,176,212,254,56,171,238,119,96,126,
233,42,159,138,75,89,87,33,32,146,58,246,
214,16,117,251,38,202,101,32,78,80,14,26,
8,119,64,178,186,128,172,222,135,119,111,95,
243,185,112,197,226,230,213,136,250,94,204,220,
202,103,119,88,2,88,205,219,124,210,233,243,
184,6,91,172,238,178,113,227,223,18,76,106,
20,99,19,199,8,25,55,13,212,195,70,102,
220,229,135,224,88,196,171,115,104,215,54,144,
213,34,105,181,8,150,73,90,200,145,204,130,
178,152,37,5,225,7,25,199,66,74,164,109,
209,236,196,220,186,246,34,227,11,63,71,71,
252,46,81,84,125,174,107,189,48,102,211,247,
99,162,135,86,252,21,33,200,91,234,79,39,
81,20,109,29,8,79,12,128,235,253,240,201,
155,16,15,13,11,163,97,151,110,240,183,228,
231,3,162,205,243,68,154,73,83,153,100,204,
156,166,244,234,25,54,214,255,123,244,153,148,
84,166,164,102,13,67,81,120,41,87,230,147,
206,24,173,74,196,196,224,144,210,250,30,74,
163,205,149,48,98,24,43,228,53,141,157,185,
83,188,112,120,157,213,141,171,24,243,47,112,
51,103,115,161,154,160,72,137,105,116,216,61,
90,224,171,179,146,153,216,167,84,173,208,13,
99,54,61,159,146,161,210,142,34,20,96,170,
178,155,109,6,141,152,211,55,122,3,190,52,
94,98,232,38,124,218,139,233,168,46,41,146,
175,76,148,121,187,209,227,75,19,37,222,174,
119,9,101,74,59,140,152,182,12,130,88,220,
47,65,100,76,108,63,234,255,144,215,85,74,
185,38,211,163,33,181,94,16,15,109,34,197,
220,238,15,143,197,207,222,107,245,120,181,90,
124,68,107,233,248,18,55,218,16,70,104,31,
223,6,85,69,79,83,236,229,25,230,39,171,
180,139,57,214,250,3,150,115,2,77,213,233,
231,171,88,197,28,109,119,136,78,8,113,112,
34,73,68,232,164,8,66,15,66,163,128,46,
134,232,105,66,136,65,137,35,140,227,223,254,
254,211,162,32,73,144,196,232,8,41,73,133,
194,216,116,1,93,137,201,231,53,102,103,109,
186,221,8,199,209,112,221,172,42,203,229,50,
79,3,161,10,20,85,48,24,196,232,186,194,
93,111,154,11,165,199,123,102,36,103,23,81,
54,246,185,100,172,113,107,247,85,206,47,102,
112,68,114,110,9,105,153,104,239,92,165,82,
49,51,39,189,195,38,175,152,6,191,31,47,
18,38,42,95,88,56,224,229,74,129,191,185,
109,146,55,28,210,129,96,97,197,103,127,231,
3,170,243,207,174,167,227,181,54,81,84,135,
210,220,127,66,109,168,209,116,159,236,11,210,
30,132,39,176,252,32,78,105,94,184,204,194,
65,155,35,221,226,30,176,150,196,146,247,62,
217,192,31,132,44,173,78,82,29,47,32,34,
47,43,42,141,132,110,199,163,219,26,48,28,
6,120,125,159,169,217,10,11,203,19,168,15,
204,25,116,93,69,183,45,254,255,62,130,56,
101,187,49,224,124,190,248,92,175,107,29,189,
67,108,12,120,59,124,149,98,174,202,196,197,
143,168,172,125,194,204,180,141,4,90,237,144,
70,221,103,114,194,34,25,14,201,153,9,211,
122,141,7,249,217,67,55,224,210,248,205,147,
157,74,106,96,40,33,114,199,38,121,97,149,
228,146,137,122,107,19,165,213,37,121,32,81,
220,185,145,85,131,150,45,201,151,63,59,88,
134,62,24,22,204,59,38,142,166,240,110,179,
247,136,255,76,170,171,248,167,52,198,75,7,
212,218,11,188,118,170,77,217,208,72,231,38,
81,54,15,64,85,72,206,44,34,199,74,200,
82,29,209,233,147,94,90,101,99,16,176,182,
223,33,222,158,195,208,85,36,9,142,26,145,
237,187,171,248,161,70,146,142,51,115,250,23,
57,184,243,39,84,202,41,138,243,149,231,186,
214,166,166,112,97,54,207,70,125,112,60,51,
187,23,187,15,58,62,115,21,251,233,137,226,
158,246,141,223,109,60,245,13,93,63,166,214,
236,102,64,253,3,166,64,71,129,74,49,121,
180,205,125,220,57,27,27,111,99,216,23,208,
196,36,197,137,251,210,5,94,115,143,245,247,
255,79,212,115,109,82,67,98,249,5,94,46,
87,168,152,42,173,70,76,97,177,193,159,237,
77,242,31,79,69,232,215,239,142,246,152,27,
200,68,229,140,89,227,214,196,235,104,75,19,
24,102,131,165,195,27,184,206,42,162,96,161,
189,127,131,43,149,34,149,184,128,187,209,166,
52,147,7,20,202,97,196,75,37,7,20,133,
156,170,242,115,147,101,234,126,196,15,118,138,
168,74,140,101,12,8,226,108,195,105,193,177,
104,247,61,86,149,50,135,158,142,50,136,17,
66,242,214,81,7,69,17,168,8,44,69,176,
59,136,142,165,49,148,254,56,105,225,241,215,
116,144,164,84,16,76,219,143,38,215,56,205,
220,113,253,52,101,191,31,32,200,72,122,198,
189,77,15,63,24,85,208,54,114,164,54,41,
43,69,56,26,169,202,142,182,43,148,173,67,
196,97,147,242,23,95,228,245,177,236,1,81,
111,108,96,88,38,20,5,141,253,144,0,3,
129,64,39,28,65,72,217,230,146,57,74,6,
129,110,144,202,24,37,25,96,144,144,34,70,
175,145,15,36,140,81,160,34,34,69,32,17,
76,87,85,28,35,165,217,215,240,98,141,163,
35,159,70,227,254,188,193,48,85,202,147,69,
26,61,137,57,26,182,141,205,149,73,85,147,
118,106,98,43,39,103,19,50,103,131,169,35,
203,5,68,16,98,68,45,246,117,201,222,237,
28,19,19,53,138,186,198,178,204,56,22,225,
66,5,99,243,128,100,97,138,143,243,14,195,
134,193,84,121,23,41,77,4,16,250,41,110,
237,46,99,150,66,73,211,185,213,185,129,229,
204,62,26,28,220,54,233,99,212,59,251,135,
27,20,166,206,209,237,5,236,181,71,9,205,
239,65,250,232,223,245,60,104,168,61,156,209,
28,111,193,144,236,212,235,184,86,142,215,175,
127,151,183,79,191,58,74,180,112,225,148,201,
112,32,72,226,22,120,67,100,52,194,222,35,
15,59,137,208,156,132,212,150,228,151,138,8,
145,64,208,128,7,47,211,19,62,195,99,128,
236,227,243,62,189,101,120,182,115,186,129,199,
158,26,50,86,124,186,76,69,26,133,164,105,
204,196,236,215,105,213,126,68,60,119,141,78,
35,32,140,230,248,210,23,35,180,163,29,60,
47,38,8,18,252,97,66,175,31,33,12,133,
48,72,201,41,29,2,85,100,124,26,64,211,
196,113,201,178,191,63,100,79,125,129,87,38,
179,223,68,120,67,180,31,125,138,44,56,164,
149,194,35,108,237,111,253,122,6,153,173,94,
176,80,158,50,215,255,247,255,119,143,211,23,
12,94,124,205,162,106,232,124,125,170,194,157,
126,198,175,186,135,40,52,122,211,204,85,119,
153,180,76,26,189,244,56,145,40,181,22,34,
142,137,47,157,38,174,22,65,74,148,82,30,
37,136,216,26,134,220,233,13,241,111,46,48,
57,247,45,132,80,8,7,45,188,254,1,249,
242,10,170,233,96,210,71,213,77,188,230,30,
170,126,135,196,75,216,219,250,61,166,22,127,
25,69,213,159,26,111,79,204,81,45,232,201,
152,195,174,79,56,232,64,28,112,48,108,147,
143,243,232,159,1,125,106,192,9,238,194,147,
120,12,113,34,89,175,123,116,253,8,116,1,
166,6,218,253,32,231,3,50,87,32,247,152,
53,171,135,207,185,88,253,213,19,255,221,220,
254,33,73,228,227,14,111,210,152,215,152,177,
178,11,252,198,146,134,30,10,218,205,8,195,
80,88,84,45,238,152,10,138,31,60,48,168,
145,116,250,9,137,200,225,108,239,162,212,246,
71,238,116,16,220,184,195,240,182,68,145,18,
217,105,49,150,118,81,72,209,220,60,201,202,
28,68,17,40,54,82,194,156,99,102,195,185,
188,70,50,223,225,189,67,135,142,55,198,68,
233,128,77,119,200,182,231,147,8,201,134,236,
96,228,21,182,164,65,216,30,99,105,198,69,
32,232,199,9,141,32,100,167,61,206,68,105,
31,18,129,52,253,39,14,21,82,41,217,25,
248,164,72,206,21,29,156,7,12,206,111,244,
60,182,30,192,66,37,80,49,116,20,50,181,
86,165,209,69,217,175,147,156,93,132,241,50,
162,214,58,54,11,58,249,38,41,98,24,160,
125,180,134,180,77,136,70,114,216,64,41,167,
176,176,224,176,179,147,213,103,33,58,9,26,
42,9,201,3,204,189,130,119,82,64,79,65,
30,39,145,199,34,67,72,84,53,83,128,221,
184,235,210,239,103,85,219,129,111,32,49,144,
64,33,175,179,184,232,112,184,221,102,117,181,
128,166,73,130,88,193,44,8,182,194,25,46,
218,27,247,171,213,48,65,72,129,54,95,36,
89,157,71,170,74,198,53,137,98,254,81,250,
35,62,180,79,115,117,253,5,86,102,111,178,
34,2,100,193,97,173,92,164,116,185,192,145,
166,114,183,151,160,138,132,113,75,231,82,121,
228,89,32,42,8,93,161,56,145,114,148,132,
168,142,124,236,125,175,168,218,99,109,64,157,
234,12,174,31,179,219,235,67,110,132,207,43,
26,216,143,135,3,118,67,133,139,227,5,140,
81,245,63,172,121,164,214,56,111,191,244,45,
94,189,243,35,62,93,186,76,160,103,157,128,
157,59,1,226,223,195,185,48,114,240,212,209,
243,103,124,134,135,134,3,199,231,229,243,58,
167,218,227,40,205,49,89,172,98,62,101,155,
42,9,125,210,36,70,183,243,56,213,95,161,
223,120,129,166,249,109,6,246,128,31,28,93,
225,202,106,9,237,230,29,130,212,165,52,105,
224,13,83,66,77,101,106,161,204,254,254,16,
167,170,211,108,4,32,5,7,45,80,108,200,
229,84,122,73,1,209,189,205,187,156,97,169,
216,162,160,184,247,110,34,32,68,168,46,177,
94,64,58,86,38,100,89,204,145,150,10,25,
198,10,217,198,85,225,241,159,253,141,175,217,
4,131,251,115,74,67,81,184,88,202,113,177,
148,99,144,36,236,15,2,24,234,124,121,162,
68,73,215,184,88,144,116,7,51,64,7,49,
218,26,84,119,14,217,200,217,180,195,136,87,
203,5,82,67,231,206,97,68,188,247,18,99,
147,171,4,238,119,73,82,155,226,196,207,19,
4,7,28,109,126,143,165,87,126,11,205,176,
81,13,139,200,239,98,104,9,111,92,177,216,
246,6,124,252,233,255,133,72,76,170,249,111,
82,156,190,248,212,24,126,28,51,19,201,81,
218,3,169,128,89,36,85,52,234,210,224,244,
103,204,179,158,25,122,218,105,13,232,14,62,
91,119,168,222,15,31,155,40,30,91,168,244,
235,120,237,13,114,165,101,6,157,93,242,151,
174,243,114,209,228,86,119,200,94,123,142,156,
217,231,251,71,29,190,49,93,97,220,209,81,
142,90,68,249,18,197,158,68,248,25,214,191,
227,207,226,167,22,190,38,208,123,71,204,23,
142,232,219,37,162,185,105,244,221,67,236,176,
143,173,167,160,235,164,115,227,12,123,49,122,
163,137,168,183,81,187,46,201,165,85,164,132,
161,151,226,228,21,214,250,3,22,28,147,78,
24,83,202,53,153,45,182,105,247,150,232,203,
62,229,124,3,91,85,88,206,219,28,14,3,
246,134,1,200,2,155,158,207,185,162,67,179,
47,169,247,103,25,248,5,162,156,137,30,41,
72,211,125,234,117,216,27,4,188,80,202,209,
141,98,10,154,150,89,57,60,36,165,240,66,
41,71,78,83,81,106,77,212,235,27,32,68,
38,107,157,179,179,42,41,239,128,222,125,50,
182,218,238,33,30,178,176,216,237,219,244,220,
232,4,2,107,16,142,18,133,70,140,134,198,
143,167,94,155,36,18,207,139,9,195,7,216,
235,104,248,228,40,83,199,20,146,205,181,1,
73,34,81,71,85,76,152,40,68,106,158,220,
3,27,44,141,122,64,173,230,115,234,116,30,
213,212,89,27,248,76,154,6,149,23,79,35,
13,29,245,230,38,175,110,93,101,254,130,203,
159,221,58,199,165,215,118,72,78,205,115,187,
222,25,221,99,58,109,119,138,153,234,22,139,
78,1,235,168,73,50,86,70,149,69,84,205,
166,238,69,236,220,92,38,246,35,210,197,232,
132,109,238,147,38,184,113,224,145,104,14,183,
107,46,242,25,37,16,194,56,101,237,48,91,
135,245,130,228,132,224,222,135,167,94,101,174,
181,199,81,105,242,56,89,252,172,30,25,143,
100,192,153,169,252,115,189,174,48,126,6,167,
180,72,237,238,119,8,195,62,7,45,149,83,
110,143,176,63,100,249,76,25,219,18,92,235,
24,12,251,23,241,91,62,0,0,32,0,73,
68,65,84,30,139,211,58,65,148,224,53,163,
140,203,16,193,193,166,143,162,8,242,121,141,
249,178,134,174,220,160,158,94,98,246,1,27,
128,227,227,238,29,146,139,43,40,12,17,27,
135,164,139,211,36,167,23,16,237,62,50,5,
183,235,32,132,64,213,193,178,149,227,123,116,
127,39,166,209,93,97,109,189,205,244,100,135,
151,94,187,223,57,57,170,202,233,130,131,17,
27,216,194,2,98,108,61,164,158,140,184,69,
99,37,104,101,26,105,150,55,100,63,74,40,
20,115,88,154,197,224,234,34,106,161,203,121,
249,7,76,22,124,246,39,198,248,100,205,97,
122,245,91,76,141,150,241,194,168,67,28,14,
217,219,252,125,126,229,202,16,253,189,59,156,
78,37,99,47,156,226,131,190,139,191,217,224,
121,64,191,166,27,226,63,196,173,104,185,33,
13,71,103,60,111,252,100,137,98,24,165,207,
244,1,74,142,78,197,121,58,149,232,96,235,
223,161,158,221,163,215,85,16,75,130,128,148,
195,97,132,165,10,38,203,123,184,126,145,70,
231,20,186,179,139,218,232,160,108,31,210,78,
170,156,141,84,20,43,11,38,11,86,70,50,
72,202,69,110,169,103,185,219,77,49,231,170,
12,99,176,43,85,150,151,14,232,169,42,7,
178,64,51,201,211,50,138,44,205,214,88,109,
220,64,120,62,193,219,119,241,79,45,83,92,
46,210,109,37,156,46,217,220,114,7,132,169,
228,139,227,69,164,20,28,90,7,124,176,55,
129,148,2,63,73,153,182,77,86,243,54,41,
176,70,202,122,10,215,219,17,181,206,28,179,
99,155,80,130,70,119,6,93,13,169,14,170,
36,249,189,167,94,139,191,62,108,83,26,153,
26,13,227,244,17,33,192,68,102,250,81,114,
68,252,73,86,102,79,106,217,56,22,201,226,
52,90,251,51,248,14,170,130,204,217,136,17,
231,96,173,87,166,21,79,48,166,237,50,95,
14,240,58,30,73,60,218,60,35,38,192,248,
137,216,152,15,38,9,128,6,115,204,115,59,
131,110,70,93,70,97,113,153,13,191,2,2,
250,161,193,140,213,96,92,59,34,78,5,154,
34,105,180,2,202,85,3,219,82,137,45,147,
154,31,114,182,224,32,201,36,219,139,166,142,
22,132,76,173,109,113,113,110,130,191,188,85,
65,232,3,2,247,69,72,19,100,238,144,153,
234,22,99,166,206,28,153,1,206,213,48,135,
233,204,162,219,37,106,59,48,190,248,243,228,
170,243,60,83,212,23,2,140,28,183,246,250,
68,137,124,174,235,49,8,19,110,236,63,250,
186,84,81,217,171,206,113,229,238,251,188,127,
250,117,126,214,143,182,23,209,29,68,148,156,
231,163,19,170,186,73,69,191,206,248,106,130,
219,7,189,156,35,215,139,184,115,225,159,98,
182,234,232,195,54,205,221,128,238,104,115,174,
92,214,105,181,238,119,182,194,206,83,19,211,
108,117,33,140,2,52,187,195,118,225,85,206,
22,235,180,147,50,101,173,75,89,235,82,213,
58,168,159,102,106,195,178,92,32,173,222,219,
84,75,193,182,143,7,187,73,156,113,153,234,
135,49,215,62,137,48,202,47,48,251,74,230,
54,119,240,209,255,65,255,59,219,124,229,27,
39,97,182,197,74,131,40,209,30,64,78,70,
118,203,83,99,217,234,123,171,199,244,120,137,
111,73,56,242,67,190,247,93,129,117,233,54,
162,26,112,152,24,252,253,245,42,230,193,105,
166,79,191,249,216,34,197,174,36,28,38,146,
120,110,129,138,150,18,42,10,184,10,157,221,
79,153,92,125,243,153,175,245,147,32,166,173,
198,0,93,17,143,253,237,158,57,22,60,11,
49,40,73,37,183,15,93,10,150,134,161,41,
232,170,32,233,7,232,98,136,130,32,103,169,
148,236,236,67,76,204,252,2,237,173,31,66,
105,19,225,132,52,221,18,161,221,165,209,155,
70,195,196,86,117,126,121,161,206,222,247,125,
150,243,25,172,210,25,24,12,18,135,182,239,
48,199,93,142,218,89,117,230,246,14,208,229,
38,8,152,105,87,16,111,190,202,39,93,201,
223,29,188,128,237,212,168,150,82,182,186,10,
83,230,14,83,134,135,41,242,200,189,33,182,
18,34,135,53,26,253,28,165,49,141,205,125,
152,30,51,216,114,125,222,109,246,73,164,164,
55,168,32,69,54,143,144,192,237,158,135,227,
107,212,218,57,14,85,149,126,52,143,161,251,
89,146,24,29,227,165,3,220,97,137,14,3,
74,221,89,210,210,254,103,94,183,121,199,228,
82,41,199,205,222,128,187,238,163,252,140,91,
125,143,68,74,206,77,87,137,75,14,152,15,
205,53,226,56,211,104,122,240,190,42,229,17,
93,55,219,156,136,162,204,160,126,126,10,101,
167,134,186,190,195,215,230,14,248,203,181,113,
58,106,149,179,249,38,75,51,42,251,251,67,
58,157,48,227,81,140,186,0,149,207,223,19,
35,194,164,166,157,101,230,112,147,138,181,199,
242,114,14,189,168,208,15,178,225,118,156,42,
12,35,193,249,115,247,235,164,94,146,48,102,
153,180,194,136,134,31,209,10,35,94,73,37,
170,109,226,159,93,196,139,14,81,115,54,50,
150,148,166,107,136,193,24,210,105,101,76,120,
1,236,214,168,21,74,236,108,191,64,121,108,
26,69,40,40,137,64,220,91,14,120,134,45,
161,84,102,107,133,126,244,227,57,31,13,238,
137,177,61,124,94,69,229,253,211,175,243,198,
173,31,240,193,233,47,16,169,250,207,116,178,
216,106,14,120,209,46,61,191,229,180,153,99,
97,117,136,33,37,124,100,49,65,137,56,104,
177,218,126,7,76,216,202,105,52,26,1,105,
122,50,175,111,242,2,229,160,205,156,118,64,
126,220,224,134,63,131,73,204,120,255,135,148,
75,69,132,150,177,183,107,209,4,215,189,115,
156,177,215,209,149,24,162,248,120,11,74,248,
17,178,92,34,145,10,13,183,192,84,161,203,
218,141,128,163,222,5,38,46,127,25,175,113,
131,253,15,254,31,102,95,249,23,204,188,252,
219,52,110,127,151,157,205,119,89,88,54,31,
10,194,143,127,94,164,34,240,202,121,246,122,
3,102,108,131,27,159,198,216,203,2,43,87,
36,137,34,54,245,14,218,133,62,241,122,239,
177,6,73,154,105,99,164,95,229,230,182,64,
213,243,124,224,255,37,66,79,73,15,231,16,
179,251,12,218,251,56,149,217,103,186,204,79,
66,125,146,84,114,235,208,165,154,51,40,57,
39,181,202,158,61,81,168,207,254,171,63,56,
85,23,94,136,76,239,227,237,115,21,155,185,
138,69,126,108,149,252,216,42,189,163,117,110,
239,124,155,220,84,143,102,127,138,201,82,141,
75,101,135,177,142,196,243,138,88,226,128,112,
118,150,244,246,46,227,28,208,168,187,244,6,
41,215,200,97,199,30,17,38,138,72,49,45,
21,219,84,168,156,157,96,115,168,226,26,13,
198,103,59,52,251,83,236,116,116,102,202,7,
172,202,28,147,119,55,33,73,239,155,192,14,
2,114,5,149,90,47,97,59,200,225,54,26,
196,45,141,36,180,168,235,19,148,115,77,198,
75,7,200,190,129,108,85,56,72,103,8,91,
62,65,169,129,177,210,102,50,255,120,56,46,
111,119,169,119,103,41,104,17,210,87,72,183,
102,112,156,21,64,195,247,118,17,231,239,7,
246,177,145,239,196,189,29,109,25,101,209,68,
104,199,197,4,123,195,128,178,161,49,245,184,
173,22,153,205,35,100,49,83,126,237,26,101,
58,137,73,177,60,164,52,147,105,219,7,249,
34,6,144,46,76,33,199,203,248,31,109,243,
82,181,142,235,70,56,134,142,97,40,204,207,
59,8,1,157,78,8,113,72,140,74,132,142,
206,79,46,117,94,165,198,46,103,17,164,24,
248,204,199,87,179,4,16,130,231,197,148,203,
6,5,51,187,111,116,5,162,84,203,182,68,
150,103,65,85,185,107,219,52,253,144,102,16,
209,141,50,207,12,125,110,130,112,117,158,235,
93,15,55,142,48,12,23,140,12,63,78,177,
201,25,26,95,172,150,176,62,93,163,181,214,
224,195,153,47,51,115,234,91,12,187,117,66,
191,199,153,51,42,55,238,254,16,167,244,107,
159,125,63,215,54,201,79,46,179,94,243,112,
253,159,158,153,212,187,103,222,96,174,185,75,
173,50,67,248,57,121,103,252,67,28,126,148,
114,208,245,153,45,63,31,148,214,13,231,113,
227,155,148,183,142,184,254,221,45,86,95,158,
100,88,24,227,104,241,69,122,29,29,187,122,
7,179,191,141,231,197,143,60,0,229,120,31,
175,11,94,23,102,173,46,65,8,173,202,69,
180,80,82,28,220,102,88,210,153,178,235,76,
233,117,238,12,87,56,109,111,100,146,57,163,
89,158,122,103,23,89,200,209,144,211,164,82,
161,182,27,81,219,75,72,149,219,116,110,94,
231,244,57,149,59,205,132,195,247,255,7,226,
52,143,110,151,249,228,71,33,83,211,38,198,
99,190,102,152,166,4,169,100,24,39,140,153,
58,7,195,144,143,70,220,166,183,247,243,232,
78,202,55,166,52,198,198,82,222,218,80,113,
201,158,119,169,251,39,100,101,164,76,105,110,
253,136,201,51,111,50,190,244,197,251,215,248,
238,54,150,54,143,152,87,105,237,188,143,181,
50,241,19,119,20,199,48,148,23,98,106,10,
73,42,143,41,15,207,142,46,200,207,231,38,
218,111,15,169,230,244,99,54,119,113,114,149,
169,206,24,169,185,143,99,186,92,42,229,88,
89,223,65,30,117,104,216,23,89,157,13,145,
81,66,109,0,145,27,208,232,9,242,69,139,
83,23,170,12,188,34,214,172,69,51,87,165,
188,190,143,45,3,250,158,138,154,143,89,208,
202,236,139,30,73,162,33,124,21,117,48,205,
170,184,157,37,137,7,209,132,56,161,153,132,
172,181,10,248,137,36,62,178,24,215,126,139,
72,9,48,218,239,17,246,2,146,162,77,57,
253,22,149,213,43,244,235,235,212,171,191,135,
83,122,232,60,143,185,68,99,230,62,221,205,
83,204,20,94,39,55,119,10,187,60,153,117,
70,7,215,104,15,238,160,142,208,163,15,91,
253,236,245,18,194,91,37,244,96,145,52,95,
71,93,61,188,95,141,198,9,215,58,30,93,
231,94,48,21,140,155,58,5,93,67,22,28,
162,175,188,12,186,198,223,253,141,79,111,176,
192,160,190,5,169,66,20,14,17,72,190,248,
141,152,83,23,140,204,75,162,148,71,154,58,
179,227,2,198,239,7,36,77,19,44,44,56,
20,10,58,205,102,38,250,23,254,24,0,212,
99,180,8,113,232,225,112,18,26,187,183,33,
181,181,229,145,207,235,36,66,35,76,4,138,
0,71,79,32,205,108,85,165,170,80,59,104,
29,147,50,1,202,186,198,237,40,229,176,223,
165,31,197,216,170,194,169,188,141,169,42,236,
12,2,218,66,50,140,83,104,180,217,255,232,
144,116,121,14,35,255,245,99,95,10,69,209,
248,224,29,27,229,51,102,59,73,20,128,76,
201,79,45,115,247,200,163,61,136,126,170,1,
54,85,84,14,170,115,92,220,190,202,39,43,
87,126,166,187,138,253,182,207,120,222,56,30,
224,63,203,33,200,160,28,165,146,231,212,47,
156,227,134,125,150,65,20,147,134,83,172,214,
190,77,211,88,32,54,43,16,156,92,220,184,
7,105,222,59,54,252,37,138,99,121,46,149,
54,217,188,211,163,47,161,86,19,172,174,22,
50,19,164,145,167,182,72,82,68,163,3,165,
60,241,107,231,145,150,69,171,150,231,194,244,
30,20,117,126,105,254,94,103,151,253,251,212,
57,216,221,244,217,184,211,167,113,176,71,48,
180,248,238,247,102,16,178,129,161,244,248,242,
155,54,134,149,37,137,155,221,1,71,65,136,
169,40,196,82,226,197,9,73,162,209,232,205,
80,202,53,153,194,102,162,154,221,235,243,37,
159,27,174,200,88,222,115,251,236,220,250,125,
114,185,51,20,42,231,25,118,119,241,253,71,
225,235,233,83,247,61,79,138,83,23,158,139,
55,211,232,63,121,25,197,54,84,242,166,198,
252,152,125,226,25,126,230,40,16,165,159,79,
166,144,192,97,55,96,101,226,62,198,62,62,
253,53,14,154,107,104,99,1,139,66,160,52,
58,108,237,122,180,213,14,225,235,58,2,104,
236,182,25,166,38,139,43,54,149,215,87,120,
203,52,240,147,148,88,74,146,134,193,138,51,
207,89,247,46,233,206,33,118,179,69,245,242,
34,65,56,65,167,87,160,109,167,204,104,77,
12,47,184,111,188,34,68,70,102,49,117,54,
189,128,94,191,128,80,93,148,250,50,133,151,
78,35,101,74,167,241,61,140,11,1,233,64,
195,59,88,199,202,77,225,182,110,162,173,158,
76,18,101,93,99,1,201,122,10,131,123,194,
95,18,244,205,55,184,116,225,151,16,170,134,
223,173,19,251,125,186,71,215,112,253,79,81,
31,42,0,100,100,144,238,142,179,116,234,95,
18,249,30,173,238,95,147,114,120,18,186,72,
18,214,250,3,94,170,228,169,26,58,230,131,
156,138,145,1,138,55,44,161,138,6,191,254,
159,229,178,149,191,84,162,189,123,13,2,29,
121,51,243,111,232,157,62,131,152,159,128,110,
253,145,236,118,79,63,167,119,143,113,72,250,
220,191,241,179,220,178,9,42,41,42,41,32,
99,201,126,83,82,176,67,246,182,251,232,134,
202,210,146,131,50,94,132,52,165,17,39,39,
146,4,64,39,140,17,64,217,208,88,206,89,
244,227,132,229,188,205,91,71,29,250,110,17,
161,31,242,245,193,0,247,221,13,244,162,197,
71,218,203,140,207,188,116,255,114,57,37,86,
223,248,207,159,8,57,165,73,194,160,185,143,
81,158,101,189,222,167,59,136,126,106,36,182,
40,74,208,180,76,31,44,86,53,62,89,185,
194,155,159,126,151,31,174,94,249,153,133,161,
82,41,217,105,13,89,157,124,118,102,120,148,
164,252,104,175,74,20,141,81,78,202,204,78,
245,57,191,118,149,250,71,251,172,245,34,138,
83,58,161,82,64,229,126,162,216,225,44,11,
35,111,106,128,61,78,51,205,38,106,51,102,
163,249,224,120,41,27,118,3,40,163,123,90,
22,28,148,122,27,177,121,64,114,229,44,8,
168,230,92,106,189,50,83,197,71,7,225,113,
8,31,127,92,96,238,245,127,69,207,253,93,
38,166,187,196,185,49,170,167,126,147,52,14,
249,171,191,248,93,190,240,122,200,228,172,198,
229,74,38,34,122,111,222,216,118,39,144,169,
202,149,217,35,44,105,98,147,34,20,149,187,
238,144,53,223,67,9,10,68,173,8,90,69,
18,189,78,215,190,77,207,253,43,210,218,4,
179,167,126,229,41,163,179,103,191,47,219,94,
196,118,107,240,153,227,131,7,99,243,115,39,
138,244,199,76,20,134,38,48,109,13,129,32,
78,37,81,146,50,124,8,227,181,138,147,208,
200,33,199,2,182,83,201,170,101,50,51,157,
18,28,236,16,137,5,186,215,246,112,125,149,
203,95,42,99,124,233,34,98,235,144,87,143,
234,184,229,28,87,85,19,217,253,2,71,106,
27,10,49,167,163,59,92,255,100,64,184,57,
160,80,57,228,236,66,204,187,233,12,190,167,
16,69,2,247,244,2,137,170,224,150,242,148,
220,33,123,150,65,207,143,216,12,11,140,13,
4,179,213,203,163,139,175,80,40,93,194,115,
143,144,131,28,177,210,230,104,253,239,137,68,
141,113,69,99,94,164,148,135,1,142,31,96,
235,42,74,146,178,120,118,137,191,60,108,18,
167,146,184,165,49,181,244,198,9,103,181,198,
222,15,8,23,222,122,152,127,136,210,157,38,
117,234,104,106,1,221,41,18,249,30,242,9,
45,92,73,215,152,54,205,39,182,143,37,165,
134,90,180,80,148,172,58,19,67,63,211,144,
241,134,12,35,149,84,106,56,155,91,136,243,
11,79,236,18,171,85,3,41,115,212,106,62,
34,130,32,214,49,136,63,191,182,18,136,48,
176,70,66,145,49,42,21,39,65,215,21,198,
199,76,38,38,44,84,77,32,117,13,116,13,
207,205,136,135,113,42,73,165,204,60,1,4,
188,58,86,196,232,121,136,90,131,116,113,154,
245,163,4,175,51,13,154,203,84,65,195,218,
27,98,156,155,226,187,222,89,138,115,39,225,
165,222,209,45,122,173,15,152,191,248,59,39,
131,65,224,177,243,225,31,51,117,238,235,232,
99,75,124,186,219,101,253,78,157,197,83,19,
159,123,48,245,220,33,27,235,71,116,219,30,
185,188,197,203,175,157,58,254,127,63,184,240,
21,102,107,119,216,175,204,18,243,179,121,52,
221,144,217,178,245,68,45,184,71,33,145,16,
51,183,195,197,174,75,239,157,91,204,246,38,
184,253,222,193,241,60,162,163,47,96,28,125,
200,131,209,35,198,224,128,83,88,120,84,168,
49,199,157,199,199,161,7,212,36,142,215,191,
71,236,103,209,115,81,63,190,77,242,210,25,
166,10,93,234,110,145,221,246,24,243,149,147,
157,203,91,127,221,71,207,93,224,224,195,223,
227,242,203,30,11,203,57,254,234,47,174,2,
223,100,255,195,63,4,117,150,195,253,53,38,
103,179,103,222,209,84,8,34,218,205,21,206,
228,5,11,165,46,5,197,65,85,5,194,150,
252,160,222,161,29,198,200,8,226,155,85,102,
206,126,139,150,249,125,130,120,11,17,152,40,
222,20,99,11,95,69,253,28,182,225,146,84,
178,213,28,124,102,55,81,205,27,20,45,109,
228,149,35,126,188,68,177,52,238,176,219,26,
210,121,74,251,173,169,130,217,178,141,109,40,
216,186,74,220,243,113,170,159,173,5,147,38,
49,40,217,227,208,142,19,146,217,9,148,224,
128,130,51,196,92,157,33,191,85,199,82,123,
172,127,220,192,74,246,232,90,99,140,239,29,
82,221,174,243,181,197,73,126,104,234,84,150,
127,147,163,219,146,179,197,26,203,211,29,90,
245,27,68,189,148,170,107,241,207,171,5,188,
47,92,224,47,218,175,80,223,43,81,157,253,
4,117,36,33,222,173,231,25,4,121,166,181,
93,38,203,191,140,161,221,215,177,159,88,122,
19,237,160,132,93,152,198,154,155,26,5,146,
1,253,237,239,225,248,223,166,246,225,109,20,
5,166,38,109,116,91,197,54,76,156,156,69,
47,141,193,183,72,172,251,179,153,112,208,195,
237,95,231,113,114,50,169,238,130,158,160,8,
235,56,88,37,105,255,164,161,72,10,40,80,
53,117,140,78,15,17,197,164,211,99,39,182,
34,16,130,175,46,54,193,208,72,162,121,208,
181,227,13,143,172,202,203,86,96,175,127,208,
67,191,218,32,73,82,150,150,114,20,10,250,
67,179,113,73,62,175,97,89,57,194,48,197,
117,99,14,155,130,56,17,24,4,159,107,146,
200,110,194,132,78,39,98,97,193,161,82,53,
81,84,1,154,74,178,58,151,117,108,134,206,
21,67,167,98,104,108,184,67,146,17,239,69,
239,123,168,159,220,206,204,106,90,61,138,203,
231,137,115,187,56,170,194,74,144,34,58,46,
223,229,21,74,43,255,26,85,63,57,116,140,
194,46,175,190,114,200,39,87,127,15,167,244,
69,114,229,69,34,223,37,244,91,204,94,250,
199,68,90,153,181,181,13,174,174,185,140,77,
22,63,247,32,218,239,13,185,254,241,54,19,
83,69,100,34,233,117,7,244,186,30,197,82,
86,129,199,170,206,208,176,25,239,55,56,180,
75,63,179,16,212,110,123,248,204,235,178,186,
150,49,153,213,219,53,130,32,101,231,211,163,
227,36,33,4,40,170,130,109,171,184,238,253,
212,185,194,167,199,157,68,133,218,147,147,178,
151,233,141,89,150,74,85,109,115,215,95,98,
165,119,72,242,194,41,148,221,35,68,167,143,
216,171,35,23,167,153,200,247,232,251,54,119,
27,83,156,26,191,127,206,211,23,76,238,28,
150,73,130,247,88,88,206,145,166,48,236,38,
248,253,6,186,118,196,165,23,98,98,93,225,
78,211,199,235,164,108,111,65,63,92,225,244,
139,29,190,56,167,210,30,228,41,58,30,97,
154,242,86,189,199,32,76,72,119,38,177,149,
139,76,190,254,117,246,110,254,17,241,236,13,
10,205,175,83,29,123,21,109,214,1,33,8,
221,206,79,14,5,118,252,207,76,18,57,83,
227,244,228,231,192,163,112,12,149,179,211,121,
250,126,204,205,77,239,137,181,229,68,193,60,
33,229,17,31,199,177,148,97,251,128,40,232,
32,132,134,80,85,12,187,138,145,43,211,111,
221,162,48,238,101,236,231,52,225,122,165,200,
197,52,33,189,84,224,219,59,29,190,244,230,
69,150,46,250,92,221,202,33,138,191,136,51,
126,145,155,193,31,240,245,211,107,104,27,123,
204,55,255,144,70,97,138,137,211,191,193,59,
55,27,156,155,185,197,249,137,58,159,94,237,
178,223,4,173,217,102,170,249,54,255,104,42,
71,50,63,205,15,123,47,19,168,26,90,148,
240,21,177,133,62,17,241,254,221,42,229,51,
23,142,89,234,247,142,202,3,112,69,182,125,
224,80,89,254,22,119,15,12,150,46,255,41,
205,171,219,108,238,12,80,144,148,26,55,153,
249,181,215,232,69,49,218,156,75,125,227,47,
89,42,253,171,17,212,81,196,201,173,16,203,
214,49,54,51,102,234,84,12,157,141,65,38,
123,30,105,123,236,222,253,125,194,228,8,109,
190,135,225,207,51,153,111,49,174,216,248,129,
128,226,144,163,32,100,211,54,89,137,34,212,
91,91,89,48,85,20,244,239,127,140,52,141,
140,100,87,247,81,14,154,200,82,14,225,13,
137,227,140,215,208,234,199,24,34,34,73,36,
209,200,124,101,125,221,101,113,49,71,175,23,
225,186,17,97,152,62,118,83,84,1,82,204,
159,248,134,13,176,208,31,147,108,142,142,124,
218,237,144,40,74,89,93,205,83,174,90,72,
199,198,141,19,62,233,184,12,147,132,139,165,
220,9,131,122,237,218,93,24,125,15,209,243,
24,120,49,5,91,227,43,82,98,109,236,243,
174,114,138,252,236,239,60,146,36,178,27,178,
207,140,13,83,151,247,184,213,253,99,54,215,
199,200,149,222,164,56,119,142,195,174,207,157,
173,22,235,159,238,226,6,22,23,94,92,252,
220,3,104,62,111,241,218,151,207,160,170,153,
218,237,181,143,183,30,1,237,6,166,131,146,
166,252,44,31,109,47,115,85,179,244,167,119,
21,150,210,195,81,85,250,251,125,162,40,37,
138,78,214,65,166,95,195,207,47,128,123,159,
144,185,203,89,98,180,99,85,129,207,58,214,
215,93,102,102,108,42,149,46,154,72,216,104,
79,179,52,12,73,86,231,81,247,142,80,218,
61,210,197,105,0,10,214,144,163,254,201,2,
97,115,35,166,120,234,2,113,252,215,35,152,
22,44,99,64,243,218,255,134,109,135,204,44,
56,252,224,131,128,173,15,29,172,210,52,165,
153,115,224,254,53,95,28,249,93,244,124,135,
138,227,113,163,59,192,245,82,148,187,47,176,
120,225,95,160,140,22,23,42,211,175,227,119,
231,24,59,245,197,167,194,122,94,144,48,8,
18,10,150,134,99,62,254,218,246,134,49,187,
237,33,227,121,227,169,28,56,93,21,143,237,
36,158,59,81,220,63,161,130,162,8,158,103,
65,176,182,249,87,148,210,247,56,155,238,19,
215,123,36,82,96,79,230,232,247,109,220,208,
96,34,72,40,199,45,204,225,128,217,137,28,
102,201,97,99,166,202,39,215,230,41,21,191,
192,219,7,155,196,97,192,212,242,47,162,89,
69,134,157,3,10,147,223,228,251,219,45,222,
252,114,133,234,159,188,75,247,230,255,132,124,
227,191,165,114,225,191,224,214,238,59,236,180,
255,156,217,217,53,26,71,30,97,172,208,106,
133,248,65,74,239,166,199,151,206,238,19,6,
9,110,39,100,113,57,207,198,248,121,242,214,
243,13,15,243,51,223,228,246,80,161,112,241,
45,198,118,62,161,217,138,56,106,73,138,191,
255,46,95,88,44,34,207,78,114,173,122,72,
224,182,142,189,188,167,87,254,9,91,183,15,
80,79,239,131,128,49,67,199,214,20,116,85,
37,1,196,76,131,148,198,241,143,146,26,123,
140,219,5,230,10,130,129,155,146,43,216,44,
56,38,63,106,246,152,159,26,195,104,247,209,
223,250,8,20,149,56,76,105,30,13,112,221,
24,207,141,144,8,52,109,72,154,74,6,94,
132,42,35,34,140,71,100,55,146,68,178,177,
113,146,24,56,61,109,97,24,42,170,42,72,
83,201,225,225,144,32,72,81,73,136,48,80,
72,73,30,232,119,20,36,218,103,108,70,101,
14,34,25,50,108,224,63,113,134,113,207,34,
85,74,50,53,77,85,33,15,188,57,89,230,
195,118,159,187,253,225,177,51,32,65,196,192,
13,16,97,130,86,206,211,47,142,177,211,214,
240,142,10,236,149,219,236,200,115,168,227,255,
33,142,243,40,171,184,91,255,152,156,115,51,
75,172,186,65,119,55,1,138,72,179,196,245,
189,30,59,187,29,214,110,238,97,38,17,99,
83,147,104,186,242,185,7,80,49,146,125,1,
80,53,133,203,175,62,234,157,157,243,61,180,
36,166,95,154,254,153,78,22,13,55,100,190,
242,116,105,15,1,156,110,180,217,127,66,78,
25,28,236,16,21,22,232,229,46,49,244,2,
20,18,170,28,96,227,18,61,80,196,212,153,
39,192,33,65,195,194,99,138,173,12,105,242,
19,54,54,92,6,3,139,249,121,72,43,69,
238,120,167,73,10,57,206,124,97,247,216,140,
234,120,25,37,231,114,212,47,49,89,200,22,
30,226,72,160,153,57,30,48,196,228,23,127,
181,200,247,127,160,51,61,22,243,206,247,6,
52,142,84,114,213,152,165,243,155,248,157,77,
190,250,79,237,71,238,247,97,146,192,193,52,
147,139,223,60,78,18,0,185,202,2,185,202,
194,19,175,143,31,37,212,122,1,141,126,72,
242,192,40,160,104,107,204,148,173,99,234,193,
253,46,65,197,15,19,54,27,131,167,94,251,
206,32,226,211,221,30,115,149,204,231,163,209,
15,56,61,153,63,30,225,61,87,162,24,132,
9,183,14,220,236,67,62,225,217,169,117,3,
28,67,101,236,1,134,95,20,213,121,249,172,
143,120,175,65,179,227,83,174,24,12,182,251,
48,72,168,106,10,19,43,21,134,157,33,141,
110,72,255,40,133,164,69,186,98,48,127,249,
63,26,121,253,190,56,10,32,146,192,109,178,
125,227,127,103,225,236,111,227,76,253,14,31,
126,242,111,120,233,171,231,41,252,219,247,56,
188,245,63,83,188,240,175,169,204,191,65,58,
125,133,53,245,127,225,202,228,251,220,189,217,
199,212,21,116,219,32,242,2,54,214,122,156,
57,87,96,242,116,145,232,141,23,16,3,241,
216,224,149,198,33,73,232,163,59,143,135,30,
172,242,121,226,177,51,20,213,255,134,197,217,
38,141,122,192,81,87,178,123,171,7,55,123,
92,57,159,231,253,241,63,101,42,255,159,102,
1,85,51,152,153,255,13,14,26,255,35,233,
64,167,181,36,40,229,99,194,68,133,72,67,
139,38,145,113,74,236,212,16,90,134,197,127,
212,239,113,24,27,156,179,242,132,129,196,52,
21,190,58,89,30,13,227,114,12,214,107,108,
213,98,134,253,33,50,58,153,4,238,42,203,
184,162,192,164,168,81,146,109,98,180,71,18,
197,35,223,201,82,153,154,178,209,52,65,24,
166,52,26,1,81,36,71,187,31,113,182,137,
139,50,154,89,140,176,205,113,155,24,139,122,
35,192,32,66,2,49,250,113,141,167,19,161,
60,7,210,94,42,233,32,64,123,239,58,225,
43,23,120,175,221,231,200,15,249,210,120,233,
120,81,32,253,112,147,198,158,71,211,85,56,
243,138,193,213,254,52,148,255,25,83,149,9,
118,251,13,242,203,139,39,230,67,199,85,86,
237,6,157,221,15,208,150,167,248,206,123,93,
132,158,98,155,191,133,54,63,201,90,211,99,
235,238,1,187,91,13,144,96,231,12,86,207,
76,253,131,5,88,215,202,163,72,201,207,250,
209,236,63,91,162,8,163,60,250,198,30,165,
146,78,163,17,176,207,42,179,172,159,44,84,
251,59,140,177,131,68,144,32,168,51,207,144,
60,99,220,71,2,38,184,111,20,54,160,200,
22,23,152,102,19,115,4,119,118,187,17,115,
115,80,234,239,83,234,239,211,154,57,75,195,
94,96,60,127,114,43,175,236,120,220,174,207,
28,39,10,93,75,57,188,241,247,172,76,157,
76,40,187,215,59,244,231,78,17,249,93,114,
249,22,111,188,38,41,18,34,23,12,228,3,
179,196,56,29,109,4,10,129,72,159,125,238,
16,196,41,135,93,143,118,160,60,177,115,8,
227,1,151,23,78,66,148,170,34,152,44,154,
236,119,252,103,122,31,63,74,89,63,242,142,
147,204,131,205,197,51,39,10,215,143,185,117,
232,158,200,100,79,106,139,214,143,60,106,221,
128,115,51,25,54,105,170,11,12,235,87,217,
221,140,88,94,204,83,116,36,197,130,78,191,
23,179,118,167,71,16,183,152,254,245,87,153,
245,67,172,254,128,52,133,191,223,157,160,248,
144,33,188,16,2,221,42,34,164,133,80,84,
12,167,194,48,254,45,126,120,251,15,121,229,
194,4,167,213,247,184,245,255,177,247,94,49,
150,165,217,149,222,119,188,187,222,135,247,233,
179,124,177,170,93,117,23,155,84,179,73,14,
135,51,2,134,28,202,140,68,81,2,8,185,
39,189,11,16,160,103,1,210,60,233,65,128,
44,49,131,33,217,195,33,57,77,211,44,182,
171,234,114,89,85,233,51,34,195,187,123,227,
122,119,252,57,122,184,145,97,50,50,179,50,
171,187,89,36,155,63,80,64,33,17,113,239,
137,115,254,179,215,191,215,94,123,237,123,187,
212,82,255,140,204,248,115,228,46,254,151,188,
123,203,224,235,47,125,151,155,239,187,92,88,
16,9,66,149,74,90,165,103,131,31,122,4,
67,133,162,46,176,226,221,2,94,29,221,176,
206,30,94,253,223,49,169,108,161,7,14,61,
79,166,149,152,32,52,222,196,202,63,52,44,
36,24,210,12,12,252,218,40,0,95,89,50,
64,128,90,221,99,101,217,227,124,226,35,238,
220,47,147,41,188,124,72,65,101,241,63,202,
80,154,255,121,90,141,63,225,74,193,96,67,
140,16,156,34,65,114,212,152,39,182,74,196,
217,218,137,13,22,163,235,2,189,70,68,174,
56,218,164,131,94,132,180,213,225,206,178,141,
23,136,232,143,0,128,233,104,147,0,17,137,
145,243,171,121,194,119,179,80,208,208,52,9,
223,143,176,237,0,199,137,8,130,136,76,70,
33,8,34,186,221,240,104,94,68,191,31,28,
117,83,11,112,168,85,58,241,66,165,36,210,
105,153,208,182,25,250,42,178,4,162,253,217,
106,25,227,227,135,147,203,106,45,194,115,211,
220,233,13,169,57,30,162,32,144,215,70,39,
38,113,239,128,187,213,28,254,160,138,46,121,
12,4,141,64,121,131,76,126,228,121,160,232,
143,175,41,136,146,137,160,137,164,74,191,142,
172,89,12,90,235,244,165,18,107,7,67,150,
111,239,112,80,29,5,11,43,161,179,120,190,
136,36,75,159,91,128,77,218,61,148,208,167,
159,42,253,157,6,10,55,136,232,218,1,41,
227,201,225,198,38,69,172,169,132,161,75,77,
57,79,217,191,255,132,236,35,166,35,228,105,
138,57,46,135,31,61,158,50,167,203,12,93,
26,140,115,192,36,10,46,175,143,53,15,69,
143,2,118,164,83,167,72,171,89,56,3,20,
163,120,118,188,215,191,240,166,197,39,239,254,
144,165,43,199,170,160,94,59,66,51,36,76,
105,141,215,191,105,162,234,38,242,187,55,15,
251,154,18,8,182,67,184,52,205,182,144,165,
29,53,248,246,110,3,127,40,16,58,62,78,
183,250,72,47,49,24,121,235,217,126,136,31,
70,108,215,7,56,145,114,202,91,239,81,65,
254,218,70,135,28,30,51,39,146,104,237,51,
102,195,41,93,121,118,234,169,239,62,29,72,
60,252,59,181,174,75,26,48,211,51,124,164,
106,188,249,213,113,250,27,93,32,160,211,245,
217,218,24,64,12,157,166,135,252,173,107,216,
95,191,200,221,246,56,138,119,142,226,249,111,
60,250,69,151,21,146,185,37,124,183,135,65,
5,35,53,73,104,254,215,188,179,254,175,121,
158,119,185,234,174,115,171,253,255,112,127,112,
131,210,212,175,33,89,75,180,131,183,25,120,
34,93,71,162,52,166,211,156,156,100,53,212,
81,7,18,201,26,184,4,132,129,193,230,135,
191,199,108,101,157,10,29,244,246,54,7,155,
93,108,224,210,185,36,145,222,231,192,104,178,
179,241,45,156,56,71,44,11,40,110,151,178,
187,131,81,223,226,126,53,6,68,170,109,7,
221,210,153,152,205,114,241,106,18,105,175,65,
42,251,9,190,123,17,157,209,140,14,51,59,
3,49,184,225,0,199,77,146,8,84,218,198,
113,209,44,206,214,144,6,147,132,230,54,8,
163,49,168,138,40,96,117,234,244,170,49,209,
88,1,83,14,96,208,35,12,32,147,20,145,
229,4,189,182,141,16,135,167,42,224,17,50,
17,226,209,220,136,7,32,49,51,243,232,194,
149,231,69,172,175,15,232,245,124,26,13,133,
68,66,70,211,68,122,189,199,63,235,189,189,
209,73,109,98,194,192,182,67,246,247,157,103,
218,148,134,49,2,172,114,217,160,92,49,136,
198,242,68,211,21,250,154,202,253,219,67,72,
10,20,30,20,220,93,159,143,223,211,201,132,
85,66,201,35,145,55,184,30,94,197,200,79,
61,213,119,89,217,9,226,129,201,214,39,127,
200,226,151,254,11,204,226,121,238,172,183,185,
119,107,135,122,173,139,149,48,152,156,205,147,
47,38,17,236,214,231,26,96,59,86,250,239,
124,141,226,168,62,80,235,179,84,73,60,113,
54,179,16,107,56,153,4,146,154,99,186,146,
197,121,231,238,147,51,207,184,73,67,172,16,
132,34,242,137,195,75,140,112,166,102,145,103,
247,72,72,113,223,57,71,99,168,35,138,163,
78,158,165,222,53,110,231,243,143,0,9,129,
149,158,203,249,67,156,22,69,120,225,245,211,
210,209,100,70,228,215,255,163,227,236,64,104,
118,17,250,67,226,76,130,240,226,44,161,235,
83,29,216,252,160,30,162,53,98,180,240,101,
146,230,56,153,203,47,226,245,207,238,175,142,
237,179,94,31,158,158,115,237,69,79,21,169,
253,48,194,143,35,234,61,143,40,142,41,36,
213,199,214,28,62,21,40,30,2,245,167,178,
25,95,175,13,9,253,19,212,129,111,143,218,
8,63,197,68,205,17,53,180,184,135,32,65,
41,140,233,173,54,25,12,92,118,219,14,185,
140,198,249,139,41,54,55,134,180,219,30,253,
174,7,27,17,150,246,13,180,108,14,167,251,
232,107,177,123,251,12,147,239,17,238,59,200,
234,177,154,194,200,189,201,199,245,20,47,202,
255,158,25,165,195,61,127,135,181,15,255,111,
146,185,11,72,9,29,133,0,13,159,79,210,
231,168,238,77,144,46,190,129,144,179,168,214,
110,19,250,55,80,141,29,46,206,180,153,115,
235,52,150,107,172,111,13,137,162,17,213,115,
255,126,159,113,59,68,232,237,98,52,60,244,
56,70,81,68,130,32,166,57,60,156,135,12,
184,168,224,122,244,93,159,29,207,70,127,105,
134,221,177,12,74,228,113,240,193,187,72,135,
124,164,39,109,209,210,223,67,78,199,52,5,
151,48,219,133,135,188,180,66,115,27,161,91,
36,78,31,28,217,139,155,221,22,253,59,53,
212,123,107,40,73,133,190,35,48,55,103,226,
5,163,23,99,48,144,136,189,144,224,16,28,
34,68,116,206,6,237,74,229,241,52,128,170,
138,44,45,37,177,237,16,73,18,168,86,237,
83,246,224,143,124,38,118,200,234,234,227,13,
16,183,56,119,244,226,202,248,84,140,22,216,
163,116,222,60,156,143,172,170,34,254,88,158,
193,252,4,154,169,211,176,3,126,244,3,21,
75,249,10,195,238,95,113,249,53,153,239,237,
196,36,247,82,92,53,110,176,187,217,166,116,
169,204,181,104,17,53,243,11,56,157,131,71,
90,130,159,57,221,14,90,60,247,149,33,141,
166,203,160,177,203,208,139,104,172,237,97,31,
212,25,203,152,204,47,89,32,56,48,112,158,
222,182,27,158,237,103,237,22,79,35,53,206,
212,55,81,194,128,161,24,255,100,175,225,167,
96,51,142,59,24,125,158,40,61,166,78,5,
183,186,85,36,65,24,81,152,113,136,32,27,
40,135,99,57,45,77,34,246,52,118,36,147,
180,22,176,151,201,144,19,4,226,39,80,111,
50,17,231,252,27,199,7,22,230,113,48,153,
96,5,149,71,31,86,20,60,158,43,110,163,
156,56,105,199,162,69,198,24,112,208,79,178,
246,141,8,0,0,32,0,73,68,65,84,81,
60,145,85,124,180,151,230,98,185,198,83,248,
246,142,222,107,87,196,186,179,62,250,204,225,
104,30,252,95,182,122,120,81,140,218,109,145,
55,254,249,209,104,93,187,181,143,63,236,30,
237,217,56,134,253,142,67,99,224,61,250,222,
74,234,167,198,91,128,150,221,162,57,28,125,
198,38,194,136,62,122,220,61,116,186,16,184,
103,158,153,36,8,200,182,195,208,121,8,40,
62,205,102,188,36,187,108,52,134,199,223,231,
13,70,246,195,242,147,85,48,86,206,68,15,
52,218,141,247,89,244,247,233,153,22,233,241,
52,197,192,69,117,92,162,40,70,18,65,211,
69,22,190,190,200,119,26,11,148,39,47,63,
185,78,210,189,143,100,249,232,78,249,204,181,
154,185,95,229,250,90,64,162,253,30,149,139,
191,139,56,39,1,2,239,175,93,227,197,185,
61,86,86,122,24,197,128,241,139,15,106,31,
224,119,191,207,215,38,183,16,55,246,216,249,
96,200,199,7,46,166,41,163,170,163,193,238,
174,27,225,186,222,200,214,226,112,53,169,144,
179,247,207,92,155,134,71,143,4,38,54,182,
43,176,208,232,211,201,100,89,91,182,40,205,
124,241,232,122,243,193,215,104,69,127,132,33,
15,25,4,33,87,51,9,222,62,232,156,14,
29,2,160,217,196,174,204,237,206,128,115,41,
147,236,149,5,74,221,1,239,95,27,16,6,
54,234,137,2,242,48,148,232,28,84,25,70,
50,165,164,66,50,249,120,169,219,242,114,151,
98,81,167,84,210,207,244,155,133,97,140,227,
140,230,23,219,246,72,182,42,203,2,65,240,
217,184,242,58,19,148,216,58,226,135,197,100,
1,209,200,18,28,2,197,112,24,176,178,210,
99,242,23,207,241,86,218,34,51,112,40,5,
17,247,62,200,51,121,254,183,105,108,190,77,
118,214,225,123,31,89,244,219,99,4,233,29,
180,130,206,204,204,2,223,222,187,68,126,225,
183,144,85,3,167,115,240,216,52,254,212,75,
160,170,40,130,72,165,16,209,244,85,66,57,
65,195,110,49,136,83,204,207,207,142,156,120,
143,211,215,167,179,216,126,214,159,69,0,235,
211,175,181,133,140,24,71,96,21,126,178,215,
240,83,178,25,71,53,71,63,255,132,21,2,
97,224,142,192,71,181,112,128,129,35,112,49,
151,164,146,171,176,178,114,147,175,37,238,177,
17,198,44,78,38,105,108,117,159,122,175,5,
200,71,114,217,135,87,12,56,36,112,48,185,
127,191,201,236,172,133,126,104,0,24,91,6,
134,226,113,109,123,142,110,218,192,13,21,46,
85,182,9,197,1,43,77,29,85,180,145,5,
129,94,16,208,15,66,46,167,45,212,135,6,
88,220,222,159,224,124,247,61,98,219,29,9,
125,94,187,196,150,227,225,69,49,209,80,66,
115,46,147,58,183,120,156,173,132,1,146,162,
161,167,139,4,97,204,114,181,79,15,29,172,
199,220,91,89,251,212,120,251,240,222,138,158,
230,217,106,169,83,207,108,50,103,96,40,18,
150,245,25,168,167,82,74,195,246,67,170,157,
103,227,157,147,186,76,220,7,55,220,100,227,
249,57,198,15,134,232,59,219,200,190,71,16,
196,135,150,17,163,14,226,40,10,33,254,244,
52,91,179,198,112,107,47,147,120,224,193,251,
112,58,58,247,143,169,191,123,19,203,233,97,
100,70,129,89,144,94,227,78,37,226,178,118,
141,228,76,204,95,44,255,30,149,115,255,2,
223,29,224,244,92,194,165,20,226,198,30,65,
16,19,134,241,17,31,255,216,3,25,42,1,
42,138,224,157,1,107,29,143,24,145,72,210,
168,127,178,199,76,177,79,75,249,202,41,117,
67,186,116,137,254,202,93,204,242,199,252,220,
205,251,68,150,73,105,170,76,245,33,215,216,
88,239,163,134,38,179,24,24,65,140,32,248,
116,26,14,185,172,66,183,110,159,254,238,192,
70,79,88,12,235,117,90,157,39,3,133,235,
70,108,111,15,169,215,93,102,103,45,12,67,
162,209,240,168,213,28,156,195,121,232,178,44,
144,205,170,156,59,151,68,215,37,154,77,143,
173,173,193,51,3,134,143,118,4,18,0,142,
53,73,209,190,73,231,4,237,52,243,210,24,
239,230,83,16,132,52,111,149,177,213,57,50,
217,89,68,73,70,209,82,12,214,126,142,242,
228,215,24,159,210,217,191,255,39,236,95,116,
169,220,94,67,17,123,136,210,179,213,16,134,
157,85,46,95,144,184,113,31,148,92,138,97,
203,199,177,125,20,69,38,149,54,255,86,209,
53,217,65,11,37,240,217,74,20,248,251,188,
162,56,102,163,62,228,210,68,18,159,89,182,
74,7,4,13,5,251,165,41,196,157,91,79,
213,236,27,33,142,166,53,62,62,124,142,176,
17,135,237,120,1,49,78,194,80,64,18,67,
132,93,25,51,233,242,229,133,59,8,66,76,
20,137,172,214,75,228,101,248,104,215,192,147,
62,56,189,135,252,24,193,157,165,98,200,136,
66,140,227,171,44,245,174,33,111,108,115,239,
126,159,197,249,4,110,103,192,109,111,148,45,
68,181,52,19,75,255,228,44,104,198,163,49,
210,7,61,247,153,104,253,159,244,58,52,169,
24,213,8,31,227,207,245,84,64,17,199,96,
123,207,230,152,169,72,2,134,42,49,4,148,
104,28,69,188,71,148,207,32,239,108,141,46,
78,132,251,171,125,250,93,159,201,41,19,243,
66,154,184,254,233,141,69,201,226,2,201,226,
2,78,231,224,56,70,122,54,113,224,35,41,
26,97,232,211,177,11,20,50,99,39,50,155,
57,60,39,197,191,95,143,145,222,137,81,18,
117,118,110,252,49,227,23,191,129,170,249,236,
86,99,102,56,238,62,87,85,241,140,77,182,
44,11,40,170,132,61,12,152,82,183,137,38,
95,98,42,190,195,242,182,128,26,246,143,126,
55,102,164,242,193,233,80,189,15,86,181,131,
246,66,243,17,247,39,199,48,132,1,26,246,
70,151,197,78,68,225,114,133,187,126,120,106,
30,197,213,162,68,217,16,240,92,104,181,32,
253,242,2,90,199,161,180,54,146,181,186,129,
132,131,138,165,65,232,246,176,123,29,58,162,
70,177,144,198,208,159,172,52,113,156,144,59,
119,186,72,39,38,134,29,29,14,74,58,99,
99,199,191,159,201,40,120,158,193,246,86,159,
102,187,77,62,151,251,76,155,82,55,85,122,
181,0,93,151,168,84,116,178,211,105,222,93,
156,162,237,28,190,228,86,159,100,242,210,81,
243,97,102,252,121,224,249,163,141,216,111,175,
241,118,115,200,243,179,115,92,178,93,62,186,
255,251,84,206,253,230,211,5,164,48,64,229,
71,136,170,68,253,96,150,201,146,74,223,182,
25,12,92,82,79,49,153,237,111,122,181,172,
236,40,163,248,25,88,125,55,96,224,6,148,
166,127,129,15,150,119,232,153,105,110,165,45,
94,157,77,115,176,218,126,10,160,24,89,194,
60,177,22,198,136,30,205,248,125,150,244,204,
40,64,154,58,209,88,129,221,100,26,65,104,
31,214,34,34,230,11,35,49,73,7,248,164,
94,38,159,172,82,239,140,17,35,80,141,36,
206,149,54,152,47,36,33,142,17,90,61,4,
53,68,148,96,97,193,66,184,48,205,71,130,
64,16,199,184,247,76,52,53,125,198,42,198,
11,34,238,87,7,56,138,252,185,222,119,89,
18,184,50,157,166,239,4,108,53,237,99,139,
163,103,5,138,24,184,127,48,160,107,63,153,
167,84,101,145,115,149,4,125,39,160,107,7,
164,205,227,143,54,18,227,40,254,50,205,161,
136,114,101,30,221,15,16,151,55,9,130,62,
162,40,96,167,243,188,127,127,140,100,241,75,
79,151,98,58,125,122,245,21,180,100,14,226,
152,221,155,127,132,43,175,34,167,5,116,251,
57,102,191,240,159,156,221,36,233,113,244,243,
62,162,25,146,11,127,131,100,113,30,65,148,
200,158,251,207,217,220,248,95,25,63,63,71,
162,113,23,81,20,24,31,55,88,91,27,21,
116,19,9,133,252,82,129,244,229,9,130,205,
26,181,91,85,170,214,75,92,72,239,147,146,
53,46,164,147,236,235,47,35,236,175,49,216,
220,60,147,239,105,42,76,120,43,172,86,255,
13,138,245,219,71,117,10,89,144,56,143,133,
119,208,166,177,63,196,219,104,195,237,3,94,
57,87,38,120,97,140,186,44,146,144,37,198,
31,140,74,141,99,228,40,160,239,202,160,154,
104,137,4,227,179,42,27,171,93,148,200,39,
66,98,24,138,200,86,134,48,22,185,119,111,
153,133,185,121,18,9,235,83,45,180,195,32,
194,15,67,20,249,248,153,237,238,218,52,155,
30,229,178,78,62,175,33,138,2,149,138,206,
96,16,144,206,148,232,118,71,251,33,20,20,
2,65,70,139,236,135,104,147,18,54,9,114,
15,117,203,138,113,64,4,92,184,144,26,1,
148,161,179,239,28,159,4,187,73,133,254,206,
159,51,102,92,101,208,252,1,94,232,161,170,
211,100,39,94,65,148,100,102,175,252,54,237,
253,15,184,230,127,192,87,94,245,41,180,182,
136,63,37,152,214,215,222,33,85,92,162,187,
255,175,248,202,107,1,203,221,33,102,122,130,
40,142,169,53,250,68,97,132,149,252,219,55,
52,232,103,37,163,120,16,104,90,67,159,201,
172,65,214,120,17,51,255,39,108,181,102,88,
127,197,39,183,243,9,174,27,62,145,114,186,
45,94,38,29,13,25,146,196,164,71,128,66,
139,242,41,169,236,201,19,244,131,87,34,154,
25,35,26,43,208,173,153,84,56,13,72,126,
20,211,139,58,168,114,130,237,218,60,186,110,
51,145,217,103,33,105,48,105,88,136,219,85,
196,221,3,132,163,209,0,2,82,54,201,70,
62,75,245,208,41,86,30,76,144,41,190,122,
166,232,124,123,183,135,31,132,15,252,6,63,
183,53,155,183,80,37,145,156,165,146,53,213,
199,134,138,79,5,138,174,237,211,236,123,159,
250,133,41,67,198,84,37,76,85,162,148,58,
205,165,101,39,94,230,222,138,76,24,12,249,
196,252,62,95,158,132,208,200,49,188,52,134,
77,137,70,250,231,41,84,174,50,108,238,19,
199,17,221,234,45,90,173,31,98,25,58,126,
172,160,160,97,10,7,168,81,147,190,172,35,
137,1,202,148,205,246,221,77,166,46,254,6,
211,47,254,51,156,238,30,245,253,183,208,147,
227,72,170,249,8,202,42,71,170,241,38,253,
221,219,164,158,91,58,13,114,99,191,195,15,
118,255,148,244,188,201,108,218,230,131,154,197,
248,23,251,148,157,22,44,78,240,193,114,1,
127,77,101,62,173,81,185,170,209,74,78,16,
79,36,70,244,73,24,81,140,225,70,242,85,
150,242,50,157,181,26,173,230,177,213,69,171,
229,97,221,90,230,124,110,151,122,127,25,33,
151,1,65,96,70,14,208,223,93,166,211,114,
57,127,62,69,173,230,176,181,53,164,250,241,
58,147,246,128,197,151,151,208,199,116,196,131,
22,177,166,226,217,34,194,123,159,208,58,112,
70,93,170,154,200,193,129,139,16,197,40,140,
58,158,147,169,12,146,32,16,34,210,244,187,
220,189,119,23,81,18,25,175,140,81,170,84,
30,217,43,226,185,46,7,7,7,12,6,67,
18,169,4,227,149,241,35,43,220,122,163,139,
227,132,212,106,14,149,138,65,16,68,36,147,
50,197,162,206,242,114,143,94,207,199,214,179,
12,180,60,197,246,29,196,195,42,203,144,20,
30,58,227,172,158,5,37,199,38,157,82,142,
38,219,133,226,232,183,194,72,166,215,155,230,
82,177,133,82,222,195,236,239,83,200,168,124,
116,71,165,55,112,49,186,147,152,217,73,100,
205,32,85,124,142,238,198,219,12,131,24,53,
25,227,12,31,207,99,199,97,64,183,246,67,
114,201,31,240,230,151,100,118,135,30,247,246,
92,202,230,24,205,190,207,254,238,40,56,72,
146,248,183,46,118,254,44,101,20,8,199,253,
22,133,169,159,99,231,254,53,126,105,182,193,
119,123,19,76,191,222,193,253,235,213,163,32,
111,89,50,115,115,9,86,86,122,108,219,101,
6,130,197,249,232,14,42,14,45,202,212,153,
32,207,46,193,209,120,223,211,135,220,66,65,
59,77,151,0,219,3,23,109,224,48,99,233,
71,255,252,81,171,71,215,15,80,101,135,116,
178,137,237,36,121,173,144,34,173,200,68,55,
215,136,182,106,72,170,200,112,16,18,11,49,
86,214,160,125,110,154,107,173,19,18,193,116,
151,84,241,226,169,239,223,168,219,248,225,231,
255,92,45,85,34,115,162,22,241,164,243,228,
167,2,69,218,80,88,42,39,88,171,15,78,
141,112,60,251,165,50,125,39,160,214,27,57,
180,198,128,42,9,20,132,209,236,225,220,212,
168,143,160,189,151,228,71,215,63,34,145,253,
77,204,249,105,52,69,67,99,100,233,220,169,
221,160,213,254,215,204,79,181,249,98,162,139,
182,215,32,86,100,68,215,135,224,144,239,147,
21,28,213,224,32,49,206,199,250,241,245,232,
169,49,10,252,34,122,234,241,167,175,194,244,
235,20,166,94,59,123,19,84,131,228,244,63,
37,140,66,222,91,253,14,249,133,95,96,167,
95,231,174,243,33,194,138,132,89,124,5,213,
115,72,68,255,27,230,116,134,137,68,196,118,
127,146,133,226,14,53,39,196,114,100,210,115,
251,124,152,122,129,95,44,124,194,245,183,118,
137,93,15,129,152,56,134,126,63,160,223,239,
0,29,68,81,64,20,161,31,198,200,25,149,
153,153,145,125,198,254,190,115,84,52,174,174,
53,40,4,49,230,139,99,176,186,71,183,15,
181,157,46,131,158,251,152,2,161,130,134,199,
206,254,38,149,202,24,177,164,147,49,22,89,
190,119,27,215,243,216,222,217,193,243,60,166,
166,167,33,6,223,247,240,195,128,65,175,207,
230,246,246,209,203,210,235,247,112,28,7,207,
113,113,28,135,48,142,168,148,43,148,42,101,
182,223,171,145,47,140,36,132,59,59,199,217,
67,194,174,145,176,107,135,89,68,25,155,4,
34,33,21,214,207,210,134,73,5,33,173,80,
176,180,163,224,112,47,151,130,24,246,91,83,
252,242,226,1,51,59,53,164,122,139,97,49,
207,95,93,123,129,242,210,111,146,125,200,69,
81,53,83,8,170,204,71,173,38,11,134,137,
219,104,160,168,169,51,89,103,191,185,142,211,
95,231,151,127,69,68,37,70,190,126,159,131,
74,153,168,174,98,44,150,185,187,217,163,94,
29,85,75,170,123,109,42,19,89,148,207,153,
10,248,153,205,40,24,245,91,180,6,62,89,
75,65,215,95,38,21,254,41,63,31,249,220,
153,46,176,88,216,163,93,183,41,20,116,116,
93,100,221,159,163,109,249,232,246,22,106,108,
31,169,156,178,84,143,60,159,154,84,30,57,
120,171,209,112,71,180,170,34,19,235,163,189,
152,48,219,188,181,153,226,11,19,54,154,0,
195,200,161,122,152,233,182,250,69,42,217,45,
134,110,130,251,61,155,87,134,67,66,59,228,
158,112,153,206,106,141,43,249,125,146,105,149,
224,185,115,220,120,184,206,168,12,8,125,247,
200,70,166,53,240,105,14,188,191,21,247,59,
103,61,253,204,147,167,122,43,178,150,130,165,
165,184,177,211,125,44,88,60,208,221,202,162,
192,254,137,162,183,101,133,156,60,223,107,137,
10,253,126,19,103,240,54,174,189,66,20,171,
196,97,132,169,126,200,23,115,53,132,213,125,
156,53,7,253,80,194,25,251,33,113,57,139,
176,63,114,114,180,237,24,147,1,59,143,40,
254,168,86,230,169,42,55,237,237,27,84,151,
191,139,102,229,168,92,248,58,122,170,120,88,
55,145,48,114,87,64,16,80,204,44,189,149,
22,182,189,201,164,53,135,128,130,58,28,32,
118,186,228,51,29,18,197,18,239,108,206,97,
7,117,222,156,13,137,122,49,77,47,66,184,
58,199,244,106,131,149,101,239,148,42,233,193,
154,156,52,209,180,145,180,214,178,228,35,11,
141,40,130,203,151,211,68,209,161,234,40,130,
205,239,174,161,42,18,205,106,247,76,205,228,
97,134,54,66,96,114,98,242,168,132,236,105,
38,11,151,95,98,229,198,7,120,129,79,237,
224,128,70,179,73,28,197,167,154,136,206,156,
98,155,45,178,185,44,169,116,134,65,242,10,
3,81,101,95,52,233,43,93,196,214,50,153,
116,138,232,176,45,191,63,232,147,48,19,71,
149,194,7,47,233,46,243,68,72,136,135,70,
47,19,19,38,157,142,199,236,197,44,235,75,
151,81,67,131,157,186,194,32,239,179,28,7,
84,91,83,148,210,59,220,238,4,168,165,28,
165,74,142,239,124,156,162,178,248,27,71,3,
92,78,241,217,245,53,72,143,2,252,253,93,
145,98,242,184,190,208,220,248,54,150,242,9,
169,36,204,44,8,180,90,17,94,44,35,123,
62,155,74,146,149,183,50,8,42,184,145,136,
227,199,204,204,21,89,95,173,33,136,2,161,
31,161,28,30,176,226,40,166,81,235,81,109,
247,200,228,18,140,79,230,158,218,206,121,109,
165,138,231,250,84,198,51,40,154,130,99,251,
12,250,54,165,114,6,77,127,122,174,225,103,
42,163,56,92,219,45,155,140,165,144,159,124,
149,235,247,63,228,75,133,251,36,111,55,232,
8,35,87,227,124,94,197,178,100,186,67,200,
12,110,99,63,198,76,200,193,66,198,103,131,
139,148,216,194,160,79,140,72,85,90,228,203,
179,163,250,102,56,55,113,52,225,238,74,193,
163,100,9,180,6,105,182,221,136,93,91,34,
147,176,105,247,11,164,205,81,141,209,113,45,
178,202,144,120,204,228,206,240,50,207,119,254,
132,120,65,66,200,228,136,198,139,196,9,131,
200,57,126,231,99,31,52,111,233,8,36,28,
63,100,245,96,240,183,226,62,171,178,248,76,
227,106,207,0,69,20,199,244,156,128,158,29,
224,248,17,94,112,248,223,19,82,165,148,113,
60,136,72,126,40,133,63,213,56,66,76,109,
247,143,89,186,50,100,233,147,191,162,223,114,
200,77,103,168,14,69,6,183,118,216,26,4,
4,135,63,95,219,119,80,117,145,114,201,64,
187,156,97,171,144,97,118,224,16,45,215,136,
211,22,97,8,50,167,79,90,162,244,116,167,
193,204,228,21,50,147,87,158,92,28,139,35,
198,150,190,65,28,71,104,201,60,141,213,111,
209,221,106,34,41,54,25,219,69,139,99,94,
157,82,216,238,167,249,235,45,131,148,5,110,
28,112,183,27,48,185,56,142,178,124,235,209,
167,38,55,36,147,81,112,156,16,89,22,16,
69,129,197,197,36,43,43,125,86,87,251,4,
65,196,220,92,2,199,139,104,238,247,158,142,
222,21,68,198,139,18,61,109,138,253,186,68,
218,94,37,138,34,156,64,224,194,133,11,212,
235,117,118,247,246,8,195,167,19,36,232,186,
193,182,116,133,176,235,145,77,130,17,5,200,
138,203,138,252,75,164,6,171,44,37,235,136,
68,104,154,206,246,238,14,153,76,134,132,53,
82,89,185,24,12,73,177,207,12,2,163,249,
22,170,154,68,159,81,88,203,20,104,172,218,
12,58,80,214,247,241,196,89,48,50,100,18,
251,40,178,135,23,193,123,125,27,75,150,112,
162,248,177,185,176,172,154,132,171,57,204,238,
69,172,204,2,70,122,156,97,99,151,198,189,
255,139,241,137,93,186,3,137,251,43,73,150,
87,5,222,120,179,207,187,239,198,56,76,161,
137,211,204,255,220,243,12,90,85,26,3,15,
215,245,216,88,63,96,114,182,192,204,108,233,
8,240,6,61,135,219,55,182,144,220,54,195,
216,162,221,28,176,181,118,128,162,74,168,170,
66,58,107,50,49,93,120,36,93,53,28,184,
236,109,55,73,166,12,106,123,29,84,77,65,
213,20,76,83,71,146,159,141,222,250,89,203,
40,96,36,154,169,247,92,138,73,141,212,220,
239,240,221,181,63,37,109,44,147,54,110,16,
30,212,104,52,70,242,245,72,144,152,158,182,
216,220,28,224,186,209,25,101,84,147,202,145,
253,199,54,231,152,228,30,2,17,19,194,10,
53,225,121,170,182,68,208,155,102,204,13,73,
104,14,65,168,80,208,125,242,90,135,166,231,
211,169,137,84,91,147,8,49,100,18,163,158,
174,153,242,93,174,31,148,233,170,99,60,23,
188,61,58,92,92,156,37,174,228,9,227,152,
229,222,144,246,161,147,97,228,136,104,123,175,
81,94,248,133,67,10,63,96,165,214,255,220,
212,77,9,93,198,245,67,52,69,34,99,40,
20,146,42,209,160,241,217,128,162,99,251,220,
236,180,121,86,123,153,7,110,177,67,47,100,
167,117,76,75,136,130,128,121,8,32,81,224,
177,191,242,103,228,23,119,41,104,10,170,169,
144,118,93,154,43,117,122,109,159,102,55,70,
57,172,4,171,154,200,194,124,2,211,148,65,
150,113,17,184,237,135,220,211,52,114,231,95,
65,78,182,112,27,49,98,156,248,169,221,88,
73,86,145,14,253,170,226,56,162,28,127,132,
37,135,216,131,144,76,26,6,27,29,212,230,
144,197,11,147,228,242,9,254,120,93,66,150,
92,254,122,35,203,151,39,108,178,19,89,156,
70,27,81,132,225,240,56,64,215,106,14,170,
42,210,239,7,8,130,203,220,92,130,102,211,
35,142,71,210,92,207,139,88,89,233,81,158,
204,62,101,13,80,64,51,84,132,243,151,209,
207,47,242,198,234,15,249,209,119,12,26,226,
56,123,181,21,198,168,83,25,47,147,74,165,
89,89,93,33,240,63,189,121,106,111,119,23,
65,108,96,8,1,51,133,231,8,196,52,221,
65,139,146,247,45,12,69,100,53,249,117,102,
185,131,34,67,33,159,71,62,1,208,93,242,
44,242,209,33,224,72,120,94,132,191,30,35,
168,34,209,61,152,206,169,72,146,192,216,132,
73,69,134,11,83,147,124,228,196,108,157,240,
45,27,4,33,226,116,141,131,141,239,81,154,
253,234,89,32,75,149,153,191,242,187,136,135,
13,72,190,221,227,96,247,47,241,109,135,72,
248,2,129,255,49,232,117,228,232,42,31,191,
183,74,178,242,159,82,176,142,85,90,138,214,
199,15,34,214,239,87,137,163,120,68,55,29,
130,68,175,107,115,243,227,13,194,32,194,60,
252,55,81,20,177,18,26,130,36,34,34,32,
8,226,200,10,253,17,203,180,52,94,255,234,
133,207,220,17,123,114,181,173,204,223,155,206,
236,103,89,163,233,120,26,130,40,161,103,95,
70,202,253,42,123,202,119,120,101,238,15,48,
147,18,120,62,210,46,36,12,153,236,194,57,
118,91,42,91,135,116,104,134,58,73,154,71,
89,239,136,22,15,120,160,156,173,148,53,42,
214,200,36,48,204,193,158,187,192,122,199,162,
239,123,220,233,184,216,174,197,100,113,5,69,
141,145,108,139,124,234,116,191,148,168,85,217,
23,107,36,166,10,204,45,86,216,15,98,250,
189,33,155,3,7,231,240,48,29,173,78,98,
170,231,41,45,125,237,40,46,222,221,239,241,
121,218,118,229,44,245,148,171,247,40,235,226,
179,1,197,147,154,248,158,88,20,209,36,186,
118,192,253,90,255,232,247,37,81,96,190,104,
161,186,54,78,183,74,173,250,255,242,213,231,
93,178,203,91,108,252,217,1,183,122,62,81,
8,126,112,232,24,42,168,168,170,136,105,200,
204,204,154,71,153,73,240,210,5,238,132,17,
248,54,97,28,19,9,18,243,9,131,108,78,
230,237,224,45,182,238,108,160,104,25,100,49,
73,118,236,37,36,85,127,108,81,179,83,189,
133,158,28,67,79,62,219,16,154,126,245,46,
115,189,6,166,28,32,102,13,106,3,141,162,
229,34,120,1,220,90,37,243,250,85,94,31,
115,120,191,54,186,230,119,171,121,102,94,127,
145,159,107,108,146,202,202,92,255,163,219,71,
253,7,113,12,91,91,67,202,101,125,228,104,
235,132,140,143,27,52,26,46,115,115,22,123,
123,14,131,65,128,166,137,36,18,242,41,239,
253,179,32,1,177,162,243,252,69,157,181,153,
89,166,179,13,122,243,23,80,214,43,76,237,
125,130,57,102,82,219,86,184,125,243,54,158,
231,61,118,24,18,128,34,43,72,178,132,101,
90,163,128,40,8,20,242,121,34,127,136,223,
111,147,150,192,51,82,212,196,89,74,180,216,
102,137,89,110,162,235,167,239,247,3,149,137,
40,10,92,190,124,44,119,14,130,8,65,145,
225,202,60,177,169,19,136,34,177,161,209,243,
67,82,74,132,34,10,200,130,128,123,56,156,
72,80,96,32,191,203,176,179,128,153,158,60,
115,189,226,137,46,213,131,205,183,64,217,98,
246,82,204,206,71,105,116,35,38,51,254,59,
24,233,50,81,24,60,50,211,236,246,93,218,
205,17,13,144,56,84,60,121,174,207,237,235,
155,132,135,123,82,86,36,206,47,77,146,43,
36,143,38,255,125,10,106,179,186,178,71,167,
53,196,247,3,66,63,66,148,69,146,41,147,
115,151,198,145,159,209,55,42,61,232,160,132,
62,195,100,241,103,10,40,220,32,226,131,245,
54,154,34,178,112,200,91,39,39,127,158,91,
107,31,240,66,243,30,189,75,23,217,18,166,
9,15,84,38,148,77,198,42,67,230,84,239,
148,11,242,73,169,236,216,184,69,176,61,170,
255,101,179,199,188,124,44,75,148,83,109,154,
241,144,186,63,164,144,134,118,191,192,192,77,
146,208,187,24,218,224,232,255,79,46,63,138,
177,20,137,31,180,7,116,78,30,188,98,8,
54,211,148,10,255,1,137,220,177,27,240,208,
13,159,62,174,198,241,167,170,20,63,203,218,
105,217,148,82,159,221,210,67,62,73,57,213,
123,254,211,118,171,159,90,215,54,58,103,210,
156,133,146,133,38,139,12,157,152,131,189,63,
230,43,95,136,72,117,60,162,112,116,114,118,
220,8,1,72,38,100,122,174,200,220,148,69,
37,123,250,143,136,69,1,199,80,89,221,111,
33,9,2,151,210,22,249,96,72,66,29,93,
246,171,25,151,251,169,85,42,146,77,79,55,
249,240,222,15,72,136,95,164,48,117,90,102,
219,174,125,72,111,112,151,249,89,159,251,91,
10,82,253,57,178,99,95,56,106,167,255,212,
103,231,190,131,230,244,88,222,112,17,5,151,
165,69,1,47,16,209,228,8,130,16,193,11,
152,181,116,110,235,13,250,78,154,124,170,202,
206,32,199,159,101,38,153,50,36,94,249,242,
128,79,222,218,56,245,153,213,234,8,207,87,
87,251,71,125,12,130,32,48,63,159,224,198,
141,54,170,42,48,49,97,34,8,112,231,78,
247,17,5,108,25,31,153,169,108,68,99,241,
5,54,104,177,181,171,81,214,82,188,248,31,
198,196,63,42,240,201,59,7,100,103,138,12,
162,73,122,213,77,118,247,118,30,107,137,80,
44,151,144,4,129,84,42,133,102,24,248,190,
135,170,168,132,65,64,223,58,71,24,201,4,
72,84,156,13,252,104,138,138,184,206,6,23,
153,225,246,35,63,47,159,63,189,145,100,89,
28,53,84,222,90,37,60,55,67,52,54,162,
83,86,250,67,36,65,32,33,203,156,79,153,
116,253,128,59,29,155,129,107,96,84,186,28,
172,254,25,51,233,223,126,114,70,187,248,77,
106,119,87,8,91,33,19,151,191,73,20,199,
168,70,138,230,192,99,187,233,16,132,17,149,
140,126,170,153,72,81,229,35,117,151,170,202,
196,49,220,187,189,131,127,216,47,148,76,25,
92,92,42,34,167,158,97,104,145,0,186,174,
50,144,61,52,125,68,57,105,154,130,105,105,
136,159,65,85,213,51,146,127,47,220,99,63,
203,138,226,120,212,187,117,162,192,57,212,190,
68,215,223,163,183,13,99,202,42,99,234,14,
162,237,17,11,194,137,209,189,16,34,98,208,
101,72,10,147,46,25,170,52,203,75,100,135,
235,52,155,30,149,138,113,24,139,71,241,38,
117,98,2,95,38,81,167,218,154,194,212,250,
28,255,196,217,85,181,189,211,32,17,65,124,
247,28,19,115,191,114,52,86,224,40,3,86,
159,97,110,184,211,38,54,50,60,221,64,225,
167,95,97,20,211,25,6,100,45,229,179,3,
197,234,193,128,214,192,39,242,195,207,4,20,
39,87,57,165,49,157,55,143,64,177,91,187,
205,75,115,107,36,223,175,82,93,109,209,168,
219,248,94,132,44,9,228,178,42,169,98,146,
89,53,66,125,84,121,33,149,96,107,232,33,
0,231,82,38,51,150,142,239,199,184,78,140,
22,123,68,31,175,114,69,13,137,74,57,148,
243,105,180,169,14,253,237,247,145,247,146,100,
42,87,143,144,217,103,159,96,172,197,109,23,
132,18,100,149,183,105,239,127,128,191,83,65,
148,50,200,226,56,201,194,121,84,51,69,171,
250,46,27,213,54,166,160,145,76,156,35,242,
251,44,9,203,132,177,140,42,4,148,38,210,
132,72,4,97,76,16,203,4,153,20,146,172,
163,137,2,75,73,19,153,93,54,27,211,140,
231,215,217,107,206,96,8,17,219,115,21,242,
55,246,105,60,194,51,201,182,195,163,19,184,
97,72,71,65,213,11,4,178,135,51,126,199,
199,13,118,119,143,41,61,27,29,5,31,29,
135,212,88,137,111,187,73,250,67,23,85,25,
48,147,233,19,6,26,246,249,115,92,54,45,
110,94,87,25,236,245,200,142,69,88,137,4,
235,235,171,120,222,89,213,197,238,238,46,179,
211,51,168,170,198,205,155,55,209,52,149,177,
202,24,174,231,49,108,127,159,56,142,81,52,
149,214,248,47,179,40,222,100,135,69,38,89,
102,151,249,83,50,216,24,40,22,52,242,121,
237,112,175,11,167,211,212,48,66,92,223,37,
28,43,176,111,187,100,21,25,4,129,231,50,
9,86,122,54,111,239,154,136,81,145,148,217,
166,214,158,64,143,135,248,118,23,197,120,146,
35,172,204,160,29,18,73,121,198,11,73,188,
32,226,214,110,143,190,19,156,162,51,78,2,
133,174,74,36,147,6,166,165,161,27,42,157,
86,159,78,235,152,255,90,60,63,142,44,61,
123,225,113,124,42,207,248,84,254,39,195,43,
59,125,228,48,248,59,239,30,251,227,1,198,
241,255,39,199,191,200,234,234,247,120,222,94,
225,45,231,87,57,208,117,94,90,108,211,241,
33,220,189,141,123,152,49,11,140,132,21,107,
92,102,134,59,212,54,15,16,11,22,65,249,
10,41,86,136,227,248,208,234,59,34,6,214,
250,167,123,128,202,217,45,246,91,83,248,129,
198,84,241,209,163,85,183,134,167,137,155,240,
192,162,50,245,75,200,143,144,230,63,11,123,
24,27,217,159,42,248,62,235,122,160,208,146,
129,163,17,121,63,14,134,9,2,204,22,76,
138,73,237,33,36,243,16,20,56,88,169,83,
221,119,144,68,129,233,41,147,108,193,192,246,
37,52,57,66,149,32,86,100,226,116,2,177,
126,216,244,162,42,56,231,102,56,112,60,190,
92,202,144,62,148,45,42,138,64,179,29,16,
212,91,164,149,81,144,21,162,8,83,16,24,
55,52,54,199,235,180,237,127,67,123,245,251,
232,194,60,217,226,107,68,177,125,148,136,198,
64,213,247,32,239,97,88,54,95,152,109,112,
171,113,157,237,157,111,35,236,38,16,166,27,
12,236,73,108,97,192,68,238,219,188,208,111,
225,44,31,96,199,17,142,27,210,106,7,156,
59,33,174,10,167,82,120,18,244,218,33,139,
73,131,9,83,227,109,113,151,173,230,4,126,
168,210,28,198,220,208,34,190,250,205,171,240,
167,215,31,9,22,163,205,52,170,79,168,170,
72,20,197,35,203,110,69,199,84,194,35,79,
154,16,145,16,25,67,240,144,165,120,164,24,
46,36,217,233,36,40,101,219,232,170,205,199,
187,147,52,99,5,35,173,35,141,141,243,66,
226,26,55,254,96,141,144,24,71,49,73,231,
202,216,189,38,131,193,67,83,10,227,152,102,
179,65,167,219,193,119,61,38,38,38,232,246,
122,184,142,195,220,252,60,97,16,16,4,1,
66,124,131,85,174,162,226,34,17,28,13,170,
63,73,95,166,82,10,150,37,19,205,142,131,
23,48,72,20,217,20,231,144,119,170,196,130,
128,155,182,104,237,181,9,37,155,189,158,137,
165,245,9,227,152,31,110,230,121,33,221,226,
197,131,219,244,10,21,190,35,183,9,59,69,
158,198,64,175,56,251,235,232,185,50,182,23,
114,119,191,143,23,68,103,94,18,63,140,80,
14,79,246,138,36,114,233,185,153,163,161,68,
169,140,133,97,168,216,135,83,105,100,69,124,
10,179,156,159,238,26,106,38,210,207,96,141,
226,20,237,235,248,156,172,68,218,198,151,240,
163,61,84,35,66,27,255,93,86,90,255,11,
139,236,227,118,250,104,15,41,160,150,248,132,
101,158,103,41,250,152,168,182,198,160,181,195,
238,194,101,150,196,227,236,254,118,103,64,219,
59,75,239,86,178,91,180,251,5,108,207,194,
80,7,68,67,145,168,158,2,201,135,56,70,
26,27,66,32,32,40,163,217,60,162,155,67,
206,157,237,238,31,122,33,43,213,254,231,118,
255,84,105,212,8,29,68,49,73,253,217,164,
223,219,45,155,221,150,115,130,122,26,28,206,
62,120,74,87,203,135,221,99,101,81,96,58,
111,98,250,14,195,135,220,42,124,167,133,72,
76,16,198,36,19,163,70,25,57,101,208,86,
83,164,114,10,36,45,2,77,33,78,154,224,
5,196,229,46,177,166,17,167,19,200,2,124,
33,113,246,230,107,154,72,10,251,40,107,17,
234,109,228,119,174,115,245,203,47,140,208,94,
112,137,102,246,240,162,61,118,234,239,34,205,
63,218,187,201,117,67,110,180,251,196,162,192,
171,23,84,54,250,93,50,170,65,115,55,98,
169,80,103,167,61,199,228,181,143,81,13,147,
56,240,41,151,117,146,121,3,78,200,94,165,
229,77,140,251,219,40,47,158,167,55,208,17,
69,129,175,150,146,220,145,91,108,55,125,250,
155,38,222,238,5,254,194,218,227,27,175,76,
227,126,127,237,177,117,135,106,213,97,106,202,
196,247,35,54,246,66,38,133,128,253,97,72,
181,234,17,163,2,34,26,14,217,172,70,225,
107,75,104,155,85,26,179,37,122,55,178,88,
122,15,215,55,249,229,124,151,124,65,4,4,
4,33,38,206,141,177,240,197,14,213,235,251,
36,195,24,195,156,134,104,18,219,118,184,185,
222,160,71,30,49,14,40,11,59,104,154,68,
68,140,105,154,52,235,13,42,227,99,68,137,
4,33,50,53,121,142,72,150,152,96,133,105,
238,178,203,60,59,44,62,82,167,254,192,14,
68,216,171,227,189,122,133,181,221,113,174,244,
127,136,216,29,109,142,96,98,158,127,213,153,
166,231,39,120,161,236,114,191,149,228,173,182,
200,82,177,206,75,107,247,17,134,14,233,238,
128,55,102,199,121,91,240,105,239,221,192,202,
204,63,249,180,20,120,4,195,62,157,97,139,
108,28,19,9,49,245,190,123,106,55,183,170,
125,18,186,140,211,169,163,104,1,178,55,60,
42,114,10,192,194,164,204,218,202,72,225,226,
28,108,163,26,225,231,234,30,171,119,247,144,
195,144,142,34,253,100,175,225,115,112,143,61,
181,2,31,226,224,169,190,191,214,238,144,58,
49,185,77,212,22,248,240,96,134,178,240,14,
251,251,227,12,188,4,162,230,50,243,218,52,
94,107,200,234,199,167,157,0,230,185,201,58,
151,40,177,133,111,76,18,135,9,60,43,133,
92,74,16,91,6,139,186,198,218,192,38,138,
33,26,10,68,123,57,4,49,134,88,34,17,
199,180,172,20,81,203,36,171,92,36,85,188,
76,28,71,196,81,68,107,229,125,68,73,197,
243,247,144,46,110,18,198,61,134,205,93,36,
89,67,84,142,105,153,214,208,35,232,61,162,
108,28,133,35,115,198,167,189,183,79,233,30,
251,240,222,242,128,251,182,196,92,193,196,182,
31,178,17,233,183,8,125,23,65,148,232,58,
1,138,40,98,28,210,100,61,39,96,183,49,
124,168,70,97,149,142,207,131,79,225,106,121,
210,61,214,212,36,206,149,19,168,143,144,254,
121,78,151,116,113,23,235,195,117,108,69,36,
123,117,22,187,146,39,74,90,88,41,145,239,
31,116,168,24,42,243,135,96,176,19,70,180,
117,157,43,190,143,242,221,15,8,46,47,16,
23,50,8,213,38,113,121,196,251,137,213,6,
137,251,123,184,61,15,221,82,16,226,152,88,
18,137,243,25,196,141,61,174,90,6,133,140,
197,135,173,62,178,44,32,149,3,194,248,136,
142,62,125,250,149,101,106,142,143,42,10,216,
189,144,215,243,105,62,110,245,16,136,152,49,
20,106,141,1,242,243,51,200,200,8,219,7,
244,214,3,148,78,15,75,214,144,101,225,136,
74,33,140,144,238,108,144,214,85,130,203,11,
216,195,136,25,73,195,42,218,220,169,101,152,
58,247,207,233,238,47,243,177,71,95,160,0,
0,32,0,73,68,65,84,94,240,7,124,245,
165,33,119,63,168,50,24,156,221,32,181,154,
131,174,139,72,146,192,192,13,216,222,10,176,
180,136,66,74,96,48,8,142,2,112,106,62,
207,91,134,78,233,249,37,110,237,141,147,207,
236,49,91,220,99,60,204,145,47,138,135,153,
225,225,95,43,10,104,175,95,68,185,50,139,
210,245,40,109,55,184,255,215,203,168,134,200,
165,169,52,59,251,91,4,195,54,61,125,145,
80,47,160,200,42,90,41,57,50,44,11,61,
36,33,38,25,27,140,9,107,28,104,231,81,
99,17,21,159,140,119,192,38,231,73,243,104,
59,248,142,144,99,71,92,68,122,215,37,113,
251,247,232,230,84,50,25,149,56,147,96,45,
105,33,58,85,42,90,154,151,138,49,207,23,
28,26,174,207,206,48,36,154,40,129,34,35,
110,238,145,219,171,51,187,96,178,181,230,63,
210,213,248,97,250,73,79,23,121,48,249,180,
99,251,212,247,78,159,228,28,77,167,116,120,
234,75,103,42,108,185,167,21,126,105,11,212,
134,72,171,217,167,225,152,164,242,202,231,234,
30,235,6,17,97,20,158,120,63,127,66,215,
240,57,186,199,142,128,226,216,61,246,211,99,
164,128,167,23,200,156,208,252,71,169,255,134,
225,141,255,137,124,244,135,244,74,255,130,13,
117,192,252,130,130,214,238,51,231,249,172,221,
62,62,173,74,4,204,114,139,125,38,145,7,
125,198,100,151,247,213,215,120,117,174,77,63,
8,217,236,13,71,32,81,79,160,245,175,80,
57,247,205,163,222,157,56,142,200,87,239,162,
140,167,49,243,227,167,105,193,210,244,225,65,
184,199,246,222,191,68,154,106,51,88,95,166,
60,251,245,83,130,154,174,228,130,55,124,212,
201,102,228,6,251,180,247,246,51,184,199,30,
83,213,176,237,43,156,175,36,78,213,201,69,
73,70,77,100,9,16,217,108,143,234,204,89,
65,65,145,68,26,158,7,86,226,108,49,251,
179,172,140,169,176,88,182,206,84,210,227,40,
164,186,246,231,36,147,215,121,227,106,140,52,
62,75,40,37,48,78,204,65,88,238,13,105,
122,62,77,207,31,233,237,5,129,219,221,1,
95,40,164,145,62,222,32,14,34,164,189,58,
65,33,115,12,18,155,251,8,173,238,104,28,
231,197,89,130,233,220,41,197,129,242,253,107,
72,126,192,84,54,73,190,152,3,203,130,132,
65,219,11,232,250,33,93,63,64,16,160,233,
250,216,39,250,66,84,81,228,181,124,26,65,
128,186,235,3,49,185,253,6,223,28,236,243,
167,27,147,20,74,10,47,46,168,168,27,7,
212,235,163,105,111,139,139,201,211,143,167,219,
71,232,130,172,41,152,75,211,96,129,216,17,
216,47,182,168,173,255,53,169,236,37,124,239,
215,248,81,240,45,94,186,228,97,183,28,250,
125,159,92,78,99,104,135,236,237,14,153,154,
26,117,105,135,97,140,142,71,132,198,220,66,
10,145,209,48,161,7,5,59,57,8,136,34,
120,111,115,146,74,110,147,87,146,26,227,164,
176,18,17,15,15,117,221,183,61,110,119,7,
12,130,81,90,126,161,82,194,204,87,177,155,
93,10,25,141,76,122,30,39,16,57,168,29,
112,176,247,67,196,68,129,176,35,49,24,244,
40,148,39,16,98,143,178,85,32,146,53,74,
175,206,178,52,102,208,250,96,147,253,85,129,
180,216,160,30,77,48,197,50,194,9,158,102,
185,59,206,184,38,82,174,127,15,199,137,168,
247,70,38,128,153,204,168,232,55,107,168,164,
198,36,118,237,14,144,64,18,4,74,186,74,
25,144,62,190,59,170,105,152,58,43,87,150,
184,247,97,154,201,249,47,60,61,199,123,232,
235,191,221,178,207,28,12,218,67,143,233,252,
104,15,74,162,128,165,201,167,234,24,0,115,
75,21,122,31,174,81,175,118,152,44,100,80,
14,183,236,176,239,226,56,30,138,38,227,187,
33,7,213,54,115,75,21,84,245,167,215,201,
173,132,62,114,24,240,179,190,106,93,247,20,
80,136,178,194,32,61,199,27,222,91,252,160,
254,135,84,179,95,197,187,253,22,23,46,38,
72,95,157,66,89,105,31,205,95,63,162,146,
216,134,16,210,164,185,90,222,101,117,183,76,
91,223,101,215,118,9,247,50,20,140,95,35,
185,184,244,16,157,46,146,174,92,60,101,64,
122,230,25,233,73,68,55,135,32,14,240,133,
157,71,82,202,159,215,50,84,137,172,165,144,
212,100,44,77,126,172,152,170,119,194,199,175,
53,240,31,95,204,254,44,43,159,80,152,47,
39,206,124,185,55,108,179,187,246,251,228,102,
108,74,9,157,190,235,35,36,114,244,219,33,
86,24,31,121,252,184,97,68,86,149,105,121,
1,55,59,163,19,142,46,137,164,20,153,40,
147,68,40,100,137,166,79,12,149,119,125,226,
108,10,12,29,242,105,188,94,116,186,238,46,
64,120,110,26,161,218,68,172,183,177,90,61,
226,124,154,224,249,115,148,116,149,210,67,170,
217,97,24,210,109,219,72,86,138,173,161,203,
123,141,46,95,40,164,81,37,17,69,242,89,
213,139,188,184,117,157,217,82,25,105,233,127,
224,71,247,127,143,175,191,222,100,229,175,214,
232,116,124,54,54,6,20,10,26,138,34,114,
112,224,96,154,50,217,172,138,208,238,35,127,
116,143,216,208,72,158,155,102,66,26,242,177,
255,125,196,166,70,97,238,117,2,111,134,191,
88,254,223,121,117,113,155,241,104,192,74,121,
130,189,134,200,87,150,182,16,26,29,210,105,
133,74,69,63,4,6,151,173,125,157,233,178,
64,175,231,99,154,18,233,180,138,82,76,178,
221,156,101,162,176,202,185,148,193,36,42,86,
234,209,202,138,123,189,225,17,72,0,220,145,
28,10,191,124,145,76,87,194,8,35,146,189,
14,254,242,22,211,229,10,141,185,10,142,61,
162,9,197,160,143,166,234,40,9,131,220,164,
193,246,194,101,246,149,36,99,215,222,103,123,
115,72,73,11,49,114,179,244,246,108,238,241,
18,243,124,130,130,71,155,18,82,171,78,189,
213,58,149,107,116,58,30,227,227,6,66,183,
143,252,222,45,242,151,231,81,172,211,15,37,
214,85,194,133,73,196,118,23,111,162,196,245,
15,202,76,46,252,230,81,103,235,147,138,116,
237,161,79,223,29,208,25,250,143,109,106,114,
252,8,219,63,190,23,105,227,44,80,24,166,
202,139,175,206,115,243,163,77,238,222,218,33,
89,6,215,11,168,87,59,167,20,99,83,51,
69,84,85,166,223,181,49,45,29,33,140,105,
30,244,24,14,28,194,32,66,148,4,20,69,
33,142,35,28,199,67,150,100,12,75,69,55,
85,116,223,199,139,109,60,47,68,18,69,18,
73,227,145,125,25,161,40,241,15,107,228,44,
123,150,151,27,103,83,158,224,101,127,133,183,
237,69,122,217,223,226,131,15,254,63,94,89,
76,48,54,97,178,185,254,232,186,64,24,70,
152,94,155,97,52,205,108,222,98,215,118,81,
252,50,201,169,165,207,94,7,136,199,241,163,
45,66,235,0,223,233,159,202,40,194,207,81,
181,54,95,180,176,180,79,223,67,143,3,135,
31,27,40,52,69,100,38,111,157,1,137,56,
142,216,219,250,125,196,243,107,180,1,55,22,
25,75,164,49,37,9,81,18,176,251,17,138,
42,162,25,2,87,50,9,90,158,207,15,14,
142,165,181,78,24,209,112,125,122,149,2,146,
32,48,46,128,186,85,69,24,58,68,99,5,
226,164,133,208,234,130,163,35,74,103,229,89,
81,57,143,212,27,18,39,12,16,68,194,217,
241,199,254,13,166,36,17,202,33,73,93,165,
160,171,244,131,16,69,20,120,163,152,193,47,
196,252,193,221,49,164,243,30,139,219,53,62,
174,221,34,59,249,143,248,97,125,155,47,190,
56,164,187,211,166,209,240,184,119,175,135,32,
140,20,75,189,94,48,2,138,222,97,90,111,
25,128,192,148,165,179,90,108,211,107,125,136,
217,158,195,72,23,41,95,254,111,185,190,119,
19,219,190,139,188,57,78,224,15,232,84,58,
100,26,29,144,68,20,64,81,68,210,121,139,
100,66,66,81,96,110,46,65,102,50,13,113,
204,78,144,37,159,175,34,137,49,243,9,3,
175,31,63,162,136,166,113,208,79,17,219,89,
164,32,164,234,122,164,204,58,170,226,80,15,
3,234,214,225,203,151,52,73,150,94,33,29,
136,35,67,65,223,39,219,239,33,185,30,93,
35,207,109,45,79,55,18,152,73,118,120,249,
131,239,208,27,72,228,22,207,17,245,91,8,
81,131,136,113,206,243,54,107,92,65,36,34,
68,98,158,235,103,132,14,115,115,199,105,108,
108,234,136,251,13,146,139,103,199,151,70,211,
21,182,10,89,110,111,187,228,139,111,34,41,
143,119,117,109,110,92,35,55,253,2,91,77,
155,106,115,8,198,167,91,133,55,251,30,15,
116,37,149,180,142,227,71,136,2,120,97,76,
103,56,122,97,84,77,225,202,75,51,236,45,
223,103,127,175,77,224,135,103,122,80,148,195,
76,226,250,181,13,226,56,198,20,6,12,162,
145,226,69,211,21,172,132,129,97,142,188,190,
136,5,6,3,155,221,237,6,97,24,145,16,
186,244,227,99,167,210,115,151,38,40,150,207,
90,236,139,113,252,51,217,112,119,134,169,10,
99,220,32,66,59,65,111,91,99,223,96,187,
119,133,237,141,255,147,197,228,15,217,136,158,
199,149,126,157,27,219,127,200,165,243,99,176,
190,252,104,208,233,7,36,87,119,120,81,173,
241,35,251,13,26,246,24,25,239,199,147,162,
230,39,190,204,238,218,14,210,220,54,253,157,
149,83,126,115,234,231,104,54,185,188,223,231,
194,120,2,253,17,53,174,206,208,167,222,114,
48,66,135,129,27,254,116,128,162,146,210,31,
153,198,28,108,124,23,166,214,16,4,40,5,
5,230,205,136,104,0,125,34,136,99,18,105,
9,215,137,233,119,66,172,148,196,158,125,86,
166,249,78,189,67,94,83,184,152,178,144,253,
0,113,125,143,218,118,143,194,210,16,97,166,
130,180,177,71,100,187,68,149,179,77,88,210,
234,14,226,230,62,225,115,75,68,249,244,83,
55,174,8,64,242,176,25,74,147,68,52,224,
215,207,215,249,227,149,10,87,133,109,196,206,
159,35,151,254,123,208,254,49,31,89,127,200,
75,19,13,50,247,182,88,94,238,225,56,33,
139,139,9,30,214,140,9,3,27,105,121,3,
206,205,240,114,62,201,119,179,13,170,91,255,
22,171,181,68,105,238,107,164,199,174,144,230,
10,118,235,30,141,141,93,222,174,62,143,158,
168,16,138,1,101,154,164,191,44,160,40,22,
110,50,100,127,16,162,10,30,63,136,52,44,
73,166,100,134,244,171,105,116,117,200,7,205,
46,147,113,130,189,158,205,98,210,64,220,175,
19,149,11,232,138,199,84,182,206,76,110,20,
224,106,142,199,95,174,23,233,12,67,4,33,
70,32,38,140,100,4,33,226,0,136,34,145,
124,170,138,22,150,89,211,11,244,213,16,195,
24,144,50,55,153,82,21,94,168,182,217,107,
120,196,165,75,44,92,177,185,219,184,140,251,
189,191,34,98,244,28,30,76,22,243,31,210,
87,63,176,109,215,77,249,72,235,40,116,7,
8,141,14,225,204,24,60,194,132,47,167,41,
244,183,52,74,87,102,142,44,196,31,229,249,
164,90,25,6,94,200,218,94,23,53,136,145,
159,98,164,68,163,239,147,181,56,50,106,91,
40,89,163,206,217,135,234,25,138,34,51,61,
87,100,250,82,142,152,152,208,143,240,253,16,
223,15,9,252,128,100,202,60,172,117,152,136,
146,64,70,87,209,178,101,18,41,253,177,198,
130,113,28,211,235,14,113,235,123,96,21,144,
85,25,93,87,49,76,245,177,60,154,64,204,
63,172,209,252,6,237,161,58,168,150,156,192,
153,249,207,8,87,254,103,92,247,251,168,133,
175,211,234,47,80,75,13,169,140,109,179,191,
103,63,178,14,216,233,248,156,59,151,228,82,
110,149,221,8,194,149,23,126,172,107,83,205,
12,185,236,27,212,107,255,6,65,60,253,44,
11,73,149,131,158,203,240,25,231,249,252,68,
238,89,24,113,115,167,199,249,74,130,196,67,
138,167,180,169,80,141,98,182,155,246,83,213,
149,62,19,80,152,143,73,103,220,112,157,148,
37,243,122,33,69,56,140,48,19,103,95,110,
77,23,136,35,129,235,235,33,91,3,13,178,
167,31,230,172,165,115,49,109,33,119,7,200,
31,221,37,244,66,6,131,128,250,219,59,196,
63,220,65,215,69,198,58,46,145,152,66,144,
116,112,61,144,37,196,157,26,226,126,131,168,
144,33,74,89,63,118,119,99,82,150,72,171,
50,65,46,205,165,253,101,174,239,223,36,93,
185,204,238,157,36,241,185,30,189,174,207,249,
243,41,90,45,143,219,183,71,195,127,38,38,
76,146,73,25,77,147,16,100,17,113,187,6,
97,68,250,194,28,149,49,168,58,49,132,199,
92,235,160,121,151,172,248,71,12,51,21,242,
179,255,241,81,112,56,216,93,225,192,13,240,
90,7,136,13,29,69,77,227,185,45,12,107,
156,253,206,26,31,234,55,208,18,163,251,86,
119,125,26,125,120,97,250,176,0,167,170,32,
112,100,249,253,96,149,116,149,223,56,63,100,
207,241,24,248,33,154,168,98,71,46,78,24,
33,31,26,244,191,191,83,68,245,67,28,9,
202,70,3,17,11,58,21,166,115,7,180,222,
93,67,184,244,26,147,217,62,59,217,203,232,
55,223,193,14,92,146,180,217,99,142,49,214,
70,1,22,143,30,57,44,218,216,201,5,10,
149,36,130,209,39,92,48,17,134,14,130,237,
18,77,20,137,13,141,253,48,34,35,70,232,
146,200,32,8,137,0,67,20,185,246,145,72,
169,248,143,176,123,123,212,119,255,29,169,212,
235,164,199,174,158,121,70,86,97,142,15,87,
26,236,109,53,89,152,214,158,234,185,58,126,
136,227,71,152,135,116,86,20,143,198,80,62,
201,246,89,64,64,86,36,100,69,226,97,44,
186,244,252,244,97,197,176,9,198,147,45,101,
4,65,32,149,182,16,228,52,241,83,24,88,
134,210,63,80,79,71,247,226,49,116,162,158,
28,99,95,159,32,231,222,161,23,191,73,106,
250,159,112,127,47,207,185,87,66,204,191,252,
240,148,125,14,64,16,196,4,65,64,163,225,
145,63,104,161,101,10,168,137,153,31,251,250,
82,165,139,184,235,111,162,29,90,197,236,221,
248,115,198,174,252,34,146,40,176,84,73,112,
123,167,247,68,191,188,159,214,26,25,186,202,
143,173,49,183,159,82,185,251,153,128,98,183,
229,176,56,118,250,197,172,111,252,128,176,184,
193,213,140,133,38,138,12,31,35,66,31,116,
35,124,45,98,83,110,67,74,64,234,141,17,
234,53,84,45,230,165,92,146,130,166,64,204,
136,98,10,35,182,182,134,180,90,199,153,135,
235,134,136,162,77,49,94,102,56,28,35,213,
120,200,139,165,222,70,184,118,23,20,153,112,
126,130,56,147,124,242,31,19,69,200,215,87,
8,175,46,17,63,100,211,16,17,18,148,178,
168,55,54,137,196,31,2,151,41,206,254,22,
127,182,249,127,240,229,82,134,123,55,246,41,
20,244,67,238,51,102,103,103,136,32,192,210,
82,10,253,242,56,209,88,17,250,67,4,215,
99,220,212,104,100,98,134,245,15,217,88,189,
73,28,123,232,233,30,87,199,53,34,99,155,
253,245,63,39,59,241,6,146,162,161,24,137,
67,149,207,197,179,193,49,63,205,224,230,58,
216,1,114,113,140,32,238,19,91,29,238,116,
70,214,2,179,185,39,52,166,9,2,19,134,
198,81,180,11,64,240,2,98,83,231,94,111,
72,33,189,75,163,81,38,99,52,105,120,25,
10,178,199,249,73,143,212,59,235,248,165,44,
251,227,51,100,198,246,104,213,12,212,157,125,
250,164,72,208,34,150,52,78,74,216,55,185,
64,81,218,231,23,23,219,200,98,131,122,144,
103,117,63,143,63,61,201,66,248,62,98,113,
68,254,52,59,131,163,76,174,227,7,52,92,
159,173,251,10,105,227,159,162,200,73,234,7,
223,97,97,161,73,187,155,34,244,93,124,100,
188,32,34,161,203,136,130,192,78,195,230,218,
181,13,146,207,56,194,180,235,248,228,78,220,
19,83,149,184,191,82,35,12,35,102,230,74,
143,111,40,10,189,145,76,241,111,104,41,129,
143,28,133,255,128,18,79,0,10,0,113,225,
191,163,191,241,47,113,15,190,133,149,251,93,
18,99,111,112,119,187,205,139,47,214,217,189,
182,121,6,44,0,182,182,6,164,82,13,44,
123,18,107,230,165,159,200,53,102,202,207,31,
217,197,100,167,95,36,10,125,68,73,65,147,
69,206,141,37,184,189,219,251,27,55,5,124,
208,35,55,147,55,145,78,196,183,32,28,201,
199,159,22,2,36,85,85,255,199,43,191,242,
95,129,219,67,112,123,160,167,71,167,163,7,
169,190,211,30,189,28,161,7,110,31,20,29,
175,215,32,12,125,50,233,52,110,175,129,221,
217,161,35,125,155,197,9,129,105,75,39,142,
193,119,99,84,77,160,227,7,232,135,60,93,
24,199,28,216,1,247,250,67,2,33,70,20,
33,84,123,228,73,241,98,206,36,119,104,85,
78,16,32,238,212,16,134,163,33,61,15,128,
226,65,146,48,53,101,98,37,100,236,166,75,
24,139,4,145,128,34,197,216,190,132,19,72,
8,126,128,236,185,8,205,238,72,110,249,24,
159,30,223,3,205,179,65,85,136,83,167,165,
122,93,63,230,163,154,194,203,238,54,90,127,
64,83,84,177,165,11,104,102,134,97,51,66,
41,55,8,238,237,80,127,168,129,174,92,54,
200,229,84,132,222,16,161,218,64,28,216,136,
205,46,73,83,167,148,23,208,205,152,165,25,
135,249,201,144,165,178,66,88,235,49,182,126,
159,252,255,207,222,123,61,89,150,93,103,126,
191,227,207,245,222,165,173,44,95,213,93,213,
221,213,232,110,144,96,3,36,28,57,32,9,
146,26,50,40,205,80,15,10,41,66,17,154,
55,61,233,47,208,147,30,21,19,161,7,106,
66,65,13,53,35,137,26,90,112,8,16,32,
8,194,52,26,237,77,185,204,50,233,243,222,
188,254,30,111,245,112,50,179,50,43,77,101,
117,87,51,240,128,29,81,81,105,78,158,187,
207,62,123,47,247,173,245,173,201,71,152,221,
31,179,237,104,72,114,145,40,240,16,21,157,
208,119,241,172,49,178,150,198,51,71,68,129,
71,101,250,53,130,145,143,29,44,33,228,13,
4,41,38,136,99,70,94,192,185,220,201,241,
23,113,179,139,252,206,45,164,213,54,210,131,
13,98,77,37,46,230,40,169,10,235,150,203,
216,76,145,75,217,116,38,83,20,115,155,44,
72,25,140,247,86,137,158,191,65,55,107,241,
112,144,227,181,201,187,116,23,183,233,198,45,
114,113,23,165,212,64,181,219,136,162,64,185,
172,81,14,87,144,130,9,237,109,7,194,144,
102,49,160,226,109,81,110,197,220,209,95,228,
221,46,220,236,106,12,188,152,130,238,177,49,
118,185,219,183,233,222,174,83,171,252,23,168,
122,5,207,28,97,218,239,35,56,160,20,95,
231,238,150,195,63,126,255,93,62,92,236,242,
112,109,200,234,218,6,139,107,6,163,78,151,
98,78,36,95,208,147,61,26,71,32,201,73,
62,185,40,39,223,187,35,80,244,164,126,192,
183,136,194,144,162,46,18,135,1,146,162,162,
251,35,204,80,228,238,205,53,180,216,36,83,
44,64,232,130,213,131,84,17,220,9,132,62,
40,169,228,190,130,152,108,70,103,8,178,154,
252,206,234,66,170,148,156,145,208,77,82,25,
237,33,16,39,181,5,206,0,36,5,66,31,
193,236,64,186,156,164,77,250,118,114,173,51,
58,52,119,217,183,8,125,11,55,91,221,155,
59,178,190,83,219,16,38,207,231,12,30,157,
85,179,147,156,223,192,73,82,96,21,253,209,
220,101,245,209,220,35,31,220,113,146,158,26,
250,8,59,231,250,208,220,5,146,239,3,103,
111,238,184,227,100,29,30,159,187,239,36,63,
119,71,199,174,251,222,220,67,47,153,135,111,
28,148,51,178,150,124,150,111,38,215,186,19,
4,119,12,122,145,116,56,38,173,201,8,130,
136,59,238,33,169,58,145,239,226,91,99,148,
116,129,64,62,143,213,126,7,189,242,34,158,
53,70,203,95,102,195,88,101,78,234,48,234,
29,77,127,55,53,147,101,85,174,34,164,95,
70,148,100,220,113,23,81,86,136,163,16,119,
210,71,214,210,4,174,133,59,233,161,229,202,
184,198,144,56,12,16,101,5,119,220,69,144,
100,204,254,10,157,149,111,49,158,124,128,209,
91,34,176,60,210,165,105,162,192,39,10,3,
36,69,195,25,119,81,100,137,82,90,101,60,
216,38,20,21,8,92,4,103,0,122,17,193,
51,146,148,225,253,235,206,206,254,146,148,228,
93,250,22,104,217,228,235,221,107,119,247,204,
238,254,18,229,228,94,145,191,183,238,66,96,
19,10,42,57,209,66,136,66,220,88,228,238,
253,21,86,6,30,158,189,243,185,74,234,208,
186,239,223,51,239,254,245,31,239,83,20,178,
134,16,122,201,230,81,82,201,203,22,196,228,
107,65,76,190,223,15,44,202,41,234,197,52,
178,150,166,179,241,45,106,103,187,188,84,78,
172,119,99,28,98,107,33,31,14,13,22,39,
54,171,166,203,125,195,230,206,196,98,195,118,
137,67,137,151,106,25,158,43,102,153,77,235,
156,45,201,224,199,120,110,140,162,10,32,137,
8,182,139,56,182,72,237,244,110,112,221,144,
185,185,12,154,38,129,32,224,107,89,10,87,
170,40,149,44,82,70,99,98,36,192,95,86,
13,8,98,1,219,151,136,131,8,89,19,137,
11,71,135,4,60,55,70,206,106,68,217,116,
210,108,105,231,159,233,73,252,237,98,141,95,
61,219,33,31,69,72,189,33,213,112,194,170,
177,142,152,123,145,124,253,60,107,171,139,156,
21,215,153,12,14,42,138,98,81,37,155,149,
147,248,114,24,34,184,94,242,44,237,30,41,
203,162,146,23,17,135,46,130,146,38,112,1,
77,35,213,73,232,77,194,205,109,228,162,137,
159,126,137,76,101,26,65,20,17,101,5,89,
75,239,104,117,125,47,3,72,75,215,24,220,
186,79,228,187,72,229,128,112,34,32,106,144,
85,100,82,146,152,112,4,237,104,86,177,211,
7,207,135,148,134,216,233,19,53,171,68,205,
10,113,41,79,52,219,0,96,203,113,89,177,
28,22,55,95,164,169,133,76,134,2,217,245,
2,74,101,68,102,123,68,83,238,243,79,209,
69,20,41,224,57,186,232,158,195,182,95,69,
117,187,72,68,232,185,44,134,29,82,174,167,
104,77,231,104,94,174,80,208,19,78,170,204,
217,42,225,11,151,136,243,25,170,217,9,149,
180,197,134,223,102,52,81,217,190,85,196,25,
189,64,42,248,28,173,179,223,64,209,11,72,
138,70,28,6,120,86,159,144,41,198,242,2,
125,51,32,83,42,146,202,166,136,4,9,219,
151,16,5,145,92,49,79,115,174,137,20,187,
137,160,150,148,228,144,41,169,228,0,137,82,
242,53,66,242,59,89,39,52,186,212,106,101,
36,81,36,10,124,100,69,65,87,69,182,251,
38,182,19,82,169,236,120,40,145,255,232,28,
8,98,146,251,46,202,137,240,143,66,16,149,
157,130,169,104,231,124,164,146,195,181,155,239,
174,232,201,103,10,194,129,115,36,196,97,114,
206,36,229,209,181,178,126,104,238,121,99,128,
134,128,145,41,239,205,61,185,86,59,120,223,
221,57,18,131,146,78,126,167,236,187,118,167,
253,238,222,181,81,144,220,75,214,14,158,235,
199,231,46,202,137,0,73,149,247,201,128,29,
67,228,241,185,199,81,242,247,106,230,216,117,
223,155,207,238,92,245,194,99,114,230,240,181,
66,232,131,154,161,144,203,83,200,166,16,4,
17,89,207,236,48,248,62,58,27,138,158,69,
212,23,208,114,101,100,45,157,144,70,234,231,
80,203,171,228,70,219,152,59,69,174,62,26,
34,33,2,208,152,207,179,218,44,225,27,243,
104,217,106,114,95,81,66,16,165,132,7,78,
16,16,101,149,56,10,145,245,12,178,170,35,
41,42,130,32,32,235,25,54,111,125,155,254,
240,123,136,65,22,100,135,104,234,62,174,124,
159,201,218,50,138,82,33,93,104,237,60,70,
26,81,82,80,21,153,90,185,72,20,11,88,
126,66,31,130,146,218,41,166,123,108,221,247,
203,222,40,72,148,132,40,31,188,86,126,108,
127,137,18,66,224,64,186,186,111,221,53,46,
183,178,100,51,25,36,69,69,18,19,62,183,
98,70,167,55,50,64,47,38,239,234,241,119,
180,111,207,188,251,159,254,215,79,22,122,18,
128,70,33,185,145,107,246,9,178,15,104,232,
201,247,65,20,115,119,146,98,219,220,36,138,
65,147,4,206,230,82,20,149,36,143,87,17,
4,98,91,32,163,39,150,196,46,163,166,158,
78,190,31,15,66,114,69,137,232,204,20,226,
208,64,232,143,152,155,75,51,61,157,218,75,
173,53,60,153,76,73,36,216,151,213,148,77,
109,96,223,110,239,0,210,17,154,20,17,68,
2,214,157,109,162,92,153,116,73,61,42,234,
132,99,29,14,145,189,181,94,230,27,250,45,
196,142,132,115,103,21,203,132,201,112,194,133,
218,109,110,13,31,80,76,93,7,33,71,225,
229,5,182,30,190,191,87,8,55,55,151,161,
86,59,62,86,46,244,71,72,253,17,217,92,
154,96,182,4,146,136,57,142,16,167,171,156,
13,124,220,108,150,31,184,151,73,41,79,142,
183,203,90,154,234,217,151,25,40,127,149,120,
107,15,106,104,207,13,40,108,118,81,162,136,
120,170,70,44,137,200,239,222,1,73,36,60,
151,100,24,69,83,9,62,0,7,11,16,123,
166,143,119,239,50,151,35,135,112,115,158,243,
138,78,253,197,95,161,187,244,111,185,112,165,
194,228,246,58,207,157,159,112,175,159,199,28,
88,172,173,89,180,230,83,56,78,140,225,89,
108,74,243,92,168,91,180,59,62,205,34,216,
207,205,114,239,98,132,236,202,88,145,202,210,
71,1,41,197,193,169,218,152,27,105,240,103,
169,196,77,234,151,190,126,108,250,107,125,238,
215,17,68,153,161,43,128,144,100,19,105,186,
2,71,209,225,216,79,177,129,163,128,149,81,
204,66,85,69,145,68,38,78,72,181,164,81,
105,86,88,91,233,61,58,136,251,15,229,137,
168,161,9,158,145,176,21,232,57,158,21,161,
155,173,165,126,145,245,180,27,42,249,4,33,
27,37,85,225,97,187,200,217,215,190,68,249,
159,190,199,182,161,49,241,50,148,73,194,213,
247,63,238,115,173,94,230,199,230,34,57,142,
79,143,29,108,253,12,191,183,65,228,251,200,
113,157,153,203,255,18,128,233,231,191,65,224,
126,9,115,176,74,161,121,153,222,242,155,184,
254,6,126,52,100,59,252,83,6,139,243,164,
213,243,148,103,94,69,148,148,189,168,72,46,
37,35,162,179,233,140,255,89,214,46,189,175,
214,71,20,4,82,138,196,122,223,121,170,123,
60,181,162,144,68,129,133,70,134,194,78,166,
70,232,155,196,118,154,110,63,196,28,123,108,
89,1,177,110,112,57,159,102,211,241,120,165,
146,79,192,210,253,56,197,9,93,187,138,102,
151,209,72,39,235,142,17,250,163,125,49,50,
137,48,151,101,178,237,82,84,93,162,194,99,
21,141,178,204,118,199,193,207,70,20,242,10,
166,21,144,203,42,228,112,8,239,44,50,62,
123,30,52,153,116,78,66,20,119,51,114,56,
18,112,79,121,22,233,7,139,72,159,127,14,
113,161,76,247,135,247,25,116,93,202,34,132,
181,45,224,58,233,226,171,252,212,94,226,250,
43,51,60,120,99,53,17,182,61,151,106,85,
59,136,163,139,194,14,31,248,163,141,46,216,
30,210,131,117,172,82,149,84,65,39,156,111,
177,157,206,243,198,59,53,102,95,252,67,156,
241,233,26,138,20,91,47,98,124,252,144,104,
212,167,84,189,70,113,242,247,228,124,131,224,
250,249,228,225,252,128,224,198,229,3,153,69,
187,74,98,47,134,233,250,116,93,143,123,15,
69,22,206,254,254,158,208,182,250,155,8,162,
68,172,93,227,157,108,159,95,105,164,121,123,
168,49,95,91,198,248,96,64,161,85,39,86,
28,38,161,128,92,157,227,87,74,203,232,41,
133,116,219,98,113,57,230,210,185,62,149,86,
141,119,151,82,100,226,27,148,90,47,51,233,
45,49,122,255,39,204,63,255,71,72,106,234,
20,202,48,9,7,166,249,4,113,250,147,240,
4,65,100,100,135,188,183,106,48,85,212,73,
171,18,78,0,23,206,84,9,34,30,133,67,
118,61,137,39,90,78,2,168,89,208,243,73,
104,6,64,205,125,234,132,138,172,99,162,4,
62,147,66,243,179,151,38,187,161,34,189,244,
115,169,40,236,167,200,26,242,194,8,85,18,
137,2,143,170,212,195,126,103,133,240,165,95,
199,235,43,156,125,240,215,12,119,50,147,39,
19,159,212,91,43,168,151,110,225,140,111,160,
231,15,175,243,112,243,3,196,169,183,184,80,
87,232,185,62,131,91,133,67,123,180,208,188,
188,3,104,95,69,16,175,33,41,42,91,247,
191,133,155,190,131,89,93,194,184,255,33,57,
253,37,42,179,175,209,55,124,30,116,77,226,
253,133,148,167,165,22,255,132,20,228,119,182,
38,73,203,135,157,172,177,238,196,123,234,118,
172,242,94,198,6,36,49,175,19,138,124,138,
41,133,102,94,71,244,109,156,177,133,164,168,
136,130,142,98,207,209,150,62,132,24,10,5,
153,43,249,12,247,38,246,1,37,225,71,49,
146,32,36,112,193,19,158,181,124,255,46,166,
39,35,202,2,178,24,19,188,116,153,40,155,
102,178,97,81,236,222,74,44,97,47,72,92,
130,29,169,31,181,170,228,11,43,60,92,236,
33,73,2,113,12,151,174,228,209,53,9,201,
178,40,46,221,33,58,55,131,105,40,4,217,
52,97,144,100,96,29,202,138,137,98,84,89,
68,203,233,4,106,226,214,181,90,41,6,3,
15,207,246,137,253,69,156,209,75,200,114,150,
222,202,117,238,45,40,180,230,71,180,151,199,
152,102,192,202,138,201,252,252,35,188,35,248,
252,117,98,73,68,220,234,33,110,15,32,138,
146,186,144,135,109,194,77,27,237,245,139,76,
124,120,115,93,166,182,240,219,120,230,8,207,
24,156,186,91,95,54,119,17,189,80,35,142,
99,28,75,34,154,107,48,186,99,210,169,92,
220,227,45,145,196,8,85,10,152,41,245,16,
108,7,233,206,114,66,61,226,122,20,206,206,
242,147,117,29,121,188,128,159,31,239,177,88,
237,206,33,91,184,74,251,206,67,182,47,250,
228,6,30,231,163,144,82,65,102,89,62,67,
211,187,71,238,124,138,137,232,96,201,77,244,
104,155,225,208,199,113,66,134,75,3,6,98,
133,188,240,58,169,194,52,190,53,198,24,220,
70,84,12,140,206,50,90,238,100,134,85,215,
24,28,240,252,176,159,96,125,57,163,211,239,
122,119,178,183,231,45,65,33,147,213,24,142,
66,26,89,133,184,41,178,109,244,159,238,158,
190,149,196,116,15,128,92,107,59,161,160,35,
50,240,158,112,206,246,110,33,68,136,178,240,
232,124,62,49,165,235,180,243,221,113,191,66,
247,128,87,164,109,255,13,110,227,155,143,66,
16,79,115,207,93,124,227,84,92,79,30,196,
225,193,207,63,233,153,68,137,137,45,96,234,
54,39,149,37,120,198,128,72,144,184,181,49,
161,148,81,104,21,83,24,118,129,203,153,128,
202,246,15,216,88,119,24,78,14,190,167,206,
134,201,141,230,199,252,192,248,43,234,241,239,
30,186,167,49,250,128,95,189,150,37,45,73,
108,76,92,76,161,117,108,165,118,224,90,8,
162,140,164,168,20,74,55,216,94,155,32,26,
45,4,84,28,163,77,239,193,59,152,206,135,
52,195,12,130,32,19,250,58,122,249,121,252,
72,196,11,66,188,32,98,60,26,32,106,121,
162,253,254,190,107,236,120,183,234,169,215,11,
64,17,5,234,121,157,140,34,17,26,14,206,
206,54,76,199,144,9,77,76,103,180,19,66,
149,78,169,40,118,249,70,162,240,72,238,145,
140,38,49,87,73,239,177,15,250,182,129,40,
201,123,21,136,234,160,138,27,39,10,192,242,
35,62,24,24,156,203,165,176,130,144,143,71,
38,86,16,98,135,17,21,77,225,243,213,194,
201,81,129,153,6,98,103,64,102,56,161,235,
234,200,18,32,101,192,129,162,230,225,184,33,
158,17,81,18,6,196,43,105,162,51,73,28,
80,24,155,104,105,133,217,43,21,68,85,33,
55,95,66,104,85,8,68,17,97,100,32,45,
173,33,125,124,143,60,16,53,43,140,155,211,
164,50,135,195,10,18,50,133,122,150,181,230,
235,108,15,75,148,114,117,206,100,123,76,77,
165,144,101,145,204,220,128,59,189,53,138,83,
47,81,154,254,28,221,237,12,141,207,91,228,
7,239,50,30,251,116,187,46,233,180,68,173,
182,239,176,41,50,209,108,131,104,182,65,224,
199,184,155,35,100,219,38,125,33,177,96,254,
105,209,1,115,158,212,116,19,65,148,136,194,
0,189,112,186,102,53,251,175,237,14,115,8,
155,107,84,198,6,21,115,45,233,225,219,40,
19,151,242,12,237,52,239,109,149,121,169,220,
78,50,202,98,24,214,26,252,211,237,121,166,
207,254,33,242,76,230,216,251,206,231,255,107,
222,190,247,127,80,159,243,65,47,98,190,120,
3,117,213,65,149,114,140,46,76,227,123,62,
214,79,92,226,124,154,250,220,50,185,153,2,
94,219,34,88,241,41,93,78,114,212,237,225,
54,81,227,67,24,100,201,54,22,158,88,109,
13,160,23,106,216,163,54,153,82,3,6,167,
80,156,167,229,90,114,39,123,215,102,138,41,
10,69,141,113,223,38,85,72,115,38,15,221,
135,3,162,40,78,236,153,83,113,241,236,36,
123,236,191,118,247,107,103,156,40,10,45,183,
127,113,79,117,223,194,100,128,18,132,152,167,
230,144,58,237,124,143,224,122,210,75,120,226,
111,66,110,234,147,221,83,144,63,19,174,167,
221,181,138,128,237,80,227,76,41,125,226,89,
80,11,53,52,67,167,239,135,24,99,145,185,
169,255,158,59,169,255,204,124,244,38,25,251,
1,217,178,142,59,113,217,218,122,20,122,89,
125,191,205,213,215,151,184,211,255,30,41,229,
2,97,96,225,135,19,192,97,118,118,147,180,
164,226,69,49,111,189,41,209,156,187,118,160,
160,238,160,254,61,40,23,231,202,255,213,129,
62,235,158,49,192,90,189,69,101,118,147,122,
211,164,55,9,232,220,237,51,115,37,185,46,
244,28,6,41,168,180,230,49,189,0,195,9,
152,56,1,195,190,156,100,22,158,38,12,186,
111,111,93,152,206,145,61,34,53,182,61,114,
49,37,25,244,16,180,252,169,222,217,35,48,
27,16,124,235,192,203,147,68,129,133,106,134,
51,213,244,129,98,151,40,240,18,160,117,199,
234,213,212,50,195,254,59,136,153,128,40,78,
50,112,94,44,229,120,171,63,97,232,5,4,
59,97,23,59,140,40,40,50,90,36,162,106,
194,137,46,86,92,43,161,61,55,67,152,201,
146,174,233,168,154,64,156,78,33,24,22,238,
192,66,191,60,67,52,223,218,203,104,18,4,
1,167,94,35,119,125,22,237,92,19,42,5,
208,84,80,21,80,21,196,229,77,28,55,196,
117,34,52,207,37,92,235,161,250,110,82,65,
189,47,52,35,9,49,99,39,141,166,134,156,
169,116,232,58,69,178,231,243,228,125,147,84,
81,39,155,18,217,182,214,9,148,203,196,65,
68,113,234,26,75,15,23,121,101,222,166,187,
50,34,142,147,84,217,106,85,123,20,150,16,
4,130,246,24,231,94,151,104,105,131,84,191,
143,46,5,132,192,27,91,34,178,243,27,52,
22,190,190,87,84,230,219,6,74,42,119,74,
227,240,209,181,238,232,46,103,154,6,113,54,
77,92,204,33,14,198,160,200,196,153,20,186,
22,18,68,49,255,184,153,161,54,45,114,71,
170,112,115,244,58,173,75,127,128,120,132,165,
178,255,190,130,32,66,156,199,137,111,178,18,
133,196,90,154,172,105,82,144,76,62,72,167,
184,239,120,92,136,92,236,108,133,247,138,175,
225,213,234,204,132,91,200,163,49,155,46,168,
197,75,64,140,223,85,200,103,94,37,85,104,
156,66,150,88,244,183,126,194,36,254,75,236,
77,135,145,56,253,132,63,176,31,1,173,79,
146,105,206,40,1,241,128,156,46,35,137,2,
67,203,167,146,85,241,205,33,109,11,198,67,
27,77,14,235,207,60,8,0,0,32,0,73,
68,65,84,78,119,207,112,199,66,62,234,218,
93,50,55,119,156,0,195,178,126,232,156,29,
251,110,119,112,138,64,61,101,250,239,105,215,
96,215,251,217,31,154,19,4,208,10,159,252,
158,129,187,3,172,138,167,19,102,113,116,170,
84,227,253,107,101,186,33,41,85,34,165,74,
199,158,5,45,157,163,148,81,48,220,16,219,
11,233,155,62,129,126,158,56,255,18,29,69,
194,45,232,204,183,68,244,192,197,243,34,130,
32,169,158,207,142,71,132,175,10,244,164,91,
108,10,62,229,218,42,55,114,107,212,202,58,
182,25,97,91,2,157,209,21,74,173,23,78,
128,190,30,201,197,246,226,15,104,223,254,7,
242,245,11,136,178,74,232,217,244,87,223,34,
147,63,143,109,230,216,94,175,96,247,155,104,
122,3,73,201,49,92,255,128,141,143,191,69,
109,225,101,20,45,133,42,139,164,229,152,106,
62,133,134,207,208,141,143,22,232,187,225,168,
29,25,187,187,94,170,40,48,87,61,188,111,
12,39,224,94,199,124,244,110,247,146,11,142,
31,79,4,179,47,54,179,167,226,48,119,140,
46,66,198,221,67,186,99,96,18,4,140,31,
235,207,156,150,37,202,154,140,235,30,4,166,
162,88,224,97,175,78,37,59,161,160,91,68,
83,143,172,233,40,47,29,136,247,171,138,72,
121,38,79,164,171,236,247,67,99,93,5,255,
104,236,35,22,69,44,73,227,238,237,45,102,
167,147,212,90,226,24,113,99,27,113,179,75,
52,85,35,188,52,15,65,8,162,192,124,249,
145,107,233,135,18,55,59,179,44,92,85,40,
220,187,133,125,175,199,153,112,192,71,193,127,
66,203,255,22,32,144,205,127,145,187,90,7,
53,223,101,52,112,241,144,137,162,24,81,20,
48,111,174,179,214,89,197,54,28,166,27,10,
141,134,78,187,216,98,217,203,208,91,89,64,
207,190,70,177,117,249,153,196,113,69,229,2,
78,97,139,204,219,31,17,188,112,158,224,198,
101,4,195,66,90,222,36,214,85,102,166,235,
44,77,34,254,234,97,139,89,229,53,154,231,
79,159,63,174,167,171,220,89,23,73,207,141,
184,55,30,176,37,159,161,56,218,192,141,34,
162,16,180,133,2,41,107,128,43,77,184,213,
190,206,149,151,46,83,252,254,251,204,140,255,
129,141,209,117,210,133,121,106,243,95,36,176,
79,207,205,31,197,235,156,111,250,60,48,141,
253,103,225,25,140,71,70,202,198,208,97,99,
232,144,223,73,205,86,179,69,220,213,14,150,
233,144,215,120,118,67,203,63,242,102,60,243,
84,236,177,37,115,128,18,248,172,102,171,191,
64,179,119,198,253,142,133,54,37,30,91,68,
6,160,202,34,87,91,57,218,19,151,181,190,
205,208,242,25,219,26,83,197,223,39,157,87,
248,241,230,223,147,153,125,159,198,194,144,243,
238,22,155,75,3,198,99,159,87,13,135,232,
220,20,15,222,28,49,92,52,217,94,123,151,
234,124,153,210,23,206,51,73,37,180,65,187,
77,143,118,255,63,110,52,46,124,145,198,133,
47,238,179,200,83,52,46,254,218,1,239,199,
51,6,123,94,187,158,125,141,234,194,107,143,
233,83,159,192,49,201,167,20,170,129,196,182,
117,2,70,33,8,168,178,72,176,35,127,189,
40,230,94,199,228,108,237,32,213,210,214,200,
253,68,181,254,39,106,129,163,168,195,143,26,
147,225,10,70,230,57,162,137,136,162,246,200,
228,55,120,163,59,62,116,52,27,186,138,42,
138,184,59,224,100,20,11,44,118,90,200,98,
200,153,234,54,55,183,234,204,149,61,10,90,
162,96,2,63,70,81,14,46,140,127,253,2,
66,16,16,11,226,233,115,75,20,25,165,156,
225,249,87,27,224,4,76,92,17,85,138,30,
105,100,207,135,48,66,249,201,7,68,211,117,
194,179,137,245,122,167,211,98,170,48,32,167,
219,124,252,176,197,139,219,3,198,253,144,241,
182,193,220,111,244,88,29,244,73,151,91,228,
235,151,89,188,93,230,87,47,228,176,223,156,
96,24,240,254,251,54,30,10,81,4,41,201,
103,118,54,3,185,52,107,149,26,239,13,190,
76,237,236,151,168,107,233,103,122,136,114,141,
27,220,91,254,71,174,11,236,121,73,113,54,
77,152,73,35,216,14,19,63,100,60,50,153,
18,62,71,105,250,233,138,140,140,193,67,92,
85,194,187,223,98,90,93,64,144,63,198,175,
84,168,168,10,67,47,96,81,2,49,151,5,
203,129,252,93,222,94,158,227,149,231,231,169,
191,179,204,120,237,143,241,212,255,17,81,205,
158,30,188,28,175,67,48,36,22,4,162,56,
164,146,85,247,138,135,62,139,49,182,3,188,
32,137,12,143,71,118,146,93,133,255,236,63,
72,203,37,158,197,30,232,157,61,214,162,27,
100,74,191,160,240,56,100,60,196,220,221,50,
121,110,58,119,178,124,18,146,110,155,165,180,
194,74,207,162,111,250,172,13,108,186,166,199,
108,249,203,248,254,13,54,21,153,7,237,255,
151,234,243,155,168,82,204,27,195,41,162,119,
11,212,212,187,164,219,183,89,117,90,140,238,
116,88,152,188,203,199,243,191,76,182,116,9,
103,188,77,224,141,136,195,8,45,91,71,205,
20,142,164,151,57,104,244,159,172,84,142,21,
206,90,26,180,52,158,49,164,81,208,233,218,
71,8,249,125,247,45,103,84,202,153,28,66,
54,199,216,9,232,77,60,238,110,25,92,104,
62,98,248,254,164,249,21,242,73,232,122,111,
236,50,85,78,61,17,137,15,164,7,20,26,
9,133,131,105,231,24,111,63,79,57,215,71,
244,33,155,54,168,165,52,154,169,16,85,140,
177,140,136,204,142,151,240,193,250,60,215,167,
151,17,133,24,215,137,56,163,111,240,183,55,
155,124,237,236,22,154,148,40,138,124,73,58,
188,46,138,124,72,73,120,78,124,114,56,235,
218,57,132,149,45,38,31,109,81,212,15,10,
0,193,245,144,63,90,2,63,72,40,65,118,
198,250,176,194,165,250,38,0,87,59,255,200,
199,193,21,102,164,247,217,52,3,102,110,223,
100,49,147,193,53,91,73,217,190,150,67,173,
87,224,205,13,98,4,156,72,65,193,167,144,
145,168,159,41,51,124,241,60,31,173,21,16,
140,179,76,93,249,23,199,78,115,123,226,50,
28,216,248,246,4,4,40,164,20,166,138,250,
169,94,176,40,201,12,173,50,225,92,19,12,
59,161,28,143,99,164,15,151,136,166,107,44,
249,57,10,206,55,41,205,191,240,212,27,165,
52,253,2,115,171,34,197,51,11,40,169,44,
131,135,29,226,86,132,29,152,123,201,10,254,
78,42,167,170,59,108,154,54,237,168,74,51,
189,194,69,123,133,187,75,255,11,238,249,255,
233,212,159,103,77,150,209,46,247,185,187,45,
80,202,93,70,43,166,24,90,62,65,24,127,
134,66,8,198,189,54,150,225,48,61,87,5,
255,17,113,159,239,5,184,110,64,38,167,29,
162,114,255,212,30,70,28,37,223,63,246,146,
247,60,138,76,229,23,26,98,191,177,24,70,
220,217,50,184,58,149,59,80,109,124,156,177,
123,190,145,101,96,249,44,119,45,28,47,100,
113,203,64,113,13,106,141,22,181,153,255,22,
87,17,176,60,135,244,78,38,222,96,120,155,
242,75,1,181,183,239,33,153,54,147,129,196,
149,226,29,220,209,255,140,86,73,19,247,199,
104,5,149,129,90,103,113,253,26,211,151,126,
239,100,255,245,83,100,191,5,158,77,28,133,
104,170,72,41,171,210,55,188,99,179,159,182,
70,14,35,207,100,62,93,163,85,208,105,21,
116,44,47,36,8,99,212,157,254,57,245,188,
70,207,120,122,131,235,104,140,98,103,18,110,
16,81,206,168,136,68,4,182,177,7,64,238,
143,197,245,215,222,198,171,255,148,221,94,163,
170,226,129,98,49,28,23,73,213,150,121,161,
161,49,87,20,72,165,4,84,77,64,81,5,
4,1,182,141,60,197,148,69,70,117,49,199,
17,178,146,216,78,173,156,205,123,195,18,151,
26,46,90,234,244,204,139,182,25,145,202,28,
127,189,48,24,227,190,191,74,86,77,122,82,
120,126,132,40,38,115,17,92,31,193,118,247,
52,81,52,93,163,103,230,104,230,70,172,14,
170,84,179,19,228,193,144,213,113,133,206,80,
197,183,29,198,109,147,27,151,35,86,189,123,
132,193,12,161,227,209,172,174,225,220,237,32,
70,1,181,178,204,244,66,142,248,213,235,124,
24,95,165,211,187,65,227,204,31,144,175,94,
57,241,57,110,110,76,112,76,3,79,76,225,
5,17,19,39,64,145,196,67,164,94,71,97,
9,0,230,36,164,90,90,33,181,188,134,184,
218,70,8,66,4,73,228,102,189,202,210,50,
52,167,126,7,225,20,89,14,71,225,36,153,
66,3,73,217,77,139,150,81,184,203,195,40,
33,107,113,163,8,63,138,247,104,149,77,71,
196,19,65,174,84,168,198,19,74,178,197,162,
51,133,154,170,30,160,97,62,86,233,9,25,
250,247,215,16,106,19,134,111,247,40,181,206,
81,202,231,217,232,91,72,71,165,190,60,53,
70,113,48,30,95,205,169,212,114,26,43,195,
136,141,205,17,149,122,158,216,183,88,91,183,
120,176,180,197,195,165,54,237,141,1,190,23,
162,105,10,81,20,33,239,178,114,158,132,81,
60,33,238,158,96,24,122,82,135,177,23,231,
23,254,249,49,138,79,123,207,127,6,140,226,
192,199,133,49,134,19,80,201,170,123,130,248,
36,108,47,165,72,212,242,26,49,49,166,19,
18,250,22,147,80,161,51,118,25,59,1,166,
159,116,117,67,128,76,174,193,208,43,34,149,
77,138,5,25,16,232,173,14,177,218,99,156,
149,45,172,237,49,86,207,97,126,42,205,150,
40,146,46,188,118,36,118,123,178,125,29,17,
122,78,82,216,119,156,151,219,223,198,54,13,
52,77,67,148,100,82,154,198,246,196,61,209,
45,8,29,147,174,43,51,180,124,36,81,32,
167,203,200,251,206,139,38,139,20,82,10,249,
148,194,96,52,62,53,70,113,34,152,29,70,
49,157,177,199,200,9,233,90,176,57,74,226,
185,189,161,65,24,11,228,51,58,74,170,200,
112,253,54,162,146,69,246,75,68,130,131,172,
217,212,149,52,183,183,158,167,109,7,44,13,
52,4,33,164,150,138,119,0,21,157,129,149,
161,150,234,99,140,35,178,121,9,73,78,148,
72,104,135,132,98,26,15,155,130,42,224,57,
17,166,21,163,42,194,177,235,19,248,49,227,
97,66,66,120,228,53,158,143,242,238,29,92,
23,116,57,177,122,71,35,31,199,14,73,165,
165,131,10,90,85,136,166,235,116,141,60,173,
194,144,200,143,136,35,129,212,250,26,241,230,
50,12,214,24,11,117,188,80,196,107,247,121,
49,103,98,202,43,24,222,2,65,180,78,179,
40,97,204,157,163,95,63,199,82,234,6,166,
250,77,10,83,95,65,150,179,79,4,168,131,
48,102,115,232,28,122,15,166,23,80,207,107,
135,26,68,29,117,56,244,92,139,173,149,159,
50,103,172,33,132,17,194,196,68,112,125,114,
11,83,220,125,168,81,174,255,242,83,131,228,
71,90,106,153,42,189,246,79,185,94,211,232,
217,62,158,16,37,217,202,118,147,200,139,25,
27,37,100,197,163,23,135,204,212,210,232,131,
1,15,168,163,166,207,158,74,81,68,190,135,
24,103,9,55,85,40,76,232,173,252,12,85,
80,216,24,165,216,222,54,200,63,206,239,244,
84,138,98,152,28,142,208,165,145,1,21,159,
166,106,99,187,1,111,190,123,159,82,81,67,
207,100,121,112,243,1,235,155,22,254,190,158,
202,198,196,97,107,99,64,24,132,84,106,249,
79,175,40,30,7,189,247,232,28,84,106,195,
77,114,182,193,56,87,253,133,162,56,234,88,
7,17,182,23,82,206,170,8,167,216,179,162,
32,36,66,82,87,24,142,198,68,59,116,195,
94,16,97,121,33,134,27,210,51,60,6,166,
71,38,63,69,182,241,69,172,210,23,233,136,
5,212,166,76,58,167,81,201,128,27,8,164,
228,128,230,115,53,238,26,53,178,197,151,63,
19,69,161,165,50,40,170,130,111,142,240,237,
17,178,24,146,201,228,144,37,241,88,54,218,
221,245,242,195,152,129,233,179,53,114,17,68,
14,96,205,170,44,50,177,3,70,227,241,179,
1,179,119,99,130,143,55,119,9,130,136,141,
161,67,181,20,162,171,105,2,171,134,97,198,
196,81,132,47,77,17,72,49,45,213,226,143,
174,61,224,7,27,10,191,215,234,115,231,110,
17,202,201,195,173,12,171,92,172,172,226,123,
49,133,242,65,235,54,87,148,56,51,233,241,
157,141,50,249,214,128,7,150,71,51,47,34,
153,50,226,200,64,112,28,194,70,117,79,176,
199,113,66,143,83,174,73,140,250,1,197,202,
225,71,146,238,44,67,20,17,33,237,11,37,
68,164,210,210,161,184,93,156,209,17,55,187,
20,138,105,134,86,134,169,165,55,89,202,188,
72,102,96,208,238,120,224,134,228,89,102,131,
115,20,173,30,66,182,206,172,215,165,43,188,
69,207,249,53,122,233,52,197,214,53,52,73,
166,254,196,205,146,172,239,174,251,124,92,245,
105,16,198,44,181,77,46,54,179,79,12,65,
9,130,136,21,205,17,103,151,16,12,139,88,
145,136,206,205,97,196,32,69,249,103,24,0,
16,176,188,58,154,216,38,184,87,71,144,138,
80,187,135,146,235,16,137,17,177,95,32,74,
247,57,91,200,224,117,2,210,142,71,156,122,
186,250,206,32,24,226,69,61,22,46,255,15,
116,30,254,3,182,177,193,245,86,137,191,107,
7,140,71,214,97,101,113,210,240,140,68,56,
37,43,143,174,72,100,84,137,122,78,101,108,
7,244,156,128,229,205,14,131,222,132,233,51,
77,162,40,34,155,75,49,157,203,96,25,46,
134,97,227,239,28,78,89,150,56,115,174,193,
103,50,180,92,34,116,3,143,65,166,132,24,
255,162,50,251,164,49,176,124,182,134,14,173,
162,126,234,191,201,165,100,46,52,178,172,216,
50,230,17,77,145,108,47,98,169,109,82,201,
249,156,169,164,201,212,94,70,41,252,6,227,
113,135,245,181,127,199,217,47,218,76,45,174,
19,72,34,97,248,116,5,138,113,28,225,25,
67,164,125,248,228,113,24,198,214,242,34,162,
36,81,44,148,16,21,149,136,152,170,174,82,
202,40,137,103,113,154,237,164,136,148,142,160,
177,31,59,79,135,191,125,170,62,142,174,31,
237,28,184,60,26,15,160,96,34,186,41,26,
121,151,246,112,138,232,254,29,174,147,193,126,
99,137,217,115,51,64,137,40,18,209,49,112,
237,248,200,174,108,130,0,153,172,196,203,147,
9,127,251,193,20,106,179,205,149,90,136,238,
219,200,139,119,129,152,80,141,18,142,162,125,
139,107,25,137,224,15,252,24,249,49,0,60,
154,174,17,231,210,100,150,54,24,187,50,121,
45,160,222,208,247,128,238,88,150,246,66,79,
98,187,79,148,210,41,180,44,30,244,234,212,
76,135,216,24,225,11,42,69,221,100,205,45,
34,18,50,197,189,164,224,26,129,248,210,28,
175,166,4,126,248,179,33,245,11,95,56,194,
128,115,25,109,253,152,241,134,65,253,202,191,
6,160,111,122,44,119,109,98,98,46,54,18,
190,120,47,56,94,40,140,236,164,162,115,161,
154,121,162,178,72,215,190,202,221,201,77,46,
113,159,240,202,89,226,106,145,229,182,129,38,
206,63,211,67,42,136,49,142,47,32,42,23,
153,189,248,85,182,238,125,155,162,244,46,114,
205,103,123,4,4,18,219,131,44,179,118,31,
51,87,68,22,159,174,194,184,126,246,43,228,
71,207,51,216,120,151,82,243,101,4,65,226,
225,237,255,141,43,173,175,241,206,82,155,235,
47,47,156,226,100,134,9,14,160,230,144,100,
153,122,78,163,16,41,228,107,83,132,81,204,
253,109,147,129,41,48,28,120,220,189,53,98,
225,252,60,130,154,70,0,234,83,197,3,53,
4,158,23,96,78,108,68,73,220,107,92,244,
89,142,95,100,61,157,110,172,15,28,202,217,
167,99,245,85,36,129,203,173,44,119,182,140,
67,134,240,238,232,77,60,12,59,96,46,21,
161,3,169,124,29,249,220,191,225,163,197,63,
97,235,138,200,108,74,67,222,40,158,234,243,
124,123,132,107,244,25,246,191,139,35,109,83,
77,255,22,146,148,65,78,229,144,143,241,176,
155,243,23,24,44,127,0,66,25,65,148,80,
119,174,123,18,46,115,80,233,133,124,176,58,
162,152,86,152,171,164,246,154,24,157,36,107,
158,185,162,80,118,98,95,173,139,223,160,243,
240,59,184,44,82,170,141,121,185,169,226,108,
191,139,62,238,209,44,164,8,126,249,34,174,
160,226,57,17,170,14,41,193,216,83,18,113,
12,182,21,17,71,16,120,17,174,27,147,78,
139,204,77,67,49,187,193,183,59,89,156,112,
76,122,56,73,20,67,20,35,45,173,34,110,
117,9,110,92,65,186,121,159,240,218,5,162,
48,38,157,21,247,240,142,131,113,18,5,233,
225,38,136,49,2,2,166,47,147,81,2,162,
249,86,210,127,123,166,129,120,127,13,177,221,
7,81,32,170,39,86,66,236,133,16,69,52,
221,37,22,141,22,225,168,67,150,33,43,92,
166,197,3,20,92,98,199,37,155,17,24,220,
233,51,93,253,136,94,251,28,158,211,33,12,
6,164,195,85,242,116,200,69,30,11,163,53,
110,183,18,37,114,175,99,30,0,148,110,111,
26,20,210,50,142,127,242,203,235,78,60,38,
78,192,116,49,69,37,167,30,11,169,106,217,
42,139,237,11,156,121,65,70,172,20,24,249,
1,203,166,67,138,103,219,223,64,244,134,44,
111,70,100,74,151,146,141,125,238,235,172,126,
220,71,154,172,80,173,182,233,183,175,194,244,
135,252,52,159,195,233,95,165,114,230,69,34,
239,116,28,51,163,205,91,52,10,95,36,85,
108,178,185,242,87,152,195,135,76,93,252,6,
177,169,33,149,211,76,198,93,140,137,77,246,
36,182,92,163,141,32,136,196,178,142,18,187,
92,170,40,200,146,207,218,218,144,123,195,24,
127,231,176,116,218,35,44,195,225,133,235,53,
52,77,78,210,87,227,56,97,72,221,215,136,
70,5,212,172,128,32,196,137,135,178,171,139,
60,243,83,211,117,28,105,45,255,34,235,233,
84,35,138,147,144,109,253,41,155,201,73,162,
192,165,86,150,187,155,70,130,79,28,101,12,
7,17,247,182,77,46,100,124,10,105,5,69,
203,210,168,255,46,221,251,127,198,133,121,7,
255,20,85,243,163,246,135,12,236,239,114,221,
125,64,107,185,203,210,236,43,104,241,247,41,
102,108,2,91,96,121,113,22,53,211,68,146,
50,148,166,94,66,16,69,236,97,27,61,95,
165,56,251,28,230,246,42,154,88,62,16,178,
157,42,234,108,12,79,207,215,52,180,124,70,
182,79,171,168,51,85,212,89,168,101,184,103,
246,79,77,145,246,84,20,30,251,212,35,8,
18,177,37,225,236,204,181,80,126,137,254,166,
135,149,121,159,161,237,81,179,122,16,249,68,
233,34,104,10,90,16,48,177,68,60,47,34,
159,246,48,199,73,104,108,50,12,169,79,41,
8,34,88,70,18,70,202,22,196,157,152,98,
140,143,139,136,64,48,215,68,18,5,196,187,
203,220,185,51,97,110,38,64,29,155,196,115,
77,194,40,70,148,4,162,40,217,52,135,54,
196,226,42,68,17,78,32,161,72,17,138,24,
98,250,50,122,46,67,84,43,130,32,16,94,
89,32,174,22,137,243,25,98,73,98,125,88,
166,94,54,9,23,166,217,254,193,50,147,193,
26,125,46,32,237,241,14,37,159,19,14,12,
58,127,250,35,132,56,162,156,205,162,137,247,
72,137,54,170,12,138,227,226,56,1,166,25,
80,122,121,10,215,207,208,95,95,167,247,216,
11,142,128,129,181,239,7,39,188,7,215,134,
251,99,88,151,37,26,121,21,221,31,30,25,
19,85,165,171,124,100,173,49,157,245,233,186,
62,241,86,17,73,76,157,216,40,254,96,164,
230,100,26,17,207,30,147,210,93,198,70,25,
73,158,236,221,183,54,243,21,2,207,166,189,
248,215,104,103,151,49,63,62,131,154,158,165,
216,120,9,207,28,18,58,22,81,120,178,219,
219,93,253,9,86,127,149,66,43,1,254,51,
218,101,122,107,111,176,21,127,151,98,245,69,
226,180,72,45,23,48,105,175,147,149,203,143,
214,108,207,133,243,146,20,84,4,98,45,7,
145,207,76,94,65,140,60,162,8,114,74,140,
233,251,4,81,152,36,80,212,211,72,205,12,
65,24,19,69,137,2,87,68,17,85,14,81,
244,152,40,138,49,156,112,143,86,225,208,14,
147,20,4,115,27,66,151,88,201,28,29,99,
220,175,72,158,116,206,194,100,14,165,193,38,
74,24,176,122,90,11,242,211,80,120,124,218,
123,126,198,20,30,79,84,22,162,138,39,78,
78,77,125,179,159,34,102,62,13,203,182,117,
172,178,136,156,17,237,142,140,86,74,237,68,
61,100,196,161,64,168,12,120,121,186,199,207,
62,136,168,205,125,149,208,179,64,16,17,136,
15,112,153,245,134,127,143,124,174,79,173,27,
177,106,121,4,3,135,144,152,57,54,73,73,
17,213,23,99,238,250,119,176,157,8,183,253,
253,36,236,36,27,240,32,15,145,12,195,10,
146,148,162,113,254,81,93,70,85,130,92,33,
98,104,5,12,76,15,47,8,119,168,239,247,
173,215,99,251,46,6,54,44,232,118,68,166,
139,58,115,25,15,35,22,233,76,194,189,140,
197,147,21,197,19,40,60,14,75,94,19,68,
25,87,45,80,206,60,178,184,166,10,191,197,
214,146,194,15,133,183,152,153,123,145,11,195,
13,10,219,29,228,206,128,104,174,137,88,110,
144,201,137,135,207,144,8,139,119,60,166,231,
100,132,8,38,163,136,92,65,68,20,4,114,
10,188,185,81,98,28,24,124,115,166,140,182,
213,103,230,133,44,246,66,157,77,37,143,63,
40,81,11,219,228,11,18,195,143,59,100,46,
213,14,1,221,174,5,130,47,161,136,49,170,
20,225,249,17,113,20,35,125,124,15,73,18,
241,95,127,41,201,118,218,105,168,115,107,107,
26,93,241,89,27,86,184,154,222,192,113,67,
210,254,152,52,135,201,250,22,239,76,246,250,
41,135,219,30,62,10,26,62,154,2,245,166,
78,161,160,210,104,165,233,79,77,17,173,95,
102,24,231,32,149,126,50,216,247,132,247,224,
2,43,54,84,5,143,179,71,208,125,232,133,
26,43,119,63,96,69,188,79,184,149,161,148,
249,58,102,111,25,201,146,40,181,94,62,5,
222,120,50,141,136,32,137,108,91,22,129,157,
165,148,146,15,93,59,28,212,144,219,42,213,
185,175,144,41,157,217,187,103,160,24,168,217,
227,221,245,205,155,223,38,14,61,230,174,255,
193,163,123,138,96,14,151,72,23,166,208,11,
211,244,238,253,57,47,126,254,191,228,189,15,
55,104,61,78,155,225,78,72,167,51,160,164,
177,39,3,166,155,13,16,160,182,47,134,29,
133,1,215,202,173,164,128,79,56,72,61,22,
197,73,165,174,36,10,56,163,237,189,57,88,
94,200,71,107,199,243,77,197,144,164,184,238,
244,79,56,17,32,124,210,251,13,18,1,58,
42,201,9,70,241,207,65,225,241,105,239,249,
25,83,120,60,105,100,202,105,212,64,62,53,
245,205,238,25,217,29,87,138,49,155,67,151,
245,161,67,244,24,86,168,201,18,231,230,102,
15,132,124,166,95,248,55,220,223,250,33,51,
206,247,120,161,121,135,251,35,139,48,22,152,
116,77,4,207,37,223,144,8,133,27,248,254,
16,121,42,81,182,118,33,147,236,181,24,162,
190,193,210,91,75,228,46,156,97,60,157,225,
122,89,162,160,200,48,15,110,24,179,189,21,
17,20,76,250,147,58,37,87,103,230,210,111,
30,74,2,209,129,66,5,102,162,152,135,93,
43,137,82,156,98,189,60,224,129,9,69,124,
206,206,181,104,77,203,116,39,30,221,137,135,
174,138,71,214,43,125,170,208,211,82,219,68,
149,109,84,89,164,145,215,168,100,85,26,231,
190,198,202,221,101,214,206,173,18,54,11,40,
233,41,46,174,223,36,187,209,37,206,85,147,
83,255,152,162,8,3,232,119,66,62,254,80,
230,242,23,100,170,90,0,70,68,38,35,242,
219,249,16,59,28,242,206,54,252,229,122,134,
230,153,41,252,40,102,117,88,160,98,103,249,
165,165,31,146,173,106,68,19,147,180,27,35,
229,98,132,70,17,79,82,113,157,164,57,82,
94,246,17,148,71,89,2,91,155,14,158,16,
209,106,104,100,212,16,97,56,33,46,61,2,
122,21,49,100,161,210,33,136,36,222,250,89,
11,85,217,6,204,99,4,196,163,77,37,17,
145,209,3,166,230,138,200,59,49,108,109,166,
128,119,166,193,79,222,201,81,94,120,9,107,
221,224,89,142,158,225,161,246,109,102,142,168,
119,105,76,255,54,27,31,255,57,141,249,223,
32,93,156,161,63,254,30,122,180,72,224,94,
222,99,102,253,196,152,107,182,130,255,176,129,
116,121,149,241,157,15,40,180,46,30,244,104,
132,38,181,179,255,242,137,197,72,143,111,134,
124,253,50,153,234,220,1,207,71,148,53,180,
98,1,99,178,136,160,127,151,198,165,8,113,
60,70,150,246,103,33,68,73,99,30,45,71,
165,148,163,85,212,177,199,33,169,252,241,32,
231,81,209,34,81,16,142,36,173,76,171,18,
146,40,156,220,161,76,148,33,85,70,112,70,
196,90,22,4,233,144,128,20,71,183,137,245,
169,83,45,71,222,158,160,4,62,102,174,206,
47,198,241,67,149,69,106,57,21,103,240,169,
16,55,90,69,157,114,86,101,185,107,49,180,
18,175,55,173,73,204,22,82,135,112,129,56,
142,201,54,127,133,182,113,30,167,247,19,210,
226,50,47,191,36,240,143,63,42,81,190,240,
175,177,87,255,45,74,253,67,202,254,26,53,
211,198,19,83,148,37,129,78,44,240,213,207,
91,60,248,255,62,64,175,166,88,121,181,194,
189,109,248,217,86,142,180,36,82,210,5,252,
80,64,20,74,84,212,136,27,103,122,248,121,
147,123,203,255,142,124,253,155,164,11,211,71,
134,208,206,213,147,162,186,174,121,250,39,30,
218,62,126,16,145,146,5,234,121,141,250,14,
21,129,237,133,152,110,120,88,81,232,138,248,
196,248,248,177,218,41,136,240,130,8,195,9,
104,143,93,206,215,51,228,178,55,24,247,186,
180,203,62,191,84,139,144,165,10,157,94,72,
54,244,240,35,153,13,219,221,107,58,87,8,
21,36,25,94,123,61,197,95,255,185,195,96,
34,176,105,233,120,196,148,139,61,26,186,202,
210,196,38,240,243,244,44,5,199,171,147,85,
36,62,159,31,83,126,255,167,172,221,218,36,
155,18,49,140,128,74,85,131,158,75,110,170,
128,244,185,11,228,2,147,184,152,37,206,103,
16,204,71,209,184,185,185,196,162,119,2,240,
35,17,169,63,38,220,167,40,68,49,89,11,
57,116,185,58,121,143,219,161,140,67,142,52,
147,99,215,97,102,58,77,54,39,163,235,18,
73,171,227,128,73,165,73,239,98,147,55,31,
154,104,202,13,58,19,255,51,57,40,91,35,
151,233,82,234,144,224,211,50,85,22,94,248,
239,30,109,40,201,231,11,47,200,124,255,141,
127,192,46,124,25,199,15,201,234,50,205,130,
182,135,55,61,205,152,189,248,71,172,63,252,
19,130,212,50,129,103,35,239,115,183,235,23,
94,127,234,251,197,196,100,170,115,135,126,190,
246,222,95,162,148,33,82,12,38,101,131,145,
7,209,232,33,158,91,58,24,250,216,241,92,
54,134,14,21,197,133,200,199,53,134,135,35,
47,214,4,87,77,33,16,35,167,242,136,39,
244,166,30,219,1,97,28,147,57,134,95,72,
12,236,189,52,203,189,231,208,11,8,158,65,
28,199,7,201,0,101,141,40,127,1,193,51,
17,39,139,196,153,5,226,19,44,240,137,158,
69,250,69,43,212,39,239,195,114,234,200,212,
241,79,100,0,201,34,23,155,217,61,69,81,
72,41,184,99,247,8,35,67,216,49,152,154,
104,217,223,35,10,60,254,233,205,239,160,151,
23,16,68,137,80,125,133,121,237,71,68,63,
188,199,192,241,9,130,152,202,133,60,158,229,
241,209,127,248,144,233,179,69,86,94,61,195,
48,130,74,126,139,254,240,12,255,162,210,163,
148,141,17,70,38,81,179,156,84,128,74,10,
147,32,36,243,146,201,79,111,254,223,84,188,
111,144,175,93,58,114,238,11,181,52,117,49,
135,86,42,50,178,124,250,166,199,200,10,142,
12,199,239,142,91,155,19,166,203,73,33,222,
238,18,158,173,103,184,181,49,57,80,220,42,
3,92,155,41,240,225,218,24,247,83,46,178,
225,4,124,180,54,230,108,253,5,42,118,137,
237,219,223,197,120,109,72,105,161,129,86,141,
64,129,190,231,243,225,208,64,136,65,16,5,
94,212,75,100,227,24,43,140,168,182,4,6,
182,66,76,76,232,10,108,244,83,108,104,30,
82,37,164,51,78,243,133,217,14,151,242,105,
164,225,132,251,127,113,147,149,158,65,70,23,
168,214,52,242,5,133,254,192,199,152,4,180,
170,35,164,155,183,17,12,11,20,229,88,146,
32,93,14,25,187,10,185,229,45,228,222,136,
104,97,138,88,85,208,197,2,166,167,145,219,
220,36,39,58,228,241,16,49,142,180,40,179,
25,153,24,40,85,84,148,122,129,56,155,134,
213,54,227,106,139,224,98,139,119,250,35,194,
173,11,164,155,23,89,25,126,54,20,20,81,
28,19,19,63,177,98,56,12,20,254,225,182,
129,39,204,236,165,215,77,156,128,206,216,165,
85,208,105,20,180,19,51,42,226,40,60,80,
172,39,171,105,234,245,223,101,179,253,167,108,
44,253,239,20,203,95,34,223,124,254,19,61,
131,107,116,105,47,255,41,249,210,151,40,78,
29,172,28,95,248,252,191,66,16,68,54,22,
255,28,103,187,141,222,127,13,39,127,150,126,
119,245,81,24,118,39,54,45,10,137,117,164,
102,82,39,64,0,54,218,78,248,43,105,115,
121,124,229,243,196,241,89,31,60,194,148,100,
81,32,220,9,79,149,51,42,211,133,52,183,
218,54,143,71,183,99,53,11,196,8,246,144,
88,207,61,242,46,148,52,120,38,177,168,19,
11,39,199,222,51,174,137,18,6,76,10,191,
80,6,199,141,221,72,198,179,30,197,180,242,
84,215,139,178,74,182,245,165,61,140,36,219,
124,149,238,198,15,80,251,54,174,27,49,61,
157,230,238,173,33,98,24,34,8,2,238,11,
23,217,24,86,25,59,53,68,79,229,43,250,
59,108,254,233,91,104,11,57,242,57,9,92,
23,97,181,77,60,63,69,228,169,52,102,179,
204,156,51,89,223,248,127,48,150,158,39,147,
189,68,28,249,164,10,211,104,251,42,247,5,
33,217,163,149,172,74,37,171,18,70,49,67,
203,167,51,118,143,196,96,130,48,102,185,103,
177,61,113,89,168,101,200,104,18,41,69,226,
185,233,60,107,125,123,47,233,70,222,189,121,
49,173,208,30,126,250,5,14,162,152,187,91,
6,115,149,25,206,92,249,111,120,239,141,255,
200,248,210,58,207,205,233,152,147,136,237,208,
67,18,4,124,67,0,57,226,125,83,97,57,
234,51,242,2,50,185,12,81,187,68,177,250,
10,122,173,137,164,106,152,131,101,122,91,127,
129,152,138,8,227,152,158,23,160,135,58,133,
84,128,94,145,41,20,85,114,57,25,81,16,
72,23,179,100,213,128,56,138,17,38,59,232,
112,24,17,107,74,34,66,119,72,10,247,99,
60,154,44,48,118,36,10,88,72,31,46,37,
30,71,234,62,139,238,57,138,209,26,253,129,
71,179,34,176,37,230,56,51,37,113,243,230,
136,104,71,211,106,170,200,153,243,121,4,93,
161,125,127,72,243,229,75,244,245,28,170,90,
64,106,230,120,119,52,198,90,205,49,61,247,
91,108,15,44,194,232,179,75,169,60,141,69,
149,87,94,67,138,106,108,72,149,196,90,217,
21,156,81,204,218,192,102,115,228,208,44,232,
180,138,218,222,58,185,65,184,151,82,23,71,
33,15,222,250,63,41,182,158,39,87,191,140,
162,101,72,229,27,200,107,117,190,242,252,61,
238,111,252,25,253,45,155,66,243,149,167,243,
36,226,8,179,253,127,241,219,103,86,121,195,
252,123,58,247,86,73,231,175,239,219,252,59,
217,117,231,191,137,213,127,153,204,243,179,44,
110,140,137,133,21,194,32,66,86,196,29,208,
91,226,92,61,179,55,223,83,133,45,178,37,
60,115,4,90,142,173,145,75,16,69,84,50,
42,133,29,65,145,211,21,32,81,20,178,36,
112,99,190,136,23,70,12,12,159,114,86,65,
145,68,46,53,101,22,31,12,240,142,8,103,
196,169,98,210,251,88,217,105,125,186,251,204,
153,217,39,206,205,86,83,120,159,178,195,221,
252,131,183,72,71,17,253,92,153,250,176,205,
122,101,150,249,205,59,60,104,44,176,208,91,
231,65,227,60,11,237,165,99,255,239,139,49,
203,23,127,245,231,82,73,100,117,153,185,74,
250,231,114,110,130,32,18,7,224,251,59,204,
217,118,128,109,39,222,161,32,10,140,205,22,
243,21,149,231,154,107,220,121,88,70,124,227,
38,97,186,194,195,85,131,153,150,72,105,171,
135,224,135,220,255,187,155,196,122,138,114,39,
197,165,151,175,178,89,26,33,232,109,6,241,
219,32,129,177,221,128,205,34,149,250,151,73,
229,167,142,12,75,237,42,13,203,11,217,26,
57,244,12,239,144,237,108,121,33,55,55,198,
92,110,229,200,233,50,154,44,114,174,158,161,
89,208,15,98,20,133,180,76,251,25,46,212,
230,208,161,145,47,80,155,249,10,131,209,144,
239,191,249,38,177,55,194,22,90,200,66,147,
124,254,18,145,237,227,152,171,12,186,125,20,
170,196,234,12,179,87,15,186,85,185,218,5,
156,135,175,209,247,63,228,225,196,97,121,236,
80,139,234,92,248,234,53,102,238,44,35,200,
82,226,57,28,0,90,147,44,166,168,94,225,
225,118,145,137,169,50,85,30,81,27,175,176,
237,21,89,187,25,162,69,46,196,49,138,228,
147,23,135,136,174,189,35,112,99,54,151,186,
244,83,117,110,142,71,216,86,200,204,153,34,
162,34,225,198,50,162,166,17,89,137,224,208,
116,9,203,151,17,94,185,74,37,186,131,247,
246,3,226,207,93,230,227,172,0,147,9,67,
35,36,203,43,40,169,34,221,213,225,169,43,
119,79,183,17,143,86,18,251,139,119,250,134,
71,123,236,34,75,2,178,40,224,71,117,70,
102,41,9,137,28,101,105,71,49,235,3,155,
158,225,114,54,13,166,27,112,107,115,194,197,
102,150,66,74,65,148,85,228,114,128,93,251,
11,70,171,63,162,85,255,67,244,124,3,145,
52,63,9,66,94,104,130,213,253,30,214,104,
154,116,225,116,113,248,73,247,46,131,246,247,
249,122,235,1,210,196,227,198,57,159,239,220,
188,75,154,107,71,30,190,76,37,17,176,41,
93,225,242,243,179,200,59,196,112,186,44,113,
117,42,255,212,89,170,187,74,104,123,236,178,
53,74,222,107,74,145,40,164,21,226,56,70,
87,196,3,214,87,123,236,210,200,107,52,10,
143,168,101,51,154,196,229,86,22,71,201,242,
176,107,29,206,81,215,139,224,153,8,161,67,
172,230,78,61,55,205,119,145,163,144,209,167,
216,39,53,163,207,7,151,94,199,83,83,180,
235,231,147,125,177,227,65,189,91,75,190,127,
183,52,125,236,255,47,126,240,45,150,127,30,
113,9,73,228,66,35,243,89,100,37,63,195,
51,154,96,164,81,4,227,241,190,176,115,237,
28,207,85,86,161,85,6,4,44,95,71,174,
191,76,48,94,100,44,212,9,54,218,116,58,
29,20,69,102,123,20,51,149,23,153,200,121,
114,178,200,171,149,60,130,105,82,40,85,8,
162,136,168,234,147,206,245,248,248,254,159,176,
121,255,121,242,197,147,201,62,43,217,132,170,
230,94,199,60,100,216,196,49,116,198,238,129,
42,238,140,38,29,84,20,121,93,65,83,68,
156,103,180,72,126,24,177,218,183,169,0,249,
250,21,224,10,73,110,200,193,55,91,32,161,
216,118,131,136,238,230,26,70,223,70,145,196,
3,7,177,58,255,5,194,135,10,74,91,68,
20,20,6,246,38,63,172,188,201,151,95,185,
74,110,99,27,105,71,81,136,66,76,24,11,
72,196,132,126,140,57,137,152,220,27,242,82,
252,62,27,157,89,110,11,117,252,72,33,47,
247,144,136,118,132,163,138,65,29,35,138,169,
43,61,228,112,64,119,36,18,101,36,60,69,
212,10,219,0,0,32,0,73,68,65,84,52,
60,2,54,58,33,87,23,64,211,98,206,76,
41,44,174,64,46,171,80,171,107,20,82,1,
147,197,53,132,139,211,164,239,220,71,77,139,
252,108,224,227,135,49,194,242,89,106,87,191,
200,74,207,194,13,66,80,158,221,38,188,216,
72,226,168,29,99,175,201,31,221,123,63,33,
215,184,140,150,45,17,199,176,58,176,113,247,
97,79,130,29,16,103,142,216,29,143,157,54,
199,143,216,26,57,148,245,136,56,134,197,45,
147,43,83,57,50,154,68,181,250,235,12,187,
255,129,95,123,53,224,141,247,254,61,25,239,
155,100,242,151,24,182,111,242,102,221,165,217,
244,153,44,221,63,149,162,176,71,27,248,163,
191,227,11,23,198,24,111,117,120,239,218,231,
88,251,142,198,244,213,223,65,16,78,94,172,
102,65,103,60,83,194,112,18,238,174,211,18,
39,30,41,116,50,5,250,27,43,32,229,118,
48,170,164,67,98,207,12,208,164,131,184,210,
74,207,34,171,201,123,7,232,241,112,197,133,
70,150,143,215,199,71,125,72,146,248,96,245,
193,183,16,60,227,16,56,186,87,157,27,6,
196,178,138,47,43,68,161,248,169,246,201,91,
215,190,78,173,179,68,58,140,88,126,44,225,
224,52,227,65,109,254,231,82,8,207,86,82,
159,8,87,251,231,28,126,32,160,107,18,65,
16,16,4,137,113,166,235,18,87,167,122,132,
249,164,134,43,136,36,84,41,224,226,21,131,
205,181,1,189,174,77,227,218,60,153,153,60,
177,233,50,59,149,69,104,86,48,140,152,88,
146,168,75,18,99,59,68,21,5,84,81,194,
36,66,0,158,63,171,19,179,200,214,154,66,
186,220,58,114,62,183,55,39,79,36,213,244,
195,163,61,88,121,191,133,58,93,74,113,223,
122,118,188,255,91,35,135,174,61,33,23,228,
72,41,18,170,44,34,75,2,146,152,88,184,
142,159,144,222,77,156,0,219,11,17,76,135,
216,119,118,64,68,159,11,205,236,158,213,215,
92,248,124,114,160,194,128,254,106,132,219,173,
147,22,151,144,214,59,123,159,151,86,66,122,
150,138,82,74,161,140,108,182,228,42,23,46,
108,18,247,75,52,106,50,173,254,58,248,1,
209,181,6,113,186,128,48,54,144,110,222,71,
112,18,221,26,34,97,154,33,2,1,195,176,
132,230,135,168,120,56,150,64,24,39,46,152,
101,5,220,184,154,66,148,69,144,4,70,173,
51,248,85,157,64,83,80,62,127,141,247,29,
31,223,17,16,87,174,48,125,233,247,25,90,
1,91,35,247,153,110,192,102,65,167,144,86,
24,236,0,110,187,209,137,234,185,95,218,187,
198,116,131,3,74,226,84,174,201,190,49,176,
124,148,157,152,102,20,199,44,182,13,158,159,
206,147,46,206,16,248,191,201,59,183,255,134,
215,111,72,188,117,231,207,240,156,95,66,231,
37,60,247,13,214,99,23,53,30,31,47,4,
1,107,180,193,176,251,51,124,113,153,215,95,
246,49,254,236,38,90,41,69,181,233,208,91,
207,161,23,154,184,79,232,27,46,75,2,87,
167,114,108,79,92,4,4,228,193,230,39,94,
207,238,196,195,17,82,73,161,157,154,97,115,
224,176,53,114,112,199,54,184,214,33,189,250,
176,107,241,220,244,209,158,65,70,59,41,59,
74,72,114,220,149,244,14,134,241,152,206,62,
160,69,7,72,97,128,252,12,192,236,237,250,
121,74,195,77,102,215,63,102,117,250,185,167,
250,219,133,237,101,222,107,61,247,115,39,132,
127,158,149,68,20,120,152,43,127,204,185,193,
93,58,230,65,92,160,213,74,17,55,203,196,
153,20,182,175,178,54,168,112,237,204,38,182,
83,102,234,95,253,50,210,127,252,49,189,7,
61,26,95,189,68,188,131,23,90,97,120,224,
156,166,179,34,195,126,72,58,35,18,69,49,
97,16,99,140,67,174,158,145,120,176,114,23,
248,26,71,165,238,233,138,132,17,6,39,206,
253,184,176,237,129,160,121,74,17,57,91,203,
176,246,152,37,250,105,70,24,193,192,244,25,
60,37,183,255,192,242,49,156,224,16,107,170,
32,201,84,206,188,204,248,195,183,176,38,48,
178,52,50,202,78,255,138,88,68,211,69,226,
235,103,145,179,50,214,3,9,253,172,206,32,
119,153,206,184,128,88,140,137,34,129,92,228,
208,16,135,196,197,92,210,5,111,71,81,72,
132,172,175,219,140,149,41,106,227,183,89,245,
231,153,229,46,2,49,91,27,54,197,178,194,
118,199,37,157,81,200,124,237,26,113,58,197,
168,31,178,22,153,20,194,136,59,94,192,198,
166,136,62,254,2,205,43,95,195,11,98,238,
111,143,159,233,38,148,68,129,169,199,120,109,
162,56,38,140,226,3,64,180,233,125,74,1,
19,39,45,19,119,135,23,68,220,109,27,92,
110,101,201,215,174,48,106,7,124,239,237,191,
39,127,198,193,114,126,140,108,183,224,246,85,
194,218,10,182,185,246,152,46,74,230,53,218,
186,137,34,229,24,77,126,196,133,23,86,153,
248,33,218,187,27,220,223,152,176,160,248,156,
183,93,182,171,219,184,230,241,96,89,20,6,
7,138,170,106,185,196,243,116,162,26,158,53,
222,21,199,196,128,154,62,29,183,149,40,146,
100,31,137,33,66,224,224,161,195,9,203,87,
122,2,208,249,76,162,33,169,18,225,184,131,
40,63,155,112,229,160,216,34,150,21,26,155,
183,24,20,26,120,242,233,186,50,61,168,205,
253,92,10,227,187,91,6,173,162,78,61,175,
254,220,41,13,179,243,14,215,27,109,250,31,
39,198,142,44,11,52,102,178,20,207,213,80,
139,233,68,1,196,144,82,60,46,212,55,49,
39,17,174,21,145,170,201,20,47,212,89,253,
97,135,168,51,64,104,150,233,123,62,183,71,
22,47,166,114,4,126,140,231,198,68,17,168,
170,136,27,168,196,130,71,224,199,20,43,73,
82,141,126,166,75,111,245,13,42,179,191,116,
104,94,211,37,157,123,109,243,88,78,185,93,
175,248,137,138,98,55,134,85,201,170,244,77,
143,149,158,253,212,156,32,207,114,108,79,188,
35,233,181,125,107,66,185,50,68,30,77,16,
20,208,207,86,17,58,3,172,73,136,246,252,
28,113,46,249,155,145,147,227,222,182,68,62,
101,115,190,190,245,200,130,52,242,244,204,28,
149,76,82,63,33,140,12,194,40,230,193,3,
147,177,13,25,182,17,125,7,133,34,54,58,
2,17,131,161,199,96,39,107,201,114,66,210,
59,89,53,75,147,152,145,226,209,118,60,162,
78,145,146,244,27,20,206,61,71,12,220,235,
24,207,188,127,66,171,168,239,213,15,236,215,
254,94,16,29,104,17,105,185,193,51,127,31,
134,19,112,103,211,224,82,43,75,161,113,13,
255,225,136,170,255,29,178,13,139,117,150,136,
134,42,74,251,57,244,236,204,49,126,237,79,
169,41,93,206,158,213,80,66,157,242,196,96,
237,131,109,90,117,153,92,86,66,232,13,185,
186,48,199,183,255,230,223,51,115,253,15,49,
186,75,140,39,31,32,160,33,145,39,242,61,
124,211,96,246,165,223,57,58,132,244,152,98,
216,85,28,137,240,22,80,210,71,123,1,229,
140,74,45,23,176,61,129,216,30,36,180,223,
39,140,252,147,200,13,133,167,248,85,124,124,
215,178,73,105,154,203,247,223,162,87,121,54,
194,122,152,173,34,166,75,220,120,255,91,188,
245,220,151,79,233,81,172,240,94,235,249,159,
59,69,17,197,9,158,182,49,180,169,229,52,
230,42,169,159,139,121,185,163,37,234,226,119,
233,254,104,145,254,32,162,88,171,35,94,189,
198,104,97,138,74,173,67,16,195,221,137,197,
184,63,70,245,52,28,187,206,84,161,199,72,
115,208,204,52,241,212,52,47,255,126,1,106,
69,220,48,226,157,254,132,172,44,145,206,138,
252,255,220,189,233,147,93,249,121,223,247,57,
251,114,247,173,119,116,55,26,235,0,179,146,
156,33,41,82,164,54,202,150,202,114,201,178,
202,118,36,39,149,23,246,43,199,169,84,222,
248,125,254,133,164,28,167,146,84,37,85,138,
229,138,165,68,229,178,28,217,20,41,82,148,
134,219,204,112,102,48,3,96,0,244,190,222,
238,190,251,61,251,242,203,139,115,123,67,119,
3,141,1,56,164,245,171,234,66,163,251,244,
185,231,252,150,103,249,62,207,243,125,60,39,
197,180,100,218,94,145,106,174,135,44,69,220,
221,185,196,173,210,6,251,65,196,186,235,19,
169,9,74,124,118,0,161,100,105,188,122,169,
196,122,219,59,151,84,48,171,141,147,14,141,
206,3,116,73,5,232,222,249,8,77,150,177,
227,46,230,32,195,78,167,128,177,84,240,104,
215,33,124,28,183,26,81,120,160,94,128,38,
216,109,35,236,193,197,130,63,143,93,171,236,
43,152,238,17,168,30,133,62,154,110,226,247,
182,240,175,122,236,206,207,81,43,216,164,113,
130,186,222,36,73,85,228,254,144,120,188,138,
36,129,38,69,44,212,155,39,152,102,37,9,
234,249,62,159,52,167,168,229,6,72,253,33,
169,16,52,183,125,122,221,144,20,3,3,159,
24,153,52,167,211,46,126,141,184,187,65,221,
91,199,102,128,4,132,169,76,42,4,178,36,
145,147,4,61,64,116,52,198,58,175,102,228,
100,171,15,217,236,248,40,78,200,97,13,178,
223,3,179,251,169,230,225,80,16,202,50,243,
114,30,121,20,221,172,5,49,131,157,13,132,
61,64,139,109,204,99,237,33,213,166,67,57,
74,94,216,90,156,128,20,183,20,230,235,57,
166,69,131,197,187,117,126,229,250,38,219,154,
6,229,144,56,188,71,174,91,194,92,125,120,
58,38,225,134,184,215,19,66,225,210,232,14,
225,7,43,52,74,80,249,181,215,16,249,28,
159,132,49,15,127,160,240,202,220,223,162,191,
242,49,27,221,119,41,214,234,72,66,34,145,
183,81,163,28,179,229,55,144,207,184,119,220,
217,193,40,119,70,107,61,82,164,143,93,19,
6,75,217,220,12,90,200,251,187,39,164,246,
130,128,49,63,102,16,196,196,189,13,20,213,
64,9,134,4,218,30,94,144,32,164,35,108,
200,12,114,200,170,116,194,251,82,220,30,178,
189,149,29,200,230,48,179,218,132,64,28,87,
2,105,130,72,67,212,148,195,185,85,68,196,
120,165,64,179,23,28,246,242,56,126,206,146,
48,164,186,126,143,244,34,30,192,5,247,88,
75,179,41,239,172,92,40,185,98,87,53,40,
111,47,93,64,66,14,179,251,61,11,133,199,
5,62,255,34,123,54,4,154,134,202,164,238,
28,202,175,167,222,119,216,193,232,62,125,174,
218,221,53,114,106,225,66,215,74,110,31,75,
254,19,228,143,62,96,224,196,216,186,194,228,
231,26,132,51,128,182,203,31,63,44,82,211,
101,48,187,12,227,24,165,99,16,21,87,184,
219,28,39,77,74,108,230,52,116,165,140,162,
134,108,45,198,20,114,187,8,82,252,36,229,
78,119,200,184,169,19,167,10,63,222,21,188,
50,17,50,101,233,148,44,151,191,88,179,113,
148,181,204,152,91,181,193,249,136,230,251,107,
84,181,107,152,70,1,77,59,121,18,94,2,
230,146,132,141,253,29,124,217,190,208,154,169,
0,139,122,198,78,169,199,130,249,210,36,197,
99,238,199,68,35,225,147,237,193,73,119,37,
116,65,86,200,229,114,24,154,76,199,9,207,
141,107,72,142,129,200,93,140,253,242,241,107,
7,192,165,137,242,137,172,158,0,216,119,223,
35,213,19,86,34,137,222,208,227,102,222,70,
154,168,33,55,93,132,157,5,53,63,89,175,
114,101,114,255,132,177,118,34,83,72,30,241,
246,88,6,195,97,204,246,142,71,128,129,76,
66,128,78,91,170,81,254,242,63,35,87,159,
101,245,207,254,7,12,207,57,148,43,114,28,
211,234,12,88,21,16,10,27,66,9,179,247,
38,185,107,95,203,168,53,90,46,187,86,0,
199,207,129,103,131,85,249,84,243,112,48,230,
235,54,81,225,72,96,136,56,101,48,76,16,
185,58,78,35,143,150,59,90,183,253,160,123,
170,208,230,121,214,226,132,165,11,132,186,206,
229,122,142,137,224,247,248,120,235,127,230,141,
89,159,119,145,144,27,62,157,78,15,123,242,
12,43,120,105,138,226,112,135,201,59,75,244,
246,61,166,111,214,72,110,94,102,77,85,216,
30,122,236,172,105,52,234,127,139,160,50,71,
215,253,115,114,181,89,36,163,200,112,127,145,
242,196,155,212,102,94,59,23,192,28,196,93,
22,151,254,148,107,211,109,162,94,66,48,245,
155,148,198,142,210,116,99,223,65,53,179,103,
74,219,219,136,51,2,126,249,209,215,145,220,
205,40,60,146,84,208,30,134,116,220,8,75,
83,208,106,214,41,190,167,180,183,135,24,81,
66,244,25,205,253,40,230,113,160,77,164,221,
31,33,138,215,145,146,228,112,110,21,89,226,
202,108,25,45,78,216,236,248,116,221,236,44,
149,108,141,126,175,199,160,54,197,237,181,247,
89,175,207,115,169,181,198,102,109,134,137,206,
14,123,197,6,181,65,139,78,174,76,222,119,
240,53,19,45,28,144,202,114,22,10,73,19,
66,213,32,23,56,12,172,34,101,167,67,171,
80,163,209,223,67,73,67,214,106,83,25,229,
198,83,198,194,206,71,124,80,159,190,128,146,
26,128,110,61,3,133,71,114,161,207,191,240,
158,181,52,42,250,0,233,156,64,238,105,157,
106,193,83,232,62,68,154,178,254,240,63,49,
247,234,111,65,249,233,180,242,189,141,63,224,
23,38,19,238,189,147,25,104,165,146,138,60,
55,197,174,171,226,4,30,5,225,210,147,247,
217,219,155,68,145,19,68,42,81,247,161,81,
202,12,140,8,8,82,133,36,86,176,173,132,
173,246,37,76,221,165,146,223,99,213,241,89,
117,50,79,97,189,95,35,49,6,60,232,43,
152,74,159,221,200,36,116,199,240,35,139,241,
249,117,144,92,188,97,139,229,7,123,104,185,
20,195,191,66,125,252,171,24,246,17,117,142,
2,140,151,243,72,102,9,55,134,182,19,209,
117,194,115,41,40,85,128,100,212,185,206,15,
53,238,181,35,106,161,196,92,205,70,85,36,
76,29,174,204,234,44,239,185,89,192,25,208,
148,148,137,74,142,137,90,86,9,52,21,165,
108,116,188,51,91,236,73,154,142,208,46,134,
135,158,117,173,35,84,242,143,53,82,215,204,
58,126,12,154,33,241,106,20,193,246,30,98,
126,10,209,93,71,84,139,188,255,104,140,98,
41,69,40,6,60,150,199,37,245,29,80,100,
42,246,144,174,155,163,154,179,49,77,229,16,
22,144,16,36,40,160,104,152,19,87,8,135,
187,164,126,64,138,66,136,130,78,132,170,64,
99,232,242,19,77,199,107,73,232,195,55,24,
191,249,183,137,211,44,208,217,242,128,199,223,
57,214,79,255,236,25,230,193,210,21,106,181,
226,137,133,212,116,72,244,236,218,72,211,17,
122,230,225,197,137,32,58,195,219,123,222,181,
56,97,105,250,144,11,160,81,168,177,189,249,
50,47,237,124,155,198,116,133,189,36,37,40,
61,192,245,191,130,85,60,121,184,26,115,223,
160,185,255,49,115,186,66,233,23,175,241,253,
98,30,215,143,80,165,8,33,36,44,255,243,
228,22,110,16,184,125,230,138,59,76,41,43,
180,10,37,214,42,191,67,174,118,147,152,243,
41,150,173,250,60,11,211,55,233,172,255,33,
37,123,13,223,125,143,130,246,21,154,143,190,
77,34,58,180,227,38,227,118,14,153,5,130,
97,159,233,137,249,167,135,105,116,3,161,27,
200,64,221,52,169,159,21,116,126,236,218,84,
28,155,123,17,157,92,243,233,175,129,223,67,
146,195,195,185,77,128,13,39,101,166,106,179,
144,179,51,74,27,33,80,36,137,119,92,15,
33,224,195,151,126,25,73,8,134,213,25,18,
89,161,87,157,37,149,21,218,141,203,164,146,
204,190,16,164,146,132,228,181,79,240,252,8,
64,30,253,110,79,164,135,127,147,196,65,230,
233,92,96,47,44,78,92,57,148,15,79,14,
70,6,217,253,46,162,40,36,50,174,167,11,
220,247,34,123,86,146,96,102,162,72,58,12,
17,250,197,246,247,193,122,61,109,220,248,181,
127,134,223,219,187,208,181,170,158,160,168,89,
243,52,85,149,201,127,241,10,171,65,204,198,
176,198,74,168,83,207,239,210,111,222,70,151,
6,200,134,79,217,8,105,39,54,105,47,79,
144,24,116,7,13,242,70,159,74,97,143,98,
174,205,100,117,21,55,200,179,219,153,97,172,
146,197,253,188,48,135,31,230,248,100,227,117,
84,37,162,104,246,144,228,132,246,96,140,171,
211,119,144,165,20,132,196,110,123,18,249,178,
96,172,178,73,204,251,108,110,47,98,183,222,
196,46,94,34,87,91,64,150,101,132,110,160,
217,22,101,69,165,92,201,224,235,102,63,96,
183,31,156,74,198,56,115,85,91,195,144,158,
23,49,83,177,40,89,26,121,67,229,149,153,
34,113,34,80,21,9,111,168,16,10,153,189,
65,128,170,200,148,109,141,43,99,57,38,203,
38,61,55,202,186,69,249,49,193,11,136,111,
180,157,232,84,156,162,49,251,117,86,31,124,
66,249,115,93,232,244,25,222,109,34,105,123,
24,99,54,82,111,136,44,165,228,204,136,230,
160,68,123,152,167,96,123,84,237,33,185,230,
6,210,39,235,200,82,202,248,252,20,247,245,
215,169,47,109,145,144,113,184,69,169,140,134,
138,66,66,46,233,211,252,247,255,2,201,115,
153,13,150,145,71,7,47,192,32,20,10,82,
16,18,197,58,37,126,145,234,205,55,233,123,
17,75,123,238,133,98,58,138,44,81,201,105,
36,105,22,132,14,162,244,204,185,146,164,44,
95,92,87,101,166,43,214,153,231,45,167,43,
12,31,195,190,131,207,40,174,180,186,239,81,
48,85,10,213,207,241,222,222,29,110,119,87,
216,191,57,133,50,225,208,92,251,55,88,173,
219,140,95,254,181,163,247,214,77,118,218,211,
252,164,40,177,43,5,68,221,16,217,132,75,
182,201,238,138,70,109,42,11,192,245,247,62,
230,74,218,102,147,10,194,245,233,122,239,17,
152,11,44,239,185,88,186,194,124,221,62,145,
235,125,100,64,20,144,173,183,88,87,134,204,
55,246,216,184,243,127,32,43,18,95,158,187,
203,183,163,121,74,150,195,107,218,183,88,139,
115,172,175,108,18,37,179,148,26,159,59,165,
208,158,103,156,136,73,9,145,113,80,29,143,
67,28,254,236,104,141,182,58,46,150,42,81,
205,107,135,214,158,27,36,136,131,116,54,145,
34,36,153,104,84,176,151,142,130,183,135,255,
30,186,201,202,41,24,33,61,252,87,57,250,
155,228,226,12,1,63,175,89,79,199,199,108,
213,198,214,21,220,231,184,199,121,205,131,158,
101,68,190,140,215,118,136,99,193,244,235,211,
188,107,44,224,250,29,132,253,144,134,37,179,
185,191,64,177,184,78,201,238,16,183,39,233,
244,26,180,80,177,141,33,182,57,64,45,38,
52,74,91,172,238,94,199,13,115,32,36,106,
197,29,82,33,179,214,188,142,161,123,116,134,
117,110,204,188,143,36,9,6,131,75,124,105,
166,195,251,173,24,69,146,50,185,167,42,172,
180,198,25,159,222,100,127,120,228,93,41,147,
3,252,228,219,184,158,196,230,131,25,108,205,
128,88,160,72,21,26,51,95,67,183,43,232,
170,204,165,170,197,68,201,100,117,223,165,237,
132,79,86,20,7,27,126,101,63,155,122,93,
149,201,25,10,81,34,240,194,132,196,31,102,
150,195,8,55,45,89,26,147,101,147,130,169,
98,151,149,195,243,208,236,7,108,186,79,76,
32,121,234,104,246,124,20,57,11,226,30,64,
80,178,162,82,175,253,109,22,127,242,159,152,
120,197,164,248,86,30,227,227,135,176,239,177,
146,187,194,84,185,143,27,23,153,46,181,105,
14,74,84,237,33,61,215,230,190,251,58,5,
189,134,210,235,81,91,233,115,163,248,61,30,
245,47,113,213,92,68,83,100,58,242,56,133,
164,143,41,134,20,69,151,226,99,184,164,4,
148,172,152,177,70,1,79,181,145,214,167,168,
222,120,147,97,16,115,127,251,226,100,127,179,
53,235,48,91,39,137,67,100,73,38,149,20,
156,32,33,78,83,196,208,161,80,47,161,95,
32,155,99,182,102,227,233,246,137,108,133,207,
42,1,33,21,130,165,61,151,121,83,67,189,
244,79,217,94,255,31,121,73,150,184,155,10,
164,169,125,210,205,211,64,209,248,252,223,67,
210,117,242,173,85,132,72,24,108,223,97,149,
46,213,210,47,161,154,25,240,99,21,198,249,
94,127,1,61,185,6,210,50,138,174,19,141,
4,176,23,38,108,247,124,192,68,83,179,236,
145,48,78,73,195,172,185,76,105,226,54,155,
223,252,51,90,253,49,94,174,62,100,188,191,
205,198,127,220,231,75,246,35,152,156,230,71,
179,175,113,169,22,240,213,47,248,236,135,119,
88,94,251,128,245,71,211,212,234,191,140,93,
158,57,71,136,164,25,61,202,25,130,36,77,
98,98,127,72,228,14,144,85,13,55,72,32,
112,50,168,41,28,34,30,19,220,82,18,142,
104,208,131,19,190,201,198,94,68,73,47,28,
10,44,103,24,30,209,112,7,105,86,213,173,
125,246,85,200,107,181,153,159,107,37,49,93,
177,78,212,92,61,115,240,57,78,113,130,24,
63,74,71,6,91,66,156,28,163,250,20,96,
196,30,211,102,114,102,237,204,210,158,195,100,
217,36,222,255,30,211,234,58,143,254,98,153,
82,73,167,117,227,42,171,67,40,230,4,58,
32,82,25,4,200,82,74,107,48,142,20,152,
36,192,116,109,25,219,28,48,240,202,232,106,
128,44,39,204,54,30,178,210,188,73,222,234,
177,215,155,102,170,182,76,171,63,65,42,100,
174,207,124,64,195,84,25,122,21,94,159,84,
232,37,41,221,225,24,191,119,211,97,217,177,
24,51,117,54,239,121,120,171,58,201,165,199,
246,158,2,74,94,224,212,83,222,154,219,38,
18,130,117,119,133,205,149,101,242,124,142,98,
253,54,70,190,134,166,72,92,29,207,209,113,
116,22,119,157,39,43,138,19,193,162,17,241,
223,147,38,219,143,18,108,93,57,204,200,145,
36,152,40,25,48,52,89,123,142,50,2,65,
214,193,106,167,23,208,40,232,140,23,77,12,
77,70,51,10,212,39,255,46,127,245,224,223,
240,185,57,155,218,173,155,12,215,60,58,174,
197,107,209,93,146,155,243,172,180,26,76,149,
58,124,235,147,87,184,86,223,230,218,165,22,
131,90,25,182,82,156,245,128,36,18,200,114,
74,34,105,140,79,90,172,237,41,184,90,149,
210,19,154,145,148,74,26,150,46,248,200,55,
177,140,207,141,224,209,172,70,228,34,2,90,
130,19,212,236,126,111,11,213,40,96,228,107,
135,217,52,110,160,92,72,73,0,200,210,81,
138,232,103,173,40,32,203,132,106,197,33,151,
102,231,112,230,254,59,156,71,255,27,95,153,
27,242,125,69,57,151,57,86,213,237,195,94,
19,101,78,103,212,228,170,11,236,237,231,145,
85,29,145,94,67,81,84,116,53,163,157,79,
133,160,235,68,116,157,147,74,72,246,134,220,
44,214,201,27,26,183,190,241,223,35,171,58,
91,173,21,86,146,183,153,120,227,46,214,254,
42,195,141,117,166,151,214,160,148,39,184,85,
103,107,170,202,84,41,225,230,236,54,43,31,
255,159,12,211,255,130,124,117,225,244,254,31,
118,89,253,241,31,115,253,87,254,233,233,223,
57,93,204,98,157,52,142,208,115,101,60,57,
2,99,212,148,203,56,157,158,43,84,3,169,
191,117,42,144,43,105,202,137,254,201,158,35,
193,65,106,172,162,67,26,253,76,4,241,84,
103,155,246,244,171,63,151,74,98,162,100,50,
93,49,159,235,30,139,77,135,225,83,50,4,
61,47,162,187,217,199,214,21,234,5,157,90,
222,64,83,36,54,218,30,251,131,144,48,78,
153,140,34,46,221,176,216,95,177,8,122,30,
221,71,54,191,19,252,128,123,215,223,228,227,
190,65,107,56,198,75,179,239,162,14,167,73,
84,7,81,111,35,233,17,2,129,31,218,196,
177,78,165,144,213,132,57,65,1,83,119,169,
151,182,65,192,230,254,2,141,210,22,121,35,
228,118,41,135,41,234,248,154,198,116,169,77,
193,211,248,145,144,209,100,153,155,197,28,219,
94,64,113,108,64,187,223,64,164,10,150,84,
162,102,249,36,2,226,52,165,29,198,128,194,
180,109,162,73,112,173,96,227,213,83,190,215,
252,22,219,123,223,65,110,78,82,200,125,129,
202,196,27,84,114,26,121,83,185,184,162,120,
210,208,20,153,91,211,5,212,115,176,227,162,
165,49,102,24,236,246,159,175,232,44,73,5,
59,189,128,221,126,200,149,177,28,22,96,149,
166,208,154,47,243,65,103,23,72,185,94,236,
115,101,253,199,12,114,54,106,32,152,47,237,
176,188,81,230,183,147,63,228,126,240,22,165,
229,251,84,154,45,238,229,223,228,165,220,22,
219,110,3,75,246,248,235,254,91,92,205,127,
130,183,169,33,61,37,245,49,10,4,162,156,
103,185,95,162,92,156,57,132,146,38,75,38,
171,173,167,59,191,214,168,32,235,80,40,214,
230,95,152,235,251,89,67,79,7,99,167,31,
80,242,98,138,249,6,225,236,127,195,71,15,
254,37,185,41,159,240,128,82,94,164,79,109,
224,254,248,208,211,25,10,229,27,228,170,71,
177,132,171,146,196,131,230,217,158,91,10,44,
237,58,188,60,83,68,30,197,8,114,181,121,
114,181,121,134,222,128,229,248,79,176,95,126,
200,245,212,71,124,210,197,127,208,164,254,195,
21,60,63,65,175,91,188,246,229,43,188,219,
255,19,118,22,175,211,152,251,117,146,208,101,
235,147,63,5,37,32,10,6,140,223,252,141,
35,193,209,109,98,149,198,8,221,62,70,161,
122,62,244,116,166,165,160,100,248,252,83,246,
250,137,158,0,63,67,158,138,157,103,232,239,
240,89,142,177,226,139,73,137,157,44,155,60,
108,94,12,13,112,195,132,181,150,199,122,219,
195,80,21,252,81,86,97,223,139,169,151,63,
207,15,239,62,226,246,47,190,70,255,143,255,
154,214,187,15,89,149,246,249,242,248,93,172,
177,175,49,28,255,144,94,103,154,56,191,5,
8,250,78,149,208,59,198,62,81,58,42,24,
13,34,139,233,250,18,142,95,192,241,74,76,
215,151,144,37,137,95,104,148,33,206,209,114,
77,230,89,68,90,117,168,175,55,185,124,249,
101,222,221,41,241,230,164,67,78,85,248,90,
53,165,55,46,88,11,60,26,234,56,215,156,
123,104,178,66,175,81,230,199,123,33,178,18,
176,220,119,120,169,217,34,188,62,203,150,23,
48,101,235,92,189,109,179,27,180,249,240,253,
239,82,225,117,64,66,29,25,172,207,173,40,
202,182,118,174,146,56,190,24,207,171,40,142,
195,29,203,123,14,55,70,1,252,210,228,23,
216,222,252,183,124,253,202,14,227,15,31,141,
122,171,250,244,223,139,145,94,154,64,221,10,
144,252,30,250,96,27,145,14,64,8,106,237,
37,6,146,205,32,201,51,107,110,144,147,92,
162,161,195,100,210,102,49,188,69,140,142,122,
6,197,219,216,152,201,216,152,73,90,43,163,
247,6,68,254,209,6,203,153,23,35,162,51,
84,249,156,128,220,139,19,6,159,121,237,139,
128,135,205,33,215,198,243,20,173,2,202,245,
127,206,254,202,59,132,233,22,43,75,255,19,
164,144,215,62,135,162,22,240,134,43,168,122,
9,211,158,192,176,235,24,249,202,153,130,176,
58,249,230,137,198,50,73,154,30,86,162,159,
15,151,166,68,113,138,161,157,156,99,205,42,
48,121,237,31,179,126,255,15,120,231,218,3,
196,181,57,202,165,152,121,215,103,174,217,37,
13,99,218,111,63,228,246,151,225,94,49,96,
245,221,54,11,111,254,62,99,115,191,132,144,
99,204,252,24,154,117,228,29,12,218,247,217,
93,254,54,147,215,127,243,148,215,244,34,88,
13,246,135,33,73,42,80,71,85,180,185,208,
195,149,21,204,48,36,208,77,140,200,39,84,
117,212,36,38,149,21,36,145,34,144,32,112,
144,81,73,100,21,53,141,137,20,21,61,14,
9,52,19,51,244,240,12,27,59,112,112,36,
137,92,224,226,200,26,185,96,136,99,228,72,
101,149,244,140,52,201,250,160,205,254,207,153,
146,168,229,117,230,235,47,6,134,171,228,52,
106,121,253,204,68,156,243,227,25,28,42,137,
67,232,112,168,240,210,228,63,230,227,181,127,
203,205,235,99,44,127,152,144,10,153,237,143,
154,228,111,110,50,181,80,226,81,109,151,71,
173,26,126,100,81,43,236,82,204,29,67,46,
132,68,107,56,70,154,42,24,90,198,59,55,
112,43,76,84,179,180,215,43,121,139,188,170,
240,97,115,156,91,245,21,212,191,254,36,59,
23,11,211,168,74,76,24,103,251,176,36,4,
202,195,117,54,175,222,228,74,46,162,168,118,
48,238,111,17,190,245,42,107,251,243,24,126,
142,235,181,247,168,90,6,68,49,90,187,207,
149,114,1,41,12,145,223,187,207,220,107,87,
249,184,208,39,112,122,232,118,153,254,232,204,
169,47,66,112,95,68,56,22,45,149,190,247,
98,138,192,226,52,131,31,114,101,176,11,85,
108,117,134,176,214,39,237,85,16,155,45,34,
47,66,146,135,168,119,30,17,182,167,25,232,
42,19,209,125,246,162,28,19,118,128,18,116,
112,141,42,41,18,94,98,98,74,62,195,216,
192,196,167,68,143,45,109,150,217,232,209,9,
248,75,2,226,40,69,175,216,184,211,13,220,
180,131,177,182,73,137,107,217,97,214,213,67,
104,228,73,67,145,179,107,118,251,1,181,252,
79,167,170,212,13,63,251,30,6,73,42,184,
191,61,160,146,211,178,4,136,75,111,210,91,
252,46,177,104,82,172,123,160,125,147,106,73,
101,95,10,113,226,20,215,135,120,211,98,106,
252,191,198,174,76,93,224,112,74,8,33,200,
25,42,150,38,147,51,85,134,126,68,123,24,
33,203,18,181,130,201,204,116,233,100,51,163,
199,32,63,41,202,19,237,169,200,197,152,174,
23,243,129,166,98,206,141,225,245,4,165,233,
73,62,191,184,206,228,120,76,171,250,58,200,
25,20,116,86,199,180,198,229,175,225,117,118,
208,115,89,214,223,222,210,95,163,25,37,204,
82,131,106,94,99,179,147,193,144,210,185,251,
65,58,195,80,56,250,222,214,21,10,166,74,
232,36,124,126,233,61,90,149,105,182,11,37,
106,189,61,218,86,142,114,224,51,52,11,88,
161,75,160,26,168,105,148,197,66,252,62,178,
230,16,170,6,86,232,225,152,57,10,222,128,
94,174,76,117,176,207,110,105,130,122,119,135,
56,95,166,222,221,38,82,53,234,237,45,242,
138,2,154,197,230,248,213,83,207,213,203,21,
127,238,148,196,66,35,247,66,239,57,95,183,
25,94,148,242,230,9,198,217,78,47,98,188,
250,27,236,26,59,204,27,63,100,41,184,138,
212,124,192,92,241,61,182,39,127,147,87,38,
82,2,95,37,213,182,112,226,147,103,116,179,
117,153,169,218,50,130,44,40,221,29,214,9,
162,204,99,50,100,153,171,18,168,239,222,195,
156,24,227,238,163,58,159,147,20,36,145,32,
15,92,44,235,10,213,130,159,197,64,22,215,
73,175,92,194,80,84,156,64,99,38,92,69,
74,98,182,31,194,75,99,31,179,90,186,202,
101,13,60,85,1,93,69,32,35,247,134,56,
111,223,35,244,34,242,55,35,66,57,37,137,
92,118,122,230,97,89,196,115,43,138,225,5,
43,128,235,121,227,133,41,10,128,158,23,114,
144,221,157,47,222,228,254,246,59,76,213,202,
180,62,220,194,117,99,38,47,105,236,238,166,
20,148,61,10,198,193,231,58,116,218,33,31,
111,149,121,117,126,31,228,18,123,113,141,234,
240,39,108,251,117,164,226,4,245,254,58,247,
229,219,212,49,176,71,29,58,14,206,176,149,
83,73,231,38,209,21,133,170,173,49,56,6,
33,72,82,198,192,219,113,158,108,245,246,189,
152,187,155,3,220,48,161,227,68,220,156,44,
188,80,100,33,136,210,83,214,206,103,57,58,
78,196,70,115,128,97,106,152,197,47,81,48,
191,74,32,69,152,193,128,253,213,21,156,112,
9,212,20,89,74,81,117,151,237,165,63,193,
84,230,41,140,207,161,232,37,186,219,119,144,
85,137,98,245,37,118,151,191,13,72,200,24,
148,167,94,103,97,44,135,16,130,216,119,144,
228,132,241,98,158,75,213,20,85,134,112,16,
157,171,36,0,90,171,63,65,210,99,140,254,
171,120,237,37,180,133,44,81,193,249,104,156,
106,253,235,72,129,202,15,180,159,240,70,243,
91,92,53,255,156,157,214,60,154,90,63,215,
251,83,141,35,216,195,42,79,208,94,123,159,
242,244,203,200,146,196,171,151,74,135,107,26,
37,41,107,173,199,82,199,21,13,146,232,4,
237,248,113,167,188,96,170,204,217,30,247,29,
147,239,223,248,18,179,221,125,250,165,41,250,
165,41,8,135,244,206,224,137,202,48,177,246,
169,54,152,7,54,107,183,146,193,164,195,210,
4,132,14,139,133,6,232,57,22,11,99,24,
126,31,249,28,58,243,188,239,254,220,40,137,
70,193,224,114,227,197,7,244,21,89,226,234,
88,158,187,91,253,231,242,8,119,251,62,141,
82,149,46,95,230,202,205,21,248,32,235,18,
227,185,17,134,173,16,239,120,124,125,60,193,
145,139,108,236,153,60,242,116,118,220,20,89,
78,168,22,155,72,146,192,245,11,12,189,18,
229,252,30,5,171,75,179,59,195,175,206,119,
49,63,122,4,142,135,182,125,159,151,172,101,
62,118,175,51,111,174,147,223,235,240,185,27,
125,190,179,61,198,213,214,29,164,238,16,231,
202,117,180,97,66,24,171,224,7,132,97,202,
252,238,123,44,110,95,101,225,230,71,168,27,
219,240,214,43,172,154,55,240,146,10,97,162,
82,121,37,207,140,178,199,187,73,138,210,153,
131,249,113,54,142,145,91,102,138,194,25,17,
235,121,29,224,2,51,21,121,89,143,92,85,
35,0,118,213,254,169,90,135,131,225,247,50,
199,213,76,5,178,51,60,108,80,127,246,70,
191,224,231,3,142,223,103,176,175,103,120,191,
80,232,238,171,244,223,200,83,88,104,208,252,
254,58,141,48,68,142,19,118,150,218,52,94,
41,163,235,50,105,42,177,181,227,97,4,1,
139,75,10,245,27,62,66,72,244,186,17,178,
183,78,50,113,3,179,191,195,130,180,140,129,
32,196,64,33,70,33,65,150,36,74,37,13,
105,224,192,68,141,107,69,139,183,157,38,110,
251,8,91,172,136,132,158,227,158,253,142,126,
31,149,132,84,128,59,218,141,3,7,118,82,
235,176,247,193,227,115,118,145,241,248,181,173,
97,8,142,255,220,243,251,60,215,42,126,196,
230,82,135,208,143,49,115,58,249,188,65,190,
96,97,26,13,76,74,228,205,50,57,67,193,
84,101,34,189,131,97,87,113,187,27,224,251,
24,118,102,193,239,174,124,15,171,48,65,42,
60,20,85,197,31,116,232,45,125,147,193,222,
34,170,82,102,252,234,47,161,140,2,194,49,
16,14,59,164,79,32,60,179,138,147,88,197,
44,93,80,221,31,99,248,232,3,20,93,65,
83,26,104,70,38,92,149,177,175,241,227,221,
30,95,151,223,69,222,253,223,121,144,124,141,
82,242,230,57,71,96,64,124,64,77,143,141,
140,197,254,210,15,177,203,167,11,13,39,85,
144,36,159,253,3,101,17,14,179,140,166,227,
2,63,82,113,219,206,49,15,74,96,133,29,
6,65,140,213,219,132,156,125,116,246,162,115,
132,183,223,127,106,252,227,240,30,0,81,246,
121,229,246,22,106,18,177,121,6,83,172,151,
70,71,242,225,137,22,138,147,221,239,66,149,
217,17,89,179,6,231,194,123,171,158,215,25,
87,76,220,118,239,194,103,225,137,30,192,99,
251,69,2,38,149,144,173,158,255,169,231,85,
68,30,27,235,67,102,26,55,217,40,94,97,
193,124,155,192,151,24,118,2,242,173,93,212,
126,143,222,59,155,104,115,227,212,172,42,215,
47,107,244,21,133,149,129,207,70,211,70,184,
57,188,200,100,188,177,118,228,93,42,18,115,
182,137,144,178,182,100,98,148,6,123,213,90,
102,55,108,144,87,28,118,186,21,222,154,116,
17,45,19,113,101,134,245,110,131,107,99,219,
60,108,78,32,185,30,75,143,134,72,50,244,
243,30,185,253,69,164,249,27,116,63,82,152,
211,30,161,92,154,1,9,126,178,107,241,173,
253,43,20,86,250,84,237,47,243,104,113,25,
113,204,235,201,164,123,110,180,146,101,149,0,
0,32,0,73,68,65,84,236,104,186,114,23,
8,94,133,206,137,244,88,71,213,25,171,158,
239,14,30,208,222,150,146,225,83,44,238,11,
126,62,25,137,91,100,213,41,140,50,136,90,
139,5,190,179,187,199,91,147,21,170,245,102,
22,221,148,51,107,225,206,71,93,102,103,115,
12,7,14,67,95,69,37,160,39,198,168,111,
127,130,83,190,65,28,36,36,66,70,236,44,
34,1,197,36,179,54,21,32,144,109,230,230,
12,246,183,250,248,126,130,181,211,130,84,96,
92,153,161,48,185,137,162,171,24,249,236,153,
109,192,174,38,236,246,3,156,32,6,164,140,
74,68,150,40,82,96,108,124,138,102,223,103,
173,149,29,84,93,149,105,76,21,207,140,241,
216,23,172,48,125,252,218,181,96,8,185,232,
185,231,247,121,174,53,115,48,87,159,38,77,
83,122,29,151,78,107,200,226,214,0,153,148,
106,65,34,223,16,84,234,86,86,148,100,20,
200,163,162,213,42,72,146,116,8,83,70,81,
155,238,250,93,110,252,234,63,71,146,36,34,
111,128,216,143,185,116,251,239,147,171,157,22,
104,178,162,158,9,19,157,55,198,171,95,57,
27,38,45,252,87,252,96,103,1,67,190,135,
225,127,64,191,5,19,215,254,238,25,71,192,
66,207,149,143,9,246,148,56,236,97,87,39,
73,147,228,84,123,213,133,42,184,27,253,17,
44,40,101,124,82,199,218,164,150,43,38,246,
99,181,50,133,72,103,208,245,185,127,109,142,
183,62,121,155,119,174,125,17,145,238,34,14,
185,168,164,147,173,86,71,61,187,47,116,126,
225,176,106,124,32,233,72,34,133,220,233,191,
213,85,227,152,124,120,146,89,222,207,42,173,
47,92,153,29,31,171,90,127,242,222,154,154,
190,116,102,79,248,167,157,133,39,141,179,246,
139,93,5,103,103,72,239,241,88,216,51,204,
107,87,82,185,148,111,16,132,191,200,204,237,
117,154,239,46,225,186,41,234,39,107,20,199,
101,218,205,1,254,106,23,33,171,168,183,39,
208,94,187,76,127,104,211,85,36,130,84,99,
60,239,80,208,84,194,52,69,150,36,174,213,
53,162,237,61,116,63,96,45,152,230,146,190,
153,173,139,20,17,139,108,143,121,161,202,36,
29,228,213,109,226,55,110,162,12,70,222,225,
208,67,222,220,67,81,36,174,93,47,176,219,
94,99,216,120,25,81,173,83,249,201,119,208,
126,241,6,169,4,107,142,79,95,116,200,185,
121,198,170,191,75,100,78,227,135,67,56,150,
76,249,220,208,147,44,73,23,110,69,88,177,
245,167,66,51,207,10,179,28,190,136,84,197,
86,219,152,63,126,196,218,110,64,167,29,18,
4,233,33,108,180,182,234,16,161,34,80,8,
48,176,226,61,246,90,49,202,112,9,145,64,
132,138,46,78,6,180,10,121,141,104,24,227,
80,225,250,173,76,224,147,36,164,181,50,74,
32,51,222,144,216,221,94,68,207,213,15,131,
209,246,168,32,236,180,181,227,157,130,152,194,
56,101,101,207,101,170,108,98,27,202,115,207,
71,146,10,6,254,207,38,141,242,204,189,33,
203,84,106,121,42,181,60,11,76,224,14,92,
134,173,125,58,237,33,155,27,109,76,83,103,
108,162,72,165,90,56,1,219,219,186,194,229,
241,55,152,188,241,235,39,2,210,51,55,255,
254,211,99,102,73,132,72,18,20,253,233,105,
147,59,15,190,73,117,230,11,232,246,17,181,
138,170,91,84,103,191,1,124,131,180,185,68,
52,188,88,219,30,69,213,201,213,110,101,231,
179,185,68,113,234,218,105,24,199,84,143,226,
71,73,152,89,170,7,239,44,18,130,225,81,
194,71,18,122,228,164,152,28,30,150,42,211,
122,249,101,102,219,27,172,24,181,35,79,36,
242,30,163,9,249,116,35,231,13,81,211,8,
231,12,69,145,72,202,207,116,15,141,23,141,
11,43,137,23,49,206,75,54,121,150,177,221,
243,89,152,248,5,30,186,119,153,41,111,208,
239,134,220,95,49,208,59,15,113,221,152,36,
17,148,74,2,205,243,208,115,10,247,183,138,
196,137,198,165,177,135,200,131,41,44,20,190,
58,94,66,14,34,36,111,157,7,251,13,110,
197,9,150,228,147,140,148,131,44,165,148,180,
193,72,136,199,200,187,109,68,33,199,94,84,
99,92,218,102,189,211,96,90,201,246,174,231,
37,236,247,101,118,173,91,248,141,121,190,176,
253,103,132,21,29,145,203,206,72,89,87,233,
59,130,66,244,57,236,234,220,97,253,220,241,
241,220,138,98,188,100,80,180,46,214,145,231,
89,251,208,62,213,214,56,38,92,20,202,188,
236,6,172,45,246,73,83,65,48,74,83,60,
65,121,65,76,146,111,224,72,121,172,193,10,
49,50,81,16,2,58,42,49,1,6,198,177,
184,68,181,97,208,30,250,252,120,127,138,72,
236,49,149,119,97,188,66,34,12,114,121,25,
53,117,73,54,131,67,37,177,244,253,127,205,
194,151,254,209,169,84,80,175,187,133,52,154,
234,199,49,208,182,19,210,118,66,198,139,6,
179,53,251,185,226,21,103,181,56,252,121,26,
118,78,199,214,202,140,153,89,128,52,21,2,
223,57,157,109,226,134,9,123,131,128,99,6,
251,211,141,84,127,64,103,176,206,176,247,136,
40,191,132,18,213,208,197,36,73,24,96,151,
103,169,78,159,132,144,132,72,241,228,187,52,
215,119,185,116,227,247,207,188,167,102,22,24,
31,255,229,103,126,207,179,148,196,129,2,60,
218,176,70,102,129,75,50,138,44,81,174,150,
79,132,184,67,87,70,179,100,174,23,43,104,
86,158,56,17,72,3,7,187,191,201,221,250,
141,3,237,153,197,58,194,225,73,24,235,25,
135,107,218,72,231,196,40,196,207,48,53,247,
82,213,162,148,250,159,217,231,137,103,136,185,
62,45,78,39,234,160,148,126,21,243,210,59,
244,187,107,52,252,187,252,208,191,201,21,62,
4,96,108,220,66,188,113,13,71,148,121,125,
204,227,97,63,51,26,114,213,93,90,131,60,
111,59,6,183,199,99,26,139,155,92,27,44,
211,141,75,200,178,32,61,38,91,106,106,22,
129,146,62,88,66,177,214,136,223,184,201,208,
55,152,92,185,143,61,55,228,99,111,158,55,
184,71,161,164,242,227,237,25,126,227,214,125,
238,138,58,105,146,160,42,18,73,185,128,27,
39,252,112,183,143,188,126,147,241,155,191,204,
135,235,131,51,99,156,207,173,40,246,6,1,
99,5,227,84,74,226,89,67,85,36,114,70,
86,129,252,34,198,1,205,118,232,118,209,245,
37,180,187,155,164,199,56,74,30,239,167,23,
96,162,137,24,85,196,232,196,28,213,95,74,
120,24,104,68,56,228,200,225,48,55,151,163,
239,43,104,196,88,102,130,39,20,10,70,76,
154,151,16,94,151,20,11,83,145,73,211,35,
43,112,238,243,127,239,204,122,129,86,243,187,
232,105,31,189,240,79,206,205,138,106,246,3,
20,69,98,166,242,233,172,167,131,54,166,255,
57,13,89,146,176,243,103,87,213,182,6,33,
243,23,62,225,130,230,238,191,71,189,186,135,
52,145,53,18,20,56,68,172,145,250,18,254,
182,12,156,84,20,65,127,31,74,29,226,96,
72,228,15,208,204,194,179,205,119,232,19,210,
37,116,251,104,86,145,200,27,32,41,79,54,
132,108,82,148,112,72,26,12,179,107,67,7,
97,20,40,89,218,83,123,88,168,138,4,183,
110,145,220,191,71,209,237,209,183,75,153,213,
161,104,231,199,44,0,68,138,20,251,136,39,
84,117,199,170,206,205,165,31,179,50,126,149,
203,205,69,86,198,22,152,221,91,97,179,62,
131,126,70,220,71,83,100,100,41,203,62,76,
210,159,142,101,50,85,54,153,44,155,184,237,
207,110,63,110,182,189,172,178,254,5,120,246,
61,47,162,80,152,192,81,236,145,145,26,98,
225,208,98,146,26,219,108,110,56,212,204,85,
170,95,211,209,162,41,222,40,60,96,91,86,
120,93,145,200,41,123,44,214,74,4,190,78,
247,234,117,20,199,161,252,201,3,30,120,11,
92,183,78,51,249,166,100,68,144,162,152,67,
233,103,172,197,250,157,79,40,95,29,103,152,
171,144,207,7,92,137,238,34,172,49,116,111,
72,24,164,40,69,11,36,9,69,146,8,214,
166,184,116,227,31,18,37,226,220,68,152,231,
86,20,113,34,232,186,209,133,203,232,77,237,
197,41,138,3,11,109,111,235,155,252,210,204,
14,155,63,60,77,185,113,242,255,9,19,118,
159,65,105,150,170,223,100,115,195,61,84,41,
22,62,17,42,182,228,49,61,109,163,104,50,
219,171,30,6,18,229,112,143,141,164,204,235,
245,14,242,198,46,34,103,145,206,77,210,48,
53,238,137,35,69,161,232,167,133,188,219,89,
167,81,94,229,101,179,199,15,182,222,39,205,
221,58,247,125,186,78,244,169,21,197,102,199,
63,164,185,248,155,48,204,39,24,30,145,55,
192,237,173,147,175,94,70,209,45,154,171,127,
142,125,173,195,66,213,198,82,20,134,113,76,
221,208,233,69,49,31,175,228,152,186,250,219,
167,161,64,103,27,165,156,160,20,83,122,59,
119,168,207,253,194,51,61,159,85,153,56,194,
241,115,101,68,18,97,228,159,236,2,25,192,
148,176,217,88,243,17,102,25,43,117,152,26,
203,81,178,46,238,105,111,155,21,38,215,151,
232,219,37,228,230,143,72,199,223,122,114,65,
158,36,63,81,73,0,4,70,158,15,94,202,
60,167,247,171,89,79,242,238,168,7,70,235,
152,91,87,207,235,76,150,205,195,190,39,66,
64,207,139,216,31,132,116,130,11,167,62,252,
220,193,77,153,193,27,178,213,125,113,222,203,
192,143,41,87,45,90,165,203,140,149,150,232,
247,66,166,88,164,79,149,125,166,168,187,91,
244,238,236,18,183,58,204,188,252,57,196,157,
187,220,124,117,26,165,94,68,89,222,224,70,
163,140,188,252,16,76,3,199,147,16,182,201,
165,36,99,153,93,139,103,153,144,119,208,229,
204,27,47,165,155,36,138,150,197,153,162,20,
105,212,26,90,108,181,16,174,199,246,166,135,
170,75,108,52,13,198,205,7,44,47,15,185,
246,249,28,164,130,119,59,3,20,105,10,73,
146,9,227,243,229,242,115,3,114,186,42,211,
40,62,61,70,17,196,89,31,230,174,251,98,
48,116,69,146,208,71,120,162,170,238,50,124,
119,5,215,121,154,219,40,81,42,233,88,122,
66,165,114,250,153,139,182,224,165,235,121,114,
229,28,170,101,113,251,86,145,98,189,64,67,
238,82,179,143,166,74,114,125,136,227,140,223,
77,30,60,1,14,25,178,223,252,67,174,109,
223,199,189,179,130,72,90,79,172,179,120,82,
122,231,147,134,31,37,52,251,254,223,24,37,
97,168,50,179,231,20,84,249,131,38,219,59,
255,10,81,253,35,54,151,254,53,251,43,111,
51,220,93,69,54,82,174,21,108,102,108,131,
155,197,28,21,93,101,199,11,208,146,217,51,
5,105,18,123,216,134,194,23,39,138,12,229,
239,176,179,244,167,63,213,119,234,110,220,1,
96,178,100,50,94,202,60,240,133,70,142,90,
94,127,166,117,159,174,152,236,95,189,205,151,
238,255,21,233,196,23,63,147,170,109,67,149,
185,49,153,103,97,44,119,162,57,150,36,101,
112,242,213,241,28,183,166,11,79,111,232,116,
129,81,47,232,204,213,63,91,78,171,157,158,
207,242,158,243,66,239,57,24,181,17,150,202,
127,7,190,252,5,166,102,178,56,82,145,246,
161,1,219,23,5,150,151,186,180,191,251,125,
6,253,144,205,31,173,130,162,144,220,186,130,
188,188,137,36,4,242,110,155,92,175,133,19,
107,88,138,207,254,228,13,228,178,133,174,132,
248,178,205,35,119,142,173,173,0,57,12,113,
191,179,132,250,104,249,80,99,151,6,219,224,
12,72,82,193,245,171,69,60,185,66,65,25,
146,47,106,136,158,139,114,231,33,151,108,131,
116,68,15,99,27,202,185,49,154,231,94,217,
169,99,100,125,231,121,28,247,183,7,47,180,
134,226,184,197,25,7,30,149,220,0,225,60,
189,242,91,0,65,148,144,26,50,221,238,105,
108,220,119,83,150,154,42,254,204,53,138,21,
139,180,82,162,250,229,24,251,221,31,35,185,
163,9,212,53,210,169,6,3,36,150,187,62,
34,58,173,112,226,192,165,179,245,62,78,112,
143,249,219,33,206,159,14,72,174,93,129,36,
193,15,207,47,234,81,229,79,167,183,215,90,
222,207,60,54,17,133,49,126,16,97,153,6,
170,246,233,237,15,77,145,184,49,153,7,247,
244,250,12,118,63,97,175,245,77,190,244,37,
104,24,69,126,164,111,82,210,154,248,249,144,
170,174,157,50,36,84,73,34,42,220,103,243,
97,202,248,236,55,80,141,35,28,63,73,60,
82,41,165,162,107,124,237,106,192,95,62,188,
71,18,254,202,153,94,225,139,24,189,189,143,
177,43,179,232,185,18,245,188,206,108,181,68,
48,124,66,203,215,56,34,141,163,145,229,158,
49,208,38,81,128,174,170,204,150,84,118,175,
92,167,208,219,102,96,21,94,156,41,127,206,
152,40,153,79,245,122,52,89,230,218,68,158,
181,118,120,110,247,180,167,141,74,78,227,242,
11,44,166,91,217,119,145,165,140,149,247,188,
244,253,141,142,199,86,231,197,25,89,51,85,
11,211,178,88,108,58,244,189,152,98,233,50,
205,238,239,50,188,34,49,87,184,79,115,185,
75,205,223,34,240,3,246,203,115,116,165,34,
241,254,71,20,139,5,134,195,152,220,119,239,
82,170,154,104,179,13,68,169,0,105,138,236,
5,24,190,67,207,46,179,150,191,198,245,217,
85,66,253,117,62,220,185,206,231,55,255,29,
66,19,44,6,87,176,21,151,57,255,193,209,
153,220,223,98,109,211,99,108,204,164,35,234,
212,213,14,8,152,158,182,144,37,137,120,106,
140,84,128,34,219,135,8,204,88,201,96,189,
229,189,88,69,113,145,140,167,253,97,72,95,
188,248,182,156,57,125,68,160,215,91,103,190,
18,95,200,42,211,137,232,123,54,82,94,224,
6,39,175,87,117,29,119,236,85,228,100,192,
53,239,67,210,166,142,209,85,105,231,190,136,
59,247,10,254,134,79,58,55,73,251,210,56,
32,241,189,181,30,233,226,101,230,94,249,157,
19,247,113,58,171,236,245,254,136,194,252,128,
25,93,229,70,179,195,74,55,160,47,213,169,
205,125,131,181,141,225,11,245,40,134,65,76,
215,251,217,100,58,9,33,104,237,245,233,111,
110,179,211,219,71,8,193,196,116,149,43,215,
39,62,157,151,40,75,220,152,40,96,106,202,
137,46,34,34,77,24,238,175,177,183,247,45,
102,94,235,80,55,10,40,146,196,151,199,179,
202,232,75,5,147,118,24,29,118,29,60,24,
47,151,243,180,195,46,137,120,143,245,71,219,
76,78,253,46,102,49,75,245,12,146,85,68,
42,24,70,9,213,40,226,234,100,192,230,198,
219,140,95,254,213,103,122,102,73,86,8,157,
238,33,123,172,72,5,170,105,163,60,214,67,
97,238,141,127,116,241,121,77,98,214,222,249,
35,38,95,254,117,204,66,3,205,202,35,210,
4,73,81,209,70,196,129,245,27,69,38,255,
244,223,241,221,185,47,160,42,233,147,22,41,
51,253,159,208,114,245,105,99,181,229,178,213,
245,209,148,172,69,166,34,75,168,178,132,161,
101,36,121,7,86,168,4,204,55,108,130,56,
121,102,195,176,104,169,92,25,203,241,162,252,
163,174,27,29,210,6,237,244,124,74,150,198,
120,201,200,226,65,18,164,34,83,36,47,138,
90,232,192,104,30,47,26,196,114,214,191,250,
209,238,144,219,211,69,172,242,77,226,194,191,
224,142,244,199,76,55,62,194,248,240,35,186,
29,129,190,252,231,4,213,151,217,217,217,65,
213,52,76,195,96,125,101,200,230,154,195,75,
146,132,97,42,164,243,83,72,174,143,182,182,
131,253,242,20,195,78,155,63,92,154,162,160,
39,124,253,210,58,194,156,70,249,112,137,134,
220,162,164,244,143,25,197,2,73,202,250,193,
219,182,194,193,106,44,45,15,153,156,178,176,
76,5,20,133,165,190,135,38,29,21,150,78,
150,50,195,127,187,227,159,232,108,250,92,138,
162,104,169,231,54,146,57,82,38,252,84,44,
158,188,53,226,215,143,66,212,97,138,19,169,
132,25,161,111,70,9,125,136,173,197,40,35,
86,126,1,116,251,49,179,181,29,150,204,91,
132,124,130,197,0,203,148,233,141,125,129,198,
240,3,220,129,199,98,148,82,174,232,92,190,
90,166,182,126,135,251,151,94,231,195,198,36,
179,179,58,127,213,236,146,174,204,80,176,127,
25,115,122,226,20,215,79,107,255,219,124,229,
182,79,195,75,241,239,175,177,252,206,22,118,
65,103,75,52,40,34,63,145,52,238,211,164,
230,117,156,232,5,0,136,207,174,32,246,118,
251,108,172,236,225,185,33,121,201,69,136,44,
147,169,189,223,231,202,181,9,158,245,196,203,
146,196,245,137,252,153,105,194,107,247,255,0,
119,216,102,234,205,62,175,87,74,168,143,9,
188,130,166,144,87,149,83,114,48,167,42,124,
165,81,226,237,189,30,92,221,102,187,245,191,
160,110,93,65,248,6,233,244,58,170,4,134,
34,145,54,42,92,93,221,102,49,94,126,118,
15,200,42,28,122,0,7,117,21,177,239,16,
37,17,154,249,236,153,72,105,146,208,92,250,
11,22,190,250,95,34,146,39,83,74,188,51,
118,147,237,205,54,151,102,159,64,179,113,48,
41,207,219,107,33,73,57,43,206,233,133,9,
87,199,115,199,192,93,184,58,150,231,227,205,
254,133,201,41,115,134,194,181,241,252,19,145,
137,103,29,222,99,15,219,243,34,122,94,132,
34,103,9,53,126,111,72,168,171,47,236,243,
114,134,194,120,73,59,133,166,124,188,209,103,
188,100,48,94,52,41,206,254,67,90,238,175,
18,221,252,191,152,92,254,1,227,152,4,206,
50,197,249,121,194,240,200,216,75,83,193,198,
98,143,133,175,94,70,20,115,8,219,68,84,
138,200,142,79,59,116,24,132,21,26,213,7,
228,141,42,82,39,70,146,68,166,36,14,12,
2,178,30,241,227,99,38,195,97,68,167,23,
18,89,21,174,153,15,89,238,132,184,78,194,
173,87,202,8,93,193,221,76,25,171,221,58,
21,35,26,120,241,197,250,81,92,100,92,132,
124,78,85,228,99,157,85,94,204,40,219,26,
249,17,45,135,93,157,165,213,77,169,17,162,
159,65,228,23,163,16,161,143,122,98,168,24,
105,132,33,66,94,175,44,179,102,76,225,71,
26,161,164,51,27,124,76,243,88,101,108,191,
23,177,185,214,71,183,34,174,78,45,241,190,
62,197,255,123,79,167,186,127,131,75,55,254,
65,214,123,224,88,85,246,209,73,81,176,16,
236,125,247,99,182,54,124,144,4,99,179,53,
144,175,63,53,163,226,211,96,188,97,44,64,
255,217,40,136,51,175,73,33,21,41,242,51,
176,197,74,18,92,25,207,157,217,140,40,14,
92,210,194,6,98,170,142,38,73,231,218,28,
231,201,152,162,166,242,43,19,85,254,106,183,
139,91,11,17,181,123,144,102,150,214,205,98,
14,125,4,247,89,65,68,46,191,79,232,116,
208,115,149,231,154,39,213,204,145,132,254,168,
149,106,157,103,209,154,161,211,38,136,86,241,
7,215,48,236,243,91,128,38,169,160,244,240,
30,106,126,26,82,59,75,145,61,111,4,131,
147,133,121,47,112,4,103,4,64,85,69,226,
218,68,158,187,155,131,167,114,159,153,154,194,
141,137,194,83,13,206,103,223,171,231,207,91,
223,139,33,121,177,66,233,82,213,230,241,110,
154,144,101,134,109,118,124,182,187,1,99,69,
131,233,74,13,253,218,127,203,102,144,80,216,
253,22,90,190,128,64,194,87,109,44,186,39,
12,167,212,212,25,36,6,178,42,241,200,77,
104,186,54,187,237,10,151,103,238,82,53,52,
76,69,70,140,148,93,124,107,1,198,170,200,
171,91,164,133,60,202,242,6,210,192,165,209,
48,89,90,116,209,106,41,177,44,24,27,183,
152,120,121,2,201,80,8,119,246,73,7,87,
48,198,78,238,179,254,99,74,34,51,184,159,
99,184,97,242,212,244,56,237,5,111,0,67,
149,79,20,180,233,86,145,117,103,129,137,235,
53,114,182,122,6,182,150,160,17,34,35,40,
89,33,95,120,57,71,49,39,48,212,148,107,
197,93,46,229,118,121,163,186,140,239,158,196,
229,84,85,162,215,13,209,107,121,246,43,37,
202,82,155,177,254,44,179,183,126,15,89,61,
31,175,53,148,9,54,84,5,163,108,35,16,
52,26,38,205,252,52,118,249,234,19,201,250,
114,134,138,125,12,71,77,210,44,155,108,163,
253,228,248,195,103,153,229,254,240,254,54,15,
239,110,158,169,36,36,89,162,90,207,115,253,
246,52,242,51,198,90,46,215,115,84,206,170,
177,17,130,157,165,255,200,149,5,137,151,198,
246,233,134,9,223,91,47,176,218,110,176,213,
187,184,48,87,37,8,143,117,148,67,206,98,
24,115,163,130,35,210,20,116,149,153,43,130,
237,181,63,33,77,158,31,202,83,116,19,179,
212,32,114,7,36,225,197,49,112,179,216,96,
254,214,63,33,63,162,159,63,111,56,65,194,
15,195,34,82,46,255,244,94,236,63,37,37,
145,121,26,103,111,78,91,87,184,57,149,127,
34,156,106,235,10,55,39,243,159,58,137,227,
231,101,148,109,237,132,145,151,158,33,19,83,
33,216,233,249,124,180,209,39,17,32,42,47,
83,170,26,164,200,168,132,180,229,73,214,185,
193,58,215,89,231,6,189,94,136,255,160,73,
249,195,59,228,246,155,108,136,62,129,181,141,
162,100,222,219,148,101,100,36,128,155,251,136,
156,133,152,168,33,100,137,116,188,134,188,189,
135,52,200,50,159,76,67,161,79,25,181,249,
33,81,146,146,155,44,33,93,157,38,153,155,
98,125,162,129,38,166,207,52,88,231,235,246,
9,198,136,231,230,122,90,92,234,60,49,149,
45,117,91,112,81,98,177,167,124,190,161,202,
204,23,114,196,125,247,4,87,139,174,222,98,
213,126,192,228,116,159,197,135,131,115,239,160,
171,242,41,201,90,48,98,246,123,18,206,177,
138,241,66,94,195,206,43,20,198,138,44,223,
190,196,114,223,133,213,155,148,103,222,196,237,
236,28,94,119,22,183,140,20,219,116,6,17,
87,71,245,1,149,154,206,178,168,18,13,90,
244,130,33,248,241,41,97,152,51,52,46,21,
76,220,253,30,65,44,177,189,189,141,187,217,
63,124,143,98,122,126,191,15,188,54,68,241,
11,153,223,167,93,123,125,86,101,97,178,66,
146,164,200,178,132,36,73,200,146,140,28,202,
25,197,129,4,224,128,227,156,135,171,64,228,
67,114,36,56,199,11,38,118,228,159,200,151,
247,135,109,250,173,247,72,205,14,115,55,251,
220,174,20,24,196,49,175,85,84,96,56,250,
58,207,148,4,101,99,135,100,172,6,134,54,
178,122,13,190,214,40,227,39,41,107,174,207,
166,27,144,34,8,210,20,91,81,50,247,98,
224,176,48,86,99,247,165,53,86,238,254,43,
138,214,91,168,74,225,137,252,81,39,60,129,
51,184,166,68,154,32,201,10,73,212,38,141,
35,52,43,127,98,223,100,74,228,252,218,151,
56,112,71,199,109,128,16,41,73,24,16,141,
200,8,31,173,117,184,221,251,132,225,188,5,
142,148,213,82,36,241,167,226,122,2,104,228,
13,198,138,198,17,194,45,50,107,120,152,184,
164,138,74,16,39,4,145,32,76,82,226,99,
197,121,66,145,113,219,46,145,59,64,245,135,
72,199,184,158,100,224,138,37,88,105,185,39,
60,15,41,137,168,229,20,198,45,133,184,239,
240,164,39,254,180,188,103,178,23,131,243,4,
185,115,81,94,172,167,92,43,1,99,249,60,
110,123,64,28,184,200,178,138,47,148,115,207,
192,246,238,128,154,212,32,23,238,163,217,58,
82,43,59,11,87,210,123,71,71,143,60,81,
237,6,190,187,75,94,3,223,54,73,71,80,
154,36,122,204,216,6,115,150,142,122,231,33,
82,187,71,58,123,20,19,148,91,93,228,189,
206,145,204,211,101,222,122,41,70,83,114,8,
73,176,185,220,162,162,69,136,217,73,118,252,
50,166,126,249,76,100,36,15,92,203,103,61,
128,142,20,197,115,112,61,117,129,52,80,201,
27,42,141,51,10,239,226,68,192,224,162,153,
36,231,127,190,161,201,220,154,42,28,210,114,
31,231,106,177,171,147,172,44,191,207,229,153,
125,228,229,33,73,124,182,48,212,117,153,237,
29,15,219,84,40,149,51,188,70,145,4,187,
219,3,252,72,69,37,34,103,171,92,191,81,
32,77,5,195,171,243,180,165,108,129,244,124,
229,20,143,140,16,226,196,207,68,154,176,215,
252,255,248,188,20,225,239,101,105,179,74,41,
79,32,205,97,149,26,56,29,13,148,147,207,
214,80,218,76,214,138,72,200,4,206,62,251,
76,227,170,14,226,216,60,200,133,194,9,111,
227,196,51,236,58,160,215,158,123,126,47,122,
173,122,6,94,41,57,242,137,231,61,223,106,
136,33,116,97,84,153,109,235,10,179,211,69,
36,73,208,217,248,0,63,220,34,102,143,208,
220,194,153,29,227,86,77,231,141,241,194,33,
132,116,238,147,118,250,96,234,72,125,7,121,
109,39,179,166,188,144,228,122,86,11,176,209,
173,113,169,188,79,213,136,168,26,26,215,10,
54,155,94,192,48,74,176,21,5,121,175,131,
212,115,208,146,132,133,162,133,119,115,31,101,
251,47,145,149,223,126,46,238,160,225,222,18,
187,139,63,100,238,141,223,29,201,231,33,178,
162,18,251,14,154,153,39,9,253,195,251,31,
111,92,117,240,125,228,13,71,113,144,44,152,
29,135,62,178,36,51,244,2,30,52,187,72,
229,155,92,157,185,74,34,43,35,162,193,240,
233,156,68,34,65,114,59,8,205,58,164,254,
144,36,152,159,175,156,9,223,153,134,126,234,
189,132,32,83,24,73,138,161,41,168,178,68,
168,91,168,86,30,89,57,189,78,175,142,65,
207,141,104,59,33,182,174,82,84,19,52,57,
61,84,156,79,27,159,134,247,76,141,83,214,
130,222,19,22,236,130,252,77,79,185,118,162,
108,82,30,25,202,135,235,27,43,224,158,109,
204,68,134,134,94,154,64,238,150,80,84,25,
133,132,20,25,249,24,62,111,49,68,149,124,
122,234,4,246,116,202,125,89,230,160,161,180,
144,125,154,110,132,8,187,164,121,11,49,63,
133,40,142,98,68,174,143,188,182,115,250,204,
170,114,198,83,103,41,24,186,76,122,105,146,
206,120,141,206,15,167,153,185,121,235,212,154,
109,118,188,81,242,130,140,80,237,231,143,81,
28,199,180,250,94,76,156,138,23,214,80,228,
44,120,226,73,189,27,172,218,111,241,32,220,
33,111,239,208,235,135,143,25,153,50,17,26,
173,150,71,189,174,51,49,97,225,12,99,186,
221,144,161,19,227,57,49,141,177,34,249,98,
129,106,94,128,4,105,181,204,70,46,71,63,
236,143,98,29,173,51,112,241,147,39,107,111,
237,123,204,191,188,69,109,24,209,73,82,20,
89,162,111,85,48,115,95,192,15,19,146,84,
61,133,171,23,115,57,92,47,66,205,143,209,
211,108,186,93,255,20,156,180,211,11,184,58,
118,122,169,50,74,241,244,51,139,81,188,232,
145,229,203,167,172,223,255,191,177,46,127,194,
203,117,11,93,150,136,133,66,148,238,211,28,
218,44,119,202,92,206,239,195,57,138,66,222,
104,162,60,88,59,45,15,243,71,251,240,122,
121,131,59,123,243,220,154,220,64,150,4,57,
85,225,122,225,232,247,202,221,37,72,5,97,
187,199,123,170,198,27,150,206,140,182,198,127,
104,253,152,226,196,181,79,253,126,249,198,2,
65,184,199,195,191,252,95,153,127,235,31,96,
20,50,44,56,242,135,104,118,1,33,146,51,
247,82,146,130,114,14,197,146,98,88,108,183,
98,246,154,61,126,219,28,208,139,2,92,227,
25,206,156,164,100,74,226,49,193,31,167,233,
133,123,163,72,82,230,221,95,52,249,66,26,
193,51,7,20,62,73,232,147,38,63,221,230,
90,186,42,163,42,210,211,59,14,2,150,174,
16,198,233,51,87,153,235,170,204,84,217,60,
51,46,113,174,173,36,196,8,106,139,145,100,
9,133,100,68,29,20,18,145,205,79,181,170,
131,41,8,52,155,255,39,29,35,23,52,143,
254,222,138,136,221,28,235,101,131,250,88,25,
59,63,202,56,11,35,148,123,203,16,68,103,
198,140,212,156,74,172,25,204,124,161,70,103,
102,156,239,255,200,166,210,56,155,24,179,217,
15,50,99,32,126,65,89,79,143,111,158,241,
162,113,6,134,249,252,27,162,81,48,78,5,
122,227,84,48,12,226,195,252,104,179,56,206,
202,226,2,223,120,101,153,143,190,191,115,98,
209,35,20,108,60,144,4,81,152,242,225,251,
157,19,27,200,208,21,106,227,5,86,220,41,
118,6,49,168,26,78,125,154,102,148,9,160,
207,87,11,188,187,246,116,122,140,48,106,178,
224,56,236,191,253,136,141,77,151,114,69,103,
37,169,97,85,166,216,223,220,58,51,224,182,
216,31,205,217,224,252,194,189,246,48,196,59,
86,17,155,193,41,233,185,109,65,255,115,24,
245,130,78,78,19,60,250,225,191,68,127,189,
201,68,201,98,82,83,80,222,185,79,154,183,
80,38,170,204,167,67,86,246,84,148,206,14,
201,194,204,153,158,132,178,180,121,182,64,189,
191,140,188,182,77,252,242,21,80,85,110,79,
110,240,193,230,28,175,207,172,156,188,238,209,
58,162,82,32,186,50,139,245,112,21,26,53,
12,199,67,18,130,168,190,200,214,195,127,199,
212,25,236,177,23,29,181,233,47,82,157,252,
194,161,119,240,164,8,83,24,167,108,116,124,
218,195,128,133,177,28,133,51,228,176,31,37,
44,173,181,73,147,148,238,141,219,132,154,241,
66,214,195,11,83,52,75,230,111,210,200,233,
42,189,115,82,199,21,73,162,86,52,104,20,
12,114,134,130,16,89,26,240,89,233,178,99,
69,3,97,26,167,234,67,102,107,214,153,65,
248,228,9,202,41,244,227,204,82,79,123,244,
119,157,17,68,39,112,177,176,113,209,52,153,
203,95,188,204,189,248,37,46,205,238,34,239,
234,52,151,95,99,124,254,3,144,192,212,93,
134,18,180,212,1,149,216,62,46,124,72,22,
166,144,91,125,228,245,157,19,168,177,176,77,
238,26,211,44,123,55,145,6,121,164,247,114,
140,95,254,10,145,219,63,195,232,143,206,84,
174,71,59,35,24,28,97,151,94,251,136,10,
216,107,103,216,114,236,131,215,29,105,47,7,
105,68,85,44,121,93,136,3,198,243,26,146,
215,38,77,98,226,208,195,239,103,120,161,231,
14,178,123,3,248,221,236,51,68,154,221,55,
137,50,119,217,107,103,82,51,242,144,130,236,
225,165,160,15,145,139,34,195,152,230,144,198,
33,73,28,226,247,246,16,34,101,121,171,197,
250,214,206,8,9,235,17,121,3,80,166,73,
45,131,90,221,56,227,72,10,16,208,235,159,
158,136,160,118,141,29,175,76,161,245,1,213,
193,61,138,251,31,80,243,91,120,94,141,217,
156,201,182,23,146,164,1,113,224,34,210,4,
191,183,55,42,136,58,120,30,65,228,13,81,
69,157,69,219,102,236,90,13,211,80,40,141,
231,232,37,147,136,36,98,224,184,163,247,76,
51,60,217,239,29,225,159,7,180,207,79,152,
247,229,61,151,192,233,17,185,125,118,122,1,
119,30,172,225,185,238,209,92,138,131,107,71,
248,100,48,60,127,222,211,248,228,188,135,238,
17,147,169,223,27,113,7,137,99,107,20,61,
251,179,139,244,196,158,33,24,28,102,230,40,
65,143,169,188,76,111,251,35,180,87,154,228,
12,133,40,21,124,228,248,48,86,102,235,157,
117,164,15,31,161,220,91,102,161,247,19,146,
203,211,103,27,39,94,240,196,236,149,116,122,
12,242,54,194,212,145,36,193,171,83,107,188,
183,57,70,60,202,14,144,247,58,8,77,229,
237,202,203,8,219,128,48,230,55,58,61,38,
31,173,145,234,26,74,37,36,28,251,9,187,
139,223,33,28,29,42,191,191,79,18,122,135,
251,64,164,9,73,232,17,12,179,121,15,156,
30,161,211,67,8,129,223,111,29,198,25,210,
56,36,141,35,146,208,199,27,116,216,237,7,
12,251,125,134,205,53,2,167,203,250,242,35,
238,220,123,200,254,254,46,105,119,13,167,221,
196,239,239,51,220,93,33,24,180,113,218,219,
184,237,45,186,237,22,222,246,34,211,149,152,
217,221,79,208,219,139,224,236,103,115,237,247,
70,107,228,157,156,247,96,120,20,119,138,131,
108,175,248,221,209,158,9,144,188,46,59,61,
159,208,237,19,58,189,209,123,182,136,3,15,
145,166,135,251,253,32,147,235,0,102,57,40,
24,12,134,157,236,108,136,131,107,195,195,179,
193,168,209,212,225,252,12,187,68,190,147,205,
79,111,47,243,46,226,232,112,46,227,192,37,
24,180,143,206,181,239,28,206,123,124,214,188,
15,50,79,63,116,251,135,202,248,224,217,45,
141,163,253,30,7,72,94,59,107,10,85,16,
92,45,103,232,135,26,245,71,159,145,50,161,
57,204,150,84,164,131,253,142,96,34,151,82,
83,3,46,55,108,38,205,16,41,114,179,118,
202,242,144,146,46,30,123,246,108,31,148,108,
21,45,113,30,155,119,127,116,54,90,168,146,
160,168,110,225,142,250,162,167,66,34,29,197,
137,163,40,197,93,223,103,242,74,192,31,237,
84,177,10,235,20,167,31,209,222,190,13,66,
34,111,245,24,120,89,188,45,95,82,112,250,
217,254,151,186,67,194,239,223,67,90,217,198,
117,98,186,157,8,161,106,236,206,206,242,103,
233,151,104,74,191,79,125,234,183,24,191,252,
119,40,213,94,25,237,201,8,191,191,127,56,
239,131,238,62,171,45,239,244,158,57,225,81,
24,133,236,101,224,36,30,119,240,189,172,100,
28,250,128,208,115,135,188,243,194,42,163,41,
50,211,181,28,138,156,63,196,107,213,81,133,
107,162,216,71,89,23,102,249,244,125,143,127,
175,89,8,35,195,175,15,254,157,40,89,228,
142,185,119,74,169,193,192,143,241,208,81,164,
236,243,14,90,82,26,70,141,158,98,211,235,
61,158,218,37,81,46,235,216,57,133,208,79,
217,111,29,89,6,194,170,80,46,24,88,131,
7,236,140,170,181,167,166,44,234,117,135,15,
147,42,143,186,62,134,238,32,71,121,212,145,
139,127,128,217,202,234,17,126,171,89,121,114,
149,57,122,238,95,147,220,152,227,170,27,176,
19,231,49,202,95,1,69,195,139,149,163,236,
20,205,134,131,36,31,179,120,198,92,171,167,
230,125,24,196,60,236,42,68,177,32,74,92,
48,75,35,105,41,31,93,171,30,173,17,199,
170,144,15,231,61,28,158,61,239,250,49,203,
228,224,190,161,115,206,26,157,241,236,145,123,
246,181,170,121,236,121,10,168,82,130,26,123,
204,205,76,96,90,26,157,65,139,68,146,185,
61,149,99,92,83,9,87,54,25,214,202,204,
126,126,26,121,127,100,148,12,92,180,239,253,
4,97,91,8,9,208,84,36,63,68,40,89,
11,199,83,214,28,74,134,251,86,75,153,34,
25,141,123,61,135,78,20,243,193,94,133,74,
201,103,204,212,177,234,101,22,123,10,137,39,
209,143,18,234,18,152,187,153,144,146,75,121,
110,154,6,134,13,159,4,63,194,237,85,208,
237,215,48,139,245,67,235,63,49,171,72,178,
130,162,91,89,239,111,192,24,237,197,108,122,
106,199,166,245,0,231,215,72,245,226,136,202,
121,180,70,142,0,106,112,80,142,96,148,144,
243,38,166,153,64,177,142,102,229,49,10,217,
156,186,93,159,118,92,230,242,213,113,182,180,
152,208,204,29,53,11,146,228,236,75,179,178,
175,131,115,125,184,46,149,163,88,145,89,6,
73,249,255,217,123,179,231,74,178,251,206,239,
147,123,222,204,187,47,216,151,42,20,170,186,
150,222,73,54,73,81,92,52,34,135,26,205,
200,146,173,177,60,178,60,139,60,225,136,121,
177,223,253,111,56,194,79,158,8,217,15,19,
30,143,253,224,152,69,17,214,50,163,141,20,
69,13,69,118,55,187,187,246,42,20,246,237,
238,55,247,204,227,135,11,220,2,10,219,5,
112,81,5,64,248,70,84,160,187,43,59,145,
121,206,201,243,59,191,237,251,5,85,65,168,
6,117,39,36,200,103,72,155,234,174,103,143,
188,246,174,103,223,166,111,223,153,95,48,210,
5,2,36,36,73,222,149,207,216,254,103,213,
180,81,183,204,90,156,27,0,0,32,0,73,
68,65,84,94,206,72,231,183,66,79,209,254,
215,26,86,239,59,211,237,92,207,208,110,143,
251,206,107,21,61,213,235,164,215,173,44,209,
150,81,217,126,246,17,37,33,22,163,221,16,
111,74,37,151,26,238,121,0,94,35,238,61,
251,206,251,142,0,41,203,228,241,170,78,193,
214,152,170,216,120,141,238,102,60,57,54,204,
80,148,224,6,49,121,171,184,207,179,167,144,
21,21,69,83,248,224,230,196,86,239,137,64,
136,12,113,210,245,88,98,189,128,174,169,221,
94,47,169,107,36,170,155,107,20,236,238,225,
86,150,37,252,134,75,169,179,78,217,206,176,
184,49,67,37,183,68,97,244,115,214,151,222,
193,176,170,196,177,198,245,116,170,219,56,152,
8,212,159,61,32,106,186,212,170,1,15,87,
189,46,101,199,59,37,190,152,186,193,179,165,
175,48,114,227,251,187,243,78,189,253,75,67,
79,23,186,77,163,24,60,106,4,68,113,188,
239,154,25,72,232,105,178,152,58,176,14,90,
57,69,19,141,42,75,140,236,67,54,184,186,
149,137,143,19,129,27,196,189,144,140,97,15,
179,222,49,240,95,161,201,40,148,13,102,166,
187,12,157,113,44,8,35,104,108,221,163,102,
94,103,218,123,192,226,234,203,74,156,149,101,
15,123,110,149,242,109,141,71,207,135,25,10,
103,24,191,241,107,123,66,106,47,54,93,70,
83,97,143,222,192,46,76,241,226,241,8,159,
100,55,144,167,239,176,185,60,65,190,52,75,
195,13,15,87,246,235,19,142,255,230,36,78,
7,129,91,195,105,244,68,69,223,26,175,116,
233,26,158,250,23,252,172,214,98,218,78,113,
79,128,253,55,247,247,41,255,137,145,154,237,
93,129,26,105,159,48,94,183,223,72,226,51,
229,125,236,107,195,148,50,27,116,252,16,47,
78,120,210,118,89,171,143,51,94,122,204,103,
141,24,67,145,217,244,5,127,190,88,224,219,
211,85,44,20,80,213,110,5,84,146,160,124,
241,140,217,119,110,160,62,90,96,114,168,192,
191,91,255,19,178,241,189,94,226,239,209,106,
155,142,31,115,107,36,125,108,250,124,219,80,
48,66,249,208,134,180,150,23,49,100,74,251,
254,247,40,136,200,100,83,76,191,248,132,167,
35,55,143,151,163,56,4,115,155,46,119,199,
50,72,23,187,90,117,87,14,225,36,210,169,
185,148,198,135,211,249,125,199,225,56,185,25,
77,145,217,34,184,166,237,71,140,229,77,30,
234,74,119,175,220,162,103,105,249,9,97,167,
129,81,236,134,86,23,164,155,196,229,235,20,
202,49,215,218,46,168,79,217,104,116,19,244,
149,241,79,88,93,248,136,15,103,63,166,108,
164,241,61,65,146,64,60,51,129,250,243,199,
164,109,149,184,104,160,168,18,155,111,223,228,
197,226,183,25,185,241,205,35,159,83,8,120,
178,222,57,52,159,115,234,160,100,218,80,41,
103,244,67,63,138,147,46,188,209,188,185,199,
0,5,113,66,109,71,51,200,78,254,120,35,
83,98,61,201,97,189,210,79,97,154,42,157,
96,171,49,5,149,116,49,141,143,142,75,138,
84,237,17,47,94,116,118,229,52,52,77,234,
70,125,218,42,227,226,6,19,183,126,171,231,
33,109,227,233,186,67,203,139,120,176,220,230,
217,186,67,156,8,100,85,103,106,246,119,89,
250,252,75,180,195,127,70,126,166,171,115,208,
112,194,11,243,113,229,44,141,201,98,138,140,
169,14,180,83,22,224,201,90,167,71,150,6,
224,59,85,146,117,11,91,85,184,237,251,200,
75,235,167,202,145,1,168,82,196,189,228,103,
100,234,75,252,209,147,97,126,184,222,228,167,
181,173,16,156,144,144,229,152,140,170,144,81,
85,254,98,190,200,59,67,45,202,186,198,74,
152,176,112,251,58,11,31,221,195,183,186,30,
152,250,240,5,120,62,138,227,114,107,186,198,
230,252,95,246,214,224,54,3,242,126,34,47,
253,174,237,195,208,116,35,252,87,186,139,19,
209,109,22,83,117,21,73,146,152,175,76,227,
15,40,71,1,208,241,35,230,171,14,23,29,
226,149,198,35,113,2,34,180,65,46,253,186,
19,242,96,185,141,174,202,200,178,76,24,70,
136,249,149,110,17,65,16,80,25,170,32,109,
245,30,13,199,79,201,36,235,172,135,10,195,
102,119,95,45,231,150,201,217,85,22,54,102,
81,181,14,27,126,192,159,172,214,8,2,129,
97,72,72,159,60,197,173,57,44,175,184,36,
66,144,27,178,120,82,53,40,142,127,212,215,
243,45,214,92,220,224,240,67,232,169,13,133,
169,31,126,11,67,147,25,203,31,159,104,77,
83,164,125,169,203,55,154,193,174,179,121,123,
199,198,35,73,50,126,146,165,60,213,117,255,
21,89,66,215,101,106,27,30,134,18,211,242,
85,130,16,214,23,107,24,4,164,112,73,241,
50,137,156,32,227,144,197,180,85,116,67,193,
121,145,101,104,250,59,251,158,234,118,110,254,
235,45,159,7,203,109,132,232,186,115,163,179,
223,39,149,29,222,181,80,206,51,100,73,98,
188,96,242,254,84,142,183,70,210,140,230,77,
166,74,41,238,140,101,6,106,44,252,40,225,
201,90,135,229,45,93,98,207,155,71,189,214,
226,94,198,70,118,125,146,225,226,137,239,237,
185,49,190,159,224,109,109,224,185,82,130,153,
121,200,114,245,165,108,170,216,122,149,91,89,
139,191,169,181,64,192,7,37,133,167,29,23,
39,142,249,207,213,38,107,65,196,220,236,47,
16,191,117,141,164,152,99,81,185,78,184,238,
112,167,81,39,22,243,123,60,187,32,74,78,
68,155,175,246,81,97,244,170,191,177,209,10,
168,215,58,228,11,221,48,206,248,198,60,122,
232,15,116,45,172,52,252,174,238,250,5,69,
228,187,123,170,17,165,55,236,34,173,54,252,
94,3,109,28,39,184,173,77,82,81,215,32,
103,50,105,82,169,151,251,163,38,199,76,223,
74,248,253,231,6,63,222,124,153,108,214,84,
31,77,241,25,78,55,144,58,227,68,173,49,
30,182,35,106,142,68,120,231,58,169,82,154,
233,73,27,36,112,98,9,159,4,209,103,15,
80,63,251,211,137,12,133,244,202,38,115,20,
198,242,230,129,236,141,7,255,63,169,93,247,
22,91,27,242,98,221,221,179,105,239,174,56,
184,129,249,246,36,182,165,114,237,90,26,9,
48,44,5,77,17,100,140,136,140,25,51,60,
244,242,52,183,192,77,22,184,201,10,215,88,
103,2,129,68,199,151,73,52,149,200,16,72,
251,212,133,239,39,16,212,246,35,230,54,247,
158,198,122,37,172,231,20,182,161,242,246,68,
134,241,66,170,71,219,190,211,27,60,137,235,
126,20,22,54,93,58,126,76,121,252,23,73,
54,109,158,187,30,47,202,5,208,250,8,225,
28,176,222,12,67,70,211,36,76,67,33,122,
239,38,43,170,130,151,196,220,27,94,97,163,
54,74,181,49,140,165,183,233,120,89,254,243,
26,60,90,29,227,78,217,227,71,155,13,70,
83,6,43,238,22,183,191,166,210,72,154,60,
180,198,248,56,247,109,138,99,9,43,155,50,
43,63,91,192,100,101,43,236,248,234,9,245,
248,39,214,163,14,185,150,161,144,210,118,215,
200,46,86,93,30,125,177,72,101,139,12,113,
165,48,70,168,14,190,54,250,201,90,135,39,
107,157,93,82,195,23,5,170,145,58,87,207,
227,133,73,175,242,106,173,233,51,59,59,132,
228,61,221,210,191,209,208,216,189,73,235,186,
140,90,76,163,9,125,207,170,146,165,152,200,
94,32,178,22,136,172,69,234,157,12,153,108,
130,70,12,158,79,202,82,200,101,53,172,124,
10,55,14,17,73,127,7,152,126,202,130,143,
157,163,144,37,137,187,19,89,12,93,71,214,
204,190,92,52,73,130,153,33,139,79,23,154,
125,209,97,27,154,188,163,75,180,43,241,185,
88,115,247,221,112,189,48,33,136,146,222,38,
87,158,252,14,63,120,252,136,177,153,20,173,
184,134,157,121,198,154,113,131,102,92,195,74,
154,60,125,214,161,213,10,169,83,161,198,48,
83,220,103,155,91,113,91,17,47,107,232,24,
25,25,173,54,185,239,169,248,32,102,204,253,
44,243,121,245,38,36,9,198,11,41,70,243,
230,161,52,32,165,180,206,98,205,59,80,249,
234,68,161,1,224,249,70,135,187,227,195,232,
43,183,168,24,143,153,176,12,146,76,106,255,
147,139,161,33,116,13,145,207,144,140,85,144,
87,171,72,141,118,151,48,45,159,65,122,190,
132,220,232,230,48,132,166,32,114,25,164,173,
78,224,144,144,98,35,193,243,50,52,134,76,
76,221,33,140,4,35,197,23,44,135,18,154,
44,241,172,237,210,222,186,254,65,203,33,78,
58,60,124,254,1,223,189,177,140,58,54,196,
88,171,198,139,159,68,148,197,38,126,224,177,
147,222,72,145,165,110,29,254,49,153,170,75,
105,141,68,88,60,91,223,63,212,51,85,178,
64,188,188,105,195,9,121,252,104,149,242,112,
14,107,43,241,89,105,174,225,26,22,145,162,
14,124,125,108,182,3,54,219,1,102,212,161,
16,165,201,89,42,182,161,14,156,147,233,178,
99,99,71,73,109,156,8,202,57,3,201,93,
128,48,68,32,237,249,38,83,150,138,176,76,
62,26,119,248,243,249,105,12,163,73,214,170,
17,197,26,138,242,114,223,81,219,99,144,91,
70,85,243,36,165,28,210,245,113,88,175,145,
82,13,30,148,71,176,54,191,133,102,101,222,
156,161,16,136,158,110,194,113,60,58,83,83,
40,218,122,95,110,237,68,33,213,35,66,124,
190,225,28,201,111,223,242,162,30,221,185,172,
104,20,174,253,83,156,181,255,133,137,122,196,
92,254,61,238,24,207,105,69,37,150,226,17,
90,25,7,175,53,135,67,134,235,252,124,235,
157,36,2,12,4,144,211,67,242,149,20,63,
168,77,80,25,255,59,123,63,160,214,193,207,
31,68,201,30,54,231,243,232,198,107,138,204,
173,17,27,187,79,47,111,40,171,243,98,115,
176,50,171,29,63,102,173,225,83,30,251,54,
247,159,222,103,226,109,3,41,136,16,25,11,
169,229,144,76,14,131,166,33,63,95,34,186,
125,13,81,202,247,172,76,60,243,74,185,108,
41,71,178,89,71,94,169,34,42,121,80,21,
134,84,133,41,219,100,209,241,193,8,72,181,
116,172,226,227,238,162,151,36,242,134,198,134,
23,226,199,2,127,71,137,237,102,59,67,199,
203,82,46,60,165,234,84,120,212,132,108,45,
79,54,187,78,51,82,80,116,131,168,253,114,
3,31,61,66,143,229,48,84,50,6,142,31,
179,250,74,237,126,38,165,146,53,213,94,181,
58,192,147,165,38,245,90,155,119,63,188,222,
251,111,53,187,120,38,70,98,247,65,44,102,
185,225,177,188,85,17,109,168,50,150,174,144,
50,20,44,189,251,199,212,148,43,139,112,0,
94,53,172,69,91,199,8,87,240,219,1,13,
134,200,179,142,182,69,102,122,251,43,35,200,
133,12,200,18,183,154,75,40,215,178,252,245,
154,196,102,107,152,36,86,40,229,186,30,109,
202,31,162,88,222,64,83,116,76,69,166,26,
132,44,230,50,172,27,38,178,4,237,199,83,
76,205,126,167,239,103,140,69,191,134,194,221,
34,218,241,26,47,75,237,14,57,13,182,106,
50,166,161,163,104,71,187,189,65,187,134,172,
168,8,1,110,189,3,135,157,76,189,6,150,
161,99,199,17,94,163,43,200,115,160,214,130,
247,178,61,191,182,209,198,142,119,39,7,55,
107,211,124,229,102,157,220,31,254,140,103,157,
144,155,55,124,198,109,133,166,98,240,103,107,
83,140,69,159,237,120,167,110,11,125,132,138,
157,86,89,170,220,32,49,255,1,113,224,19,
7,187,19,172,235,107,109,216,174,86,121,101,
188,44,77,193,111,190,180,250,27,237,128,246,
182,188,162,215,56,198,215,121,244,60,156,244,
90,83,215,185,150,183,80,188,0,207,59,122,
222,0,50,2,100,183,117,112,229,86,191,207,
144,196,91,253,28,221,49,90,92,170,145,30,
205,16,183,114,64,68,50,62,132,40,229,81,
62,127,74,50,57,130,48,117,226,241,74,183,
26,105,251,212,21,69,136,125,186,180,69,41,
79,92,218,45,69,250,110,62,205,132,101,240,
67,81,35,201,215,123,231,183,72,8,194,125,
78,81,141,78,153,106,171,76,33,189,129,23,
216,124,105,102,147,244,39,143,120,248,96,17,
95,135,182,100,33,53,54,104,110,116,215,177,
38,75,228,137,240,26,173,94,159,64,63,216,
57,182,21,21,26,81,123,151,183,172,72,42,
94,195,235,149,134,118,58,29,238,127,252,136,
107,227,89,36,255,229,239,201,212,23,104,202,
9,225,118,66,187,223,53,182,109,129,98,191,
191,245,181,211,163,222,250,83,219,21,191,150,
48,132,131,105,165,73,25,26,134,42,163,171,
50,186,34,239,123,152,140,195,0,145,68,135,
242,92,237,55,86,131,188,246,56,243,213,239,
181,145,239,32,201,234,174,125,49,15,212,146,
14,109,63,70,2,84,63,66,138,35,4,18,
25,170,84,25,101,152,57,20,69,66,191,54,
132,152,30,69,121,182,8,142,79,59,99,99,
25,30,45,55,143,72,100,100,41,65,120,58,
215,172,132,97,205,198,222,234,200,204,106,42,
73,10,86,230,21,54,94,36,84,42,111,247,
250,93,142,122,175,40,138,16,142,211,167,161,
216,174,123,79,226,190,248,79,60,45,38,159,
77,247,106,170,15,223,27,186,245,210,115,155,
14,142,170,30,238,195,36,49,83,147,227,152,
41,173,43,155,138,10,135,133,28,183,158,213,
81,100,204,92,110,215,95,165,179,14,34,155,
166,82,49,105,181,67,150,87,29,130,32,193,
113,98,70,197,110,18,44,153,24,89,150,184,
49,165,240,176,242,30,97,254,183,200,149,222,
218,155,135,240,34,124,77,125,217,71,240,202,
120,229,243,38,102,174,251,192,110,16,179,82,
109,65,202,218,243,188,125,109,168,103,112,173,
165,194,91,55,167,15,38,24,220,103,222,122,
39,161,208,98,227,32,111,170,223,103,72,34,
80,140,94,255,69,4,136,84,6,164,105,188,
248,17,166,34,35,173,110,34,76,3,130,16,
76,125,15,117,135,208,250,63,65,187,113,66,
51,136,16,49,132,15,115,140,124,165,131,27,
199,120,113,130,19,197,200,72,61,227,183,90,
155,38,155,106,50,61,252,144,48,210,25,205,
85,49,69,30,201,245,240,131,136,124,201,166,
165,220,36,107,149,232,108,173,227,145,98,10,
107,71,245,210,171,156,72,253,142,237,172,85,
228,243,165,86,143,117,84,75,235,152,57,187,
215,64,246,100,57,160,21,154,204,142,238,14,
133,118,114,30,137,85,130,157,121,138,126,230,
65,217,106,146,212,251,84,147,59,226,158,9,
224,122,10,110,98,81,243,213,174,37,217,10,
227,234,154,140,169,42,152,154,140,169,117,127,
106,90,136,46,245,199,245,244,234,88,13,234,
218,227,204,87,191,215,110,115,61,189,186,47,
222,202,150,249,124,177,137,23,38,152,185,2,
73,77,224,6,18,38,46,66,174,179,104,124,
137,143,138,95,32,63,94,64,88,38,242,220,
10,241,219,55,94,122,186,66,66,150,147,173,
13,59,207,204,104,2,66,166,221,76,48,83,
18,154,222,85,224,241,54,134,41,230,111,81,
154,254,168,239,247,138,245,44,164,204,62,13,
197,49,209,116,35,70,242,199,115,95,183,123,
31,14,67,218,80,201,166,52,218,126,196,124,
213,61,198,253,119,231,41,0,34,233,46,207,
229,117,102,11,58,241,35,65,189,22,30,158,
67,185,97,241,176,124,143,32,253,219,100,75,
179,251,94,183,118,68,8,44,187,213,31,144,
8,193,147,245,206,145,92,252,175,19,5,75,
99,92,179,250,50,18,251,97,56,107,28,108,
40,78,8,93,233,110,30,145,113,141,255,244,
226,199,252,226,68,158,236,88,5,33,75,136,
157,100,71,71,168,179,5,177,96,105,35,192,
141,99,102,71,82,104,178,68,59,138,249,235,
205,38,245,103,26,234,24,196,29,9,199,143,
49,116,25,47,78,200,233,42,237,32,102,181,
85,164,227,229,249,238,245,77,90,145,199,188,
147,48,147,19,188,219,10,208,190,120,70,140,
220,45,99,156,28,33,149,121,159,229,250,75,
55,172,96,107,3,25,7,75,87,184,81,182,
120,188,238,32,132,216,69,213,18,199,130,249,
229,6,19,211,123,181,41,82,190,139,148,62,
191,9,231,174,244,112,130,31,38,52,220,93,
71,111,20,98,172,180,232,134,175,12,5,75,
87,177,141,203,25,194,234,246,131,153,60,223,
232,206,111,172,166,208,21,153,24,152,40,120,
12,23,92,138,217,12,178,20,34,30,189,96,
205,156,164,42,222,34,162,43,107,42,18,137,
225,210,66,247,80,43,105,248,177,203,154,23,
32,105,80,14,117,90,94,194,95,175,117,200,
230,127,25,195,236,95,93,210,13,18,22,27,
29,250,17,43,56,145,161,104,123,17,113,34,
232,119,90,55,219,253,37,116,135,115,6,66,
192,179,117,231,216,26,208,115,155,14,55,135,
95,158,80,52,179,196,98,43,228,70,198,70,
86,55,187,140,178,18,152,166,66,161,160,163,
171,114,247,44,41,4,118,206,224,126,241,54,
65,225,159,161,198,234,1,49,245,232,200,141,
210,208,186,247,124,178,218,57,87,205,113,105,
67,229,198,176,141,87,107,157,248,30,182,161,
246,77,178,214,47,166,183,56,239,21,205,196,
175,234,196,227,162,39,196,178,203,227,91,171,
146,12,151,8,124,193,134,23,18,198,2,191,
45,88,223,84,112,130,18,178,116,29,149,34,
170,101,243,199,207,126,76,74,95,97,124,58,
164,98,104,212,59,42,196,18,250,237,58,190,
36,122,149,60,113,34,88,110,148,201,164,218,
124,127,58,34,76,18,202,166,137,46,75,204,
100,82,232,235,85,252,133,117,86,151,61,70,
38,211,252,92,190,78,108,78,178,190,254,146,
62,58,25,224,57,160,144,214,121,199,84,104,
185,17,229,29,242,194,155,142,143,231,133,76,
94,219,107,40,66,85,71,72,23,147,159,41,
78,4,45,47,218,85,181,152,183,52,174,149,
173,61,213,119,151,197,88,232,170,140,44,73,
8,186,210,205,49,16,4,9,183,114,43,60,
116,110,112,203,122,130,228,5,12,41,243,56,
114,158,207,86,134,145,165,132,182,151,71,136,
5,36,4,34,145,249,163,149,42,154,44,241,
118,62,77,168,37,124,188,26,17,190,152,101,
236,222,7,251,139,169,177,149,19,108,250,56,
65,132,23,38,36,66,32,156,54,24,217,30,
203,198,192,13,133,16,93,105,193,138,217,95,
41,154,165,31,109,82,74,105,29,83,245,88,
174,123,71,54,127,236,135,90,39,100,185,238,
245,26,153,52,61,139,37,52,72,18,108,91,
197,233,68,20,75,6,67,67,6,198,142,231,
9,239,92,231,47,230,43,232,197,223,38,149,
169,236,59,208,235,173,128,23,155,71,199,241,
158,174,117,8,226,228,92,149,21,154,154,204,
173,145,193,200,76,26,170,66,20,15,78,255,
124,251,4,233,181,255,130,219,119,21,100,137,
46,61,135,169,131,235,163,172,110,34,173,108,
34,76,157,103,142,205,243,213,175,162,152,101,
100,73,69,183,74,164,71,75,100,182,54,202,
174,154,92,133,76,229,6,34,142,248,226,139,
127,201,175,126,203,163,241,118,147,213,143,243,
92,127,199,69,24,130,165,45,90,143,229,182,
14,146,64,85,29,126,180,1,155,205,97,12,
89,163,104,228,184,93,123,140,188,176,134,124,
99,140,248,163,60,31,63,206,34,167,254,238,
158,230,186,249,77,151,91,35,233,129,53,103,
153,170,130,153,217,253,173,172,108,118,251,2,
246,235,5,144,16,151,106,51,173,59,33,159,
46,52,153,42,165,168,100,140,75,245,110,110,
24,147,53,85,64,208,142,96,234,70,145,213,
143,157,94,184,49,217,22,251,82,21,226,233,
81,158,198,14,195,249,110,31,69,37,183,196,
114,117,154,156,189,73,118,75,128,43,76,4,
143,154,14,145,16,116,26,10,229,225,131,195,
77,75,53,143,197,186,203,105,2,28,39,46,
153,104,56,33,149,62,69,198,114,86,183,203,
119,103,40,38,111,105,140,23,82,221,18,60,
77,166,146,49,104,172,215,89,236,156,188,186,
102,161,234,98,27,42,217,148,138,145,173,176,
185,106,144,140,148,152,253,80,2,215,103,67,
205,50,31,153,196,178,130,162,201,108,198,22,
181,185,183,40,140,255,125,124,161,237,113,155,
171,237,128,229,186,119,168,50,221,78,188,218,
211,241,166,161,200,18,183,70,50,3,83,16,
83,7,92,26,217,116,35,114,154,143,164,174,
51,27,4,104,65,183,153,242,193,134,96,105,
62,143,164,94,67,200,32,60,13,157,33,74,
215,191,209,215,125,37,69,165,80,250,59,252,
217,211,255,139,214,194,56,182,57,78,198,250,
25,78,172,179,82,29,162,214,30,66,83,3,
102,199,62,165,233,20,112,188,12,133,244,6,
95,25,146,41,155,26,157,168,204,23,134,206,
162,36,225,62,206,50,90,254,29,214,2,155,
68,236,206,254,55,220,144,231,27,29,174,87,
236,99,191,123,18,133,36,113,216,227,53,218,
215,115,247,35,30,63,94,99,248,218,254,164,
136,114,18,35,137,203,101,44,226,68,240,108,
221,33,73,216,183,225,246,162,162,233,70,148,
51,6,205,181,71,40,149,58,201,156,67,140,
210,51,20,37,171,197,134,62,65,113,52,225,
105,185,64,163,190,155,199,108,180,56,71,171,
154,193,121,2,202,112,17,121,164,70,91,196,
68,13,25,179,250,33,153,91,55,15,140,180,
244,19,246,63,51,67,209,116,163,163,66,199,
187,60,16,77,145,240,119,8,10,153,154,140,
109,40,216,91,13,50,94,152,240,162,234,34,
244,147,55,120,9,186,205,66,247,38,50,232,
138,10,226,23,248,143,139,63,33,137,43,36,
202,24,118,238,235,152,153,33,36,89,65,36,
49,170,110,146,13,19,238,175,181,81,229,144,
219,163,153,222,98,253,108,177,53,208,222,129,
55,129,107,101,11,83,27,140,27,239,133,9,
77,119,176,61,33,115,155,14,239,76,100,49,
172,255,146,255,248,248,143,49,83,45,218,173,
18,249,194,55,177,75,185,99,37,27,95,69,
182,114,139,185,79,167,201,151,62,100,227,249,
95,240,233,15,211,84,39,117,42,249,101,70,
138,243,4,145,193,243,213,219,100,204,58,35,
197,46,157,252,39,117,184,155,179,89,114,125,
234,137,0,33,72,37,239,160,90,37,214,54,
246,175,40,90,111,5,84,178,198,177,63,36,
73,146,120,250,87,191,199,228,251,191,65,42,
187,191,33,120,177,210,4,68,175,111,98,255,
208,211,229,236,107,152,219,116,64,130,204,37,
120,151,40,17,248,81,66,193,214,88,93,249,
24,169,20,34,101,82,200,212,136,99,137,141,
13,159,92,105,141,39,197,111,82,200,44,49,
239,236,95,142,152,41,182,72,98,200,38,223,
69,108,4,36,73,132,149,157,198,190,53,177,
239,245,43,13,111,32,70,226,84,134,98,91,
15,34,99,170,71,24,148,144,249,70,125,151,
219,163,43,50,227,133,212,46,239,228,241,90,
135,100,0,34,60,97,156,240,120,181,195,157,
209,12,197,241,175,33,196,71,56,155,203,216,
229,87,63,70,141,106,39,232,241,52,205,14,
219,59,6,216,191,240,70,162,146,49,122,189,
37,167,69,183,159,165,51,240,64,71,156,8,
158,174,119,152,29,186,65,166,52,75,18,71,
228,71,212,94,40,41,17,130,48,18,196,137,
32,74,186,63,227,68,144,8,241,146,219,73,
150,73,155,251,135,54,167,223,249,39,136,56,
66,183,10,172,60,248,67,204,118,7,165,176,
45,159,235,115,109,248,62,171,181,9,130,200,
64,87,3,64,160,74,82,175,1,41,90,51,
24,25,122,159,213,166,127,104,83,210,252,166,
203,245,99,58,21,146,162,82,152,156,166,181,
254,111,208,140,255,1,213,72,239,25,155,141,
134,199,244,141,161,131,67,85,129,139,36,18,
46,43,230,54,28,38,141,16,235,130,191,71,
219,139,152,42,165,144,147,144,64,157,231,94,
38,69,123,185,133,132,32,8,4,115,115,29,
38,53,13,237,3,155,164,152,231,171,73,194,
159,174,213,119,245,247,244,188,200,74,139,206,
179,207,152,188,249,143,15,253,157,181,78,56,
208,222,167,83,117,235,212,58,225,145,134,34,
99,106,140,164,76,54,219,1,65,148,32,1,
51,67,118,175,17,197,9,98,30,173,118,43,
132,164,1,78,204,147,181,14,55,134,108,36,
73,238,17,110,237,196,106,195,239,81,110,228,
45,141,162,253,114,83,245,47,184,145,72,233,
10,211,229,193,81,25,188,216,116,14,236,70,
31,132,75,190,29,151,86,100,9,39,240,112,
252,152,78,189,131,191,217,255,242,76,69,29,
110,152,69,172,87,42,103,36,69,197,202,143,
51,243,213,127,198,242,227,255,64,152,172,236,
34,174,25,46,44,176,217,28,38,138,117,222,
170,52,152,170,215,185,47,117,239,33,183,75,
168,99,5,86,95,52,142,12,57,110,138,144,
241,220,241,222,189,60,245,43,60,127,48,71,
242,236,199,152,185,33,242,163,111,239,50,206,
153,220,225,91,100,199,176,73,228,203,221,236,
182,84,243,40,14,39,125,51,182,158,71,100,
83,221,208,251,230,252,143,49,198,234,140,45,
184,124,90,19,180,153,160,204,34,138,12,217,
225,52,181,164,251,142,186,44,147,81,149,158,
161,144,187,105,236,222,65,45,50,86,240,59,
155,24,118,233,192,144,229,147,181,206,64,223,
225,84,134,162,238,132,76,149,82,71,184,216,
93,26,242,201,98,138,182,31,33,75,210,174,
228,246,102,59,56,147,50,210,106,39,128,53,
184,49,180,247,168,183,88,115,89,172,117,221,
59,85,145,246,240,25,37,23,56,236,43,75,
18,179,195,246,192,200,252,226,68,28,89,22,
124,90,132,113,178,119,97,71,241,203,126,149,
62,224,134,49,159,45,53,153,44,166,24,201,
237,95,23,62,124,237,123,204,205,63,68,157,
172,242,145,109,210,82,100,42,134,78,67,219,
32,211,236,160,254,116,149,133,197,38,51,191,
246,1,247,195,8,89,216,108,182,131,125,184,
157,246,119,243,203,199,220,208,36,69,165,84,
254,101,150,31,253,7,92,191,72,110,228,46,
210,86,114,190,31,170,140,172,219,164,97,23,
8,21,141,203,138,68,116,195,201,119,198,50,
72,103,180,246,250,149,128,61,205,55,9,208,
241,63,227,195,74,154,234,159,222,167,147,100,
208,240,9,242,179,232,184,168,105,147,157,212,
76,186,44,163,72,18,51,233,20,211,105,147,
143,107,45,214,189,173,208,175,20,35,31,80,
169,20,39,221,170,203,65,239,169,167,26,33,
47,140,143,21,162,73,27,234,158,10,168,179,
228,142,169,118,2,30,175,181,119,109,252,115,
27,78,207,72,0,76,151,172,61,11,229,34,
243,217,92,43,91,123,200,228,78,3,69,150,
40,88,23,67,148,91,136,110,55,255,131,229,
246,46,189,223,222,98,87,117,82,220,33,110,
43,108,68,49,149,141,58,169,181,42,242,162,
135,107,26,40,211,195,196,223,127,151,207,231,
44,164,249,25,76,109,154,229,122,127,70,50,
22,130,167,39,56,197,101,135,110,147,26,85,
152,253,112,153,165,71,191,199,252,227,127,69,
109,249,227,190,62,244,106,166,68,164,92,126,
250,140,182,23,177,218,240,6,118,191,48,238,
230,67,255,102,174,206,79,231,26,71,82,4,
13,204,232,73,14,150,128,122,45,162,34,45,
35,210,21,198,245,69,52,83,167,83,151,176,
229,38,247,171,218,150,113,129,138,169,243,86,
214,194,143,18,170,91,90,216,73,71,193,240,
239,162,165,246,119,95,23,107,238,190,107,255,
141,122,20,219,94,197,72,238,228,139,117,56,
107,176,82,247,14,21,36,63,109,120,172,237,
181,25,215,124,106,157,112,151,134,110,46,165,
237,27,199,215,148,139,105,40,202,105,253,80,
109,144,147,226,198,176,77,170,38,179,84,247,
184,8,69,54,13,55,228,147,249,38,99,5,
147,209,220,110,226,202,145,235,191,194,242,3,
120,182,210,226,145,188,134,144,35,210,234,135,
180,31,127,129,176,171,200,139,195,76,221,249,
29,100,85,103,125,117,25,175,211,255,65,168,
229,69,44,214,60,198,11,102,255,223,207,242,
199,220,157,232,48,243,224,25,111,151,214,249,
163,96,132,116,238,239,247,197,191,51,84,95,
197,209,45,162,75,236,81,108,99,173,233,31,
232,41,30,7,155,237,128,185,13,135,200,9,
33,245,122,23,179,38,85,8,233,32,43,18,
145,128,247,203,243,44,27,31,16,151,203,124,
146,46,147,77,92,254,116,46,197,139,214,26,
137,30,16,37,130,31,46,36,108,174,106,36,
126,17,77,26,34,99,220,166,48,251,165,3,
188,90,191,91,121,105,15,254,217,79,109,40,
106,157,240,84,19,168,108,117,45,46,212,220,
51,155,160,40,22,251,10,204,76,20,247,127,
238,179,118,69,207,2,166,166,48,93,62,155,
180,159,68,151,105,182,156,49,88,168,186,23,
66,175,32,17,130,133,170,203,122,203,199,214,
85,162,36,193,80,21,138,105,141,92,229,61,
172,226,40,98,43,17,44,73,50,165,248,107,
136,56,70,153,122,185,38,78,146,151,89,170,
185,100,204,110,137,246,209,30,80,66,179,250,
231,92,143,22,120,252,241,42,229,239,141,64,
235,23,208,172,60,181,206,209,21,102,103,69,
51,126,30,225,133,9,45,239,232,226,153,195,
48,95,117,119,117,214,119,191,27,153,114,250,
245,148,225,150,204,58,197,182,195,156,19,211,
97,136,41,217,71,155,45,51,34,109,96,122,
79,120,212,186,67,49,87,163,45,154,52,107,
35,16,36,104,245,219,84,134,239,97,12,21,
14,165,76,234,248,241,153,138,78,117,71,125,
75,216,187,43,164,221,135,149,13,93,144,84,
80,53,218,29,104,233,153,3,195,53,94,99,
227,104,247,59,17,168,78,155,120,75,208,188,
191,149,211,236,114,7,245,131,3,222,107,109,
169,65,54,165,98,106,47,85,248,188,198,6,
174,20,64,199,59,241,125,207,234,121,15,186,
86,145,96,178,98,227,215,15,15,125,244,51,
23,71,93,59,170,66,222,74,88,109,122,47,
213,5,251,125,222,36,134,208,131,216,27,236,
120,29,114,253,54,137,221,54,54,128,49,163,
191,67,137,83,91,3,213,57,214,239,23,192,
227,167,235,220,24,74,31,232,153,238,28,91,
219,254,58,255,222,11,249,165,223,40,243,163,
249,28,150,53,133,83,93,198,119,35,112,54,
145,0,37,114,144,100,9,169,43,179,220,101,
41,6,198,23,239,243,116,104,154,104,91,125,
177,223,49,219,38,5,12,59,131,155,7,191,
211,189,95,63,201,245,40,236,18,67,246,243,
251,119,172,45,183,214,70,57,132,241,248,176,
245,189,222,10,88,109,122,123,222,107,168,100,
225,214,14,159,227,160,93,35,233,163,209,52,
242,29,100,89,69,62,128,44,213,82,154,180,
30,172,116,53,111,180,17,230,37,155,210,143,
255,128,154,23,49,59,155,230,131,184,193,240,
237,119,248,211,133,113,108,115,5,93,234,96,
164,190,69,232,180,72,194,0,73,213,247,109,
71,72,146,46,101,144,136,18,164,227,238,73,
145,223,215,156,117,71,221,30,122,121,118,180,
251,168,93,15,58,221,182,111,181,75,205,29,
152,246,161,165,152,86,113,244,200,91,78,25,
1,207,158,211,223,239,135,238,239,239,151,56,
239,128,247,90,23,176,238,116,19,238,182,174,
98,104,50,65,232,210,86,179,96,103,79,124,
223,179,122,222,131,174,157,153,185,78,161,79,
221,230,126,230,226,168,107,45,160,52,210,173,
134,90,105,248,253,63,111,18,65,224,244,72,
1,7,55,94,253,95,47,128,21,103,157,74,
118,184,71,21,225,53,215,240,218,171,164,75,
55,80,245,20,113,224,33,27,41,252,101,183,
39,46,127,156,223,31,1,75,161,202,157,202,
193,93,0,194,255,57,168,25,10,227,31,224,
189,120,198,243,70,27,211,254,7,164,43,211,
47,199,184,216,253,221,175,146,231,197,137,224,
241,90,135,57,201,196,55,237,151,31,122,191,
99,22,28,131,20,176,223,123,42,77,208,173,
190,232,32,136,252,238,90,232,139,148,176,187,
182,84,89,162,50,146,63,178,111,107,191,53,
219,112,67,86,235,237,221,223,244,214,123,89,
133,12,214,17,222,159,172,168,167,34,5,220,
94,121,234,134,64,73,18,36,4,19,241,103,
240,52,38,123,61,77,106,72,39,73,64,206,
216,124,178,82,195,168,67,93,181,9,54,45,
38,110,93,71,36,49,138,110,30,232,81,60,
89,235,224,27,41,48,78,176,39,157,37,133,
199,158,240,147,19,158,186,102,191,156,209,217,
76,169,52,120,253,16,162,91,82,214,246,65,
242,35,132,202,133,193,80,214,232,219,72,12,
26,83,37,11,85,145,89,236,112,161,144,8,
120,81,117,153,221,170,136,243,58,107,52,204,
255,135,234,166,1,158,134,236,20,24,190,245,
187,7,211,170,247,129,150,23,177,214,244,119,
9,112,245,54,148,246,50,55,243,127,133,155,
168,44,46,60,36,118,52,150,248,128,145,27,
183,251,14,215,206,14,217,240,232,167,124,81,
158,193,49,46,122,167,193,209,40,216,250,137,
169,82,170,135,112,205,205,87,29,238,142,101,
57,235,190,197,36,10,145,18,129,254,214,36,
252,120,13,146,24,1,204,207,59,68,81,194,
91,111,101,177,23,215,40,149,115,216,249,95,
103,88,86,209,135,243,40,122,138,160,93,63,
240,190,77,55,124,45,161,224,129,4,227,27,
78,56,144,36,231,88,62,117,33,243,3,111,
10,197,180,78,37,243,102,99,212,99,121,147,
130,165,94,184,177,171,182,131,94,14,34,63,
114,15,165,94,66,29,114,81,167,154,200,22,
125,211,182,28,134,245,3,72,36,213,248,175,
40,110,46,50,211,122,142,174,9,166,222,249,
109,70,110,252,131,99,221,91,145,37,226,183,
238,224,107,198,165,95,231,138,44,29,152,79,
236,7,135,81,207,116,252,152,167,235,103,127,
210,241,26,207,169,110,10,30,124,17,162,203,
49,195,55,198,41,15,89,68,91,98,103,178,
44,129,44,161,142,203,84,87,254,130,84,110,
20,69,63,186,23,234,56,44,219,111,220,80,
108,51,65,158,126,65,192,76,229,242,159,142,
6,129,188,165,113,163,98,159,139,103,25,203,
167,72,155,23,207,88,244,136,30,37,9,41,
25,194,82,21,16,160,196,5,58,167,101,255,
21,98,223,210,113,183,250,41,211,222,3,130,
166,203,143,243,19,132,198,215,78,252,43,38,
86,159,98,132,254,165,95,235,251,149,176,31,
107,156,138,169,67,213,28,55,219,193,190,197,
46,131,68,236,111,16,169,163,200,181,26,185,
114,154,82,70,208,40,124,200,236,173,28,153,
140,138,174,203,120,215,199,89,143,98,68,216,
159,123,83,237,4,167,95,167,175,211,80,108,
63,244,32,144,179,180,125,221,245,43,236,240,
36,108,157,155,195,105,206,13,205,143,212,109,
108,188,104,253,39,78,16,247,202,165,115,133,
143,104,61,205,193,23,119,41,148,63,194,9,
78,121,240,145,164,61,205,119,110,237,1,185,
167,191,135,247,249,83,156,247,175,179,182,62,
78,166,242,214,137,127,197,231,217,201,75,239,
81,164,77,245,212,37,223,146,4,111,141,164,
119,233,124,188,138,181,166,207,163,213,246,137,
152,171,251,122,143,225,143,48,203,223,37,165,
36,108,110,122,180,171,117,110,79,45,241,120,
234,123,20,126,249,23,152,127,235,91,252,81,
124,135,23,247,71,41,95,255,187,125,221,115,
173,249,250,170,15,7,102,40,6,217,97,61,
85,74,93,233,240,30,128,114,70,231,198,176,
205,121,227,130,51,84,121,79,135,251,69,192,
54,105,90,186,52,195,228,196,191,96,242,157,
127,132,85,152,220,37,77,122,82,236,220,152,
218,43,127,192,104,248,175,153,181,235,148,191,
127,143,191,126,56,129,165,221,62,241,189,221,
48,198,126,250,224,82,123,20,154,34,31,171,
39,229,48,168,138,196,237,209,244,161,205,168,
181,78,151,230,252,225,74,123,224,148,53,146,
162,34,68,27,93,248,196,49,116,58,49,27,
159,175,114,239,157,38,76,148,72,87,20,198,
45,147,162,234,99,101,142,46,30,136,18,65,
203,11,95,219,92,12,204,80,196,137,232,171,
246,187,175,135,146,164,45,158,166,43,195,176,
19,67,89,131,153,138,205,121,29,150,162,173,
51,94,72,93,168,121,171,59,97,79,63,225,
176,88,74,0,0,32,0,73,68,65,84,68,
213,83,61,221,135,65,112,11,109,111,74,161,
87,103,172,241,199,92,75,59,152,182,202,223,
188,24,103,114,246,119,41,77,126,253,68,247,
13,99,193,195,149,54,207,43,215,47,173,71,
33,73,48,59,108,15,148,218,94,83,100,222,
26,75,31,57,183,117,39,228,254,114,139,159,
206,53,152,219,112,6,146,175,2,136,157,57,
146,150,67,77,30,161,153,185,67,16,74,168,
181,6,249,148,195,211,205,44,75,155,79,176,
181,247,251,94,183,175,179,249,117,160,153,227,
65,182,194,219,134,194,68,49,197,21,94,158,
78,167,75,231,255,196,62,94,48,185,55,158,
37,109,92,156,156,69,175,31,100,7,172,1,
200,114,154,186,76,18,71,200,115,255,43,179,
102,21,127,122,20,223,206,32,180,41,56,161,
50,93,156,8,30,174,180,240,195,132,233,245,
103,151,214,163,152,42,89,167,106,174,59,8,
186,34,115,123,52,131,38,31,61,254,97,156,
176,218,244,121,188,214,97,190,234,158,184,6,
46,137,67,68,28,161,182,30,98,73,62,197,
100,153,130,115,159,112,252,109,30,6,55,249,
124,101,130,213,185,77,50,242,239,80,153,254,
86,95,247,124,221,196,165,3,157,137,166,27,
225,135,9,198,128,52,16,70,115,38,77,39,
218,69,187,241,183,21,51,21,235,194,156,212,
45,93,225,238,120,134,181,166,207,74,195,27,
72,24,231,44,225,239,195,141,211,143,42,99,
63,227,208,169,62,33,103,8,126,122,227,58,
203,171,53,180,141,73,204,210,135,39,186,159,
16,240,120,181,211,75,96,206,93,82,143,162,
146,209,25,62,195,60,165,161,201,84,178,58,
75,199,176,177,203,117,143,142,31,159,200,203,
233,44,252,17,35,155,127,128,213,106,83,173,
250,140,12,219,88,57,133,145,239,164,65,93,
167,227,36,44,43,31,97,23,167,143,113,96,
120,189,115,50,240,90,212,65,19,108,205,12,
89,23,150,123,105,80,40,216,218,161,85,27,
231,21,67,89,131,119,39,115,220,27,207,50,
154,55,207,37,85,180,44,73,20,109,109,159,
77,94,61,245,125,77,77,33,83,185,69,60,
245,63,179,240,249,91,132,143,238,144,157,250,
239,145,78,72,13,254,116,189,179,235,208,116,
25,61,10,219,80,185,86,62,251,106,190,224,
4,135,151,166,27,242,104,165,125,236,144,79,
208,89,103,36,29,32,86,87,80,100,9,59,
173,161,32,144,130,238,92,134,106,130,43,253,
4,183,177,116,44,207,242,194,122,20,0,27,
237,96,160,33,35,77,145,153,169,216,60,88,
105,255,173,53,20,67,23,92,63,120,91,201,
112,60,167,225,118,36,48,50,4,81,87,245,
43,216,250,179,253,207,225,107,62,42,141,23,
204,125,11,39,44,93,65,130,19,135,27,236,
29,161,43,73,86,24,191,245,155,47,67,26,
238,241,235,246,87,26,222,158,198,170,197,210,
20,193,37,226,122,210,20,153,155,103,92,168,
33,4,60,219,232,176,209,9,224,4,219,84,
203,139,152,175,58,76,29,35,12,92,188,253,
59,60,121,248,140,130,181,72,218,214,88,91,
245,40,166,5,25,199,67,88,38,170,44,33,
77,172,179,244,236,95,113,45,245,63,30,202,
233,180,141,68,92,112,67,17,68,9,13,39,
36,55,192,110,225,156,165,49,146,51,89,25,
32,213,240,69,129,161,202,123,198,50,112,155,
104,186,133,164,92,60,47,67,147,101,116,67,
229,0,117,79,132,128,32,78,104,215,124,68,
202,238,25,146,48,78,186,74,119,241,110,197,
187,211,26,176,131,8,45,37,169,75,180,232,
158,226,222,131,130,19,196,251,54,86,141,212,
150,240,180,235,68,23,112,29,236,25,111,186,
201,107,253,12,189,206,68,8,158,172,118,168,
57,167,11,101,175,52,124,108,67,237,155,141,
66,146,21,54,152,161,242,43,54,238,143,62,
69,248,155,100,111,20,72,178,54,2,88,116,
124,12,69,38,20,18,138,222,223,161,112,40,
107,208,246,163,94,33,198,133,51,20,221,240,
83,48,80,67,1,93,166,215,186,19,156,251,
120,247,160,177,95,82,181,182,252,19,28,231,
33,186,152,34,63,250,206,165,122,95,73,234,
26,71,161,43,152,125,124,136,219,242,168,94,
35,68,207,102,17,162,235,5,116,127,10,90,
110,196,106,211,223,195,209,47,73,112,189,114,
248,233,53,165,203,184,39,92,110,131,12,21,
174,54,252,125,195,29,107,249,97,66,245,114,
80,140,151,51,198,153,36,175,119,174,147,135,
43,237,129,52,6,3,60,91,119,80,21,137,
92,170,191,241,79,165,60,74,247,127,78,80,
177,232,172,84,73,189,115,141,100,203,64,214,
131,16,231,121,137,74,241,239,109,253,151,163,
145,49,85,222,153,200,178,92,247,216,104,7,
103,110,48,206,196,124,215,156,128,40,30,176,
194,146,36,157,25,141,246,121,198,126,97,17,
51,61,140,174,13,49,122,247,123,39,142,119,
95,22,40,178,132,174,202,232,106,55,39,144,
210,21,44,93,193,54,20,210,134,202,104,222,
228,237,137,236,158,15,250,90,217,58,50,97,
125,154,143,111,144,155,94,253,128,19,112,185,
185,142,22,71,23,126,14,37,9,198,6,212,
47,113,144,39,113,127,121,112,70,98,251,158,
143,86,58,52,251,44,180,17,66,102,249,69,
11,51,173,49,117,187,140,212,118,64,215,112,
227,132,134,27,147,150,191,66,166,124,243,216,
123,226,120,33,197,123,219,121,192,156,121,102,
249,220,51,49,225,66,116,115,21,35,185,193,
198,214,183,133,134,46,130,30,194,160,176,31,
13,68,110,232,46,185,161,187,92,161,207,69,
46,75,220,26,77,51,183,225,176,222,242,169,
100,117,42,71,228,125,90,94,68,39,136,79,
20,199,78,155,234,64,67,40,7,197,163,235,
86,254,82,132,157,42,25,227,204,10,29,94,
86,138,13,222,160,38,66,240,112,165,195,205,
17,251,72,207,66,36,2,215,147,241,231,214,
177,127,237,43,196,106,247,144,242,179,90,11,
119,37,197,228,200,123,167,122,150,237,60,96,
73,74,163,100,114,52,189,144,150,27,209,244,
162,129,40,222,157,90,143,226,64,119,217,87,
200,196,54,254,0,52,16,118,157,162,36,193,
102,167,117,166,250,14,103,114,237,9,159,183,
209,145,232,168,233,94,35,216,73,198,236,188,
92,43,146,152,200,115,136,130,163,35,255,253,
106,0,28,231,250,33,25,74,25,65,216,106,
225,84,151,15,188,206,9,98,230,54,28,112,
27,253,207,217,142,249,205,106,38,78,181,125,
234,241,138,252,46,255,144,234,10,226,104,239,
129,193,174,47,80,151,34,216,46,145,189,128,
122,20,18,80,200,100,112,170,141,51,89,135,
139,53,143,186,19,156,238,123,60,228,218,4,
120,240,24,74,105,157,178,145,160,40,251,235,
81,200,146,0,33,72,144,144,151,214,73,166,
70,8,19,193,218,115,25,169,62,66,144,110,
16,116,246,31,131,208,105,161,104,198,129,58,
23,175,142,129,73,87,228,206,86,97,36,13,
97,36,104,7,33,29,47,166,229,69,47,213,
19,95,167,30,197,129,235,6,112,53,27,43,
55,24,13,132,222,160,197,9,180,82,103,170,
239,112,38,215,158,240,121,19,192,55,109,138,
246,233,244,62,206,195,181,73,28,17,185,109,
244,116,254,232,225,234,83,3,224,36,215,59,
178,116,224,243,174,53,125,94,52,92,18,203,
6,73,233,127,206,182,230,55,155,82,25,31,
201,156,72,55,97,239,126,222,53,54,186,16,
248,251,132,77,188,92,136,72,87,96,187,242,
233,2,234,81,100,83,26,185,74,250,76,214,
225,102,59,160,38,117,246,151,6,61,206,247,
216,199,181,155,2,218,142,203,244,80,134,124,
54,179,143,173,81,17,36,212,59,144,111,117,
199,191,29,197,72,181,113,174,189,243,79,121,
117,193,136,173,205,92,146,36,124,205,68,53,
82,125,85,67,29,52,94,185,29,94,80,181,
29,178,209,246,105,86,95,179,30,197,65,88,
172,185,220,28,96,73,180,16,48,183,241,122,
104,117,207,19,22,170,30,5,75,191,162,52,
57,67,8,1,143,215,218,167,162,161,201,89,
26,179,103,64,61,35,239,115,195,145,218,18,
35,181,101,150,138,227,23,122,220,139,233,179,
75,198,191,42,123,122,214,240,163,132,135,43,
29,70,67,149,201,210,238,152,101,168,86,184,
121,35,69,44,52,228,213,42,34,151,38,51,
54,132,102,105,236,183,96,118,70,16,90,235,
143,105,180,126,132,22,143,99,149,134,41,79,
127,227,84,107,169,156,209,41,103,116,154,134,
75,83,164,216,236,196,251,54,157,238,250,255,
206,122,224,170,157,193,117,85,63,88,105,157,
136,165,86,237,37,60,101,46,226,94,235,133,
49,107,205,203,79,39,253,166,81,63,225,90,
205,152,42,55,135,108,222,26,73,159,9,131,
238,206,28,69,202,119,248,198,231,127,198,106,
126,148,191,185,241,229,11,61,222,146,4,5,
235,108,250,64,26,110,56,48,142,166,99,27,
168,134,199,98,109,183,145,18,137,75,152,43,
96,170,49,237,64,197,121,90,101,126,33,198,
221,220,64,36,135,63,103,18,133,168,148,112,
155,139,164,178,19,3,123,78,93,145,24,203,
155,188,59,149,227,122,197,58,52,175,118,230,
153,176,245,150,199,104,34,6,242,1,101,76,
13,69,146,144,36,9,69,209,49,243,41,20,
73,34,140,19,194,173,250,250,56,17,168,138,
132,165,171,216,70,183,2,198,175,123,88,197,
174,243,21,39,130,166,27,241,162,234,188,182,
26,228,65,121,103,165,180,142,170,92,185,21,
103,181,105,233,170,124,228,201,234,85,104,138,
196,205,145,52,81,235,108,78,175,27,45,159,
150,39,35,33,248,242,163,191,226,225,248,109,
126,112,247,91,151,98,204,179,166,118,102,235,
121,163,21,188,241,239,85,87,101,42,25,157,
56,244,81,237,247,249,100,245,17,239,76,166,
248,172,145,161,169,90,180,63,174,160,106,16,
184,155,24,189,240,255,94,228,199,238,82,210,
63,4,73,66,57,3,202,22,137,110,65,65,
57,109,176,218,244,89,172,185,123,122,148,206,
220,80,68,9,60,93,235,112,115,36,125,234,
123,237,164,28,246,26,14,102,238,248,37,117,
138,44,81,176,53,156,192,96,177,118,113,194,
88,81,34,120,177,233,48,51,100,115,133,179,
193,80,214,56,182,98,88,218,84,81,101,137,
179,40,82,109,123,17,11,85,151,153,234,50,
142,97,241,241,245,15,47,85,39,246,89,133,
157,4,93,213,205,55,141,249,170,67,209,214,
80,84,157,165,207,254,28,35,51,206,95,54,
190,65,216,106,99,151,110,162,217,159,18,231,
230,168,45,125,204,200,205,239,29,189,119,157,
194,72,8,33,14,44,136,217,121,88,26,201,
25,228,45,149,135,43,157,93,21,151,175,133,
124,167,230,132,44,156,163,77,57,17,98,96,
66,75,175,19,27,237,96,224,92,90,87,120,
137,225,156,177,199,253,30,202,26,76,20,83,
251,246,92,108,215,177,159,5,194,56,97,121,
185,198,91,139,15,88,41,140,178,82,24,187,
84,70,2,32,127,70,97,167,181,134,79,244,
154,185,144,246,61,220,197,130,213,166,15,146,
196,204,55,254,9,1,77,220,230,10,237,206,
99,52,35,77,34,66,80,100,124,229,17,129,
219,60,99,143,249,112,35,225,183,215,169,46,
252,132,36,244,49,53,133,123,227,153,93,170,
149,175,141,165,109,169,230,157,155,254,135,231,
27,206,153,41,89,157,53,230,54,220,55,22,
123,189,236,144,37,137,201,98,10,89,146,80,
101,137,169,98,138,107,101,139,177,173,166,189,
27,67,187,153,67,135,178,198,64,88,102,247,
158,254,224,201,90,7,17,6,44,22,199,112,
140,203,231,69,218,166,122,38,205,97,45,175,
27,86,62,47,88,169,123,196,137,64,86,84,
38,110,254,183,88,217,41,102,63,252,23,196,
113,128,87,173,49,148,138,185,113,183,206,242,
252,191,62,22,41,224,192,215,190,106,224,180,
159,243,248,7,255,146,36,142,81,100,137,155,
59,40,85,94,107,183,206,147,181,14,45,47,
98,178,152,122,35,178,153,66,192,66,205,125,
227,241,203,211,122,67,143,87,219,220,30,203,
160,43,50,87,24,44,74,105,157,156,213,205,
133,249,205,112,207,223,101,83,42,79,214,58,
56,65,204,88,254,224,208,167,23,198,196,201,
201,56,159,158,109,116,104,186,17,101,223,69,
77,34,90,151,112,156,243,169,193,111,61,13,
55,100,169,217,70,136,243,243,158,81,34,88,
111,249,140,228,76,140,116,9,35,221,21,171,
170,173,252,21,74,201,103,101,65,97,185,170,
33,25,53,22,150,254,119,242,155,191,68,101,
230,235,175,253,57,53,51,203,196,237,223,36,
153,13,145,183,180,58,52,69,102,118,216,62,
216,80,200,146,68,74,151,207,68,184,123,173,
233,211,112,66,102,134,236,51,229,118,217,239,
195,125,178,214,121,109,98,228,103,251,46,9,
159,47,182,184,53,128,188,207,21,246,226,48,
189,1,109,75,248,38,17,98,223,178,85,232,
146,198,205,87,29,36,36,238,77,100,14,149,
223,124,21,139,53,175,119,144,241,53,131,56,
81,47,229,24,103,83,131,203,79,36,66,48,
183,225,178,81,245,16,118,230,220,189,235,106,
211,223,67,62,57,113,247,55,89,249,252,143,
16,66,144,207,191,139,179,57,135,61,114,13,
171,48,246,102,189,234,87,154,168,183,5,200,
118,175,66,9,202,105,157,137,98,10,93,149,
121,186,222,57,147,211,183,31,37,124,177,212,
98,36,103,50,81,52,15,252,224,78,178,96,
162,68,16,199,221,159,81,44,232,4,17,205,
13,23,167,209,122,237,212,188,103,137,96,107,
12,135,228,16,179,32,6,54,134,87,232,63,
76,181,231,244,24,11,158,174,119,122,220,76,
2,193,243,117,135,183,70,211,135,206,79,24,
11,156,32,98,165,225,239,74,194,106,113,136,
122,9,184,156,94,133,161,201,251,114,152,157,
4,78,16,243,100,181,131,27,198,231,182,244,
221,15,247,103,212,46,78,125,169,215,36,106,
191,210,15,19,39,130,106,39,36,138,19,10,
138,224,77,51,186,169,0,239,76,100,241,194,
4,213,245,200,236,16,13,41,167,141,51,13,
211,108,115,236,87,50,71,115,239,236,23,70,
106,186,17,53,39,160,225,132,4,145,56,208,
16,72,126,132,80,197,165,251,224,226,68,176,
210,242,216,20,77,110,142,164,7,74,109,125,
133,99,206,133,128,199,203,173,61,249,163,150,
23,241,249,98,171,183,49,10,4,113,211,37,
118,91,91,101,221,201,129,161,146,248,146,18,
62,14,34,236,36,68,183,161,110,177,238,114,
17,206,127,171,77,191,111,70,237,133,170,203,
106,179,203,24,60,150,55,207,197,33,80,133,
174,30,115,74,87,112,94,41,168,49,245,179,
143,129,135,113,194,82,221,99,169,238,145,137,
92,134,141,128,108,74,219,215,253,15,99,65,
211,13,169,117,66,26,155,45,98,83,189,218,
161,232,234,55,84,219,1,182,113,165,49,254,
166,176,222,244,113,34,245,192,83,239,78,3,
34,121,17,66,57,218,83,144,133,64,78,46,
31,173,126,238,148,213,78,110,24,243,116,205,
57,19,162,191,179,66,221,9,241,163,228,72,
242,195,197,154,199,114,221,163,146,53,24,47,
164,208,20,137,160,253,230,117,120,14,221,105,
117,69,70,150,164,215,22,178,105,251,17,173,
213,46,15,138,170,116,105,163,77,77,38,78,
4,29,63,222,205,130,40,4,87,120,137,166,
23,114,34,170,211,43,12,4,87,203,177,63,
40,178,68,218,80,17,201,201,54,249,149,134,
199,66,213,187,144,97,228,103,235,29,102,135,
211,7,230,192,154,110,196,70,219,231,222,68,
246,76,170,233,206,204,80,192,22,13,201,27,
152,147,40,22,180,227,136,182,119,245,113,245,
131,142,31,19,37,226,216,194,239,87,24,12,
10,105,141,245,198,96,239,153,72,18,146,124,
185,42,219,242,150,134,36,29,127,75,241,195,
132,167,235,157,129,106,74,188,246,195,156,27,
241,241,139,6,89,83,197,84,5,146,31,96,
42,1,138,36,225,133,49,27,237,128,219,163,
153,115,169,45,127,168,161,112,130,248,181,139,
120,95,225,52,11,49,60,148,101,246,10,103,
7,83,149,177,13,117,160,225,16,53,142,80,
147,203,213,51,99,104,199,223,4,215,154,62,
243,85,247,82,236,69,113,34,186,82,172,73,
4,190,215,165,103,223,58,144,223,25,59,159,
70,226,72,67,113,69,68,119,241,78,44,87,
134,226,13,110,130,170,76,103,128,159,76,160,
234,68,226,114,229,40,142,227,241,250,81,194,
243,13,151,182,234,92,250,181,51,81,72,245,
74,81,47,148,161,240,194,248,138,46,226,130,
161,225,132,8,193,21,29,249,27,130,174,14,
118,224,141,208,71,77,34,26,151,104,140,218,
94,76,37,117,116,252,125,181,233,51,191,233,
34,252,232,53,183,5,191,126,228,82,26,163,
121,243,92,63,227,190,126,78,148,8,30,173,
118,174,18,116,23,12,126,148,240,100,189,115,
53,16,111,8,169,1,39,32,93,35,69,231,
146,209,119,84,59,65,87,98,246,144,3,234,
23,75,45,230,54,156,75,213,247,116,16,50,
166,202,205,145,243,63,199,42,128,215,88,239,
186,186,237,26,200,42,207,214,59,135,115,33,
133,110,87,253,171,31,146,50,175,209,159,60,
226,73,174,237,23,103,249,12,231,236,121,171,
46,60,115,245,125,79,40,65,187,134,220,167,
198,242,89,92,155,36,49,177,231,144,196,71,
51,123,250,237,218,241,140,228,49,174,239,247,
121,143,115,79,191,93,67,209,99,112,59,3,
155,95,187,177,210,165,240,208,245,193,174,177,
109,41,212,216,31,220,61,253,118,247,126,125,
188,215,211,185,136,137,188,78,58,253,146,40,
52,22,176,222,242,217,104,6,36,59,83,221,
231,229,123,236,119,92,251,221,23,147,24,130,
54,19,102,134,160,121,120,197,78,232,180,144,
53,3,165,15,41,212,227,124,183,126,187,70,
146,196,200,253,74,161,110,119,7,198,81,196,
130,151,162,163,168,135,87,90,42,71,75,161,
238,26,144,126,37,7,143,115,45,156,205,125,
47,193,243,174,199,32,2,157,169,146,181,43,
38,156,196,81,223,114,161,103,113,109,18,71,
68,90,127,82,168,59,215,101,191,56,139,119,
59,206,51,104,153,50,180,234,3,155,223,118,
46,70,22,98,240,107,76,57,134,20,106,191,
247,148,212,190,165,80,131,200,103,174,19,51,
156,74,147,77,105,180,253,136,149,186,71,148,
168,144,178,207,223,247,120,156,113,237,123,95,
140,64,86,232,40,57,70,114,135,95,47,43,
26,138,110,246,37,133,26,71,225,174,53,187,
83,82,117,251,223,119,50,201,234,233,66,95,
134,69,238,186,123,9,47,54,93,30,44,183,
122,244,3,87,184,216,216,104,5,124,58,223,
188,144,116,234,23,21,138,44,145,183,6,199,
97,148,117,155,20,218,213,75,57,86,113,34,
88,170,123,220,95,110,177,80,117,207,5,45,
248,155,192,124,213,25,40,27,244,171,116,226,
73,232,225,62,249,75,98,175,189,239,223,247,
11,25,224,147,249,6,43,13,143,248,42,39,
113,161,241,234,26,8,227,132,199,171,29,30,
173,182,9,226,228,106,128,94,3,198,6,152,
148,172,217,5,54,178,229,171,65,189,196,16,
2,94,108,158,172,170,43,142,250,59,4,202,
149,155,68,225,233,10,147,174,56,48,46,1,
74,233,110,78,34,165,41,116,252,136,106,39,
100,181,233,245,138,17,106,157,144,166,219,100,
76,11,177,174,134,235,76,145,54,85,134,179,
93,73,201,211,162,208,169,161,69,33,243,233,
43,99,113,153,113,162,46,108,33,88,254,225,
255,205,196,183,126,231,112,47,87,79,161,232,
167,103,108,184,50,20,231,121,1,25,10,37,
91,239,113,5,121,97,220,219,252,37,169,75,
218,88,76,167,201,237,144,71,77,155,42,105,
83,165,156,209,121,182,254,146,86,61,78,4,
11,85,15,223,112,152,46,89,87,37,180,103,
136,233,178,69,24,159,94,69,177,102,23,144,
197,149,39,120,81,81,206,232,84,219,225,129,
213,91,186,34,51,61,156,166,96,107,59,246,
127,209,95,120,72,146,142,52,18,131,196,149,
161,56,99,108,207,249,73,162,122,166,170,236,
170,94,18,162,75,136,22,39,130,148,174,160,
202,18,78,181,113,224,41,229,238,120,150,133,
170,203,114,253,101,85,197,90,211,71,150,96,
170,116,229,91,156,37,102,135,109,106,29,157,
197,218,201,21,9,175,60,138,139,141,225,172,
201,100,209,98,185,238,177,217,246,9,183,98,
251,154,34,83,202,152,148,53,176,108,237,149,
253,98,127,35,145,196,33,10,111,174,215,226,
202,80,156,17,74,105,157,241,66,10,83,147,
73,132,96,109,197,101,61,86,142,37,193,186,
139,4,113,203,232,28,199,77,149,128,201,45,
189,231,103,235,78,207,88,85,59,33,83,165,
171,57,58,107,20,108,141,130,173,81,237,4,
44,213,61,156,99,138,102,93,121,20,23,27,
115,155,14,119,199,50,76,149,82,76,149,82,
56,65,140,166,72,104,138,76,18,71,199,98,
133,93,255,233,239,51,250,181,127,136,244,134,
168,231,175,12,197,128,81,76,235,76,20,204,
93,194,44,178,212,173,134,25,206,102,183,40,
213,143,230,208,87,100,137,161,172,49,144,103,
42,165,117,76,77,230,225,227,117,66,24,104,
101,206,21,250,88,19,182,78,209,214,105,121,
17,11,85,151,118,159,61,145,87,30,197,197,
70,219,139,88,107,250,189,239,248,52,140,176,
195,95,254,245,55,102,36,174,12,197,0,97,
106,10,179,195,246,161,139,65,146,96,188,96,
82,76,107,172,54,124,154,110,72,20,11,36,
169,235,142,218,134,138,109,40,40,182,77,113,
40,63,208,60,130,109,168,220,24,182,217,72,
244,35,235,182,175,112,54,200,152,42,119,198,
50,44,38,53,86,35,233,200,146,208,43,143,
226,226,163,214,9,7,118,224,235,186,22,15,
208,133,71,39,61,67,82,123,129,90,156,64,
73,229,174,12,197,155,134,36,117,115,5,110,
24,31,26,98,152,169,216,40,125,18,158,165,
52,133,107,229,131,115,4,78,36,159,73,178,
89,149,37,102,202,246,165,157,171,56,240,136,
215,31,161,143,191,115,174,159,179,96,105,12,
229,178,60,223,112,168,117,194,43,143,226,178,
238,29,116,19,218,39,69,228,187,72,73,72,
214,157,39,43,155,88,178,76,144,10,89,238,
40,40,186,137,156,41,35,94,147,176,213,149,
161,56,2,55,135,211,228,45,141,229,186,199,
124,213,221,243,247,150,174,48,59,148,190,170,
34,234,23,66,224,183,54,193,107,32,217,101,
116,43,219,251,171,176,190,68,134,14,25,29,
214,90,13,18,59,191,71,236,253,80,3,92,
127,68,165,100,176,120,1,134,65,83,100,110,
14,167,169,118,2,158,111,56,68,251,52,49,
93,121,20,23,23,170,44,49,59,108,147,77,
157,44,204,27,215,151,80,171,15,185,62,61,
197,247,126,227,151,105,181,90,252,229,127,254,
152,7,203,45,50,179,95,235,202,163,102,135,
95,223,251,92,77,233,193,176,116,165,23,207,
31,205,155,212,156,144,246,43,194,41,147,197,
212,161,70,66,136,4,214,31,80,148,67,108,
167,78,181,19,17,228,174,247,213,142,255,186,
224,44,63,196,26,189,117,172,255,39,9,28,
226,218,60,66,179,80,91,171,228,181,38,178,
36,227,135,49,181,196,38,137,3,108,201,195,
54,52,162,68,16,197,96,42,49,69,91,99,
246,218,4,195,149,41,94,44,46,179,176,242,
20,4,40,138,204,187,95,187,65,185,212,205,
178,39,73,194,31,255,224,175,89,216,112,137,
133,132,42,75,104,138,132,34,75,132,177,32,
140,5,25,83,69,149,187,255,220,104,183,176,
45,149,185,149,42,234,181,139,179,198,138,182,
78,214,212,152,219,116,216,108,7,87,30,197,
101,216,55,12,133,155,195,233,19,106,75,8,
148,245,207,25,73,193,127,241,143,127,139,78,
167,195,191,253,255,254,132,185,6,40,185,97,
114,250,58,222,243,31,161,95,255,250,235,53,
124,87,211,122,248,71,188,19,163,57,147,71,
91,173,240,219,176,143,208,237,214,214,62,225,
119,254,254,183,49,12,163,183,1,254,248,167,
159,242,131,159,252,156,212,236,55,80,141,55,
87,166,26,7,30,122,237,1,211,70,196,220,
243,159,96,91,22,81,34,72,20,147,116,176,
73,197,109,227,6,49,155,129,142,156,27,67,
111,60,97,36,107,34,68,130,231,121,252,87,
191,241,75,212,235,117,178,217,247,80,213,151,
227,176,180,188,66,218,182,200,102,179,135,254,
254,59,183,102,185,115,107,118,223,191,147,101,
153,239,125,243,171,199,122,159,102,179,73,173,
209,226,179,199,79,0,193,102,59,66,136,132,
68,82,112,66,16,118,5,45,93,60,119,235,
76,85,36,110,12,217,20,109,157,231,27,14,
225,86,23,125,221,206,95,74,205,236,203,12,
89,146,184,53,156,70,63,161,0,81,180,250,
144,127,242,171,95,199,52,77,62,253,226,33,
127,250,249,10,73,106,8,173,146,70,209,77,
98,235,107,40,193,235,151,253,188,50,20,7,
133,49,116,133,225,87,146,190,57,75,237,202,
56,138,151,238,229,81,66,44,57,59,213,51,
18,219,27,224,87,63,124,151,39,243,43,212,
234,139,248,128,146,29,69,77,165,207,56,226,
211,221,112,194,245,167,24,82,128,18,121,124,
105,118,140,247,190,249,75,132,97,136,16,2,
203,234,26,173,78,167,131,109,191,204,101,52,
154,77,238,63,122,202,189,143,62,34,109,239,
206,113,20,139,123,55,222,177,209,145,55,50,
103,217,108,150,108,54,203,244,228,248,222,15,
48,138,248,252,193,99,22,214,30,179,212,217,
164,211,94,67,207,13,161,228,70,207,205,154,
43,216,26,153,84,150,23,155,14,27,173,128,
92,167,129,22,135,56,153,202,213,7,121,81,
66,78,138,116,34,35,145,196,33,202,234,207,
121,107,172,136,105,154,196,113,204,143,190,88,
68,27,186,77,232,180,122,215,73,178,130,98,
190,254,60,227,149,161,216,111,195,73,169,220,
28,78,239,73,78,47,213,188,93,101,173,59,
75,96,15,194,74,71,162,217,108,238,58,93,
87,171,85,174,141,150,248,205,247,222,70,146,
36,62,187,255,136,31,60,89,1,35,139,145,
63,163,77,225,201,127,98,186,50,196,183,127,
241,75,20,11,187,217,91,119,122,3,192,46,
35,1,144,203,102,249,234,151,222,191,216,31,
176,170,242,238,189,219,188,123,239,165,215,243,
7,63,248,41,65,118,228,92,41,61,169,178,
196,76,197,102,36,103,82,75,138,52,79,32,
26,47,73,160,72,221,48,157,34,75,8,1,
137,16,221,159,8,132,44,117,255,253,234,83,
31,56,130,40,33,138,5,170,114,188,53,37,
43,26,229,140,206,47,127,227,43,0,44,44,
44,160,7,85,234,75,95,160,228,39,222,252,
186,188,154,218,87,60,128,148,202,205,145,204,
158,189,195,11,19,150,27,222,30,15,99,63,
248,245,85,140,252,48,254,252,207,240,125,159,
249,165,21,238,237,48,20,165,82,137,111,148,
94,118,188,125,248,222,219,164,237,167,164,76,
157,251,79,158,81,147,59,232,78,139,148,38,
179,214,244,105,5,18,137,98,32,103,134,80,
76,27,73,234,255,196,18,182,107,40,237,101,
190,252,254,59,124,249,189,183,175,38,120,135,
215,243,91,191,250,109,254,245,239,255,25,126,
249,30,138,182,127,9,163,72,226,55,82,191,
110,233,10,101,51,1,25,30,23,82,108,180,
3,188,173,202,59,89,146,48,117,25,75,87,
176,116,133,148,174,96,168,50,81,59,196,202,
231,187,137,206,195,214,132,219,93,183,178,97,
227,110,209,195,244,254,248,241,223,10,193,160,
179,12,61,245,91,253,40,146,152,200,109,131,
44,19,53,86,120,231,189,235,56,142,195,191,
253,195,63,99,161,234,98,100,43,164,70,110,
19,180,235,111,252,189,174,12,197,14,140,228,
12,202,210,254,201,105,83,147,185,53,146,166,
229,70,248,81,130,161,201,228,147,189,167,189,
176,190,68,174,249,136,78,42,3,118,153,89,
123,133,122,253,104,225,147,91,179,51,0,76,
78,236,127,122,112,93,151,7,143,159,177,178,
177,194,226,90,149,104,236,203,135,110,96,73,
20,80,108,222,231,131,187,51,220,184,254,45,
100,89,190,154,224,87,55,99,203,226,215,190,
243,101,158,189,88,228,15,255,244,135,228,166,
223,69,18,49,36,33,202,208,109,0,26,15,
126,64,246,214,47,244,45,6,51,72,248,166,
133,172,27,140,21,76,198,10,38,113,34,136,
133,64,87,246,159,75,79,150,142,52,18,59,
161,200,82,143,27,172,183,121,9,232,4,17,
45,55,162,233,69,180,61,137,248,106,169,244,
141,76,74,237,219,65,77,117,230,73,69,6,
43,205,128,148,183,198,236,204,215,248,183,255,
254,247,169,102,238,146,45,159,47,138,157,75,
109,40,100,73,194,210,21,116,52,178,101,139,
148,174,236,178,246,81,44,122,100,123,217,148,
74,209,214,15,228,78,234,122,27,26,185,29,
229,110,78,245,165,165,79,226,136,120,253,49,
186,95,229,191,251,71,191,201,239,253,155,127,
135,215,113,249,242,247,191,201,196,196,248,169,
223,37,149,74,241,254,59,119,1,248,223,254,
143,127,69,208,105,96,100,14,78,204,134,141,
21,190,254,225,237,3,13,207,21,186,168,148,
203,148,138,69,238,220,156,65,150,101,84,85,
101,109,99,147,249,197,21,86,157,58,234,84,
153,245,181,143,145,116,187,27,86,64,69,41,
207,188,22,195,161,5,30,74,20,18,238,216,
216,21,206,54,76,38,73,144,54,84,210,134,
202,40,224,166,92,18,51,139,187,245,157,120,
97,210,251,249,183,193,243,144,164,110,81,139,
171,42,123,56,187,50,166,186,43,196,36,233,
58,227,197,254,153,90,67,217,196,111,181,25,
75,9,102,238,188,7,192,199,15,158,50,252,
181,47,157,187,113,184,20,134,66,2,12,77,
38,181,229,134,119,93,114,21,83,147,183,54,
244,54,214,1,221,145,217,212,233,135,32,108,
87,49,27,79,248,221,95,255,110,47,113,253,
207,255,209,175,19,69,17,138,50,216,176,69,
189,209,164,221,118,208,70,246,63,85,198,161,
143,178,249,144,175,206,12,95,25,137,126,15,
20,178,76,38,147,233,253,251,228,248,24,147,
227,99,251,94,219,106,181,249,244,139,71,84,
91,14,43,237,152,164,124,251,204,158,43,210,
116,18,69,121,227,223,150,109,40,216,198,222,
231,8,162,4,47,76,240,163,24,175,25,16,
235,6,177,80,8,99,65,20,119,99,245,97,
146,240,166,236,201,182,119,117,82,131,166,171,
50,183,71,211,224,68,248,90,138,7,43,221,
138,199,180,169,50,83,177,246,228,40,67,23,
100,165,127,207,61,182,71,208,202,105,212,230,
23,188,115,231,38,0,99,35,35,68,113,140,
164,156,175,173,249,66,25,10,73,234,118,53,
155,154,130,166,153,100,138,54,166,166,144,210,
148,55,146,143,76,156,26,5,185,202,189,27,
19,220,121,235,87,247,48,63,190,154,36,30,
4,238,63,126,134,249,214,119,118,181,237,199,
129,71,180,248,49,88,69,148,230,2,255,252,
31,255,195,129,27,168,43,108,157,34,51,105,
126,225,163,15,0,248,228,231,159,243,39,115,
171,152,249,179,105,124,82,163,16,57,142,56,
175,26,133,186,42,111,85,248,168,4,146,129,
154,178,246,245,180,226,68,244,140,71,152,8,
2,79,38,10,67,132,102,130,4,50,18,72,
221,239,91,66,218,250,217,253,247,208,48,49,
114,54,130,46,5,119,34,94,254,12,227,164,
151,99,217,175,97,241,90,197,194,214,21,238,
47,183,123,37,199,199,49,50,55,135,211,152,
154,130,7,100,83,26,138,44,97,168,50,111,
141,164,251,206,67,244,99,138,151,22,23,49,
205,110,95,85,189,229,96,206,255,4,237,218,
87,207,213,92,159,91,67,161,171,221,100,157,
109,116,189,131,148,174,96,104,114,207,241,246,
26,14,166,173,159,217,239,247,86,31,99,84,
174,239,155,7,16,73,140,177,246,49,223,127,
239,6,247,110,223,124,173,227,242,181,47,189,
199,227,127,255,103,56,91,134,66,120,77,174,
101,28,190,251,223,252,10,65,16,224,121,183,
175,140,196,107,194,187,111,223,5,233,62,159,
60,249,148,78,43,34,8,218,8,89,70,203,
143,17,52,55,81,77,11,53,149,61,249,65,
68,86,224,18,132,119,182,171,175,216,242,240,
99,213,32,137,21,180,212,209,97,26,39,212,
176,210,71,127,231,65,156,176,178,208,164,38,
201,248,91,172,203,219,116,252,215,42,22,143,
86,218,199,122,230,235,21,107,151,23,37,73,
221,242,229,177,124,106,128,70,2,36,89,230,
253,183,239,244,14,149,255,211,63,255,109,254,
207,127,247,31,241,206,217,28,158,11,67,97,
106,50,150,174,162,43,6,185,161,52,182,174,
30,187,188,108,144,8,218,117,70,229,26,180,
2,214,99,27,181,56,189,59,36,224,117,248,
230,189,235,220,125,205,70,98,27,5,75,97,
91,60,49,39,187,252,189,239,252,93,0,52,
77,219,83,218,122,133,51,54,22,247,110,243,
238,189,219,8,33,16,66,208,104,52,249,236,
225,19,174,125,48,194,255,251,151,143,225,20,
134,226,10,199,56,88,42,50,229,140,206,84,
49,71,221,9,187,93,238,98,251,123,209,40,
166,117,170,237,254,124,179,209,188,73,105,31,
227,52,83,25,204,183,37,146,152,246,163,31,
162,143,222,33,172,45,49,122,119,168,247,119,
154,166,33,233,38,157,39,63,194,156,124,111,
32,234,116,131,192,107,45,133,217,214,83,40,
103,116,166,74,22,119,198,50,124,233,90,158,
119,39,115,204,14,219,148,51,58,185,148,246,
70,141,4,128,158,206,211,22,38,255,245,175,
126,135,91,249,4,111,237,9,201,242,207,137,
253,238,246,108,181,231,184,243,214,173,55,246,
124,105,83,71,196,221,174,227,129,50,83,94,
225,20,107,91,66,150,101,10,133,60,191,248,
213,47,177,81,107,34,231,198,78,119,79,145,
32,93,113,61,29,27,121,75,227,198,144,189,
171,97,118,166,98,245,69,175,159,179,52,38,
139,103,183,57,71,94,155,66,227,115,190,249,
222,12,238,250,115,126,229,237,34,183,95,97,
39,24,206,232,88,186,74,84,91,184,252,30,
197,182,81,48,182,42,142,108,163,155,104,150,
207,33,123,158,72,98,226,141,39,20,244,132,
60,62,206,70,157,196,235,118,67,126,247,155,
95,101,230,249,255,223,222,217,236,68,13,69,
1,248,235,109,103,218,233,212,153,129,128,146,
152,12,152,32,46,212,96,134,24,37,98,72,
136,11,23,18,214,186,50,190,130,79,224,43,
184,117,39,15,96,76,92,185,211,200,66,87,
70,9,2,25,35,138,38,4,73,181,210,249,
233,116,122,235,162,136,252,201,20,24,226,44,
250,173,155,211,228,156,222,211,123,239,249,251,
204,153,254,34,143,159,60,167,146,59,203,208,
169,92,188,113,133,199,196,229,75,231,153,123,
246,10,47,76,113,99,106,52,241,12,29,200,
226,151,239,104,249,163,5,186,165,154,100,175,
183,109,71,172,40,12,245,89,172,56,81,115,
207,112,207,155,13,149,193,147,199,123,34,15,
126,126,101,234,214,24,66,8,102,102,151,89,
119,119,55,26,29,191,90,226,253,244,83,244,
66,231,164,200,106,0,165,254,2,78,205,199,
94,249,65,69,19,187,38,171,197,49,130,153,
86,49,55,178,35,178,186,182,25,96,174,218,
149,127,102,28,181,77,249,190,183,89,48,37,
131,38,66,168,160,40,173,139,165,194,144,96,
117,129,98,94,48,113,115,132,204,150,59,83,
207,243,254,238,70,6,162,171,167,59,147,19,
188,152,121,205,248,232,181,255,106,52,211,52,
185,59,121,157,15,139,229,205,32,88,66,103,
113,113,240,52,47,23,150,80,186,7,14,45,
67,245,27,168,65,19,47,81,103,219,232,203,
27,228,50,41,202,75,206,182,56,192,65,71,
5,28,218,166,93,69,222,206,206,115,101,100,
152,92,86,167,252,109,141,210,142,166,7,186,
174,211,211,219,139,107,21,59,231,196,108,89,
86,120,239,193,163,200,57,174,219,232,39,186,
145,97,136,223,220,200,84,144,18,41,163,242,
255,40,243,0,2,191,134,80,84,244,140,129,
161,9,210,91,130,204,59,225,178,57,96,0,
0,0,216,73,68,65,84,249,35,51,14,113,
159,13,165,164,177,178,64,90,19,8,36,61,
57,19,183,86,71,17,26,94,189,138,153,49,
144,50,36,157,214,112,235,1,191,60,137,106,
245,146,202,22,80,54,10,207,164,107,99,120,
171,148,46,156,219,214,139,41,33,225,168,120,
94,131,249,79,203,216,97,14,177,199,80,153,
184,223,185,112,108,148,80,18,20,226,117,143,
109,84,28,210,217,214,67,108,154,94,228,34,
53,221,104,155,76,191,86,65,213,141,104,147,
214,234,164,228,55,144,82,198,122,255,113,248,
15,0,207,117,168,10,147,106,61,192,212,5,
93,251,36,198,28,68,175,66,8,68,106,255,
224,187,92,251,200,216,240,32,111,102,203,132,
34,197,216,240,238,198,152,239,230,22,177,51,
253,40,138,18,233,86,75,181,148,123,80,29,
52,42,14,90,198,106,105,179,135,247,111,71,
63,10,215,117,147,213,157,144,144,144,144,176,
39,191,1,229,210,15,128,154,237,161,94,0,
0,0,0,73,69,78,68,174,66,96,130,
};
  static imagedef_t images[]= {
 {394, 261, img0data},
};
  canvas->set_images(images, 1);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_0);
  tf_0(canvas);
}
static void tf_1(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,2);
  canvas->pbreak();
  canvas->println(3,25);
  canvas->pbreak();
  canvas->println(26,43);
  canvas->verbatim_vspace();
  canvas->verbatim_line(44,44);
  canvas->verbatim_vspace();
  canvas->println(45,102);
  canvas->pbreak();
  canvas->println(103,110);
  canvas->bullet_println(1, 111, 120);
  canvas->bullet_println(1, 121, 139);
  canvas->bullet_println(1, 140, 160);
  canvas->println(161,177);

 canvas->end_doc(tf_1);
}
static void cf_1(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xb,0x4b,0xb,0x4c,0xb,0x4d,0x0,0x15,0x0,0x4e,0x0,0x4f,0x0,0x50,0x0,0x51,0x0,0x52,0x0,0x53,0x0,0x54,0x0,0x55,0x0,0x56,0x0,0x51,0x0,0x52,0x0,0x57,0x0,0x58,0x0,0x59,0x0,0x5a,0x0,0x5b,0x0,0x5c,0x0,0x4e,0x0,0x5b,0x0,0x5d,0x0,0x54,0x0,0x5e,0x0,0x5f,0x0,0x4e,0x0,0x16,0x0,0x60,0x40000,0x61,0x0,0x28,0x0,0x62,0x0,0x16,0x0,0x63,0x0,0x64,0x0,0x65,0x0,0x18,0x0,0x24,0x0,0x66,0x0,0x52,0x0,0x5b,0x0,0x59,0x0,0x67,0x50004,0x68,0x0,0x5f,0x0,0x69,0x0,0x6a,0x0,0x16,0x0,0x6b,0x0,0x5b,0x0,0x65,0x0,0x6c,0x0,0x6d,0x0,0x6e,0x0,0x6f,0x0,0x70,0x0,0x71,0x0,0x72,0x0,0x73,0x0,0x65,0x0,0x74,0x0,0x65,0x0,0x75,0x0,0x76,0x0,0x64,0x0,0x3e,0x0,0x77,0x0,0x78,0x0,0x79,0x0,0x65,0x0,0x7a,0x0,0x7b,0x0,0x7c,0x0,0x1d,0x0,0x65,0x0,0x75,0x0,0x1a,0x0,0x59,0x0,0x7d,0x0,0x7e,0x0,0x7f,0x0,0x65,0x0,0x80,0x0,0x81,0x0,0x15,0x0,0x16,0x0,0x82,0x0,0x65,0x0,0x4e,0x0,0x16,0x0,0x83,0x0,0x84,0x0,0x85,0x0,0x86,0x0,0x87,0x0,0x88,0x0,0x89,0x0,0x8a,0x0,0x7a,0x0,0x8b,0x0,0x8c,0x0,0x8d,0x0,0x8e,0x0,0x70,0x0,0x6,0x0,0x8f,0x0,0x7a,0x0,0x90,0x0,0x91,0x0,0x92,0x0,0x0,0x40000,0x93,0x0,0x94,0x0,0x95,0x0,0x96,0x0,0x88,0x0,0x97,0x0,0x8f,0x0,0x7a,0x0,0x98,0x0,0x0,0x40000,0x99,0x0,0x94,0x0,0x9a,0x0,0x9b,0x0,0x9c,0x0,0x65,0x0,0x9d,0x0,0x54,0x0,0x9e,0x0,0x9f,0x0,0x1a,0x0,0xa0,0x0,0x3e,0x0,0xa1,0x0,0xa2,0x0,0xa3,0x0,0xa4,0x0,0xa5,0x0,0x0,0x40000,0xa6,0x40000,0xa7,0x0,0xa8,0x0,0x13,0x0,0xa9,0x0,0xaa,0x0,0x64,0x0,0x65,0x0,0x7c,0x0,0xab,0x0,0x6e,0x0,0x9a,0x0,0xac,0x0,0xad,0x0,0xae,0x0,0xaf,0x0,0xa9,0x0,0xaa,0x0,0x16,0x0,0xb0,0x0,0xb1,0x0,0xb2,0x0,0xb3,0x0,0xb4,0x0,0x52,0x0,0x5b,0x4,0xb5,0x0,0xb6,0x0,0xb7,0x0,0x70,0x0,0xb8,0x0,0x88,0x0,0xb9,0x0,0x91,0x0,0x13,0x0,0xba,0x0,0xbb,};
  canvas->set_elts(seq,178);
static const unsigned int bgrects[]= {
0x1e,0x1e,0x4,0x70,0x70,0x4,0x7a,0x7a,0x4,0x8d,0x8e,0x4,};
  canvas->set_bgrects(bgrects, 4);
  canvas->set_underlines(NULL, 0);
static const unsigned int verboxes[]= {
0x2c,0x2c,0x5,};
  canvas->set_vboxes(verboxes, 1);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_1);
  tf_1(canvas);
}
static void tf_2(FYGfxCanvas* canvas) {
  int *TL_ROW= canvas->get_row_lines(7);
  int *TL_COL= canvas->get_col_lines(4);
canvas->reset_doc();
  canvas->println(0,1);
  canvas->pbreak();
  canvas->println(2,17);
  canvas->bullet_println(1, 18, 25);
  canvas->bullet_println(1, 26, 27);
  canvas->bullet_println(1, 28, 29);
  canvas->bullet_println(1, 30, 31);
  canvas->bullet_println(1, 32, 33);
  canvas->bullet_println(1, 34, 42);
  canvas->bullet_println(1, 43, 45);
  canvas->bullet_println(2, 46, 51);
  canvas->bullet_println(2, 52, 53);
  canvas->bullet_println(2, 54, 58);
  canvas->println(59,71);
  canvas->verbatim_vspace();
  canvas->verbatim_line(72,73);
  canvas->verbatim_vspace();
  canvas->println(74,109);
  canvas->pbreak();
  canvas->println(110,128);
  canvas->pbreak();
  canvas->println(129,160);
  canvas->pbreak();
  canvas->println(161,162);
  canvas->pbreak();
  canvas->println(163,190);
 /// BEGIN TABLE 
 {
  int cw0, ch0,ca0;
  canvas->get_seq_metrics(191,191, &cw0, &ch0, &ca0);
  int cw1, ch1,ca1;
  canvas->get_seq_metrics(192,192, &cw1, &ch1, &ca1);
  int cw2, ch2,ca2;
  canvas->get_seq_metrics(193,193, &cw2, &ch2, &ca2);
  int cw3, ch3,ca3;
  canvas->get_seq_metrics(194,194, &cw3, &ch3, &ca3);
  int cw4, ch4,ca4;
  canvas->get_seq_metrics(195,195, &cw4, &ch4, &ca4);
  int cw5, ch5,ca5;
  canvas->get_seq_metrics(196,198, &cw5, &ch5, &ca5);
  int cw6, ch6,ca6;
  canvas->get_seq_metrics(199,199, &cw6, &ch6, &ca6);
  int cw7, ch7,ca7;
  canvas->get_seq_metrics(200,200, &cw7, &ch7, &ca7);
  int cw8, ch8,ca8;
  canvas->get_seq_metrics(201,203, &cw8, &ch8, &ca8);
  int cw9, ch9,ca9;
  canvas->get_seq_metrics(204,204, &cw9, &ch9, &ca9);
  int cw10, ch10,ca10;
  canvas->get_seq_metrics(205,205, &cw10, &ch10, &ca10);
  int cw11, ch11,ca11;
  canvas->get_seq_metrics(206,208, &cw11, &ch11, &ca11);
  int cw12, ch12,ca12;
  canvas->get_seq_metrics(209,209, &cw12, &ch12, &ca12);
  int cw13, ch13,ca13;
  canvas->get_seq_metrics(210,210, &cw13, &ch13, &ca13);
  int cw14, ch14,ca14;
  canvas->get_seq_metrics(211,213, &cw14, &ch14, &ca14);
  int cw15, ch15,ca15;
  canvas->get_seq_metrics(214,214, &cw15, &ch15, &ca15);
  int cw16, ch16,ca16;
  canvas->get_seq_metrics(215,215, &cw16, &ch16, &ca16);
  int cw17, ch17,ca17;
  canvas->get_seq_metrics(216,216, &cw17, &ch17, &ca17);
 int rh0=0, ra0=0;
 int rh1=0, ra1=0;
 int rh2=0, ra2=0;
 int rh3=0, ra3=0;
 int rh4=0, ra4=0;
 int rh5=0, ra5=0;
  rh0= max_i(rh0, ch0); ra0= max_i(ra0, ca0);
  rh0= max_i(rh0, ch1); ra0= max_i(ra0, ca1);
  rh0= max_i(rh0, ch2); ra0= max_i(ra0, ca2);
  rh1= max_i(rh1, ch3); ra1= max_i(ra1, ca3);
  rh1= max_i(rh1, ch4); ra1= max_i(ra1, ca4);
  rh1= max_i(rh1, ch5); ra1= max_i(ra1, ca5);
  rh2= max_i(rh2, ch6); ra2= max_i(ra2, ca6);
  rh2= max_i(rh2, ch7); ra2= max_i(ra2, ca7);
  rh2= max_i(rh2, ch8); ra2= max_i(ra2, ca8);
  rh3= max_i(rh3, ch9); ra3= max_i(ra3, ca9);
  rh3= max_i(rh3, ch10); ra3= max_i(ra3, ca10);
  rh3= max_i(rh3, ch11); ra3= max_i(ra3, ca11);
  rh4= max_i(rh4, ch12); ra4= max_i(ra4, ca12);
  rh4= max_i(rh4, ch13); ra4= max_i(ra4, ca13);
  rh4= max_i(rh4, ch14); ra4= max_i(ra4, ca14);
  rh5= max_i(rh5, ch15); ra5= max_i(ra5, ca15);
  rh5= max_i(rh5, ch16); ra5= max_i(ra5, ca16);
  rh5= max_i(rh5, ch17); ra5= max_i(ra5, ca17);
  int cow0=0;
  int cow1=0;
  int cow2=0;
  cow0= max_i(cow0, cw0);
  cow0= max_i(cow0, cw3);
  cow0= max_i(cow0, cw6);
  cow0= max_i(cow0, cw9);
  cow0= max_i(cow0, cw12);
  cow0= max_i(cow0, cw15);
  cow1= max_i(cow1, cw1);
  cow1= max_i(cow1, cw4);
  cow1= max_i(cow1, cw7);
  cow1= max_i(cow1, cw10);
  cow1= max_i(cow1, cw13);
  cow1= max_i(cow1, cw16);
  cow2= max_i(cow2, cw2);
  cow2= max_i(cow2, cw5);
  cow2= max_i(cow2, cw8);
  cow2= max_i(cow2, cw11);
  cow2= max_i(cow2, cw14);
  cow2= max_i(cow2, cw17);
  int D=0;
  int rp0;
  int rp1;
  int rp2;
  int rp3;
  int rp4;
  int rp5;
  int rp6;
  int cp0;
  int cp1;
  int cp2;
  int cp3;
  cp0= rp0= 0;
  rp1= rp0 + 2*CELL_PAD + rh0;
  rp2= rp1 + 2*CELL_PAD + rh1;
  rp3= rp2 + 2*CELL_PAD + rh2;
  rp4= rp3 + 2*CELL_PAD + rh3;
  rp5= rp4 + 2*CELL_PAD + rh4;
  rp6= rp5 + 2*CELL_PAD + rh5;
  cp1= cp0 + 2*CELL_PAD + cow0;
  cp2= cp1 + 2*CELL_PAD + cow1;
  cp3= cp2 + 2*CELL_PAD + cow2;
  int off_x= canvas->get_width() - cp3;
  if (off_x<0) off_x= 0; else off_x/= 2;
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 191, 191);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 192, 192);
  canvas->cell_center(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 193, 193);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 194, 194);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 195, 195);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 196, 198);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 199, 199);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 200, 200);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 201, 203);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 204, 204);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 205, 205);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 206, 208);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 209, 209);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 210, 210);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 211, 213);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 214, 214);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 215, 215);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 216, 216);
  int off_y= canvas->get_cursor_y();
  TL_ROW[0]= off_y+rp0; TL_ROW[1]= off_x+cp0; TL_ROW[2]= off_x+cp3;
  TL_ROW[3]= off_y+rp1; TL_ROW[4]= off_x+cp0; TL_ROW[5]= off_x+cp3;
  TL_ROW[6]= off_y+rp2; TL_ROW[7]= off_x+cp0; TL_ROW[8]= off_x+cp3;
  TL_ROW[9]= off_y+rp3; TL_ROW[10]= off_x+cp0; TL_ROW[11]= off_x+cp3;
  TL_ROW[12]= off_y+rp4; TL_ROW[13]= off_x+cp0; TL_ROW[14]= off_x+cp3;
  TL_ROW[15]= off_y+rp5; TL_ROW[16]= off_x+cp0; TL_ROW[17]= off_x+cp3;
  TL_ROW[18]= off_y+rp6; TL_ROW[19]= off_x+cp0; TL_ROW[20]= off_x+cp3;
  TL_COL[0]= off_y+rp0; TL_COL[1]= off_y+rp6; TL_COL[2]= off_x+cp0;
  TL_COL[3]= off_y+rp0; TL_COL[4]= off_y+rp6; TL_COL[5]= off_x+cp1;
  TL_COL[6]= off_y+rp0; TL_COL[7]= off_y+rp6; TL_COL[8]= off_x+cp2;
  TL_COL[9]= off_y+rp0; TL_COL[10]= off_y+rp6; TL_COL[11]= off_x+cp3;
  canvas->xcoord(cp3+3);
  canvas->advance_y(rp6+2);
 /// END TABLE 
 }
  canvas->println(217,253);

 canvas->end_doc(tf_2);
}
static void cf_2(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xb,0xbc,0xb,0xbd,0x0,0xbe,0x0,0x66,0x0,0x16,0x0,0x13,0x0,0xbf,0x0,0x36,0x0,0xaa,0x0,0x64,0x0,0x6,0x0,0xc0,0x0,0x5f,0x0,0xc1,0x0,0xc2,0x0,0xc3,0x0,0x70,0x0,0xc4,0x0,0x0,0x0,0xc5,0x0,0xc6,0x0,0x88,0x0,0xc7,0x0,0xc8,0x0,0x24,0x0,0xc9,0x0,0x0,0x0,0xca,0x0,0x0,0x0,0xcb,0x0,0x0,0x0,0xcc,0x0,0x0,0x0,0xcd,0x0,0x0,0x0,0xce,0x0,0x36,0x0,0xcf,0x0,0x6a,0x0,0xc,0x0,0xd0,0x0,0x8f,0x0,0xd1,0x0,0x0,0x0,0xd2,0x0,0xc2,0x0,0x0,0x0,0xd3,0x0,0xd4,0x0,0x88,0x0,0xd5,0x0,0xd6,0x0,0x0,0x0,0xd7,0x0,0x0,0x0,0xd8,0x0,0xd9,0x0,0x88,0x0,0xda,0x0,0xdb,0x0,0x16,0x0,0xdc,0x0,0x9c,0x0,0xdd,0x0,0xde,0x0,0x5f,0x0,0xdf,0x0,0xe0,0x0,0xa3,0x0,0x13,0x0,0xe1,0x0,0xe2,0x4,0xe3,0x50004,0xe4,0x0,0xe5,0x40000,0xe1,0x0,0x16,0x0,0x13,0x0,0xe6,0x0,0xe7,0x0,0x24,0x0,0xe8,0x0,0xe9,0x40000,0xea,0x0,0x16,0x0,0xeb,0x0,0x65,0x0,0x36,0x0,0xb2,0x0,0xec,0x0,0x5b,0x0,0xe0,0x0,0xc,0x0,0x6,0x0,0xed,0x0,0xee,0x0,0x65,0x0,0xef,0x0,0xf0,0x0,0xf1,0x0,0xed,0x0,0x5b,0x40000,0xf2,0x0,0xf3,0x0,0x70,0x0,0x3e,0x0,0x7,0x0,0x5b,0x0,0x65,0x0,0xf4,0x0,0x8e,0x0,0x70,0x0,0xc7,0x0,0xf5,0x0,0x24,0x0,0xf6,0x1000200,0xf7,0x0,0x88,0x0,0xf8,0x0,0xd2,0x0,0xf3,0x0,0xf9,0x0,0xe0,0x0,0xfa,0x0,0xed,0x0,0x84,0x0,0x13,0x0,0xfb,0x0,0xfc,0x0,0xfd,0x0,0xf3,0x0,0xfe,0x0,0x13,0x0,0x66,0x0,0xff,0x0,0x100,0x0,0x91,0x0,0x13,0x0,0x101,0x0,0x102,0x0,0x5f,0x0,0x66,0x0,0xe0,0x0,0x16,0x0,0x7a,0x0,0x103,0x0,0x104,0x0,0x105,0x0,0xf3,0x0,0x70,0x0,0x106,0x0,0x107,0x0,0x108,0x0,0xb7,0x0,0x70,0x0,0x109,0x0,0x65,0x0,0x10a,0x0,0x24,0x0,0x13,0x0,0x10b,0x9,0x10c,0x9,0x10d,0x0,0x10e,0x0,0x10f,0x0,0x5b,0x0,0xa2,0x0,0x110,0x0,0x111,0x0,0x112,0x0,0x113,0x0,0xc,0x0,0x114,0x0,0x13,0x0,0xe1,0x0,0x115,0x0,0xb2,0x0,0x58,0x0,0x116,0x0,0x65,0x0,0x117,0x0,0x118,0x0,0x56,0x0,0x22,0x0,0x119,0x0,0x11a,0x0,0xb2,0x0,0x11b,0x0,0x11c,0x0,0x11d,0x0,0x11e,0x0,0x11f,0x0,0x120,0x0,0x121,0x0,0x122,0x0,0x123,0x0,0x11d,0x0,0x124,0x0,0x125,0x0,0x126,0x0,0x127,0x0,0x11d,0x0,0x128,0x0,0x125,0x0,0x129,0x0,0x12a,0x0,0x11d,0x0,0x124,0x0,0x12b,0x0,0x12c,0x0,0x12d,0x0,0x11d,0x0,0x128,0x0,0x12b,0x0,0x12e,0x0,0x12f,0x0,0x130,0x0,0x131,0x0,0x132,0x0,0x70,0x0,0x133,0x0,0x134,0x0,0x24,0x0,0x65,0x0,0x45,0x0,0x135,0x0,0x88,0x0,0x70,0x0,0x136,0x0,0xa3,0x0,0x137,0x0,0x65,0x0,0x36,0x0,0x138,0x0,0x139,0x0,0x6e,0x0,0xb2,0x0,0xec,0x0,0x5b,0x0,0x13a,0x0,0x13,0x0,0x13b,0x0,0x130,0x0,0xf1,0x0,0x13c,0x0,0x84,0x0,0x5c,0x0,0x13d,0x0,0xb2,0x0,0x58,0x0,0x116,0x0,0x9a,0x0,0x117,0x0,0x13e,};
  canvas->set_elts(seq,254);
static const unsigned int bgrects[]= {
0x4b,0x4b,0x4,0x53,0x53,0x4,0x66,0x66,0x4,};
  canvas->set_bgrects(bgrects, 3);
static const unsigned int underlines[]= {
0x74,0x74,0x2,};
  canvas->set_underlines(underlines, 1);
static const unsigned int verboxes[]= {
0x48,0x49,0x5,};
  canvas->set_vboxes(verboxes, 1);
  canvas->set_images(NULL,0);
static const unsigned int links[]= {
0x74,0x74,0xc,};
  canvas->set_links(links, 1);
  win->set_section_menu(secmen_2);
  tf_2(canvas);
}
static void tf_4(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,2);
  canvas->pbreak();
  canvas->println(3,26);
  canvas->pbreak();
  canvas->println(27,67);
  canvas->verbatim_vspace();
  canvas->verbatim_line(68,68);
  canvas->verbatim_line(69,69);
  canvas->verbatim_line(70,70);
  canvas->verbatim_line(71,71);
  canvas->verbatim_vspace();
  canvas->println(72,86);

 canvas->end_doc(tf_4);
}
static void cf_4(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x13f,0xa,0x140,0xa,0x141,0x0,0x5f,0x0,0x66,0x0,0x106,0x0,0x1d,0x0,0x61,0x0,0x16,0x0,0x142,0x0,0x5b,0x0,0x59,0x0,0x13,0x0,0x143,0x0,0x5f,0x0,0x144,0x0,0x145,0x0,0xc6,0x0,0x70,0x0,0x146,0x0,0x88,0x0,0x70,0x0,0x147,0x0,0x84,0x0,0x65,0x0,0x146,0x0,0x148,0x0,0x10e,0x0,0x10f,0x0,0x5b,0x0,0x149,0x0,0x13,0x0,0x14a,0x0,0x14b,0x0,0xc,0x0,0x13,0x0,0x14c,0x0,0xb2,0x0,0xa0,0x0,0x5b,0x0,0xb3,0x0,0x13,0x0,0x14d,0x0,0x14e,0x0,0x14f,0x0,0x14d,0x0,0x150,0x0,0x151,0x0,0x64,0x0,0x13,0x0,0x7b,0x0,0x24,0x0,0x152,0x0,0x111,0x0,0x88,0x0,0x16,0x0,0x153,0x0,0x1d,0x0,0x65,0x0,0x150,0x0,0x154,0x0,0x155,0x0,0xc8,0x0,0x24,0x0,0x14d,0x0,0x156,0x0,0x70,0x0,0x157,0x50004,0x158,0x50004,0x159,0x50004,0x15a,0x50004,0x15b,0x0,0x56,0x0,0x156,0x0,0x15c,0x0,0x15d,0x0,0x109,0x0,0x65,0x0,0x10a,0x0,0x24,0x0,0x13,0x0,0x7a,0x0,0x7e,0x0,0x15e,0x0,0x15f,0x0,0x16,0x0,0x160,};
  canvas->set_elts(seq,87);
  canvas->set_bgrects(NULL, 0);
  canvas->set_underlines(NULL, 0);
static const unsigned int verboxes[]= {
0x44,0x47,0x5,};
  canvas->set_vboxes(verboxes, 1);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_2);
  tf_4(canvas);
}
static void tf_5(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,0);
  canvas->pbreak();
  canvas->println(1,42);
  canvas->pbreak();
  canvas->println(43,62);

 canvas->end_doc(tf_5);
}
static void cf_5(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x161,0x0,0x162,0x0,0x11a,0x0,0xad,0x0,0xff,0x0,0x24,0x0,0x163,0x0,0xe1,0x0,0x16,0x0,0x142,0x0,0x5b,0x0,0x59,0x0,0xff,0x0,0x24,0x0,0x13,0x0,0x164,0x0,0x14f,0x0,0x165,0x0,0x166,0x0,0x108,0x0,0x13,0x0,0x14a,0x0,0xe1,0x0,0x151,0x0,0x109,0x0,0x65,0x0,0x167,0x0,0x7a,0x0,0xc,0x0,0x168,0x0,0x65,0x0,0x167,0x0,0x7a,0x0,0x169,0x0,0x16a,0x0,0x7e,0x0,0x16b,0x0,0x16c,0x0,0x70,0x0,0xe6,0x0,0x16d,0x0,0x108,0x0,0x16e,0x0,0xd2,0x0,0xc2,0x0,0x113,0x0,0x84,0x0,0x13,0x0,0x165,0x0,0x15c,0x0,0x16f,0x0,0x170,0x0,0x65,0x0,0x164,0x0,0x171,0x0,0x3e,0x0,0x77,0x0,0x1a,0x0,0x59,0x0,0x7c,0x0,0x1d,0x0,0x65,0x0,0x172,};
  canvas->set_elts(seq,63);
  canvas->set_bgrects(NULL, 0);
  canvas->set_underlines(NULL, 0);
  canvas->set_vboxes(NULL, 0);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_2);
  tf_5(canvas);
}
static void tf_6(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,0);
  canvas->pbreak();
  canvas->println(1,36);
  canvas->pbreak();
  canvas->println(37,40);
  canvas->verbatim_vspace();
  canvas->verbatim_line(41,41);
  canvas->verbatim_line(42,42);
  canvas->verbatim_line(43,43);
  canvas->verbatim_line(44,44);
  canvas->verbatim_line(45,45);
  canvas->verbatim_vspace();
  canvas->println(46,50);
  canvas->bullet_println(1, 51, 55);
  canvas->bullet_println(2, 56, 60);
  canvas->bullet_println(2, 61, 65);
  canvas->bullet_println(1, 66, 70);
  canvas->bullet_println(1, 71, 75);
  canvas->println(76,126);

 canvas->end_doc(tf_6);
}
static void cf_6(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x173,0x0,0x174,0x0,0x175,0x0,0x176,0x0,0x70,0x0,0x177,0x0,0x14f,0x0,0x18,0x0,0x151,0x0,0x64,0x0,0x13,0x0,0x178,0x0,0x109,0x0,0x65,0x0,0x10a,0x0,0x24,0x0,0x13,0x0,0x10b,0x0,0x9f,0x0,0x58,0x0,0x11b,0x0,0xb4,0x0,0x179,0x0,0x84,0x0,0x17a,0x0,0x5b,0x0,0x17b,0x0,0x1e,0x0,0x176,0x0,0xc,0x0,0x176,0x0,0x64,0x0,0x17c,0x0,0x17d,0x0,0x17e,0x0,0x65,0x0,0x17f,0x0,0x3d,0x0,0x16,0x0,0x3e,0x0,0x180,0x50004,0x181,0x50004,0x182,0x50004,0x183,0x50004,0x184,0x50004,0x185,0x0,0x15,0x0,0x1a,0x0,0x59,0x0,0x186,0x0,0x187,0x0,0x0,0x0,0x188,0x0,0x18,0x0,0x189,0x0,0x18a,0x0,0x0,0x0,0x18b,0x0,0x18,0x0,0x189,0x0,0x18a,0x0,0x0,0x0,0x18b,0x0,0x18,0x0,0x18c,0x0,0x18a,0x0,0x0,0x0,0x188,0x0,0x18,0x0,0x18c,0x0,0x18a,0x0,0x0,0x0,0x188,0x0,0x18,0x0,0x18d,0x0,0x18a,0x0,0x14f,0x0,0x18,0x0,0x166,0x0,0x18e,0x0,0x64,0x0,0x3e,0x0,0x18f,0x0,0x7a,0x0,0xc,0x0,0x108,0x0,0x13,0x0,0x14a,0x0,0xe1,0x0,0x16,0x0,0x190,0x0,0x191,0x0,0x65,0x0,0x167,0x0,0x10b,0x0,0x6e,0x0,0x65,0x0,0x36,0x0,0x24,0x0,0x13,0x0,0x18,0x0,0x18a,0x0,0x192,0x0,0xb4,0x0,0x193,0x0,0x57,0x0,0x65,0x0,0x194,0x0,0x195,0x0,0x15c,0x0,0x149,0x0,0x64,0x0,0x16b,0x0,0x196,0x0,0x65,0x0,0x194,0x0,0x7a,0x0,0x1a,0x0,0x59,0x0,0x197,0x0,0x91,0x0,0x65,0x0,0x149,0x0,0x24,0x0,0x13,0x0,0x14a,0x0,0x164,};
  canvas->set_elts(seq,127);
  canvas->set_bgrects(NULL, 0);
  canvas->set_underlines(NULL, 0);
static const unsigned int verboxes[]= {
0x29,0x2d,0x5,};
  canvas->set_vboxes(verboxes, 1);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_2);
  tf_6(canvas);
}
static void tf_7(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,1);
  canvas->pbreak();
  canvas->println(2,31);
  canvas->pbreak();
  canvas->println(32,41);
  canvas->bullet_println(1, 42, 43);
  canvas->bullet_println(1, 44, 45);
  canvas->bullet_println(1, 46, 47);
  canvas->println(48,55);
  canvas->pbreak();
  canvas->println(56,100);
  canvas->pbreak();
  canvas->println(101,150);
  canvas->verbatim_vspace();
  canvas->verbatim_line(151,152);
  canvas->verbatim_line(153,153);
  canvas->verbatim_line(154,154);
  canvas->verbatim_line(155,156);
  canvas->verbatim_vspace();
  canvas->println(157,160);
  canvas->verbatim_vspace();
  canvas->verbatim_line(161,163);
  canvas->verbatim_line(164,164);
  canvas->verbatim_vspace();

 canvas->end_doc(tf_7);
}
static void cf_7(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x198,0xa,0x199,0x0,0xce,0x0,0x36,0x0,0x16,0x0,0x13,0x0,0x105,0x0,0x24,0x0,0x36,0x0,0x37,0x0,0x16,0x0,0x19a,0x0,0xe6,0x0,0x16d,0x0,0x88,0x0,0x19b,0x0,0x65,0x0,0x130,0x0,0x111,0x0,0x84,0x0,0x1b,0x0,0x108,0x0,0x16e,0x0,0x19c,0x0,0x19d,0x0,0xa3,0x0,0x19e,0x0,0x6a,0x0,0xc6,0x0,0xc,0x0,0xd0,0x0,0x19f,0x0,0x14f,0x0,0x45,0x0,0x36,0x0,0x105,0x0,0x151,0x0,0x64,0x0,0x1a0,0x0,0x24,0x0,0x1a1,0x0,0xf6,0x0,0x0,0x0,0x1a2,0x0,0x0,0x0,0x1a3,0x0,0x0,0x0,0x1a4,0x0,0x62,0x0,0x1a5,0x0,0x1a6,0x0,0x65,0x40000,0x1a7,0x0,0xe1,0x0,0x16,0x0,0x1a8,0x0,0xd3,0x0,0x1a9,0x0,0x1a1,0x0,0xc7,0x0,0x195,0x0,0x70,0x0,0x1aa,0x0,0x45,0x0,0x7e,0x0,0xf7,0x0,0xc2,0x0,0x16,0x0,0x83,0x0,0x1ab,0x0,0x5f,0x0,0x1ac,0x0,0xa3,0x0,0xf7,0x0,0xc2,0x0,0x84,0x0,0x100,0x0,0x13,0x0,0x1ad,0x0,0x16,0x0,0x65,0x0,0x1ae,0x0,0x91,0x0,0xa3,0x0,0x1af,0x0,0xf7,0x0,0xc2,0x0,0x64,0x0,0x1a0,0x0,0x1b0,0x0,0x1b1,0x0,0xf7,0x0,0xc2,0x0,0xf3,0x0,0x15c,0x0,0x16f,0x0,0x191,0x0,0x65,0x0,0x7a,0x0,0xb7,0x0,0x1b2,0x0,0x9f,0x0,0x58,0x0,0x1b3,0x0,0x1b4,0x0,0x65,0x0,0xe1,0x0,0x111,0x0,0xa3,0x0,0xf7,0x0,0x1b5,0x0,0x1b6,0x0,0x1b7,0x0,0x88,0x0,0x1b8,0x0,0x111,0x0,0x70,0x0,0x136,0x0,0x5b,0x0,0x17b,0x0,0x13,0x0,0xc2,0x0,0xef,0x0,0x1b9,0x0,0xb2,0x0,0x58,0x0,0x1b4,0x0,0x9a,0x0,0x1ba,0x0,0x1d,0x0,0x1bb,0x0,0x1b,0x0,0x91,0x0,0x3e,0x0,0x1bc,0x0,0x5b,0x0,0x65,0x0,0x45,0x0,0xef,0x0,0x1bd,0x0,0x1be,0x0,0x65,0x0,0xc1,0x0,0x1bf,0x0,0x65,0x0,0xe1,0x0,0x111,0x0,0x5b,0x0,0x1c0,0x0,0x88,0x0,0x1c1,0x4,0x1c2,0x50004,0xe4,0x50004,0x1c3,0x50004,0x1c4,0x4,0x1c5,0x50004,0xe4,0x0,0x15,0x0,0x16,0x0,0x1aa,0x0,0x187,0x50004,0x1c6,0x30004,0x1c7,0x50004,0x1c8,0x50004,0x1c4,};
  canvas->set_elts(seq,165);
static const unsigned int bgrects[]= {
0x34,0x34,0x4,0xa2,0xa2,0x3,};
  canvas->set_bgrects(bgrects, 2);
  canvas->set_underlines(NULL, 0);
static const unsigned int verboxes[]= {
0x97,0x9c,0x5,0xa1,0xa4,0x5,};
  canvas->set_vboxes(verboxes, 2);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_2);
  tf_7(canvas);
}
static void tf_8(FYGfxCanvas* canvas) {
  int *TL_ROW= canvas->get_row_lines(4);
  int *TL_COL= canvas->get_col_lines(6);
canvas->reset_doc();
  canvas->println(0,0);
  canvas->pbreak();
  canvas->println(1,47);
  canvas->pbreak();
  canvas->println(48,74);
  canvas->pbreak();
  canvas->println(75,120);
  canvas->pbreak();
  canvas->println(121,126);
  canvas->bullet_println(1, 127, 131);
  canvas->bullet_println(1, 132, 136);
  canvas->bullet_println(1, 137, 141);
  canvas->bullet_println(1, 142, 148);
  canvas->println(149,153);
  canvas->verbatim_vspace();
  canvas->verbatim_line(154,155);
  canvas->verbatim_line(156,164);
  canvas->verbatim_line(165,171);
  canvas->verbatim_line(172,180);
  canvas->verbatim_line(181,182);
  canvas->verbatim_vspace();
  canvas->println(183,187);
 /// BEGIN TABLE 
 {
  int cw0, ch0,ca0;
  canvas->get_seq_metrics(188,188, &cw0, &ch0, &ca0);
  int cw1, ch1,ca1;
  canvas->get_seq_metrics(189,189, &cw1, &ch1, &ca1);
  int cw2, ch2,ca2;
  canvas->get_seq_metrics(190,190, &cw2, &ch2, &ca2);
  int cw3, ch3,ca3;
  canvas->get_seq_metrics(191,191, &cw3, &ch3, &ca3);
  int cw4, ch4,ca4;
  canvas->get_seq_metrics(192,193, &cw4, &ch4, &ca4);
  int cw5, ch5,ca5;
  canvas->get_seq_metrics(194,194, &cw5, &ch5, &ca5);
  int cw6, ch6,ca6;
  canvas->get_seq_metrics(195,196, &cw6, &ch6, &ca6);
  int cw7, ch7,ca7;
  canvas->get_seq_metrics(197,198, &cw7, &ch7, &ca7);
  int cw8, ch8,ca8;
  canvas->get_seq_metrics(199,199, &cw8, &ch8, &ca8);
  int cw9, ch9,ca9;
  canvas->get_seq_metrics(200,201, &cw9, &ch9, &ca9);
  int cw10, ch10,ca10;
  canvas->get_seq_metrics(202,203, &cw10, &ch10, &ca10);
 int rh0=0, ra0=0;
 int rh1=0, ra1=0;
 int rh2=0, ra2=0;
  rh0= max_i(rh0, ch0); ra0= max_i(ra0, ca0);
  rh0= max_i(rh0, ch1); ra0= max_i(ra0, ca1);
  rh0= max_i(rh0, ch2); ra0= max_i(ra0, ca2);
  rh0= max_i(rh0, ch3); ra0= max_i(ra0, ca3);
  rh1= max_i(rh1, ch4); ra1= max_i(ra1, ca4);
  rh1= max_i(rh1, ch5); ra1= max_i(ra1, ca5);
  rh1= max_i(rh1, ch6); ra1= max_i(ra1, ca6);
  rh2= max_i(rh2, ch7); ra2= max_i(ra2, ca7);
  rh2= max_i(rh2, ch8); ra2= max_i(ra2, ca8);
  rh2= max_i(rh2, ch9); ra2= max_i(ra2, ca9);
  rh2= max_i(rh2, ch10); ra2= max_i(ra2, ca10);
  int cow0=0;
  int cow1=0;
  int cow2=0;
  int cow3=0;
  cow0= max_i(cow0, cw0);
  cow0= max_i(cow0, cw4);
  cow0= max_i(cow0, cw7);
  cow1= max_i(cow1, cw1);
  cow1= max_i(cow1, cw5);
  cow1= max_i(cow1, cw8);
  cow2= max_i(cow2, cw2);
  cow2= max_i(cow2, cw9);
  cow3= max_i(cow3, cw3);
  cow3= max_i(cow3, cw10);
  int D=0;
  D= cw6  - cow2 - cow3;
  if (D>0) { 
    cow2 += D / 2; D -= D / 2;
    cow3 += D; }
  int rp0;
  int rp1;
  int rp2;
  int rp3;
  int cp0;
  int cp1;
  int cp2;
  int cp3;
  int cp4;
  cp0= rp0= 0;
  rp1= rp0 + 2*CELL_PAD + rh0;
  rp2= rp1 + 2*CELL_PAD + rh1;
  rp3= rp2 + 2*CELL_PAD + rh2;
  cp1= cp0 + 2*CELL_PAD + cow0;
  cp2= cp1 + 2*CELL_PAD + cow1;
  cp3= cp2 + 2*CELL_PAD + cow2;
  cp4= cp3 + 2*CELL_PAD + cow3;
  int off_x= canvas->get_width() - cp4;
  if (off_x<0) off_x= 0; else off_x/= 2;
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 188, 188);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 189, 189);
  canvas->cell_center(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 190, 190);
  canvas->cell_right(off_x+cp3+CELL_PAD, off_x+cp4-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 191, 191);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 192, 193);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 194, 194);
  canvas->cell_center(off_x+cp2+CELL_PAD, off_x+cp4-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 195, 196);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 197, 198);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 199, 199);
  canvas->cell_left(off_x+cp2+CELL_PAD, off_x+cp3-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 200, 201);
  canvas->cell_left(off_x+cp3+CELL_PAD, off_x+cp4-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 202, 203);
  int off_y= canvas->get_cursor_y();
  TL_ROW[0]= off_y+rp0; TL_ROW[1]= off_x+cp0; TL_ROW[2]= off_x+cp4;
  TL_ROW[3]= off_y+rp1; TL_ROW[4]= off_x+cp0; TL_ROW[5]= off_x+cp4;
  TL_ROW[6]= off_y+rp2; TL_ROW[7]= off_x+cp0; TL_ROW[8]= off_x+cp4;
  TL_ROW[9]= off_y+rp3; TL_ROW[10]= off_x+cp0; TL_ROW[11]= off_x+cp4;
  TL_COL[0]= off_y+rp0; TL_COL[1]= off_y+rp1; TL_COL[2]= off_x+cp3;
  TL_COL[3]= off_y+rp0; TL_COL[4]= off_y+rp3; TL_COL[5]= off_x+cp0;
  TL_COL[6]= off_y+rp0; TL_COL[7]= off_y+rp3; TL_COL[8]= off_x+cp1;
  TL_COL[9]= off_y+rp0; TL_COL[10]= off_y+rp3; TL_COL[11]= off_x+cp2;
  TL_COL[12]= off_y+rp0; TL_COL[13]= off_y+rp3; TL_COL[14]= off_x+cp4;
  TL_COL[15]= off_y+rp2; TL_COL[16]= off_y+rp3; TL_COL[17]= off_x+cp3;
  canvas->xcoord(cp4+3);
  canvas->advance_y(rp3+2);
 /// END TABLE 
 }

 canvas->end_doc(tf_8);
}
static void cf_8(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x1c9,0x0,0x15,0x0,0x16,0x0,0x1ca,0x0,0x65,0x0,0x1cb,0x0,0x1cc,0x0,0xff,0x0,0x24,0x0,0x9a,0x0,0xbb,0x0,0x10e,0x0,0x9a,0x0,0x1cd,0x0,0x1ce,0x0,0x70,0x0,0x1cf,0x0,0x9c,0x0,0x1d0,0x0,0x88,0x0,0x1d1,0x0,0x9f,0x0,0x149,0x0,0x13,0x0,0xf2,0x0,0x64,0x0,0x65,0x40000,0x1d2,0x0,0xe1,0x0,0x88,0x0,0x57,0x0,0x5d,0x0,0x65,0x0,0x1d3,0x0,0x5f,0x0,0xf2,0x0,0x166,0x0,0x108,0x0,0x3e,0x0,0xe1,0x0,0x16,0x0,0x1d4,0x0,0x109,0x0,0x65,0x0,0x149,0x0,0x24,0x0,0x13,0x0,0x10b,0x0,0xbe,0x0,0x1d5,0x0,0x24,0x0,0x1d6,0x0,0x15c,0x0,0x59,0x0,0x1d7,0x0,0x64,0x0,0x13,0x0,0x1d8,0x0,0x1d9,0x0,0x1da,0x0,0x1db,0x0,0xa3,0x0,0x65,0x0,0x1dc,0x0,0x1dd,0x0,0x9f,0x0,0x58,0x0,0x1de,0x0,0x24,0x0,0x9a,0x0,0xf1,0x0,0x91,0x0,0x13,0x0,0x7a,0x0,0x1df,0x0,0xbe,0x0,0x1e0,0x0,0x16,0x0,0x1e1,0x0,0x1a9,0x0,0x1e2,0x0,0x1e3,0x0,0x88,0x0,0x58,0x0,0x1e4,0x0,0x1e5,0x0,0xc2,0x0,0x96,0x0,0x88,0x0,0x1d3,0x0,0x56,0x0,0xc7,0x0,0x70,0x0,0x1e6,0x0,0x1d,0x0,0x13,0x0,0x1e7,0x0,0x1e8,0x0,0x1e9,0x0,0x111,0x0,0x1ea,0x0,0x65,0x0,0x189,0x0,0x1e7,0x0,0x1eb,0x0,0x1ec,0x0,0x1ed,0x0,0x84,0x0,0x65,0x0,0x1e0,0x0,0x70,0x0,0x142,0x0,0x5b,0x0,0x59,0x0,0xc2,0x0,0x1ee,0x0,0x5f,0x0,0x1ef,0x0,0x16,0x0,0x1e0,0x0,0x1d3,0x0,0x5f,0x0,0xc1,0x0,0xc2,0x0,0x111,0x0,0x70,0x0,0x1f0,0x0,0x0,0x0,0x1f1,0x0,0xa3,0x0,0x1f2,0x0,0x1f3,0x0,0x0,0x0,0x1f4,0x0,0xa3,0x0,0x1f5,0x0,0x1f3,0x0,0x0,0x0,0x1f6,0x0,0xa3,0x0,0x1f7,0x0,0x1f3,0x0,0x0,0x0,0x13,0x0,0x7b,0x0,0xa3,0x0,0x1f8,0x0,0x1f9,0x0,0x1fa,0x0,0x3d,0x0,0x16,0x0,0x3e,0x0,0x7,0x0,0x1fb,0x4,0x1fc,0x50004,0xe4,0x50004,0x1fd,0x4,0x1fe,0x50004,0x1ff,0x4,0x200,0x50004,0x1ff,0x4,0x201,0x50004,0x1ff,0x4,0x202,0x50004,0x203,0x50004,0x1fd,0x4,0x204,0x50004,0x205,0x4,0x206,0x50004,0x205,0x4,0x207,0x50004,0x203,0x50004,0x1fd,0x4,0x208,0x50004,0x1ff,0x4,0x209,0x50004,0x1ff,0x4,0x20a,0x50004,0x20b,0x4,0x20c,0x50004,0x20d,0x4,0x1c5,0x50004,0xe4,0x0,0x3d,0x0,0x16,0x0,0x42,0x0,0x20e,0x0,0x20f,0x0,0x210,0x0,0x124,0x0,0x211,0x0,0x128,0x0,0x212,0x0,0x1fa,0x0,0x213,0x0,0x214,0x0,0x215,0x0,0xd2,0x0,0xe0,0x0,0x216,0x30000,0x217,0x30000,0x18a,0x1000000,0x218,0x1000000,0x18a,};
  canvas->set_elts(seq,204);
static const unsigned int bgrects[]= {
0x1b,0x1b,0x4,0xc8,0xc9,0x3,};
  canvas->set_bgrects(bgrects, 2);
static const unsigned int underlines[]= {
0xca,0xcb,0x0,};
  canvas->set_underlines(underlines, 1);
static const unsigned int verboxes[]= {
0x9a,0xb6,0x5,};
  canvas->set_vboxes(verboxes, 1);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_2);
  tf_8(canvas);
}
static void tf_9(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,0);
  canvas->pbreak();
  canvas->println(1,57);
  canvas->pbreak();
  canvas->println(58,84);
  canvas->verbatim_vspace();
  canvas->verbatim_line(85,86);
  canvas->verbatim_vspace();
  canvas->println(87,89);
  canvas->image(0);
  canvas->println(90,107);

 canvas->end_doc(tf_9);
}
static void cf_9(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x219,0x0,0xcd,0x0,0x84,0x0,0x21a,0x0,0xe0,0x0,0x58,0x0,0x59,0x0,0x136,0x0,0x64,0x0,0x9a,0x0,0x21b,0x0,0x62,0x0,0x21c,0x0,0x65,0x0,0x7c,0x0,0x21d,0x0,0xaa,0x0,0x138,0x0,0x65,0x0,0x69,0x0,0x51,0x0,0xaa,0x0,0x88,0x0,0x21e,0x0,0x11a,0x0,0x1d6,0x0,0x109,0x0,0x21f,0x0,0x220,0x0,0x221,0x0,0x65,0x0,0x69,0x0,0x51,0x0,0x6a,0x0,0x58,0x0,0x222,0x0,0x223,0x0,0x224,0x0,0x5f,0x0,0x4e,0x0,0x225,0x0,0x226,0x0,0x227,0x0,0x7e,0x0,0xfa,0x0,0x50,0x0,0x228,0x0,0x16,0x0,0x91,0x0,0x229,0x0,0x91,0x0,0x22a,0x0,0xc,0x0,0x22b,0x0,0x22c,0x0,0x65,0x0,0x22d,0x0,0x22e,0x0,0x5f,0x0,0xe1,0x0,0x5b,0x0,0x59,0x0,0x136,0x0,0x16,0x40000,0x22f,0x40000,0x230,0x0,0x28,0x0,0x5f,0x0,0x231,0x0,0x16,0x0,0x197,0x0,0x232,0x0,0x5b,0x0,0x65,0x0,0x233,0x0,0x24,0x0,0x65,0x0,0x66,0x0,0x22e,0x0,0x3d,0x0,0x16,0x0,0x3e,0x0,0x7,0x0,0x21d,0x0,0x234,0x4,0x235,0x50004,0xe4,0x0,0x88,0x0,0x85,0x0,0x236,0x0,0x237,0x0,0x11a,0x0,0x227,0x0,0x70,0x0,0x238,0x0,0x105,0x0,0x239,0x0,0xb2,0x0,0x89,0x0,0x11b,0x0,0x23a,0x0,0x91,0x0,0xff,0x0,0x24,0x0,0x23b,0x0,0xc,0x0,0xf2,0x0,0x23c,};
  canvas->set_elts(seq,108);
static const unsigned int bgrects[]= {
0x40,0x41,0x4,};
  canvas->set_bgrects(bgrects, 1);
  canvas->set_underlines(NULL, 0);
static const unsigned int verboxes[]= {
0x55,0x56,0x5,};
  canvas->set_vboxes(verboxes, 1);
static const unsigned char img0data[]={

137,80,78,71,13,10,26,10,0,0,0,13,
73,72,68,82,0,0,1,138,0,0,1,5,
8,6,0,0,0,253,152,160,29,0,0,0,
6,98,75,71,68,0,255,0,255,0,255,160,
189,167,147,0,0,0,9,112,72,89,115,0,
0,3,211,0,0,3,211,1,231,116,74,61,
0,0,0,7,116,73,77,69,7,222,7,6,
17,40,4,52,127,247,199,0,0,32,0,73,
68,65,84,120,218,236,189,215,119,92,215,157,
239,249,57,251,132,58,149,51,170,10,25,32,
8,230,40,82,209,146,108,75,178,220,14,237,
52,178,187,237,118,247,117,247,117,207,234,213,
179,238,125,156,63,97,222,102,205,154,181,122,
110,247,189,221,237,110,231,32,203,114,148,100,
73,150,68,5,42,48,139,9,36,65,228,88,
40,84,142,39,205,67,81,0,33,130,36,0,
18,20,105,243,247,34,17,117,194,62,123,255,
246,47,124,127,97,75,169,84,202,225,46,221,
165,187,116,151,238,210,93,186,10,73,62,159,
207,41,22,139,43,190,225,216,201,51,92,28,
157,96,239,246,126,14,28,60,66,173,253,193,
107,94,95,153,159,194,19,73,45,251,219,116,
190,206,104,166,178,56,152,114,26,199,27,95,
114,141,34,75,236,237,12,129,180,244,222,242,
196,89,218,212,60,251,247,238,230,153,151,222,
102,119,127,39,163,147,51,60,176,103,43,155,
251,251,248,191,254,191,239,17,217,241,56,245,
233,115,184,234,115,124,245,115,143,19,14,5,
23,238,207,102,179,124,239,213,115,232,45,61,
43,30,239,106,190,237,195,84,203,167,209,131,
241,155,254,220,117,187,54,51,73,113,224,53,
58,55,110,163,98,169,124,230,222,62,186,58,
218,150,92,211,104,52,248,151,95,29,196,114,
133,174,249,92,203,113,56,55,93,162,88,174,
65,163,2,122,224,250,3,168,206,131,59,178,
114,78,190,206,245,18,176,183,39,132,44,73,
43,158,135,213,172,217,205,90,223,82,205,100,
96,186,132,101,59,120,138,105,132,227,80,10,
180,220,148,57,88,92,184,114,243,191,154,247,
230,61,179,86,0,205,3,66,185,254,181,102,
29,108,115,69,239,151,202,105,240,198,217,218,
22,192,235,146,111,26,127,207,29,254,53,177,
123,62,191,240,111,247,220,41,158,122,242,1,
106,181,26,161,80,104,217,123,190,255,204,115,
228,77,133,71,182,119,178,99,235,166,149,173,
117,165,194,63,127,239,105,122,186,58,217,209,
215,193,239,223,57,133,210,125,63,182,101,210,
40,101,175,224,25,163,48,203,99,27,116,54,
247,247,221,62,74,66,146,16,171,189,105,247,
246,45,244,117,181,241,219,35,19,215,85,18,
215,35,77,145,174,207,83,150,67,169,110,94,
57,120,179,204,151,62,247,105,84,69,224,118,
170,196,66,126,226,177,8,179,153,28,150,101,
17,82,13,26,115,195,232,212,249,212,199,246,
225,214,93,24,134,177,112,255,155,135,79,225,
138,117,222,53,21,174,206,29,36,30,252,75,
74,158,46,246,245,132,72,182,196,174,184,164,
88,44,130,230,191,238,163,100,73,162,63,233,
195,175,43,31,217,231,56,64,205,176,110,251,
105,247,233,10,155,82,126,100,33,225,171,149,
8,84,242,127,242,172,24,245,105,215,85,18,
171,158,231,13,247,45,229,101,79,27,255,247,
63,255,39,231,135,198,174,122,207,150,158,20,
33,185,206,75,239,158,97,108,124,242,234,188,
230,56,76,76,76,144,201,100,240,120,60,236,
216,212,203,197,121,147,231,6,45,228,174,123,
175,179,89,52,26,151,201,169,219,133,214,180,
115,103,231,114,136,96,219,13,191,92,147,87,
166,167,106,166,141,239,67,127,179,213,0,135,
143,28,163,175,111,3,223,252,218,151,40,149,
43,204,204,231,41,213,77,100,89,230,31,190,
253,183,252,207,167,95,134,214,221,60,243,230,
49,92,158,9,220,245,89,62,255,216,199,248,
221,129,67,212,220,73,132,91,230,46,93,135,
65,220,62,222,185,56,65,177,124,148,79,62,
124,255,146,223,198,38,166,144,116,63,212,43,
148,71,79,160,185,220,168,137,141,87,85,22,
27,19,62,206,55,202,20,63,162,111,169,27,
54,94,237,246,159,115,159,75,102,115,202,207,
68,193,75,221,180,254,164,249,79,146,160,61,
226,89,243,253,182,101,98,20,102,113,133,91,
175,124,240,229,252,233,14,208,219,213,206,224,
197,139,140,79,205,226,243,121,217,210,219,65,
123,251,162,156,187,103,207,110,182,108,234,231,
165,55,15,243,238,169,65,58,218,91,151,247,
86,50,25,190,243,253,31,83,51,29,186,54,
110,199,244,165,240,119,246,129,180,2,121,87,
78,211,213,190,227,182,91,7,177,150,155,82,
137,24,141,204,200,13,191,92,85,86,246,122,
219,190,50,140,34,187,253,252,230,165,3,56,
142,141,223,239,39,149,76,176,161,171,149,169,
177,225,133,107,246,111,74,97,103,46,34,7,
18,104,45,125,152,109,247,241,163,215,206,210,
72,238,69,4,91,239,106,129,149,108,84,33,
163,165,182,146,41,213,104,52,26,75,126,123,
235,240,251,168,30,63,78,97,154,71,55,199,
104,204,158,167,58,114,248,234,204,38,36,122,
98,94,2,238,143,198,179,176,236,59,39,28,
231,117,201,244,7,5,94,179,246,39,205,127,
49,191,107,69,200,195,178,8,219,244,121,250,
165,81,30,237,118,81,31,126,23,179,156,189,
134,66,18,204,7,183,145,214,186,152,214,251,
120,237,141,131,60,255,226,203,148,203,229,165,
198,173,166,81,46,149,72,151,76,42,149,202,
178,207,10,5,131,36,19,9,60,186,134,207,
163,97,90,214,245,149,132,227,96,25,117,146,
110,131,96,48,112,219,173,195,154,20,69,46,
95,64,120,194,84,6,223,166,81,94,187,107,
44,173,112,253,203,245,43,173,42,179,48,75,
176,173,143,51,103,207,45,252,109,67,119,23,
159,121,226,19,11,255,222,179,99,43,155,162,
50,86,118,28,219,52,144,132,140,59,213,127,
87,250,175,129,210,196,248,215,31,255,154,83,
103,207,47,252,109,215,182,77,216,141,26,45,
110,147,189,123,118,243,201,135,246,241,208,214,
118,204,82,230,154,202,162,63,233,39,248,17,
40,11,231,14,75,219,80,60,30,146,169,208,
138,247,201,31,27,185,53,65,220,183,118,23,
208,22,50,178,144,104,77,181,96,229,167,216,
19,174,96,228,166,174,62,223,46,15,178,230,
34,213,24,228,191,255,253,55,249,175,223,250,
107,188,222,102,28,101,98,114,138,11,67,195,
252,207,31,253,146,137,162,77,41,51,201,203,
175,189,9,112,133,194,56,246,254,25,106,22,
212,44,153,122,189,1,102,227,234,99,44,207,
225,205,158,161,219,26,98,183,119,142,199,31,
188,231,246,228,197,181,220,116,102,34,135,28,
218,128,212,210,135,230,13,174,249,229,134,101,
175,76,72,21,235,184,53,153,100,208,181,232,
81,20,199,113,180,32,187,118,108,107,194,32,
227,19,188,253,222,17,158,124,236,81,28,199,
65,186,180,187,62,241,208,126,60,218,33,14,
101,242,8,127,236,174,196,95,43,163,4,90,
32,208,194,129,161,25,206,14,189,66,123,75,
152,98,169,140,89,85,232,109,79,54,21,199,
142,237,56,142,195,169,103,158,167,238,13,95,
213,138,18,18,108,76,250,57,55,93,162,80,
189,53,120,172,95,87,136,249,93,119,214,156,
155,6,1,201,166,39,230,225,98,186,242,167,
227,197,210,140,75,180,69,220,88,133,181,127,
183,222,210,203,241,161,119,56,122,242,12,255,
248,237,255,130,162,40,92,252,197,115,84,166,
231,192,157,196,106,212,96,234,24,118,100,3,
170,191,25,84,118,108,139,222,174,54,226,177,
165,178,226,103,207,254,134,68,71,47,197,170,
129,215,45,99,104,30,6,198,231,56,255,79,
255,11,191,223,199,183,255,250,47,23,174,77,
196,35,60,245,249,39,57,248,238,97,210,117,
5,87,188,123,121,15,215,168,67,110,140,111,
124,253,11,183,253,154,172,218,163,248,238,51,
47,80,146,2,212,242,115,104,55,40,120,77,
115,229,38,222,88,166,194,92,177,190,8,91,
5,147,4,189,26,63,248,213,75,140,142,141,
113,236,204,32,147,243,101,126,246,155,151,56,
121,250,44,0,133,66,145,127,127,250,57,222,
58,54,128,191,54,137,81,41,220,149,248,55,
72,114,32,65,38,176,141,35,165,24,35,202,
6,116,171,68,71,123,234,50,47,81,226,75,
79,60,68,109,118,232,218,140,39,65,127,210,
71,208,163,174,105,28,170,44,104,13,233,244,
39,125,244,37,188,116,134,221,180,133,117,2,
186,130,144,36,132,4,178,16,248,116,133,238,
152,151,254,148,15,113,135,89,230,150,44,99,
203,77,5,23,245,173,125,158,122,226,94,34,
107,188,255,86,83,64,87,216,214,22,160,183,
197,139,107,5,208,180,99,153,56,51,167,17,
227,239,226,44,227,50,42,173,59,176,125,173,
76,78,77,33,73,18,223,248,242,103,248,204,
3,219,176,243,19,24,185,73,190,252,248,3,
236,111,177,48,242,211,205,235,253,49,14,31,
61,177,228,89,182,109,19,12,71,153,156,201,
240,133,143,223,195,63,254,213,23,248,199,175,
127,150,253,219,250,168,215,106,124,243,171,95,
90,242,206,206,142,118,18,45,113,190,248,185,
79,243,216,222,62,26,249,244,82,47,194,108,
32,165,207,178,51,92,231,91,95,249,244,157,
97,180,172,246,134,10,46,84,127,20,33,223,
56,116,208,176,86,174,40,28,96,40,93,193,
235,82,112,107,50,166,191,157,252,192,171,120,
83,155,112,185,116,114,229,58,90,247,125,20,
39,79,241,238,161,35,244,247,245,242,194,27,
135,168,135,251,217,215,54,143,170,200,28,202,
222,45,25,89,30,146,89,244,192,46,255,255,
229,126,95,80,24,74,19,18,104,113,219,180,
183,46,141,247,140,140,79,33,175,32,165,83,
72,176,49,225,227,220,84,137,66,109,101,158,
133,215,165,144,10,185,8,123,212,37,99,170,
153,42,122,208,13,225,63,34,203,218,113,144,
236,166,215,221,26,114,147,41,173,206,251,242,
235,10,253,73,31,178,144,144,112,152,47,25,
183,245,247,118,68,220,164,66,250,202,161,165,
249,97,122,252,6,15,125,106,31,178,44,243,
175,207,188,12,209,165,188,232,204,95,100,67,
210,135,121,89,124,202,52,13,108,217,133,176,
12,188,30,55,247,238,221,201,228,11,7,152,
46,10,132,191,133,188,43,201,191,252,251,247,
248,214,95,125,141,119,14,29,197,227,214,73,
207,231,137,38,59,216,220,223,76,214,208,52,
141,71,31,220,79,34,26,68,215,175,28,115,
165,82,225,233,103,127,67,89,248,80,82,59,
151,252,230,201,156,226,175,190,244,4,66,136,
59,134,23,87,53,210,114,185,140,33,221,60,
247,189,97,218,171,19,104,192,76,161,190,96,
185,134,182,127,10,57,152,228,39,7,206,82,
14,111,65,205,14,240,87,159,121,144,191,251,
155,111,224,114,185,48,81,144,117,63,135,47,
206,241,238,169,65,34,213,161,5,203,225,46,
177,196,11,88,238,255,175,245,55,0,163,156,
163,191,51,177,160,76,10,197,34,111,188,115,
152,183,207,140,161,186,253,43,99,64,9,54,
38,189,120,181,107,27,30,46,69,208,215,226,
99,107,155,159,136,87,187,234,152,254,168,214,
197,113,46,113,253,202,19,63,62,160,176,87,
99,243,165,52,91,22,158,114,251,82,212,167,
174,74,73,88,70,29,59,59,198,224,116,158,
239,254,242,101,94,124,235,40,166,178,76,118,
84,160,131,90,185,196,129,247,222,231,153,223,
191,206,185,11,23,73,37,19,120,236,10,90,
75,31,47,190,241,30,0,79,62,178,31,10,
19,77,37,16,104,225,190,123,247,161,105,26,
245,122,157,19,167,78,227,78,245,19,244,106,
140,141,79,80,40,20,56,114,252,125,0,54,
111,90,62,230,169,235,58,79,125,233,243,216,
146,130,80,22,227,44,182,101,176,161,61,126,
71,41,137,85,123,20,233,185,12,142,30,188,
105,47,175,26,230,170,239,201,148,26,116,70,
220,205,193,123,155,133,49,117,217,133,60,254,
14,127,245,149,207,46,104,247,90,173,70,161,
222,196,59,29,163,202,183,191,246,57,76,211,
228,187,47,30,135,96,242,174,118,184,25,84,
152,192,235,237,231,95,126,248,44,134,26,0,
73,65,105,233,67,36,59,112,108,139,226,240,
81,36,205,139,191,125,203,181,225,44,33,209,
159,242,113,102,178,120,69,173,131,34,36,218,
194,58,45,1,215,159,132,114,88,98,49,95,
230,181,219,171,136,196,199,252,46,122,99,158,
37,69,170,215,154,59,33,129,199,165,160,8,
137,92,229,214,123,29,110,77,208,29,247,174,
248,122,171,81,69,168,46,164,13,15,99,214,
171,216,133,41,198,50,101,44,225,190,18,122,
243,6,201,123,119,147,59,253,26,70,235,189,
188,244,254,121,254,50,26,161,37,232,102,78,
150,25,154,41,242,210,43,175,243,201,71,31,
66,150,101,28,64,241,134,121,111,120,152,243,
99,175,50,52,54,133,105,24,132,91,83,76,
78,156,224,135,207,156,33,18,111,69,182,42,
236,217,185,253,170,243,154,207,231,249,231,239,
252,128,224,238,207,45,249,187,89,201,227,139,
232,119,28,47,174,74,81,204,231,139,88,213,
18,178,170,95,194,218,12,132,178,118,236,179,
214,176,87,125,143,101,59,204,151,27,31,170,
171,112,232,73,197,248,245,31,14,50,58,120,
142,167,190,240,25,10,165,10,118,176,19,25,
176,3,237,28,63,117,150,251,247,237,225,222,
222,48,199,7,143,99,39,119,221,21,244,55,
106,241,122,98,252,250,232,12,134,191,231,138,
138,88,73,200,4,122,247,173,248,89,170,44,
177,173,205,207,84,174,78,182,92,71,88,50,
193,176,78,42,168,47,88,197,127,106,36,27,
13,100,203,164,190,10,69,225,117,41,244,196,
220,87,116,50,208,150,241,72,92,138,32,25,
116,17,11,184,144,37,9,199,129,67,67,217,
91,235,125,72,208,19,247,34,175,194,8,80,
210,167,177,20,31,178,42,179,43,161,211,218,
147,228,173,227,3,52,164,171,11,224,224,230,
143,97,141,188,135,163,7,121,245,157,99,56,
118,211,72,213,66,41,58,90,3,156,56,61,
0,234,101,138,38,210,77,14,8,7,182,82,
43,204,161,184,60,88,209,13,120,19,91,145,
114,231,248,155,175,254,111,72,146,196,59,239,
190,199,125,247,238,95,242,174,159,62,251,59,
66,193,0,193,150,118,20,109,169,242,210,252,
49,222,26,188,136,205,25,146,209,16,169,84,
234,142,224,197,85,249,63,199,79,157,163,67,
204,34,231,154,129,202,249,83,175,172,249,197,
14,80,93,99,165,236,108,97,105,186,153,203,
23,230,216,180,193,233,83,167,248,218,151,62,
71,103,103,7,83,153,2,178,171,185,72,114,
37,205,206,173,155,200,229,114,180,182,68,169,
54,238,198,42,110,138,149,17,76,160,70,218,
47,91,84,135,185,145,55,200,142,31,90,155,
96,20,18,237,17,157,29,29,65,54,180,120,
105,15,187,255,100,149,4,128,169,185,104,184,
46,25,101,214,74,54,179,196,134,22,207,178,
86,238,229,129,97,183,38,211,21,243,176,171,
51,64,34,168,47,8,233,66,205,188,229,16,
85,34,160,227,115,173,46,222,105,250,59,112,
172,58,93,65,193,67,247,238,101,108,106,150,
76,161,138,36,150,22,208,150,198,79,99,91,
230,130,225,34,233,1,36,33,51,69,130,145,
90,211,212,84,194,109,252,238,248,12,111,158,
184,64,80,49,208,210,167,176,43,185,203,93,
177,230,117,179,239,243,165,93,17,34,149,65,
76,213,135,101,89,28,59,117,150,116,238,202,
242,209,134,236,97,88,237,67,233,121,112,217,
26,0,57,214,203,107,167,166,249,206,143,158,
225,212,217,115,100,50,25,114,249,2,191,121,
233,0,213,106,245,206,247,40,246,237,220,196,
192,133,33,74,233,113,36,91,187,161,22,24,
117,211,90,115,94,123,169,110,82,115,217,92,
110,63,36,180,26,127,254,247,223,92,128,157,
6,103,75,136,75,8,147,163,122,249,143,231,
222,165,54,125,158,96,162,3,87,231,222,187,
82,254,38,83,163,146,101,122,252,151,56,237,
23,193,18,148,207,142,208,190,249,43,119,39,
230,6,72,173,215,144,45,3,99,133,30,69,
91,68,71,87,151,239,54,224,82,4,125,137,
166,229,238,145,212,69,115,254,50,202,92,150,
85,120,43,200,165,8,58,86,8,195,216,150,
137,52,125,2,199,223,134,28,72,64,160,133,
153,217,163,88,150,69,123,170,133,179,147,69,
62,156,72,235,87,44,148,204,105,140,150,102,
48,89,36,54,45,88,198,202,165,36,1,33,
43,184,147,27,241,205,29,227,241,251,119,34,
132,224,135,175,158,70,120,66,52,242,211,40,
133,113,8,116,209,144,220,248,252,126,62,253,
241,7,248,209,31,78,240,253,223,188,70,126,
126,142,175,125,246,19,228,114,185,37,253,161,
220,154,76,229,58,5,118,122,106,19,166,35,
113,240,212,48,229,250,16,154,34,33,82,59,
56,240,246,97,158,252,196,199,238,108,69,177,
99,219,86,92,154,198,104,69,71,13,196,113,
249,35,107,126,113,117,13,176,211,229,148,43,
27,132,46,203,206,173,9,15,154,166,225,118,
187,57,116,236,125,236,112,239,2,83,72,158,
48,74,56,129,226,142,97,5,91,238,74,160,
117,160,185,233,87,160,119,240,146,232,177,48,
19,239,83,152,217,78,32,177,233,238,228,172,
85,249,234,30,164,75,174,196,245,18,4,125,
46,153,240,117,82,141,35,151,250,151,24,213,
43,21,130,229,56,100,203,183,46,62,33,75,
18,125,73,47,226,122,144,83,45,79,184,84,
2,219,162,179,55,193,153,11,23,168,251,98,
72,66,166,20,220,204,191,62,243,18,117,111,
27,174,228,54,200,206,44,53,70,149,0,237,
222,50,195,181,18,178,238,187,186,39,38,203,
20,163,59,248,193,27,67,56,245,18,122,106,
115,115,140,102,13,143,83,166,110,91,72,241,
141,252,228,217,223,242,247,127,253,23,60,185,
167,147,35,103,135,232,232,239,166,53,149,228,
7,63,253,5,150,35,209,209,154,4,73,98,
174,38,144,124,215,159,3,45,16,199,240,4,
168,158,126,13,211,169,178,49,236,230,161,7,
110,79,35,118,213,57,174,253,27,251,248,100,
221,96,104,120,152,137,249,50,90,164,99,77,
47,174,53,110,172,135,77,182,98,208,117,153,
77,100,183,108,229,63,127,247,54,15,109,105,
101,54,91,68,113,37,150,222,32,9,244,63,
2,37,49,117,254,55,88,78,149,88,235,163,
232,190,91,251,61,182,81,39,59,249,62,209,
174,165,177,7,171,81,163,174,94,224,114,91,
214,153,143,224,235,234,185,43,237,111,128,244,
74,17,217,52,40,196,219,80,174,3,193,197,
252,58,56,107,247,8,210,197,6,214,173,42,
93,151,160,59,230,185,110,166,155,99,91,116,
249,12,190,244,228,35,0,28,61,126,146,98,
160,15,215,37,136,73,184,220,208,186,135,171,
249,36,106,164,157,161,153,179,16,184,118,210,
140,227,216,88,181,10,194,27,65,141,46,202,
51,57,218,77,201,29,65,50,234,152,249,41,
2,254,16,146,36,209,215,219,77,95,111,247,
162,209,91,171,227,245,184,25,154,76,83,113,
181,160,197,55,174,64,128,13,209,161,27,120,
165,2,206,182,62,106,181,42,173,81,63,62,
159,247,182,228,197,85,41,138,74,165,194,187,
71,142,177,169,175,151,221,219,183,240,157,95,
190,198,90,109,144,234,13,42,10,195,182,201,
87,12,66,151,172,40,73,18,72,201,109,12,
142,157,65,250,80,141,135,227,252,113,196,36,
26,213,60,85,207,73,148,68,153,236,176,70,
170,239,214,86,116,206,156,127,13,111,252,74,
184,177,148,27,68,244,44,197,106,237,134,69,
53,63,133,55,218,117,87,226,175,145,42,190,
32,226,18,68,162,202,215,86,20,110,77,192,
26,245,132,227,56,76,231,110,13,54,46,209,
172,151,8,172,160,33,167,152,57,197,231,190,
240,104,147,247,27,13,58,219,83,216,23,6,
176,140,18,182,101,162,126,216,91,253,208,62,
151,132,140,156,220,186,36,102,83,47,101,209,
60,126,172,198,226,247,58,99,239,241,233,253,
91,41,148,203,156,30,190,64,209,183,97,193,
3,81,60,1,164,137,163,60,182,119,35,253,
125,187,151,29,231,183,190,241,20,178,44,99,
24,6,223,127,230,121,106,99,243,200,29,215,
110,197,97,59,130,123,183,247,211,214,118,103,
244,156,91,149,162,40,149,74,188,115,244,52,
71,166,33,32,157,164,33,123,89,107,217,221,
141,66,79,208,44,192,235,79,250,150,180,32,
158,178,66,48,63,142,124,153,60,251,163,72,
171,116,108,166,71,126,137,210,215,108,82,102,
88,229,91,62,132,212,214,79,45,207,244,118,
117,161,91,71,64,149,105,117,187,56,237,22,
184,131,41,50,99,239,98,153,85,132,208,16,
194,133,89,43,130,98,17,239,122,244,138,224,
227,93,90,74,222,98,22,217,48,200,71,147,
168,178,64,72,112,181,190,134,226,6,120,124,
174,212,160,97,222,10,99,74,162,39,238,38,
236,114,22,130,204,87,35,43,59,198,159,221,
183,25,85,85,121,238,229,3,148,75,37,90,
91,83,168,129,56,86,163,2,206,82,197,230,
100,46,210,106,103,200,20,85,228,203,218,113,
248,102,15,161,185,125,164,149,20,149,194,60,
78,110,18,127,208,141,89,152,199,152,59,133,
212,121,31,82,116,35,233,92,153,7,246,237,
100,239,174,237,252,211,119,127,142,236,139,97,
10,21,37,222,71,204,171,113,122,224,28,110,
151,70,48,24,36,16,240,51,58,54,78,162,
37,206,197,161,97,218,218,90,9,248,253,168,
170,202,183,190,246,121,126,245,251,87,153,90,
166,72,117,201,122,69,186,120,230,249,87,248,
63,254,246,235,119,132,124,90,149,156,111,105,
105,161,183,183,151,89,87,146,244,185,55,137,
238,122,114,141,22,12,84,110,194,217,0,134,
101,115,102,178,64,111,139,119,1,127,149,131,
41,8,166,238,104,1,97,91,6,165,244,69,
52,119,16,213,27,161,154,159,102,110,242,15,
56,61,231,23,160,182,27,73,75,190,17,143,
38,51,242,22,209,142,7,209,188,1,102,135,
94,197,118,106,52,60,231,145,37,208,132,196,
189,209,32,121,195,68,54,34,76,156,123,6,
167,231,20,97,175,74,213,176,113,76,168,89,
22,142,12,99,3,227,116,110,249,235,187,218,
224,90,134,89,48,186,80,153,45,73,144,10,
185,153,200,46,111,249,171,178,196,154,76,47,
7,166,242,235,31,196,22,18,244,37,188,132,
60,106,179,199,210,181,246,117,122,144,46,191,
69,119,71,51,163,110,124,174,72,200,239,225,
196,200,60,1,49,69,107,60,194,96,166,190,
36,67,43,238,106,240,229,39,63,203,179,207,
253,129,89,226,11,40,131,99,89,124,241,137,
143,241,220,43,111,50,231,84,200,212,243,244,
166,218,120,228,169,207,2,240,63,126,254,42,
114,114,43,39,38,7,217,86,44,18,240,251,
249,203,207,125,130,72,36,194,79,159,249,21,
83,19,117,234,181,28,13,203,97,240,87,207,
83,175,150,248,250,83,95,230,103,47,188,137,
207,173,97,148,179,152,106,128,71,247,239,224,
158,157,91,48,77,147,137,146,64,246,94,91,
248,55,10,115,108,233,233,184,170,146,176,44,
11,33,196,109,163,68,86,237,16,60,249,240,
126,126,240,252,219,232,170,160,94,204,224,242,
71,87,239,82,55,172,155,6,7,217,14,92,
152,169,176,191,87,227,143,37,145,114,102,248,
119,136,238,105,156,89,15,82,221,15,145,52,
114,191,177,248,125,14,200,150,255,150,141,167,
81,206,34,43,46,166,206,60,143,112,129,234,
9,0,18,37,227,40,74,111,118,33,54,177,
41,224,69,151,5,233,186,67,195,53,76,172,
207,102,119,52,132,95,93,100,179,186,101,243,
110,166,64,38,56,74,163,156,67,243,134,110,
189,115,118,135,64,145,190,124,166,233,81,132,
155,177,168,182,176,78,200,173,146,175,25,228,
43,6,165,122,115,31,105,178,64,85,196,154,
144,167,92,213,184,225,120,225,245,72,22,43,
63,184,170,145,157,160,83,205,177,111,251,61,
204,166,231,120,253,141,55,49,212,40,51,117,
129,84,207,240,213,167,158,108,182,250,126,241,
0,83,211,39,33,185,189,185,103,234,26,111,
189,123,132,246,84,156,169,76,9,179,81,71,
243,133,40,68,119,241,219,23,94,230,254,253,
123,248,206,211,47,112,207,182,205,28,60,113,
142,125,123,119,225,241,44,86,114,75,209,13,
124,255,183,175,227,114,185,168,153,18,155,83,
62,62,247,233,199,57,114,252,36,7,143,141,
18,220,249,103,88,102,3,189,94,229,149,67,
167,209,91,122,145,227,61,72,150,137,46,4,
39,6,142,144,140,5,121,229,221,147,72,209,
173,215,253,78,45,16,227,244,116,6,237,224,
33,30,185,255,30,36,73,194,182,109,94,127,
231,8,182,99,115,126,108,14,67,82,216,218,
26,226,225,251,247,162,40,202,71,202,139,202,
114,155,232,154,46,147,44,144,173,26,106,251,
94,100,109,109,7,138,148,235,230,77,253,136,
68,224,143,67,73,152,181,18,179,163,175,160,
108,154,66,242,216,16,46,194,135,143,249,49,
36,156,145,110,34,29,251,111,137,103,51,51,
252,2,21,235,52,138,20,192,142,229,49,171,
18,229,244,8,170,219,79,52,244,105,230,39,
158,71,110,203,226,85,100,146,186,134,84,40,
209,30,240,17,216,45,19,80,149,43,214,197,
37,11,250,252,110,178,190,34,181,242,236,186,
43,10,199,177,73,143,28,160,94,155,161,99,
211,83,216,166,193,252,216,123,120,163,183,127,
199,206,82,48,6,206,82,63,193,171,203,120,
117,153,214,144,142,237,56,84,27,214,117,131,
194,215,162,217,117,78,137,21,18,108,73,249,
241,172,240,132,58,201,29,98,116,240,24,195,
153,55,105,143,7,25,156,156,199,219,22,195,
44,206,227,209,20,84,181,233,73,127,254,137,
71,152,73,207,241,244,139,239,160,118,237,199,
145,221,72,194,102,223,238,29,148,223,122,151,
84,87,148,3,39,78,35,37,118,48,154,245,
240,184,215,203,159,61,186,159,96,192,199,19,
143,62,120,9,50,181,177,28,22,12,29,165,
99,31,22,160,2,231,166,79,211,159,153,231,
119,47,189,70,203,182,102,48,93,86,52,100,
69,195,240,238,197,181,32,15,155,115,223,72,
238,225,103,7,135,17,238,54,116,101,101,173,
209,37,95,11,167,171,94,6,126,241,42,81,
183,68,166,234,96,69,250,16,178,130,212,222,
133,138,196,217,90,153,19,63,125,153,128,71,
99,99,194,203,246,45,155,8,133,130,183,156,
23,175,224,176,235,185,58,178,16,32,171,200,
186,23,89,91,91,41,250,114,231,75,220,8,
181,134,239,188,146,248,43,220,237,90,129,209,
179,223,69,237,79,35,105,75,133,131,61,30,
199,101,245,96,139,50,170,20,163,101,211,99,
172,231,33,5,86,163,70,102,234,45,204,236,
4,114,71,26,85,0,20,145,105,110,170,116,
246,59,144,9,144,8,60,133,53,22,6,173,
192,222,173,126,244,124,9,145,158,199,9,248,
8,126,224,69,52,12,208,150,194,100,113,93,
67,82,193,42,174,127,235,236,217,161,63,80,
79,30,192,145,28,38,6,116,84,53,134,209,
122,140,201,11,224,15,238,192,23,233,188,34,
249,225,246,241,40,230,154,30,69,40,126,21,
33,44,225,117,173,125,236,134,101,147,43,155,
235,250,13,137,128,190,98,37,1,160,234,94,
156,45,159,70,18,50,147,229,60,114,216,226,
177,109,9,94,61,209,32,233,147,150,200,167,
68,60,198,206,158,56,199,210,99,216,8,148,
75,253,147,30,125,176,121,220,232,187,103,39,
168,0,178,203,203,119,126,252,44,223,250,139,
47,162,235,250,194,51,138,197,34,13,195,194,
78,15,55,97,42,217,133,30,138,131,36,176,
133,74,60,30,227,255,252,111,127,207,155,239,
29,103,120,230,12,82,168,27,217,229,198,170,
22,208,139,35,84,203,5,100,79,24,211,17,
136,88,47,106,48,65,238,252,65,92,59,30,
95,49,100,36,187,60,144,216,70,6,32,0,
31,158,41,89,247,34,71,123,168,105,58,71,
243,85,206,189,116,144,111,125,229,201,91,14,
73,173,138,203,242,249,60,94,175,151,29,61,
9,14,103,235,55,160,40,110,46,115,42,119,
120,245,174,101,214,25,61,247,31,168,219,210,
75,107,160,28,176,46,180,211,222,245,23,168,
158,91,99,69,20,210,103,153,43,253,22,101,
99,142,171,157,84,43,135,109,8,231,72,143,
60,143,43,24,193,54,220,212,109,27,36,16,
211,25,28,85,65,50,109,164,92,1,169,84,
197,218,209,135,29,95,108,235,90,179,108,236,
154,132,230,94,255,243,65,12,107,30,52,7,
9,48,58,14,81,62,239,69,139,217,84,171,
67,52,98,135,200,140,180,147,140,125,1,151,
63,113,219,5,21,47,143,81,172,7,205,21,
27,172,119,187,192,76,185,65,60,160,93,181,
16,112,89,75,251,82,146,131,230,13,98,204,
11,94,61,60,128,233,73,96,25,115,216,182,
141,16,130,23,95,121,157,39,62,241,48,15,
63,112,47,193,147,103,112,128,128,111,41,194,
97,35,97,230,38,209,156,26,141,158,135,248,
143,223,31,167,158,30,228,191,255,221,215,81,
20,133,66,161,136,48,171,180,73,85,138,229,
42,143,61,242,0,191,125,119,0,145,216,130,
75,50,209,117,29,93,215,249,204,227,143,96,
219,54,71,79,156,34,91,152,34,154,8,176,
231,241,199,22,121,204,48,120,251,240,113,6,
166,167,9,111,250,216,186,241,145,203,31,198,
46,13,127,36,124,186,42,69,49,51,151,225,
185,159,191,64,123,212,131,149,55,176,236,58,
242,198,79,174,14,206,176,157,27,78,141,189,
66,208,218,160,220,193,9,52,19,131,63,66,
221,180,84,73,216,243,42,114,102,35,93,27,
191,178,164,251,228,122,82,173,48,195,124,245,
57,148,142,220,202,54,116,215,24,149,241,28,
138,225,80,175,213,145,207,143,81,204,214,17,
133,49,220,186,140,105,59,168,170,64,140,207,
46,81,20,37,211,194,41,120,209,91,227,235,
11,157,153,13,140,82,105,193,74,19,46,208,
183,151,49,39,93,40,114,4,201,151,193,170,
230,104,212,242,204,143,191,143,230,241,34,20,
55,145,182,93,43,59,223,120,221,61,138,165,
49,138,155,77,233,82,99,221,191,161,97,218,
156,154,40,210,219,226,189,110,65,224,178,60,
230,141,33,34,41,148,137,99,220,255,200,30,
132,16,76,76,78,113,124,96,136,253,123,118,
48,59,151,193,6,118,111,95,218,120,210,178,
44,10,53,11,203,202,160,82,67,141,110,4,
125,35,146,30,224,212,217,243,236,218,190,133,
64,192,143,18,76,48,90,175,241,153,253,91,
232,104,111,167,237,236,8,83,92,209,42,11,
33,4,247,236,94,254,44,107,85,85,121,248,
254,125,180,39,135,121,246,88,26,37,222,185,
126,243,169,199,25,159,156,162,189,117,245,9,
59,215,11,43,220,52,69,49,59,51,75,195,
145,25,169,122,232,74,168,36,67,30,142,206,
167,209,2,43,223,240,149,134,117,211,109,152,
82,221,92,168,167,184,147,200,177,45,70,79,
254,8,167,103,136,15,50,69,157,170,132,93,
151,209,230,118,209,182,245,214,225,232,86,163,
194,244,204,51,136,158,249,43,55,43,160,10,
9,195,110,90,230,151,219,184,69,205,207,182,
246,89,186,199,50,100,199,114,12,95,44,33,
201,18,146,3,186,46,179,121,71,4,115,203,
98,225,93,222,48,57,158,45,226,50,183,33,
171,235,123,226,156,80,52,132,255,74,163,68,
110,169,83,189,144,67,111,21,56,25,63,129,
237,155,208,244,48,211,197,255,192,150,12,138,
131,135,208,229,94,2,161,109,127,180,30,133,
105,217,235,30,196,94,52,228,28,46,204,148,
216,146,242,227,211,215,6,149,137,182,221,60,
115,120,18,183,113,10,204,6,74,172,151,127,
250,206,79,80,116,15,95,124,226,33,0,94,
124,229,117,26,245,42,229,90,3,159,47,128,
240,70,208,2,201,102,6,209,7,66,61,152,
224,245,145,73,222,59,243,60,170,166,35,123,
219,145,200,98,92,202,59,214,181,230,70,212,
213,213,9,212,116,122,142,137,169,25,156,233,
1,156,72,235,186,193,153,114,168,149,131,199,
78,243,212,26,20,197,141,120,34,171,250,154,
143,61,120,63,167,71,126,77,126,126,148,25,
215,38,102,42,85,212,216,234,10,70,102,139,
55,223,138,41,223,177,138,194,198,112,50,104,
238,166,48,112,234,2,235,66,27,129,192,78,
98,91,238,187,101,227,200,140,189,75,161,254,
14,114,207,236,135,32,61,65,151,199,69,183,
207,77,197,106,246,230,26,44,86,136,235,26,
69,195,164,110,59,204,170,130,77,18,136,116,
150,208,182,54,250,188,115,100,38,139,4,252,
42,161,160,134,185,165,27,244,166,71,84,183,
28,222,30,41,209,152,232,166,165,237,145,91,
242,109,225,224,35,100,210,191,64,120,171,216,
195,29,40,94,21,187,102,33,251,115,72,231,
118,208,190,253,241,166,96,8,180,16,40,127,
130,130,253,26,82,247,40,213,236,20,74,206,
143,59,252,209,181,164,95,79,143,162,102,216,
183,150,215,29,56,55,83,98,119,231,218,33,
84,53,212,138,73,83,222,120,199,222,225,91,
127,247,117,2,126,31,133,66,145,255,124,246,
37,50,134,27,87,168,23,17,212,200,169,46,
20,119,83,48,10,121,41,220,160,133,90,169,
72,42,122,48,222,140,187,185,3,188,57,48,
136,209,104,80,44,22,33,4,37,188,252,240,
23,191,227,201,71,31,96,124,114,146,142,182,
54,34,225,165,137,23,213,106,149,247,207,156,
227,158,93,219,153,78,103,56,120,106,4,119,
207,189,235,26,243,146,36,193,156,214,197,225,
227,167,184,103,215,173,51,100,86,252,69,153,
249,121,76,211,162,167,173,133,211,245,6,142,
47,126,41,77,114,229,148,175,26,75,142,51,
189,89,84,172,90,183,213,201,102,215,59,49,
110,97,62,166,79,65,96,241,120,86,107,36,
68,107,239,151,154,197,72,183,8,135,156,25,
122,153,106,248,13,228,148,137,4,180,121,92,
116,122,220,248,13,3,109,46,135,60,49,141,
213,145,196,157,104,246,245,138,186,130,136,90,
3,220,26,13,203,230,204,148,137,236,56,216,
209,32,47,4,3,132,239,141,178,219,173,161,
84,235,156,56,109,179,53,26,162,97,219,164,
107,6,231,7,20,228,218,39,209,228,34,245,
202,28,110,255,250,215,187,4,226,155,32,253,
69,230,135,223,166,107,235,55,16,138,74,101,
126,10,79,56,121,197,28,71,82,251,241,20,
186,72,15,188,130,199,147,192,27,249,104,91,
144,44,151,245,116,51,121,244,86,147,105,57,
20,107,38,62,249,218,26,101,126,226,8,13,
99,6,167,33,19,140,239,92,222,195,240,132,
57,115,126,144,106,181,206,197,116,25,171,101,
39,98,126,106,213,50,105,65,0,71,55,112,
112,58,143,45,183,161,1,82,168,131,18,29,
252,248,181,51,52,26,13,220,3,179,116,134,
85,54,117,183,33,132,96,114,38,205,217,241,
44,245,80,31,135,158,125,29,124,73,60,221,
251,110,202,201,159,215,245,42,220,126,6,198,
198,185,231,22,158,148,176,226,175,154,153,77,
243,139,151,222,97,95,127,138,168,87,38,151,
185,8,158,221,43,87,18,21,131,243,51,235,
83,77,92,170,27,216,182,131,184,77,130,218,
215,59,49,14,154,103,121,228,38,78,160,237,
109,122,88,118,94,193,239,221,131,30,140,83,
153,159,186,101,99,173,152,39,144,125,38,138,
36,177,55,226,111,102,36,1,242,133,81,196,
76,166,249,13,166,5,153,60,98,100,10,9,
7,41,87,2,69,70,234,239,34,112,94,229,
144,87,224,74,181,96,154,38,33,77,161,174,
41,200,110,23,129,182,6,182,229,112,112,62,
79,113,52,72,220,247,21,208,100,102,39,126,
131,144,85,96,199,45,249,198,64,124,51,129,
248,230,15,47,210,178,215,234,129,22,58,182,
125,173,105,117,127,232,172,227,91,239,81,92,
59,235,233,70,168,121,98,158,125,171,119,198,
53,127,173,151,210,76,79,62,139,211,58,134,
208,29,140,19,41,12,107,154,152,252,73,28,
211,193,29,106,93,16,196,86,180,159,35,133,
42,178,162,33,181,220,156,0,165,226,9,98,
27,117,108,163,142,184,4,139,170,45,125,88,
249,52,34,24,103,28,184,120,62,139,227,88,
168,158,54,68,178,179,41,64,147,205,186,9,
163,90,186,101,51,153,175,89,183,116,229,86,
164,40,108,219,230,133,63,188,129,99,217,156,
29,155,67,138,244,172,10,114,154,43,53,123,
214,85,39,0,0,32,0,73,68,65,84,24,
74,87,214,205,138,177,29,40,214,76,130,119,
10,252,228,216,140,158,254,33,210,230,11,56,
22,72,10,24,231,98,196,238,127,228,150,15,
69,182,99,192,60,221,62,157,233,90,3,143,
34,227,83,100,236,254,78,68,182,0,13,3,
121,96,248,74,235,176,35,201,241,89,31,158,
96,39,115,7,103,80,183,204,210,214,169,178,
193,183,120,80,139,156,180,121,57,157,165,158,
246,16,113,127,6,151,175,133,201,243,207,160,
168,81,18,189,79,172,175,53,158,30,196,27,
237,94,210,38,164,60,63,68,113,254,28,129,
200,206,59,130,77,202,254,240,66,175,167,155,
77,186,42,227,214,164,155,158,88,178,28,105,
138,32,228,81,137,249,52,124,186,130,117,149,
119,22,102,207,226,180,143,34,46,229,110,168,
59,155,6,211,204,196,44,142,92,69,154,73,
210,209,247,205,133,195,128,62,124,40,208,135,
201,106,84,155,61,161,46,59,150,55,59,121,
132,106,101,28,183,218,133,203,23,166,82,152,
196,27,110,6,159,135,79,252,39,150,146,69,
242,85,145,230,19,200,62,129,223,187,19,183,
103,177,81,160,162,123,169,229,103,49,106,35,
56,142,141,191,165,15,199,177,113,22,90,146,
220,26,79,205,116,133,153,153,157,37,209,114,
107,26,131,42,215,246,34,102,121,238,229,55,
104,111,77,240,223,254,247,255,66,181,90,229,
183,47,29,96,86,95,121,145,84,177,106,48,
146,91,255,190,68,133,59,72,81,148,231,199,
176,106,85,132,36,33,20,135,198,136,78,56,
185,7,233,150,101,218,56,228,167,79,81,40,
30,35,231,205,19,2,38,171,13,20,73,194,
167,200,92,40,86,41,155,22,187,119,247,163,
188,119,106,9,239,59,65,31,217,182,4,163,
186,139,11,167,60,56,141,41,188,247,204,178,
189,221,67,232,82,225,151,3,28,153,47,144,
174,155,212,222,239,32,154,186,15,89,214,24,
27,250,14,162,111,18,123,62,192,196,185,159,
211,214,191,178,243,42,44,163,142,80,212,85,
205,143,109,55,152,31,61,66,180,123,177,48,
177,86,110,198,96,114,51,39,208,124,161,235,
10,154,143,154,60,165,124,179,123,108,116,125,
226,36,137,128,206,240,220,250,237,205,168,79,
35,21,210,241,104,43,179,248,67,173,123,40,
206,190,137,104,93,58,38,185,237,18,60,155,
24,99,252,194,143,104,239,251,58,138,122,237,
212,252,185,145,55,40,72,175,35,41,14,202,
208,6,58,182,54,189,196,98,253,24,116,12,
145,29,185,72,110,228,15,88,5,193,134,192,
63,82,202,142,96,123,51,168,93,217,230,3,
82,67,32,65,121,208,187,160,40,102,134,95,
166,106,159,196,9,207,129,4,174,244,125,88,
245,58,153,194,11,200,4,176,204,6,194,91,
195,37,109,0,71,16,77,62,136,203,23,93,
151,185,117,69,218,56,244,254,57,62,251,216,
109,160,40,14,30,59,75,94,111,35,119,246,
12,247,95,2,196,146,209,16,211,133,149,215,
65,212,173,91,163,97,87,123,248,252,71,73,
222,104,23,113,243,147,204,156,56,128,136,200,
72,102,137,200,214,181,85,90,87,242,227,152,
86,142,64,252,234,231,82,59,142,141,101,212,
169,151,211,100,230,127,143,41,229,145,98,243,
56,17,9,179,24,71,0,85,203,98,87,200,
207,161,57,139,225,2,244,71,0,211,90,170,
36,60,58,185,77,221,188,158,175,176,67,83,
184,247,225,10,46,151,68,76,15,47,128,10,
13,219,97,180,92,163,108,218,152,182,141,39,
152,66,117,133,152,45,254,20,121,67,51,237,
86,138,23,104,84,135,153,185,248,50,150,83,
196,31,220,129,39,216,190,108,22,148,227,216,
140,158,253,79,36,183,73,208,253,0,225,214,
221,43,138,223,4,18,91,174,232,38,26,237,
184,143,209,129,239,193,134,115,140,94,60,67,
247,134,127,88,128,24,110,71,170,122,3,8,
123,253,44,254,152,95,99,60,91,193,188,201,
123,212,163,201,116,69,61,248,221,171,195,235,
11,115,239,35,130,213,107,34,87,82,207,16,
179,131,47,209,210,253,9,20,253,234,45,185,
109,44,68,160,142,240,89,200,181,69,3,210,
110,24,8,25,148,222,38,172,234,156,107,1,
33,168,22,47,162,116,102,151,188,203,206,184,
137,37,154,85,220,147,231,127,73,163,245,48,
194,213,156,43,251,66,55,225,196,62,198,207,
60,141,210,159,71,202,5,145,220,5,228,206,
50,38,205,19,30,211,67,85,218,55,126,117,
221,214,111,238,22,158,31,178,236,74,230,11,
5,222,122,239,24,231,46,12,177,123,203,6,
70,170,1,254,237,103,207,35,249,226,72,222,
8,242,42,44,177,168,87,35,83,18,24,214,
250,225,161,126,183,66,50,160,113,167,144,213,
168,49,87,252,13,218,254,44,230,164,78,50,
248,55,107,110,242,151,175,29,64,88,174,5,
69,81,156,187,128,16,42,222,200,98,123,239,
137,51,63,199,27,232,163,82,26,197,217,184,
120,110,196,76,174,157,251,219,138,76,149,226,
140,229,85,94,204,129,75,173,82,173,71,232,
104,47,66,185,2,18,56,161,0,181,13,109,
216,62,15,239,205,230,8,168,50,33,85,161,
34,44,100,73,44,65,158,39,171,53,206,22,
202,11,94,184,85,177,153,117,158,69,116,126,
168,54,35,84,164,60,56,138,178,231,34,115,
185,99,56,99,94,100,43,132,46,119,226,210,
91,209,189,77,43,90,146,4,154,30,198,234,
62,65,190,242,44,217,139,175,34,36,13,197,
138,145,234,254,226,181,83,108,63,164,80,102,
47,28,192,176,211,136,146,132,210,155,33,51,
121,136,120,215,67,183,45,159,232,149,18,178,
101,80,140,173,79,43,106,33,65,42,164,51,
150,185,57,45,198,21,33,209,22,117,211,18,
212,215,212,82,167,218,24,66,120,175,46,39,
28,19,156,11,157,24,82,134,225,145,255,151,
182,228,223,98,27,117,188,209,118,178,227,239,
227,242,54,173,235,177,19,191,192,114,242,136,
220,54,172,228,25,130,158,102,124,42,51,250,
14,137,196,23,152,30,254,9,114,247,28,118,
89,16,208,239,109,86,101,59,226,138,16,138,
82,234,198,147,104,99,230,220,171,52,90,143,
44,40,9,43,167,18,246,223,75,49,59,128,
39,220,129,33,207,33,112,97,217,26,176,232,
13,217,114,110,93,249,163,104,123,200,231,243,
4,131,235,95,140,187,172,162,208,93,46,66,
1,63,170,75,231,253,139,83,96,219,168,129,
24,122,107,83,24,173,38,104,99,88,246,186,
246,97,242,186,20,250,19,62,184,67,186,61,
85,139,83,76,14,255,12,101,99,22,73,128,
147,245,227,233,92,27,180,96,27,117,156,124,
0,111,106,241,160,20,163,158,195,237,111,186,
202,150,217,96,252,196,207,137,117,63,132,59,
216,70,185,56,184,20,122,8,141,243,222,100,
10,143,94,36,30,76,95,182,97,189,212,44,
27,39,28,192,124,96,23,89,33,120,51,157,
67,45,55,104,216,54,88,54,175,205,102,145,
36,184,47,218,100,210,188,97,146,107,152,156,
188,12,102,156,203,37,241,155,231,113,93,110,
169,125,0,39,248,44,106,229,6,138,13,114,
200,130,80,1,40,80,103,148,202,140,135,88,
229,171,232,193,102,16,87,86,124,88,128,240,
216,136,174,102,157,135,237,76,51,51,226,165,
181,239,115,43,158,175,150,190,71,104,113,30,
38,51,114,28,51,63,143,238,79,220,214,188,
82,215,61,235,234,81,0,164,130,46,50,197,
6,149,27,140,85,248,116,153,190,246,0,154,
182,54,131,205,113,28,12,59,203,114,32,149,
99,1,163,237,200,213,56,13,207,24,74,215,
40,202,172,31,73,150,153,26,250,57,158,76,
47,245,114,14,105,30,186,35,127,131,227,153,
67,116,143,98,93,76,162,207,63,64,112,203,
86,234,165,121,114,246,243,88,153,79,162,216,
45,56,204,225,212,85,10,179,231,136,116,238,
193,178,11,75,223,89,21,184,148,246,5,247,
66,82,22,21,152,83,118,33,92,26,229,200,
75,200,222,166,194,85,114,161,43,158,97,106,
25,140,90,25,85,95,159,195,136,212,88,23,
111,31,61,197,147,31,127,240,163,81,20,46,
151,139,183,15,31,71,117,251,80,100,129,95,
145,152,175,231,87,245,224,116,177,206,68,182,
138,145,47,227,120,61,235,50,120,183,42,216,
148,244,33,223,65,45,60,172,70,13,201,95,
65,82,192,174,10,34,241,135,214,148,119,109,
25,117,70,206,252,27,34,98,82,179,7,25,
59,63,70,75,234,207,136,180,237,99,252,253,
103,73,246,63,129,144,53,92,158,4,115,179,
47,160,102,90,16,184,175,200,115,137,5,167,
150,241,240,29,70,202,85,82,110,141,17,219,
230,100,166,112,9,86,106,222,173,72,18,178,
144,216,17,244,18,115,169,152,142,195,225,249,
34,21,115,81,216,76,102,186,105,9,77,160,
132,151,119,143,173,10,212,83,6,250,50,136,
161,220,82,97,226,220,115,216,57,47,126,151,
192,82,51,136,101,96,136,70,228,40,243,19,
9,34,109,171,128,237,36,9,119,32,129,39,
178,251,182,231,21,181,81,67,182,76,214,23,
96,144,232,142,121,56,51,89,92,115,24,182,
37,224,34,229,2,69,94,59,252,91,203,78,
65,120,102,57,12,9,233,194,22,52,87,156,
106,242,16,74,164,217,31,204,78,7,201,228,
14,34,247,100,169,143,79,16,106,249,24,115,
179,47,146,30,57,128,42,165,48,24,69,238,
158,198,25,105,10,123,205,29,64,194,75,217,
243,38,82,67,67,0,114,164,142,37,206,51,
54,240,125,84,41,1,54,216,115,126,148,106,
55,46,165,133,120,87,51,185,196,180,242,72,
151,107,48,111,5,179,92,5,161,129,175,153,
181,232,216,13,52,179,11,70,146,8,161,97,
99,162,152,98,93,139,74,27,133,57,222,59,
249,14,143,63,124,31,178,188,190,173,41,150,
149,80,182,109,179,127,71,63,195,211,57,210,
133,42,247,237,218,202,145,129,81,86,106,115,
100,203,205,44,39,214,217,206,239,138,121,81,
228,59,171,207,147,109,84,145,66,151,152,125,
214,71,164,119,207,154,158,83,152,29,64,244,
78,33,116,128,52,230,180,111,161,213,135,161,
78,145,29,63,6,178,141,227,216,216,129,28,
13,109,6,207,252,199,169,143,133,145,59,178,
215,129,36,108,74,70,179,192,206,178,29,62,
145,140,160,72,18,175,205,102,169,89,54,237,
151,138,240,124,138,140,3,28,203,46,85,18,
233,124,43,33,223,28,138,124,117,17,231,184,
85,26,49,249,170,88,180,103,211,12,117,67,
103,170,20,39,226,175,240,129,157,106,87,192,
154,115,163,180,85,17,129,6,141,252,220,202,
45,244,114,134,236,244,187,212,140,81,196,188,
130,36,105,8,52,36,35,64,188,243,161,53,
231,224,175,155,81,161,168,235,238,81,52,189,
1,133,120,192,197,108,97,181,53,78,18,93,
49,55,137,128,139,198,13,182,3,41,151,70,
145,147,87,170,42,179,32,240,235,61,20,67,
207,161,4,23,127,87,182,143,99,217,227,72,
117,176,212,18,249,224,207,208,34,80,158,60,
133,79,217,133,225,0,2,12,121,178,57,82,
89,65,54,195,72,237,195,75,224,33,57,228,
224,232,195,232,217,189,20,70,83,8,220,180,
247,45,141,43,24,98,154,203,109,81,169,226,
70,117,251,113,166,163,16,153,194,158,115,35,
219,17,42,214,0,30,185,143,74,110,12,130,
121,132,112,147,153,120,139,120,231,250,100,51,
202,149,52,143,125,252,97,138,197,34,161,208,
250,118,97,94,214,4,16,66,240,224,3,247,
83,204,103,217,220,25,231,245,83,19,140,159,
59,137,109,174,140,25,138,85,243,150,108,164,
181,182,3,248,168,168,81,201,50,59,254,218,
2,214,41,121,107,84,215,152,171,95,171,141,
92,82,18,151,214,172,30,67,213,3,24,213,
2,142,90,162,156,25,163,210,242,123,234,201,
55,112,104,32,194,117,42,230,57,34,238,79,
97,207,94,255,44,139,154,109,115,34,87,162,
219,235,198,45,11,170,150,77,205,178,105,117,
187,216,18,244,226,83,100,38,171,117,14,204,
230,152,174,94,178,170,28,137,137,76,15,97,
95,26,143,235,218,240,228,108,174,141,120,112,
242,154,215,184,212,26,201,240,24,185,82,140,
90,163,233,149,10,15,212,171,17,42,175,71,
177,11,26,146,179,114,168,99,122,248,121,26,
29,111,33,250,198,161,123,24,127,223,32,117,
103,12,127,100,19,67,239,252,132,225,163,63,
100,226,226,211,204,141,188,113,89,186,227,71,
71,21,127,8,83,213,232,62,241,214,186,191,
171,43,234,38,176,138,224,179,34,75,108,105,
245,145,8,220,28,139,217,180,114,203,75,163,
162,23,161,57,8,83,95,86,122,9,55,168,
189,69,132,27,36,13,108,81,66,211,163,88,
151,194,46,246,101,231,136,7,125,247,98,78,
94,57,94,107,58,68,160,165,31,197,9,99,
250,199,168,230,22,61,27,219,50,176,148,165,
109,109,156,134,11,199,177,144,218,155,215,57,
243,33,170,156,66,238,31,39,151,62,142,29,
156,64,180,205,33,117,141,81,14,190,194,220,
216,193,155,191,96,142,67,72,179,104,111,77,
50,56,50,190,238,252,113,85,95,81,8,193,
63,252,237,55,248,179,199,30,225,137,189,189,
68,182,127,124,197,205,233,26,183,40,211,169,
102,88,220,73,148,157,62,140,182,107,18,233,
82,220,218,49,20,180,53,116,133,117,28,155,
134,51,245,33,215,48,136,99,91,76,142,62,
141,156,42,224,107,233,193,25,107,67,248,45,
148,68,243,68,49,59,54,129,109,57,196,92,
95,193,26,186,122,90,157,237,52,217,98,172,
82,227,189,249,2,14,160,203,130,123,163,1,
182,133,188,8,73,194,114,28,90,116,141,126,
255,34,172,56,53,223,69,91,100,248,154,158,
196,2,92,17,154,96,46,191,178,202,236,150,
208,4,133,114,100,81,128,38,192,189,55,143,
121,166,21,111,168,119,201,230,105,148,175,238,
45,105,36,48,39,117,60,178,64,19,130,130,
97,33,185,77,230,74,191,64,219,63,142,216,
113,6,171,243,56,229,196,11,12,159,249,95,
216,150,241,145,243,76,57,16,97,120,231,131,
116,158,122,151,208,204,216,250,1,80,146,68,
95,194,215,60,123,251,122,144,175,38,179,173,
45,176,162,131,136,86,12,179,105,33,236,249,
43,147,100,36,195,67,161,118,24,17,189,122,
192,189,80,9,83,174,93,242,6,125,101,172,
186,1,245,230,119,56,194,88,168,110,15,37,
119,208,26,252,54,214,217,212,210,108,62,26,
216,166,129,84,247,163,101,118,163,7,23,247,
134,144,21,68,117,105,219,7,185,53,75,49,
123,6,117,118,103,51,192,30,154,69,244,76,
55,121,108,83,22,181,179,182,0,85,9,175,
73,169,113,152,90,97,250,166,174,151,81,206,
82,202,206,114,113,120,148,11,163,235,95,160,
123,5,87,204,204,204,48,50,186,148,33,35,
33,63,141,169,179,20,39,47,172,76,81,152,
183,70,128,143,207,87,185,147,200,237,107,195,
156,12,224,92,50,86,229,88,133,236,212,145,
85,63,103,230,226,139,208,190,116,141,20,41,
196,220,232,27,208,213,204,255,22,138,74,107,
199,95,32,157,223,142,115,201,17,84,162,13,
114,190,167,153,155,127,30,39,179,60,204,98,
152,46,100,217,188,12,134,186,212,20,80,18,
52,12,157,115,133,10,47,77,103,120,63,87,
66,145,36,146,186,70,167,87,103,112,106,27,
45,161,9,144,86,102,36,40,178,177,160,144,
86,164,28,47,115,253,117,173,66,67,215,80,
182,76,83,41,12,47,42,226,137,99,140,15,
254,24,199,177,169,20,102,169,151,22,97,169,
252,244,105,92,129,8,34,108,82,177,108,92,
114,51,91,75,78,148,145,59,115,8,221,94,
178,43,140,106,241,186,71,118,222,124,35,241,
234,115,55,186,237,94,74,225,22,186,78,190,
179,126,56,180,144,216,148,244,95,179,203,107,
208,173,176,181,213,143,235,38,167,163,71,219,
238,199,93,186,127,97,111,0,216,53,9,185,
30,71,163,13,59,211,52,72,236,250,229,243,
213,100,10,143,171,68,169,218,52,184,228,128,
69,190,240,206,130,49,134,106,96,94,182,142,
122,48,65,215,230,255,138,57,28,197,177,192,
58,215,74,50,242,117,36,33,83,15,158,160,
33,141,125,168,109,138,68,192,119,15,118,105,
17,38,149,52,135,70,203,113,124,129,62,196,
200,54,228,208,226,245,242,50,219,74,116,206,
48,53,251,83,230,70,222,188,105,198,135,85,
205,243,192,254,61,84,12,137,121,195,197,63,
253,203,191,82,46,175,95,77,204,21,171,29,
139,197,248,254,211,191,226,119,47,31,192,178,
154,2,63,209,210,130,223,239,167,49,63,114,
91,121,20,185,138,65,161,106,114,167,80,160,
101,11,65,231,227,56,239,111,195,42,131,164,
66,197,56,181,170,138,245,220,204,73,106,222,
163,72,151,25,115,86,70,35,24,223,69,195,
156,65,82,154,207,45,148,14,145,158,254,61,
78,67,90,178,185,100,175,141,19,155,69,146,
150,183,6,51,133,4,81,255,162,235,221,162,
107,164,107,6,63,57,27,228,104,126,142,145,
114,141,186,229,160,8,193,92,221,96,178,90,
231,208,120,146,120,112,146,186,177,186,243,73,
132,100,163,10,137,79,38,194,184,132,32,226,
82,175,121,237,194,60,122,178,228,203,17,68,
160,70,197,117,136,242,37,190,172,214,71,145,
55,78,49,50,240,111,12,101,126,204,249,153,
31,48,240,254,79,49,107,37,210,51,207,147,
203,191,141,93,110,126,119,209,48,175,26,188,
149,4,216,189,18,217,204,192,45,229,143,203,
219,189,152,245,10,229,204,48,229,204,8,102,
189,105,16,153,154,139,76,91,47,145,201,97,
212,250,77,48,146,28,135,114,102,233,158,214,
20,193,198,164,143,254,164,143,100,192,69,80,
87,112,107,50,62,151,76,103,212,77,255,58,
38,143,180,116,61,130,51,213,130,221,0,78,
238,32,152,255,50,177,182,135,105,219,248,101,
2,141,79,81,57,16,163,113,50,76,113,54,
72,58,223,202,92,33,69,58,223,74,174,20,
35,17,110,26,78,146,2,98,203,48,194,103,
55,189,6,181,113,133,194,23,138,134,110,247,
163,142,222,67,247,166,111,227,14,182,50,61,
252,91,132,44,16,29,51,20,102,151,174,187,
55,220,133,152,238,195,46,168,151,41,4,131,
236,212,17,252,190,109,216,163,215,105,177,34,
129,220,157,166,220,242,60,35,163,255,15,67,
39,254,253,134,231,74,11,183,241,234,225,115,
140,230,26,52,100,15,182,43,200,208,58,66,
80,87,72,139,239,62,253,107,58,58,186,40,
22,10,252,243,79,158,163,61,30,228,11,79,
60,204,199,239,221,206,91,39,135,185,110,148,
194,1,195,188,117,61,100,134,210,101,182,182,
5,80,87,16,212,182,204,6,245,98,26,163,
158,163,86,153,197,237,107,37,63,113,22,220,
85,82,61,127,142,226,242,172,251,120,131,201,
77,148,231,39,48,106,32,123,129,214,49,38,
6,158,161,125,211,151,175,91,72,102,25,117,
178,213,23,144,219,47,89,14,54,24,227,62,
36,213,160,238,204,18,138,236,35,157,25,68,
68,43,136,190,49,44,154,155,231,138,144,241,
120,10,61,26,92,54,155,230,114,111,2,224,
68,182,68,161,28,65,82,230,144,46,19,214,
195,165,42,115,53,147,179,211,157,180,70,135,
48,76,13,195,108,226,191,113,189,153,85,82,
179,108,98,186,74,166,110,80,48,44,236,203,
20,98,195,208,209,212,26,134,237,32,46,245,
153,210,101,193,108,173,193,169,252,149,150,145,
99,47,111,193,202,241,2,217,225,119,136,197,
63,129,201,28,146,234,32,245,141,226,7,38,
230,122,73,182,159,97,248,212,20,202,166,44,
194,181,168,252,102,107,215,230,228,172,17,163,
170,140,16,203,143,227,9,182,47,97,112,163,
86,198,170,87,81,116,239,186,241,76,110,242,
8,185,226,59,216,24,204,12,249,232,255,216,
63,34,73,18,165,112,28,17,8,211,122,254,
56,51,61,91,49,92,107,175,46,207,140,191,
71,94,121,30,215,192,78,84,17,161,165,239,
225,5,30,12,121,84,66,30,149,154,234,65,
15,222,154,32,191,36,43,68,253,159,162,158,
206,16,221,209,236,194,250,65,223,179,124,254,
16,158,135,155,21,209,58,224,103,5,89,152,
18,72,170,133,81,43,224,242,71,150,162,36,
201,253,11,233,215,217,241,67,216,221,39,17,
151,244,64,121,106,128,96,114,233,185,215,29,
27,255,130,129,3,255,3,109,251,28,74,216,
198,49,192,105,29,35,83,153,68,233,89,153,
231,41,169,32,119,20,176,203,21,166,46,252,
142,96,104,7,4,215,214,199,75,40,26,174,
222,7,154,74,163,81,165,210,168,115,240,204,
24,181,122,157,125,123,118,174,191,162,80,85,
141,114,205,32,151,43,224,142,117,48,89,22,
216,182,77,127,223,6,222,60,53,177,162,197,
17,66,194,178,111,141,87,81,55,109,46,204,
150,216,156,244,93,179,223,186,99,91,92,120,
235,159,9,36,54,81,175,77,225,138,204,80,
205,234,84,10,18,94,29,134,79,255,27,193,
208,30,188,129,222,117,29,175,162,7,104,223,
249,25,138,153,77,148,46,14,96,152,25,108,
39,131,89,175,92,179,210,20,160,81,205,114,
57,184,106,205,233,200,153,30,236,248,48,197,
198,113,218,147,95,39,123,97,3,118,244,253,
107,235,242,196,36,185,145,10,222,238,229,150,
207,105,42,6,201,198,239,201,129,35,81,172,
5,105,139,14,93,113,237,185,153,54,218,99,
205,218,140,82,45,72,212,63,131,34,73,236,
9,251,208,196,82,193,94,181,108,206,23,43,
232,178,96,32,95,99,54,215,70,123,124,16,
151,44,48,28,135,232,37,111,66,44,179,134,
233,124,43,209,224,244,146,81,154,150,198,254,
104,128,146,105,113,38,124,129,137,115,121,228,
13,211,75,92,100,77,173,33,203,38,242,206,
165,153,81,217,186,129,196,181,187,242,84,13,
15,109,27,46,50,51,245,67,92,233,62,144,
29,76,114,88,228,113,180,10,146,203,128,162,
134,94,219,69,106,21,181,28,43,165,88,207,
199,8,86,155,41,188,170,238,93,98,68,216,
178,194,248,230,123,104,61,119,156,154,63,196,
124,170,107,229,49,143,236,16,133,244,89,76,
105,30,203,59,141,18,55,152,159,29,198,27,
58,75,99,104,156,218,92,142,158,221,127,187,
230,211,43,87,67,165,204,16,186,55,134,162,
47,38,87,248,99,155,240,47,115,240,161,174,
183,83,203,204,128,98,227,212,93,40,70,8,
199,22,216,254,105,228,160,121,213,104,171,236,
135,244,208,111,17,202,159,227,9,119,44,123,
77,173,49,177,8,85,1,141,232,41,114,211,
27,8,37,119,46,81,98,145,222,237,84,26,
135,129,60,146,10,74,220,128,53,36,47,11,
175,137,213,251,30,179,83,39,41,13,238,37,
185,225,211,107,151,39,19,239,144,138,132,153,
112,41,148,45,63,233,185,204,250,123,20,23,
135,71,41,149,42,116,164,162,196,55,247,112,
116,96,24,39,185,139,151,223,124,143,88,192,
139,140,181,162,105,113,41,226,134,11,120,86,
67,197,170,201,72,166,74,119,236,234,214,93,
173,56,75,172,237,33,230,243,175,210,214,215,
96,246,164,224,193,143,155,76,119,152,228,48,
216,30,87,145,141,131,28,59,114,10,79,228,
239,215,181,205,183,80,52,130,137,45,4,19,
75,219,110,204,12,254,129,224,53,90,113,184,
3,73,228,185,16,92,178,166,228,150,26,182,
247,36,138,215,193,26,45,54,149,128,184,62,
78,41,66,130,154,164,178,156,90,42,84,194,
36,195,99,56,72,100,10,73,108,71,144,138,
140,92,37,206,176,232,125,4,61,243,228,203,
81,194,190,121,222,158,43,16,214,20,182,135,
124,11,233,209,110,89,176,57,224,97,160,80,
97,114,174,135,246,248,224,7,232,7,186,16,
24,182,195,72,185,198,88,165,105,157,149,107,
1,114,165,24,138,210,64,72,54,178,88,124,
215,68,166,155,135,186,166,137,24,14,73,7,
138,41,155,76,96,2,71,113,48,108,9,33,
65,165,230,199,165,212,112,9,129,225,216,124,
96,183,116,121,117,250,253,30,78,230,203,76,
85,235,248,85,153,168,166,54,21,88,161,130,
3,228,203,81,118,181,142,33,201,50,165,84,
17,147,163,139,130,103,41,87,81,155,63,74,
37,187,27,33,214,177,21,200,85,120,113,178,
127,23,194,50,217,242,214,115,12,220,119,237,
38,139,141,114,150,252,220,73,202,206,17,68,
111,83,113,22,203,81,140,124,10,119,111,9,
197,93,198,230,12,106,74,98,252,252,79,80,
245,16,66,184,144,12,29,217,181,15,85,247,
221,244,207,106,84,231,72,79,188,72,207,206,
191,191,238,181,169,190,207,82,43,222,139,99,
89,104,129,224,66,119,136,244,249,183,201,165,
95,198,213,95,187,58,236,211,59,77,122,252,
89,82,174,111,162,121,66,203,76,239,82,147,
65,4,234,228,134,222,34,152,216,190,164,191,
88,188,253,97,46,188,57,136,80,10,168,209,
27,52,132,37,16,161,58,245,145,57,202,243,
99,120,35,29,107,128,14,109,76,71,176,161,
35,193,3,177,8,191,123,241,85,82,61,125,
235,175,40,90,98,17,20,97,113,240,173,183,
112,197,186,232,74,70,168,205,188,207,187,115,
25,240,68,209,124,33,212,249,41,204,122,5,
33,20,132,186,124,22,148,92,173,64,237,210,
198,254,144,21,124,77,90,205,181,181,2,216,
139,194,99,182,12,74,85,39,226,93,28,83,
105,254,2,178,240,97,214,170,228,242,111,33,
130,53,92,106,148,125,125,50,135,75,85,98,
97,5,79,64,226,181,153,50,163,35,2,173,
33,176,26,114,211,221,93,129,162,168,173,34,
135,191,81,202,54,207,153,184,6,85,42,23,
40,141,156,34,228,125,228,170,152,178,85,179,
150,8,43,225,109,206,151,99,64,101,126,10,
211,44,112,189,210,155,186,169,163,186,174,84,
249,249,114,148,13,45,35,116,5,44,132,36,
81,49,231,41,153,22,166,3,123,195,65,42,
150,205,241,92,17,211,82,152,203,167,216,20,
203,146,179,62,80,26,6,166,165,226,85,100,
162,46,149,177,114,141,77,1,47,218,101,120,
246,171,211,69,70,50,173,60,214,59,207,104,
69,166,98,89,184,21,129,42,36,170,150,77,
166,222,160,124,41,17,162,84,13,210,22,187,
184,236,248,189,106,141,141,126,29,247,137,243,
32,73,236,242,123,176,219,226,152,170,202,96,
169,138,97,59,20,29,157,17,195,226,241,176,
15,81,174,50,162,105,68,116,21,127,161,140,
52,49,75,193,213,20,236,123,67,62,252,154,
74,205,178,25,40,52,61,158,170,108,211,234,
118,179,49,172,112,174,80,165,104,152,148,77,
155,146,121,229,250,201,145,6,115,3,111,18,
240,239,184,238,250,126,224,217,22,103,134,87,
196,51,102,189,89,111,99,84,139,215,188,238,
200,166,157,68,207,29,197,180,12,42,203,140,
193,106,212,152,201,254,24,173,175,66,135,199,
69,151,47,136,227,192,43,165,16,159,232,158,
39,168,41,28,201,42,248,20,25,85,146,200,
108,31,162,120,41,163,208,49,97,108,250,109,
148,114,39,178,20,66,145,125,205,212,80,36,
116,127,10,213,221,132,165,140,74,17,165,86,
90,210,173,119,57,50,170,5,50,211,175,18,
142,63,12,142,130,33,101,72,15,190,133,55,
220,131,81,41,52,59,189,94,182,247,46,223,
99,213,252,56,149,250,121,64,198,145,44,108,
169,132,109,24,184,54,94,31,250,145,218,103,
25,63,255,61,168,122,104,105,127,28,179,86,
198,182,76,28,199,193,174,130,211,104,166,215,
46,188,43,147,97,94,62,142,59,148,92,178,
119,91,183,60,73,37,63,78,117,118,20,83,
155,132,72,14,217,119,233,224,49,171,217,222,
131,138,142,218,93,196,54,224,90,246,131,53,
175,32,121,38,153,154,248,9,173,124,227,234,
115,86,41,34,171,174,43,228,173,153,155,100,
119,187,143,87,222,124,143,72,56,64,36,18,
198,90,167,110,195,75,20,133,207,231,35,22,
14,18,217,119,31,247,238,218,66,123,123,27,
0,35,163,99,252,252,133,215,113,71,183,33,
235,94,140,106,9,33,43,87,117,79,83,186,
65,113,186,180,168,58,189,241,149,171,217,149,
94,43,20,112,47,197,29,75,178,66,123,100,
209,141,205,100,254,128,89,46,83,159,174,33,
201,14,157,91,37,220,162,128,34,233,220,183,
167,233,125,188,63,161,19,177,91,192,93,97,
38,27,66,211,83,20,139,39,113,28,139,104,
226,1,84,207,181,11,89,60,145,149,165,120,
10,89,89,192,68,175,70,73,245,11,76,13,
253,156,106,113,136,154,115,1,191,111,59,190,
104,207,98,33,93,173,132,227,89,222,181,84,
213,64,115,44,185,235,11,43,183,86,38,87,
138,17,240,44,77,37,21,239,26,132,158,172,
179,43,87,133,134,129,29,15,97,39,194,212,
44,27,203,244,112,100,58,192,84,165,136,38,
91,108,75,204,176,63,230,231,120,214,96,131,
223,77,193,48,249,67,1,238,139,6,208,133,
96,71,185,138,116,126,20,199,165,97,119,165,
56,151,117,147,43,122,120,106,83,30,69,72,
100,12,153,186,109,179,61,216,180,84,47,150,
170,164,235,151,41,175,171,100,79,101,10,9,
90,124,69,228,108,13,50,5,36,1,106,195,
192,220,208,142,12,248,20,25,125,62,135,215,
235,48,48,223,130,152,185,136,50,56,206,6,
85,1,203,226,3,215,226,147,65,31,149,104,
8,223,192,8,98,62,143,28,11,241,121,221,
5,138,204,249,68,131,183,198,130,108,10,155,
108,14,52,249,100,172,82,231,120,246,42,2,
91,45,163,7,227,205,179,68,114,19,72,142,
140,230,13,45,236,15,219,50,25,59,249,115,
36,108,66,137,123,40,202,127,160,158,61,129,
16,50,154,176,145,80,105,212,66,180,246,125,
113,201,193,55,31,180,202,81,221,215,183,230,
27,129,56,222,209,179,136,101,248,113,122,240,
247,184,251,170,108,14,121,233,185,212,6,126,
44,27,101,111,84,166,92,106,163,181,101,154,
93,33,31,30,69,70,145,36,202,166,197,80,
169,138,34,36,162,46,13,39,225,160,136,73,
242,141,49,198,10,53,84,33,225,85,20,102,
254,127,222,222,44,72,174,252,58,239,252,253,
255,119,205,155,251,90,251,130,2,10,64,163,
129,70,239,100,147,108,146,162,184,136,150,199,
35,141,45,203,182,102,177,39,194,17,19,243,
50,239,122,152,176,222,253,48,79,142,176,103,
108,137,99,75,118,140,165,144,236,160,24,67,
137,100,139,18,187,155,189,177,187,177,47,181,
239,185,239,119,95,230,33,171,11,64,3,141,
70,83,228,220,8,68,161,162,50,51,50,243,
222,123,190,115,206,247,157,239,28,170,140,187,
79,81,174,190,130,146,175,162,89,217,135,22,
247,36,81,136,80,84,226,40,160,179,251,51,
70,222,251,200,213,125,122,107,239,128,30,163,
157,245,169,223,94,227,194,242,203,52,246,191,
143,229,156,38,86,199,24,198,12,197,217,103,
30,184,199,250,189,119,72,206,220,188,119,189,
242,24,125,255,125,71,60,82,136,71,42,82,
87,240,251,3,204,92,133,208,176,8,252,46,
173,163,215,152,59,253,219,116,143,210,56,165,
55,144,153,144,36,4,211,152,162,184,116,17,
169,104,15,221,187,102,190,74,123,223,71,81,
78,161,135,101,188,163,38,190,211,199,25,110,
97,102,103,152,57,251,109,182,175,254,7,146,
56,68,94,222,249,196,247,165,205,251,128,79,
176,158,39,85,168,125,34,200,250,122,10,69,
55,31,138,183,78,224,48,59,83,226,75,175,
124,142,63,250,222,223,208,232,73,246,143,238,
242,194,243,207,253,114,129,2,224,55,126,253,
193,126,89,20,69,188,254,254,77,50,23,190,
254,196,47,90,176,52,50,166,202,200,253,255,
79,145,148,54,20,86,106,233,19,210,186,121,
231,39,4,74,131,133,139,99,196,42,92,94,
200,34,132,246,64,15,124,56,138,88,173,134,
40,42,188,125,0,218,66,147,216,241,8,179,
117,188,155,89,10,229,23,248,219,26,151,55,
182,94,67,79,85,48,205,79,247,115,74,101,
167,88,121,230,127,101,243,230,191,70,89,218,
163,227,126,72,119,63,139,244,167,40,22,191,
72,186,124,138,112,63,139,58,231,60,52,242,
30,19,144,68,33,137,124,178,239,188,86,216,
231,176,179,132,34,35,44,115,64,198,28,160,
207,134,76,153,58,118,38,133,52,116,140,70,
27,245,250,6,205,217,151,137,98,201,87,79,
213,121,191,51,228,153,98,6,145,100,80,251,
99,46,23,39,237,37,63,138,17,76,170,3,
229,206,22,114,191,201,216,42,179,147,93,65,
238,232,84,220,125,126,59,231,146,40,37,154,
97,68,211,245,121,161,148,165,168,171,36,64,
73,87,25,155,58,183,91,121,162,72,227,155,
139,49,253,200,100,107,124,47,91,244,143,201,
242,83,249,24,253,246,17,131,129,79,110,177,
72,248,236,185,251,139,46,170,94,192,13,77,
35,101,140,104,164,76,102,165,164,239,90,28,
154,43,12,103,138,104,194,230,82,180,65,182,
222,6,67,35,201,101,136,83,6,3,67,199,
213,117,122,126,72,204,61,169,152,31,199,159,
12,18,64,50,179,67,107,251,39,164,70,179,
200,212,107,76,233,1,110,51,225,48,76,17,
196,42,81,44,208,228,18,249,153,139,196,161,
71,118,6,190,58,211,103,55,129,51,18,68,
119,72,47,76,120,125,109,72,166,240,121,242,
83,231,249,172,158,6,34,142,208,93,7,231,
161,34,52,198,77,238,242,98,57,195,108,234,
94,122,59,95,232,32,146,8,142,185,164,156,
118,47,20,164,85,101,210,54,140,34,146,251,
172,33,74,186,198,138,34,16,221,1,104,10,
148,7,236,46,95,99,115,107,147,214,208,5,
63,197,236,252,63,194,204,222,11,170,245,181,
215,40,204,92,166,190,251,125,156,102,27,243,
165,14,197,172,74,254,149,61,222,223,173,240,
92,78,229,189,41,143,157,187,191,79,92,104,
50,236,183,120,249,133,152,122,231,61,118,238,
190,141,22,45,147,42,78,33,132,164,84,123,
133,195,157,13,228,226,103,235,195,71,67,133,
240,86,25,179,80,230,204,139,191,137,84,53,
66,207,102,56,186,133,242,212,30,187,107,127,
192,194,226,63,69,235,22,233,173,93,67,44,
109,66,181,203,184,187,77,182,242,112,43,71,
72,133,145,120,27,89,28,17,15,117,148,209,
34,139,23,127,135,250,250,143,72,146,4,169,
106,156,122,238,159,209,61,120,159,126,183,137,
44,126,178,66,45,172,27,20,140,175,124,106,
37,246,200,228,83,55,185,187,115,68,165,92,
132,56,66,45,44,163,15,236,95,74,124,85,
116,93,255,23,191,251,187,191,251,137,15,248,
247,127,246,3,198,229,139,136,251,200,201,56,
244,17,82,62,118,237,95,41,163,19,197,9,
246,104,0,250,147,153,98,137,192,126,226,199,
18,58,160,165,48,84,201,82,197,98,169,98,
161,30,183,57,58,219,239,226,140,15,48,173,
49,151,150,5,115,55,214,233,95,107,144,153,
201,128,161,223,151,177,37,152,41,201,79,91,
125,210,82,39,25,103,208,178,29,150,178,6,
67,205,165,223,185,198,176,125,147,113,111,147,
208,246,72,229,102,62,86,70,143,30,88,138,
242,168,163,222,254,83,156,241,54,105,235,169,
79,125,236,9,79,208,188,78,102,170,207,124,
214,96,170,150,208,182,234,140,199,183,9,91,
48,115,246,219,244,142,62,68,230,124,98,15,
156,119,242,104,179,30,241,80,165,88,126,133,
206,206,135,200,81,25,109,116,26,186,69,34,
181,135,208,31,46,71,165,136,201,166,250,164,
205,1,189,81,149,180,57,64,173,4,19,210,
217,246,200,107,42,153,32,68,217,220,167,95,
94,98,169,218,65,8,65,78,10,178,27,251,
104,237,62,162,213,69,24,58,118,160,162,26,
146,43,45,131,185,172,67,166,144,69,180,7,
104,222,152,170,189,67,165,189,78,102,216,66,
182,122,40,245,54,154,101,160,101,211,44,164,
39,46,163,2,136,18,184,61,28,35,101,132,
20,42,95,152,137,80,165,224,192,241,79,154,
145,205,222,28,159,159,239,114,65,87,57,252,
171,219,100,139,6,226,242,25,212,205,125,146,
108,6,121,212,162,92,111,90,156,223,71,0,
0,32,0,73,68,65,84,17,23,242,252,108,
88,100,42,55,224,210,120,140,28,140,48,113,
169,36,117,170,75,42,235,6,188,230,44,16,
206,231,104,149,211,244,203,5,198,185,12,111,
185,1,251,81,76,99,156,130,216,226,153,202,
4,116,175,116,199,12,195,79,6,96,161,37,
4,73,11,175,221,225,215,141,13,106,238,128,
249,254,33,171,197,128,51,207,22,233,89,67,
188,65,66,161,250,42,86,97,150,230,254,123,
60,19,30,82,146,2,217,233,131,166,146,94,
223,101,53,223,162,157,90,167,117,36,73,23,
150,79,156,16,20,237,211,7,93,101,28,35,
135,29,130,242,131,215,104,123,231,117,180,185,
235,60,93,78,163,12,109,100,103,128,60,106,
161,108,236,161,220,217,65,214,59,200,222,144,
68,215,192,156,0,137,236,143,72,84,5,245,
234,26,194,245,144,107,123,8,18,228,97,19,
229,246,54,178,217,69,30,78,206,103,78,83,
153,234,213,57,245,82,153,65,182,79,103,107,
159,76,238,226,189,77,116,129,207,209,246,255,
139,60,189,133,204,68,84,170,146,87,42,121,
102,82,6,69,83,35,8,83,172,150,34,238,
132,26,142,174,145,158,242,185,80,84,89,42,
152,76,205,56,180,244,13,234,107,183,49,245,
57,204,236,20,74,84,38,104,196,147,185,138,
108,143,71,173,40,73,66,136,118,139,36,182,
6,34,68,41,132,40,51,67,18,171,75,212,
178,72,23,231,105,237,188,133,175,174,33,114,
99,148,146,131,183,15,149,185,87,232,7,63,
70,153,114,144,25,143,222,237,13,50,153,115,
244,27,31,48,238,111,144,46,156,66,8,129,
144,10,189,250,135,104,181,17,83,101,201,112,
16,17,15,21,138,51,207,145,41,158,58,249,
236,169,236,12,170,59,79,80,87,8,195,1,
194,242,31,209,126,210,176,235,77,12,163,134,
158,126,244,62,231,200,119,39,131,127,31,139,
183,106,42,75,63,73,113,117,253,16,57,117,
158,56,112,113,27,155,124,238,249,95,236,230,
200,223,251,189,223,123,252,62,138,215,94,127,
139,113,110,5,245,231,48,173,83,196,196,108,
76,25,155,28,254,18,120,237,180,174,80,173,
88,84,178,198,3,62,44,113,20,162,24,22,
106,246,128,111,174,14,224,205,77,122,35,5,
61,114,25,125,184,143,114,105,153,80,170,164,
210,2,55,22,188,219,234,227,68,49,209,208,
192,55,250,188,82,177,176,163,152,84,198,39,
73,141,248,250,180,142,31,15,249,241,214,117,
198,183,215,31,106,15,124,234,13,108,151,41,
212,158,127,224,57,131,230,45,218,155,239,178,
252,210,63,57,33,203,34,223,166,185,243,19,
114,213,11,8,47,203,216,139,168,26,9,251,
246,113,176,168,56,168,193,143,169,111,238,83,
202,255,29,186,251,223,71,153,235,163,159,25,
129,2,177,217,39,112,70,44,157,253,167,168,
166,69,119,239,10,249,197,167,217,221,248,14,
100,182,62,165,225,119,15,72,198,97,196,74,
38,69,38,10,81,175,175,51,26,6,44,57,
183,72,218,105,196,112,76,193,11,136,15,219,
212,247,134,84,74,38,178,239,240,150,125,150,
124,81,146,46,69,236,218,1,211,89,29,145,
196,147,86,207,71,55,175,169,35,92,159,184,
144,69,147,202,3,155,240,0,10,186,202,87,
107,69,94,59,136,137,68,12,72,138,250,132,
243,24,4,33,65,104,144,53,61,206,231,44,
148,15,238,48,182,67,102,86,103,144,7,77,
226,233,10,40,130,184,82,32,209,85,26,249,
12,195,142,96,69,23,136,173,46,163,190,79,
38,163,66,146,160,95,91,227,11,151,206,48,
183,152,240,250,126,10,59,200,32,72,168,230,
15,16,2,58,131,41,64,242,247,78,143,136,
18,133,205,145,67,221,253,116,15,164,40,175,
34,162,1,107,11,211,12,21,201,115,251,117,
100,189,131,58,118,121,254,226,105,76,243,10,
127,177,254,135,152,229,175,128,229,50,202,166,
201,54,58,132,237,17,31,57,204,123,135,61,
94,204,123,92,171,253,148,173,59,71,84,230,
190,253,25,50,205,228,196,23,42,244,108,218,
251,111,17,49,68,49,239,242,234,92,110,98,
228,120,103,27,49,184,79,236,32,37,241,169,
89,146,116,10,225,250,136,131,137,157,140,114,
123,11,153,77,51,218,235,81,127,247,128,98,
94,163,60,124,180,72,66,110,31,34,2,5,
235,238,46,151,15,122,132,23,18,254,102,237,
59,168,162,68,140,138,211,218,192,124,161,11,
18,76,43,228,197,114,1,229,184,170,95,202,
59,108,182,107,156,202,55,88,206,7,28,218,
62,31,54,21,254,228,86,133,111,44,143,152,
201,192,151,231,243,236,150,122,92,189,253,135,
104,205,167,113,122,77,82,197,26,179,167,127,
147,205,59,255,10,117,181,69,226,8,146,72,
16,247,82,24,238,121,34,39,34,149,201,144,
216,33,140,5,177,217,39,140,6,152,44,83,
61,253,242,228,59,10,123,196,233,17,74,50,
201,84,156,228,54,221,131,41,130,166,64,57,
230,149,245,115,67,220,209,17,110,188,9,243,
123,212,55,18,166,79,127,147,36,137,49,130,
21,178,131,30,233,12,40,75,109,70,219,239,
147,227,194,67,223,79,166,124,138,76,249,20,
187,119,7,36,149,59,15,223,123,86,192,212,
242,55,201,84,127,190,253,236,170,145,70,53,
142,147,235,246,93,158,57,51,255,203,231,40,
62,58,238,110,108,241,214,149,59,12,210,167,
80,255,150,70,105,217,148,194,225,47,104,149,
108,74,147,148,51,6,229,172,78,50,14,49,
63,230,51,227,141,90,236,108,252,1,43,231,
125,190,34,198,216,239,215,137,71,1,41,19,
180,233,2,227,83,203,24,106,130,153,145,56,
227,152,66,94,240,92,152,37,74,18,14,80,
217,194,229,189,142,79,152,36,44,90,38,187,
182,139,174,8,70,65,196,133,105,131,122,246,
6,59,183,218,204,44,252,125,140,76,229,33,
85,201,206,205,127,143,89,170,145,203,60,75,
174,54,241,192,95,190,244,63,29,147,114,77,
220,65,131,78,227,77,252,252,117,204,202,37,
2,167,79,115,255,71,36,74,64,32,246,144,
11,125,14,222,221,65,89,221,69,81,56,105,
187,24,138,100,38,101,240,116,46,205,126,121,
151,107,239,101,38,98,114,64,173,70,199,164,
170,199,160,113,141,252,212,37,118,110,125,135,
168,188,73,114,16,144,49,158,102,232,109,159,
248,75,61,90,84,243,224,223,54,70,14,185,
98,150,220,202,28,173,31,222,198,120,99,29,
43,173,160,42,10,173,182,75,191,231,99,234,
10,149,10,104,113,192,171,151,60,54,250,41,
78,21,135,216,118,149,48,55,68,251,136,111,
176,76,176,93,162,103,207,33,119,142,192,52,
136,171,5,252,80,37,136,21,130,72,33,99,
120,168,50,98,125,16,114,212,157,231,233,153,
3,64,178,59,118,113,142,193,166,61,152,98,
182,188,75,66,137,196,212,169,150,13,152,171,
33,238,236,16,149,242,32,160,225,250,92,73,
4,97,103,136,16,22,121,77,37,206,101,184,
243,55,251,60,247,66,101,146,125,70,49,178,
211,39,153,51,41,103,219,100,227,152,36,17,
108,213,207,99,25,35,138,217,6,47,149,53,
202,27,71,188,63,85,97,199,126,60,81,186,
121,248,20,150,57,194,247,114,60,191,124,23,
87,10,100,146,176,63,59,197,92,16,34,134,
99,76,219,65,233,13,41,171,219,108,141,254,
24,109,201,197,77,18,172,102,15,55,81,113,
91,54,113,156,96,232,146,84,38,226,66,193,
37,42,172,179,121,245,63,83,158,249,246,19,
113,20,247,3,198,209,198,247,40,94,188,193,
32,8,249,250,76,249,164,137,149,228,38,96,
127,82,162,197,49,178,209,3,247,8,49,252,
88,203,98,48,70,74,193,220,76,10,211,146,
143,85,100,197,185,12,194,212,73,103,116,220,
119,111,241,229,23,99,110,105,71,92,61,154,
102,250,212,61,30,236,185,82,22,227,99,178,
233,143,70,107,182,90,83,204,230,59,124,107,
49,160,237,53,216,111,47,82,50,155,24,106,
200,162,101,98,93,8,184,221,127,27,167,23,
225,138,187,28,220,245,152,153,251,71,12,119,
175,147,74,47,34,20,141,72,29,51,20,87,
137,79,95,195,217,155,34,176,123,44,158,249,
231,164,10,15,219,201,87,230,94,69,53,190,
193,193,245,239,145,152,54,153,212,41,10,51,
79,51,238,110,2,29,52,41,8,141,132,225,
214,93,162,244,0,85,75,136,147,201,119,100,
119,247,241,106,111,114,113,41,77,63,8,137,
246,74,168,126,241,83,90,131,169,71,203,116,
98,73,115,235,39,152,185,169,71,42,178,158,
168,181,230,187,120,27,111,224,121,30,175,252,
119,95,255,229,2,69,20,69,252,213,27,111,
179,209,114,241,173,25,180,234,179,252,109,157,
92,6,78,200,90,99,12,214,207,63,148,164,
43,146,114,70,167,156,209,177,140,123,217,213,
163,110,95,69,53,17,35,133,11,217,24,255,
173,29,90,251,14,139,171,121,176,12,228,108,
137,92,85,103,60,140,81,129,84,90,210,30,
103,41,167,39,189,103,71,13,24,171,58,77,
215,39,163,42,12,194,16,33,4,71,142,207,
214,216,229,92,214,98,57,99,114,211,108,177,
190,249,111,208,246,86,201,229,159,191,247,62,
211,69,172,252,2,225,226,207,104,183,239,34,
219,255,132,76,249,244,201,223,91,7,63,33,
40,222,64,89,118,97,115,22,195,152,229,96,
231,255,153,24,212,1,23,243,105,246,108,149,
222,124,11,165,248,96,171,40,167,169,156,205,
90,8,1,243,25,131,235,153,91,96,62,152,
225,9,21,130,160,71,115,231,199,112,102,29,
5,112,182,55,168,78,127,131,65,247,71,48,
253,201,125,210,251,173,52,114,154,74,86,145,
84,118,143,8,130,144,217,255,230,89,162,158,
131,183,121,136,59,114,168,148,116,22,231,172,
201,46,13,41,32,78,184,125,59,128,90,154,
37,203,228,142,11,90,189,5,113,12,170,66,
112,126,133,157,65,149,196,78,17,85,22,1,
129,104,39,152,90,128,34,98,164,140,185,93,
159,229,233,153,93,118,122,37,230,107,183,184,
92,46,146,28,191,23,41,4,237,65,141,114,
174,206,185,156,133,214,232,50,184,117,200,216,
9,201,199,9,113,121,2,18,126,28,179,54,
116,88,201,164,24,135,17,227,177,194,105,41,
80,250,67,138,69,29,226,152,184,82,68,132,
33,162,55,98,169,82,228,174,184,7,148,153,
84,255,196,164,176,104,90,108,46,78,211,26,
126,250,244,115,218,28,16,39,10,181,140,195,
75,165,44,250,79,39,51,44,201,84,153,100,
182,74,84,61,139,250,227,247,184,115,119,200,
153,213,3,14,206,204,65,2,137,166,48,30,
4,88,89,193,214,161,131,239,199,44,46,89,
16,197,40,55,55,185,248,204,42,169,151,91,
92,189,241,135,24,157,51,84,230,190,134,158,
42,60,190,46,148,202,49,8,8,76,79,3,
3,14,108,143,57,107,146,80,69,103,151,136,
103,171,8,199,71,116,250,8,219,69,52,59,
143,124,181,126,223,39,147,209,62,221,157,57,
73,136,167,43,36,167,74,136,217,42,217,171,
107,176,190,203,23,117,21,187,50,67,39,48,
49,52,23,75,145,24,82,112,224,120,76,155,
250,9,87,216,25,103,136,99,201,98,169,69,
99,152,71,81,98,102,114,93,106,115,135,220,
56,154,231,194,244,228,254,168,24,26,86,57,
199,15,157,14,209,213,211,84,206,191,138,145,
45,145,202,77,17,71,33,66,74,156,110,157,
176,17,19,95,63,139,172,116,57,255,138,96,
239,230,31,163,180,207,145,45,173,146,46,126,
108,222,68,168,132,153,29,132,91,164,182,244,
85,14,214,254,12,229,220,21,78,101,83,156,
201,90,180,188,128,91,195,27,84,107,49,71,
13,133,108,126,66,174,143,122,107,168,75,49,
53,83,99,24,132,16,75,22,158,250,45,2,
231,225,170,43,142,34,252,113,155,169,197,111,
177,119,184,142,50,243,96,214,172,214,2,168,
109,177,127,231,63,178,184,250,207,32,73,62,
211,98,184,143,206,65,40,13,212,180,197,120,
108,99,24,191,56,169,118,183,215,187,199,81,
84,47,124,145,119,111,237,176,231,152,232,181,
213,79,125,163,79,194,81,192,241,60,197,120,
128,195,147,13,239,124,156,163,152,43,154,172,
78,165,201,91,218,67,171,78,31,53,160,102,
247,119,9,148,219,44,55,119,80,252,128,106,
65,162,38,17,172,76,54,132,37,89,139,48,
72,80,213,137,129,209,209,160,128,166,134,232,
74,68,198,18,68,182,198,124,126,66,174,118,
253,144,32,142,57,116,124,52,41,209,20,73,
197,208,72,0,207,244,81,203,13,90,71,27,
164,205,115,168,250,4,8,135,157,53,226,226,
62,73,171,128,33,103,49,179,211,184,131,22,
27,239,255,159,147,30,109,126,146,101,219,141,
42,177,121,7,185,52,177,202,40,27,26,167,
210,38,186,34,56,61,171,114,224,120,100,53,
149,32,142,201,168,10,203,233,20,69,253,222,
119,189,37,251,196,70,252,144,212,174,156,253,
54,113,232,18,228,142,61,185,6,5,220,225,
33,204,239,60,254,60,105,46,125,187,76,214,
116,248,218,116,137,249,157,3,146,148,193,206,
108,149,234,81,19,57,83,70,59,51,139,37,
98,180,233,34,201,210,12,178,209,33,153,169,
114,125,63,197,82,37,164,122,161,68,99,52,
203,243,179,29,180,159,221,102,223,155,166,237,
22,104,231,22,56,101,223,36,191,164,83,180,
198,147,127,198,144,156,93,39,99,6,88,86,
68,37,51,228,202,193,18,45,223,230,243,69,
143,124,146,112,117,236,114,163,63,38,136,19,
28,47,75,41,61,228,92,218,68,251,224,46,
119,111,244,64,8,74,254,144,228,252,50,40,
146,237,177,75,213,156,136,21,14,29,159,214,
168,68,33,51,36,83,239,177,183,222,39,138,
18,50,207,157,162,53,83,163,61,85,226,102,
16,49,8,238,241,14,142,151,193,50,71,72,
96,69,145,220,114,188,19,137,232,39,29,35,
39,79,198,8,113,3,157,231,203,3,190,187,
81,100,249,156,134,213,232,34,91,61,100,171,
79,82,201,35,27,61,100,28,115,180,213,231,
133,211,101,22,20,73,121,191,137,33,194,73,
27,70,192,96,24,96,219,17,166,169,32,18,
80,45,131,219,170,202,234,156,130,151,63,164,
190,182,137,149,58,141,212,140,71,14,150,42,
97,128,217,220,199,157,90,32,87,189,64,123,
39,196,22,187,44,23,117,82,234,113,130,229,
7,8,77,37,201,90,160,107,196,51,21,132,
42,17,189,225,137,232,240,36,64,116,124,12,
67,162,62,129,167,147,95,44,32,242,25,214,
187,115,68,51,53,82,21,21,101,231,136,217,
105,193,95,53,86,41,100,90,156,207,167,249,
160,51,34,165,42,84,205,123,156,203,108,161,
75,41,61,66,145,49,185,148,67,228,196,164,
252,1,178,63,98,184,237,81,110,174,161,236,
55,16,221,1,90,214,226,230,122,26,45,174,
161,167,74,56,131,67,134,205,59,180,91,127,
201,168,177,137,34,115,196,202,152,204,234,45,
190,124,81,99,46,109,48,59,27,112,96,220,
161,215,93,35,234,73,136,161,125,248,58,163,
254,26,195,225,21,66,111,132,153,81,136,93,
147,177,245,55,156,159,54,56,155,179,144,199,
251,227,151,171,58,115,150,193,78,215,37,37,
62,135,59,58,98,48,126,11,189,60,230,92,
46,141,19,199,212,131,33,162,87,67,51,11,
15,197,197,131,245,255,66,103,240,35,74,213,
47,97,119,14,73,10,143,118,139,150,165,33,
189,221,107,244,214,183,201,79,95,120,160,229,
248,73,28,197,177,222,154,246,149,191,68,36,
49,225,184,203,254,225,33,207,93,122,250,23,
6,20,127,241,215,239,240,159,254,195,239,79,
138,134,164,250,212,68,106,214,249,5,187,16,
10,152,45,152,88,34,197,94,247,179,121,211,
100,76,149,185,226,147,35,107,18,71,52,15,
126,192,51,149,38,225,238,16,5,31,153,209,
38,8,125,123,155,232,233,201,196,117,42,45,
177,71,49,99,45,68,53,58,188,83,87,248,
194,220,36,232,150,245,132,33,176,51,118,185,
63,145,26,4,33,227,65,136,46,4,11,105,
147,41,83,103,103,236,50,88,238,51,216,187,
78,245,212,100,238,33,72,14,137,7,10,181,
252,111,146,46,79,178,151,253,205,63,194,124,
190,117,162,227,235,142,170,164,87,14,144,218,
189,154,40,171,41,92,239,219,116,188,128,132,
132,4,88,78,155,92,237,141,152,54,117,186,
126,64,89,73,145,110,28,178,93,206,227,62,
98,173,172,232,206,98,157,155,163,219,188,223,
146,58,66,160,19,123,226,177,173,39,85,9,
136,99,73,24,39,28,216,30,181,66,142,215,
17,60,19,199,200,70,23,217,30,16,62,189,
66,116,102,1,17,71,40,183,38,3,120,141,
76,141,45,169,242,212,194,144,88,132,168,114,
18,88,227,172,197,92,124,124,45,237,29,146,
88,6,31,133,100,49,24,163,188,127,27,36,
68,175,76,50,52,197,113,56,85,221,96,217,
247,40,221,222,97,231,204,2,166,156,144,158,
239,118,38,75,147,46,23,51,100,59,125,236,
142,77,20,39,140,134,1,163,142,67,246,168,
77,188,48,197,169,116,138,3,199,227,70,127,
196,171,138,96,121,170,201,245,118,141,197,87,
210,60,61,91,102,240,225,14,205,31,223,34,
245,229,243,52,77,131,170,161,81,51,117,246,
108,151,59,141,41,42,185,201,228,119,90,83,
201,180,251,84,74,121,122,126,248,200,139,186,
121,236,122,107,168,33,95,175,12,120,223,72,
240,199,42,23,179,14,55,92,201,171,31,137,
29,220,16,237,221,155,160,72,10,101,19,219,
14,177,214,247,176,196,71,188,141,65,92,46,
48,37,26,84,42,6,71,71,14,205,166,135,
101,133,76,135,33,2,216,28,185,12,131,16,
101,101,159,253,198,191,34,222,54,153,153,249,
109,172,226,252,199,58,24,10,174,117,47,113,
170,46,255,10,135,119,29,250,181,171,20,116,
21,37,142,81,110,109,78,246,160,199,9,98,
228,64,28,79,72,108,41,33,142,9,159,59,
143,122,123,27,198,14,83,51,230,35,167,228,
31,44,167,82,140,231,151,88,31,207,99,222,
116,57,91,190,206,238,110,25,119,181,68,122,
170,76,88,43,161,53,61,218,131,105,222,116,
109,92,127,10,63,236,115,58,147,160,221,79,
44,58,30,234,218,46,118,211,166,217,52,200,
22,71,152,106,204,153,68,158,120,124,137,62,
196,179,53,236,124,133,133,149,235,40,206,77,
178,166,164,148,213,40,234,42,123,227,22,87,
126,216,192,200,21,57,85,86,73,29,47,81,
74,171,10,175,214,10,188,46,250,140,236,63,
103,224,74,148,197,248,164,165,251,140,153,102,
247,112,76,123,235,22,242,162,199,240,19,52,
11,169,162,160,125,243,39,100,115,151,72,170,
135,132,241,196,138,63,173,40,200,76,196,168,
125,13,43,255,240,132,124,54,191,74,220,242,
81,20,21,53,169,226,197,60,146,128,71,128,
178,216,37,174,116,233,238,191,79,121,233,229,
39,138,123,129,51,34,37,60,62,247,236,69,
20,85,227,224,240,128,56,142,145,242,111,111,
218,104,219,54,251,35,241,201,28,197,47,250,
152,45,154,132,73,204,81,239,201,23,163,20,
211,159,77,152,58,106,173,147,82,186,148,118,
15,24,181,135,12,162,4,71,209,41,124,225,
28,152,6,201,113,86,37,4,216,1,220,176,
199,92,46,74,252,68,229,79,214,4,223,90,
10,49,133,114,50,193,235,247,21,148,220,125,
100,44,96,71,49,175,213,187,44,88,230,100,
144,76,64,124,44,72,76,162,144,36,138,145,
86,68,123,243,39,164,203,75,4,118,15,181,
56,38,171,43,12,143,7,201,162,72,69,215,
30,108,156,109,141,38,191,87,13,141,166,23,
160,73,193,173,193,120,226,151,20,199,172,213,
23,153,91,24,144,221,218,167,156,121,24,60,
99,79,144,181,158,197,183,187,248,230,230,201,
192,93,196,16,141,105,226,122,14,185,248,201,
222,56,81,172,162,200,136,213,172,133,38,5,
239,40,42,231,50,38,149,227,32,141,31,160,
190,127,155,68,85,17,247,169,127,52,17,113,
249,178,32,174,21,81,111,109,162,153,105,156,
156,70,54,140,8,163,24,145,8,196,98,121,
210,195,118,61,208,117,148,27,155,19,217,101,
202,154,156,12,63,64,249,240,46,37,69,146,
168,10,241,83,203,204,89,169,19,98,221,84,
36,73,34,112,194,8,213,243,9,252,251,44,
76,130,132,36,61,121,172,28,219,24,9,172,
102,45,82,166,206,130,231,179,57,54,64,140,
97,121,134,124,57,79,120,117,131,163,239,93,
225,236,239,124,17,180,73,229,120,237,168,202,
66,161,69,197,84,232,135,48,10,67,118,43,
5,118,71,143,150,25,78,246,104,236,243,57,
17,179,47,37,185,157,14,47,155,6,255,149,
179,252,189,74,139,63,218,91,226,244,162,205,
84,163,205,230,219,187,36,9,44,204,165,73,
149,76,166,255,193,231,72,110,108,32,6,99,
226,233,50,98,104,19,90,38,106,57,143,116,
60,230,102,39,55,101,82,41,48,152,169,98,
132,49,103,12,141,110,108,48,109,233,116,10,
30,195,200,99,247,131,171,196,97,128,102,230,
79,60,140,100,28,97,141,6,15,180,100,107,
75,95,227,195,15,238,114,253,116,139,146,161,
241,133,103,206,222,127,210,17,67,27,97,59,
40,183,182,136,107,37,146,66,150,224,153,85,
68,16,34,226,24,174,172,193,241,249,78,202,
121,68,127,52,1,26,32,58,191,204,122,180,
138,209,108,114,182,249,99,44,37,128,58,24,
115,89,104,142,8,207,45,99,29,182,40,153,
57,20,99,200,130,165,81,74,13,184,89,159,
230,70,239,144,203,165,123,160,38,59,125,222,
252,153,96,228,23,80,100,194,192,83,249,252,
124,7,114,38,78,181,132,146,36,136,40,226,
205,16,98,98,190,84,123,184,5,183,146,75,
177,253,185,6,227,183,115,8,87,192,125,212,
170,46,37,85,67,99,28,70,40,199,10,192,
103,139,89,102,45,29,137,32,214,92,26,59,
29,180,182,206,249,217,71,171,46,107,166,78,
119,225,26,237,214,6,74,1,34,15,246,198,
30,165,227,74,63,252,216,26,212,147,86,110,
237,34,185,218,69,146,36,102,212,59,164,151,
159,35,183,27,96,60,213,120,192,54,228,228,
94,30,24,140,154,187,248,163,62,249,217,11,
88,197,185,79,73,198,37,249,124,1,169,40,
92,126,250,28,159,127,241,23,183,185,241,251,
127,253,54,202,49,215,170,232,186,254,47,190,
245,223,255,111,79,44,247,252,44,173,167,251,
95,51,173,171,28,14,220,39,110,61,45,87,
172,199,174,86,188,191,245,228,219,93,24,253,
17,191,170,111,178,113,187,207,169,69,147,114,
217,192,44,89,96,232,36,165,28,72,73,146,
64,215,15,184,209,180,24,202,1,117,215,199,
210,61,146,160,204,179,213,24,207,137,233,19,
210,232,134,112,231,60,201,72,69,20,71,8,
49,1,10,33,224,233,124,26,41,38,158,68,
97,4,241,240,34,86,126,14,33,37,233,204,
42,189,250,85,196,116,157,222,222,117,6,222,
219,100,23,29,236,48,62,33,178,198,94,142,
180,57,124,20,39,200,108,202,160,227,135,76,
155,6,253,32,36,12,77,214,26,179,60,51,
83,231,244,214,58,241,165,85,14,17,212,157,
128,112,171,4,195,60,113,43,75,202,121,134,
218,210,151,57,216,248,83,196,210,61,63,46,
145,31,19,102,119,81,138,143,7,104,63,48,
137,19,137,169,187,172,13,29,94,46,231,72,
16,100,215,118,16,199,230,121,137,162,16,189,
112,30,121,172,140,65,8,82,174,141,117,110,
106,18,236,239,108,83,24,31,114,71,121,134,
106,251,14,123,155,35,164,2,6,49,253,221,
128,244,254,54,162,209,65,28,147,195,194,15,
192,245,81,239,238,34,124,127,242,187,161,19,
207,213,16,8,16,2,93,74,218,94,64,125,
148,98,54,235,82,113,61,228,96,72,171,57,
249,60,66,17,20,43,38,232,26,234,59,55,
72,11,65,126,182,130,114,115,11,177,113,128,
57,53,69,172,106,24,106,136,236,15,209,15,
26,108,109,142,48,250,125,204,229,42,215,134,
14,251,67,147,111,29,220,96,169,55,64,155,
169,208,244,67,246,93,159,71,25,32,219,94,
134,48,214,248,186,53,98,102,251,128,130,166,
162,122,1,158,161,113,35,172,241,210,120,139,
124,201,231,167,246,12,215,181,25,106,165,57,
20,153,167,47,171,148,205,1,241,202,28,202,
230,1,209,153,5,146,108,154,164,144,225,79,
189,34,202,92,154,98,60,9,134,73,165,128,
108,116,208,253,128,25,199,37,115,212,162,28,
133,24,157,1,197,78,31,45,157,226,250,213,
30,149,213,159,209,239,189,135,63,40,144,202,
77,33,142,239,137,168,112,111,134,65,42,42,
233,212,57,188,29,201,160,225,161,148,71,228,
52,117,210,238,146,2,76,157,36,155,158,72,
198,115,25,18,93,1,67,159,216,175,127,112,
7,17,132,147,11,83,64,146,50,113,159,61,
135,226,251,12,102,103,121,51,120,142,11,233,
109,170,91,87,8,34,129,166,76,190,176,221,
120,145,217,206,13,148,221,58,221,27,135,164,
87,106,236,5,146,211,133,128,51,217,20,203,
5,143,31,110,101,209,245,1,37,99,210,42,
76,114,105,230,151,84,78,89,61,74,186,71,
170,100,176,95,169,240,150,243,18,215,215,203,
28,21,92,152,203,48,118,42,44,101,85,106,
105,255,17,114,111,65,201,208,216,218,12,168,
76,5,84,114,15,71,225,186,235,159,232,251,
138,134,134,46,36,134,34,241,227,152,119,70,
89,94,93,245,40,155,143,78,80,45,85,97,
211,113,144,217,0,18,240,222,45,49,8,5,
73,214,165,183,159,34,167,189,130,145,170,126,
98,92,60,184,251,103,176,114,141,180,49,164,
40,191,5,157,60,65,216,68,166,31,108,111,
38,1,80,105,18,136,35,134,157,91,36,182,
130,97,85,62,177,245,164,104,58,163,250,22,
223,248,210,75,100,179,217,95,24,72,220,217,
216,228,74,29,148,84,158,239,254,187,127,249,
139,1,10,223,238,161,104,230,99,129,66,74,
65,107,232,63,214,44,240,126,160,88,172,88,
143,29,57,186,31,40,90,59,127,194,215,204,
107,200,222,144,116,214,36,165,29,111,144,11,
66,226,149,57,2,117,66,138,222,24,140,105,
122,62,253,113,134,88,27,112,212,153,195,243,
74,152,70,151,213,130,66,224,195,64,132,116,
226,128,36,219,66,29,174,96,56,23,8,122,
30,177,112,112,213,144,3,219,227,208,241,241,
226,132,246,205,12,149,217,111,209,222,121,3,
18,65,28,133,140,149,119,145,185,16,81,24,
33,179,30,65,156,60,160,118,24,185,121,50,
230,195,217,135,46,37,45,71,227,168,55,141,
84,28,14,123,51,248,161,193,175,204,28,177,
144,51,81,43,5,222,31,166,248,171,157,50,
182,157,33,229,184,164,196,25,138,197,47,83,
154,127,129,222,209,85,156,226,235,15,205,76,
136,39,168,64,133,72,232,14,107,196,114,140,
34,35,134,65,68,78,83,200,84,11,19,53,
140,34,17,81,68,188,50,71,98,26,136,222,
8,44,99,66,98,22,178,136,222,8,97,59,
36,217,52,85,123,155,109,237,60,167,204,54,
70,74,34,163,136,110,152,199,137,82,100,226,
1,73,38,69,60,91,35,94,168,33,15,154,
19,128,248,72,65,52,247,50,145,97,97,121,
189,73,144,210,53,252,56,161,229,187,116,29,
147,249,197,52,233,48,192,62,26,226,249,49,
82,8,106,70,132,232,143,16,65,72,146,179,
72,42,5,146,82,142,96,183,69,65,14,217,
146,171,84,59,107,136,177,3,253,49,141,134,
75,191,233,160,236,55,88,78,43,56,185,37,
210,215,175,208,220,233,179,172,71,148,79,205,
176,99,63,8,172,253,113,153,161,93,68,85,
18,74,217,38,79,217,14,122,187,143,230,249,
40,99,135,102,165,72,35,41,241,116,127,147,
242,112,200,194,188,202,243,237,53,140,173,219,
148,180,30,211,169,30,34,142,137,167,170,196,
153,52,202,126,29,185,223,32,169,20,89,71,
112,171,89,96,118,89,103,52,83,97,87,215,
40,171,10,137,148,184,138,70,107,182,66,162,
171,19,147,69,199,69,76,149,216,141,61,190,
120,38,77,181,162,176,126,112,29,45,89,193,
82,12,114,123,107,216,211,147,246,71,107,247,
77,154,205,63,39,28,5,84,151,191,204,222,
235,63,162,113,168,208,200,141,168,165,117,116,
41,39,109,192,181,189,147,185,136,164,86,2,
67,3,69,65,118,250,8,215,35,92,158,225,
246,169,121,218,122,133,63,63,204,241,78,184,
202,141,192,96,165,182,205,76,41,61,1,168,
246,24,77,73,216,116,23,89,138,214,193,119,
145,73,130,231,198,204,46,152,108,134,243,60,
93,82,49,212,8,69,128,23,71,220,234,10,
118,221,46,49,9,53,215,135,116,138,59,106,
150,15,70,103,104,168,207,19,164,190,65,161,
246,121,6,205,31,50,93,149,236,173,43,124,
229,233,4,5,131,214,56,71,62,245,112,197,
151,82,37,83,75,1,134,46,72,235,15,202,
138,211,170,130,34,196,201,244,127,203,11,216,
25,185,56,81,204,110,167,198,108,169,201,197,
242,131,49,204,141,98,164,128,142,31,98,169,
10,59,99,15,231,70,14,181,191,2,249,54,
154,243,20,221,102,204,84,233,183,200,85,207,
127,98,2,29,135,62,237,254,247,81,42,147,
133,70,113,35,143,149,89,198,222,237,33,74,
125,18,79,33,222,156,66,86,70,200,84,130,
48,98,100,54,36,217,175,49,117,234,87,72,
226,228,19,129,34,24,182,153,75,185,12,250,
125,150,151,22,127,110,96,136,227,24,215,117,
233,118,187,196,113,204,119,223,184,129,172,76,
196,56,79,12,20,73,20,210,185,245,31,209,
140,2,138,145,39,14,125,146,36,70,170,42,
129,211,198,219,251,183,28,108,92,33,149,89,
122,136,96,190,255,53,247,187,46,143,91,189,
240,17,80,8,248,84,126,226,30,80,36,52,
175,253,37,103,178,93,220,81,68,70,15,31,
108,252,169,10,237,140,197,149,222,136,182,23,
208,15,66,68,144,161,229,164,201,103,90,164,
82,109,158,45,107,228,52,149,48,76,72,4,
28,186,30,82,135,216,108,161,142,167,152,63,
253,143,209,221,51,68,71,22,241,32,71,210,
179,8,90,75,100,82,159,167,215,122,7,175,
242,14,99,239,38,246,112,13,101,250,193,106,
33,165,72,132,56,113,142,192,241,50,39,21,
197,208,41,208,29,214,176,189,12,99,63,133,
23,164,40,152,54,237,113,150,211,213,58,151,
115,46,11,82,112,221,11,73,233,42,175,237,
166,152,41,109,147,73,15,80,74,30,113,110,
159,97,99,141,97,247,22,118,189,137,152,106,
35,212,207,110,86,38,69,140,169,59,12,237,
34,150,49,34,1,86,115,22,158,162,160,206,
213,136,23,166,224,88,73,145,84,139,196,243,
53,146,153,10,73,202,128,40,134,124,6,217,
25,66,20,227,165,242,164,7,117,244,200,69,
32,72,166,203,100,189,22,7,222,52,37,181,
71,82,72,67,214,2,43,133,210,234,145,228,
50,39,64,148,155,85,25,106,69,110,247,150,
153,170,216,8,1,65,28,227,38,46,7,253,
60,79,149,67,76,219,193,176,199,180,59,62,
105,75,37,151,87,9,134,147,76,81,145,2,
252,16,217,29,224,13,125,104,246,24,12,83,
212,90,119,16,131,49,107,107,67,28,39,162,
80,208,209,84,1,174,79,84,94,97,126,62,
166,191,222,196,238,218,148,102,242,108,222,55,
89,50,118,179,199,164,107,155,103,202,146,11,
249,52,86,189,131,28,59,136,48,130,4,26,
211,85,236,56,207,138,49,68,54,187,152,237,
30,245,187,29,14,15,93,20,69,224,56,17,
214,98,153,235,218,101,106,237,53,72,27,32,
36,178,222,70,157,205,177,49,204,97,154,125,
146,100,162,230,123,87,72,70,133,44,215,239,
212,216,14,2,206,167,67,148,56,33,94,93,
36,144,146,11,217,8,37,137,177,90,125,42,
167,179,220,184,177,70,174,244,60,65,28,17,
231,202,212,215,127,136,58,245,215,164,166,71,
216,135,6,186,94,65,79,103,145,83,91,188,
176,156,162,212,31,161,174,237,18,47,205,64,
33,67,108,26,160,107,36,179,21,72,18,180,
55,175,32,130,0,111,166,194,247,213,89,182,
7,83,248,82,229,139,229,38,95,10,214,217,
18,11,124,121,46,66,142,109,228,198,62,161,
23,163,41,9,65,162,145,140,218,52,91,30,
133,188,78,42,165,32,109,23,95,59,69,193,
105,98,166,18,148,157,35,204,156,228,90,55,
207,87,141,62,43,7,77,148,245,61,100,189,
205,190,145,167,217,74,19,138,58,227,250,30,
133,233,203,168,202,44,210,252,34,206,232,128,
211,139,46,150,238,211,177,51,20,173,71,207,
117,164,84,249,16,72,156,100,245,109,159,214,
78,138,100,104,17,108,88,4,187,22,189,64,
167,163,192,215,23,146,147,129,93,128,142,31,
240,218,118,143,45,199,101,199,113,217,24,185,
68,73,130,176,66,210,201,231,9,187,30,181,
229,95,33,147,59,71,42,63,243,9,157,150,
201,144,134,144,10,118,187,78,82,60,36,140,
52,250,245,33,83,139,87,120,241,82,72,59,
14,241,55,231,152,93,250,135,180,63,184,139,
44,122,136,227,68,55,137,37,57,235,57,72,
248,68,160,16,138,202,222,205,247,120,238,226,
89,106,181,234,207,13,20,63,121,235,61,110,
223,185,203,81,179,205,95,190,117,29,117,254,
185,147,25,175,239,254,187,127,249,233,28,69,
18,133,52,174,254,103,50,131,171,152,222,53,
252,11,255,59,221,235,223,65,87,7,12,109,
157,149,243,9,47,159,238,179,171,70,172,31,
212,31,88,35,248,113,189,244,147,90,143,203,
207,224,220,26,216,67,158,59,55,36,220,239,
83,111,68,152,179,10,170,166,144,20,50,136,
238,128,104,113,154,131,145,75,219,11,136,26,
105,164,155,35,112,117,148,234,8,77,245,17,
66,80,59,86,97,164,44,73,52,80,176,148,
137,97,157,76,71,184,188,69,239,112,129,194,
204,165,7,28,30,227,208,103,243,198,255,133,
118,254,112,226,130,144,13,184,127,105,251,253,
131,100,3,63,36,56,174,43,196,241,207,70,
111,14,93,245,152,46,237,96,123,25,52,197,
103,57,39,232,249,33,47,21,109,22,218,67,
228,200,230,234,169,121,92,63,224,90,111,76,
146,124,108,166,69,130,186,220,7,250,72,230,
137,15,203,112,234,231,219,193,173,169,30,65,
164,157,100,82,63,174,247,56,155,179,200,28,
115,34,241,76,249,126,246,123,162,12,139,33,
74,18,60,199,99,201,247,113,58,49,253,200,
99,86,239,158,224,116,120,102,1,53,74,56,
163,238,115,203,191,200,83,173,107,196,86,138,
120,170,68,180,56,67,60,91,153,72,105,143,
7,192,170,12,200,153,35,110,55,75,156,175,
181,169,154,58,65,146,112,183,41,216,24,57,
60,179,56,67,166,221,103,222,142,216,219,179,
177,111,132,168,170,100,105,57,141,62,180,81,
134,54,195,65,200,216,153,40,138,68,60,166,
109,123,52,26,30,206,241,146,171,110,215,39,
157,182,104,63,245,101,150,219,239,51,218,62,
100,249,43,167,105,188,183,195,232,253,77,190,
245,245,203,252,168,51,164,231,26,120,65,138,
90,174,201,217,92,154,154,162,144,218,60,160,
125,237,128,74,73,71,30,43,30,230,5,4,
137,138,104,116,65,8,186,13,155,131,3,135,
133,5,139,84,74,65,164,12,174,77,173,178,
233,182,185,104,40,188,85,200,49,51,91,163,
234,123,236,6,9,82,70,44,167,83,220,30,
140,233,7,17,113,2,71,142,71,162,77,147,
116,37,71,51,61,122,90,72,212,29,17,37,
9,47,168,18,55,72,216,10,114,92,253,139,
136,76,110,1,61,142,49,15,214,57,240,127,
192,51,23,70,212,82,6,210,15,25,189,122,
196,206,254,31,161,205,198,84,166,39,138,35,
229,198,6,225,115,231,240,162,152,171,67,135,
72,85,81,43,69,94,56,22,27,36,41,131,
224,204,34,255,105,175,198,133,180,199,175,45,
15,16,251,77,236,159,110,161,155,112,186,214,
102,107,148,112,42,245,160,12,179,162,182,104,
56,17,11,115,199,254,88,254,28,118,237,20,
103,211,27,33,151,85,169,0,0,32,0,73,
68,65,84,152,225,24,249,78,19,130,144,242,
216,33,76,102,168,27,6,51,134,78,244,244,
10,202,237,29,158,17,3,230,190,148,226,181,
215,76,10,83,207,162,104,6,190,215,197,107,
30,226,218,99,188,56,198,144,146,230,40,71,
37,61,36,107,126,54,129,204,158,91,35,30,
69,68,61,15,53,173,35,75,33,74,236,240,
234,114,128,161,220,83,98,181,199,1,111,188,
47,73,186,103,17,249,26,97,124,7,57,215,
68,90,144,52,43,40,134,130,158,45,62,214,
187,109,231,218,31,18,167,142,48,101,134,152,
52,129,214,65,69,80,111,204,178,120,97,135,
203,211,69,172,181,29,86,10,171,188,227,135,
168,154,197,202,179,255,156,221,237,127,11,43,
77,220,159,149,208,115,22,142,126,128,153,254,
100,63,57,169,234,152,51,231,120,253,173,119,
185,240,212,249,159,139,196,14,130,128,245,131,
46,67,99,154,184,223,198,90,253,210,147,13,
220,221,127,116,118,222,196,235,221,97,110,74,
80,54,35,174,188,251,127,240,43,249,59,72,
145,16,157,171,98,138,16,186,49,243,75,25,
182,213,63,103,216,76,145,173,62,236,143,18,
126,6,87,67,241,25,62,171,221,124,141,74,
210,99,119,215,35,155,98,50,69,30,199,147,
94,183,101,34,162,136,174,31,144,56,10,150,
247,60,83,43,223,4,18,118,238,124,7,138,
45,10,247,121,220,56,227,152,150,226,99,223,
55,81,140,18,35,228,71,118,4,30,82,42,
8,69,165,127,116,13,101,229,240,83,93,201,
14,29,159,156,166,50,171,107,28,56,30,154,
234,83,239,46,80,200,180,48,180,201,133,110,
25,163,227,184,106,112,65,215,152,187,177,14,
2,188,51,139,52,189,224,68,198,153,36,143,
1,208,185,61,162,125,3,133,159,255,16,15,
128,53,24,82,18,39,147,255,15,195,136,172,
250,224,171,23,117,149,119,218,3,58,126,128,
121,126,153,218,250,30,113,125,204,29,103,133,
176,88,130,153,34,73,55,5,249,25,102,79,
245,89,106,214,185,22,127,147,243,165,45,80,
213,9,72,28,3,196,3,228,186,8,184,211,
79,56,95,131,189,227,54,80,206,234,114,173,
153,194,84,60,206,190,244,52,53,253,14,221,
222,46,227,81,72,46,171,97,153,10,190,23,
179,177,49,34,138,19,116,77,226,56,17,102,
38,65,81,4,182,253,160,156,101,56,12,136,
14,6,184,245,3,182,54,199,148,122,27,84,
230,50,212,119,134,228,214,118,113,242,57,134,
118,129,197,82,131,47,84,139,16,132,164,223,
191,69,123,167,207,238,206,152,66,94,67,87,
4,40,18,245,160,133,98,77,51,58,115,150,
236,237,155,236,237,217,72,41,176,237,136,210,
139,203,180,102,171,188,177,158,225,27,167,219,
172,39,85,58,131,49,45,63,100,224,102,177,
140,193,164,245,231,7,244,252,240,120,126,68,
161,164,234,236,198,26,83,51,127,159,245,253,
58,222,184,131,157,121,147,175,174,8,228,110,
157,209,210,44,219,219,47,113,234,210,75,104,
102,150,48,10,233,205,174,16,58,91,76,229,
244,137,120,192,50,72,171,10,79,45,43,40,
107,187,136,77,15,49,118,136,75,121,252,36,
225,141,86,159,130,174,242,98,49,75,156,36,
36,9,8,67,195,185,184,202,107,123,69,150,
203,251,60,91,157,84,84,178,219,199,16,49,
174,155,240,108,116,151,31,119,158,231,78,44,
121,118,241,57,74,215,223,153,4,226,113,149,
174,145,166,51,74,8,92,143,121,235,128,133,
246,33,52,38,51,53,31,17,225,106,171,199,
63,152,223,226,135,135,75,216,229,50,47,79,
133,52,50,105,218,174,202,237,107,58,83,243,
127,23,123,188,193,198,219,191,15,36,156,123,
245,136,198,77,80,196,4,128,94,94,90,251,
204,215,246,102,187,198,139,53,155,194,178,67,
219,11,112,99,7,67,74,106,166,254,192,117,
159,0,7,253,128,124,78,16,23,246,8,131,
58,250,48,135,123,117,30,161,8,102,158,250,
59,12,219,215,176,131,91,236,222,24,179,240,
212,111,63,210,105,90,181,76,46,94,10,88,
160,75,146,113,137,73,24,250,121,126,120,103,
64,98,95,226,53,251,38,223,58,187,76,216,
44,115,233,197,13,118,239,190,193,212,242,215,
80,236,57,188,155,62,83,139,191,70,110,250,
28,66,72,252,81,239,241,226,179,233,51,20,
250,254,207,5,18,63,125,239,3,222,186,185,
143,177,112,25,67,213,225,17,195,137,79,4,
20,197,197,47,144,159,125,145,250,254,45,90,
237,255,194,203,165,61,82,241,132,208,137,14,
14,144,199,229,90,82,206,241,185,66,196,15,
182,111,61,18,40,38,101,221,227,215,197,72,
1,133,140,254,153,236,208,146,222,58,86,122,
76,54,175,83,203,65,152,8,148,108,106,34,
61,60,59,233,215,22,157,128,97,125,134,169,
51,223,188,247,60,233,35,142,75,204,97,24,
82,210,53,20,69,144,142,31,12,134,210,4,
175,217,132,228,60,59,183,127,31,169,170,84,
107,223,38,63,125,145,222,237,119,224,220,167,
175,31,92,74,155,44,166,77,78,249,38,141,
172,195,221,225,163,77,205,166,77,131,105,199,
65,36,9,206,11,23,216,12,66,6,199,38,
95,94,96,146,54,7,159,28,228,85,208,151,
60,254,118,71,66,78,83,241,226,24,5,129,
42,38,3,82,51,41,157,187,3,155,179,57,
11,47,138,41,27,26,253,32,36,175,169,24,
138,68,151,146,138,235,35,251,35,178,74,72,
54,53,34,137,119,25,4,75,100,125,143,68,
85,88,115,159,102,42,108,113,70,222,230,186,
247,28,139,94,139,140,225,62,226,29,76,0,
170,233,57,216,97,196,116,74,231,71,71,93,
76,61,166,63,46,177,51,238,99,74,201,210,
202,28,229,155,117,198,163,144,86,219,99,56,
14,16,137,192,245,38,193,200,117,35,76,67,
65,138,24,223,143,209,117,137,239,199,24,134,
130,174,75,162,236,28,243,238,26,59,135,14,
113,156,208,233,248,152,166,67,191,31,16,118,
199,164,74,5,226,68,82,208,39,159,81,6,
19,165,80,169,108,224,121,17,154,246,145,245,
74,196,221,215,119,56,243,156,207,110,225,75,
100,204,117,226,56,97,97,57,131,148,130,195,
215,55,208,87,134,156,155,181,120,125,183,76,
34,251,64,30,199,207,97,25,3,14,218,203,
120,65,138,43,117,27,117,127,10,213,114,144,
153,144,173,198,12,181,153,87,81,116,147,98,
113,146,81,238,220,217,165,206,1,153,156,69,
119,221,193,42,207,161,29,47,253,41,29,108,
97,251,46,254,236,111,113,109,243,15,184,120,
230,88,13,118,216,66,120,62,59,83,53,244,
192,99,186,209,38,201,90,52,221,128,146,174,
113,169,144,225,106,215,229,122,189,76,76,8,
24,184,161,32,159,57,32,47,99,222,104,246,
81,4,188,146,182,168,215,247,232,118,125,140,
173,49,207,77,183,40,204,102,57,100,158,70,
106,25,101,227,42,205,227,21,162,31,65,242,
22,208,176,84,44,75,33,138,18,76,115,242,
115,126,222,162,182,187,193,63,150,155,236,36,
43,252,65,227,34,137,58,96,53,23,82,155,
111,49,216,250,83,70,254,8,169,231,88,122,
106,204,249,124,154,167,94,121,242,171,120,191,
87,98,228,153,24,234,100,47,123,24,43,44,
20,91,164,52,31,184,215,65,56,73,26,194,
8,77,8,252,56,70,149,130,75,179,22,204,
222,187,30,175,246,154,19,201,252,214,101,172,
252,28,253,206,207,200,174,142,9,194,155,28,
222,249,62,211,171,223,120,96,238,193,27,53,
9,212,61,102,253,96,34,80,200,164,232,121,
33,8,208,226,50,243,23,126,139,237,187,255,
134,237,153,14,82,198,156,42,152,220,181,239,
208,63,156,97,225,242,111,144,196,49,227,246,
206,3,187,48,30,155,224,41,42,77,81,225,
195,171,215,89,156,159,165,88,44,62,225,156,
94,194,149,205,38,169,83,47,125,234,99,63,
21,40,164,20,236,93,253,30,145,191,195,175,
45,180,208,219,93,54,118,109,28,59,66,81,
4,243,243,22,187,251,54,229,230,45,174,205,
127,149,218,210,55,62,129,48,21,148,51,26,
237,209,163,87,80,90,134,194,82,62,75,182,
252,100,166,128,190,211,163,87,255,175,124,53,
191,129,54,118,177,44,149,48,241,209,145,196,
115,85,68,167,79,146,181,168,187,147,10,33,
73,252,147,158,97,125,227,53,146,153,253,19,
64,218,24,58,236,74,143,51,153,20,202,120,
66,122,69,247,145,41,129,223,2,33,80,173,
20,241,244,26,71,246,191,70,238,84,113,133,
196,233,207,80,201,63,126,254,68,17,130,219,
131,49,2,241,192,144,215,67,23,184,227,49,
87,239,96,175,46,240,198,96,204,253,27,101,
131,208,64,83,125,126,185,135,224,76,54,197,
148,169,243,163,122,151,15,122,67,194,56,161,
229,25,180,189,128,32,142,25,135,17,101,67,
99,103,236,82,49,52,12,41,73,171,10,162,
219,39,182,76,100,127,116,162,106,202,239,173,
29,183,173,42,44,157,107,177,27,45,179,18,
222,230,105,243,22,251,238,2,7,253,60,118,
236,112,166,210,197,144,18,85,8,156,104,178,
11,99,174,216,224,237,253,89,190,186,52,228,
197,114,150,150,23,48,176,199,52,134,121,76,
101,200,72,215,120,106,38,131,118,228,34,36,
120,238,195,21,107,169,164,227,18,33,4,152,
166,130,239,199,168,170,160,92,214,233,26,89,
162,225,214,73,162,19,199,9,251,251,54,154,
38,241,154,67,94,154,182,185,27,154,108,118,
10,36,244,120,65,145,36,249,12,66,177,153,
187,79,177,184,181,53,38,138,18,84,18,100,
224,227,54,71,132,97,130,169,9,204,175,92,
34,223,31,177,247,195,91,124,110,240,6,95,
158,43,114,117,246,52,219,94,204,249,74,157,
5,203,228,167,173,45,162,200,0,20,8,53,
230,230,254,23,66,207,102,241,124,238,196,102,
252,36,145,152,255,111,217,187,117,157,245,96,
11,223,62,160,84,141,80,3,159,213,183,127,
192,173,47,124,27,103,208,194,204,86,57,90,
171,177,60,223,102,28,69,28,142,116,198,227,
28,227,237,203,100,202,63,99,90,1,97,187,
92,247,230,56,114,99,174,55,160,106,164,89,
174,30,49,159,210,153,75,25,220,29,218,24,
138,202,230,40,160,162,169,8,63,192,41,231,
153,251,202,89,42,215,119,216,223,183,233,247,
124,138,5,151,89,214,184,158,125,145,218,11,
207,177,116,235,42,219,155,163,143,105,241,195,
135,170,185,56,78,88,90,74,67,156,176,216,
94,231,119,22,199,252,223,206,121,190,200,22,
178,54,77,84,77,120,191,147,240,92,41,62,
169,34,158,40,38,184,49,59,141,2,197,114,
192,92,161,243,201,252,102,156,240,102,171,127,
162,54,236,249,33,70,34,241,194,152,233,156,
206,229,66,150,173,177,195,145,227,19,39,201,
137,180,93,85,242,199,1,54,32,103,41,228,
52,133,45,249,6,59,55,27,44,61,253,63,
222,23,51,53,98,92,118,12,157,36,72,184,
115,117,68,148,243,152,206,234,196,197,58,71,
107,127,65,156,106,98,42,58,190,18,179,213,
156,103,245,105,151,237,189,63,69,237,255,15,
164,139,11,100,167,78,127,230,36,239,141,43,
235,220,94,223,226,31,254,198,175,63,209,51,
110,221,185,139,159,153,71,127,130,199,126,58,
153,45,4,249,217,11,228,103,95,98,183,117,
131,169,241,1,233,153,2,83,231,43,248,157,
49,59,59,99,52,203,228,96,241,18,234,220,
255,252,208,243,239,127,205,188,165,17,199,9,
78,16,63,64,106,11,33,56,55,157,65,6,
227,39,115,88,77,18,14,182,255,4,243,252,
30,103,43,25,252,253,62,66,72,210,38,36,
167,231,38,18,76,85,33,46,100,121,171,213,
199,241,19,148,254,83,100,75,231,176,123,187,
244,227,191,64,230,239,101,223,163,48,98,20,
70,108,219,14,206,40,203,88,121,144,107,136,
205,38,227,237,54,229,218,87,177,215,93,228,
66,29,81,24,163,151,71,104,106,64,111,84,
61,105,31,61,234,104,186,62,65,2,135,142,
119,178,152,231,81,199,217,156,69,126,100,179,
155,182,216,245,67,34,18,84,49,33,86,133,
72,176,189,44,41,221,254,165,193,196,216,205,
49,159,115,40,27,26,37,93,165,231,135,248,
113,204,32,136,136,146,132,67,199,159,88,170,
232,26,215,250,99,118,109,143,126,16,50,10,
35,236,124,150,100,182,74,102,104,159,72,106,
79,206,239,216,102,67,57,207,242,220,0,114,
105,48,117,178,166,75,57,61,198,73,28,190,
183,145,230,131,166,142,170,198,84,82,9,215,
250,35,164,140,233,142,203,204,229,198,188,217,
234,19,37,9,137,180,177,221,28,91,173,121,
20,227,136,213,169,2,83,170,79,105,202,194,
144,96,59,17,186,33,49,12,5,77,147,44,
46,167,65,40,104,150,69,228,216,140,199,33,
65,16,227,22,86,89,205,213,217,220,24,82,
173,154,232,186,196,182,63,154,115,73,136,99,
48,102,22,232,102,52,66,57,100,187,83,225,
134,109,48,189,146,34,173,107,200,206,0,146,
73,21,114,116,228,80,42,233,228,115,58,131,
161,193,104,123,23,69,17,212,106,38,106,163,
205,78,249,25,74,43,69,140,197,2,206,108,
13,87,9,25,198,14,75,150,73,144,196,212,
93,159,47,214,50,52,253,49,241,160,68,38,
115,14,69,209,17,82,125,200,61,86,209,76,
210,133,69,114,229,103,200,149,94,96,105,48,
68,13,125,14,206,61,75,34,229,137,192,195,
72,175,178,185,153,165,217,92,68,234,95,67,
207,126,145,76,121,133,104,244,46,203,195,61,
14,103,79,179,102,107,124,123,197,230,82,57,
230,92,41,68,17,48,107,25,8,49,113,10,
56,114,125,206,214,187,44,30,52,48,175,237,
145,201,234,36,211,101,212,98,154,188,8,104,
28,140,25,141,66,138,69,157,244,184,206,142,
121,158,153,75,57,114,90,136,26,60,8,14,
82,138,7,238,119,219,142,168,86,13,20,69,
144,20,178,56,213,18,165,180,199,29,167,198,
162,58,64,152,6,179,150,241,68,92,229,238,
150,199,107,127,30,240,214,15,134,28,238,133,
172,127,56,96,127,99,76,253,192,103,241,180,
254,200,253,99,82,8,70,97,196,145,235,159,
12,175,58,55,139,4,219,57,198,210,161,169,
56,28,56,62,94,28,227,116,84,226,163,44,
177,163,80,46,252,42,129,211,103,224,252,12,
39,211,167,235,134,120,119,179,136,244,24,213,
175,162,26,25,132,84,80,205,12,241,64,82,
63,82,25,116,158,162,88,248,187,36,157,105,
154,155,14,202,114,157,210,212,33,231,231,12,
106,166,78,49,229,80,201,12,49,122,3,230,
51,62,183,246,58,228,43,15,206,66,60,118,
50,251,163,64,158,202,17,249,14,138,219,225,
153,139,23,30,187,22,250,35,149,211,31,127,
247,47,81,166,206,61,224,12,254,168,227,137,
200,236,147,19,242,193,31,83,156,137,240,206,
158,167,120,212,198,222,108,226,197,18,229,252,
25,14,202,191,74,126,249,215,17,159,50,87,
161,8,193,82,197,98,174,152,162,49,244,232,
140,125,146,56,97,161,108,97,233,10,79,26,
2,59,187,239,17,105,77,236,142,32,82,135,
184,1,4,227,49,150,105,146,84,138,147,233,
230,140,197,214,200,193,30,67,178,177,196,210,
197,111,19,135,62,141,198,119,145,43,15,42,
147,50,170,138,4,132,20,4,99,129,98,60,
88,81,8,43,34,152,187,66,253,160,142,48,
82,147,157,58,226,30,9,252,241,62,191,252,
88,69,162,30,251,220,60,238,123,89,78,155,
204,37,9,114,191,129,95,204,159,8,0,74,
230,100,80,104,119,88,164,144,110,253,210,64,
194,15,76,10,41,151,211,217,20,113,50,145,
19,126,174,146,35,1,118,199,30,189,32,96,
20,68,236,217,30,251,182,71,152,36,204,28,
147,153,171,217,20,29,47,96,107,236,114,103,
110,138,203,51,21,10,2,196,208,70,238,214,
33,129,80,213,79,136,252,251,143,83,89,201,
236,121,155,126,16,242,253,109,147,145,175,160,
232,147,44,175,23,13,249,206,149,105,76,61,
203,190,72,38,250,126,25,178,88,91,99,183,
181,196,91,211,109,206,125,254,18,69,18,42,
255,31,121,111,250,36,201,121,223,249,125,242,
206,172,251,234,58,250,158,158,233,185,49,192,
224,32,64,130,0,69,130,160,68,29,212,177,
138,149,87,90,217,94,31,225,55,182,255,23,
135,67,246,134,29,235,136,181,34,100,249,144,
180,34,165,149,40,94,34,72,144,0,113,3,
51,152,171,239,238,234,174,234,186,207,204,202,
235,241,139,236,233,158,198,204,96,26,160,64,
175,185,249,170,163,163,58,59,235,201,231,249,
157,223,239,247,183,89,163,16,10,164,110,31,
98,38,34,102,18,22,50,100,111,172,179,222,
41,147,77,52,217,223,191,91,78,150,104,29,
140,230,219,220,28,49,55,23,99,122,218,162,
209,152,224,121,33,178,162,80,27,21,249,134,
182,198,43,153,20,203,153,38,111,86,167,152,
26,217,200,59,251,4,139,21,148,157,125,146,
201,200,200,40,138,68,224,122,72,94,31,1,
156,59,151,58,12,104,22,118,126,198,223,22,
191,204,153,236,14,174,235,99,40,50,231,45,
157,156,34,83,29,59,148,76,29,39,8,113,
67,129,94,169,178,189,247,63,144,148,190,196,
212,226,11,15,207,251,132,160,92,219,162,21,
58,172,237,254,20,185,27,146,47,124,25,69,
137,178,113,213,136,147,155,123,242,1,196,74,
141,222,147,23,248,206,122,138,111,44,247,137,
41,202,161,49,190,251,46,241,124,164,221,38,
149,247,182,216,89,237,178,21,70,25,64,178,
121,155,124,126,11,213,80,72,60,119,142,89,
199,167,181,59,98,103,103,76,50,151,32,174,
56,232,173,6,166,152,32,103,52,116,93,34,
22,83,15,215,119,119,55,234,199,217,118,180,
238,189,158,71,46,167,35,15,70,164,111,110,
96,46,84,120,203,47,241,194,94,147,32,117,
50,17,196,107,111,141,185,246,134,203,185,171,
6,234,57,147,229,139,38,205,186,207,107,63,
176,153,76,100,30,102,255,2,31,38,31,81,
55,208,47,30,101,32,3,15,252,154,129,108,
8,212,246,25,242,133,47,225,7,67,116,43,
207,250,7,255,19,250,197,168,103,16,108,21,
40,149,190,206,126,227,111,216,87,255,12,107,
253,115,148,206,124,5,17,6,36,115,23,177,
188,121,98,153,136,65,159,139,103,209,99,89,
218,27,1,237,169,29,158,241,6,232,213,6,
127,211,93,38,51,235,241,204,178,142,254,225,
62,83,139,10,157,234,219,100,103,158,252,228,
135,120,80,227,119,126,231,87,31,233,36,238,
58,10,97,166,79,124,235,19,243,40,210,229,
11,184,163,34,235,141,24,45,33,232,72,49,
106,149,223,68,90,250,207,137,151,30,127,168,
87,122,208,61,101,89,34,105,170,20,83,6,
165,180,137,169,41,39,230,113,216,221,27,48,
124,21,199,17,92,60,237,147,184,190,131,153,
214,176,190,122,21,49,95,6,69,198,214,117,
90,19,143,110,47,100,178,251,34,217,226,115,
232,177,20,59,183,254,47,164,211,107,247,205,
132,113,195,16,79,132,56,65,72,24,168,204,
196,36,84,85,98,176,171,225,174,38,80,203,
19,12,93,98,113,214,167,155,104,220,247,247,
67,59,77,220,24,130,20,193,97,207,165,226,
236,59,46,49,85,193,11,5,129,16,140,131,
135,55,243,231,226,38,139,9,11,195,15,144,
171,13,118,202,5,186,7,153,199,200,15,240,
66,193,200,73,17,183,250,159,153,163,104,13,
202,164,19,53,134,126,192,157,161,205,141,222,
136,190,31,176,16,55,201,25,26,211,150,193,
98,194,34,161,42,100,117,21,59,8,169,196,
12,46,164,226,152,18,20,110,108,48,107,104,
148,11,105,118,100,153,73,204,36,94,202,131,
162,16,150,11,88,25,137,213,246,12,217,216,
8,249,64,173,182,239,69,205,91,83,137,136,
117,91,125,133,181,118,129,174,163,50,180,51,
168,178,79,37,191,73,194,234,145,176,250,36,
172,62,73,107,4,82,64,50,214,101,171,147,
103,109,16,178,19,244,216,52,77,118,98,105,
86,99,167,152,148,19,88,197,52,106,204,68,
228,211,232,59,187,196,116,143,122,61,234,137,
20,204,1,19,171,204,184,213,38,12,161,223,
247,72,36,52,122,61,143,49,41,182,252,5,
158,140,221,198,28,15,153,157,45,49,211,234,
160,38,100,250,118,130,169,253,77,228,129,141,
72,196,232,215,135,116,58,46,195,161,79,167,
227,162,140,58,228,150,242,216,231,230,48,82,
49,68,58,1,49,147,93,210,140,229,1,9,
77,102,201,113,201,214,91,104,190,79,102,42,
131,161,42,180,93,143,103,108,135,139,131,14,
203,51,58,187,94,29,137,11,72,146,122,44,
163,128,136,129,157,223,93,167,95,152,102,146,
46,32,124,129,228,239,50,108,109,18,203,156,
69,187,7,158,30,6,30,34,12,177,251,187,
236,109,253,59,150,22,251,12,213,16,199,141,
35,105,61,100,73,34,118,15,64,97,223,113,
73,108,214,184,254,205,235,116,26,54,97,120,
164,238,234,186,33,169,148,202,250,218,16,246,
59,100,191,120,14,122,35,90,245,49,253,161,
32,185,88,68,82,100,140,65,7,195,80,72,
36,52,12,67,161,84,50,49,77,133,120,92,
69,85,101,70,35,31,33,162,53,223,219,115,
200,101,52,84,25,20,63,224,109,109,134,83,
75,26,250,193,51,249,110,244,127,85,237,232,
192,173,14,109,114,122,132,206,43,20,53,46,
60,105,50,51,175,81,154,214,112,198,1,223,
255,230,128,202,101,149,207,61,99,161,155,199,
109,210,245,119,29,126,240,205,49,203,151,52,
20,45,210,4,123,168,51,182,2,188,107,69,
210,83,143,161,91,121,226,249,121,100,69,35,
157,123,156,254,70,21,41,215,69,235,158,165,
48,247,2,217,220,51,72,163,60,153,242,21,
100,69,101,212,218,161,190,242,93,188,81,159,
84,233,136,13,175,91,89,76,109,158,118,243,
3,236,162,160,18,4,40,49,133,59,55,166,
217,237,250,204,94,140,209,26,193,254,246,152,
76,249,137,79,148,81,0,4,70,134,149,183,
126,200,213,199,31,61,143,226,213,159,189,195,
182,155,68,179,18,255,180,25,133,36,43,164,
42,231,73,113,30,207,126,145,160,40,147,53,
98,252,34,175,192,29,227,183,191,199,139,211,
219,12,84,149,212,110,143,198,40,64,76,28,
226,59,117,130,83,51,76,130,144,238,196,143,
234,235,107,115,44,44,127,137,113,123,143,253,
141,127,36,156,189,133,36,193,148,169,51,242,
3,198,247,148,129,238,34,119,61,179,199,140,
150,165,46,57,180,138,54,106,113,130,240,32,
145,144,153,4,81,41,200,255,8,25,36,102,
14,104,244,43,44,21,234,12,188,128,149,3,
201,102,39,120,120,153,169,98,25,76,91,58,
123,182,203,214,200,161,57,241,184,148,142,51,
93,204,210,126,144,150,147,36,62,211,181,205,
167,106,236,52,79,67,97,13,136,178,133,75,
233,248,3,159,91,110,116,88,52,52,26,7,
28,17,36,153,224,236,60,234,107,31,16,191,
189,201,197,132,133,72,196,16,153,36,225,124,
52,217,207,66,80,84,55,248,230,157,10,139,
153,17,219,253,72,174,68,146,36,206,229,251,
204,165,192,241,44,78,149,110,222,87,38,8,
133,32,103,104,88,138,204,92,204,100,117,104,
51,99,25,172,105,13,110,213,103,112,124,11,
69,10,80,100,31,203,232,112,107,163,68,33,
181,207,139,21,141,130,166,146,152,146,9,91,
114,20,249,7,2,73,146,80,251,59,135,145,
255,230,230,136,189,189,40,226,221,103,142,69,
239,58,205,154,65,124,33,142,186,177,139,52,
28,243,132,57,224,90,233,249,131,34,183,79,
235,78,131,205,205,168,68,169,235,50,229,178,
133,16,2,75,21,236,100,146,124,167,33,24,
185,57,226,154,132,37,203,148,53,133,161,31,
160,53,218,200,251,109,194,48,68,171,181,40,
201,50,101,199,133,184,201,164,152,163,109,154,
60,117,81,227,199,175,189,67,126,230,185,251,
214,127,249,141,239,179,242,228,151,8,52,29,
5,200,47,60,3,60,131,16,130,73,191,121,
12,214,190,246,225,159,160,196,5,185,156,195,
87,190,96,16,83,116,190,189,97,145,75,214,
81,37,137,188,17,25,220,247,187,67,122,174,
207,130,169,83,123,125,29,223,143,246,154,170,
74,148,74,22,213,234,248,176,44,183,176,16,
103,107,107,196,240,47,223,101,230,124,30,69,
145,208,227,26,165,100,151,91,163,51,168,54,
228,212,46,5,173,125,216,143,104,54,39,56,
78,136,127,79,211,237,238,17,234,245,188,72,
8,113,48,98,62,27,178,229,120,92,60,120,
174,31,126,123,72,99,79,65,150,92,126,239,
95,165,113,165,128,15,247,198,20,22,53,210,
154,138,172,194,250,208,166,49,244,112,106,6,
86,202,227,204,215,13,214,238,184,136,7,152,
166,102,67,16,47,150,145,229,33,99,47,122,
22,191,165,161,230,143,207,143,15,39,18,222,
141,28,114,169,131,170,29,135,164,171,102,156,
84,234,10,126,117,145,252,252,179,56,131,58,
237,250,107,196,99,203,135,13,237,120,97,1,
100,65,114,234,254,62,131,145,44,96,213,47,
177,85,253,25,163,114,154,23,124,151,234,249,
61,236,238,231,249,222,91,91,200,70,149,202,
210,111,127,170,51,172,199,211,96,36,89,219,
216,100,105,113,225,99,63,187,82,235,163,232,
197,19,223,251,211,107,61,73,242,47,204,65,
132,129,79,243,195,255,157,133,204,14,179,97,
23,247,70,13,3,193,192,14,233,116,4,153,
25,21,113,48,157,43,4,222,172,246,241,55,
138,204,156,250,198,17,18,33,220,68,178,34,
195,109,127,76,159,64,120,50,205,192,163,234,
77,14,65,90,189,127,48,16,95,20,200,105,
135,153,152,129,34,73,108,221,51,162,51,97,
246,81,101,159,219,181,57,202,185,77,236,3,
67,159,80,21,250,15,81,32,221,179,39,244,
61,159,83,137,72,204,174,104,234,132,66,128,
162,224,221,3,37,46,154,58,251,142,139,170,
120,145,90,169,213,251,76,214,88,145,125,102,
242,235,212,59,51,148,178,59,204,197,12,76,
69,142,122,3,130,99,100,164,187,34,110,229,
98,14,194,144,224,212,12,76,92,68,204,140,
164,171,7,99,164,193,152,80,145,161,82,56,
132,143,72,72,232,137,91,212,220,4,197,172,
195,69,67,103,218,210,249,238,102,130,181,182,
78,76,63,14,3,204,233,26,115,113,131,152,
162,144,51,180,195,68,174,112,96,72,74,166,
142,169,212,184,86,43,80,200,212,200,235,26,
109,215,99,58,191,193,126,119,134,239,108,141,
120,174,226,113,234,137,179,200,111,220,64,81,
186,4,129,56,116,10,158,23,210,110,31,143,
42,83,180,144,101,137,233,105,11,212,200,128,
133,249,12,114,173,137,209,223,57,52,114,253,
190,135,44,75,84,42,22,186,46,147,126,250,
20,82,127,196,234,116,153,55,62,212,152,153,
78,51,91,108,98,202,50,67,63,64,114,125,
78,59,19,36,215,167,81,119,208,218,53,50,
25,29,82,113,6,229,18,111,54,75,56,235,
23,81,212,4,174,87,39,95,185,191,244,144,
175,174,113,231,233,47,19,200,202,125,200,192,
143,150,27,234,91,111,208,41,24,92,41,11,
158,41,31,157,213,129,43,241,100,82,103,234,
96,13,199,65,192,214,200,225,75,165,44,153,
91,27,220,25,30,25,205,133,133,56,107,107,
145,132,77,62,111,224,56,33,139,139,22,205,
230,132,225,192,227,214,27,145,144,162,37,251,
132,55,183,152,119,55,144,145,216,110,38,176,
115,211,204,25,17,42,82,211,101,100,69,98,
191,238,17,124,68,27,165,209,152,80,44,154,
72,18,100,173,14,183,155,41,206,165,60,20,
73,226,165,223,74,208,106,248,36,147,6,170,
10,8,25,163,23,231,251,223,137,241,220,151,
7,180,109,159,213,15,179,196,149,199,200,148,
31,103,229,245,127,77,242,249,1,165,203,202,
177,76,233,238,117,230,172,66,171,213,66,55,
45,74,138,78,107,224,177,115,187,72,123,46,
32,94,24,98,104,14,116,34,35,82,154,255,
58,134,85,192,72,230,113,122,199,185,73,185,
153,167,15,225,242,181,218,255,3,146,68,185,
248,117,134,141,117,28,123,135,169,133,47,49,
232,93,163,63,248,25,178,148,64,37,131,162,
197,201,205,94,5,36,42,167,127,157,198,157,
12,221,250,26,255,80,216,229,169,217,128,91,
198,107,120,253,175,209,220,238,34,185,239,80,
74,126,229,83,157,227,78,96,240,215,175,222,
228,191,204,101,73,165,30,60,75,232,250,173,
21,108,171,28,145,101,63,115,71,241,11,186,
252,137,205,96,237,127,229,243,11,13,244,219,
43,108,109,140,168,59,1,149,178,69,38,171,
33,66,31,77,147,144,106,45,194,74,1,69,
2,51,41,35,226,103,48,147,71,152,96,25,
139,224,158,230,245,67,81,94,227,44,155,70,
135,201,129,177,246,170,49,22,159,253,125,218,
245,31,66,122,147,234,120,114,223,240,21,0,
83,31,83,201,111,82,109,158,162,152,217,65,
83,93,6,94,64,119,88,96,226,69,80,197,
82,118,155,238,176,128,44,133,164,226,109,70,
126,192,181,238,240,48,196,138,29,96,205,45,
69,57,116,54,119,123,29,233,120,139,70,119,
26,89,14,137,25,131,207,6,243,116,192,208,
118,220,216,129,18,235,232,80,152,239,124,42,
78,136,96,46,102,18,206,151,17,83,89,164,
78,31,185,218,64,123,237,3,132,34,35,217,
147,99,32,8,17,51,35,241,185,225,24,169,
213,37,177,56,195,148,169,113,41,173,18,87,
143,234,163,95,158,31,242,106,163,143,47,162,
239,28,87,21,150,147,49,114,134,74,236,96,
102,179,56,40,195,73,18,135,191,211,100,137,
32,80,209,20,143,23,139,89,246,109,151,190,
23,48,151,48,40,24,251,220,220,47,146,55,
70,72,99,7,197,182,153,158,182,216,216,56,
2,42,124,212,73,44,46,198,89,100,64,175,
167,209,237,186,20,42,185,187,227,29,16,137,
24,115,253,53,246,253,2,126,115,155,78,199,
229,242,227,89,184,176,128,42,193,15,93,139,
199,206,22,184,246,211,105,44,89,160,152,235,
92,235,186,156,138,91,156,241,125,98,215,86,
32,8,25,219,62,59,213,113,228,92,166,44,
110,77,157,98,115,239,73,242,139,95,37,253,
145,224,235,94,212,147,234,185,152,195,30,161,
162,62,18,62,238,59,67,108,233,53,74,83,
109,54,218,151,184,60,181,71,199,245,216,233,
27,72,178,67,74,85,24,249,1,123,246,132,
188,161,69,128,9,63,68,26,29,41,39,228,
242,6,137,39,151,48,119,175,51,53,101,146,
207,235,135,70,62,149,210,24,141,238,145,104,
183,125,182,182,60,242,5,131,84,74,161,18,
244,241,66,153,27,227,101,206,89,171,236,85,
109,130,64,48,61,99,161,153,6,59,155,189,
195,94,197,100,18,17,39,231,78,167,89,242,
218,180,133,201,183,214,44,126,109,113,140,53,
118,40,58,125,104,57,248,231,23,81,37,137,
39,46,201,252,172,147,227,157,183,114,200,146,
73,62,115,137,86,247,175,24,239,190,69,242,
133,1,95,42,101,136,169,202,3,215,104,102,
65,103,230,32,208,110,78,92,6,178,143,122,
97,143,152,22,163,61,44,226,143,20,230,218,
18,90,204,64,228,66,140,100,254,99,215,121,
220,169,162,137,105,242,229,207,35,171,58,131,
222,29,4,99,154,91,175,147,91,254,128,199,
243,9,170,246,38,186,44,211,29,250,108,221,
88,99,246,220,63,67,146,100,10,167,159,195,
29,158,67,179,18,188,183,246,58,102,234,117,
198,238,53,22,46,254,17,170,153,248,212,103,
88,78,20,209,172,36,219,213,61,46,61,196,
81,212,26,45,36,165,0,129,243,203,227,40,
134,187,255,158,207,93,24,224,191,122,135,219,
27,3,130,32,154,4,150,76,170,12,135,62,
149,105,139,189,218,136,202,124,9,121,175,73,
63,155,98,18,132,24,178,254,145,134,113,146,
71,78,100,29,199,9,124,155,113,205,64,41,
218,228,117,141,14,18,146,34,161,233,121,98,
218,14,105,77,61,36,129,61,8,162,54,83,
88,99,167,121,154,124,170,70,119,88,32,147,
104,146,73,52,9,67,133,122,103,142,108,162,
65,32,20,38,7,3,93,100,41,210,121,218,
24,57,76,41,50,146,231,227,249,58,16,69,
118,173,201,81,132,55,149,217,165,217,175,32,
4,15,20,22,252,68,96,58,33,211,232,85,
162,161,61,102,15,203,24,29,58,164,122,103,
150,29,253,136,31,226,9,193,59,157,193,97,
137,110,33,110,34,44,3,97,77,69,131,112,
250,35,164,225,56,26,139,121,215,89,8,129,
114,123,11,17,51,145,155,93,68,38,137,38,
75,124,46,31,109,94,105,236,32,223,217,2,
203,36,219,29,240,242,185,5,110,34,81,183,
93,76,89,102,198,212,142,213,78,63,236,141,
88,31,70,153,192,179,249,20,113,85,227,221,
122,140,91,237,52,95,94,170,147,84,77,118,
194,144,231,10,41,98,170,194,43,245,46,65,
168,146,212,84,164,253,54,193,185,69,98,163,
149,136,79,17,106,72,193,228,62,57,153,189,
61,155,249,249,56,243,243,113,92,55,196,219,
239,161,42,18,178,42,19,156,154,198,26,110,
17,184,67,110,237,217,88,150,194,228,137,51,
188,49,86,177,107,41,70,123,62,164,70,232,
70,22,75,191,78,115,226,178,164,107,60,182,
94,5,103,2,65,136,176,12,122,187,54,169,
164,70,50,103,242,86,225,44,237,253,231,41,
156,122,225,227,75,10,206,152,226,198,77,118,
206,159,172,193,217,217,123,7,169,216,62,40,
117,220,228,207,111,149,8,195,24,170,226,147,
79,213,184,209,55,232,123,1,49,69,102,109,
232,144,209,85,18,68,14,61,12,35,130,98,
241,66,9,111,166,204,185,75,85,182,55,134,
236,237,218,204,206,199,48,13,153,210,83,115,
140,126,180,129,59,9,112,156,0,207,11,233,
118,61,178,89,157,193,192,39,22,83,48,130,
1,121,227,14,183,199,167,241,139,101,68,245,
67,54,54,70,132,185,10,75,103,211,212,111,
111,29,58,139,253,125,7,33,4,243,174,199,
75,198,62,175,205,63,205,191,121,183,194,191,
200,127,72,113,107,143,189,51,179,140,71,14,
211,150,193,91,111,24,148,151,94,70,79,228,
104,172,255,132,118,251,21,190,250,5,133,152,
10,144,63,241,254,95,76,88,216,65,200,202,
80,66,181,96,174,176,10,121,48,164,23,41,
45,189,124,162,123,36,139,75,36,139,75,71,
149,133,212,105,244,88,14,213,176,168,238,252,
144,203,185,144,184,40,97,16,48,91,28,129,
116,155,181,155,255,11,165,242,55,48,226,83,
7,77,98,139,169,197,95,97,208,156,70,247,
247,88,127,231,223,114,230,217,255,230,231,171,
192,216,29,76,253,225,92,138,11,75,115,124,
248,214,30,40,230,47,143,163,48,115,207,80,
223,191,78,62,25,67,150,6,248,128,235,133,
244,122,30,211,115,113,92,15,172,98,28,169,
213,99,18,51,145,133,64,120,160,27,199,71,
149,202,124,252,212,39,73,40,72,88,200,141,
121,114,115,207,176,191,243,151,140,23,219,88,
11,46,171,55,126,196,180,126,142,126,47,196,
202,9,236,15,12,172,139,19,30,70,131,158,
45,172,178,182,119,145,165,202,135,247,52,240,
3,74,217,109,84,73,58,208,48,18,135,136,
39,132,78,86,202,144,163,15,182,67,127,92,
68,50,215,31,120,239,66,106,143,70,111,26,
85,241,81,100,159,238,168,0,2,52,213,69,
150,2,130,80,69,150,3,82,177,3,25,13,
33,49,116,82,140,38,73,184,135,217,45,132,
68,57,183,245,144,50,84,128,46,203,8,4,
143,101,18,100,117,141,31,212,59,8,33,136,
61,64,209,87,164,226,136,84,212,207,80,110,
110,220,133,131,17,92,56,133,200,164,144,54,
107,4,51,69,164,70,39,82,42,5,36,63,
64,84,10,96,79,8,23,167,209,235,237,104,
254,69,167,143,208,84,2,117,150,65,50,142,
121,224,44,196,129,85,239,14,202,188,239,197,
73,104,48,149,108,113,126,74,144,211,205,195,
146,84,234,46,211,62,140,145,48,134,128,68,
88,200,130,161,97,156,179,73,239,59,180,212,
108,154,11,0,0,32,0,73,68,65,84,204,
115,92,202,110,69,50,24,43,125,12,67,102,
56,244,153,76,66,86,86,162,137,120,243,243,
113,110,124,216,65,146,36,206,158,79,161,109,
236,17,4,130,141,141,33,243,243,113,18,25,
131,239,111,230,49,180,47,81,156,89,166,111,
252,9,175,85,77,52,50,164,226,19,6,72,
92,104,247,162,108,36,21,39,56,51,199,36,
25,167,148,222,102,226,195,223,239,158,167,192,
55,200,207,207,60,242,12,36,218,251,84,207,
93,61,121,16,224,73,136,145,12,70,136,44,
7,135,147,251,100,73,98,62,110,161,72,80,
62,40,231,41,99,135,171,217,100,196,228,38,
226,155,84,78,165,121,71,156,195,121,197,227,
165,132,201,204,227,73,98,59,125,134,29,155,
193,208,99,250,55,22,57,213,232,210,173,141,
14,251,52,0,221,174,139,21,83,137,199,212,
8,225,101,200,156,141,173,50,54,45,26,210,
52,205,205,29,228,246,10,183,244,39,200,43,
18,75,75,137,200,121,132,130,241,56,64,196,
45,94,41,62,75,43,24,243,242,217,117,222,
110,47,160,205,207,227,140,2,250,3,5,117,
103,66,169,242,27,232,137,28,253,250,10,67,
245,21,202,103,92,98,106,250,19,219,21,9,
24,251,1,242,172,123,52,86,77,2,87,84,
63,181,173,138,101,231,144,21,149,126,227,38,
185,66,200,254,196,229,213,234,132,151,230,29,
132,144,184,84,136,35,61,214,98,237,189,191,
103,102,233,95,28,119,58,133,179,36,11,103,
73,20,206,252,92,25,5,128,150,42,243,55,
111,175,112,110,103,159,95,251,210,115,247,161,
157,254,230,135,111,160,207,61,203,164,223,250,
229,113,20,118,103,19,67,113,209,2,15,207,
23,40,146,196,242,114,130,216,233,34,227,197,
105,228,173,38,98,59,34,196,253,80,213,112,
59,67,34,110,157,248,200,2,185,119,249,118,
15,78,217,236,10,190,190,131,34,205,17,75,
207,146,117,190,194,234,250,143,32,37,200,45,
111,208,88,233,33,203,9,246,157,1,214,99,
31,31,169,55,251,229,195,65,56,247,149,5,
132,64,18,160,73,50,158,8,241,66,193,110,
183,192,239,157,29,162,41,22,152,6,65,32,
125,236,139,153,74,239,210,234,85,112,60,139,
153,194,26,32,49,57,144,10,183,148,17,161,
144,217,239,206,18,134,50,189,97,158,217,226,
42,165,204,206,201,13,13,18,166,34,243,66,
46,137,36,75,32,203,60,147,75,82,29,59,
116,92,63,50,48,15,128,224,137,116,2,17,
183,34,165,86,207,71,222,216,37,44,100,32,
8,144,108,23,101,99,23,105,48,70,148,114,
72,173,30,193,133,83,132,11,21,122,94,64,
98,42,131,130,4,107,59,4,75,179,8,224,
149,109,135,218,32,135,16,2,65,140,148,38,
243,210,252,144,74,236,160,79,19,42,17,157,
255,224,42,181,186,72,189,17,238,133,69,26,
253,28,203,197,26,18,73,80,21,148,235,171,
200,251,109,230,79,37,40,251,53,182,75,207,
178,112,218,225,172,250,46,97,32,168,30,200,
110,248,126,200,236,108,236,144,76,10,176,185,
62,68,136,168,167,81,169,88,228,22,179,52,
167,202,104,173,207,147,173,60,142,221,221,165,
213,84,152,42,125,13,53,157,100,195,11,121,
58,21,67,105,117,232,95,62,77,124,48,230,
166,80,169,238,15,185,50,91,225,103,175,107,
148,206,252,243,71,150,54,0,206,124,240,83,
234,151,158,69,124,2,253,179,226,217,231,105,
108,10,70,27,215,160,84,67,182,162,146,222,
172,101,240,152,34,35,226,38,72,17,249,243,
98,58,142,188,83,167,181,49,162,209,206,144,
42,91,188,105,60,141,198,132,180,62,32,184,
120,138,239,189,158,36,54,147,33,83,217,224,
212,244,132,214,200,33,216,28,80,171,29,215,
90,234,116,92,146,9,13,89,134,68,252,104,
7,199,100,155,249,60,52,183,162,192,104,106,
178,194,48,54,79,54,219,101,48,240,80,213,
8,20,177,153,189,66,57,59,225,212,208,37,
47,233,156,55,54,163,97,75,178,140,183,60,
207,157,69,133,155,123,223,167,221,147,112,228,
22,221,120,142,76,176,251,169,109,203,19,185,
36,63,222,239,30,149,162,5,104,84,126,110,
155,229,216,187,184,98,66,175,59,33,36,206,
141,126,159,107,253,128,83,9,139,75,153,56,
153,167,27,188,249,222,255,140,98,151,152,79,
255,225,71,156,205,12,66,132,39,102,101,63,
208,150,169,26,214,244,5,54,186,219,252,31,
127,251,10,249,132,198,215,190,20,81,219,183,
119,170,184,233,37,244,79,120,127,53,130,161,
217,135,205,153,187,63,127,60,250,200,65,40,
42,130,71,35,113,78,122,207,135,125,86,4,
125,114,139,38,154,150,37,177,214,99,56,242,
49,74,25,222,155,41,209,26,187,100,43,57,
150,37,29,67,114,184,148,176,120,183,63,70,
142,67,119,244,109,38,235,13,10,51,95,36,
240,38,100,203,207,177,119,103,27,121,121,43,
26,56,228,128,104,166,81,252,60,184,6,142,
223,64,61,7,110,250,38,131,198,26,178,176,
200,200,125,212,116,84,39,78,46,183,216,107,
205,163,141,60,244,252,253,181,189,80,200,52,
186,51,164,205,144,169,212,46,60,0,165,116,
23,49,37,0,239,160,30,47,35,17,10,137,
55,90,125,190,28,51,80,195,144,152,57,166,
239,198,48,63,134,92,151,79,239,177,223,153,
61,52,237,119,117,163,14,141,197,129,99,176,
140,17,182,27,255,88,249,143,251,158,83,246,
217,31,233,120,195,26,49,219,33,52,116,166,
230,203,84,214,118,144,198,19,194,92,42,154,
24,248,17,227,37,226,22,254,83,23,80,223,
186,129,100,79,34,117,210,131,223,75,206,4,
238,150,236,70,14,97,54,197,158,89,225,198,
173,24,154,31,48,140,59,60,81,30,144,92,
172,112,167,59,228,78,163,194,110,111,154,63,
126,124,139,180,166,48,9,197,193,212,50,9,
201,113,81,110,172,131,237,224,127,254,10,72,
18,82,163,131,114,123,147,176,152,231,118,127,
130,237,165,121,44,147,184,27,37,68,207,61,
61,133,48,13,244,213,29,250,70,138,137,238,
163,127,238,2,122,183,143,82,91,97,225,229,
243,224,122,76,74,57,60,89,38,217,232,34,
239,238,147,72,104,212,106,54,203,203,73,194,
217,18,173,233,2,63,121,55,77,121,225,50,
190,107,163,197,178,148,43,191,143,145,42,208,
92,255,11,190,122,58,32,181,94,165,93,202,
243,234,7,67,166,76,147,78,243,41,38,126,
149,159,170,93,50,249,95,67,49,98,143,60,
23,83,107,215,185,115,230,10,178,166,193,73,
206,165,55,161,187,251,1,195,206,109,202,203,
191,142,84,211,25,137,191,39,232,66,42,15,
103,109,27,245,195,42,222,231,46,131,166,70,
217,151,16,140,81,249,219,247,115,124,109,185,
198,78,250,2,10,69,94,152,125,11,121,183,
129,159,184,64,46,61,198,182,126,147,190,250,
171,252,96,237,85,208,118,120,233,197,139,228,
246,155,108,189,89,101,228,169,88,9,11,119,
226,71,28,166,16,118,170,99,146,73,141,108,
78,63,216,161,2,73,68,252,162,84,34,100,
44,197,129,46,158,39,168,84,12,52,77,230,
150,149,102,112,171,202,251,55,38,252,254,217,
163,89,40,132,1,218,141,117,42,79,156,163,
157,222,33,107,104,172,12,108,152,120,108,117,
115,220,180,34,212,86,76,85,142,222,249,73,
202,218,94,112,140,0,27,116,85,146,153,139,
247,189,151,79,106,23,11,179,47,178,93,255,
0,105,102,136,161,143,217,233,102,201,38,247,
185,217,27,161,203,50,55,250,35,148,133,16,
113,51,123,223,125,59,213,87,32,80,201,206,
127,225,216,125,239,174,225,39,178,161,177,2,
29,10,52,251,109,238,252,233,95,242,220,99,
103,120,235,250,10,242,244,83,248,174,77,224,
77,8,92,135,80,81,78,230,40,66,55,58,
192,161,239,30,254,252,177,15,227,79,16,97,
112,162,151,113,210,123,62,236,179,178,118,138,
183,174,191,199,211,242,144,225,216,167,80,48,
144,82,22,77,55,98,4,15,122,42,157,172,
197,98,50,67,76,145,15,161,171,74,113,136,
93,125,27,127,252,68,36,255,43,160,60,253,
219,52,238,124,31,73,129,184,177,68,34,119,
230,176,14,62,216,95,165,118,235,39,140,167,
117,252,230,95,163,36,7,104,149,227,176,185,
82,124,139,193,90,129,45,251,20,167,166,239,
160,200,81,67,111,104,167,233,143,179,124,126,
190,198,56,8,105,56,247,191,80,77,150,184,
152,142,243,94,103,136,68,52,254,116,224,5,
84,44,157,222,74,15,39,27,50,233,143,208,
99,38,137,88,139,157,122,241,176,73,254,208,
245,18,143,142,10,210,241,22,3,59,67,111,
148,39,29,63,89,170,153,77,238,179,219,90,
164,95,116,136,111,236,162,0,114,181,1,113,
11,2,31,121,191,141,220,31,17,156,158,33,
44,125,36,42,86,21,194,82,14,101,173,26,
149,161,28,55,26,198,179,83,71,210,20,8,
2,208,85,250,167,206,51,220,215,120,185,241,
125,130,92,138,110,205,224,123,195,115,136,68,
13,99,56,203,75,203,125,254,110,61,32,171,
171,72,99,39,154,251,28,4,200,173,30,242,
78,157,240,212,12,34,97,129,36,161,172,85,
145,182,246,192,50,233,205,151,121,123,171,66,
220,240,143,198,109,106,42,193,153,89,228,90,
11,101,99,143,176,82,64,77,9,174,247,2,
2,224,153,233,34,115,207,7,140,103,138,172,
14,109,118,134,14,50,18,24,6,23,31,91,
102,97,109,135,105,25,198,23,151,216,141,89,
188,247,102,142,108,230,197,99,123,85,248,30,
237,245,255,155,151,138,239,227,5,26,223,162,
66,240,254,20,42,51,216,249,121,114,149,89,
60,103,25,207,237,131,224,68,103,66,182,135,
120,194,71,59,233,249,241,38,136,192,199,243,
218,120,195,30,4,33,126,71,35,172,165,40,
38,6,36,234,235,17,210,69,85,241,133,96,
101,96,115,62,21,227,77,77,33,118,69,197,
60,59,207,112,227,113,240,182,232,25,25,106,
169,52,181,215,98,72,198,115,152,70,84,243,
158,154,254,50,129,55,225,167,187,215,145,104,
16,127,162,138,38,59,108,118,20,158,153,235,
97,172,174,49,118,124,2,33,113,123,211,227,
233,180,134,44,75,4,174,64,51,36,220,137,
160,215,245,200,206,195,181,225,57,46,159,142,
244,161,60,61,134,26,78,120,74,219,226,169,
115,46,210,71,56,14,34,159,70,203,36,120,
246,224,188,46,37,44,220,32,228,47,110,155,
172,12,34,84,82,42,20,84,199,19,202,150,
254,192,140,247,163,215,192,247,143,155,222,110,
26,173,146,188,239,221,132,222,228,19,217,69,
5,240,198,89,186,189,36,2,233,208,78,56,
158,201,119,55,146,148,204,44,146,235,163,136,
194,177,251,134,129,71,70,125,135,81,80,36,
116,159,58,102,23,57,97,70,249,32,27,42,
233,113,196,212,37,94,217,24,33,165,78,163,
222,181,247,222,132,192,115,144,3,245,100,142,
66,79,68,163,5,125,215,62,252,249,99,235,
123,7,228,15,69,127,116,51,228,164,247,124,
216,103,237,246,27,76,79,203,244,175,57,88,
150,194,252,92,12,79,87,233,238,72,40,69,
80,197,20,131,96,155,157,209,145,8,169,240,
96,242,126,10,61,111,161,39,50,132,161,127,
120,223,185,204,127,130,211,175,209,109,191,67,
187,85,5,100,2,199,197,205,94,39,57,31,
144,4,184,231,17,44,85,97,18,68,210,207,
178,9,233,139,77,156,142,69,179,95,62,76,
26,226,134,205,76,126,131,93,91,58,236,61,
124,244,242,66,193,251,157,40,59,41,90,58,
79,102,147,252,164,217,163,187,29,195,28,107,
232,242,144,68,16,34,146,113,6,158,79,49,
83,165,51,156,162,209,157,97,118,106,21,73,
58,14,101,107,13,74,100,147,251,39,107,186,
89,93,186,195,194,39,114,22,113,115,192,107,
147,36,191,117,249,52,234,181,85,36,33,8,
166,11,40,235,19,130,165,217,8,54,90,184,
167,97,230,7,104,175,190,27,241,39,14,8,
81,34,102,70,205,110,215,67,234,141,162,185,
204,79,158,103,224,38,120,189,122,129,175,245,
255,156,192,15,185,253,131,117,178,25,157,223,
204,237,80,207,95,102,222,123,5,79,186,192,
149,108,130,129,43,147,217,170,33,55,58,224,
249,132,133,12,193,197,165,200,113,141,108,100,
215,139,216,223,18,56,87,150,249,214,86,150,
231,23,154,188,181,151,56,106,170,175,239,34,
111,238,130,128,149,213,1,75,137,24,147,242,
46,111,85,167,88,44,12,25,165,21,212,229,
57,148,102,135,43,251,93,140,98,145,235,183,
50,132,126,192,123,25,155,121,85,66,150,37,
38,65,200,181,15,19,204,44,254,225,125,196,
208,246,206,183,120,254,98,21,163,39,241,250,
36,77,138,223,37,245,216,249,227,77,233,131,
61,56,110,239,125,236,153,144,3,159,165,119,
126,196,141,75,207,162,3,154,117,178,40,57,
12,60,146,233,211,100,231,162,126,134,149,43,
147,153,60,142,151,30,208,114,254,77,84,26,
204,36,163,26,139,128,181,161,77,167,227,179,
251,122,138,92,89,227,218,134,73,102,225,37,
154,55,255,154,247,87,37,140,226,111,96,228,
121,160,156,182,149,61,174,50,106,25,111,178,
205,63,114,106,198,97,243,213,53,166,206,23,
201,46,101,89,187,125,243,16,81,118,241,124,
154,241,216,199,245,4,237,237,155,228,146,22,
31,120,103,48,77,16,241,12,167,99,85,164,
213,193,131,224,120,132,51,197,99,104,67,85,
138,148,123,203,233,1,245,97,129,212,120,66,
183,53,197,27,90,159,220,204,128,171,11,49,
146,154,242,8,64,199,71,99,156,220,3,223,
75,24,120,159,200,46,74,138,130,174,219,88,
233,35,94,139,161,200,36,165,4,47,86,70,
76,29,100,68,43,123,243,199,238,91,95,251,
46,95,158,238,179,223,236,177,49,233,144,200,
159,58,234,27,234,230,207,109,111,245,100,246,
254,239,21,207,156,104,82,233,127,240,61,138,
244,220,87,121,239,213,13,126,165,152,161,162,
4,72,134,198,59,74,14,189,119,30,207,174,
34,42,14,238,245,41,26,65,28,245,202,198,
97,83,74,206,79,72,104,95,64,86,143,70,
27,142,218,155,232,86,158,221,221,255,19,245,
76,244,18,133,35,67,173,130,90,8,62,186,
55,137,43,42,211,49,29,55,16,108,140,162,
116,206,175,105,36,90,1,241,75,71,181,209,
180,174,210,115,121,168,147,184,219,60,43,91,
6,93,215,35,169,40,248,66,96,200,50,157,
94,18,37,101,243,68,54,129,180,185,199,56,
21,63,36,255,101,19,13,178,137,6,213,230,
210,65,47,226,222,232,69,139,112,223,39,188,
50,137,38,174,103,82,107,207,63,180,137,125,
23,165,117,55,19,169,117,230,120,47,61,225,
241,207,63,134,80,21,36,1,161,31,48,44,
100,184,105,24,76,187,62,146,20,241,25,170,
195,41,178,75,23,72,223,185,126,216,6,146,
250,67,164,254,48,26,138,147,138,35,114,105,
54,135,211,40,43,91,156,158,185,67,115,173,
205,84,206,100,118,54,70,50,165,17,22,179,
204,205,65,152,186,130,226,7,148,252,17,221,
81,150,140,20,141,70,189,117,250,37,164,132,
137,186,222,32,101,22,200,245,163,158,135,95,
41,242,158,117,145,157,93,141,103,203,67,178,
106,146,146,97,34,215,182,145,87,171,52,134,
41,250,254,41,240,108,132,24,48,41,231,216,
217,150,41,244,226,132,195,207,243,234,134,132,
96,76,90,181,208,68,154,205,31,171,84,46,
253,30,8,137,94,243,93,122,103,60,20,201,
160,77,156,180,58,160,122,227,91,44,62,121,
84,91,222,189,243,215,92,188,188,75,170,86,
167,126,122,150,254,155,139,204,44,159,255,212,
123,62,213,170,177,241,248,243,17,53,249,231,
188,84,35,134,172,104,212,54,77,126,90,41,
210,31,6,152,183,199,180,235,42,110,63,73,
120,181,207,229,23,58,92,172,54,249,81,127,
25,16,20,206,31,17,190,28,183,113,159,113,
29,78,124,18,134,122,44,200,213,147,103,112,
228,51,188,221,253,51,158,250,195,18,59,117,
88,185,25,112,78,179,152,201,73,136,16,20,
93,34,153,210,104,236,79,232,247,61,36,25,
18,246,53,156,73,192,178,154,66,185,246,224,
12,57,56,61,27,245,186,14,174,158,231,83,
179,3,222,174,230,240,122,46,230,208,34,180,
33,62,213,199,51,199,152,25,193,200,15,30,
233,40,188,143,120,138,64,109,209,219,187,73,
44,123,164,202,251,105,174,173,27,127,202,210,
147,35,242,102,18,89,146,48,228,72,133,56,
90,174,200,30,237,140,29,236,160,197,100,208,
196,72,22,8,92,27,199,221,34,148,65,177,
29,28,111,227,152,163,248,255,250,250,133,59,
10,33,4,187,183,255,26,69,31,225,15,67,
102,46,255,203,143,215,38,145,36,148,188,201,
248,202,52,83,76,243,65,31,58,181,47,82,
152,189,64,179,245,109,2,229,22,146,187,136,
52,221,192,121,55,131,249,68,23,17,72,100,
180,95,161,176,240,252,241,168,175,250,54,169,
169,101,84,53,3,68,6,49,236,235,136,70,
12,117,241,158,13,211,149,73,231,37,18,154,
204,237,254,248,30,36,13,184,27,22,149,197,
231,233,238,253,3,114,37,106,170,142,188,147,
149,225,116,89,34,174,41,100,13,13,77,146,
145,37,144,146,125,28,39,228,253,238,16,115,
161,76,102,104,83,144,100,234,247,8,235,85,
242,27,212,58,115,148,179,219,0,244,71,57,
82,241,246,39,94,123,93,115,72,39,90,52,
123,211,104,234,4,63,208,8,133,140,8,163,
3,154,140,117,216,107,47,160,41,46,133,244,
30,217,68,147,237,110,138,139,11,1,170,36,
241,70,205,226,150,29,195,187,3,96,113,29,
21,89,142,164,153,191,54,55,162,234,158,38,
89,174,33,215,90,7,25,133,5,154,134,188,
85,195,91,156,229,70,123,129,165,245,159,144,
20,125,110,116,178,232,154,204,157,149,1,169,
148,74,34,161,210,186,182,71,97,226,17,206,
20,97,226,82,217,111,115,103,254,69,164,206,
128,13,227,60,75,238,77,252,141,17,147,190,
143,253,216,85,222,209,207,145,44,142,185,177,
55,203,215,231,126,198,19,246,24,58,54,173,
228,2,75,105,3,105,187,199,238,184,136,46,
108,242,246,45,182,182,198,204,46,198,248,192,
141,17,247,94,38,119,246,233,7,174,211,233,
131,137,146,78,175,65,96,183,121,99,115,6,
217,248,6,70,63,143,100,78,88,188,122,4,
137,236,55,110,193,212,251,44,186,145,58,241,
7,183,2,74,243,95,255,212,231,163,188,122,
141,113,58,143,175,233,255,36,142,2,96,216,
92,71,245,202,4,147,47,144,50,99,212,110,
125,143,236,244,83,244,228,31,3,125,242,16,
245,3,98,101,38,195,54,122,44,115,76,54,
251,238,229,250,33,119,234,35,38,126,192,213,
249,251,163,86,197,72,96,148,190,193,181,173,
191,162,85,83,153,120,19,228,171,23,233,175,
175,177,127,189,193,242,217,36,241,184,74,16,
132,200,178,196,104,20,48,59,19,163,189,49,
100,101,101,192,220,124,140,120,252,126,147,36,
242,17,170,169,55,81,120,171,150,160,227,57,
12,29,141,151,150,90,188,247,150,198,192,240,
201,159,239,241,194,108,10,120,180,90,68,40,
96,123,236,112,231,64,194,255,46,200,69,158,
110,179,94,251,54,201,59,112,250,242,127,127,
226,114,207,125,25,161,9,91,182,77,136,201,
114,210,58,70,254,235,78,66,254,253,90,130,
113,96,145,42,215,168,246,255,53,218,214,25,
36,179,69,249,98,155,137,98,82,12,234,188,
94,127,147,252,194,139,15,124,15,191,244,142,
162,190,242,61,20,57,129,171,172,243,114,97,
159,215,196,23,79,36,96,53,189,244,187,188,
253,254,55,145,244,9,170,40,83,94,122,17,
167,215,32,110,94,160,83,223,67,191,186,65,
208,83,144,45,29,89,130,176,86,166,112,250,
192,73,28,52,141,17,130,48,176,49,19,21,
92,251,52,99,111,13,73,11,81,138,14,178,
214,63,28,125,73,8,65,95,101,152,118,113,
14,72,119,135,210,224,50,24,103,199,236,87,
127,204,176,96,80,184,7,201,244,168,107,33,
97,161,75,18,29,215,231,102,127,140,151,16,
36,100,139,93,57,199,204,220,58,186,172,16,
202,10,140,29,174,206,148,248,78,189,125,72,
182,147,165,144,152,62,162,63,206,146,138,117,
112,220,216,167,114,20,0,150,62,98,104,167,
17,72,228,31,128,204,178,140,17,126,160,81,
109,46,81,201,111,176,213,152,225,79,71,99,
64,162,16,31,241,27,167,66,36,9,108,63,
164,100,233,120,161,192,11,67,226,170,66,115,
64,164,172,10,160,170,136,100,12,121,115,151,
238,252,57,246,250,57,30,243,127,70,175,213,
132,156,78,145,58,59,79,255,46,233,238,54,
229,214,7,136,80,176,181,53,66,81,36,146,
125,27,165,144,138,230,117,247,70,244,118,123,
44,102,62,128,177,76,117,117,64,175,231,114,
69,26,145,29,249,12,135,62,47,157,158,103,
247,78,153,197,78,36,107,94,88,20,172,166,
159,160,89,185,192,184,152,98,122,231,71,84,
235,62,153,156,134,117,186,204,238,202,28,115,
87,158,62,209,122,205,62,254,71,31,91,234,
233,116,126,196,197,39,52,164,205,6,40,50,
163,137,76,86,209,63,213,187,201,212,183,233,
21,103,177,147,153,127,210,115,151,42,159,37,
85,62,210,28,154,185,252,91,180,171,111,33,
230,54,48,100,153,132,7,175,251,75,168,153,
23,216,187,241,15,88,233,25,138,203,47,28,
59,155,99,55,224,214,222,16,47,8,153,201,
90,15,181,161,70,114,22,146,255,29,86,120,
141,64,93,225,122,77,80,200,106,164,172,62,
161,174,49,241,4,229,138,69,190,96,112,235,
86,31,89,130,233,138,197,238,158,77,117,119,
204,217,229,251,73,98,82,127,132,136,91,124,
119,211,228,244,212,14,246,120,194,215,23,179,
196,84,3,235,115,30,223,217,48,249,92,37,
113,112,204,37,2,33,163,202,31,67,170,149,
34,46,208,214,200,161,55,241,105,127,83,33,
247,141,128,238,184,64,162,48,64,54,61,70,
157,109,226,185,79,55,131,58,110,92,100,48,
222,100,91,56,140,130,128,47,20,34,71,87,
239,103,216,31,90,44,21,110,49,10,124,206,
167,226,188,219,153,160,150,63,228,92,58,78,
209,140,163,236,119,209,131,9,223,40,221,226,
7,155,127,67,254,212,111,255,7,225,40,228,
95,232,63,83,76,20,205,98,62,209,229,237,
141,24,66,212,217,93,255,183,12,90,183,31,
145,62,39,152,59,247,135,204,158,250,87,148,
151,190,142,8,3,90,91,111,146,41,63,129,
132,196,197,76,156,151,207,166,184,252,204,132,
138,105,162,72,38,32,81,91,251,54,123,55,
191,23,25,115,207,65,49,117,252,201,0,51,
54,141,24,202,200,227,60,218,120,14,73,24,
199,86,68,159,119,145,36,120,58,159,98,62,
126,188,46,168,228,124,204,199,246,48,50,46,
251,157,89,234,221,89,106,237,249,71,54,150,
119,70,14,113,77,225,153,124,138,43,153,136,
115,240,70,221,160,146,219,33,103,168,56,65,
64,219,245,144,252,0,85,145,56,159,58,30,
25,165,226,109,100,57,160,214,153,35,25,239,
252,92,239,97,42,189,139,235,25,135,140,241,
251,214,91,241,152,41,172,97,132,121,190,188,
180,135,38,75,92,157,110,242,27,11,81,10,
157,214,84,202,86,52,96,74,151,37,226,158,
207,200,53,136,203,67,112,189,232,36,250,62,
34,157,164,181,244,24,189,161,197,197,189,31,
98,223,222,99,99,107,68,163,57,33,239,238,
113,126,240,26,57,183,198,237,216,21,222,29,
94,70,81,36,214,215,135,92,123,171,129,187,
213,136,234,28,181,14,66,64,167,237,210,110,
56,196,19,209,108,243,78,203,161,94,119,208,
77,25,37,109,17,38,19,17,27,60,25,71,
40,50,167,119,127,74,59,204,178,20,220,132,
137,135,101,41,100,43,73,126,52,60,75,182,
242,229,159,123,47,143,187,59,236,109,253,143,
36,150,118,200,7,126,212,67,81,21,174,94,
128,218,157,239,126,154,84,27,109,226,252,147,
59,137,135,214,178,131,38,146,42,136,201,10,
111,239,206,51,74,189,68,173,243,231,132,241,
230,3,157,196,205,189,1,94,16,162,41,50,
149,140,241,104,64,196,220,101,220,73,131,80,
76,104,218,115,200,167,22,9,242,121,246,170,
35,66,17,41,42,231,114,6,189,190,135,235,
133,76,21,77,134,125,159,209,248,254,121,45,
242,94,243,176,128,123,37,147,228,82,38,113,
24,165,231,12,13,113,79,176,118,125,111,22,
69,122,180,52,197,155,63,25,19,174,233,12,
221,20,185,223,9,240,133,198,192,78,163,171,
19,148,76,200,176,119,27,17,248,159,106,109,
243,115,207,161,52,162,225,109,237,137,199,207,
154,3,222,173,150,81,228,144,199,166,247,184,
156,141,243,82,57,199,230,200,97,49,97,241,
220,84,26,93,150,136,41,202,161,64,170,230,
58,148,252,183,63,245,51,252,255,58,163,200,
148,31,103,229,173,63,225,234,227,54,233,73,
15,117,120,139,213,210,2,91,254,51,39,63,
160,157,93,70,221,85,60,101,59,210,126,247,
203,236,141,215,168,152,6,103,147,49,222,110,
15,80,194,92,148,9,84,87,56,245,244,31,
227,217,35,234,27,223,38,60,243,30,141,157,
29,252,110,18,253,178,79,72,139,144,22,24,
18,56,6,152,71,104,129,167,115,41,10,134,
70,222,208,208,37,137,149,161,125,95,131,56,
105,29,105,19,237,181,23,168,228,54,31,126,
48,133,96,165,63,38,163,107,180,39,46,125,
87,198,113,115,60,94,234,115,37,147,160,225,
184,36,52,149,96,41,34,96,221,213,128,234,
184,71,27,37,97,246,73,152,255,52,10,178,
197,76,245,88,57,235,48,34,148,101,2,4,
186,44,83,74,12,57,159,142,211,30,232,52,
39,30,110,24,98,126,132,112,39,181,186,72,
246,132,154,62,207,217,214,107,81,164,61,27,
53,58,123,177,50,141,65,154,243,173,31,211,
219,238,176,189,57,34,12,162,204,33,25,87,
209,85,149,222,106,27,121,188,134,81,122,138,
187,133,150,48,20,172,173,13,185,112,33,29,
105,109,245,189,104,116,169,39,200,100,117,144,
160,217,116,89,94,78,66,54,133,91,73,211,
234,78,83,122,250,73,134,205,144,210,214,7,
8,77,101,121,252,46,202,118,29,38,30,190,
170,114,43,127,137,120,234,159,33,130,159,127,
253,58,245,127,228,43,207,75,196,54,70,200,
107,13,252,103,46,34,44,147,198,186,67,182,
242,201,37,162,47,253,248,111,185,254,194,111,
254,66,206,97,224,77,80,204,77,158,203,165,
185,245,174,140,185,240,159,81,91,253,59,36,
61,192,16,243,135,78,194,15,4,181,238,132,
102,123,16,233,144,1,51,89,243,196,51,237,
167,22,158,167,181,243,51,16,113,246,210,79,
49,91,89,39,181,190,139,55,9,233,245,92,
226,49,21,129,71,173,230,112,106,49,78,91,
149,88,93,25,114,254,124,10,93,63,218,103,
82,119,128,188,93,103,49,94,68,150,134,44,
222,19,188,73,128,33,43,236,12,21,58,189,
105,158,153,175,158,72,64,115,115,85,194,76,
202,84,230,85,86,58,179,200,132,199,122,126,
131,230,77,198,226,125,116,177,8,174,74,201,
124,9,213,136,159,120,141,117,169,194,132,40,
0,94,105,167,248,234,252,14,5,51,50,183,
137,3,39,247,249,169,244,97,63,47,125,183,
188,157,56,10,16,207,153,77,190,191,249,67,
178,197,167,254,227,114,20,154,149,32,59,245,
12,215,59,175,240,66,208,162,181,52,203,205,
221,115,148,79,95,60,241,61,26,245,191,195,
151,91,40,82,154,192,181,73,37,158,164,217,
185,205,29,109,140,47,4,126,87,37,83,136,
202,10,75,207,254,23,168,186,133,103,143,240,
68,11,101,156,66,42,244,113,210,130,65,191,
140,166,78,34,6,179,34,144,28,11,65,228,
40,52,89,34,111,70,77,167,125,219,165,104,
26,108,140,156,143,45,49,25,170,67,103,80,
196,208,108,108,55,142,101,12,239,27,102,52,
240,131,195,105,89,251,189,57,42,185,77,100,
76,4,80,48,142,15,89,105,77,188,99,78,
226,51,201,240,30,112,160,150,18,22,11,113,
147,81,16,28,162,76,166,45,147,82,230,193,
145,174,220,25,32,87,247,97,106,1,185,214,
188,219,1,37,156,47,83,235,102,56,23,188,
133,188,223,102,111,215,198,61,80,236,84,36,
137,205,205,17,249,177,79,189,238,48,59,27,
99,199,11,8,239,17,140,179,237,128,70,195,
65,202,79,240,253,16,119,18,253,109,171,21,
189,163,100,226,160,57,56,24,161,118,58,60,
53,254,33,43,195,243,168,227,1,141,73,142,
41,111,159,240,198,22,190,0,53,109,113,205,
184,140,149,253,99,172,68,129,113,123,239,231,
115,18,187,239,18,20,215,208,187,32,111,238,
17,166,19,188,231,133,44,24,33,115,179,26,
175,254,248,239,152,79,252,17,178,122,178,18,
84,102,127,135,59,207,124,229,23,114,6,91,
219,63,68,18,111,144,87,5,215,175,157,35,
177,240,235,200,138,74,229,204,175,51,108,158,
39,185,184,76,16,10,106,61,135,90,111,66,
48,154,128,21,25,72,77,145,41,36,79,94,
86,75,149,46,18,120,62,138,106,34,203,1,
213,238,14,198,88,97,123,187,199,244,116,36,
166,168,123,10,170,42,211,31,248,44,45,37,
216,223,119,216,221,179,89,92,136,211,246,178,
52,189,28,72,2,245,131,144,165,23,134,15,
108,56,44,39,108,222,222,151,248,131,229,234,
35,157,88,24,194,95,253,233,144,108,14,190,
242,91,81,70,29,42,123,220,218,91,100,246,
30,192,136,126,181,30,57,85,162,236,125,103,
117,151,233,185,63,64,143,101,79,248,237,143,
158,99,33,219,161,104,102,62,230,19,247,36,
150,134,126,40,70,58,148,76,180,79,193,58,
255,165,104,102,39,10,167,233,212,59,124,120,
89,226,182,235,195,216,64,8,113,162,94,5,
128,208,199,104,243,3,188,219,26,222,100,68,
232,122,136,113,150,205,120,27,177,91,64,119,
151,169,156,137,136,104,170,126,84,90,17,138,
141,136,245,33,128,20,19,228,80,195,246,13,
90,253,50,133,100,29,217,75,19,118,53,132,
57,192,51,29,234,206,132,25,203,164,233,122,
108,12,109,158,200,38,89,25,142,25,60,164,
113,157,75,213,241,3,13,207,215,201,167,106,
12,157,20,181,246,60,134,230,28,131,177,182,
6,37,130,64,61,36,192,213,28,151,210,196,
163,231,122,156,73,30,69,19,121,67,163,96,
104,52,39,222,103,231,40,14,234,184,87,178,
9,122,174,79,115,18,205,15,87,101,137,180,
28,109,141,238,56,78,202,180,15,200,110,15,
56,120,165,60,82,173,133,149,145,89,219,88,
96,201,220,68,222,111,227,186,50,9,219,71,
241,118,241,220,16,123,28,96,153,10,133,41,
3,207,21,212,234,54,195,3,97,185,86,219,
195,49,71,247,141,100,244,253,104,188,104,42,
169,49,232,249,135,81,45,16,57,158,73,72,
58,171,35,53,87,163,95,238,175,226,91,73,
234,102,133,102,43,90,223,133,115,5,94,113,
159,36,179,252,95,159,216,112,63,234,26,140,
223,225,229,217,16,253,250,70,100,68,102,75,
76,5,1,237,161,205,181,117,19,85,77,157,
248,127,165,154,123,104,206,24,183,56,251,153,
158,187,65,227,54,195,254,53,210,210,10,174,
40,18,230,126,155,84,226,200,232,73,178,66,
178,184,76,115,224,178,213,30,227,7,247,7,
17,73,75,61,113,54,113,88,130,154,189,194,
164,223,100,235,195,63,199,47,61,134,117,70,
144,24,189,141,105,201,52,219,19,242,121,131,
76,90,163,221,113,201,101,116,138,69,51,234,
19,250,2,213,222,103,42,168,147,72,168,200,
23,23,9,147,247,115,138,228,173,61,174,238,
119,144,206,46,96,7,22,113,85,121,196,158,
135,116,86,240,149,223,74,210,243,124,26,142,
199,110,47,195,84,166,250,64,146,236,225,223,
157,174,178,122,231,47,152,47,191,76,60,187,
240,200,239,237,5,30,251,221,25,4,18,67,
99,192,186,101,115,42,97,157,204,191,88,38,
140,29,10,147,14,150,248,71,236,126,156,68,
225,212,127,100,142,34,191,128,221,223,100,213,
11,120,34,105,225,95,188,195,230,238,123,100,
103,158,56,97,83,197,66,184,32,143,74,108,
127,248,103,168,41,157,112,162,194,190,69,86,
255,10,153,249,251,245,53,60,103,8,70,180,
201,36,5,80,70,132,140,208,251,73,10,228,
104,212,151,17,214,144,169,100,19,236,56,8,
184,214,25,147,210,52,102,45,131,214,196,165,
104,234,220,236,143,30,250,110,197,65,109,95,
85,188,99,101,162,177,147,100,187,113,26,67,
115,16,66,38,21,111,99,233,35,114,186,198,
200,151,81,36,137,125,199,229,194,65,79,66,
26,140,17,7,14,99,46,102,126,102,142,162,
209,157,38,155,108,28,206,120,152,55,67,132,
44,31,143,114,132,160,62,72,115,62,181,142,
84,31,17,150,114,72,157,62,82,111,24,77,
18,76,88,136,132,69,240,228,121,166,99,99,
236,186,79,173,94,164,172,239,179,161,157,227,
116,231,85,108,59,96,52,242,169,156,74,179,
229,150,208,7,235,216,163,227,153,82,224,6,
56,138,202,108,209,100,48,240,176,237,0,213,
48,217,211,47,115,94,219,37,94,52,49,12,
133,181,141,33,186,42,227,250,33,197,130,137,
61,9,168,215,109,12,67,33,244,5,131,129,
71,82,12,200,171,54,106,222,132,39,151,248,
94,245,2,149,179,127,244,200,233,139,159,172,
89,233,98,170,74,212,135,73,198,200,199,45,
226,239,222,194,207,167,185,227,252,42,229,11,
39,67,61,197,250,29,84,215,161,49,127,246,
51,118,18,43,236,173,255,59,102,166,44,72,
189,76,60,182,124,223,103,38,126,200,70,99,
76,207,126,248,126,27,58,209,208,161,79,10,
6,210,227,25,228,156,77,203,189,70,50,60,
207,242,227,123,4,251,117,226,113,5,77,147,
208,117,133,124,222,160,223,247,201,229,52,6,
182,143,51,9,24,12,124,134,67,143,139,23,
211,40,235,85,164,254,144,112,166,20,17,45,
21,57,98,228,251,1,210,200,230,137,149,173,
8,176,82,46,16,206,22,63,246,33,191,250,
141,8,250,234,133,130,155,253,17,158,151,229,
233,178,77,115,162,209,117,189,7,242,159,29,
55,134,63,61,100,127,252,231,196,214,158,162,
120,234,43,15,149,217,176,123,187,248,210,26,
197,76,148,185,102,228,41,182,218,5,132,163,
83,74,245,137,235,19,36,47,64,136,16,244,
35,248,190,100,79,16,186,134,208,14,178,101,
199,229,139,234,7,188,231,15,105,183,254,144,
84,229,236,47,183,163,16,66,208,220,248,46,
19,183,139,101,228,145,100,139,244,68,195,84,
4,229,90,149,59,227,15,128,19,58,138,32,
139,215,221,135,48,140,116,223,219,55,153,190,
248,219,248,238,24,35,145,123,112,169,96,255,
167,24,75,25,66,71,38,52,122,135,57,159,
170,164,9,226,155,20,82,145,70,211,126,119,
6,211,24,145,177,243,120,177,22,175,53,122,
188,84,206,242,68,54,197,187,157,193,161,244,
247,253,37,28,137,140,174,210,158,68,155,236,
222,225,70,161,144,201,37,247,143,169,189,78,
91,6,83,166,70,92,85,248,105,163,199,36,
8,145,37,9,169,59,68,222,111,19,36,231,
15,179,138,43,153,4,31,246,70,71,101,47,
33,163,216,83,72,146,134,16,130,208,245,8,
205,38,146,121,114,109,249,86,191,76,204,28,
160,42,30,110,40,69,189,135,145,141,122,125,
45,106,6,167,19,72,253,33,123,133,139,76,
197,122,72,215,214,240,155,125,140,132,206,126,
117,72,62,111,160,172,86,33,102,32,166,178,
132,133,12,202,118,13,127,223,71,147,36,208,
84,188,76,6,127,51,0,9,134,67,143,85,
107,25,69,248,244,187,238,125,7,49,157,214,
80,11,6,37,45,134,109,15,136,205,204,209,
23,89,158,202,220,161,190,231,208,240,66,6,
35,31,66,56,117,58,129,101,41,132,11,21,
132,105,32,239,54,144,250,104,183,232,196,0,
0,32,0,73,68,65,84,35,198,182,143,61,
9,88,94,78,34,146,49,106,51,37,94,191,
115,137,185,11,255,252,231,210,205,121,96,22,
37,228,35,118,167,231,19,127,227,58,194,208,
240,147,9,68,231,100,44,106,57,12,48,198,
125,218,211,159,93,164,56,216,191,77,237,214,
15,88,122,246,63,101,225,210,127,133,17,143,
206,199,189,242,229,2,168,247,28,118,218,206,
177,140,237,65,151,235,135,52,135,19,166,146,
198,39,122,14,73,81,153,46,253,75,236,126,
149,120,102,158,155,235,215,121,33,215,1,63,
164,221,157,16,179,20,82,105,147,209,200,103,
60,14,48,13,133,149,149,1,17,125,70,98,
48,240,201,105,50,114,189,141,92,111,71,160,
5,67,67,20,115,132,59,109,100,33,144,134,
99,80,85,148,59,91,136,108,18,145,120,52,
76,54,165,169,36,189,75,124,225,76,141,156,
17,227,28,96,7,33,155,35,135,158,231,71,
80,254,145,76,123,80,68,149,125,10,233,61,
136,131,227,253,144,245,245,119,35,161,81,79,
33,147,120,145,204,204,229,3,91,23,210,216,
253,62,242,242,30,57,93,35,103,168,88,138,
205,108,44,60,100,139,203,91,53,148,245,42,
254,249,83,136,82,244,78,164,221,38,114,173,
201,181,90,156,90,175,192,213,84,64,126,206,
64,110,245,120,124,184,194,74,242,127,227,214,
141,171,100,102,94,196,76,125,122,61,170,32,
20,216,110,84,73,248,36,179,208,62,115,71,
33,194,128,198,202,171,136,212,251,4,211,93,
70,213,41,230,78,253,183,244,107,25,222,175,
191,66,225,170,198,108,115,157,173,149,111,145,
206,61,26,178,104,234,37,194,226,187,132,141,
128,192,157,80,152,121,145,126,227,38,146,144,
209,227,25,134,205,85,70,189,219,184,82,157,
192,119,144,39,69,148,98,15,208,145,253,36,
98,148,64,228,171,7,181,242,144,233,88,52,
234,243,189,206,128,98,118,135,90,103,142,138,
25,48,58,104,54,245,188,128,141,145,77,221,
113,177,20,153,243,233,24,215,58,227,67,173,
38,136,102,70,180,38,30,178,36,17,87,20,
4,33,26,50,9,85,102,191,155,160,152,57,
174,72,121,57,147,32,20,2,89,138,178,134,
148,166,32,121,1,202,7,119,8,174,68,209,
222,205,254,136,221,177,139,29,4,136,113,2,
77,202,17,134,19,2,169,79,16,171,31,221,
204,2,6,73,208,6,15,85,179,61,22,237,
184,113,20,217,63,116,92,119,33,184,34,25,
39,88,158,71,253,96,5,234,45,130,211,179,
244,130,36,23,86,127,140,219,236,177,182,54,
196,117,67,124,95,176,187,107,147,203,233,204,
206,10,100,187,134,212,232,32,217,19,146,170,
74,221,45,34,180,49,129,35,176,98,10,187,
123,54,173,216,12,227,212,19,44,245,190,67,
160,201,120,222,209,218,25,166,66,79,157,33,
116,97,99,123,64,177,104,176,183,191,143,105,
111,210,81,45,42,51,49,6,3,143,169,146,
137,51,14,162,153,8,138,66,88,204,31,100,
52,49,164,159,125,72,163,49,161,144,51,144,
52,153,224,210,41,214,6,1,104,67,234,171,
127,79,241,212,203,12,26,183,113,39,45,114,
149,103,14,140,232,42,146,172,146,40,44,124,
226,61,237,133,26,98,18,181,222,165,3,137,
146,215,228,69,58,59,143,81,168,60,199,234,
254,136,148,165,126,172,65,61,247,218,63,112,
231,233,207,182,47,33,73,6,11,87,255,0,
69,183,80,244,227,101,143,32,16,180,122,19,
26,131,201,161,225,56,201,85,237,56,20,18,
198,39,206,42,70,157,13,36,85,193,72,22,
241,226,6,222,208,103,107,107,132,174,41,228,
178,50,190,23,160,27,50,137,184,74,40,4,
229,178,69,173,102,35,132,160,90,29,211,233,
184,204,205,197,162,38,183,16,72,142,139,180,
85,227,205,193,85,158,78,188,7,165,28,34,
25,143,102,126,59,46,34,30,59,214,4,8,
67,25,89,62,30,76,41,146,194,98,42,36,
103,28,125,208,82,100,206,167,98,116,237,24,
181,94,134,225,68,227,115,75,183,241,195,144,
155,253,40,0,148,52,80,231,35,14,85,176,
157,33,93,185,112,148,153,141,218,4,241,72,
246,38,161,41,156,78,196,14,165,100,36,219,
65,185,185,129,116,32,217,47,5,97,68,225,
240,2,214,223,233,113,167,87,34,23,115,137,
101,146,188,61,201,241,252,112,5,235,220,34,
202,214,30,139,41,159,129,182,194,250,45,151,
133,103,254,248,19,239,133,106,199,166,53,244,
112,238,41,157,155,254,136,139,137,204,137,160,
175,159,185,163,168,221,249,14,210,255,75,221,
155,254,72,114,231,103,126,159,95,220,17,121,
103,221,119,85,87,223,108,54,217,36,135,156,
25,205,80,115,172,180,154,93,239,74,90,93,
182,86,182,97,96,129,5,236,23,6,252,23,
248,165,13,248,133,1,191,177,12,99,109,11,
150,229,181,181,150,4,105,101,105,36,205,104,
36,206,12,57,188,217,236,171,186,171,235,62,
178,242,206,140,200,140,251,231,23,145,93,221,
213,7,187,123,134,99,97,2,32,8,54,59,
35,51,35,35,190,199,243,253,62,207,179,176,
135,90,10,208,252,25,226,208,65,8,133,210,
204,5,142,54,118,233,185,239,176,172,38,212,
199,62,197,239,205,225,84,159,156,45,253,126,
11,69,205,211,235,84,200,175,222,161,231,30,
208,27,38,136,66,136,223,40,211,184,241,125,
212,217,67,212,229,145,205,97,100,19,4,1,
197,124,139,99,176,195,182,160,83,65,51,115,
84,74,117,78,23,28,246,31,208,149,81,133,
160,90,110,226,183,166,217,31,246,168,234,58,
113,42,169,24,89,91,186,233,6,153,192,217,
168,26,59,9,63,73,28,77,97,194,50,41,
234,26,113,42,57,42,116,78,48,158,1,106,
126,192,130,147,109,110,132,105,202,148,237,32,
106,205,76,122,123,36,215,61,103,91,108,55,
109,68,172,128,221,4,213,227,114,41,207,135,
237,225,35,184,151,40,244,33,176,80,250,85,
210,242,3,106,154,143,81,203,245,252,34,57,
243,228,230,84,63,74,48,77,5,165,231,30,
127,43,113,103,23,43,95,38,62,59,135,225,
13,57,119,78,225,198,141,30,113,156,5,235,
122,61,99,215,198,177,228,226,197,18,134,161,
160,139,152,72,102,183,212,11,202,85,234,158,
195,64,81,168,23,95,70,109,125,194,234,184,
203,173,195,209,166,138,173,225,217,11,12,115,
21,10,225,30,78,107,135,238,32,102,119,55,
97,113,49,135,229,148,81,151,167,145,182,137,
62,128,186,62,205,92,165,155,65,115,205,46,
200,20,101,99,31,229,176,193,214,182,135,109,
171,24,186,66,20,166,132,239,223,230,181,233,
10,222,27,54,55,214,154,236,126,188,77,172,
9,138,151,14,168,221,94,167,60,241,75,164,
73,64,109,227,79,56,59,254,95,61,247,61,
109,171,2,101,4,9,74,211,224,93,125,158,
161,255,79,152,90,185,72,211,13,105,121,30,
139,99,39,3,115,156,74,92,63,70,83,5,
147,181,93,110,189,241,11,164,234,79,225,241,
75,99,182,63,252,19,22,95,253,53,242,19,
39,147,96,156,164,244,252,152,70,211,163,61,
8,65,203,101,234,192,207,113,132,113,74,173,
23,48,93,122,190,174,34,145,46,65,180,193,
24,175,227,250,121,62,48,47,176,252,98,155,
66,227,128,126,63,68,246,5,174,23,97,153,
42,19,147,22,227,227,6,249,188,198,246,182,
71,24,166,132,97,72,154,74,38,39,45,74,
165,251,112,205,23,242,31,177,21,204,19,222,
213,41,22,99,166,163,189,140,197,253,0,147,
123,175,83,101,24,26,168,106,138,23,88,156,
159,222,163,83,15,120,231,251,17,151,222,208,
31,249,172,119,27,147,24,210,231,214,223,222,
225,155,255,65,142,188,147,125,215,9,203,224,
131,86,159,110,116,31,58,85,229,73,114,162,
149,31,71,116,28,160,207,225,48,164,29,198,
252,220,68,9,77,8,212,171,235,89,231,115,
239,209,245,3,68,24,161,94,91,231,140,221,
103,222,232,242,110,127,137,215,212,235,120,227,
211,216,50,68,86,139,184,229,50,127,243,183,
54,154,30,81,152,57,224,96,237,219,84,231,
94,59,238,16,159,218,89,250,49,123,237,71,
85,28,252,40,97,175,61,100,105,66,255,135,
79,20,138,106,32,81,16,26,36,218,1,194,
24,31,5,253,58,126,184,203,123,126,196,132,
174,208,47,14,72,15,174,81,229,209,213,194,
222,209,45,58,205,31,145,148,55,144,137,129,
238,64,221,91,192,178,7,199,218,69,201,68,
138,22,196,168,249,251,85,67,187,63,241,168,
92,133,234,35,212,34,177,189,75,45,130,163,
90,112,34,224,167,82,178,19,4,232,138,198,
218,136,185,89,210,53,46,149,243,188,221,232,
114,185,156,39,150,146,187,238,144,67,63,51,
191,17,192,197,114,142,49,195,192,139,99,130,
68,82,213,245,145,189,99,135,253,246,212,113,
162,48,85,133,7,39,1,103,138,14,182,170,
32,154,221,204,17,110,84,170,21,116,149,85,
7,110,196,217,235,166,44,243,184,50,177,84,
133,146,174,209,8,34,46,150,114,196,169,228,
70,207,67,232,29,68,144,39,85,7,232,209,
28,105,18,146,26,45,48,238,227,206,227,197,
3,186,222,24,238,176,196,68,121,159,179,69,
39,235,104,90,93,148,141,125,154,205,128,122,
61,96,97,193,33,214,65,109,117,24,180,134,
212,235,1,190,127,178,242,12,70,155,72,235,
235,46,23,46,20,71,185,73,32,6,62,250,
214,14,69,67,240,142,60,135,134,96,89,28,
112,227,122,143,52,149,44,159,155,164,45,166,
152,74,14,216,219,216,34,82,96,113,53,79,
146,72,230,231,29,204,138,67,125,252,44,187,
225,4,86,163,129,173,4,84,138,13,110,220,
174,226,229,151,121,249,165,54,170,34,209,62,
184,9,73,202,236,156,13,18,134,126,194,208,
141,232,187,3,186,159,212,57,247,181,136,55,
206,45,146,92,244,241,226,132,31,54,82,202,
75,62,141,91,255,47,194,210,209,198,229,143,
37,233,156,164,57,24,13,77,131,74,137,218,
33,84,102,199,72,82,201,118,115,72,217,214,
209,71,195,255,182,23,113,216,245,233,251,89,
112,49,226,16,109,175,142,122,86,65,255,156,
159,181,221,171,127,202,228,169,55,153,60,251,
230,253,223,40,74,105,244,124,214,182,90,28,
28,118,73,226,20,37,29,160,105,26,139,103,
151,81,149,231,103,31,239,119,134,76,20,140,
231,122,109,161,124,14,119,125,3,86,192,89,
252,77,250,237,29,90,226,143,24,30,109,208,
106,100,182,195,185,188,198,220,156,205,96,152,
176,182,214,103,98,194,68,121,224,61,122,189,
136,40,74,41,20,138,199,127,46,132,100,217,
202,86,188,91,227,231,184,157,251,6,203,211,
45,130,88,231,163,189,101,28,61,100,185,122,
196,92,249,62,65,245,118,125,134,219,223,223,
100,252,252,191,228,230,251,127,192,210,220,253,
164,119,231,104,154,217,114,27,199,8,112,47,
42,168,74,194,61,186,89,78,83,249,210,68,
137,111,31,180,142,97,58,153,8,246,175,255,
5,186,109,83,154,124,25,84,13,145,106,199,
69,160,34,192,79,82,242,170,138,72,78,62,
63,202,230,126,150,44,70,29,134,173,38,188,
84,104,176,29,172,80,94,45,242,105,238,117,
44,55,100,125,183,201,204,217,62,175,185,91,
224,216,120,171,63,226,175,62,168,177,244,194,
179,117,22,65,244,100,88,186,214,13,24,43,
102,146,44,63,149,68,17,198,41,246,51,44,
118,76,174,188,201,145,251,135,192,112,20,40,
45,164,148,236,92,255,3,46,127,109,192,166,
7,197,156,197,248,48,224,64,171,83,219,248,
54,99,115,95,61,222,88,170,111,125,15,175,
240,61,148,165,20,69,36,8,61,66,227,62,
195,216,245,139,228,173,30,166,62,36,28,148,
79,108,206,60,184,79,45,165,66,16,89,88,
198,0,69,227,216,237,78,62,166,250,86,252,
113,34,107,239,248,207,23,115,22,5,77,225,
114,57,79,78,87,241,162,152,86,16,158,16,
21,187,222,245,56,157,151,228,52,149,130,174,
34,4,236,187,10,183,14,103,143,37,191,1,
130,81,187,121,239,45,77,69,201,236,66,253,
0,89,185,207,74,245,122,41,43,147,10,55,
70,13,194,233,130,131,169,40,20,117,141,43,
213,2,5,77,37,145,242,24,247,244,211,148,
221,193,16,25,150,80,41,16,219,217,195,163,
14,38,73,212,38,168,247,111,210,82,174,73,
16,230,57,106,47,161,135,41,103,243,3,212,
245,93,14,14,134,236,239,15,41,151,13,114,
57,141,188,91,227,163,239,244,144,157,207,230,
110,12,6,49,119,238,244,57,117,42,143,37,
2,154,81,149,49,189,197,247,26,11,40,86,
129,153,163,239,83,82,123,228,78,229,25,184,
49,145,235,98,39,61,142,186,209,113,149,168,
42,130,106,213,36,150,64,4,27,155,38,95,
88,184,138,226,143,180,134,6,80,230,46,9,
121,62,186,251,21,46,159,110,146,44,205,160,
52,58,24,163,37,131,94,63,226,112,223,39,
24,65,41,238,250,17,183,198,74,180,20,133,
25,219,100,210,50,184,214,49,152,112,92,82,
49,139,82,240,8,251,45,204,226,248,115,222,
253,83,132,94,132,13,124,167,179,66,117,254,
183,33,77,217,239,248,68,73,138,162,104,15,
52,117,242,56,73,88,161,207,234,193,26,87,
167,78,99,29,184,92,152,45,156,244,34,255,
9,231,128,147,43,95,165,56,123,154,40,73,
217,111,251,220,218,106,178,189,221,164,211,242,
40,149,115,44,175,78,226,228,76,8,71,75,
25,218,143,55,187,137,19,73,173,27,48,91,
121,118,135,52,167,178,72,121,44,43,2,53,
211,33,63,177,202,225,225,37,236,5,88,206,
221,100,111,171,151,249,82,40,130,156,163,177,
188,156,195,52,21,124,63,219,150,107,119,178,
174,127,56,76,56,56,24,50,51,99,159,72,
34,0,213,131,91,228,74,13,110,138,47,225,
11,155,156,238,115,105,118,231,145,207,114,102,
226,128,213,95,49,249,243,63,254,35,180,220,
50,3,119,15,39,175,80,119,139,76,21,187,
56,70,54,103,58,127,217,122,236,226,202,162,
99,178,229,249,72,64,206,110,17,172,3,193,
4,253,250,77,202,179,47,161,203,25,98,246,
142,161,221,68,74,68,16,194,48,160,22,77,
208,143,243,140,235,45,202,90,23,113,120,82,
164,179,164,246,40,57,61,216,130,169,226,45,
210,83,115,156,42,121,40,166,142,114,228,67,
171,135,57,94,126,196,152,237,179,151,47,30,
197,164,151,199,29,204,40,196,200,149,24,70,
41,79,241,117,67,53,12,227,191,254,199,191,
243,95,30,15,185,30,86,197,124,220,209,233,
123,220,60,28,208,30,102,76,77,219,80,159,
208,5,135,28,110,254,123,114,179,53,98,37,
37,57,114,176,184,140,149,155,195,245,110,114,
110,57,68,8,193,134,235,99,105,42,126,193,
35,210,119,80,218,51,88,197,41,146,208,167,
209,255,51,148,9,23,37,44,128,30,28,7,
115,75,85,16,34,101,24,100,190,13,170,146,
208,233,172,146,43,28,80,212,85,76,85,69,
40,1,245,222,52,158,95,228,160,181,132,101,
12,233,120,227,12,221,49,250,177,137,231,151,
178,127,130,34,113,98,96,25,131,172,160,239,
85,153,176,108,210,176,128,28,20,208,108,151,
122,16,113,182,232,32,128,24,50,61,252,135,
142,165,188,197,188,99,98,107,42,55,91,42,
111,237,171,76,148,247,70,85,9,136,208,66,
4,38,78,98,81,206,75,234,65,196,90,127,
64,201,212,49,4,200,137,42,24,58,253,110,
138,105,11,52,77,48,97,25,180,195,136,89,
199,196,86,21,22,115,214,49,191,225,193,85,
197,73,203,200,240,80,51,160,149,184,199,146,
36,82,247,208,252,121,82,61,11,246,227,166,
206,11,186,198,235,210,227,149,170,207,180,38,
233,237,37,168,189,6,157,67,15,223,207,160,
37,33,160,185,91,135,201,51,168,110,141,52,
253,236,201,87,16,164,244,251,49,11,227,33,
219,225,18,159,182,44,250,84,185,98,92,103,
177,56,164,221,9,81,4,236,31,12,25,12,
98,10,5,157,114,217,64,215,21,138,37,131,
118,39,202,148,69,253,20,75,73,25,134,38,
21,195,69,125,88,118,62,12,193,48,233,170,
37,242,179,153,58,173,156,168,128,174,99,14,
50,47,234,36,78,89,57,149,167,88,54,176,
22,167,89,40,231,153,115,76,134,73,202,158,
151,176,50,5,237,77,147,177,213,54,253,125,
141,124,101,245,241,1,49,24,160,89,143,146,
172,204,220,52,7,141,15,89,72,90,108,24,
167,40,78,125,21,207,243,216,234,201,209,28,
40,161,239,199,228,45,141,130,165,227,71,41,
195,48,97,188,119,196,90,156,99,239,176,129,
98,23,25,196,146,146,163,63,49,89,164,35,
173,39,85,127,122,69,38,132,160,59,12,217,
115,5,235,7,46,63,124,103,157,173,141,58,
154,166,114,230,194,44,11,75,227,232,198,40,
129,37,209,189,167,255,25,178,194,16,244,71,
215,58,251,65,76,42,161,96,105,8,33,72,
66,31,69,55,78,216,215,62,2,217,149,103,
16,66,32,147,24,41,83,156,137,151,72,220,
15,177,246,239,50,28,102,182,170,150,169,96,
59,26,158,27,35,133,196,182,53,92,89,38,
201,77,34,189,140,211,224,186,49,119,228,139,
180,147,49,12,61,51,70,58,174,124,131,33,
147,241,30,193,212,52,51,197,14,134,22,63,
225,122,65,235,104,136,55,172,208,173,215,88,
88,49,200,25,1,166,246,217,220,165,157,65,
192,90,255,1,248,72,147,224,184,40,195,9,
114,165,51,232,102,158,65,247,144,186,18,48,
8,10,116,135,37,124,191,76,74,133,134,95,
166,52,56,98,206,60,196,80,194,19,168,112,
186,56,157,25,126,61,240,156,137,32,202,164,
241,59,253,204,78,88,130,172,22,241,166,171,
108,236,230,41,142,63,202,63,123,92,12,215,
85,133,90,47,56,62,181,166,10,198,11,38,
3,207,165,23,41,116,134,201,103,206,211,254,
236,223,252,119,63,94,71,209,31,65,16,131,
48,225,118,205,101,121,220,97,178,104,62,110,
221,129,56,238,227,55,198,177,184,72,117,246,
117,52,211,225,104,253,45,52,171,79,195,139,
112,101,76,152,166,132,35,93,37,117,239,2,
229,243,217,6,193,238,250,223,80,31,115,160,
99,51,163,75,80,122,104,66,176,100,105,76,
107,58,157,129,202,119,98,139,90,103,158,177,
66,141,233,98,147,131,246,2,105,46,224,151,
150,50,94,128,152,236,178,238,14,89,59,146,
148,114,77,74,185,38,74,47,66,26,39,171,
228,56,214,56,220,61,75,213,10,112,205,62,
99,118,11,93,130,176,96,82,201,35,204,251,
21,121,55,124,252,10,161,46,4,110,28,243,
214,62,52,125,73,181,248,64,181,16,169,8,
52,82,199,229,46,30,91,183,230,72,138,77,
148,206,44,3,2,114,179,147,25,108,35,179,
133,26,221,200,110,163,138,161,157,224,152,124,
86,253,169,180,187,172,84,75,24,138,114,236,
113,13,16,105,251,168,65,9,172,30,175,84,
11,88,181,22,245,239,221,34,142,83,74,37,
157,86,178,138,223,110,208,110,103,129,41,138,
82,118,118,6,88,182,134,170,235,244,74,231,
73,82,129,215,15,168,196,219,104,50,251,123,
49,26,9,42,34,171,173,192,139,216,221,15,
184,19,244,113,165,195,171,185,107,20,241,105,
119,98,124,63,161,211,9,41,20,116,150,151,
115,236,238,14,168,84,12,170,85,131,91,55,
123,196,177,68,10,201,216,152,73,46,167,33,
251,2,221,125,124,39,51,219,190,197,71,206,
60,179,227,217,3,43,243,14,236,212,208,52,
65,249,133,25,138,177,36,218,174,147,86,138,
168,142,117,252,128,184,113,66,24,86,169,218,
45,14,237,10,227,169,193,122,122,244,220,247,
191,170,155,116,147,51,72,123,19,25,100,195,
215,221,230,0,249,0,57,170,55,140,185,190,
215,231,210,124,145,213,201,28,182,174,34,203,
43,252,179,247,254,134,255,85,159,224,195,31,
173,115,241,242,34,195,48,97,166,108,50,85,
180,208,212,167,119,23,97,156,210,247,99,188,
32,147,122,209,20,193,32,76,232,14,34,106,
155,71,180,253,14,237,150,135,109,233,188,240,
210,18,229,106,142,159,198,33,37,28,116,124,
218,94,200,249,153,103,83,90,125,152,43,21,
122,29,166,6,91,180,154,1,250,104,201,97,
107,107,144,17,240,220,8,215,141,57,123,174,
136,123,216,103,144,155,32,169,92,34,109,239,
211,102,138,241,218,26,115,11,17,251,254,21,
198,242,39,117,207,164,99,145,132,144,175,248,
153,55,251,214,1,233,226,52,50,119,50,225,
125,241,231,109,174,189,191,193,135,63,136,121,
237,231,210,19,108,240,199,126,103,96,211,123,
20,235,87,114,49,201,202,7,212,175,53,152,
86,255,57,177,178,207,88,41,211,81,59,95,
116,88,53,251,40,66,65,77,110,34,244,1,
168,10,194,208,145,73,10,134,6,113,146,205,
85,130,8,165,213,67,22,108,68,199,29,233,
160,75,210,133,105,48,117,54,236,50,55,239,
22,240,255,190,128,250,156,161,187,104,233,180,
188,240,184,35,92,59,116,97,56,4,83,231,
252,220,211,73,132,63,86,162,112,30,186,160,
219,205,33,99,249,71,49,75,69,85,153,89,
250,85,156,202,20,60,128,3,79,174,126,133,
214,158,197,181,247,254,14,253,252,0,219,80,
40,40,42,205,3,13,117,160,176,125,243,223,
146,136,14,162,220,101,198,202,33,82,157,125,
79,101,202,206,241,45,191,129,210,142,232,174,
204,114,122,208,160,63,155,109,27,109,237,46,
242,229,222,199,228,205,132,91,250,18,63,220,
43,240,230,194,16,49,170,182,245,52,38,76,
38,152,200,123,212,67,133,200,9,24,196,153,
55,244,149,74,129,118,24,161,230,215,50,175,
235,241,45,122,163,92,240,106,181,128,57,20,
148,116,19,63,77,233,71,9,219,94,128,174,
8,20,33,208,132,56,118,201,186,221,31,50,
109,25,52,6,21,202,133,221,17,140,53,137,
34,45,146,212,37,125,64,200,47,41,236,161,
13,102,177,202,13,42,149,2,3,79,34,148,
44,162,25,102,22,122,131,36,165,19,198,244,
163,209,218,233,103,37,137,195,38,234,157,29,
226,11,203,204,149,139,76,76,148,168,71,9,
219,195,128,38,17,50,72,153,12,199,72,253,
20,209,247,232,118,67,92,55,99,70,163,221,
224,80,76,83,230,228,60,71,85,64,236,127,
128,230,129,38,19,28,153,80,103,30,137,96,
146,29,82,84,204,17,155,61,69,16,97,208,
172,15,152,23,159,34,37,232,154,129,40,152,
199,112,86,181,106,208,235,69,220,188,217,67,
81,96,119,119,192,234,106,33,115,142,75,37,
187,187,67,90,141,144,153,41,155,32,126,2,
130,175,40,72,67,35,109,246,72,87,4,202,
232,154,137,32,202,252,47,166,199,232,255,221,
117,138,142,10,154,202,90,127,136,151,36,20,
149,60,111,239,77,243,210,236,30,219,235,38,
19,167,190,74,236,93,197,212,234,207,69,248,
188,119,24,185,203,252,121,99,131,241,149,223,
96,163,62,200,20,132,31,122,154,226,84,210,
232,135,204,86,172,99,152,230,253,43,223,224,
23,63,248,14,159,92,190,196,208,210,73,165,
100,175,237,179,223,241,113,12,13,199,80,113,
76,149,188,169,162,75,136,211,148,174,27,210,
31,198,244,252,136,32,74,9,195,152,86,163,
143,63,8,169,84,243,148,70,201,32,87,180,
209,203,5,150,87,167,158,251,251,252,184,135,
31,165,220,216,239,179,90,74,159,59,152,4,
94,147,77,243,28,231,95,9,153,50,3,214,
63,109,210,239,71,116,122,33,227,227,38,139,
11,57,132,2,203,147,112,55,209,112,90,55,
169,85,206,176,212,254,56,171,238,119,96,126,
233,42,159,138,75,89,87,33,32,146,58,246,
214,16,117,251,38,202,101,32,78,80,14,26,
8,119,64,178,186,128,172,222,135,119,111,95,
243,185,112,197,226,230,213,136,250,94,204,220,
202,103,119,88,2,88,205,219,124,210,233,243,
184,6,91,172,238,178,113,227,223,18,76,106,
20,99,19,199,8,25,55,13,212,195,70,102,
220,229,135,224,88,196,171,115,104,215,54,144,
213,34,105,181,8,150,73,90,200,145,204,130,
178,152,37,5,225,7,25,199,66,74,164,109,
209,236,196,220,186,246,34,227,11,63,71,71,
252,46,81,84,125,174,107,189,48,102,211,247,
99,162,135,86,252,21,33,200,91,234,79,39,
81,20,109,29,8,79,12,128,235,253,240,201,
155,16,15,13,11,163,97,151,110,240,183,228,
231,3,162,205,243,68,154,73,83,153,100,204,
156,166,244,234,25,54,214,255,123,244,153,148,
84,166,164,102,13,67,81,120,41,87,230,147,
206,24,173,74,196,196,224,144,210,250,30,74,
163,205,149,48,98,24,43,228,53,141,157,185,
83,188,112,120,157,213,141,171,24,243,47,112,
51,103,115,161,154,160,72,137,105,116,216,61,
90,224,171,179,146,153,216,167,84,173,208,13,
99,54,61,159,146,161,210,142,34,20,96,170,
178,155,109,6,141,152,211,55,122,3,190,52,
94,98,232,38,124,218,139,233,168,46,41,146,
175,76,148,121,187,209,227,75,19,37,222,174,
119,9,101,74,59,140,152,182,12,130,88,220,
47,65,100,76,108,63,234,255,144,215,85,74,
185,38,211,163,33,181,94,16,15,109,34,197,
220,238,15,143,197,207,222,107,245,120,181,90,
124,68,107,233,248,18,55,218,16,70,104,31,
223,6,85,69,79,83,236,229,25,230,39,171,
180,139,57,214,250,3,150,115,2,77,213,233,
231,171,88,197,28,109,119,136,78,8,113,112,
34,73,68,232,164,8,66,15,66,163,128,46,
134,232,105,66,136,65,137,35,140,227,223,254,
254,211,162,32,73,144,196,232,8,41,73,133,
194,216,116,1,93,137,201,231,53,102,103,109,
186,221,8,199,209,112,221,172,42,203,229,50,
79,3,161,10,20,85,48,24,196,232,186,194,
93,111,154,11,165,199,123,102,36,103,23,81,
54,246,185,100,172,113,107,247,85,206,47,102,
112,68,114,110,9,105,153,104,239,92,165,82,
49,51,39,189,195,38,175,152,6,191,31,47,
18,38,42,95,88,56,224,229,74,129,191,185,
109,146,55,28,210,129,96,97,197,103,127,231,
3,170,243,207,174,167,227,181,54,81,84,135,
210,220,127,66,109,168,209,116,159,236,11,210,
30,132,39,176,252,32,78,105,94,184,204,194,
65,155,35,221,226,30,176,150,196,146,247,62,
217,192,31,132,44,173,78,82,29,47,32,34,
47,43,42,141,132,110,199,163,219,26,48,28,
6,120,125,159,169,217,10,11,203,19,168,15,
204,25,116,93,69,183,45,254,255,62,130,56,
101,187,49,224,124,190,248,92,175,107,29,189,
67,108,12,120,59,124,149,98,174,202,196,197,
143,168,172,125,194,204,180,141,4,90,237,144,
70,221,103,114,194,34,25,14,201,153,9,211,
122,141,7,249,217,67,55,224,210,248,205,147,
157,74,106,96,40,33,114,199,38,121,97,149,
228,146,137,122,107,19,165,213,37,121,32,81,
220,185,145,85,131,150,45,201,151,63,59,88,
134,62,24,22,204,59,38,142,166,240,110,179,
247,136,255,76,170,171,248,167,52,198,75,7,
212,218,11,188,118,170,77,217,208,72,231,38,
81,54,15,64,85,72,206,44,34,199,74,200,
82,29,209,233,147,94,90,101,99,16,176,182,
223,33,222,158,195,208,85,36,9,142,26,145,
237,187,171,248,161,70,146,142,51,115,250,23,
57,184,243,39,84,202,41,138,243,149,231,186,
214,166,166,112,97,54,207,70,125,112,60,51,
187,23,187,15,58,62,115,21,251,233,137,226,
158,246,141,223,109,60,245,13,93,63,166,214,
236,102,64,253,3,166,64,71,129,74,49,121,
180,205,125,220,57,27,27,111,99,216,23,208,
196,36,197,137,251,210,5,94,115,143,245,247,
255,79,212,115,109,82,67,98,249,5,94,46,
87,168,152,42,173,70,76,97,177,193,159,237,
77,242,31,79,69,232,215,239,142,246,152,27,
200,68,229,140,89,227,214,196,235,104,75,19,
24,102,131,165,195,27,184,206,42,162,96,161,
189,127,131,43,149,34,149,184,128,187,209,166,
52,147,7,20,202,97,196,75,37,7,20,133,
156,170,242,115,147,101,234,126,196,15,118,138,
168,74,140,101,12,8,226,108,195,105,193,177,
104,247,61,86,149,50,135,158,142,50,136,17,
66,242,214,81,7,69,17,168,8,44,69,176,
59,136,142,165,49,148,254,56,105,225,241,215,
116,144,164,84,16,76,219,143,38,215,56,205,
220,113,253,52,101,191,31,32,200,72,122,198,
189,77,15,63,24,85,208,54,114,164,54,41,
43,69,56,26,169,202,142,182,43,148,173,67,
196,97,147,242,23,95,228,245,177,236,1,81,
111,108,96,88,38,20,5,141,253,144,0,3,
129,64,39,28,65,72,217,230,146,57,74,6,
129,110,144,202,24,37,25,96,144,144,34,70,
175,145,15,36,140,81,160,34,34,69,32,17,
76,87,85,28,35,165,217,215,240,98,141,163,
35,159,70,227,254,188,193,48,85,202,147,69,
26,61,137,57,26,182,141,205,149,73,85,147,
118,106,98,43,39,103,19,50,103,131,169,35,
203,5,68,16,98,68,45,246,117,201,222,237,
28,19,19,53,138,186,198,178,204,56,22,225,
66,5,99,243,128,100,97,138,143,243,14,195,
134,193,84,121,23,41,77,4,16,250,41,110,
237,46,99,150,66,73,211,185,213,185,129,229,
204,62,26,28,220,54,233,99,212,59,251,135,
27,20,166,206,209,237,5,236,181,71,9,205,
239,65,250,232,223,245,60,104,168,61,156,209,
28,111,193,144,236,212,235,184,86,142,215,175,
127,151,183,79,191,58,74,180,112,225,148,201,
112,32,72,226,22,120,67,100,52,194,222,35,
15,59,137,208,156,132,212,150,228,151,138,8,
145,64,208,128,7,47,211,19,62,195,99,128,
236,227,243,62,189,101,120,182,115,186,129,199,
158,26,50,86,124,186,76,69,26,133,164,105,
204,196,236,215,105,213,126,68,60,119,141,78,
35,32,140,230,248,210,23,35,180,163,29,60,
47,38,8,18,252,97,66,175,31,33,12,133,
48,72,201,41,29,2,85,100,124,26,64,211,
196,113,201,178,191,63,100,79,125,129,87,38,
179,223,68,120,67,180,31,125,138,44,56,164,
149,194,35,108,237,111,253,122,6,153,173,94,
176,80,158,50,215,255,247,255,119,143,211,23,
12,94,124,205,162,106,232,124,125,170,194,157,
126,198,175,186,135,40,52,122,211,204,85,119,
153,180,76,26,189,244,56,145,40,181,22,34,
142,137,47,157,38,174,22,65,74,148,82,30,
37,136,216,26,134,220,233,13,241,111,46,48,
57,247,45,132,80,8,7,45,188,254,1,249,
242,10,170,233,96,210,71,213,77,188,230,30,
170,126,135,196,75,216,219,250,61,166,22,127,
25,69,213,159,26,111,79,204,81,45,232,201,
152,195,174,79,56,232,64,28,112,48,108,147,
143,243,232,159,1,125,106,192,9,238,194,147,
120,12,113,34,89,175,123,116,253,8,116,1,
166,6,218,253,32,231,3,50,87,32,247,152,
53,171,135,207,185,88,253,213,19,255,221,220,
254,33,73,228,227,14,111,210,152,215,152,177,
178,11,252,198,146,134,30,10,218,205,8,195,
80,88,84,45,238,152,10,138,31,60,48,168,
145,116,250,9,137,200,225,108,239,162,212,246,
71,238,116,16,220,184,195,240,182,68,145,18,
217,105,49,150,118,81,72,209,220,60,201,202,
28,68,17,40,54,82,194,156,99,102,195,185,
188,70,50,223,225,189,67,135,142,55,198,68,
233,128,77,119,200,182,231,147,8,201,134,236,
96,228,21,182,164,65,216,30,99,105,198,69,
32,232,199,9,141,32,100,167,61,206,68,105,
31,18,129,52,253,39,14,21,82,41,217,25,
248,164,72,206,21,29,156,7,12,206,111,244,
60,182,30,192,66,37,80,49,116,20,50,181,
86,165,209,69,217,175,147,156,93,132,241,50,
162,214,58,54,11,58,249,38,41,98,24,160,
125,180,134,180,77,136,70,114,216,64,41,167,
176,176,224,176,179,147,213,103,33,58,9,26,
42,9,201,3,204,189,130,119,82,64,79,65,
30,39,145,199,34,67,72,84,53,83,128,221,
184,235,210,239,103,85,219,129,111,32,49,144,
64,33,175,179,184,232,112,184,221,102,117,181,
128,166,73,130,88,193,44,8,182,194,25,46,
218,27,247,171,213,48,65,72,129,54,95,36,
89,157,71,170,74,198,53,137,98,254,81,250,
35,62,180,79,115,117,253,5,86,102,111,178,
34,2,100,193,97,173,92,164,116,185,192,145,
166,114,183,151,160,138,132,113,75,231,82,121,
228,89,32,42,8,93,161,56,145,114,148,132,
168,142,124,236,125,175,168,218,99,109,64,157,
234,12,174,31,179,219,235,67,110,132,207,43,
26,216,143,135,3,118,67,133,139,227,5,140,
81,245,63,172,121,164,214,56,111,191,244,45,
94,189,243,35,62,93,186,76,160,103,157,128,
157,59,1,226,223,195,185,48,114,240,212,209,
243,103,124,134,135,134,3,199,231,229,243,58,
167,218,227,40,205,49,89,172,98,62,101,155,
42,9,125,210,36,70,183,243,56,213,95,161,
223,120,129,166,249,109,6,246,128,31,28,93,
225,202,106,9,237,230,29,130,212,165,52,105,
224,13,83,66,77,101,106,161,204,254,254,16,
167,170,211,108,4,32,5,7,45,80,108,200,
229,84,122,73,1,209,189,205,187,156,97,169,
216,162,160,184,247,110,34,32,68,168,46,177,
94,64,58,86,38,100,89,204,145,150,10,25,
198,10,217,198,85,225,241,159,253,141,175,217,
4,131,251,115,74,67,81,184,88,202,113,177,
148,99,144,36,236,15,2,24,234,124,121,162,
68,73,215,184,88,144,116,7,51,64,7,49,
218,26,84,119,14,217,200,217,180,195,136,87,
203,5,82,67,231,206,97,68,188,247,18,99,
147,171,4,238,119,73,82,155,226,196,207,19,
4,7,28,109,126,143,165,87,126,11,205,176,
81,13,139,200,239,98,104,9,111,92,177,216,
246,6,124,252,233,255,133,72,76,170,249,111,
82,156,190,248,212,24,126,28,51,19,201,81,
218,3,169,128,89,36,85,52,234,210,224,244,
103,204,179,158,25,122,218,105,13,232,14,62,
91,119,168,222,15,31,155,40,30,91,168,244,
235,120,237,13,114,165,101,6,157,93,242,151,
174,243,114,209,228,86,119,200,94,123,142,156,
217,231,251,71,29,190,49,93,97,220,209,81,
142,90,68,249,18,197,158,68,248,25,214,191,
227,207,226,167,22,190,38,208,123,71,204,23,
142,232,219,37,162,185,105,244,221,67,236,176,
143,173,167,160,235,164,115,227,12,123,49,122,
163,137,168,183,81,187,46,201,165,85,164,132,
161,151,226,228,21,214,250,3,22,28,147,78,
24,83,202,53,153,45,182,105,247,150,232,203,
62,229,124,3,91,85,88,206,219,28,14,3,
246,134,1,200,2,155,158,207,185,162,67,179,
47,169,247,103,25,248,5,162,156,137,30,41,
72,211,125,234,117,216,27,4,188,80,202,209,
141,98,10,154,150,89,57,60,36,165,240,66,
41,71,78,83,81,106,77,212,235,27,32,68,
38,107,157,179,179,42,41,239,128,222,125,50,
182,218,238,33,30,178,176,216,237,219,244,220,
232,4,2,107,16,142,18,133,70,140,134,198,
143,167,94,155,36,18,207,139,9,195,7,216,
235,104,248,228,40,83,199,20,146,205,181,1,
73,34,81,71,85,76,152,40,68,106,158,220,
3,27,44,141,122,64,173,230,115,234,116,30,
213,212,89,27,248,76,154,6,149,23,79,35,
13,29,245,230,38,175,110,93,101,254,130,203,
159,221,58,199,165,215,118,72,78,205,115,187,
222,25,221,99,58,109,119,138,153,234,22,139,
78,1,235,168,73,50,86,70,149,69,84,205,
166,238,69,236,220,92,38,246,35,210,197,232,
132,109,238,147,38,184,113,224,145,104,14,183,
107,46,242,25,37,16,194,56,101,237,48,91,
135,245,130,228,132,224,222,135,167,94,101,174,
181,199,81,105,242,56,89,252,172,30,25,143,
100,192,153,169,252,115,189,174,48,126,6,167,
180,72,237,238,119,8,195,62,7,45,149,83,
110,143,176,63,100,249,76,25,219,18,92,235,
24,12,251,23,241,91,62,0,0,32,0,73,
68,65,84,30,139,211,58,65,148,224,53,163,
140,203,16,193,193,166,143,162,8,242,121,141,
249,178,134,174,220,160,158,94,98,246,1,27,
128,227,227,238,29,146,139,43,40,12,17,27,
135,164,139,211,36,167,23,16,237,62,50,5,
183,235,32,132,64,213,193,178,149,227,123,116,
127,39,166,209,93,97,109,189,205,244,100,135,
151,94,187,223,57,57,170,202,233,130,131,17,
27,216,194,2,98,108,61,164,158,140,184,69,
99,37,104,101,26,105,150,55,100,63,74,40,
20,115,88,154,197,224,234,34,106,161,203,121,
249,7,76,22,124,246,39,198,248,100,205,97,
122,245,91,76,141,150,241,194,168,67,28,14,
217,219,252,125,126,229,202,16,253,189,59,156,
78,37,99,47,156,226,131,190,139,191,217,224,
121,64,191,166,27,226,63,196,173,104,185,33,
13,71,103,60,111,252,100,137,98,24,165,207,
244,1,74,142,78,197,121,58,149,232,96,235,
223,161,158,221,163,215,85,16,75,130,128,148,
195,97,132,165,10,38,203,123,184,126,145,70,
231,20,186,179,139,218,232,160,108,31,210,78,
170,156,141,84,20,43,11,38,11,86,70,50,
72,202,69,110,169,103,185,219,77,49,231,170,
12,99,176,43,85,150,151,14,232,169,42,7,
178,64,51,201,211,50,138,44,205,214,88,109,
220,64,120,62,193,219,119,241,79,45,83,92,
46,210,109,37,156,46,217,220,114,7,132,169,
228,139,227,69,164,20,28,90,7,124,176,55,
129,148,2,63,73,153,182,77,86,243,54,41,
176,70,202,122,10,215,219,17,181,206,28,179,
99,155,80,130,70,119,6,93,13,169,14,170,
36,249,189,167,94,139,191,62,108,83,26,153,
26,13,227,244,17,33,192,68,102,250,81,114,
68,252,73,86,102,79,106,217,56,22,201,226,
52,90,251,51,248,14,170,130,204,217,136,17,
231,96,173,87,166,21,79,48,166,237,50,95,
14,240,58,30,73,60,218,60,35,38,192,248,
137,216,152,15,38,9,128,6,115,204,115,59,
131,110,70,93,70,97,113,153,13,191,2,2,
250,161,193,140,213,96,92,59,34,78,5,154,
34,105,180,2,202,85,3,219,82,137,45,147,
154,31,114,182,224,32,201,36,219,139,166,142,
22,132,76,173,109,113,113,110,130,191,188,85,
65,232,3,2,247,69,72,19,100,238,144,153,
234,22,99,166,206,28,153,1,206,213,48,135,
233,204,162,219,37,106,59,48,190,248,243,228,
170,243,60,83,212,23,2,140,28,183,246,250,
68,137,124,174,235,49,8,19,110,236,63,250,
186,84,81,217,171,206,113,229,238,251,188,127,
250,117,126,214,143,182,23,209,29,68,148,156,
231,163,19,170,186,73,69,191,206,248,106,130,
219,7,189,156,35,215,139,184,115,225,159,98,
182,234,232,195,54,205,221,128,238,104,115,174,
92,214,105,181,238,119,182,194,206,83,19,211,
108,117,33,140,2,52,187,195,118,225,85,206,
22,235,180,147,50,101,173,75,89,235,82,213,
58,168,159,102,106,195,178,92,32,173,222,219,
84,75,193,182,143,7,187,73,156,113,153,234,
135,49,215,62,137,48,202,47,48,251,74,230,
54,119,240,209,255,65,255,59,219,124,229,27,
39,97,182,197,74,131,40,209,30,64,78,70,
118,203,83,99,217,234,123,171,199,244,120,137,
111,73,56,242,67,190,247,93,129,117,233,54,
162,26,112,152,24,252,253,245,42,230,193,105,
166,79,191,249,216,34,197,174,36,28,38,146,
120,110,129,138,150,18,42,10,184,10,157,221,
79,153,92,125,243,153,175,245,147,32,166,173,
198,0,93,17,143,253,237,158,57,22,60,11,
49,40,73,37,183,15,93,10,150,134,161,41,
232,170,32,233,7,232,98,136,130,32,103,169,
148,236,236,67,76,204,252,2,237,173,31,66,
105,19,225,132,52,221,18,161,221,165,209,155,
70,195,196,86,117,126,121,161,206,222,247,125,
150,243,25,172,210,25,24,12,18,135,182,239,
48,199,93,142,218,89,117,230,246,14,208,229,
38,8,152,105,87,16,111,190,202,39,93,201,
223,29,188,128,237,212,168,150,82,182,186,10,
83,230,14,83,134,135,41,242,200,189,33,182,
18,34,135,53,26,253,28,165,49,141,205,125,
152,30,51,216,114,125,222,109,246,73,164,164,
55,168,32,69,54,143,144,192,237,158,135,227,
107,212,218,57,14,85,149,126,52,143,161,251,
89,146,24,29,227,165,3,220,97,137,14,3,
74,221,89,210,210,254,103,94,183,121,199,228,
82,41,199,205,222,128,187,238,163,252,140,91,
125,143,68,74,206,77,87,137,75,14,152,15,
205,53,226,56,211,104,122,240,190,42,229,17,
93,55,219,156,136,162,204,160,126,126,10,101,
167,134,186,190,195,215,230,14,248,203,181,113,
58,106,149,179,249,38,75,51,42,251,251,67,
58,157,48,227,81,140,186,0,149,207,223,19,
35,194,164,166,157,101,230,112,147,138,181,199,
242,114,14,189,168,208,15,178,225,118,156,42,
12,35,193,249,115,247,235,164,94,146,48,102,
153,180,194,136,134,31,209,10,35,94,73,37,
170,109,226,159,93,196,139,14,81,115,54,50,
150,148,166,107,136,193,24,210,105,101,76,120,
1,236,214,168,21,74,236,108,191,64,121,108,
26,69,40,40,137,64,220,91,14,120,134,45,
161,84,102,107,133,126,244,227,57,31,13,238,
137,177,61,124,94,69,229,253,211,175,243,198,
173,31,240,193,233,47,16,169,250,207,116,178,
216,106,14,120,209,46,61,191,229,180,153,99,
97,117,136,33,37,124,100,49,65,137,56,104,
177,218,126,7,76,216,202,105,52,26,1,105,
122,50,175,111,242,2,229,160,205,156,118,64,
126,220,224,134,63,131,73,204,120,255,135,148,
75,69,132,150,177,183,107,209,4,215,189,115,
156,177,215,209,149,24,162,248,120,11,74,248,
17,178,92,34,145,10,13,183,192,84,161,203,
218,141,128,163,222,5,38,46,127,25,175,113,
131,253,15,254,31,102,95,249,23,204,188,252,
219,52,110,127,151,157,205,119,89,88,54,31,
10,194,143,127,94,164,34,240,202,121,246,122,
3,102,108,131,27,159,198,216,203,2,43,87,
36,137,34,54,245,14,218,133,62,241,122,239,
177,6,73,154,105,99,164,95,229,230,182,64,
213,243,124,224,255,37,66,79,73,15,231,16,
179,251,12,218,251,56,149,217,103,186,204,79,
66,125,146,84,114,235,208,165,154,51,40,57,
39,181,202,158,61,81,168,207,254,171,63,56,
85,23,94,136,76,239,227,237,115,21,155,185,
138,69,126,108,149,252,216,42,189,163,117,110,
239,124,155,220,84,143,102,127,138,201,82,141,
75,101,135,177,142,196,243,138,88,226,128,112,
118,150,244,246,46,227,28,208,168,187,244,6,
41,215,200,97,199,30,17,38,138,72,49,45,
21,219,84,168,156,157,96,115,168,226,26,13,
198,103,59,52,251,83,236,116,116,102,202,7,
172,202,28,147,119,55,33,73,239,155,192,14,
2,114,5,149,90,47,97,59,200,225,54,26,
196,45,141,36,180,168,235,19,148,115,77,198,
75,7,200,190,129,108,85,56,72,103,8,91,
62,65,169,129,177,210,102,50,255,120,56,46,
111,119,169,119,103,41,104,17,210,87,72,183,
102,112,156,21,64,195,247,118,17,231,239,7,
246,177,145,239,196,189,29,109,25,101,209,68,
104,199,197,4,123,195,128,178,161,49,245,184,
173,22,153,205,35,100,49,83,126,237,26,101,
58,137,73,177,60,164,52,147,105,219,7,249,
34,6,144,46,76,33,199,203,248,31,109,243,
82,181,142,235,70,56,134,142,97,40,204,207,
59,8,1,157,78,8,113,72,140,74,132,142,
206,79,46,117,94,165,198,46,103,17,164,24,
248,204,199,87,179,4,16,130,231,197,148,203,
6,5,51,187,111,116,5,162,84,203,182,68,
150,103,65,85,185,107,219,52,253,144,102,16,
209,141,50,207,12,125,110,130,112,117,158,235,
93,15,55,142,48,12,23,140,12,63,78,177,
201,25,26,95,172,150,176,62,93,163,181,214,
224,195,153,47,51,115,234,91,12,187,117,66,
191,199,153,51,42,55,238,254,16,167,244,107,
159,125,63,215,54,201,79,46,179,94,243,112,
253,159,158,153,212,187,103,222,96,174,185,75,
173,50,67,248,57,121,103,252,67,28,126,148,
114,208,245,153,45,63,31,148,214,13,231,113,
227,155,148,183,142,184,254,221,45,86,95,158,
100,88,24,227,104,241,69,122,29,29,187,122,
7,179,191,141,231,197,143,60,0,229,120,31,
175,11,94,23,102,173,46,65,8,173,202,69,
180,80,82,28,220,102,88,210,153,178,235,76,
233,117,238,12,87,56,109,111,100,146,57,163,
89,158,122,103,23,89,200,209,144,211,164,82,
161,182,27,81,219,75,72,149,219,116,110,94,
231,244,57,149,59,205,132,195,247,255,7,226,
52,143,110,151,249,228,71,33,83,211,38,198,
99,190,102,152,166,4,169,100,24,39,140,153,
58,7,195,144,143,70,220,166,183,247,243,232,
78,202,55,166,52,198,198,82,222,218,80,113,
201,158,119,169,251,39,100,101,164,76,105,110,
253,136,201,51,111,50,190,244,197,251,215,248,
238,54,150,54,143,152,87,105,237,188,143,181,
50,241,19,119,20,199,48,148,23,98,106,10,
73,42,143,41,15,207,142,46,200,207,231,38,
218,111,15,169,230,244,99,54,119,113,114,149,
169,206,24,169,185,143,99,186,92,42,229,88,
89,223,65,30,117,104,216,23,89,157,13,145,
81,66,109,0,145,27,208,232,9,242,69,139,
83,23,170,12,188,34,214,172,69,51,87,165,
188,190,143,45,3,250,158,138,154,143,89,208,
202,236,139,30,73,162,33,124,21,117,48,205,
170,184,157,37,137,7,209,132,56,161,153,132,
172,181,10,248,137,36,62,178,24,215,126,139,
72,9,48,218,239,17,246,2,146,162,77,57,
253,22,149,213,43,244,235,235,212,171,191,135,
83,122,232,60,143,185,68,99,230,62,221,205,
83,204,20,94,39,55,119,10,187,60,153,117,
70,7,215,104,15,238,160,142,208,163,15,91,
253,236,245,18,194,91,37,244,96,145,52,95,
71,93,61,188,95,141,198,9,215,58,30,93,
231,94,48,21,140,155,58,5,93,67,22,28,
162,175,188,12,186,198,223,253,141,79,111,176,
192,160,190,5,169,66,20,14,17,72,190,248,
141,152,83,23,140,204,75,162,148,71,154,58,
179,227,2,198,239,7,36,77,19,44,44,56,
20,10,58,205,102,38,250,23,254,24,0,212,
99,180,8,113,232,225,112,18,26,187,183,33,
181,181,229,145,207,235,36,66,35,76,4,138,
0,71,79,32,205,108,85,165,170,80,59,104,
29,147,50,1,202,186,198,237,40,229,176,223,
165,31,197,216,170,194,169,188,141,169,42,236,
12,2,218,66,50,140,83,104,180,217,255,232,
144,116,121,14,35,255,245,99,95,10,69,209,
248,224,29,27,229,51,102,59,73,20,128,76,
201,79,45,115,247,200,163,61,136,126,170,1,
54,85,84,14,170,115,92,220,190,202,39,43,
87,126,166,187,138,253,182,207,120,222,56,30,
224,63,203,33,200,160,28,165,146,231,212,47,
156,227,134,125,150,65,20,147,134,83,172,214,
190,77,211,88,32,54,43,16,156,92,220,184,
7,105,222,59,54,252,37,138,99,121,46,149,
54,217,188,211,163,47,161,86,19,172,174,22,
50,19,164,145,167,182,72,82,68,163,3,165,
60,241,107,231,145,150,69,171,150,231,194,244,
30,20,117,126,105,254,94,103,151,253,251,212,
57,216,221,244,217,184,211,167,113,176,71,48,
180,248,238,247,102,16,178,129,161,244,248,242,
155,54,134,149,37,137,155,221,1,71,65,136,
169,40,196,82,226,197,9,73,162,209,232,205,
80,202,53,153,194,102,162,154,221,235,243,37,
159,27,174,200,88,222,115,251,236,220,250,125,
114,185,51,20,42,231,25,118,119,241,253,71,
225,235,233,83,247,61,79,138,83,23,158,139,
55,211,232,63,121,25,197,54,84,242,166,198,
252,152,125,226,25,126,230,40,16,165,159,79,
166,144,192,97,55,96,101,226,62,198,62,62,
253,53,14,154,107,104,99,1,139,66,160,52,
58,108,237,122,180,213,14,225,235,58,2,104,
236,182,25,166,38,139,43,54,149,215,87,120,
203,52,240,147,148,88,74,146,134,193,138,51,
207,89,247,46,233,206,33,118,179,69,245,242,
34,65,56,65,167,87,160,109,167,204,104,77,
12,47,184,111,188,34,68,70,102,49,117,54,
189,128,94,191,128,80,93,148,250,50,133,151,
78,35,101,74,167,241,61,140,11,1,233,64,
195,59,88,199,202,77,225,182,110,162,173,158,
76,18,101,93,99,1,201,122,10,131,123,194,
95,18,244,205,55,184,116,225,151,16,170,134,
223,173,19,251,125,186,71,215,112,253,79,81,
31,42,0,100,100,144,238,142,179,116,234,95,
18,249,30,173,238,95,147,114,120,18,186,72,
18,214,250,3,94,170,228,169,26,58,230,131,
156,138,145,1,138,55,44,161,138,6,191,254,
159,229,178,149,191,84,162,189,123,13,2,29,
121,51,243,111,232,157,62,131,152,159,128,110,
253,145,236,118,79,63,167,119,143,113,72,250,
220,191,241,179,220,178,9,42,41,42,41,32,
99,201,126,83,82,176,67,246,182,251,232,134,
202,210,146,131,50,94,132,52,165,17,39,39,
146,4,64,39,140,17,64,217,208,88,206,89,
244,227,132,229,188,205,91,71,29,250,110,17,
161,31,242,245,193,0,247,221,13,244,162,197,
71,218,203,140,207,188,116,255,114,57,37,86,
223,248,207,159,8,57,165,73,194,160,185,143,
81,158,101,189,222,167,59,136,126,106,36,182,
40,74,208,180,76,31,44,86,53,62,89,185,
194,155,159,126,151,31,174,94,249,153,133,161,
82,41,217,105,13,89,157,124,118,102,120,148,
164,252,104,175,74,20,141,81,78,202,204,78,
245,57,191,118,149,250,71,251,172,245,34,138,
83,58,161,82,64,229,126,162,216,225,44,11,
35,111,106,128,61,78,51,205,38,106,51,102,
163,249,224,120,41,27,118,3,40,163,123,90,
22,28,148,122,27,177,121,64,114,229,44,8,
168,230,92,106,189,50,83,197,71,7,225,113,
8,31,127,92,96,238,245,127,69,207,253,93,
38,166,187,196,185,49,170,167,126,147,52,14,
249,171,191,248,93,190,240,122,200,228,172,198,
229,74,38,34,122,111,222,216,118,39,144,169,
202,149,217,35,44,105,98,147,34,20,149,187,
238,144,53,223,67,9,10,68,173,8,90,69,
18,189,78,215,190,77,207,253,43,210,218,4,
179,167,126,229,41,163,179,103,191,47,219,94,
196,118,107,240,153,227,131,7,99,243,115,39,
138,244,199,76,20,134,38,48,109,13,129,32,
78,37,81,146,50,124,8,227,181,138,147,208,
200,33,199,2,182,83,201,170,101,50,51,157,
18,28,236,16,137,5,186,215,246,112,125,149,
203,95,42,99,124,233,34,98,235,144,87,143,
234,184,229,28,87,85,19,217,253,2,71,106,
27,10,49,167,163,59,92,255,100,64,184,57,
160,80,57,228,236,66,204,187,233,12,190,167,
16,69,2,247,244,2,137,170,224,150,242,148,
220,33,123,150,65,207,143,216,12,11,140,13,
4,179,213,203,163,139,175,80,40,93,194,115,
143,144,131,28,177,210,230,104,253,239,137,68,
141,113,69,99,94,164,148,135,1,142,31,96,
235,42,74,146,178,120,118,137,191,60,108,18,
167,146,184,165,49,181,244,198,9,103,181,198,
222,15,8,23,222,122,152,127,136,210,157,38,
117,234,104,106,1,221,41,18,249,30,242,9,
45,92,73,215,152,54,205,39,182,143,37,165,
134,90,180,80,148,172,58,19,67,63,211,144,
241,134,12,35,149,84,106,56,155,91,136,243,
11,79,236,18,171,85,3,41,115,212,106,62,
34,130,32,214,49,136,63,191,182,18,136,48,
176,70,66,145,49,42,21,39,65,215,21,198,
199,76,38,38,44,84,77,32,117,13,116,13,
207,205,136,135,113,42,73,165,204,60,1,4,
188,58,86,196,232,121,136,90,131,116,113,154,
245,163,4,175,51,13,154,203,84,65,195,218,
27,98,156,155,226,187,222,89,138,115,39,225,
165,222,209,45,122,173,15,152,191,248,59,39,
131,65,224,177,243,225,31,51,117,238,235,232,
99,75,124,186,219,101,253,78,157,197,83,19,
159,123,48,245,220,33,27,235,71,116,219,30,
185,188,197,203,175,157,58,254,127,63,184,240,
21,102,107,119,216,175,204,18,243,179,121,52,
221,144,217,178,245,68,45,184,71,33,145,16,
51,183,195,197,174,75,239,157,91,204,246,38,
184,253,222,193,241,60,162,163,47,96,28,125,
200,131,209,35,198,224,128,83,88,120,84,168,
49,199,157,199,199,161,7,212,36,142,215,191,
71,236,103,209,115,81,63,190,77,242,210,25,
166,10,93,234,110,145,221,246,24,243,149,147,
157,203,91,127,221,71,207,93,224,224,195,223,
227,242,203,30,11,203,57,254,234,47,174,2,
223,100,255,195,63,4,117,150,195,253,53,38,
103,179,103,222,209,84,8,34,218,205,21,206,
228,5,11,165,46,5,197,65,85,5,194,150,
252,160,222,161,29,198,200,8,226,155,85,102,
206,126,139,150,249,125,130,120,11,17,152,40,
222,20,99,11,95,69,253,28,182,225,146,84,
178,213,28,124,102,55,81,205,27,20,45,109,
228,149,35,126,188,68,177,52,238,176,219,26,
210,121,74,251,173,169,130,217,178,141,109,40,
216,186,74,220,243,113,170,159,173,5,147,38,
49,40,217,227,208,142,19,146,217,9,148,224,
128,130,51,196,92,157,33,191,85,199,82,123,
172,127,220,192,74,246,232,90,99,140,239,29,
82,221,174,243,181,197,73,126,104,234,84,150,
127,147,163,219,146,179,197,26,203,211,29,90,
245,27,68,189,148,170,107,241,207,171,5,188,
47,92,224,47,218,175,80,223,43,81,157,253,
4,117,36,33,222,173,231,25,4,121,166,181,
93,38,203,191,140,161,221,215,177,159,88,122,
19,237,160,132,93,152,198,154,155,26,5,146,
1,253,237,239,225,248,223,166,246,225,109,20,
5,166,38,109,116,91,197,54,76,156,156,69,
47,141,193,183,72,172,251,179,153,112,208,195,
237,95,231,113,114,50,169,238,130,158,160,8,
235,56,88,37,105,255,164,161,72,10,40,80,
53,117,140,78,15,17,197,164,211,99,39,182,
34,16,130,175,46,54,193,208,72,162,121,208,
181,227,13,143,172,202,203,86,96,175,127,208,
67,191,218,32,73,82,150,150,114,20,10,250,
67,179,113,73,62,175,97,89,57,194,48,197,
117,99,14,155,130,56,17,24,4,159,107,146,
200,110,194,132,78,39,98,97,193,161,82,53,
81,84,1,154,74,178,58,151,117,108,134,206,
21,67,167,98,104,108,184,67,146,17,239,69,
239,123,168,159,220,206,204,106,90,61,138,203,
231,137,115,187,56,170,194,74,144,34,58,46,
223,229,21,74,43,255,26,85,63,57,116,140,
194,46,175,190,114,200,39,87,127,15,167,244,
69,114,229,69,34,223,37,244,91,204,94,250,
199,68,90,153,181,181,13,174,174,185,140,77,
22,63,247,32,218,239,13,185,254,241,54,19,
83,69,100,34,233,117,7,244,186,30,197,82,
86,129,199,170,206,208,176,25,239,55,56,180,
75,63,179,16,212,110,123,248,204,235,178,186,
150,49,153,213,219,53,130,32,101,231,211,163,
227,36,33,4,40,170,130,109,171,184,238,253,
212,185,194,167,199,157,68,133,218,147,147,178,
151,233,141,89,150,74,85,109,115,215,95,98,
165,119,72,242,194,41,148,221,35,68,167,143,
216,171,35,23,167,153,200,247,232,251,54,119,
27,83,156,26,191,127,206,211,23,76,238,28,
150,73,130,247,88,88,206,145,166,48,236,38,
248,253,6,186,118,196,165,23,98,98,93,225,
78,211,199,235,164,108,111,65,63,92,225,244,
139,29,190,56,167,210,30,228,41,58,30,97,
154,242,86,189,199,32,76,72,119,38,177,149,
139,76,190,254,117,246,110,254,17,241,236,13,
10,205,175,83,29,123,21,109,214,1,33,8,
221,206,79,14,5,118,252,207,76,18,57,83,
227,244,228,231,192,163,112,12,149,179,211,121,
250,126,204,205,77,239,137,181,229,68,193,60,
33,229,17,31,199,177,148,97,251,128,40,232,
32,132,134,80,85,12,187,138,145,43,211,111,
221,162,48,238,101,236,231,52,225,122,165,200,
197,52,33,189,84,224,219,59,29,190,244,230,
69,150,46,250,92,221,202,33,138,191,136,51,
126,145,155,193,31,240,245,211,107,104,27,123,
204,55,255,144,70,97,138,137,211,191,193,59,
55,27,156,155,185,197,249,137,58,159,94,237,
178,223,4,173,217,102,170,249,54,255,104,42,
71,50,63,205,15,123,47,19,168,26,90,148,
240,21,177,133,62,17,241,254,221,42,229,51,
23,142,89,234,247,142,202,3,112,69,182,125,
224,80,89,254,22,119,15,12,150,46,255,41,
205,171,219,108,238,12,80,144,148,26,55,153,
249,181,215,232,69,49,218,156,75,125,227,47,
89,42,253,171,17,212,81,196,201,173,16,203,
214,49,54,51,102,234,84,12,157,141,65,38,
123,30,105,123,236,222,253,125,194,228,8,109,
190,135,225,207,51,153,111,49,174,216,248,129,
128,226,144,163,32,100,211,54,89,137,34,212,
91,91,89,48,85,20,244,239,127,140,52,141,
140,100,87,247,81,14,154,200,82,14,225,13,
137,227,140,215,208,234,199,24,34,34,73,36,
209,200,124,101,125,221,101,113,49,71,175,23,
225,186,17,97,152,62,118,83,84,1,82,204,
159,248,134,13,176,208,31,147,108,142,142,124,
218,237,144,40,74,89,93,205,83,174,90,72,
199,198,141,19,62,233,184,12,147,132,139,165,
220,9,131,122,237,218,93,24,125,15,209,243,
24,120,49,5,91,227,43,82,98,109,236,243,
174,114,138,252,236,239,60,146,36,178,27,178,
207,140,13,83,151,247,184,213,253,99,54,215,
199,200,149,222,164,56,119,142,195,174,207,157,
173,22,235,159,238,226,6,22,23,94,92,252,
220,3,104,62,111,241,218,151,207,160,170,153,
218,237,181,143,183,30,1,237,6,166,131,146,
166,252,44,31,109,47,115,85,179,244,167,119,
21,150,210,195,81,85,250,251,125,162,40,37,
138,78,214,65,166,95,195,207,47,128,123,159,
144,185,203,89,98,180,99,85,129,207,58,214,
215,93,102,102,108,42,149,46,154,72,216,104,
79,179,52,12,73,86,231,81,247,142,80,218,
61,210,197,105,0,10,214,144,163,254,201,2,
97,115,35,166,120,234,2,113,252,215,35,152,
22,44,99,64,243,218,255,134,109,135,204,44,
56,252,224,131,128,173,15,29,172,210,52,165,
153,115,224,254,53,95,28,249,93,244,124,135,
138,227,113,163,59,192,245,82,148,187,47,176,
120,225,95,160,140,22,23,42,211,175,227,119,
231,24,59,245,197,167,194,122,94,144,48,8,
18,10,150,134,99,62,254,218,246,134,49,187,
237,33,227,121,227,169,28,56,93,21,143,237,
36,158,59,81,220,63,161,130,162,8,158,103,
65,176,182,249,87,148,210,247,56,155,238,19,
215,123,36,82,96,79,230,232,247,109,220,208,
96,34,72,40,199,45,204,225,128,217,137,28,
102,201,97,99,166,202,39,215,230,41,21,191,
192,219,7,155,196,97,192,212,242,47,162,89,
69,134,157,3,10,147,223,228,251,219,45,222,
252,114,133,234,159,188,75,247,230,255,132,124,
227,191,165,114,225,191,224,214,238,59,236,180,
255,156,217,217,53,26,71,30,97,172,208,106,
133,248,65,74,239,166,199,151,206,238,19,6,
9,110,39,100,113,57,207,198,248,121,242,214,
243,13,15,243,51,223,228,246,80,161,112,241,
45,198,118,62,161,217,138,56,106,73,138,191,
255,46,95,88,44,34,207,78,114,173,122,72,
224,182,142,189,188,167,87,254,9,91,183,15,
80,79,239,131,128,49,67,199,214,20,116,85,
37,1,196,76,131,148,198,241,143,146,26,123,
140,219,5,230,10,130,129,155,146,43,216,44,
56,38,63,106,246,152,159,26,195,104,247,209,
223,250,8,20,149,56,76,105,30,13,112,221,
24,207,141,144,8,52,109,72,154,74,6,94,
132,42,35,34,140,71,100,55,146,68,178,177,
113,146,24,56,61,109,97,24,42,170,42,72,
83,201,225,225,144,32,72,81,73,136,48,80,
72,73,30,232,119,20,36,218,103,108,70,101,
14,34,25,50,108,224,63,113,134,113,207,34,
85,74,50,53,77,85,33,15,188,57,89,230,
195,118,159,187,253,225,177,51,32,65,196,192,
13,16,97,130,86,206,211,47,142,177,211,214,
240,142,10,236,149,219,236,200,115,168,227,255,
33,142,243,40,171,184,91,255,152,156,115,51,
75,172,186,65,119,55,1,138,72,179,196,245,
189,30,59,187,29,214,110,238,97,38,17,99,
83,147,104,186,242,185,7,80,49,146,125,1,
80,53,133,203,175,62,234,157,157,243,61,180,
36,166,95,154,254,153,78,22,13,55,100,190,
242,116,105,15,1,156,110,180,217,127,66,78,
25,28,236,16,21,22,232,229,46,49,244,2,
20,18,170,28,96,227,18,61,80,196,212,153,
39,192,33,65,195,194,99,138,173,12,105,242,
19,54,54,92,6,3,139,249,121,72,43,69,
238,120,167,73,10,57,206,124,97,247,216,140,
234,120,25,37,231,114,212,47,49,89,200,22,
30,226,72,160,153,57,30,48,196,228,23,127,
181,200,247,127,160,51,61,22,243,206,247,6,
52,142,84,114,213,152,165,243,155,248,157,77,
190,250,79,237,71,238,247,97,146,192,193,52,
147,139,223,60,78,18,0,185,202,2,185,202,
194,19,175,143,31,37,212,122,1,141,126,72,
242,192,40,160,104,107,204,148,173,99,234,193,
253,46,65,197,15,19,54,27,131,167,94,251,
206,32,226,211,221,30,115,149,204,231,163,209,
15,56,61,153,63,30,225,61,87,162,24,132,
9,183,14,220,236,67,62,225,217,169,117,3,
28,67,101,236,1,134,95,20,213,121,249,172,
143,120,175,65,179,227,83,174,24,12,182,251,
48,72,168,106,10,19,43,21,134,157,33,141,
110,72,255,40,133,164,69,186,98,48,127,249,
63,26,121,253,190,56,10,32,146,192,109,178,
125,227,127,103,225,236,111,227,76,253,14,31,
126,242,111,120,233,171,231,41,252,219,247,56,
188,245,63,83,188,240,175,169,204,191,65,58,
125,133,53,245,127,225,202,228,251,220,189,217,
199,212,21,116,219,32,242,2,54,214,122,156,
57,87,96,242,116,145,232,141,23,16,3,241,
216,224,149,198,33,73,232,163,59,143,135,30,
172,242,121,226,177,51,20,213,255,134,197,217,
38,141,122,192,81,87,178,123,171,7,55,123,
92,57,159,231,253,241,63,101,42,255,159,102,
1,85,51,152,153,255,13,14,26,255,35,233,
64,167,181,36,40,229,99,194,68,133,72,67,
139,38,145,113,74,236,212,16,90,134,197,127,
212,239,113,24,27,156,179,242,132,129,196,52,
21,190,58,89,30,13,227,114,12,214,107,108,
213,98,134,253,33,50,58,153,4,238,42,203,
184,162,192,164,168,81,146,109,98,180,71,18,
197,35,223,201,82,153,154,178,209,52,65,24,
166,52,26,1,81,36,71,187,31,113,182,137,
139,50,154,89,140,176,205,113,155,24,139,122,
35,192,32,66,2,49,250,113,141,167,19,161,
60,7,210,94,42,233,32,64,123,239,58,225,
43,23,120,175,221,231,200,15,249,210,120,233,
120,81,32,253,112,147,198,158,71,211,85,56,
243,138,193,213,254,52,148,255,25,83,149,9,
118,251,13,242,203,139,39,230,67,199,85,86,
237,6,157,221,15,208,150,167,248,206,123,93,
132,158,98,155,191,133,54,63,201,90,211,99,
235,238,1,187,91,13,144,96,231,12,86,207,
76,253,131,5,88,215,202,163,72,201,207,250,
209,236,63,91,162,8,163,60,250,198,30,165,
146,78,163,17,176,207,42,179,172,159,44,84,
251,59,140,177,131,68,144,32,168,51,207,144,
60,99,220,71,2,38,184,111,20,54,160,200,
22,23,152,102,19,115,4,119,118,187,17,115,
115,80,234,239,83,234,239,211,154,57,75,195,
94,96,60,127,114,43,175,236,120,220,174,207,
28,39,10,93,75,57,188,241,247,172,76,157,
76,40,187,215,59,244,231,78,17,249,93,114,
249,22,111,188,38,41,18,34,23,12,228,3,
179,196,56,29,109,4,10,129,72,159,125,238,
16,196,41,135,93,143,118,160,60,177,115,8,
227,1,151,23,78,66,148,170,34,152,44,154,
236,119,252,103,122,31,63,74,89,63,242,142,
147,204,131,205,197,51,39,10,215,143,185,117,
232,158,200,100,79,106,139,214,143,60,106,221,
128,115,51,25,54,105,170,11,12,235,87,217,
221,140,88,94,204,83,116,36,197,130,78,191,
23,179,118,167,71,16,183,152,254,245,87,153,
245,67,172,254,128,52,133,191,223,157,160,248,
144,33,188,16,2,221,42,34,164,133,80,84,
12,167,194,48,254,45,126,120,251,15,121,229,
194,4,167,213,247,184,245,255,177,247,94,49,
150,165,217,149,222,119,188,187,222,135,247,233,
179,124,177,170,93,117,23,155,84,179,73,14,
135,51,2,134,28,202,140,68,81,2,8,185,
39,189,11,16,160,103,1,210,60,233,65,128,
44,49,131,33,217,195,33,57,77,211,44,182,
171,234,114,89,85,233,51,34,195,187,123,227,
122,119,252,57,122,184,145,97,50,50,179,50,
171,187,89,36,155,63,80,64,33,17,113,239,
137,115,254,179,215,191,215,94,123,237,123,187,
212,82,255,140,204,248,115,228,46,254,151,188,
123,203,224,235,47,125,151,155,239,187,92,88,
16,9,66,149,74,90,165,103,131,31,122,4,
67,133,162,46,176,226,221,2,94,29,221,176,
206,30,94,253,223,49,169,108,161,7,14,61,
79,166,149,152,32,52,222,196,202,63,52,44,
36,24,210,12,12,252,218,40,0,95,89,50,
64,128,90,221,99,101,217,227,124,226,35,238,
220,47,147,41,188,124,72,65,101,241,63,202,
80,154,255,121,90,141,63,225,74,193,96,67,
140,16,156,34,65,114,212,152,39,182,74,196,
217,218,137,13,22,163,235,2,189,70,68,174,
56,218,164,131,94,132,180,213,225,206,178,141,
23,136,232,143,0,128,233,104,147,0,17,137,
145,243,171,121,194,119,179,80,208,208,52,9,
223,143,176,237,0,199,137,8,130,136,76,70,
33,8,34,186,221,240,104,94,68,191,31,28,
117,83,11,112,168,85,58,241,66,165,36,210,
105,153,208,182,25,250,42,178,4,162,253,217,
106,25,227,227,135,147,203,106,45,194,115,211,
220,233,13,169,57,30,162,32,144,215,70,39,
38,113,239,128,187,213,28,254,160,138,46,121,
12,4,141,64,121,131,76,126,228,121,160,232,
143,175,41,136,146,137,160,137,164,74,191,142,
172,89,12,90,235,244,165,18,107,7,67,150,
111,239,112,80,29,5,11,43,161,179,120,190,
136,36,75,159,91,128,77,218,61,148,208,167,
159,42,253,157,6,10,55,136,232,218,1,41,
227,201,225,198,38,69,172,169,132,161,75,77,
57,79,217,191,255,132,236,35,166,35,228,105,
138,57,46,135,31,61,158,50,167,203,12,93,
26,140,115,192,36,10,46,175,143,53,15,69,
143,2,118,164,83,167,72,171,89,56,3,20,
163,120,118,188,215,191,240,166,197,39,239,254,
144,165,43,199,170,160,94,59,66,51,36,76,
105,141,215,191,105,162,234,38,242,187,55,15,
251,154,18,8,182,67,184,52,205,182,144,165,
29,53,248,246,110,3,127,40,16,58,62,78,
183,250,72,47,49,24,121,235,217,126,136,31,
70,108,215,7,56,145,114,202,91,239,81,65,
254,218,70,135,28,30,51,39,146,104,237,51,
102,195,41,93,121,118,234,169,239,62,29,72,
60,252,59,181,174,75,26,48,211,51,124,164,
106,188,249,213,113,250,27,93,32,160,211,245,
217,218,24,64,12,157,166,135,252,173,107,216,
95,191,200,221,246,56,138,119,142,226,249,111,
60,250,69,151,21,146,185,37,124,183,135,65,
5,35,53,73,104,254,215,188,179,254,175,121,
158,119,185,234,174,115,171,253,255,112,127,112,
131,210,212,175,33,89,75,180,131,183,25,120,
34,93,71,162,52,166,211,156,156,100,53,212,
81,7,18,201,26,184,4,132,129,193,230,135,
191,199,108,101,157,10,29,244,246,54,7,155,
93,108,224,210,185,36,145,222,231,192,104,178,
179,241,45,156,56,71,44,11,40,110,151,178,
187,131,81,223,226,126,53,6,68,170,109,7,
221,210,153,152,205,114,241,106,18,105,175,65,
42,251,9,190,123,17,157,209,140,14,51,59,
3,49,184,225,0,199,77,146,8,84,218,198,
113,209,44,206,214,144,6,147,132,230,54,8,
163,49,168,138,40,96,117,234,244,170,49,209,
88,1,83,14,96,208,35,12,32,147,20,145,
229,4,189,182,141,16,135,167,42,224,17,50,
17,226,209,220,136,7,32,49,51,243,232,194,
149,231,69,172,175,15,232,245,124,26,13,133,
68,66,70,211,68,122,189,199,63,235,189,189,
209,73,109,98,194,192,182,67,246,247,157,103,
218,148,134,49,2,172,114,217,160,92,49,136,
198,242,68,211,21,250,154,202,253,219,67,72,
10,20,30,20,220,93,159,143,223,211,201,132,
85,66,201,35,145,55,184,30,94,197,200,79,
61,213,119,89,217,9,226,129,201,214,39,127,
200,226,151,254,11,204,226,121,238,172,183,185,
119,107,135,122,173,139,149,48,152,156,205,147,
47,38,17,236,214,231,26,96,59,86,250,239,
124,141,226,168,62,80,235,179,84,73,60,113,
54,179,16,107,56,153,4,146,154,99,186,146,
197,121,231,238,147,51,207,184,73,67,172,16,
132,34,242,137,195,75,140,112,166,102,145,103,
247,72,72,113,223,57,71,99,168,35,138,163,
78,158,165,222,53,110,231,243,143,0,9,129,
149,158,203,249,67,156,22,69,120,225,245,211,
210,209,100,70,228,215,255,163,227,236,64,104,
118,17,250,67,226,76,130,240,226,44,161,235,
83,29,216,252,160,30,162,53,98,180,240,101,
146,230,56,153,203,47,226,245,207,238,175,142,
237,179,94,31,158,158,115,237,69,79,21,169,
253,48,194,143,35,234,61,143,40,142,41,36,
213,199,214,28,62,21,40,30,2,245,167,178,
25,95,175,13,9,253,19,212,129,111,143,218,
8,63,197,68,205,17,53,180,184,135,32,65,
41,140,233,173,54,25,12,92,118,219,14,185,
140,198,249,139,41,54,55,134,180,219,30,253,
174,7,27,17,150,246,13,180,108,14,167,251,
232,107,177,123,251,12,147,239,17,238,59,200,
234,177,154,194,200,189,201,199,245,20,47,202,
255,158,25,165,195,61,127,135,181,15,255,111,
146,185,11,72,9,29,133,0,13,159,79,210,
231,168,238,77,144,46,190,129,144,179,168,214,
110,19,250,55,80,141,29,46,206,180,153,115,
235,52,150,107,172,111,13,137,162,17,213,115,
255,126,159,113,59,68,232,237,98,52,60,244,
56,70,81,68,130,32,166,57,60,156,135,12,
184,168,224,122,244,93,159,29,207,70,127,105,
134,221,177,12,74,228,113,240,193,187,72,135,
124,164,39,109,209,210,223,67,78,199,52,5,
151,48,219,133,135,188,180,66,115,27,161,91,
36,78,31,28,217,139,155,221,22,253,59,53,
212,123,107,40,73,133,190,35,48,55,103,226,
5,163,23,99,48,144,136,189,144,224,16,28,
34,68,116,206,6,237,74,229,241,52,128,170,
138,44,45,37,177,237,16,73,18,168,86,237,
83,246,224,143,124,38,118,200,234,234,227,13,
16,183,56,119,244,226,202,248,84,140,22,216,
163,116,222,60,156,143,172,170,34,254,88,158,
193,252,4,154,169,211,176,3,126,244,3,21,
75,249,10,195,238,95,113,249,53,153,239,237,
196,36,247,82,92,53,110,176,187,217,166,116,
169,204,181,104,17,53,243,11,56,157,131,71,
90,130,159,57,221,14,90,60,247,149,33,141,
166,203,160,177,203,208,139,104,172,237,97,31,
212,25,203,152,204,47,89,32,56,48,112,158,
222,182,27,158,237,103,237,22,79,35,53,206,
212,55,81,194,128,161,24,255,100,175,225,167,
96,51,142,59,24,125,158,40,61,166,78,5,
183,186,85,36,65,24,81,152,113,136,32,27,
40,135,99,57,45,77,34,246,52,118,36,147,
180,22,176,151,201,144,19,4,226,39,80,111,
50,17,231,252,27,199,7,22,230,113,48,153,
96,5,149,71,31,86,20,60,158,43,110,163,
156,56,105,199,162,69,198,24,112,208,79,178,
246,141,8,0,0,32,0,73,68,65,84,81,
60,145,85,124,180,151,230,98,185,198,83,248,
246,142,222,107,87,196,186,179,62,250,204,225,
104,30,252,95,182,122,120,81,140,218,109,145,
55,254,249,209,104,93,187,181,143,63,236,30,
237,217,56,134,253,142,67,99,224,61,250,222,
74,234,167,198,91,128,150,221,162,57,28,125,
198,38,194,136,62,122,220,61,116,186,16,184,
103,158,153,36,8,200,182,195,208,121,8,40,
62,205,102,188,36,187,108,52,134,199,223,231,
13,70,246,195,242,147,85,48,86,206,68,15,
52,218,141,247,89,244,247,233,153,22,233,241,
52,197,192,69,117,92,162,40,70,18,65,211,
69,22,190,190,200,119,26,11,148,39,47,63,
185,78,210,189,143,100,249,232,78,249,204,181,
154,185,95,229,250,90,64,162,253,30,149,139,
191,139,56,39,1,2,239,175,93,227,197,185,
61,86,86,122,24,197,128,241,139,15,106,31,
224,119,191,207,215,38,183,16,55,246,216,249,
96,200,199,7,46,166,41,163,170,163,193,238,
174,27,225,186,222,200,214,226,112,53,169,144,
179,247,207,92,155,134,71,143,4,38,54,182,
43,176,208,232,211,201,100,89,91,182,40,205,
124,241,232,122,243,193,215,104,69,127,132,33,
15,25,4,33,87,51,9,222,62,232,156,14,
29,2,160,217,196,174,204,237,206,128,115,41,
147,236,149,5,74,221,1,239,95,27,16,6,
54,234,137,2,242,48,148,232,28,84,25,70,
50,165,164,66,50,249,120,169,219,242,114,151,
98,81,167,84,210,207,244,155,133,97,140,227,
140,230,23,219,246,72,182,42,203,2,65,240,
217,184,242,58,19,148,216,58,226,135,197,100,
1,209,200,18,28,2,197,112,24,176,178,210,
99,242,23,207,241,86,218,34,51,112,40,5,
17,247,62,200,51,121,254,183,105,108,190,77,
118,214,225,123,31,89,244,219,99,4,233,29,
180,130,206,204,204,2,223,222,187,68,126,225,
183,144,85,3,167,115,240,216,52,254,212,75,
160,170,40,130,72,165,16,209,244,85,66,57,
65,195,110,49,136,83,204,207,207,142,156,120,
143,211,215,167,179,216,126,214,159,69,0,235,
211,175,181,133,140,24,71,96,21,126,178,215,
240,83,178,25,71,53,71,63,255,132,21,2,
97,224,142,192,71,181,112,128,129,35,112,49,
151,164,146,171,176,178,114,147,175,37,238,177,
17,198,44,78,38,105,108,117,159,122,175,5,
200,71,114,217,135,87,12,56,36,112,48,185,
127,191,201,236,172,133,126,104,0,24,91,6,
134,226,113,109,123,142,110,218,192,13,21,46,
85,182,9,197,1,43,77,29,85,180,145,5,
129,94,16,208,15,66,46,167,45,212,135,6,
88,220,222,159,224,124,247,61,98,219,29,9,
125,94,187,196,150,227,225,69,49,209,80,66,
115,46,147,58,183,120,156,173,132,1,146,162,
161,167,139,4,97,204,114,181,79,15,29,172,
199,220,91,89,251,212,120,251,240,222,138,158,
230,217,106,169,83,207,108,50,103,96,40,18,
150,245,25,168,167,82,74,195,246,67,170,157,
103,227,157,147,186,76,220,7,55,220,100,227,
249,57,198,15,134,232,59,219,200,190,71,16,
196,135,150,17,163,14,226,40,10,33,254,244,
52,91,179,198,112,107,47,147,120,224,193,251,
112,58,58,247,143,169,191,123,19,203,233,97,
100,70,129,89,144,94,227,78,37,226,178,118,
141,228,76,204,95,44,255,30,149,115,255,2,
223,29,224,244,92,194,165,20,226,198,30,65,
16,19,134,241,17,31,255,216,3,25,42,1,
42,138,224,157,1,107,29,143,24,145,72,210,
168,127,178,199,76,177,79,75,249,202,41,117,
67,186,116,137,254,202,93,204,242,199,252,220,
205,251,68,150,73,105,170,76,245,33,215,216,
88,239,163,134,38,179,24,24,65,140,32,248,
116,26,14,185,172,66,183,110,159,254,238,192,
70,79,88,12,235,117,90,157,39,3,133,235,
70,108,111,15,169,215,93,102,103,45,12,67,
162,209,240,168,213,28,156,195,121,232,178,44,
144,205,170,156,59,151,68,215,37,154,77,143,
173,173,193,51,3,134,143,118,4,18,0,142,
53,73,209,190,73,231,4,237,52,243,210,24,
239,230,83,16,132,52,111,149,177,213,57,50,
217,89,68,73,70,209,82,12,214,126,142,242,
228,215,24,159,210,217,191,255,39,236,95,116,
169,220,94,67,17,123,136,210,179,213,16,134,
157,85,46,95,144,184,113,31,148,92,138,97,
203,199,177,125,20,69,38,149,54,255,86,209,
53,217,65,11,37,240,217,74,20,248,251,188,
162,56,102,163,62,228,210,68,18,159,89,182,
74,7,4,13,5,251,165,41,196,157,91,79,
213,236,27,33,142,166,53,62,62,124,142,176,
17,135,237,120,1,49,78,194,80,64,18,67,
132,93,25,51,233,242,229,133,59,8,66,76,
20,137,172,214,75,228,101,248,104,215,192,147,
62,56,189,135,252,24,193,157,165,98,200,136,
66,140,227,171,44,245,174,33,111,108,115,239,
126,159,197,249,4,110,103,192,109,111,148,45,
68,181,52,19,75,255,228,44,104,198,163,49,
210,7,61,247,153,104,253,159,244,58,52,169,
24,213,8,31,227,207,245,84,64,17,199,96,
123,207,230,152,169,72,2,134,42,49,4,148,
104,28,69,188,71,148,207,32,239,108,141,46,
78,132,251,171,125,250,93,159,201,41,19,243,
66,154,184,254,233,141,69,201,226,2,201,226,
2,78,231,224,56,70,122,54,113,224,35,41,
26,97,232,211,177,11,20,50,99,39,50,155,
57,60,39,197,191,95,143,145,222,137,81,18,
117,118,110,252,49,227,23,191,129,170,249,236,
86,99,102,56,238,62,87,85,241,140,77,182,
44,11,40,170,132,61,12,152,82,183,137,38,
95,98,42,190,195,242,182,128,26,246,143,126,
55,102,164,242,193,233,80,189,15,86,181,131,
246,66,243,17,247,39,199,48,132,1,26,246,
70,151,197,78,68,225,114,133,187,126,120,106,
30,197,213,162,68,217,16,240,92,104,181,32,
253,242,2,90,199,161,180,54,146,181,186,129,
132,131,138,165,65,232,246,176,123,29,58,162,
70,177,144,198,208,159,172,52,113,156,144,59,
119,186,72,39,38,134,29,29,14,74,58,99,
99,199,191,159,201,40,120,158,193,246,86,159,
102,187,77,62,151,251,76,155,82,55,85,122,
181,0,93,151,168,84,116,178,211,105,222,93,
156,162,237,28,190,228,86,159,100,242,210,81,
243,97,102,252,121,224,249,163,141,216,111,175,
241,118,115,200,243,179,115,92,178,93,62,186,
255,251,84,206,253,230,211,5,164,48,64,229,
71,136,170,68,253,96,150,201,146,74,223,182,
25,12,92,82,79,49,153,237,111,122,181,172,
236,40,163,248,25,88,125,55,96,224,6,148,
166,127,129,15,150,119,232,153,105,110,165,45,
94,157,77,115,176,218,126,10,160,24,89,194,
60,177,22,198,136,30,205,248,125,150,244,204,
40,64,154,58,209,88,129,221,100,26,65,104,
31,214,34,34,230,11,35,49,73,7,248,164,
94,38,159,172,82,239,140,17,35,80,141,36,
206,149,54,152,47,36,33,142,17,90,61,4,
53,68,148,96,97,193,66,184,48,205,71,130,
64,16,199,184,247,76,52,53,125,198,42,198,
11,34,238,87,7,56,138,252,185,222,119,89,
18,184,50,157,166,239,4,108,53,237,99,139,
163,103,5,138,24,184,127,48,160,107,63,153,
167,84,101,145,115,149,4,125,39,160,107,7,
164,205,227,143,54,18,227,40,254,50,205,161,
136,114,101,30,221,15,16,151,55,9,130,62,
162,40,96,167,243,188,127,127,140,100,241,75,
79,151,98,58,125,122,245,21,180,100,14,226,
152,221,155,127,132,43,175,34,167,5,116,251,
57,102,191,240,159,156,221,36,233,113,244,243,
62,162,25,146,11,127,131,100,113,30,65,148,
200,158,251,207,217,220,248,95,25,63,63,71,
162,113,23,81,20,24,31,55,88,91,27,21,
116,19,9,133,252,82,129,244,229,9,130,205,
26,181,91,85,170,214,75,92,72,239,147,146,
53,46,164,147,236,235,47,35,236,175,49,216,
220,60,147,239,105,42,76,120,43,172,86,255,
13,138,245,219,71,117,10,89,144,56,143,133,
119,208,166,177,63,196,219,104,195,237,3,94,
57,87,38,120,97,140,186,44,146,144,37,198,
31,140,74,141,99,228,40,160,239,202,160,154,
104,137,4,227,179,42,27,171,93,148,200,39,
66,98,24,138,200,86,134,48,22,185,119,111,
153,133,185,121,18,9,235,83,45,180,195,32,
194,15,67,20,249,248,153,237,238,218,52,155,
30,229,178,78,62,175,33,138,2,149,138,206,
96,16,144,206,148,232,118,71,251,33,20,20,
2,65,70,139,236,135,104,147,18,54,9,114,
15,117,203,138,113,64,4,92,184,144,26,1,
148,161,179,239,28,159,4,187,73,133,254,206,
159,51,102,92,101,208,252,1,94,232,161,170,
211,100,39,94,65,148,100,102,175,252,54,237,
253,15,184,230,127,192,87,94,245,41,180,182,
136,63,37,152,214,215,222,33,85,92,162,187,
255,175,248,202,107,1,203,221,33,102,122,130,
40,142,169,53,250,68,97,132,149,252,219,55,
52,232,103,37,163,120,16,104,90,67,159,201,
172,65,214,120,17,51,255,39,108,181,102,88,
127,197,39,183,243,9,174,27,62,145,114,186,
45,94,38,29,13,25,146,196,164,71,128,66,
139,242,41,169,236,201,19,244,131,87,34,154,
25,35,26,43,208,173,153,84,56,13,72,126,
20,211,139,58,168,114,130,237,218,60,186,110,
51,145,217,103,33,105,48,105,88,136,219,85,
196,221,3,132,163,209,0,2,82,54,201,70,
62,75,245,208,41,86,30,76,144,41,190,122,
166,232,124,123,183,135,31,132,15,252,6,63,
183,53,155,183,80,37,145,156,165,146,53,213,
199,134,138,79,5,138,174,237,211,236,123,159,
250,133,41,67,198,84,37,76,85,162,148,58,
205,165,101,39,94,230,222,138,76,24,12,249,
196,252,62,95,158,132,208,200,49,188,52,134,
77,137,70,250,231,41,84,174,50,108,238,19,
199,17,221,234,45,90,173,31,98,25,58,126,
172,160,160,97,10,7,168,81,147,190,172,35,
137,1,202,148,205,246,221,77,166,46,254,6,
211,47,254,51,156,238,30,245,253,183,208,147,
227,72,170,249,8,202,42,71,170,241,38,253,
221,219,164,158,91,58,13,114,99,191,195,15,
118,255,148,244,188,201,108,218,230,131,154,197,
248,23,251,148,157,22,44,78,240,193,114,1,
127,77,101,62,173,81,185,170,209,74,78,16,
79,36,70,244,73,24,81,140,225,70,242,85,
150,242,50,157,181,26,173,230,177,213,69,171,
229,97,221,90,230,124,110,151,122,127,25,33,
151,1,65,96,70,14,208,223,93,166,211,114,
57,127,62,69,173,230,176,181,53,164,250,241,
58,147,246,128,197,151,151,208,199,116,196,131,
22,177,166,226,217,34,194,123,159,208,58,112,
70,93,170,154,200,193,129,139,16,197,40,140,
58,158,147,169,12,146,32,16,34,210,244,187,
220,189,119,23,81,18,25,175,140,81,170,84,
30,217,43,226,185,46,7,7,7,12,6,67,
18,169,4,227,149,241,35,43,220,122,163,139,
227,132,212,106,14,149,138,65,16,68,36,147,
50,197,162,206,242,114,143,94,207,199,214,179,
12,180,60,197,246,29,196,195,42,203,144,20,
30,58,227,172,158,5,37,199,38,157,82,142,
38,219,133,226,232,183,194,72,166,215,155,230,
82,177,133,82,222,195,236,239,83,200,168,124,
116,71,165,55,112,49,186,147,152,217,73,100,
205,32,85,124,142,238,198,219,12,131,24,53,
25,227,12,31,207,99,199,97,64,183,246,67,
114,201,31,240,230,151,100,118,135,30,247,246,
92,202,230,24,205,190,207,254,238,40,56,72,
146,248,183,46,118,254,44,101,20,8,199,253,
22,133,169,159,99,231,254,53,126,105,182,193,
119,123,19,76,191,222,193,253,235,213,163,32,
111,89,50,115,115,9,86,86,122,108,219,101,
6,130,197,249,232,14,42,14,45,202,212,153,
32,207,46,193,209,120,223,211,135,220,66,65,
59,77,151,0,219,3,23,109,224,48,99,233,
71,255,252,81,171,71,215,15,80,101,135,116,
178,137,237,36,121,173,144,34,173,200,68,55,
215,136,182,106,72,170,200,112,16,18,11,49,
86,214,160,125,110,154,107,173,19,18,193,116,
151,84,241,226,169,239,223,168,219,248,225,231,
255,92,45,85,34,115,162,22,241,164,243,228,
167,2,69,218,80,88,42,39,88,171,15,78,
141,112,60,251,165,50,125,39,160,214,27,57,
180,198,128,42,9,20,132,209,236,225,220,212,
168,143,160,189,151,228,71,215,63,34,145,253,
77,204,249,105,52,69,67,99,100,233,220,169,
221,160,213,254,215,204,79,181,249,98,162,139,
182,215,32,86,100,68,215,135,224,144,239,147,
21,28,213,224,32,49,206,199,250,241,245,232,
169,49,10,252,34,122,234,241,167,175,194,244,
235,20,166,94,59,123,19,84,131,228,244,63,
37,140,66,222,91,253,14,249,133,95,96,167,
95,231,174,243,33,194,138,132,89,124,5,213,
115,72,68,255,27,230,116,134,137,68,196,118,
127,146,133,226,14,53,39,196,114,100,210,115,
251,124,152,122,129,95,44,124,194,245,183,118,
137,93,15,129,152,56,134,126,63,160,223,239,
0,29,68,81,64,20,161,31,198,200,25,149,
153,153,145,125,198,254,190,115,84,52,174,174,
53,40,4,49,230,139,99,176,186,71,183,15,
181,157,46,131,158,251,152,2,161,130,134,199,
206,254,38,149,202,24,177,164,147,49,22,89,
190,119,27,215,243,216,222,217,193,243,60,166,
166,167,33,6,223,247,240,195,128,65,175,207,
230,246,246,209,203,210,235,247,112,28,7,207,
113,113,28,135,48,142,168,148,43,148,42,101,
182,223,171,145,47,140,36,132,59,59,199,217,
67,194,174,145,176,107,135,89,68,25,155,4,
34,33,21,214,207,210,134,73,5,33,173,80,
176,180,163,224,112,47,151,130,24,246,91,83,
252,242,226,1,51,59,53,164,122,139,97,49,
207,95,93,123,129,242,210,111,146,125,200,69,
81,53,83,8,170,204,71,173,38,11,134,137,
219,104,160,168,169,51,89,103,191,185,142,211,
95,231,151,127,69,68,37,70,190,126,159,131,
74,153,168,174,98,44,150,185,187,217,163,94,
29,85,75,170,123,109,42,19,89,148,207,153,
10,248,153,205,40,24,245,91,180,6,62,89,
75,65,215,95,38,21,254,41,63,31,249,220,
153,46,176,88,216,163,93,183,41,20,116,116,
93,100,221,159,163,109,249,232,246,22,106,108,
31,169,156,178,84,143,60,159,154,84,30,57,
120,171,209,112,71,180,170,34,19,235,163,189,
152,48,219,188,181,153,226,11,19,54,154,0,
195,200,161,122,152,233,182,250,69,42,217,45,
134,110,130,251,61,155,87,134,67,66,59,228,
158,112,153,206,106,141,43,249,125,146,105,149,
224,185,115,220,120,184,206,168,12,8,125,247,
200,70,166,53,240,105,14,188,191,21,247,59,
103,61,253,204,147,167,122,43,178,150,130,165,
165,184,177,211,125,44,88,60,208,221,202,162,
192,254,137,162,183,101,133,156,60,223,107,137,
10,253,126,19,103,240,54,174,189,66,20,171,
196,97,132,169,126,200,23,115,53,132,213,125,
156,53,7,253,80,194,25,251,33,113,57,139,
176,63,114,114,180,237,24,147,1,59,143,40,
254,168,86,230,169,42,55,237,237,27,84,151,
191,139,102,229,168,92,248,58,122,170,120,88,
55,145,48,114,87,64,16,80,204,44,189,149,
22,182,189,201,164,53,135,128,130,58,28,32,
118,186,228,51,29,18,197,18,239,108,206,97,
7,117,222,156,13,137,122,49,77,47,66,184,
58,199,244,106,131,149,101,239,148,42,233,193,
154,156,52,209,180,145,180,214,178,228,35,11,
141,40,130,203,151,211,68,209,161,234,40,130,
205,239,174,161,42,18,205,106,247,76,205,228,
97,134,54,66,96,114,98,242,168,132,236,105,
38,11,151,95,98,229,198,7,120,129,79,237,
224,128,70,179,73,28,197,167,154,136,206,156,
98,155,45,178,185,44,169,116,134,65,242,10,
3,81,101,95,52,233,43,93,196,214,50,153,
116,138,232,176,45,191,63,232,147,48,19,71,
149,194,7,47,233,46,243,68,72,136,135,70,
47,19,19,38,157,142,199,236,197,44,235,75,
151,81,67,131,157,186,194,32,239,179,28,7,
84,91,83,148,210,59,220,238,4,168,165,28,
165,74,142,239,124,156,162,178,248,27,71,3,
92,78,241,217,245,53,72,143,2,252,253,93,
145,98,242,184,190,208,220,248,54,150,242,9,
169,36,204,44,8,180,90,17,94,44,35,123,
62,155,74,146,149,183,50,8,42,184,145,136,
227,199,204,204,21,89,95,173,33,136,2,161,
31,161,28,30,176,226,40,166,81,235,81,109,
247,200,228,18,140,79,230,158,218,206,121,109,
165,138,231,250,84,198,51,40,154,130,99,251,
12,250,54,165,114,6,77,127,122,174,225,103,
42,163,56,92,219,45,155,140,165,144,159,124,
149,235,247,63,228,75,133,251,36,111,55,232,
8,35,87,227,124,94,197,178,100,186,67,200,
12,110,99,63,198,76,200,193,66,198,103,131,
139,148,216,194,160,79,140,72,85,90,228,203,
179,163,250,102,56,55,113,52,225,238,74,193,
163,100,9,180,6,105,182,221,136,93,91,34,
147,176,105,247,11,164,205,81,141,209,113,45,
178,202,144,120,204,228,206,240,50,207,119,254,
132,120,65,66,200,228,136,198,139,196,9,131,
200,57,126,231,99,31,52,111,233,8,36,28,
63,100,245,96,240,183,226,62,171,178,248,76,
227,106,207,0,69,20,199,244,156,128,158,29,
224,248,17,94,112,248,223,19,82,165,148,113,
60,136,72,126,40,133,63,213,56,66,76,109,
247,143,89,186,50,100,233,147,191,162,223,114,
200,77,103,168,14,69,6,183,118,216,26,4,
4,135,63,95,219,119,80,117,145,114,201,64,
187,156,97,171,144,97,118,224,16,45,215,136,
211,22,97,8,50,167,79,90,162,244,116,167,
193,204,228,21,50,147,87,158,92,28,139,35,
198,150,190,65,28,71,104,201,60,141,213,111,
209,221,106,34,41,54,25,219,69,139,99,94,
157,82,216,238,167,249,235,45,131,148,5,110,
28,112,183,27,48,185,56,142,178,124,235,209,
167,38,55,36,147,81,112,156,16,89,22,16,
69,129,197,197,36,43,43,125,86,87,251,4,
65,196,220,92,2,199,139,104,238,247,158,142,
222,21,68,198,139,18,61,109,138,253,186,68,
218,94,37,138,34,156,64,224,194,133,11,212,
235,117,118,247,246,8,195,167,19,36,232,186,
193,182,116,133,176,235,145,77,130,17,5,200,
138,203,138,252,75,164,6,171,44,37,235,136,
68,104,154,206,246,238,14,153,76,134,132,53,
82,89,185,24,12,73,177,207,12,2,163,249,
22,170,154,68,159,81,88,203,20,104,172,218,
12,58,80,214,247,241,196,89,48,50,100,18,
251,40,178,135,23,193,123,125,27,75,150,112,
162,248,177,185,176,172,154,132,171,57,204,238,
69,172,204,2,70,122,156,97,99,151,198,189,
255,139,241,137,93,186,3,137,251,43,73,150,
87,5,222,120,179,207,187,239,198,56,76,161,
137,211,204,255,220,243,12,90,85,26,3,15,
215,245,216,88,63,96,114,182,192,204,108,233,
8,240,6,61,135,219,55,182,144,220,54,195,
216,162,221,28,176,181,118,128,162,74,168,170,
66,58,107,50,49,93,120,36,93,53,28,184,
236,109,55,73,166,12,106,123,29,84,77,65,
213,20,76,83,71,146,159,141,222,250,89,203,
40,96,36,154,169,247,92,138,73,141,212,220,
239,240,221,181,63,37,109,44,147,54,110,16,
30,212,104,52,70,242,245,72,144,152,158,182,
216,220,28,224,186,209,25,101,84,147,202,145,
253,199,54,231,152,228,30,2,17,19,194,10,
53,225,121,170,182,68,208,155,102,204,13,73,
104,14,65,168,80,208,125,242,90,135,166,231,
211,169,137,84,91,147,8,49,100,18,163,158,
174,153,242,93,174,31,148,233,170,99,60,23,
188,61,58,92,92,156,37,174,228,9,227,152,
229,222,144,246,161,147,97,228,136,104,123,175,
81,94,248,133,67,10,63,96,165,214,255,220,
212,77,9,93,198,245,67,52,69,34,99,40,
20,146,42,209,160,241,217,128,162,99,251,220,
236,180,121,86,123,153,7,110,177,67,47,100,
167,117,76,75,136,130,128,121,8,32,81,224,
177,191,242,103,228,23,119,41,104,10,170,169,
144,118,93,154,43,117,122,109,159,102,55,70,
57,172,4,171,154,200,194,124,2,211,148,65,
150,113,17,184,237,135,220,211,52,114,231,95,
65,78,182,112,27,49,98,156,248,169,221,88,
73,86,145,14,253,170,226,56,162,28,127,132,
37,135,216,131,144,76,26,6,27,29,212,230,
144,197,11,147,228,242,9,254,120,93,66,150,
92,254,122,35,203,151,39,108,178,19,89,156,
70,27,81,132,225,240,56,64,215,106,14,170,
42,210,239,7,8,130,203,220,92,130,102,211,
35,142,71,210,92,207,139,88,89,233,81,158,
204,62,101,13,80,64,51,84,132,243,151,209,
207,47,242,198,234,15,249,209,119,12,26,226,
56,123,181,21,198,168,83,25,47,147,74,165,
89,89,93,33,240,63,189,121,106,111,119,23,
65,108,96,8,1,51,133,231,8,196,52,221,
65,139,146,247,45,12,69,100,53,249,117,102,
185,131,34,67,33,159,71,62,1,208,93,242,
44,242,209,33,224,72,120,94,132,191,30,35,
168,34,209,61,152,206,169,72,146,192,216,132,
73,69,134,11,83,147,124,228,196,108,157,240,
45,27,4,33,226,116,141,131,141,239,81,154,
253,234,89,32,75,149,153,191,242,187,136,135,
13,72,190,221,227,96,247,47,241,109,135,72,
248,2,129,255,49,232,117,228,232,42,31,191,
183,74,178,242,159,82,176,142,85,90,138,214,
199,15,34,214,239,87,137,163,120,68,55,29,
130,68,175,107,115,243,227,13,194,32,194,60,
252,55,81,20,177,18,26,130,36,34,34,32,
8,226,200,10,253,17,203,180,52,94,255,234,
133,207,220,17,123,114,181,173,204,223,155,206,
236,103,89,163,233,120,26,130,40,161,103,95,
70,202,253,42,123,202,119,120,101,238,15,48,
147,18,120,62,210,46,36,12,153,236,194,57,
118,91,42,91,135,116,104,134,58,73,154,71,
89,239,136,22,15,120,160,156,173,148,53,42,
214,200,36,48,204,193,158,187,192,122,199,162,
239,123,220,233,184,216,174,197,100,113,5,69,
141,145,108,139,124,234,116,191,148,168,85,217,
23,107,36,166,10,204,45,86,216,15,98,250,
189,33,155,3,7,231,240,48,29,173,78,98,
170,231,41,45,125,237,40,46,222,221,239,241,
121,218,118,229,44,245,148,171,247,40,235,226,
179,1,197,147,154,248,158,88,20,209,36,186,
118,192,253,90,255,232,247,37,81,96,190,104,
161,186,54,78,183,74,173,250,255,242,213,231,
93,178,203,91,108,252,217,1,183,122,62,81,
8,126,112,232,24,42,168,168,170,136,105,200,
204,204,154,71,153,73,240,210,5,238,132,17,
248,54,97,28,19,9,18,243,9,131,108,78,
230,237,224,45,182,238,108,160,104,25,100,49,
73,118,236,37,36,85,127,108,81,179,83,189,
133,158,28,67,79,62,219,16,154,126,245,46,
115,189,6,166,28,32,102,13,106,3,141,162,
229,34,120,1,220,90,37,243,250,85,94,31,
115,120,191,54,186,230,119,171,121,102,94,127,
145,159,107,108,146,202,202,92,255,163,219,71,
253,7,113,12,91,91,67,202,101,125,228,104,
235,132,140,143,27,52,26,46,115,115,22,123,
123,14,131,65,128,166,137,36,18,242,41,239,
253,179,32,1,177,162,243,252,69,157,181,153,
89,166,179,13,122,243,23,80,214,43,76,237,
125,130,57,102,82,219,86,184,125,243,54,158,
231,61,118,24,18,128,34,43,72,178,132,101,
90,163,128,40,8,20,242,121,34,127,136,223,
111,147,150,192,51,82,212,196,89,74,180,216,
102,137,89,110,162,235,167,239,247,3,149,137,
40,10,92,190,124,44,119,14,130,8,65,145,
225,202,60,177,169,19,136,34,177,161,209,243,
67,82,74,132,34,10,200,130,128,123,56,156,
72,80,96,32,191,203,176,179,128,153,158,60,
115,189,226,137,46,213,131,205,183,64,217,98,
246,82,204,206,71,105,116,35,38,51,254,59,
24,233,50,81,24,60,50,211,236,246,93,218,
205,17,13,144,56,84,60,121,174,207,237,235,
155,132,135,123,82,86,36,206,47,77,146,43,
36,143,38,255,125,10,106,179,186,178,71,167,
53,196,247,3,66,63,66,148,69,146,41,147,
115,151,198,145,159,209,55,42,61,232,160,132,
62,195,100,241,103,10,40,220,32,226,131,245,
54,154,34,178,112,200,91,39,39,127,158,91,
107,31,240,66,243,30,189,75,23,217,18,166,
9,15,84,38,148,77,198,42,67,230,84,239,
148,11,242,73,169,236,216,184,69,176,61,170,
255,101,179,199,188,124,44,75,148,83,109,154,
241,144,186,63,164,144,134,118,191,192,192,77,
146,208,187,24,218,224,232,255,79,46,63,138,
177,20,137,31,180,7,116,78,30,188,98,8,
54,211,148,10,255,1,137,220,177,27,240,208,
13,159,62,174,198,241,167,170,20,63,203,218,
105,217,148,82,159,221,210,67,62,73,57,213,
123,254,211,118,171,159,90,215,54,58,103,210,
156,133,146,133,38,139,12,157,152,131,189,63,
230,43,95,136,72,117,60,162,112,116,114,118,
220,8,1,72,38,100,122,174,200,220,148,69,
37,123,250,143,136,69,1,199,80,89,221,111,
33,9,2,151,210,22,249,96,72,66,29,93,
246,171,25,151,251,169,85,42,146,77,79,55,
249,240,222,15,72,136,95,164,48,117,90,102,
219,174,125,72,111,112,151,249,89,159,251,91,
10,82,253,57,178,99,95,56,106,167,255,212,
103,231,190,131,230,244,88,222,112,17,5,151,
165,69,1,47,16,209,228,8,130,16,193,11,
152,181,116,110,235,13,250,78,154,124,170,202,
206,32,199,159,101,38,153,50,36,94,249,242,
128,79,222,218,56,245,153,213,234,8,207,87,
87,251,71,125,12,130,32,48,63,159,224,198,
141,54,170,42,48,49,97,34,8,112,231,78,
247,17,5,108,25,31,153,169,108,68,99,241,
5,54,104,177,181,171,81,214,82,188,248,31,
198,196,63,42,240,201,59,7,100,103,138,12,
162,73,122,213,77,118,247,118,30,107,137,80,
44,151,144,4,129,84,42,133,102,24,248,190,
135,170,168,132,65,64,223,58,71,24,201,4,
72,84,156,13,252,104,138,138,184,206,6,23,
153,225,246,35,63,47,159,63,189,145,100,89,
28,53,84,222,90,37,60,55,67,52,54,162,
83,86,250,67,36,65,32,33,203,156,79,153,
116,253,128,59,29,155,129,107,96,84,186,28,
172,254,25,51,233,223,126,114,70,187,248,77,
106,119,87,8,91,33,19,151,191,73,20,199,
168,70,138,230,192,99,187,233,16,132,17,149,
140,126,170,153,72,81,229,35,117,151,170,202,
196,49,220,187,189,131,127,216,47,148,76,25,
92,92,42,34,167,158,97,104,145,0,186,174,
50,144,61,52,125,68,57,105,154,130,105,105,
136,159,65,85,213,51,146,127,47,220,99,63,
203,138,226,120,212,187,117,162,192,57,212,190,
68,215,223,163,183,13,99,202,42,99,234,14,
162,237,17,11,194,137,209,189,16,34,98,208,
101,72,10,147,46,25,170,52,203,75,100,135,
235,52,155,30,149,138,113,24,139,71,241,38,
117,98,2,95,38,81,167,218,154,194,212,250,
28,255,196,217,85,181,189,211,32,17,65,124,
247,28,19,115,191,114,52,86,224,40,3,86,
159,97,110,184,211,38,54,50,60,221,64,225,
167,95,97,20,211,25,6,100,45,229,179,3,
197,234,193,128,214,192,39,242,195,207,4,20,
39,87,57,165,49,157,55,143,64,177,91,187,
205,75,115,107,36,223,175,82,93,109,209,168,
219,248,94,132,44,9,228,178,42,169,98,146,
89,53,66,125,84,121,33,149,96,107,232,33,
0,231,82,38,51,150,142,239,199,184,78,140,
22,123,68,31,175,114,69,13,137,74,57,148,
243,105,180,169,14,253,237,247,145,247,146,100,
42,87,143,144,217,103,159,96,172,197,109,23,
132,18,100,149,183,105,239,127,128,191,83,65,
148,50,200,226,56,201,194,121,84,51,69,171,
250,46,27,213,54,166,160,145,76,156,35,242,
251,44,9,203,132,177,140,42,4,148,38,210,
132,72,4,97,76,16,203,4,153,20,146,172,
163,137,2,75,73,19,153,93,54,27,211,140,
231,215,217,107,206,96,8,17,219,115,21,242,
55,246,105,60,194,51,201,182,195,163,19,184,
97,72,71,65,213,11,4,178,135,51,126,199,
199,13,118,119,143,41,61,27,29,5,31,29,
135,212,88,137,111,187,73,250,67,23,85,25,
48,147,233,19,6,26,246,249,115,92,54,45,
110,94,87,25,236,245,200,142,69,88,137,4,
235,235,171,120,222,89,213,197,238,238,46,179,
211,51,168,170,198,205,155,55,209,52,149,177,
202,24,174,231,49,108,127,159,56,142,81,52,
149,214,248,47,179,40,222,100,135,69,38,89,
102,151,249,83,50,216,24,40,22,52,242,121,
237,112,175,11,167,211,212,48,66,92,223,37,
28,43,176,111,187,100,21,25,4,129,231,50,
9,86,122,54,111,239,154,136,81,145,148,217,
166,214,158,64,143,135,248,118,23,197,120,146,
35,172,204,160,29,18,73,121,198,11,73,188,
32,226,214,110,143,190,19,156,162,51,78,2,
133,174,74,36,147,6,166,165,161,27,42,157,
86,159,78,235,152,255,90,60,63,142,44,61,
123,225,113,124,42,207,248,84,254,39,195,43,
59,125,228,48,248,59,239,30,251,227,1,198,
241,255,39,199,191,200,234,234,247,120,222,94,
225,45,231,87,57,208,117,94,90,108,211,241,
33,220,189,141,123,152,49,11,140,132,21,107,
92,102,134,59,212,54,15,16,11,22,65,249,
10,41,86,136,227,248,208,234,59,34,6,214,
250,167,123,128,202,217,45,246,91,83,248,129,
198,84,241,209,163,85,183,134,167,137,155,240,
192,162,50,245,75,200,143,144,230,63,11,123,
24,27,217,159,42,248,62,235,122,160,208,146,
129,163,17,121,63,14,134,9,2,204,22,76,
138,73,237,33,36,243,16,20,56,88,169,83,
221,119,144,68,129,233,41,147,108,193,192,246,
37,52,57,66,149,32,86,100,226,116,2,177,
126,216,244,162,42,56,231,102,56,112,60,190,
92,202,144,62,148,45,42,138,64,179,29,16,
212,91,164,149,81,144,21,162,8,83,16,24,
55,52,54,199,235,180,237,127,67,123,245,251,
232,194,60,217,226,107,68,177,125,148,136,198,
64,213,247,32,239,97,88,54,95,152,109,112,
171,113,157,237,157,111,35,236,38,16,166,27,
12,236,73,108,97,192,68,238,219,188,208,111,
225,44,31,96,199,17,142,27,210,106,7,156,
59,33,174,10,167,82,120,18,244,218,33,139,
73,131,9,83,227,109,113,151,173,230,4,126,
168,210,28,198,220,208,34,190,250,205,171,240,
167,215,31,9,22,163,205,52,170,79,168,170,
72,20,197,35,203,110,69,199,84,194,35,79,
154,16,145,16,25,67,240,144,165,120,164,24,
46,36,217,233,36,40,101,219,232,170,205,199,
187,147,52,99,5,35,173,35,141,141,243,66,
226,26,55,254,96,141,144,24,71,49,73,231,
202,216,189,38,131,193,67,83,10,227,152,102,
179,65,167,219,193,119,61,38,38,38,232,246,
122,184,142,195,220,252,60,97,16,16,4,1,
66,124,131,85,174,162,226,34,17,28,13,170,
63,73,95,166,82,10,150,37,19,205,142,131,
23,48,72,20,217,20,231,144,119,170,196,130,
128,155,182,104,237,181,9,37,155,189,158,137,
165,245,9,227,152,31,110,230,121,33,221,226,
197,131,219,244,10,21,190,35,183,9,59,69,
158,198,64,175,56,251,235,232,185,50,182,23,
114,119,191,143,23,68,103,94,18,63,140,80,
14,79,246,138,36,114,233,185,153,163,161,68,
169,140,133,97,168,216,135,83,105,100,69,124,
10,179,156,159,238,26,106,38,210,207,96,141,
226,20,237,235,248,156,172,68,218,198,151,240,
163,61,84,35,66,27,255,93,86,90,255,11,
139,236,227,118,250,104,15,41,160,150,248,132,
101,158,103,41,250,152,168,182,198,160,181,195,
238,194,101,150,196,227,236,254,118,103,64,219,
59,75,239,86,178,91,180,251,5,108,207,194,
80,7,68,67,145,168,158,2,201,135,56,70,
26,27,66,32,32,40,163,217,60,162,155,67,
206,157,237,238,31,122,33,43,213,254,231,118,
255,84,105,212,8,29,68,49,73,253,217,164,
223,219,45,155,221,150,115,130,122,26,28,206,
62,120,74,87,203,135,221,99,101,81,96,58,
111,98,250,14,195,135,220,42,124,167,133,72,
76,16,198,36,19,163,70,25,57,101,208,86,
83,164,114,10,36,45,2,77,33,78,154,224,
5,196,229,46,177,166,17,167,19,200,2,124,
33,113,246,230,107,154,72,10,251,40,107,17,
234,109,228,119,174,115,245,203,47,140,208,94,
112,137,102,246,240,162,61,118,234,239,34,205,
63,218,187,201,117,67,110,180,251,196,162,192,
171,23,84,54,250,93,50,170,65,115,55,98,
169,80,103,167,61,199,228,181,143,81,13,147,
56,240,41,151,117,146,121,3,78,200,94,165,
229,77,140,251,219,40,47,158,167,55,208,17,
69,129,175,150,146,220,145,91,108,55,125,250,
155,38,222,238,5,254,194,218,227,27,175,76,
227,126,127,237,177,117,135,106,213,97,106,202,
196,247,35,54,246,66,38,133,128,253,97,72,
181,234,17,163,2,34,26,14,217,172,70,225,
107,75,104,155,85,26,179,37,122,55,178,88,
122,15,215,55,249,229,124,151,124,65,4,4,
4,33,38,206,141,177,240,197,14,213,235,251,
36,195,24,195,156,134,104,18,219,118,184,185,
222,160,71,30,49,14,40,11,59,104,154,68,
68,140,105,154,52,235,13,42,227,99,68,137,
4,33,50,53,121,142,72,150,152,96,133,105,
238,178,203,60,59,44,62,82,167,254,192,14,
68,216,171,227,189,122,133,181,221,113,174,244,
127,136,216,29,109,142,96,98,158,127,213,153,
166,231,39,120,161,236,114,191,149,228,173,182,
200,82,177,206,75,107,247,17,134,14,233,238,
128,55,102,199,121,91,240,105,239,221,192,202,
204,63,249,180,20,120,4,195,62,157,97,139,
108,28,19,9,49,245,190,123,106,55,183,170,
125,18,186,140,211,169,163,104,1,178,55,60,
42,114,10,192,194,164,204,218,202,72,225,226,
28,108,163,26,225,231,234,30,171,119,247,144,
195,144,142,34,253,100,175,225,115,112,143,61,
181,2,31,226,224,169,190,191,214,238,144,58,
49,185,77,212,22,248,240,96,134,178,240,14,
251,251,227,12,188,4,162,230,50,243,218,52,
94,107,200,234,199,167,157,0,230,185,201,58,
151,40,177,133,111,76,18,135,9,60,43,133,
92,74,16,91,6,139,186,198,218,192,38,138,
33,26,10,68,123,57,4,49,134,88,34,17,
199,180,172,20,81,203,36,171,92,36,85,188,
76,28,71,196,81,68,107,229,125,68,73,197,
243,247,144,46,110,18,198,61,134,205,93,36,
89,67,84,142,105,153,214,208,35,232,61,162,
108,28,133,35,115,198,167,189,183,79,233,30,
251,240,222,242,128,251,182,196,92,193,196,182,
31,178,17,233,183,8,125,23,65,148,232,58,
1,138,40,98,28,210,100,61,39,96,183,49,
124,168,70,97,149,142,207,131,79,225,106,121,
210,61,214,212,36,206,149,19,168,143,144,254,
121,78,151,116,113,23,235,195,117,108,69,36,
123,117,22,187,146,39,74,90,88,41,145,239,
31,116,168,24,42,243,135,96,176,19,70,180,
117,157,43,190,143,242,221,15,8,46,47,16,
23,50,8,213,38,113,121,196,251,137,213,6,
137,251,123,184,61,15,221,82,16,226,152,88,
18,137,243,25,196,141,61,174,90,6,133,140,
197,135,173,62,178,44,32,149,3,194,248,136,
142,62,125,250,149,101,106,142,143,42,10,216,
189,144,215,243,105,62,110,245,16,136,152,49,
20,106,141,1,242,243,51,200,200,8,219,7,
244,214,3,148,78,15,75,214,144,101,225,136,
74,33,140,144,238,108,144,214,85,130,203,11,
216,195,136,25,73,195,42,218,220,169,101,152,
58,247,207,233,238,47,243,177,71,95,160,0,
0,32,0,73,68,65,84,94,240,7,124,245,
165,33,119,63,168,50,24,156,221,32,181,154,
131,174,139,72,146,192,192,13,216,222,10,176,
180,136,66,74,96,48,8,142,2,112,106,62,
207,91,134,78,233,249,37,110,237,141,147,207,
236,49,91,220,99,60,204,145,47,138,135,153,
225,225,95,43,10,104,175,95,68,185,50,139,
210,245,40,109,55,184,255,215,203,168,134,200,
165,169,52,59,251,91,4,195,54,61,125,145,
80,47,160,200,42,90,41,57,50,44,11,61,
36,33,38,25,27,140,9,107,28,104,231,81,
99,17,21,159,140,119,192,38,231,73,243,104,
59,248,142,144,99,71,92,68,122,215,37,113,
251,247,232,230,84,50,25,149,56,147,96,45,
105,33,58,85,42,90,154,151,138,49,207,23,
28,26,174,207,206,48,36,154,40,129,34,35,
110,238,145,219,171,51,187,96,178,181,230,63,
210,213,248,97,250,73,79,23,121,48,249,180,
99,251,212,247,78,159,228,28,77,167,116,120,
234,75,103,42,108,185,167,21,126,105,11,212,
134,72,171,217,167,225,152,164,242,202,231,234,
30,235,6,17,97,20,158,120,63,127,66,215,
240,57,186,199,142,128,226,216,61,246,211,99,
164,128,167,23,200,156,208,252,71,169,255,134,
225,141,255,137,124,244,135,244,74,255,130,13,
117,192,252,130,130,214,238,51,231,249,172,221,
62,62,173,74,4,204,114,139,125,38,145,7,
125,198,100,151,247,213,215,120,117,174,77,63,
8,217,236,13,71,32,81,79,160,245,175,80,
57,247,205,163,222,157,56,142,200,87,239,162,
140,167,49,243,227,167,105,193,210,244,225,65,
184,199,246,222,191,68,154,106,51,88,95,166,
60,251,245,83,130,154,174,228,130,55,124,212,
201,102,228,6,251,180,247,246,51,184,199,30,
83,213,176,237,43,156,175,36,78,213,201,69,
73,70,77,100,9,16,217,108,143,234,204,89,
65,65,145,68,26,158,7,86,226,108,49,251,
179,172,140,169,176,88,182,206,84,210,227,40,
164,186,246,231,36,147,215,121,227,106,140,52,
62,75,40,37,48,78,204,65,88,238,13,105,
122,62,77,207,31,233,237,5,129,219,221,1,
95,40,164,145,62,222,32,14,34,164,189,58,
65,33,115,12,18,155,251,8,173,238,104,28,
231,197,89,130,233,220,41,197,129,242,253,107,
72,126,192,84,54,73,190,152,3,203,130,132,
65,219,11,232,250,33,93,63,64,16,160,233,
250,216,39,250,66,84,81,228,181,124,26,65,
128,186,235,3,49,185,253,6,223,28,236,243,
167,27,147,20,74,10,47,46,168,168,27,7,
212,235,163,105,111,139,139,201,211,143,167,219,
71,232,130,172,41,152,75,211,96,129,216,17,
216,47,182,168,173,255,53,169,236,37,124,239,
215,248,81,240,45,94,186,228,97,183,28,250,
125,159,92,78,99,104,135,236,237,14,153,154,
26,117,105,135,97,140,142,71,132,198,220,66,
10,145,209,48,161,7,5,59,57,8,136,34,
120,111,115,146,74,110,147,87,146,26,227,164,
176,18,17,15,15,117,221,183,61,110,119,7,
12,130,81,90,126,161,82,194,204,87,177,155,
93,10,25,141,76,122,30,39,16,57,168,29,
112,176,247,67,196,68,129,176,35,49,24,244,
40,148,39,16,98,143,178,85,32,146,53,74,
175,206,178,52,102,208,250,96,147,253,85,129,
180,216,160,30,77,48,197,50,194,9,158,102,
185,59,206,184,38,82,174,127,15,199,137,168,
247,70,38,128,153,204,168,232,55,107,168,164,
198,36,118,237,14,144,64,18,4,74,186,74,
25,144,62,190,59,170,105,152,58,43,87,150,
184,247,97,154,201,249,47,60,61,199,123,232,
235,191,221,178,207,28,12,218,67,143,233,252,
104,15,74,162,128,165,201,167,234,24,0,115,
75,21,122,31,174,81,175,118,152,44,100,80,
14,183,236,176,239,226,56,30,138,38,227,187,
33,7,213,54,115,75,21,84,245,167,215,201,
173,132,62,114,24,240,179,190,106,93,247,20,
80,136,178,194,32,61,199,27,222,91,252,160,
254,135,84,179,95,197,187,253,22,23,46,38,
72,95,157,66,89,105,31,205,95,63,162,146,
216,134,16,210,164,185,90,222,101,117,183,76,
91,223,101,215,118,9,247,50,20,140,95,35,
185,184,244,16,157,46,146,174,92,60,101,64,
122,230,25,233,73,68,55,135,32,14,240,133,
157,71,82,202,159,215,50,84,137,172,165,144,
212,100,44,77,126,172,152,170,119,194,199,175,
53,240,31,95,204,254,44,43,159,80,152,47,
39,206,124,185,55,108,179,187,246,251,228,102,
108,74,9,157,190,235,35,36,114,244,219,33,
86,24,31,121,252,184,97,68,86,149,105,121,
1,55,59,163,19,142,46,137,164,20,153,40,
147,68,40,100,137,166,79,12,149,119,125,226,
108,10,12,29,242,105,188,94,116,186,238,46,
64,120,110,26,161,218,68,172,183,177,90,61,
226,124,154,224,249,115,148,116,149,210,67,170,
217,97,24,210,109,219,72,86,138,173,161,203,
123,141,46,95,40,164,81,37,17,69,242,89,
213,139,188,184,117,157,217,82,25,105,233,127,
224,71,247,127,143,175,191,222,100,229,175,214,
232,116,124,54,54,6,20,10,26,138,34,114,
112,224,96,154,50,217,172,138,208,238,35,127,
116,143,216,208,72,158,155,102,66,26,242,177,
255,125,196,166,70,97,238,117,2,111,134,191,
88,254,223,121,117,113,155,241,104,192,74,121,
130,189,134,200,87,150,182,16,26,29,210,105,
133,74,69,63,4,6,151,173,125,157,233,178,
64,175,231,99,154,18,233,180,138,82,76,178,
221,156,101,162,176,202,185,148,193,36,42,86,
234,209,202,138,123,189,225,17,72,0,220,145,
28,10,191,124,145,76,87,194,8,35,146,189,
14,254,242,22,211,229,10,141,185,10,142,61,
162,9,197,160,143,166,234,40,9,131,220,164,
193,246,194,101,246,149,36,99,215,222,103,123,
115,72,73,11,49,114,179,244,246,108,238,241,
18,243,124,130,130,71,155,18,82,171,78,189,
213,58,149,107,116,58,30,227,227,6,66,183,
143,252,222,45,242,151,231,81,172,211,15,37,
214,85,194,133,73,196,118,23,111,162,196,245,
15,202,76,46,252,230,81,103,235,147,138,116,
237,161,79,223,29,208,25,250,143,109,106,114,
252,8,219,63,190,23,105,227,44,80,24,166,
202,139,175,206,115,243,163,77,238,222,218,33,
89,6,215,11,168,87,59,167,20,99,83,51,
69,84,85,166,223,181,49,45,29,33,140,105,
30,244,24,14,28,194,32,66,148,4,20,69,
33,142,35,28,199,67,150,100,12,75,69,55,
85,116,223,199,139,109,60,47,68,18,69,18,
73,227,145,125,25,161,40,241,15,107,228,44,
123,150,151,27,103,83,158,224,101,127,133,183,
237,69,122,217,223,226,131,15,254,63,94,89,
76,48,54,97,178,185,254,232,186,64,24,70,
152,94,155,97,52,205,108,222,98,215,118,81,
252,50,201,169,165,207,94,7,136,199,241,163,
45,66,235,0,223,233,159,202,40,194,207,81,
181,54,95,180,176,180,79,223,67,143,3,135,
31,27,40,52,69,100,38,111,157,1,137,56,
142,216,219,250,125,196,243,107,180,1,55,22,
25,75,164,49,37,9,81,18,176,251,17,138,
42,162,25,2,87,50,9,90,158,207,15,14,
142,165,181,78,24,209,112,125,122,149,2,146,
32,48,46,128,186,85,69,24,58,68,99,5,
226,164,133,208,234,130,163,35,74,103,229,89,
81,57,143,212,27,18,39,12,16,68,194,217,
241,199,254,13,166,36,17,202,33,73,93,165,
160,171,244,131,16,69,20,120,163,152,193,47,
196,252,193,221,49,164,243,30,139,219,53,62,
174,221,34,59,249,143,248,97,125,155,47,190,
56,164,187,211,166,209,240,184,119,175,135,32,
140,20,75,189,94,48,2,138,222,97,90,111,
25,128,192,148,165,179,90,108,211,107,125,136,
217,158,195,72,23,41,95,254,111,185,190,119,
19,219,190,139,188,57,78,224,15,232,84,58,
100,26,29,144,68,20,64,81,68,210,121,139,
100,66,66,81,96,110,46,65,102,50,13,113,
204,78,144,37,159,175,34,137,49,243,9,3,
175,31,63,162,136,166,113,208,79,17,219,89,
164,32,164,234,122,164,204,58,170,226,80,15,
3,234,214,225,203,151,52,73,150,94,33,29,
136,35,67,65,223,39,219,239,33,185,30,93,
35,207,109,45,79,55,18,152,73,118,120,249,
131,239,208,27,72,228,22,207,17,245,91,8,
81,131,136,113,206,243,54,107,92,65,36,34,
68,98,158,235,103,132,14,115,115,199,105,108,
108,234,136,251,13,146,139,103,199,151,70,211,
21,182,10,89,110,111,187,228,139,111,34,41,
143,119,117,109,110,92,35,55,253,2,91,77,
155,106,115,8,198,167,91,133,55,251,30,15,
116,37,149,180,142,227,71,136,2,120,97,76,
103,56,122,97,84,77,225,202,75,51,236,45,
223,103,127,175,77,224,135,103,122,80,148,195,
76,226,250,181,13,226,56,198,20,6,12,162,
145,226,69,211,21,172,132,129,97,142,188,190,
136,5,6,3,155,221,237,6,97,24,145,16,
186,244,227,99,167,210,115,151,38,40,150,207,
90,236,139,113,252,51,217,112,119,134,169,10,
99,220,32,66,59,65,111,91,99,223,96,187,
119,133,237,141,255,147,197,228,15,217,136,158,
199,149,126,157,27,219,127,200,165,243,99,176,
190,252,104,208,233,7,36,87,119,120,81,173,
241,35,251,13,26,246,24,25,239,199,147,162,
230,39,190,204,238,218,14,210,220,54,253,157,
149,83,126,115,234,231,104,54,185,188,223,231,
194,120,2,253,17,53,174,206,208,167,222,114,
48,66,135,129,27,254,116,128,162,146,210,31,
153,198,28,108,124,23,166,214,16,4,40,5,
5,230,205,136,104,0,125,34,136,99,18,105,
9,215,137,233,119,66,172,148,196,158,125,86,
166,249,78,189,67,94,83,184,152,178,144,253,
0,113,125,143,218,118,143,194,210,16,97,166,
130,180,177,71,100,187,68,149,179,77,88,210,
234,14,226,230,62,225,115,75,68,249,244,83,
55,174,8,64,242,176,25,74,147,68,52,224,
215,207,215,249,227,149,10,87,133,109,196,206,
159,35,151,254,123,208,254,49,31,89,127,200,
75,19,13,50,247,182,88,94,238,225,56,33,
139,139,9,30,214,140,9,3,27,105,121,3,
206,205,240,114,62,201,119,179,13,170,91,255,
22,171,181,68,105,238,107,164,199,174,144,230,
10,118,235,30,141,141,93,222,174,62,143,158,
168,16,138,1,101,154,164,191,44,160,40,22,
110,50,100,127,16,162,10,30,63,136,52,44,
73,166,100,134,244,171,105,116,117,200,7,205,
46,147,113,130,189,158,205,98,210,64,220,175,
19,149,11,232,138,199,84,182,206,76,110,20,
224,106,142,199,95,174,23,233,12,67,4,33,
70,32,38,140,100,4,33,226,0,136,34,145,
124,170,138,22,150,89,211,11,244,213,16,195,
24,144,50,55,153,82,21,94,168,182,217,107,
120,196,165,75,44,92,177,185,219,184,140,251,
189,191,34,98,244,28,30,76,22,243,31,210,
87,63,176,109,215,77,249,72,235,40,116,7,
8,141,14,225,204,24,60,194,132,47,167,41,
244,183,52,74,87,102,142,44,196,31,229,249,
164,90,25,6,94,200,218,94,23,53,136,145,
159,98,164,68,163,239,147,181,56,50,106,91,
40,89,163,206,217,135,234,25,138,34,51,61,
87,100,250,82,142,152,152,208,143,240,253,16,
223,15,9,252,128,100,202,60,172,117,152,136,
146,64,70,87,209,178,101,18,41,253,177,198,
130,113,28,211,235,14,113,235,123,96,21,144,
85,25,93,87,49,76,245,177,60,154,64,204,
63,172,209,252,6,237,161,58,168,150,156,192,
153,249,207,8,87,254,103,92,247,251,168,133,
175,211,234,47,80,75,13,169,140,109,179,191,
103,63,178,14,216,233,248,156,59,151,228,82,
110,149,221,8,194,149,23,126,172,107,83,205,
12,185,236,27,212,107,255,6,65,60,253,44,
11,73,149,131,158,203,240,25,231,249,252,68,
238,89,24,113,115,167,199,249,74,130,196,67,
138,167,180,169,80,141,98,182,155,246,83,213,
149,62,19,80,152,143,73,103,220,112,157,148,
37,243,122,33,69,56,140,48,19,103,95,110,
77,23,136,35,129,235,235,33,91,3,13,178,
167,31,230,172,165,115,49,109,33,119,7,200,
31,221,37,244,66,6,131,128,250,219,59,196,
63,220,65,215,69,198,58,46,145,152,66,144,
116,112,61,144,37,196,157,26,226,126,131,168,
144,33,74,89,63,118,119,99,82,150,72,171,
50,65,46,205,165,253,101,174,239,223,36,93,
185,204,238,157,36,241,185,30,189,174,207,249,
243,41,90,45,143,219,183,71,195,127,38,38,
76,146,73,25,77,147,16,100,17,113,187,6,
97,68,250,194,28,149,49,168,58,49,132,199,
92,235,160,121,151,172,248,71,12,51,21,242,
179,255,241,81,112,56,216,93,225,192,13,240,
90,7,136,13,29,69,77,227,185,45,12,107,
156,253,206,26,31,234,55,208,18,163,251,86,
119,125,26,125,120,97,250,176,0,167,170,32,
112,100,249,253,96,149,116,149,223,56,63,100,
207,241,24,248,33,154,168,98,71,46,78,24,
33,31,26,244,191,191,83,68,245,67,28,9,
202,70,3,17,11,58,21,166,115,7,180,222,
93,67,184,244,26,147,217,62,59,217,203,232,
55,223,193,14,92,146,180,217,99,142,49,214,
70,1,22,143,30,57,44,218,216,201,5,10,
149,36,130,209,39,92,48,17,134,14,130,237,
18,77,20,137,13,141,253,48,34,35,70,232,
146,200,32,8,137,0,67,20,185,246,145,72,
169,248,143,176,123,123,212,119,255,29,169,212,
235,164,199,174,158,121,70,86,97,142,15,87,
26,236,109,53,89,152,214,158,234,185,58,126,
136,227,71,152,135,116,86,20,143,198,80,62,
201,246,89,64,64,86,36,100,69,226,97,44,
186,244,252,244,97,197,176,9,198,147,45,101,
4,65,32,149,182,16,228,52,241,83,24,88,
134,210,63,80,79,71,247,226,49,116,162,158,
28,99,95,159,32,231,222,161,23,191,73,106,
250,159,112,127,47,207,185,87,66,204,191,252,
240,148,125,14,64,16,196,4,65,64,163,225,
145,63,104,161,101,10,168,137,153,31,251,250,
82,165,139,184,235,111,162,29,90,197,236,221,
248,115,198,174,252,34,146,40,176,84,73,112,
123,167,247,68,191,188,159,214,26,25,186,202,
143,173,49,183,159,82,185,251,153,128,98,183,
229,176,56,118,250,197,172,111,252,128,176,184,
193,213,140,133,38,138,12,31,35,66,31,116,
35,124,45,98,83,110,67,74,64,234,141,17,
234,53,84,45,230,165,92,146,130,166,64,204,
136,98,10,35,182,182,134,180,90,199,153,135,
235,134,136,162,77,49,94,102,56,28,35,213,
120,200,139,165,222,70,184,118,23,20,153,112,
126,130,56,147,124,242,31,19,69,200,215,87,
8,175,46,17,63,100,211,16,17,18,148,178,
168,55,54,137,196,31,2,151,41,206,254,22,
127,182,249,127,240,229,82,134,123,55,246,41,
20,244,67,238,51,102,103,103,136,32,192,210,
82,10,253,242,56,209,88,17,250,67,4,215,
99,220,212,104,100,98,134,245,15,217,88,189,
73,28,123,232,233,30,87,199,53,34,99,155,
253,245,63,39,59,241,6,146,162,161,24,137,
67,149,207,197,179,193,49,63,205,224,230,58,
216,1,114,113,140,32,238,19,91,29,238,116,
70,214,2,179,185,39,52,166,9,2,19,134,
198,81,180,11,64,240,2,98,83,231,94,111,
72,33,189,75,163,81,38,99,52,105,120,25,
10,178,199,249,73,143,212,59,235,248,165,44,
251,227,51,100,198,246,104,213,12,212,157,125,
250,164,72,208,34,150,52,78,74,216,55,185,
64,81,218,231,23,23,219,200,98,131,122,144,
103,117,63,143,63,61,201,66,248,62,98,113,
68,254,52,59,131,163,76,174,227,7,52,92,
159,173,251,10,105,227,159,162,200,73,234,7,
223,97,97,161,73,187,155,34,244,93,124,100,
188,32,34,161,203,136,130,192,78,195,230,218,
181,13,146,207,56,194,180,235,248,228,78,220,
19,83,149,184,191,82,35,12,35,102,230,74,
143,111,40,10,189,145,76,241,111,104,41,129,
143,28,133,255,128,18,79,0,10,0,113,225,
191,163,191,241,47,113,15,190,133,149,251,93,
18,99,111,112,119,187,205,139,47,214,217,189,
182,121,6,44,0,182,182,6,164,82,13,44,
123,18,107,230,165,159,200,53,102,202,207,31,
217,197,100,167,95,36,10,125,68,73,65,147,
69,206,141,37,184,189,219,251,27,55,5,124,
208,35,55,147,55,145,78,196,183,32,28,201,
199,159,22,2,36,85,85,255,199,43,191,242,
95,129,219,67,112,123,160,167,71,167,163,7,
169,190,211,30,189,28,161,7,110,31,20,29,
175,215,32,12,125,50,233,52,110,175,129,221,
217,161,35,125,155,197,9,129,105,75,39,142,
193,119,99,84,77,160,227,7,232,135,60,93,
24,199,28,216,1,247,250,67,2,33,70,20,
33,84,123,228,73,241,98,206,36,119,104,85,
78,16,32,238,212,16,134,163,33,61,15,128,
226,65,146,48,53,101,98,37,100,236,166,75,
24,139,4,145,128,34,197,216,190,132,19,72,
8,126,128,236,185,8,205,238,72,110,249,24,
159,30,223,3,205,179,65,85,136,83,167,165,
122,93,63,230,163,154,194,203,238,54,90,127,
64,83,84,177,165,11,104,102,134,97,51,66,
41,55,8,238,237,80,127,168,129,174,92,54,
200,229,84,132,222,16,161,218,64,28,216,136,
205,46,73,83,167,148,23,208,205,152,165,25,
135,249,201,144,165,178,66,88,235,49,182,126,
159,252,255,207,222,123,61,89,150,93,103,126,
191,227,207,245,222,165,173,44,95,213,93,213,
221,213,232,110,144,96,3,36,28,57,32,9,
146,26,50,40,205,80,15,10,41,66,17,154,
55,61,233,47,208,147,30,21,19,161,7,106,
66,65,13,53,35,137,26,90,112,8,16,32,
8,194,52,26,237,77,185,204,50,233,243,222,
188,254,30,111,245,112,50,179,50,43,77,101,
117,87,51,240,128,29,81,81,105,78,158,187,
207,62,123,47,247,173,245,173,201,71,152,221,
31,179,237,104,72,114,145,40,240,16,21,157,
208,119,241,172,49,178,150,198,51,71,68,129,
71,101,250,53,130,145,143,29,44,33,228,13,
4,41,38,136,99,70,94,192,185,220,201,241,
23,113,179,139,252,206,45,164,213,54,210,131,
13,98,77,37,46,230,40,169,10,235,150,203,
216,76,145,75,217,116,38,83,20,115,155,44,
72,25,140,247,86,137,158,191,65,55,107,241,
112,144,227,181,201,187,116,23,183,233,198,45,
114,113,23,165,212,64,181,219,136,162,64,185,
172,81,14,87,144,130,9,237,109,7,194,144,
102,49,160,226,109,81,110,197,220,209,95,228,
221,46,220,236,106,12,188,152,130,238,177,49,
118,185,219,183,233,222,174,83,171,252,23,168,
122,5,207,28,97,218,239,35,56,160,20,95,
231,238,150,195,63,126,255,93,62,92,236,242,
112,109,200,234,218,6,139,107,6,163,78,151,
98,78,36,95,208,147,61,26,71,32,201,73,
62,185,40,39,223,187,35,80,244,164,126,192,
183,136,194,144,162,46,18,135,1,146,162,162,
251,35,204,80,228,238,205,53,180,216,36,83,
44,64,232,130,213,131,84,17,220,9,132,62,
40,169,228,190,130,152,108,70,103,8,178,154,
252,206,234,66,170,148,156,145,208,77,82,25,
237,33,16,39,181,5,206,0,36,5,66,31,
193,236,64,186,156,164,77,250,118,114,173,51,
58,52,119,217,183,8,125,11,55,91,221,155,
59,178,190,83,219,16,38,207,231,12,30,157,
85,179,147,156,223,192,73,82,96,21,253,209,
220,101,245,209,220,35,31,220,113,146,158,26,
250,8,59,231,250,208,220,5,146,239,3,103,
111,238,184,227,100,29,30,159,187,239,36,63,
119,71,199,174,251,222,220,67,47,153,135,111,
28,148,51,178,150,124,150,111,38,215,186,19,
4,119,12,122,145,116,56,38,173,201,8,130,
136,59,238,33,169,58,145,239,226,91,99,148,
116,129,64,62,143,213,126,7,189,242,34,158,
53,70,203,95,102,195,88,101,78,234,48,234,
29,77,127,55,53,147,101,85,174,34,164,95,
70,148,100,220,113,23,81,86,136,163,16,119,
210,71,214,210,4,174,133,59,233,161,229,202,
184,198,144,56,12,16,101,5,119,220,69,144,
100,204,254,10,157,149,111,49,158,124,128,209,
91,34,176,60,210,165,105,162,192,39,10,3,
36,69,195,25,119,81,100,137,82,90,101,60,
216,38,20,21,8,92,4,103,0,122,17,193,
51,146,148,225,253,235,206,206,254,146,148,228,
93,250,22,104,217,228,235,221,107,119,247,204,
238,254,18,229,228,94,145,191,183,238,66,96,
19,10,42,57,209,66,136,66,220,88,228,238,
253,21,86,6,30,158,189,243,185,74,234,208,
186,239,223,51,239,254,245,31,239,83,20,178,
134,16,122,201,230,81,82,201,203,22,196,228,
107,65,76,190,223,15,44,202,41,234,197,52,
178,150,166,179,241,45,106,103,187,188,84,78,
172,119,99,28,98,107,33,31,14,13,22,39,
54,171,166,203,125,195,230,206,196,98,195,118,
137,67,137,151,106,25,158,43,102,153,77,235,
156,45,201,224,199,120,110,140,162,10,32,137,
8,182,139,56,182,72,237,244,110,112,221,144,
185,185,12,154,38,129,32,224,107,89,10,87,
170,40,149,44,82,70,99,98,36,192,95,86,
13,8,98,1,219,151,136,131,8,89,19,137,
11,71,135,4,60,55,70,206,106,68,217,116,
210,108,105,231,159,233,73,252,237,98,141,95,
61,219,33,31,69,72,189,33,213,112,194,170,
177,142,152,123,145,124,253,60,107,171,139,156,
21,215,153,12,14,42,138,98,81,37,155,149,
147,248,114,24,34,184,94,242,44,237,30,41,
203,162,146,23,17,135,46,130,146,38,112,1,
77,35,213,73,232,77,194,205,109,228,162,137,
159,126,137,76,101,26,65,20,17,101,5,89,
75,239,104,117,125,47,3,72,75,215,24,220,
186,79,228,187,72,229,128,112,34,32,106,144,
85,100,82,146,152,112,4,237,104,86,177,211,
7,207,135,148,134,216,233,19,53,171,68,205,
10,113,41,79,52,219,0,96,203,113,89,177,
28,22,55,95,164,169,133,76,134,2,217,245,
2,74,101,68,102,123,68,83,238,243,79,209,
69,20,41,224,57,186,232,158,195,182,95,69,
117,187,72,68,232,185,44,134,29,82,174,167,
104,77,231,104,94,174,80,208,19,78,170,204,
217,42,225,11,151,136,243,25,170,217,9,149,
180,197,134,223,102,52,81,217,190,85,196,25,
189,64,42,248,28,173,179,223,64,209,11,72,
138,70,28,6,120,86,159,144,41,198,242,2,
125,51,32,83,42,146,202,166,136,4,9,219,
151,16,5,145,92,49,79,115,174,137,20,187,
137,160,150,148,228,144,41,169,228,0,137,82,
242,53,66,242,59,89,39,52,186,212,106,101,
36,81,36,10,124,100,69,65,87,69,182,251,
38,182,19,82,169,236,120,40,145,255,232,28,
8,98,146,251,46,202,137,240,143,66,16,149,
157,130,169,104,231,124,164,146,195,181,155,239,
174,232,201,103,10,194,129,115,36,196,97,114,
206,36,229,209,181,178,126,104,238,121,99,128,
134,128,145,41,239,205,61,185,86,59,120,223,
221,57,18,131,146,78,126,167,236,187,118,167,
253,238,222,181,81,144,220,75,214,14,158,235,
199,231,46,202,137,0,73,149,247,201,128,29,
67,228,241,185,199,81,242,247,106,230,216,117,
223,155,207,238,92,245,194,99,114,230,240,181,
66,232,131,154,161,144,203,83,200,166,16,4,
17,89,207,236,48,248,62,58,27,138,158,69,
212,23,208,114,101,100,45,157,144,70,234,231,
80,203,171,228,70,219,152,59,69,174,62,26,
34,33,2,208,152,207,179,218,44,225,27,243,
104,217,106,114,95,81,66,16,165,132,7,78,
16,16,101,149,56,10,145,245,12,178,170,35,
41,42,130,32,32,235,25,54,111,125,155,254,
240,123,136,65,22,100,135,104,234,62,174,124,
159,201,218,50,138,82,33,93,104,237,60,70,
26,81,82,80,21,153,90,185,72,20,11,88,
126,66,31,130,146,218,41,166,123,108,221,247,
203,222,40,72,148,132,40,31,188,86,126,108,
127,137,18,66,224,64,186,186,111,221,53,46,
183,178,100,51,25,36,69,69,18,19,62,183,
98,70,167,55,50,64,47,38,239,234,241,119,
180,111,207,188,251,159,254,215,79,22,122,18,
128,70,33,185,145,107,246,9,178,15,104,232,
201,247,65,20,115,119,146,98,219,220,36,138,
65,147,4,206,230,82,20,149,36,143,87,17,
4,98,91,32,163,39,150,196,46,163,166,158,
78,190,31,15,66,114,69,137,232,204,20,226,
208,64,232,143,152,155,75,51,61,157,218,75,
173,53,60,153,76,73,36,216,151,213,148,77,
109,96,223,110,239,0,210,17,154,20,17,68,
2,214,157,109,162,92,153,116,73,61,42,234,
132,99,29,14,145,189,181,94,230,27,250,45,
196,142,132,115,103,21,203,132,201,112,194,133,
218,109,110,13,31,80,76,93,7,33,71,225,
229,5,182,30,190,191,87,8,55,55,151,161,
86,59,62,86,46,244,71,72,253,17,217,92,
154,96,182,4,146,136,57,142,16,167,171,156,
13,124,220,108,150,31,184,151,73,41,79,142,
183,203,90,154,234,217,151,25,40,127,149,120,
107,15,106,104,207,13,40,108,118,81,162,136,
120,170,70,44,137,200,239,222,1,73,36,60,
151,100,24,69,83,9,62,0,7,11,16,123,
166,143,119,239,50,151,35,135,112,115,158,243,
138,78,253,197,95,161,187,244,111,185,112,165,
194,228,246,58,207,157,159,112,175,159,199,28,
88,172,173,89,180,230,83,56,78,140,225,89,
108,74,243,92,168,91,180,59,62,205,34,216,
207,205,114,239,98,132,236,202,88,145,202,210,
71,1,41,197,193,169,218,152,27,105,240,103,
169,196,77,234,151,190,126,108,250,107,125,238,
215,17,68,153,161,43,128,144,100,19,105,186,
2,71,209,225,216,79,177,129,163,128,149,81,
204,66,85,69,145,68,38,78,72,181,164,81,
105,86,88,91,233,61,58,136,251,15,229,137,
168,161,9,158,145,176,21,232,57,158,21,161,
155,173,165,126,145,245,180,27,42,249,4,33,
27,37,85,225,97,187,200,217,215,190,68,249,
159,190,199,182,161,49,241,50,148,73,194,213,
247,63,238,115,173,94,230,199,230,34,57,142,
79,143,29,108,253,12,191,183,65,228,251,200,
113,157,153,203,255,18,128,233,231,191,65,224,
126,9,115,176,74,161,121,153,222,242,155,184,
254,6,126,52,100,59,252,83,6,139,243,164,
213,243,148,103,94,69,148,148,189,168,72,46,
37,35,162,179,233,140,255,89,214,46,189,175,
214,71,20,4,82,138,196,122,223,121,170,123,
60,181,162,144,68,129,133,70,134,194,78,166,
70,232,155,196,118,154,110,63,196,28,123,108,
89,1,177,110,112,57,159,102,211,241,120,165,
146,79,192,210,253,56,197,9,93,187,138,102,
151,209,72,39,235,142,17,250,163,125,49,50,
137,48,151,101,178,237,82,84,93,162,194,99,
21,141,178,204,118,199,193,207,70,20,242,10,
166,21,144,203,42,228,112,8,239,44,50,62,
123,30,52,153,116,78,66,20,119,51,114,56,
18,112,79,121,22,233,7,139,72,159,127,14,
113,161,76,247,135,247,25,116,93,202,34,132,
181,45,224,58,233,226,171,252,212,94,226,250,
43,51,60,120,99,53,17,182,61,151,106,85,
59,136,163,139,194,14,31,248,163,141,46,216,
30,210,131,117,172,82,149,84,65,39,156,111,
177,157,206,243,198,59,53,102,95,252,67,156,
241,233,26,138,20,91,47,98,124,252,144,104,
212,167,84,189,70,113,242,247,228,124,131,224,
250,249,228,225,252,128,224,198,229,3,153,69,
187,74,98,47,134,233,250,116,93,143,123,15,
69,22,206,254,254,158,208,182,250,155,8,162,
68,172,93,227,157,108,159,95,105,164,121,123,
168,49,95,91,198,248,96,64,161,85,39,86,
28,38,161,128,92,157,227,87,74,203,232,41,
133,116,219,98,113,57,230,210,185,62,149,86,
141,119,151,82,100,226,27,148,90,47,51,233,
45,49,122,255,39,204,63,255,71,72,106,234,
20,202,48,9,7,166,249,4,113,250,147,240,
4,65,100,100,135,188,183,106,48,85,212,73,
171,18,78,0,23,206,84,9,34,30,133,67,
118,61,137,39,90,78,2,168,89,208,243,73,
104,6,64,205,125,234,132,138,172,99,162,4,
62,147,66,243,179,151,38,187,161,34,189,244,
115,169,40,236,167,200,26,242,194,8,85,18,
137,2,143,170,212,195,126,103,133,240,165,95,
199,235,43,156,125,240,215,12,119,50,147,39,
19,159,212,91,43,168,151,110,225,140,111,160,
231,15,175,243,112,243,3,196,169,183,184,80,
87,232,185,62,131,91,133,67,123,180,208,188,
188,3,104,95,69,16,175,33,41,42,91,247,
191,133,155,190,131,89,93,194,184,255,33,57,
253,37,42,179,175,209,55,124,30,116,77,226,
253,133,148,167,165,22,255,132,20,228,119,182,
38,73,203,135,157,172,177,238,196,123,234,118,
172,242,94,198,6,36,49,175,19,138,124,138,
41,133,102,94,71,244,109,156,177,133,164,168,
136,130,142,98,207,209,150,62,132,24,10,5,
153,43,249,12,247,38,246,1,37,225,71,49,
146,32,36,112,193,19,158,181,124,255,46,166,
39,35,202,2,178,24,19,188,116,153,40,155,
102,178,97,81,236,222,74,44,97,47,72,92,
130,29,169,31,181,170,228,11,43,60,92,236,
33,73,2,113,12,151,174,228,209,53,9,201,
178,40,46,221,33,58,55,131,105,40,4,217,
52,97,144,100,96,29,202,138,137,98,84,89,
68,203,233,4,106,226,214,181,90,41,6,3,
15,207,246,137,253,69,156,209,75,200,114,150,
222,202,117,238,45,40,180,230,71,180,151,199,
152,102,192,202,138,201,252,252,35,188,35,248,
252,117,98,73,68,220,234,33,110,15,32,138,
146,186,144,135,109,194,77,27,237,245,139,76,
124,120,115,93,166,182,240,219,120,230,8,207,
24,156,186,91,95,54,119,17,189,80,35,142,
99,28,75,34,154,107,48,186,99,210,169,92,
220,227,45,145,196,8,85,10,152,41,245,16,
108,7,233,206,114,66,61,226,122,20,206,206,
242,147,117,29,121,188,128,159,31,239,177,88,
237,206,33,91,184,74,251,206,67,182,47,250,
228,6,30,231,163,144,82,65,102,89,62,67,
211,187,71,238,124,138,137,232,96,201,77,244,
104,155,225,208,199,113,66,134,75,3,6,98,
133,188,240,58,169,194,52,190,53,198,24,220,
70,84,12,140,206,50,90,238,100,134,85,215,
24,28,240,252,176,159,96,125,57,163,211,239,
122,119,178,183,231,45,65,33,147,213,24,142,
66,26,89,133,184,41,178,109,244,159,238,158,
190,149,196,116,15,128,92,107,59,161,160,35,
50,240,158,112,206,246,110,33,68,136,178,240,
232,124,62,49,165,235,180,243,221,113,191,66,
247,128,87,164,109,255,13,110,227,155,143,66,
16,79,115,207,93,124,227,84,92,79,30,196,
225,193,207,63,233,153,68,137,137,45,96,234,
54,39,149,37,120,198,128,72,144,184,181,49,
161,148,81,104,21,83,24,118,129,203,153,128,
202,246,15,216,88,119,24,78,14,190,167,206,
134,201,141,230,199,252,192,248,43,234,241,239,
30,186,167,49,250,128,95,189,150,37,45,73,
108,76,92,76,161,117,108,165,118,224,90,8,
162,140,164,168,20,74,55,216,94,155,32,26,
45,4,84,28,163,77,239,193,59,152,206,135,
52,195,12,130,32,19,250,58,122,249,121,252,
72,196,11,66,188,32,98,60,26,32,106,121,
162,253,254,190,107,236,120,183,234,169,215,11,
64,17,5,234,121,157,140,34,17,26,14,206,
206,54,76,199,144,9,77,76,103,180,19,66,
149,78,169,40,118,249,70,162,240,72,238,145,
140,38,49,87,73,239,177,15,250,182,129,40,
201,123,21,136,234,160,138,27,39,10,192,242,
35,62,24,24,156,203,165,176,130,144,143,71,
38,86,16,98,135,17,21,77,225,243,213,194,
201,81,129,153,6,98,103,64,102,56,161,235,
234,200,18,32,101,192,129,162,230,225,184,33,
158,17,81,18,6,196,43,105,162,51,73,28,
80,24,155,104,105,133,217,43,21,68,85,33,
55,95,66,104,85,8,68,17,97,100,32,45,
173,33,125,124,143,60,16,53,43,140,155,211,
164,50,135,195,10,18,50,133,122,150,181,230,
235,108,15,75,148,114,117,206,100,123,76,77,
165,144,101,145,204,220,128,59,189,53,138,83,
47,81,154,254,28,221,237,12,141,207,91,228,
7,239,50,30,251,116,187,46,233,180,68,173,
182,239,176,41,50,209,108,131,104,182,65,224,
199,184,155,35,100,219,38,125,33,177,96,254,
105,209,1,115,158,212,116,19,65,148,136,194,
0,189,112,186,102,53,251,175,237,14,115,8,
155,107,84,198,6,21,115,45,233,225,219,40,
19,151,242,12,237,52,239,109,149,121,169,220,
78,50,202,98,24,214,26,252,211,237,121,166,
207,254,33,242,76,230,216,251,206,231,255,107,
222,190,247,127,80,159,243,65,47,98,190,120,
3,117,213,65,149,114,140,46,76,227,123,62,
214,79,92,226,124,154,250,220,50,185,153,2,
94,219,34,88,241,41,93,78,114,212,237,225,
54,81,227,67,24,100,201,54,22,158,88,109,
13,160,23,106,216,163,54,153,82,3,6,167,
80,156,167,229,90,114,39,123,215,102,138,41,
10,69,141,113,223,38,85,72,115,38,15,221,
135,3,162,40,78,236,153,83,113,241,236,36,
123,236,191,118,247,107,103,156,40,10,45,183,
127,113,79,117,223,194,100,128,18,132,152,167,
230,144,58,237,124,143,224,122,210,75,120,226,
111,66,110,234,147,221,83,144,63,19,174,167,
221,181,138,128,237,80,227,76,41,125,226,89,
80,11,53,52,67,167,239,135,24,99,145,185,
169,255,158,59,169,255,204,124,244,38,25,251,
1,217,178,142,59,113,217,218,122,20,122,89,
125,191,205,213,215,151,184,211,255,30,41,229,
2,97,96,225,135,19,192,97,118,118,147,180,
164,226,69,49,111,189,41,209,156,187,118,160,
160,238,160,254,61,40,23,231,202,255,213,129,
62,235,158,49,192,90,189,69,101,118,147,122,
211,164,55,9,232,220,237,51,115,37,185,46,
244,28,6,41,168,180,230,49,189,0,195,9,
152,56,1,195,190,156,100,22,158,38,12,186,
111,111,93,152,206,145,61,34,53,182,61,114,
49,37,25,244,16,180,252,169,222,217,35,48,
27,16,124,235,192,203,147,68,129,133,106,134,
51,213,244,129,98,151,40,240,18,160,117,199,
234,213,212,50,195,254,59,136,153,128,40,78,
50,112,94,44,229,120,171,63,97,232,5,4,
59,97,23,59,140,40,40,50,90,36,162,106,
194,137,46,86,92,43,161,61,55,67,152,201,
146,174,233,168,154,64,156,78,33,24,22,238,
192,66,191,60,67,52,223,218,203,104,18,4,
1,167,94,35,119,125,22,237,92,19,42,5,
208,84,80,21,80,21,196,229,77,28,55,196,
117,34,52,207,37,92,235,161,250,110,82,65,
189,47,52,35,9,49,99,39,141,166,134,156,
169,116,232,58,69,178,231,243,228,125,147,84,
81,39,155,18,217,182,214,9,148,203,196,65,
68,113,234,26,75,15,23,121,101,222,166,187,
50,34,142,147,84,217,106,85,123,20,150,16,
4,130,246,24,231,94,151,104,105,131,84,191,
143,46,5,132,192,27,91,34,178,243,27,52,
22,190,190,87,84,230,219,6,74,42,119,74,
227,240,209,181,238,232,46,103,154,6,113,54,
77,92,204,33,14,198,160,200,196,153,20,186,
22,18,68,49,255,184,153,161,54,45,114,71,
170,112,115,244,58,173,75,127,128,120,132,165,
178,255,190,130,32,66,156,199,137,111,178,18,
133,196,90,154,172,105,82,144,76,62,72,167,
184,239,120,92,136,92,236,108,133,247,138,175,
225,213,234,204,132,91,200,163,49,155,46,168,
197,75,64,140,223,85,200,103,94,37,85,104,
156,66,150,88,244,183,126,194,36,254,75,236,
77,135,145,56,253,132,63,176,31,1,173,79,
146,105,206,40,1,241,128,156,46,35,137,2,
67,203,167,146,85,241,205,33,109,11,198,67,
27,77,14,235,207,60,8,0,0,32,0,73,
68,65,84,78,119,207,112,199,66,62,234,218,
93,50,55,119,156,0,195,178,126,232,156,29,
251,110,119,112,138,64,61,101,250,239,105,215,
96,215,251,217,31,154,19,4,208,10,159,252,
158,129,187,3,172,138,167,19,102,113,116,170,
84,227,253,107,101,186,33,41,85,34,165,74,
199,158,5,45,157,163,148,81,48,220,16,219,
11,233,155,62,129,126,158,56,255,18,29,69,
194,45,232,204,183,68,244,192,197,243,34,130,
32,169,158,207,142,71,132,175,10,244,164,91,
108,10,62,229,218,42,55,114,107,212,202,58,
182,25,97,91,2,157,209,21,74,173,23,78,
128,190,30,201,197,246,226,15,104,223,254,7,
242,245,11,136,178,74,232,217,244,87,223,34,
147,63,143,109,230,216,94,175,96,247,155,104,
122,3,73,201,49,92,255,128,141,143,191,69,
109,225,101,20,45,133,42,139,164,229,152,106,
62,133,134,207,208,141,143,22,232,187,225,168,
29,25,187,187,94,170,40,48,87,61,188,111,
12,39,224,94,199,124,244,110,247,146,11,142,
31,79,4,179,47,54,179,167,226,48,119,140,
46,66,198,221,67,186,99,96,18,4,140,31,
235,207,156,150,37,202,154,140,235,30,4,166,
162,88,224,97,175,78,37,59,161,160,91,68,
83,143,172,233,40,47,29,136,247,171,138,72,
121,38,79,164,171,236,247,67,99,93,5,255,
104,236,35,22,69,44,73,227,238,237,45,102,
167,147,212,90,226,24,113,99,27,113,179,75,
52,85,35,188,52,15,65,8,162,192,124,249,
145,107,233,135,18,55,59,179,44,92,85,40,
220,187,133,125,175,199,153,112,192,71,193,127,
66,203,255,22,32,144,205,127,145,187,90,7,
53,223,101,52,112,241,144,137,162,24,81,20,
48,111,174,179,214,89,197,54,28,166,27,10,
141,134,78,187,216,98,217,203,208,91,89,64,
207,190,70,177,117,249,153,196,113,69,229,2,
78,97,139,204,219,31,17,188,112,158,224,198,
101,4,195,66,90,222,36,214,85,102,166,235,
44,77,34,254,234,97,139,89,229,53,154,231,
79,159,63,174,167,171,220,89,23,73,207,141,
184,55,30,176,37,159,161,56,218,192,141,34,
162,16,180,133,2,41,107,128,43,77,184,213,
190,206,149,151,46,83,252,254,251,204,140,255,
129,141,209,117,210,133,121,106,243,95,36,176,
79,207,205,31,197,235,156,111,250,60,48,141,
253,103,225,25,140,71,70,202,198,208,97,99,
232,144,223,73,205,86,179,69,220,213,14,150,
233,144,215,120,118,67,203,63,242,102,60,243,
84,236,177,37,115,128,18,248,172,102,171,191,
64,179,119,198,253,142,133,54,37,30,91,68,
6,160,202,34,87,91,57,218,19,151,181,190,
205,208,242,25,219,26,83,197,223,39,157,87,
248,241,230,223,147,153,125,159,198,194,144,243,
238,22,155,75,3,198,99,159,87,13,135,232,
220,20,15,222,28,49,92,52,217,94,123,151,
234,124,153,210,23,206,51,73,37,180,65,187,
77,143,118,255,63,110,52,46,124,145,198,133,
47,238,179,200,83,52,46,254,218,1,239,199,
51,6,123,94,187,158,125,141,234,194,107,143,
233,83,159,192,49,201,167,20,170,129,196,182,
117,2,70,33,8,168,178,72,176,35,127,189,
40,230,94,199,228,108,237,32,213,210,214,200,
253,68,181,254,39,106,129,163,168,195,143,26,
147,225,10,70,230,57,162,137,136,162,246,200,
228,55,120,163,59,62,116,52,27,186,138,42,
138,184,59,224,100,20,11,44,118,90,200,98,
200,153,234,54,55,183,234,204,149,61,10,90,
162,96,2,63,70,81,14,46,140,127,253,2,
66,16,16,11,226,233,115,75,20,25,165,156,
225,249,87,27,224,4,76,92,17,85,138,30,
105,100,207,135,48,66,249,201,7,68,211,117,
194,179,137,245,122,167,211,98,170,48,32,167,
219,124,252,176,197,139,219,3,198,253,144,241,
182,193,220,111,244,88,29,244,73,151,91,228,
235,151,89,188,93,230,87,47,228,176,223,156,
96,24,240,254,251,54,30,10,81,4,41,201,
103,118,54,3,185,52,107,149,26,239,13,190,
76,237,236,151,168,107,233,103,122,136,114,141,
27,220,91,254,71,174,11,236,121,73,113,54,
77,152,73,35,216,14,19,63,100,60,50,153,
18,62,71,105,250,233,138,140,140,193,67,92,
85,194,187,223,98,90,93,64,144,63,198,175,
84,168,168,10,67,47,96,81,2,49,151,5,
203,129,252,93,222,94,158,227,149,231,231,169,
191,179,204,120,237,143,241,212,255,17,81,205,
158,30,188,28,175,67,48,36,22,4,162,56,
164,146,85,247,138,135,62,139,49,182,3,188,
32,137,12,143,71,118,146,93,133,255,236,63,
72,203,37,158,197,30,232,157,61,214,162,27,
100,74,191,160,240,56,100,60,196,220,221,50,
121,110,58,119,178,124,18,146,110,155,165,180,
194,74,207,162,111,250,172,13,108,186,166,199,
108,249,203,248,254,13,54,21,153,7,237,255,
151,234,243,155,168,82,204,27,195,41,162,119,
11,212,212,187,164,219,183,89,117,90,140,238,
116,88,152,188,203,199,243,191,76,182,116,9,
103,188,77,224,141,136,195,8,45,91,71,205,
20,142,164,151,57,104,244,159,172,84,142,21,
206,90,26,180,52,158,49,164,81,208,233,218,
71,8,249,125,247,45,103,84,202,153,28,66,
54,199,216,9,232,77,60,238,110,25,92,104,
62,98,248,254,164,249,21,242,73,232,122,111,
236,50,85,78,61,17,137,15,164,7,20,26,
9,133,131,105,231,24,111,63,79,57,215,71,
244,33,155,54,168,165,52,154,169,16,85,140,
177,140,136,204,142,151,240,193,250,60,215,167,
151,17,133,24,215,137,56,163,111,240,183,55,
155,124,237,236,22,154,148,40,138,124,73,58,
188,46,138,124,72,73,120,78,124,114,56,235,
218,57,132,149,45,38,31,109,81,212,15,10,
0,193,245,144,63,90,2,63,72,40,65,118,
198,250,176,194,165,250,38,0,87,59,255,200,
199,193,21,102,164,247,217,52,3,102,110,223,
100,49,147,193,53,91,73,217,190,150,67,173,
87,224,205,13,98,4,156,72,65,193,167,144,
145,168,159,41,51,124,241,60,31,173,21,16,
140,179,76,93,249,23,199,78,115,123,226,50,
28,216,248,246,4,4,40,164,20,166,138,250,
169,94,176,40,201,12,173,50,225,92,19,12,
59,161,28,143,99,164,15,151,136,166,107,44,
249,57,10,206,55,41,205,191,240,212,27,165,
52,253,2,115,171,34,197,51,11,40,169,44,
131,135,29,226,86,132,29,152,123,201,10,254,
78,42,167,170,59,108,154,54,237,168,74,51,
189,194,69,123,133,187,75,255,11,238,249,255,
233,212,159,103,77,150,209,46,247,185,187,45,
80,202,93,70,43,166,24,90,62,65,24,127,
134,66,8,198,189,54,150,225,48,61,87,5,
255,17,113,159,239,5,184,110,64,38,167,29,
162,114,255,212,30,70,28,37,223,63,246,146,
247,60,138,76,229,23,26,98,191,177,24,70,
220,217,50,184,58,149,59,80,109,124,156,177,
123,190,145,101,96,249,44,119,45,28,47,100,
113,203,64,113,13,106,141,22,181,153,255,22,
87,17,176,60,135,244,78,38,222,96,120,155,
242,75,1,181,183,239,33,153,54,147,129,196,
149,226,29,220,209,255,140,86,73,19,247,199,
104,5,149,129,90,103,113,253,26,211,151,126,
239,100,255,245,83,100,191,5,158,77,28,133,
104,170,72,41,171,210,55,188,99,179,159,182,
70,14,35,207,100,62,93,163,85,208,105,21,
116,44,47,36,8,99,212,157,254,57,245,188,
70,207,120,122,131,235,104,140,98,103,18,110,
16,81,206,168,136,68,4,182,177,7,64,238,
143,197,245,215,222,198,171,255,148,221,94,163,
170,226,129,98,49,28,23,73,213,150,121,161,
161,49,87,20,72,165,4,84,77,64,81,5,
4,1,182,141,60,197,148,69,70,117,49,199,
17,178,146,216,78,173,156,205,123,195,18,151,
26,46,90,234,244,204,139,182,25,145,202,28,
127,189,48,24,227,190,191,74,86,77,122,82,
120,126,132,40,38,115,17,92,31,193,118,247,
52,81,52,93,163,103,230,104,230,70,172,14,
170,84,179,19,228,193,144,213,113,133,206,80,
197,183,29,198,109,147,27,151,35,86,189,123,
132,193,12,161,227,209,172,174,225,220,237,32,
70,1,181,178,204,244,66,142,248,213,235,124,
24,95,165,211,187,65,227,204,31,144,175,94,
57,241,57,110,110,76,112,76,3,79,76,225,
5,17,19,39,64,145,196,67,164,94,71,97,
9,0,230,36,164,90,90,33,181,188,134,184,
218,70,8,66,4,73,228,102,189,202,210,50,
52,167,126,7,225,20,89,14,71,225,36,153,
66,3,73,217,77,139,150,81,184,203,195,40,
33,107,113,163,8,63,138,247,104,149,77,71,
196,19,65,174,84,168,198,19,74,178,197,162,
51,133,154,170,30,160,97,62,86,233,9,25,
250,247,215,16,106,19,134,111,247,40,181,206,
81,202,231,217,232,91,72,71,165,190,60,53,
70,113,48,30,95,205,169,212,114,26,43,195,
136,141,205,17,149,122,158,216,183,88,91,183,
120,176,180,197,195,165,54,237,141,1,190,23,
162,105,10,81,20,33,239,178,114,158,132,81,
60,33,238,158,96,24,122,82,135,177,23,231,
23,254,249,49,138,79,123,207,127,6,140,226,
192,199,133,49,134,19,80,201,170,123,130,248,
36,108,47,165,72,212,242,26,49,49,166,19,
18,250,22,147,80,161,51,118,25,59,1,166,
159,116,117,67,128,76,174,193,208,43,34,149,
77,138,5,25,16,232,173,14,177,218,99,156,
149,45,172,237,49,86,207,97,126,42,205,150,
40,146,46,188,118,36,118,123,178,125,29,17,
122,78,82,216,119,156,151,219,223,198,54,13,
52,77,67,148,100,82,154,198,246,196,61,209,
45,8,29,147,174,43,51,180,124,36,81,32,
167,203,200,251,206,139,38,139,20,82,10,249,
148,194,96,52,62,53,70,113,34,152,29,70,
49,157,177,199,200,9,233,90,176,57,74,226,
185,189,161,65,24,11,228,51,58,74,170,200,
112,253,54,162,146,69,246,75,68,130,131,172,
217,212,149,52,183,183,158,167,109,7,44,13,
52,4,33,164,150,138,119,0,21,157,129,149,
161,150,234,99,140,35,178,121,9,73,78,148,
72,104,135,132,98,26,15,155,130,42,224,57,
17,166,21,163,42,194,177,235,19,248,49,227,
97,66,66,120,228,53,158,143,242,238,29,92,
23,116,57,177,122,71,35,31,199,14,73,165,
165,131,10,90,85,136,166,235,116,141,60,173,
194,144,200,143,136,35,129,212,250,26,241,230,
50,12,214,24,11,117,188,80,196,107,247,121,
49,103,98,202,43,24,222,2,65,180,78,179,
40,97,204,157,163,95,63,199,82,234,6,166,
250,77,10,83,95,65,150,179,79,4,168,131,
48,102,115,232,28,122,15,166,23,80,207,107,
135,26,68,29,117,56,244,92,139,173,149,159,
50,103,172,33,132,17,194,196,68,112,125,114,
11,83,220,125,168,81,174,255,242,83,131,228,
71,90,106,153,42,189,246,79,185,94,211,232,
217,62,158,16,37,217,202,118,147,200,139,25,
27,37,100,197,163,23,135,204,212,210,232,131,
1,15,168,163,166,207,158,74,81,68,190,135,
24,103,9,55,85,40,76,232,173,252,12,85,
80,216,24,165,216,222,54,200,63,206,239,244,
84,138,98,152,28,142,208,165,145,1,21,159,
166,106,99,187,1,111,190,123,159,82,81,67,
207,100,121,112,243,1,235,155,22,254,190,158,
202,198,196,97,107,99,64,24,132,84,106,249,
79,175,40,30,7,189,247,232,28,84,106,195,
77,114,182,193,56,87,253,133,162,56,234,88,
7,17,182,23,82,206,170,8,167,216,179,162,
32,36,66,82,87,24,142,198,68,59,116,195,
94,16,97,121,33,134,27,210,51,60,6,166,
71,38,63,69,182,241,69,172,210,23,233,136,
5,212,166,76,58,167,81,201,128,27,8,164,
228,128,230,115,53,238,26,53,178,197,151,63,
19,69,161,165,50,40,170,130,111,142,240,237,
17,178,24,146,201,228,144,37,241,88,54,218,
221,245,242,195,152,129,233,179,53,114,17,68,
14,96,205,170,44,50,177,3,70,227,241,179,
1,179,119,99,130,143,55,119,9,130,136,141,
161,67,181,20,162,171,105,2,171,134,97,198,
196,81,132,47,77,17,72,49,45,213,226,143,
174,61,224,7,27,10,191,215,234,115,231,110,
17,202,201,195,173,12,171,92,172,172,226,123,
49,133,242,65,235,54,87,148,56,51,233,241,
157,141,50,249,214,128,7,150,71,51,47,34,
153,50,226,200,64,112,28,194,70,117,79,176,
199,113,66,143,83,174,73,140,250,1,197,202,
225,71,146,238,44,67,20,17,33,237,11,37,
68,164,210,210,161,184,93,156,209,17,55,187,
20,138,105,134,86,134,169,165,55,89,202,188,
72,102,96,208,238,120,224,134,228,89,102,131,
115,20,173,30,66,182,206,172,215,165,43,188,
69,207,249,53,122,233,52,197,214,53,52,73,
166,254,196,205,146,172,239,174,251,124,92,245,
105,16,198,44,181,77,46,54,179,79,12,65,
9,130,136,21,205,17,103,151,16,12,139,88,
145,136,206,205,97,196,32,69,249,103,24,0,
16,176,188,58,154,216,38,184,87,71,144,138,
80,187,135,146,235,16,137,17,177,95,32,74,
247,57,91,200,224,117,2,210,142,71,156,122,
186,250,206,32,24,226,69,61,22,46,255,15,
116,30,254,3,182,177,193,245,86,137,191,107,
7,140,71,214,97,101,113,210,240,140,68,56,
37,43,143,174,72,100,84,137,122,78,101,108,
7,244,156,128,229,205,14,131,222,132,233,51,
77,162,40,34,155,75,49,157,203,96,25,46,
134,97,227,239,28,78,89,150,56,115,174,193,
103,50,180,92,34,116,3,143,65,166,132,24,
255,162,50,251,164,49,176,124,182,134,14,173,
162,126,234,191,201,165,100,46,52,178,172,216,
50,230,17,77,145,108,47,98,169,109,82,201,
249,156,169,164,201,212,94,70,41,252,6,227,
113,135,245,181,127,199,217,47,218,76,45,174,
19,72,34,97,248,116,5,138,113,28,225,25,
67,164,125,248,228,113,24,198,214,242,34,162,
36,81,44,148,16,21,149,136,152,170,174,82,
202,40,137,103,113,154,237,164,136,148,142,160,
177,31,59,79,135,191,125,170,62,142,174,31,
237,28,184,60,26,15,160,96,34,186,41,26,
121,151,246,112,138,232,254,29,174,147,193,126,
99,137,217,115,51,64,137,40,18,209,49,112,
237,248,200,174,108,130,0,153,172,196,203,147,
9,127,251,193,20,106,179,205,149,90,136,238,
219,200,139,119,129,152,80,141,18,142,162,125,
139,107,25,137,224,15,252,24,249,49,0,60,
154,174,17,231,210,100,150,54,24,187,50,121,
45,160,222,208,247,128,238,88,150,246,66,79,
98,187,79,148,210,41,180,44,30,244,234,212,
76,135,216,24,225,11,42,69,221,100,205,45,
34,18,50,197,189,164,224,26,129,248,210,28,
175,166,4,126,248,179,33,245,11,95,56,194,
128,115,25,109,253,152,241,134,65,253,202,191,
6,160,111,122,44,119,109,98,98,46,54,18,
190,120,47,56,94,40,140,236,164,162,115,161,
154,121,162,178,72,215,190,202,221,201,77,46,
113,159,240,202,89,226,106,145,229,182,129,38,
206,63,211,67,42,136,49,142,47,32,42,23,
153,189,248,85,182,238,125,155,162,244,46,114,
205,103,123,4,4,18,219,131,44,179,118,31,
51,87,68,22,159,174,194,184,126,246,43,228,
71,207,51,216,120,151,82,243,101,4,65,226,
225,237,255,141,43,173,175,241,206,82,155,235,
47,47,156,226,100,134,9,14,160,230,144,100,
153,122,78,163,16,41,228,107,83,132,81,204,
253,109,147,129,41,48,28,120,220,189,53,98,
225,252,60,130,154,70,0,234,83,197,3,53,
4,158,23,96,78,108,68,73,220,107,92,244,
89,142,95,100,61,157,110,172,15,28,202,217,
167,99,245,85,36,129,203,173,44,119,182,140,
67,134,240,238,232,77,60,12,59,96,46,21,
161,3,169,124,29,249,220,191,225,163,197,63,
97,235,138,200,108,74,67,222,40,158,234,243,
124,123,132,107,244,25,246,191,139,35,109,83,
77,255,22,146,148,65,78,229,144,143,241,176,
155,243,23,24,44,127,0,66,25,65,148,80,
119,174,123,18,46,115,80,233,133,124,176,58,
162,152,86,152,171,164,246,154,24,157,36,107,
158,185,162,80,118,98,95,173,139,223,160,243,
240,59,184,44,82,170,141,121,185,169,226,108,
191,139,62,238,209,44,164,8,126,249,34,174,
160,226,57,17,170,14,41,193,216,83,18,113,
12,182,21,17,71,16,120,17,174,27,147,78,
139,204,77,67,49,187,193,183,59,89,156,112,
76,122,56,73,20,67,20,35,45,173,34,110,
117,9,110,92,65,186,121,159,240,218,5,162,
48,38,157,21,247,240,142,131,113,18,5,233,
225,38,136,49,2,2,166,47,147,81,2,162,
249,86,210,127,123,166,129,120,127,13,177,221,
7,81,32,170,39,86,66,236,133,16,69,52,
221,37,22,141,22,225,168,67,150,33,43,92,
166,197,3,20,92,98,199,37,155,17,24,220,
233,51,93,253,136,94,251,28,158,211,33,12,
6,164,195,85,242,116,200,69,30,11,163,53,
110,183,18,37,114,175,99,30,0,148,110,111,
26,20,210,50,142,127,242,203,235,78,60,38,
78,192,116,49,69,37,167,30,11,169,106,217,
42,139,237,11,156,121,65,70,172,20,24,249,
1,203,166,67,138,103,219,223,64,244,134,44,
111,70,100,74,151,146,141,125,238,235,172,126,
220,71,154,172,80,173,182,233,183,175,194,244,
135,252,52,159,195,233,95,165,114,230,69,34,
239,116,28,51,163,205,91,52,10,95,36,85,
108,178,185,242,87,152,195,135,76,93,252,6,
177,169,33,149,211,76,198,93,140,137,77,246,
36,182,92,163,141,32,136,196,178,142,18,187,
92,170,40,200,146,207,218,218,144,123,195,24,
127,231,176,116,218,35,44,195,225,133,235,53,
52,77,78,210,87,227,56,97,72,221,215,136,
70,5,212,172,128,32,196,137,135,178,171,139,
60,243,83,211,117,28,105,45,255,34,235,233,
84,35,138,147,144,109,253,41,155,201,73,162,
192,165,86,150,187,155,70,130,79,28,101,12,
7,17,247,182,77,46,100,124,10,105,5,69,
203,210,168,255,46,221,251,127,198,133,121,7,
255,20,85,243,163,246,135,12,236,239,114,221,
125,64,107,185,203,210,236,43,104,241,247,41,
102,108,2,91,96,121,113,22,53,211,68,146,
50,148,166,94,66,16,69,236,97,27,61,95,
165,56,251,28,230,246,42,154,88,62,16,178,
157,42,234,108,12,79,207,215,52,180,124,70,
182,79,171,168,51,85,212,89,168,101,184,103,
246,79,77,145,246,84,20,30,251,212,35,8,
18,177,37,225,236,204,181,80,126,137,254,166,
135,149,121,159,161,237,81,179,122,16,249,68,
233,34,104,10,90,16,48,177,68,60,47,34,
159,246,48,199,73,104,108,50,12,169,79,41,
8,34,88,70,18,70,202,22,196,157,152,98,
140,143,139,136,64,48,215,68,18,5,196,187,
203,220,185,51,97,110,38,64,29,155,196,115,
77,194,40,70,148,4,162,40,217,52,135,54,
196,226,42,68,17,78,32,161,72,17,138,24,
98,250,50,122,46,67,84,43,130,32,16,94,
89,32,174,22,137,243,25,98,73,98,125,88,
166,94,54,9,23,166,217,254,193,50,147,193,
26,125,46,32,237,241,14,37,159,19,14,12,
58,127,250,35,132,56,162,156,205,162,137,247,
72,137,54,170,12,138,227,226,56,1,166,25,
80,122,121,10,215,207,208,95,95,167,247,216,
11,142,128,129,181,239,7,39,188,7,215,134,
251,99,88,151,37,26,121,21,221,31,30,25,
19,85,165,171,124,100,173,49,157,245,233,186,
62,241,86,17,73,76,157,216,40,254,96,164,
230,100,26,17,207,30,147,210,93,198,70,25,
73,158,236,221,183,54,243,21,2,207,166,189,
248,215,104,103,151,49,63,62,131,154,158,165,
216,120,9,207,28,18,58,22,81,120,178,219,
219,93,253,9,86,127,149,66,43,1,254,51,
218,101,122,107,111,176,21,127,151,98,245,69,
226,180,72,45,23,48,105,175,147,149,203,143,
214,108,207,133,243,146,20,84,4,98,45,7,
145,207,76,94,65,140,60,162,8,114,74,140,
233,251,4,81,152,36,80,212,211,72,205,12,
65,24,19,69,137,2,87,68,17,85,14,81,
244,152,40,138,49,156,112,143,86,225,208,14,
147,20,4,115,27,66,151,88,201,28,29,99,
220,175,72,158,116,206,194,100,14,165,193,38,
74,24,176,122,90,11,242,211,80,120,124,218,
123,126,198,20,30,79,84,22,162,138,39,78,
78,77,125,179,159,34,102,62,13,203,182,117,
172,178,136,156,17,237,142,140,86,74,237,68,
61,100,196,161,64,168,12,120,121,186,199,207,
62,136,168,205,125,149,208,179,64,16,17,136,
15,112,153,245,134,127,143,124,174,79,173,27,
177,106,121,4,3,135,144,152,57,54,73,73,
17,213,23,99,238,250,119,176,157,8,183,253,
253,36,236,36,27,240,32,15,145,12,195,10,
146,148,162,113,254,81,93,70,85,130,92,33,
98,104,5,12,76,15,47,8,119,168,239,247,
173,215,99,251,46,6,54,44,232,118,68,166,
139,58,115,25,15,35,22,233,76,194,189,140,
197,147,21,197,19,40,60,14,75,94,19,68,
25,87,45,80,206,60,178,184,166,10,191,197,
214,146,194,15,133,183,152,153,123,145,11,195,
13,10,219,29,228,206,128,104,174,137,88,110,
144,201,137,135,207,144,8,139,119,60,166,231,
100,132,8,38,163,136,92,65,68,20,4,114,
10,188,185,81,98,28,24,124,115,166,140,182,
213,103,230,133,44,246,66,157,77,37,143,63,
40,81,11,219,228,11,18,195,143,59,100,46,
213,14,1,221,174,5,130,47,161,136,49,170,
20,225,249,17,113,20,35,125,124,15,73,18,
241,95,127,41,201,118,218,105,168,115,107,107,
26,93,241,89,27,86,184,154,222,192,113,67,
210,254,152,52,135,201,250,22,239,76,246,250,
41,135,219,30,62,10,26,62,154,2,245,166,
78,161,160,210,104,165,233,79,77,17,173,95,
102,24,231,32,149,126,50,216,247,132,247,224,
2,43,54,84,5,143,179,71,208,125,232,133,
26,43,119,63,96,69,188,79,184,149,161,148,
249,58,102,111,25,201,146,40,181,94,62,5,
222,120,50,141,136,32,137,108,91,22,129,157,
165,148,146,15,93,59,28,212,144,219,42,213,
185,175,144,41,157,217,187,103,160,24,168,217,
227,221,245,205,155,223,38,14,61,230,174,255,
193,163,123,138,96,14,151,72,23,166,208,11,
211,244,238,253,57,47,126,254,191,228,189,15,
55,104,61,78,155,225,78,72,167,51,160,164,
177,39,3,166,155,13,16,160,182,47,134,29,
133,1,215,202,173,164,128,79,56,72,61,22,
197,73,165,174,36,10,56,163,237,189,57,88,
94,200,71,107,199,243,77,197,144,164,184,238,
244,79,56,17,32,124,210,251,13,18,1,58,
42,201,9,70,241,207,65,225,241,105,239,249,
25,83,120,60,105,100,202,105,212,64,62,53,
245,205,238,25,217,29,87,138,49,155,67,151,
245,161,67,244,24,86,168,201,18,231,230,102,
15,132,124,166,95,248,55,220,223,250,33,51,
206,247,120,161,121,135,251,35,139,48,22,152,
116,77,4,207,37,223,144,8,133,27,248,254,
16,121,42,81,182,118,33,147,236,181,24,162,
190,193,210,91,75,228,46,156,97,60,157,225,
122,89,162,160,200,48,15,110,24,179,189,21,
17,20,76,250,147,58,37,87,103,230,210,111,
30,74,2,209,129,66,5,102,162,152,135,93,
43,137,82,156,98,189,60,224,129,9,69,124,
206,206,181,104,77,203,116,39,30,221,137,135,
174,138,71,214,43,125,170,208,211,82,219,68,
149,109,84,89,164,145,215,168,100,85,26,231,
190,198,202,221,101,214,206,173,18,54,11,40,
233,41,46,174,223,36,187,209,37,206,85,147,
83,255,152,162,8,3,232,119,66,62,254,80,
230,242,23,100,170,90,0,70,68,38,35,242,
219,249,16,59,28,242,206,54,252,229,122,134,
230,153,41,252,40,102,117,88,160,98,103,249,
165,165,31,146,173,106,68,19,147,180,27,35,
229,98,132,70,17,79,82,113,157,164,57,82,
94,246,17,148,71,89,2,91,155,14,158,16,
209,106,104,100,212,16,97,56,33,46,61,2,
122,21,49,100,161,210,33,136,36,222,250,89,
11,85,217,6,204,99,4,196,163,77,37,17,
145,209,3,166,230,138,200,59,49,108,109,166,
128,119,166,193,79,222,201,81,94,120,9,107,
221,224,89,142,158,225,161,246,109,102,142,168,
119,105,76,255,54,27,31,255,57,141,249,223,
32,93,156,161,63,254,30,122,180,72,224,94,
222,99,102,253,196,152,107,182,130,255,176,129,
116,121,149,241,157,15,40,180,46,30,244,104,
132,38,181,179,255,242,137,197,72,143,111,134,
124,253,50,153,234,220,1,207,71,148,53,180,
98,1,99,178,136,160,127,151,198,165,8,113,
60,70,150,246,103,33,68,73,99,30,45,71,
165,148,163,85,212,177,199,33,169,252,241,32,
231,81,209,34,81,16,142,36,173,76,171,18,
146,40,156,220,161,76,148,33,85,70,112,70,
196,90,22,4,233,144,128,20,71,183,137,245,
169,83,45,71,222,158,160,4,62,102,174,206,
47,198,241,67,149,69,106,57,21,103,240,169,
16,55,90,69,157,114,86,101,185,107,49,180,
18,175,55,173,73,204,22,82,135,112,129,56,
142,201,54,127,133,182,113,30,167,247,19,210,
226,50,47,191,36,240,143,63,42,81,190,240,
175,177,87,255,45,74,253,67,202,254,26,53,
211,198,19,83,148,37,129,78,44,240,213,207,
91,60,248,255,62,64,175,166,88,121,181,194,
189,109,248,217,86,142,180,36,82,210,5,252,
80,64,20,74,84,212,136,27,103,122,248,121,
147,123,203,255,142,124,253,155,164,11,211,71,
134,208,206,213,147,162,186,174,121,250,39,30,
218,62,126,16,145,146,5,234,121,141,250,14,
21,129,237,133,152,110,120,88,81,232,138,248,
196,248,248,177,218,41,136,240,130,8,195,9,
104,143,93,206,215,51,228,178,55,24,247,186,
180,203,62,191,84,139,144,165,10,157,94,72,
54,244,240,35,153,13,219,221,107,58,87,8,
21,36,25,94,123,61,197,95,255,185,195,96,
34,176,105,233,120,196,148,139,61,26,186,202,
210,196,38,240,243,244,44,5,199,171,147,85,
36,62,159,31,83,126,255,167,172,221,218,36,
155,18,49,140,128,74,85,131,158,75,110,170,
128,244,185,11,228,2,147,184,152,37,206,103,
16,204,71,209,184,185,185,196,162,119,2,240,
35,17,169,63,38,220,167,40,68,49,89,11,
57,116,185,58,121,143,219,161,140,67,142,52,
147,99,215,97,102,58,77,54,39,163,235,18,
73,171,227,128,73,165,73,239,98,147,55,31,
154,104,202,13,58,19,255,51,57,40,91,35,
151,233,82,234,144,224,211,50,85,22,94,248,
239,30,109,40,201,231,11,47,200,124,255,141,
127,192,46,124,25,199,15,201,234,50,205,130,
182,135,55,61,205,152,189,248,71,172,63,252,
19,130,212,50,129,103,35,239,115,183,235,23,
94,127,234,251,197,196,100,170,115,135,126,190,
246,222,95,162,148,33,82,12,38,101,131,145,
7,209,232,33,158,91,58,24,250,216,241,92,
54,134,14,21,197,133,200,199,53,134,135,35,
47,214,4,87,77,33,16,35,167,242,136,39,
244,166,30,219,1,97,28,147,57,134,95,72,
12,236,189,52,203,189,231,208,11,8,158,65,
28,199,7,201,0,101,141,40,127,1,193,51,
17,39,139,196,153,5,226,19,44,240,137,158,
69,250,69,43,212,39,239,195,114,234,200,212,
241,79,100,0,201,34,23,155,217,61,69,81,
72,41,184,99,247,8,35,67,216,49,152,154,
104,217,223,35,10,60,254,233,205,239,160,151,
23,16,68,137,80,125,133,121,237,71,68,63,
188,199,192,241,9,130,152,202,133,60,158,229,
241,209,127,248,144,233,179,69,86,94,61,195,
48,130,74,126,139,254,240,12,255,162,210,163,
148,141,17,70,38,81,179,156,84,128,74,10,
147,32,36,243,146,201,79,111,254,223,84,188,
111,144,175,93,58,114,238,11,181,52,117,49,
135,86,42,50,178,124,250,166,199,200,10,142,
12,199,239,142,91,155,19,166,203,73,33,222,
238,18,158,173,103,184,181,49,57,80,220,42,
3,92,155,41,240,225,218,24,247,83,46,178,
225,4,124,180,54,230,108,253,5,42,118,137,
237,219,223,197,120,109,72,105,161,129,86,141,
64,129,190,231,243,225,208,64,136,65,16,5,
94,212,75,100,227,24,43,140,168,182,4,6,
182,66,76,76,232,10,108,244,83,108,104,30,
82,37,164,51,78,243,133,217,14,151,242,105,
164,225,132,251,127,113,147,149,158,65,70,23,
168,214,52,242,5,133,254,192,199,152,4,180,
170,35,164,155,183,17,12,11,20,229,88,146,
32,93,14,25,187,10,185,229,45,228,222,136,
104,97,138,88,85,208,197,2,166,167,145,219,
220,36,39,58,228,241,16,49,142,180,40,179,
25,153,24,40,85,84,148,122,129,56,155,134,
213,54,227,106,139,224,98,139,119,250,35,194,
173,11,164,155,23,89,25,126,54,20,20,81,
28,19,19,63,177,98,56,12,20,254,225,182,
129,39,204,236,165,215,77,156,128,206,216,165,
85,208,105,20,180,19,51,42,226,40,60,80,
172,39,171,105,234,245,223,101,179,253,167,108,
44,253,239,20,203,95,34,223,124,254,19,61,
131,107,116,105,47,255,41,249,210,151,40,78,
29,172,28,95,248,252,191,66,16,68,54,22,
255,28,103,187,141,222,127,13,39,127,150,126,
119,245,81,24,118,39,54,45,10,137,117,164,
102,82,39,64,0,54,218,78,248,43,105,115,
121,124,229,243,196,241,89,31,60,194,148,100,
81,32,220,9,79,149,51,42,211,133,52,183,
218,54,143,71,183,99,53,11,196,8,246,144,
88,207,61,242,46,148,52,120,38,177,168,19,
11,39,199,222,51,174,137,18,6,76,10,191,
80,6,199,141,221,72,198,179,30,197,180,242,
84,215,139,178,74,182,245,165,61,140,36,219,
124,149,238,198,15,80,251,54,174,27,49,61,
157,230,238,173,33,98,24,34,8,2,238,11,
23,217,24,86,25,59,53,68,79,229,43,250,
59,108,254,233,91,104,11,57,242,57,9,92,
23,97,181,77,60,63,69,228,169,52,102,179,
204,156,51,89,223,248,127,48,150,158,39,147,
189,68,28,249,164,10,211,104,251,42,247,5,
33,217,163,149,172,74,37,171,18,70,49,67,
203,167,51,118,143,196,96,130,48,102,185,103,
177,61,113,89,168,101,200,104,18,41,69,226,
185,233,60,107,125,123,47,233,70,222,189,121,
49,173,208,30,126,250,5,14,162,152,187,91,
6,115,149,25,206,92,249,111,120,239,141,255,
200,248,210,58,207,205,233,152,147,136,237,208,
67,18,4,124,67,0,57,226,125,83,97,57,
234,51,242,2,50,185,12,81,187,68,177,250,
10,122,173,137,164,106,152,131,101,122,91,127,
129,152,138,8,227,152,158,23,160,135,58,133,
84,128,94,145,41,20,85,114,57,25,81,16,
72,23,179,100,213,128,56,138,17,38,59,232,
112,24,17,107,74,34,66,119,72,10,247,99,
60,154,44,48,118,36,10,88,72,31,46,37,
30,71,234,62,139,238,57,138,209,26,253,129,
71,179,34,176,37,230,56,51,37,113,243,230,
136,104,71,211,106,170,200,153,243,121,4,93,
161,125,127,72,243,229,75,244,245,28,170,90,
64,106,230,120,119,52,198,90,205,49,61,247,
91,108,15,44,194,232,179,75,169,60,141,69,
149,87,94,67,138,106,108,72,149,196,90,217,
21,156,81,204,218,192,102,115,228,208,44,232,
180,138,218,222,58,185,65,184,151,82,23,71,
33,15,222,250,63,41,182,158,39,87,191,140,
162,101,72,229,27,200,107,117,190,242,252,61,
238,111,252,25,253,45,155,66,243,149,167,243,
36,226,8,179,253,127,241,219,103,86,121,195,
252,123,58,247,86,73,231,175,239,219,252,59,
217,117,231,191,137,213,127,153,204,243,179,44,
110,140,137,133,21,194,32,66,86,196,29,208,
91,226,92,61,179,55,223,83,133,45,178,37,
60,115,4,90,142,173,145,75,16,69,84,50,
42,133,29,65,145,211,21,32,81,20,178,36,
112,99,190,136,23,70,12,12,159,114,86,65,
145,68,46,53,101,22,31,12,240,142,8,103,
196,169,98,210,251,88,217,105,125,186,251,204,
153,217,39,206,205,86,83,120,159,178,195,221,
252,131,183,72,71,17,253,92,153,250,176,205,
122,101,150,249,205,59,60,104,44,176,208,91,
231,65,227,60,11,237,165,99,255,239,139,49,
203,23,127,245,231,82,73,100,117,153,185,74,
250,231,114,110,130,32,18,7,224,251,59,204,
217,118,128,109,39,222,161,32,10,140,205,22,
243,21,149,231,154,107,220,121,88,70,124,227,
38,97,186,194,195,85,131,153,150,72,105,171,
135,224,135,220,255,187,155,196,122,138,114,39,
197,165,151,175,178,89,26,33,232,109,6,241,
219,32,129,177,221,128,205,34,149,250,151,73,
229,167,142,12,75,237,42,13,203,11,217,26,
57,244,12,239,144,237,108,121,33,55,55,198,
92,110,229,200,233,50,154,44,114,174,158,161,
89,208,15,98,20,133,180,76,251,25,46,212,
230,208,161,145,47,80,155,249,10,131,209,144,
239,191,249,38,177,55,194,22,90,200,66,147,
124,254,18,145,237,227,152,171,12,186,125,20,
170,196,234,12,179,87,15,186,85,185,218,5,
156,135,175,209,247,63,228,225,196,97,121,236,
80,139,234,92,248,234,53,102,238,44,35,200,
82,226,57,28,0,90,147,44,166,168,94,225,
225,118,145,137,169,50,85,30,81,27,175,176,
237,21,89,187,25,162,69,46,196,49,138,228,
147,23,135,136,174,189,35,112,99,54,151,186,
244,83,117,110,142,71,216,86,200,204,153,34,
162,34,225,198,50,162,166,17,89,137,224,208,
116,9,203,151,17,94,185,74,37,186,131,247,
246,3,226,207,93,230,227,172,0,147,9,67,
35,36,203,43,40,169,34,221,213,225,169,43,
119,79,183,17,143,86,18,251,139,119,250,134,
71,123,236,34,75,2,178,40,224,71,117,70,
102,41,9,137,28,101,105,71,49,235,3,155,
158,225,114,54,13,166,27,112,107,115,194,197,
102,150,66,74,65,148,85,228,114,128,93,251,
11,70,171,63,162,85,255,67,244,124,3,145,
52,63,9,66,94,104,130,213,253,30,214,104,
154,116,225,116,113,248,73,247,46,131,246,247,
249,122,235,1,210,196,227,198,57,159,239,220,
188,75,154,107,71,30,190,76,37,17,176,41,
93,225,242,243,179,200,59,196,112,186,44,113,
117,42,255,212,89,170,187,74,104,123,236,178,
53,74,222,107,74,145,40,164,21,226,56,70,
87,196,3,214,87,123,236,210,200,107,52,10,
143,168,101,51,154,196,229,86,22,71,201,242,
176,107,29,206,81,215,139,224,153,8,161,67,
172,230,78,61,55,205,119,145,163,144,209,167,
216,39,53,163,207,7,151,94,199,83,83,180,
235,231,147,125,177,227,65,189,91,75,190,127,
183,52,125,236,255,47,126,240,45,150,127,30,
113,9,73,228,66,35,243,89,100,37,63,195,
51,154,96,164,81,4,227,241,190,176,115,237,
28,207,85,86,161,85,6,4,44,95,71,174,
191,76,48,94,100,44,212,9,54,218,116,58,
29,20,69,102,123,20,51,149,23,153,200,121,
114,178,200,171,149,60,130,105,82,40,85,8,
162,136,168,234,147,206,245,248,248,254,159,176,
121,255,121,242,197,147,201,62,43,217,132,170,
230,94,199,60,100,216,196,49,116,198,238,129,
42,238,140,38,29,84,20,121,93,65,83,68,
156,103,180,72,126,24,177,218,183,169,0,249,
250,21,224,10,73,110,200,193,55,91,32,161,
216,118,131,136,238,230,26,70,223,70,145,196,
3,7,177,58,255,5,194,135,10,74,91,68,
20,20,6,246,38,63,172,188,201,151,95,185,
74,110,99,27,105,71,81,136,66,76,24,11,
72,196,132,126,140,57,137,152,220,27,242,82,
252,62,27,157,89,110,11,117,252,72,33,47,
247,144,136,118,132,163,138,65,29,35,138,169,
43,61,228,112,64,119,36,18,101,36,60,69,
212,10,219,0,0,32,0,73,68,65,84,52,
60,2,54,58,33,87,23,64,211,98,206,76,
41,44,174,64,46,171,80,171,107,20,82,1,
147,197,53,132,139,211,164,239,220,71,77,139,
252,108,224,227,135,49,194,242,89,106,87,191,
200,74,207,194,13,66,80,158,221,38,188,216,
72,226,168,29,99,175,201,31,221,123,63,33,
215,184,140,150,45,17,199,176,58,176,113,247,
97,79,130,29,16,103,142,216,29,143,157,54,
199,143,216,26,57,148,245,136,56,134,197,45,
147,43,83,57,50,154,68,181,250,235,12,187,
255,129,95,123,53,224,141,247,254,61,25,239,
155,100,242,151,24,182,111,242,102,221,165,217,
244,153,44,221,63,149,162,176,71,27,248,163,
191,227,11,23,198,24,111,117,120,239,218,231,
88,251,142,198,244,213,223,65,16,78,94,172,
102,65,103,60,83,194,112,18,238,174,211,18,
39,30,41,116,50,5,250,27,43,32,229,118,
48,170,164,67,98,207,12,208,164,131,184,210,
74,207,34,171,201,123,7,232,241,112,197,133,
70,150,143,215,199,71,125,72,146,248,96,245,
193,183,16,60,227,16,56,186,87,157,27,6,
196,178,138,47,43,68,161,248,169,246,201,91,
215,190,78,173,179,68,58,140,88,126,44,225,
224,52,227,65,109,254,231,82,8,207,86,82,
159,8,87,251,231,28,126,32,160,107,18,65,
16,16,4,137,113,166,235,18,87,167,122,132,
249,164,134,43,136,36,84,41,224,226,21,131,
205,181,1,189,174,77,227,218,60,153,153,60,
177,233,50,59,149,69,104,86,48,140,152,88,
146,168,75,18,99,59,68,21,5,84,81,194,
36,66,0,158,63,171,19,179,200,214,154,66,
186,220,58,114,62,183,55,39,79,36,213,244,
195,163,61,88,121,191,133,58,93,74,113,223,
122,118,188,255,91,35,135,174,61,33,23,228,
72,41,18,170,44,34,75,2,146,152,88,184,
142,159,144,222,77,156,0,219,11,17,76,135,
216,119,118,64,68,159,11,205,236,158,213,215,
92,248,124,114,160,194,128,254,106,132,219,173,
147,22,151,144,214,59,123,159,151,86,66,122,
150,138,82,74,161,140,108,182,228,42,23,46,
108,18,247,75,52,106,50,173,254,58,248,1,
209,181,6,113,186,128,48,54,144,110,222,71,
112,18,221,26,34,97,154,33,2,1,195,176,
132,230,135,168,120,56,150,64,24,39,46,152,
101,5,220,184,154,66,148,69,144,4,70,173,
51,248,85,157,64,83,80,62,127,141,247,29,
31,223,17,16,87,174,48,125,233,247,25,90,
1,91,35,247,153,110,192,102,65,167,144,86,
24,236,0,110,187,209,137,234,185,95,218,187,
198,116,131,3,74,226,84,174,201,190,49,176,
124,148,157,152,102,20,199,44,182,13,158,159,
206,147,46,206,16,248,191,201,59,183,255,134,
215,111,72,188,117,231,207,240,156,95,66,231,
37,60,247,13,214,99,23,53,30,31,47,4,
1,107,180,193,176,251,51,124,113,153,215,95,
246,49,254,236,38,90,41,69,181,233,208,91,
207,161,23,154,184,79,232,27,46,75,2,87,
167,114,108,79,92,4,4,228,193,230,39,94,
207,238,196,195,17,82,73,161,157,154,97,115,
224,176,53,114,112,199,54,184,214,33,189,250,
176,107,241,220,244,209,158,65,70,59,41,59,
74,72,114,220,149,244,14,134,241,152,206,62,
160,69,7,72,97,128,252,12,192,236,237,250,
121,74,195,77,102,215,63,102,117,250,185,167,
250,219,133,237,101,222,107,61,247,115,39,132,
127,158,149,68,20,120,152,43,127,204,185,193,
93,58,230,65,92,160,213,74,17,55,203,196,
153,20,182,175,178,54,168,112,237,204,38,182,
83,102,234,95,253,50,210,127,252,49,189,7,
61,26,95,189,68,188,131,23,90,97,120,224,
156,166,179,34,195,126,72,58,35,18,69,49,
97,16,99,140,67,174,158,145,120,176,114,23,
248,26,71,165,238,233,138,132,17,6,39,206,
253,184,176,237,129,160,121,74,17,57,91,203,
176,246,152,37,250,105,70,24,193,192,244,25,
60,37,183,255,192,242,49,156,224,16,107,170,
32,201,84,206,188,204,248,195,183,176,38,48,
178,52,50,202,78,255,138,88,68,211,69,226,
235,103,145,179,50,214,3,9,253,172,206,32,
119,153,206,184,128,88,140,137,34,129,92,228,
208,16,135,196,197,92,210,5,111,71,81,72,
132,172,175,219,140,149,41,106,227,183,89,245,
231,153,229,46,2,49,91,27,54,197,178,194,
118,199,37,157,81,200,124,237,26,113,58,197,
168,31,178,22,153,20,194,136,59,94,192,198,
166,136,62,254,2,205,43,95,195,11,98,238,
111,143,159,233,38,148,68,129,169,199,120,109,
162,56,38,140,226,3,64,180,233,125,74,1,
19,39,45,19,119,135,23,68,220,109,27,92,
110,101,201,215,174,48,106,7,124,239,237,191,
39,127,198,193,114,126,140,108,183,224,246,85,
194,218,10,182,185,246,152,46,74,230,53,218,
186,137,34,229,24,77,126,196,133,23,86,153,
248,33,218,187,27,220,223,152,176,160,248,156,
183,93,182,171,219,184,230,241,96,89,20,6,
7,138,170,106,185,196,243,116,162,26,158,53,
222,21,199,196,128,154,62,29,183,149,40,146,
100,31,137,33,66,224,224,161,195,9,203,87,
122,2,208,249,76,162,33,169,18,225,184,131,
40,63,155,112,229,160,216,34,150,21,26,155,
183,24,20,26,120,242,233,186,50,61,168,205,
253,92,10,227,187,91,6,173,162,78,61,175,
254,220,41,13,179,243,14,215,27,109,250,31,
39,198,142,44,11,52,102,178,20,207,213,80,
139,233,68,1,196,144,82,60,46,212,55,49,
39,17,174,21,145,170,201,20,47,212,89,253,
97,135,168,51,64,104,150,233,123,62,183,71,
22,47,166,114,4,126,140,231,198,68,17,168,
170,136,27,168,196,130,71,224,199,20,43,73,
82,141,126,166,75,111,245,13,42,179,191,116,
104,94,211,37,157,123,109,243,88,78,185,93,
175,248,137,138,98,55,134,85,201,170,244,77,
143,149,158,253,212,156,32,207,114,108,79,188,
35,233,181,125,107,66,185,50,68,30,77,16,
20,208,207,86,17,58,3,172,73,136,246,252,
28,113,46,249,155,145,147,227,222,182,68,62,
101,115,190,190,245,200,130,52,242,244,204,28,
149,76,82,63,33,140,12,194,40,230,193,3,
147,177,13,25,182,17,125,7,133,34,54,58,
2,17,131,161,199,96,39,107,201,114,66,210,
59,89,53,75,147,152,145,226,209,118,60,162,
78,145,146,244,27,20,206,61,71,12,220,235,
24,207,188,127,66,171,168,239,213,15,236,215,
254,94,16,29,104,17,105,185,193,51,127,31,
134,19,112,103,211,224,82,43,75,161,113,13,
255,225,136,170,255,29,178,13,139,117,150,136,
134,42,74,251,57,244,236,204,49,126,237,79,
169,41,93,206,158,213,80,66,157,242,196,96,
237,131,109,90,117,153,92,86,66,232,13,185,
186,48,199,183,255,230,223,51,115,253,15,49,
186,75,140,39,31,32,160,33,145,39,242,61,
124,211,96,246,165,223,57,58,132,244,152,98,
216,85,28,137,240,22,80,210,71,123,1,229,
140,74,45,23,176,61,129,216,30,36,180,223,
39,140,252,147,200,13,133,167,248,85,124,124,
215,178,73,105,154,203,247,223,162,87,121,54,
194,122,152,173,34,166,75,220,120,255,91,188,
245,220,151,79,233,81,172,240,94,235,249,159,
59,69,17,197,9,158,182,49,180,169,229,52,
230,42,169,159,139,121,185,163,37,234,226,119,
233,254,104,145,254,32,162,88,171,35,94,189,
198,104,97,138,74,173,67,16,195,221,137,197,
184,63,70,245,52,28,187,206,84,161,199,72,
115,208,204,52,241,212,52,47,255,126,1,106,
69,220,48,226,157,254,132,172,44,145,206,138,
252,255,220,189,233,147,93,249,121,223,247,57,
251,114,247,173,119,116,55,26,235,0,179,146,
156,33,41,82,164,54,202,150,202,114,201,178,
202,118,36,39,149,23,246,43,199,169,84,222,
248,125,254,133,164,28,167,146,84,37,85,138,
229,138,165,68,229,178,28,217,20,41,82,148,
134,219,204,112,102,48,3,96,0,244,190,222,
238,190,251,61,251,242,203,139,115,123,67,119,
3,141,1,56,164,245,171,234,66,163,251,244,
185,231,252,150,103,249,62,207,243,125,60,39,
197,180,100,218,94,145,106,174,135,44,69,220,
221,185,196,173,210,6,251,65,196,186,235,19,
169,9,74,124,118,0,161,100,105,188,122,169,
196,122,219,59,151,84,48,171,141,147,14,141,
206,3,116,73,5,232,222,249,8,77,150,177,
227,46,230,32,195,78,167,128,177,84,240,104,
215,33,124,28,183,26,81,120,160,94,128,38,
216,109,35,236,193,197,130,63,143,93,171,236,
43,152,238,17,168,30,133,62,154,110,226,247,
182,240,175,122,236,206,207,81,43,216,164,113,
130,186,222,36,73,85,228,254,144,120,188,138,
36,129,38,69,44,212,155,39,152,102,37,9,
234,249,62,159,52,167,168,229,6,72,253,33,
169,16,52,183,125,122,221,144,20,3,3,159,
24,153,52,167,211,46,126,141,184,187,65,221,
91,199,102,128,4,132,169,76,42,4,178,36,
145,147,4,61,64,116,52,198,58,175,102,228,
100,171,15,217,236,248,40,78,200,97,13,178,
223,3,179,251,169,230,225,80,16,202,50,243,
114,30,121,20,221,172,5,49,131,157,13,132,
61,64,139,109,204,99,237,33,213,166,67,57,
74,94,216,90,156,128,20,183,20,230,235,57,
166,69,131,197,187,117,126,229,250,38,219,154,
6,229,144,56,188,71,174,91,194,92,125,120,
58,38,225,134,184,215,19,66,225,210,232,14,
225,7,43,52,74,80,249,181,215,16,249,28,
159,132,49,15,127,160,240,202,220,223,162,191,
242,49,27,221,119,41,214,234,72,66,34,145,
183,81,163,28,179,229,55,144,207,184,119,220,
217,193,40,119,70,107,61,82,164,143,93,19,
6,75,217,220,12,90,200,251,187,39,164,246,
130,128,49,63,102,16,196,196,189,13,20,213,
64,9,134,4,218,30,94,144,32,164,35,108,
200,12,114,200,170,116,194,251,82,220,30,178,
189,149,29,200,230,48,179,218,132,64,28,87,
2,105,130,72,67,212,148,195,185,85,68,196,
120,165,64,179,23,28,246,242,56,126,206,146,
48,164,186,126,143,244,34,30,192,5,247,88,
75,179,41,239,172,92,40,185,98,87,53,40,
111,47,93,64,66,14,179,251,61,11,133,199,
5,62,255,34,123,54,4,154,134,202,164,238,
28,202,175,167,222,119,216,193,232,62,125,174,
218,221,53,114,106,225,66,215,74,110,31,75,
254,19,228,143,62,96,224,196,216,186,194,228,
231,26,132,51,128,182,203,31,63,44,82,211,
101,48,187,12,227,24,165,99,16,21,87,184,
219,28,39,77,74,108,230,52,116,165,140,162,
134,108,45,198,20,114,187,8,82,252,36,229,
78,119,200,184,169,19,167,10,63,222,21,188,
50,17,50,101,233,148,44,151,191,88,179,113,
148,181,204,152,91,181,193,249,136,230,251,107,
84,181,107,152,70,1,77,59,121,18,94,2,
230,146,132,141,253,29,124,217,190,208,154,169,
0,139,122,198,78,169,199,130,249,210,36,197,
99,238,199,68,35,225,147,237,193,73,119,37,
116,65,86,200,229,114,24,154,76,199,9,207,
141,107,72,142,129,200,93,140,253,242,241,107,
7,192,165,137,242,137,172,158,0,216,119,223,
35,213,19,86,34,137,222,208,227,102,222,70,
154,168,33,55,93,132,157,5,53,63,89,175,
114,101,114,255,132,177,118,34,83,72,30,241,
246,88,6,195,97,204,246,142,71,128,129,76,
66,128,78,91,170,81,254,242,63,35,87,159,
101,245,207,254,7,12,207,57,148,43,114,28,
211,234,12,88,21,16,10,27,66,9,179,247,
38,185,107,95,203,168,53,90,46,187,86,0,
199,207,129,103,131,85,249,84,243,112,48,230,
235,54,81,225,72,96,136,56,101,48,76,16,
185,58,78,35,143,150,59,90,183,253,160,123,
170,208,230,121,214,226,132,165,11,132,186,206,
229,122,142,137,224,247,248,120,235,127,230,141,
89,159,119,145,144,27,62,157,78,15,123,242,
12,43,120,105,138,226,112,135,201,59,75,244,
246,61,166,111,214,72,110,94,102,77,85,216,
30,122,236,172,105,52,234,127,139,160,50,71,
215,253,115,114,181,89,36,163,200,112,127,145,
242,196,155,212,102,94,59,23,192,28,196,93,
22,151,254,148,107,211,109,162,94,66,48,245,
155,148,198,142,210,116,99,223,65,53,179,103,
74,219,219,136,51,2,126,249,209,215,145,220,
205,40,60,146,84,208,30,134,116,220,8,75,
83,208,106,214,41,190,167,180,183,135,24,81,
66,244,25,205,253,40,230,113,160,77,164,221,
31,33,138,215,145,146,228,112,110,21,89,226,
202,108,25,45,78,216,236,248,116,221,236,44,
149,108,141,126,175,199,160,54,197,237,181,247,
89,175,207,115,169,181,198,102,109,134,137,206,
14,123,197,6,181,65,139,78,174,76,222,119,
240,53,19,45,28,144,202,114,22,10,73,19,
66,213,32,23,56,12,172,34,101,167,67,171,
80,163,209,223,67,73,67,214,106,83,25,229,
198,83,198,194,206,71,124,80,159,190,128,146,
26,128,110,61,3,133,71,114,161,207,191,240,
158,181,52,42,250,0,233,156,64,238,105,157,
106,193,83,232,62,68,154,178,254,240,63,49,
247,234,111,65,249,233,180,242,189,141,63,224,
23,38,19,238,189,147,25,104,165,146,138,60,
55,197,174,171,226,4,30,5,225,210,147,247,
217,219,155,68,145,19,68,42,81,247,161,81,
202,12,140,8,8,82,133,36,86,176,173,132,
173,246,37,76,221,165,146,223,99,213,241,89,
117,50,79,97,189,95,35,49,6,60,232,43,
152,74,159,221,200,36,116,199,240,35,139,241,
249,117,144,92,188,97,139,229,7,123,104,185,
20,195,191,66,125,252,171,24,246,17,117,142,
2,140,151,243,72,102,9,55,134,182,19,209,
117,194,115,41,40,85,128,100,212,185,206,15,
53,238,181,35,106,161,196,92,205,70,85,36,
76,29,174,204,234,44,239,185,89,192,25,208,
148,148,137,74,142,137,90,86,9,52,21,165,
108,116,188,51,91,236,73,154,142,208,46,134,
135,158,117,173,35,84,242,143,53,82,215,204,
58,126,12,154,33,241,106,20,193,246,30,98,
126,10,209,93,71,84,139,188,255,104,140,98,
41,69,40,6,60,150,199,37,245,29,80,100,
42,246,144,174,155,163,154,179,49,77,229,16,
22,144,16,36,40,160,104,152,19,87,8,135,
187,164,126,64,138,66,136,130,78,132,170,64,
99,232,242,19,77,199,107,73,232,195,55,24,
191,249,183,137,211,44,208,217,242,128,199,223,
57,214,79,255,236,25,230,193,210,21,106,181,
226,137,133,212,116,72,244,236,218,72,211,17,
122,230,225,197,137,32,58,195,219,123,222,181,
56,97,105,250,144,11,160,81,168,177,189,249,
50,47,237,124,155,198,116,133,189,36,37,40,
61,192,245,191,130,85,60,121,184,26,115,223,
160,185,255,49,115,186,66,233,23,175,241,253,
98,30,215,143,80,165,8,33,36,44,255,243,
228,22,110,16,184,125,230,138,59,76,41,43,
180,10,37,214,42,191,67,174,118,147,152,243,
41,150,173,250,60,11,211,55,233,172,255,33,
37,123,13,223,125,143,130,246,21,154,143,190,
77,34,58,180,227,38,227,118,14,153,5,130,
97,159,233,137,249,167,135,105,116,3,161,27,
200,64,221,52,169,159,21,116,126,236,218,84,
28,155,123,17,157,92,243,233,175,129,223,67,
146,195,195,185,77,128,13,39,101,166,106,179,
144,179,51,74,27,33,80,36,137,119,92,15,
33,224,195,151,126,25,73,8,134,213,25,18,
89,161,87,157,37,149,21,218,141,203,164,146,
204,190,16,164,146,132,228,181,79,240,252,8,
64,30,253,110,79,164,135,127,147,196,65,230,
233,92,96,47,44,78,92,57,148,15,79,14,
70,6,217,253,46,162,40,36,50,174,167,11,
220,247,34,123,86,146,96,102,162,72,58,12,
17,250,197,246,247,193,122,61,109,220,248,181,
127,134,223,219,187,208,181,170,158,160,168,89,
243,52,85,149,201,127,241,10,171,65,204,198,
176,198,74,168,83,207,239,210,111,222,70,151,
6,200,134,79,217,8,105,39,54,105,47,79,
144,24,116,7,13,242,70,159,74,97,143,98,
174,205,100,117,21,55,200,179,219,153,97,172,
146,197,253,188,48,135,31,230,248,100,227,117,
84,37,162,104,246,144,228,132,246,96,140,171,
211,119,144,165,20,132,196,110,123,18,249,178,
96,172,178,73,204,251,108,110,47,98,183,222,
196,46,94,34,87,91,64,150,101,132,110,160,
217,22,101,69,165,92,201,224,235,102,63,96,
183,31,156,74,198,56,115,85,91,195,144,158,
23,49,83,177,40,89,26,121,67,229,149,153,
34,113,34,80,21,9,111,168,16,10,153,189,
65,128,170,200,148,109,141,43,99,57,38,203,
38,61,55,202,186,69,249,49,193,11,136,111,
180,157,232,84,156,162,49,251,117,86,31,124,
66,249,115,93,232,244,25,222,109,34,105,123,
24,99,54,82,111,136,44,165,228,204,136,230,
160,68,123,152,167,96,123,84,237,33,185,230,
6,210,39,235,200,82,202,248,252,20,247,245,
215,169,47,109,145,144,113,184,69,169,140,134,
138,66,66,46,233,211,252,247,255,2,201,115,
153,13,150,145,71,7,47,192,32,20,10,82,
16,18,197,58,37,126,145,234,205,55,233,123,
17,75,123,238,133,98,58,138,44,81,201,105,
36,105,22,132,14,162,244,204,185,146,164,44,
95,92,87,101,166,43,214,153,231,45,167,43,
12,31,195,190,131,207,40,174,180,186,239,81,
48,85,10,213,207,241,222,222,29,110,119,87,
216,191,57,133,50,225,208,92,251,55,88,173,
219,140,95,254,181,163,247,214,77,118,218,211,
252,164,40,177,43,5,68,221,16,217,132,75,
182,201,238,138,70,109,42,11,192,245,247,62,
230,74,218,102,147,10,194,245,233,122,239,17,
152,11,44,239,185,88,186,194,124,221,62,145,
235,125,100,64,20,144,173,183,88,87,134,204,
55,246,216,184,243,127,32,43,18,95,158,187,
203,183,163,121,74,150,195,107,218,183,88,139,
115,172,175,108,18,37,179,148,26,159,59,165,
208,158,103,156,136,73,9,145,113,80,29,143,
67,28,254,236,104,141,182,58,46,150,42,81,
205,107,135,214,158,27,36,136,131,116,54,145,
34,36,153,104,84,176,151,142,130,183,135,255,
30,186,201,202,41,24,33,61,252,87,57,250,
155,228,226,12,1,63,175,89,79,199,199,108,
213,198,214,21,220,231,184,199,121,205,131,158,
101,68,190,140,215,118,136,99,193,244,235,211,
188,107,44,224,250,29,132,253,144,134,37,179,
185,191,64,177,184,78,201,238,16,183,39,233,
244,26,180,80,177,141,33,182,57,64,45,38,
52,74,91,172,238,94,199,13,115,32,36,106,
197,29,82,33,179,214,188,142,161,123,116,134,
117,110,204,188,143,36,9,6,131,75,124,105,
166,195,251,173,24,69,146,50,185,167,42,172,
180,198,25,159,222,100,127,120,228,93,41,147,
3,252,228,219,184,158,196,230,131,25,108,205,
128,88,160,72,21,26,51,95,67,183,43,232,
170,204,165,170,197,68,201,100,117,223,165,237,
132,79,86,20,7,27,126,101,63,155,122,93,
149,201,25,10,81,34,240,194,132,196,31,102,
150,195,8,55,45,89,26,147,101,147,130,169,
98,151,149,195,243,208,236,7,108,186,79,76,
32,121,234,104,246,124,20,57,11,226,30,64,
80,178,162,82,175,253,109,22,127,242,159,152,
120,197,164,248,86,30,227,227,135,176,239,177,
146,187,194,84,185,143,27,23,153,46,181,105,
14,74,84,237,33,61,215,230,190,251,58,5,
189,134,210,235,81,91,233,115,163,248,61,30,
245,47,113,213,92,68,83,100,58,242,56,133,
164,143,41,134,20,69,151,226,99,184,164,4,
148,172,152,177,70,1,79,181,145,214,167,168,
222,120,147,97,16,115,127,251,226,100,127,179,
53,235,48,91,39,137,67,100,73,38,149,20,
156,32,33,78,83,196,208,161,80,47,161,95,
32,155,99,182,102,227,233,246,137,108,133,207,
42,1,33,21,130,165,61,151,121,83,67,189,
244,79,217,94,255,31,121,73,150,184,155,10,
164,169,125,210,205,211,64,209,248,252,223,67,
210,117,242,173,85,132,72,24,108,223,97,149,
46,213,210,47,161,154,25,240,99,21,198,249,
94,127,1,61,185,6,210,50,138,174,19,141,
4,176,23,38,108,247,124,192,68,83,179,236,
145,48,78,73,195,172,185,76,105,226,54,155,
223,252,51,90,253,49,94,174,62,100,188,191,
205,198,127,220,231,75,246,35,152,156,230,71,
179,175,113,169,22,240,213,47,248,236,135,119,
88,94,251,128,245,71,211,212,234,191,140,93,
158,57,71,136,164,25,61,202,25,130,36,77,
98,98,127,72,228,14,144,85,13,55,72,32,
112,50,168,41,28,34,30,19,220,82,18,142,
104,208,131,19,190,201,198,94,68,73,47,28,
10,44,103,24,30,209,112,7,105,86,213,173,
125,246,85,200,107,181,153,159,107,37,49,93,
177,78,212,92,61,115,240,57,78,113,130,24,
63,74,71,6,91,66,156,28,163,250,20,96,
196,30,211,102,114,102,237,204,210,158,195,100,
217,36,222,255,30,211,234,58,143,254,98,153,
82,73,167,117,227,42,171,67,40,230,4,58,
32,82,25,4,200,82,74,107,48,142,20,152,
36,192,116,109,25,219,28,48,240,202,232,106,
128,44,39,204,54,30,178,210,188,73,222,234,
177,215,155,102,170,182,76,171,63,65,42,100,
174,207,124,64,195,84,25,122,21,94,159,84,
232,37,41,221,225,24,191,119,211,97,217,177,
24,51,117,54,239,121,120,171,58,201,165,199,
246,158,2,74,94,224,212,83,222,154,219,38,
18,130,117,119,133,205,149,101,242,124,142,98,
253,54,70,190,134,166,72,92,29,207,209,113,
116,22,119,157,39,43,138,19,193,162,17,241,
223,147,38,219,143,18,108,93,57,204,200,145,
36,152,40,25,48,52,89,123,142,50,2,65,
214,193,106,167,23,208,40,232,140,23,77,12,
77,70,51,10,212,39,255,46,127,245,224,223,
240,185,57,155,218,173,155,12,215,60,58,174,
197,107,209,93,146,155,243,172,180,26,76,149,
58,124,235,147,87,184,86,223,230,218,165,22,
131,90,25,182,82,156,245,128,36,18,200,114,
74,34,105,140,79,90,172,237,41,184,90,149,
210,19,154,145,148,74,26,150,46,248,200,55,
177,140,207,141,224,209,172,70,228,34,2,90,
130,19,212,236,126,111,11,213,40,96,228,107,
135,217,52,110,160,92,72,73,0,200,210,81,
138,232,103,173,40,32,203,132,106,197,33,151,
102,231,112,230,254,59,156,71,255,27,95,153,
27,242,125,69,57,151,57,86,213,237,195,94,
19,101,78,103,212,228,170,11,236,237,231,145,
85,29,145,94,67,81,84,116,53,163,157,79,
133,160,235,68,116,157,147,74,72,246,134,220,
44,214,201,27,26,183,190,241,223,35,171,58,
91,173,21,86,146,183,153,120,227,46,214,254,
42,195,141,117,166,151,214,160,148,39,184,85,
103,107,170,202,84,41,225,230,236,54,43,31,
255,159,12,211,255,130,124,117,225,244,254,31,
118,89,253,241,31,115,253,87,254,233,233,223,
57,93,204,98,157,52,142,208,115,101,60,57,
2,99,212,148,203,56,157,158,43,84,3,169,
191,117,42,144,43,105,202,137,254,201,158,35,
193,65,106,172,162,67,26,253,76,4,241,84,
103,155,246,244,171,63,151,74,98,162,100,50,
93,49,159,235,30,139,77,135,225,83,50,4,
61,47,162,187,217,199,214,21,234,5,157,90,
222,64,83,36,54,218,30,251,131,144,48,78,
153,140,34,46,221,176,216,95,177,8,122,30,
221,71,54,191,19,252,128,123,215,223,228,227,
190,65,107,56,198,75,179,239,162,14,167,73,
84,7,81,111,35,233,17,2,129,31,218,196,
177,78,165,144,213,132,57,65,1,83,119,169,
151,182,65,192,230,254,2,141,210,22,121,35,
228,118,41,135,41,234,248,154,198,116,169,77,
193,211,248,145,144,209,100,153,155,197,28,219,
94,64,113,108,64,187,223,64,164,10,150,84,
162,102,249,36,2,226,52,165,29,198,128,194,
180,109,162,73,112,173,96,227,213,83,190,215,
252,22,219,123,223,65,110,78,82,200,125,129,
202,196,27,84,114,26,121,83,185,184,162,120,
210,208,20,153,91,211,5,212,115,176,227,162,
165,49,102,24,236,246,159,175,232,44,73,5,
59,189,128,221,126,200,149,177,28,22,96,149,
166,208,154,47,243,65,103,23,72,185,94,236,
115,101,253,199,12,114,54,106,32,152,47,237,
176,188,81,230,183,147,63,228,126,240,22,165,
229,251,84,154,45,238,229,223,228,165,220,22,
219,110,3,75,246,248,235,254,91,92,205,127,
130,183,169,33,61,37,245,49,10,4,162,156,
103,185,95,162,92,156,57,132,146,38,75,38,
171,173,167,59,191,214,168,32,235,80,40,214,
230,95,152,235,251,89,67,79,7,99,167,31,
80,242,98,138,249,6,225,236,127,195,71,15,
254,37,185,41,159,240,128,82,94,164,79,109,
224,254,248,208,211,25,10,229,27,228,170,71,
177,132,171,146,196,131,230,217,158,91,10,44,
237,58,188,60,83,68,30,197,8,114,181,121,
114,181,121,134,222,128,229,248,79,176,95,126,
200,245,212,71,124,210,197,127,208,164,254,195,
21,60,63,65,175,91,188,246,229,43,188,219,
255,19,118,22,175,211,152,251,117,146,208,101,
235,147,63,5,37,32,10,6,140,223,252,141,
35,193,209,109,98,149,198,8,221,62,70,161,
122,62,244,116,166,165,160,100,248,252,83,246,
250,137,158,0,63,67,158,138,157,103,232,239,
240,89,142,177,226,139,73,137,157,44,155,60,
108,94,12,13,112,195,132,181,150,199,122,219,
195,80,21,252,81,86,97,223,139,169,151,63,
207,15,239,62,226,246,47,190,70,255,143,255,
154,214,187,15,89,149,246,249,242,248,93,172,
177,175,49,28,255,144,94,103,154,56,191,5,
8,250,78,149,208,59,198,62,81,58,42,24,
13,34,139,233,250,18,142,95,192,241,74,76,
215,151,144,37,137,95,104,148,33,206,209,114,
77,230,89,68,90,117,168,175,55,185,124,249,
101,222,221,41,241,230,164,67,78,85,248,90,
53,165,55,46,88,11,60,26,234,56,215,156,
123,104,178,66,175,81,230,199,123,33,178,18,
176,220,119,120,169,217,34,188,62,203,150,23,
48,101,235,92,189,109,179,27,180,249,240,253,
239,82,225,117,64,66,29,25,172,207,173,40,
202,182,118,174,146,56,190,24,207,171,40,142,
195,29,203,123,14,55,70,1,252,210,228,23,
216,222,252,183,124,253,202,14,227,15,31,141,
122,171,250,244,223,139,145,94,154,64,221,10,
144,252,30,250,96,27,145,14,64,8,106,237,
37,6,146,205,32,201,51,107,110,144,147,92,
162,161,195,100,210,102,49,188,69,140,142,122,
6,197,219,216,152,201,216,152,73,90,43,163,
247,6,68,254,209,6,203,153,23,35,162,51,
84,249,156,128,220,139,19,6,159,121,237,139,
128,135,205,33,215,198,243,20,173,2,202,245,
127,206,254,202,59,132,233,22,43,75,255,19,
164,144,215,62,135,162,22,240,134,43,168,122,
9,211,158,192,176,235,24,249,202,153,130,176,
58,249,230,137,198,50,73,154,30,86,162,159,
15,151,166,68,113,138,161,157,156,99,205,42,
48,121,237,31,179,126,255,15,120,231,218,3,
196,181,57,202,165,152,121,215,103,174,217,37,
13,99,218,111,63,228,246,151,225,94,49,96,
245,221,54,11,111,254,62,99,115,191,132,144,
99,204,252,24,154,117,228,29,12,218,247,217,
93,254,54,147,215,127,243,148,215,244,34,88,
13,246,135,33,73,42,80,71,85,180,185,208,
195,149,21,204,48,36,208,77,140,200,39,84,
117,212,36,38,149,21,36,145,34,144,32,112,
144,81,73,100,21,53,141,137,20,21,61,14,
9,52,19,51,244,240,12,27,59,112,112,36,
137,92,224,226,200,26,185,96,136,99,228,72,
101,149,244,140,52,201,250,160,205,254,207,153,
146,168,229,117,230,235,47,6,134,171,228,52,
106,121,253,204,68,156,243,227,25,28,42,137,
67,232,112,168,240,210,228,63,230,227,181,127,
203,205,235,99,44,127,152,144,10,153,237,143,
154,228,111,110,50,181,80,226,81,109,151,71,
173,26,126,100,81,43,236,82,204,29,67,46,
132,68,107,56,70,154,42,24,90,198,59,55,
112,43,76,84,179,180,215,43,121,139,188,170,
240,97,115,156,91,245,21,212,191,254,36,59,
23,11,211,168,74,76,24,103,251,176,36,4,
202,195,117,54,175,222,228,74,46,162,168,118,
48,238,111,17,190,245,42,107,251,243,24,126,
142,235,181,247,168,90,6,68,49,90,187,207,
149,114,1,41,12,145,223,187,207,220,107,87,
249,184,208,39,112,122,232,118,153,254,232,204,
169,47,66,112,95,68,56,22,45,149,190,247,
98,138,192,226,52,131,31,114,101,176,11,85,
108,117,134,176,214,39,237,85,16,155,45,34,
47,66,146,135,168,119,30,17,182,167,25,232,
42,19,209,125,246,162,28,19,118,128,18,116,
112,141,42,41,18,94,98,98,74,62,195,216,
192,196,167,68,143,45,109,150,217,232,209,9,
248,75,2,226,40,69,175,216,184,211,13,220,
180,131,177,182,73,137,107,217,97,214,213,67,
104,228,73,67,145,179,107,118,251,1,181,252,
79,167,170,212,13,63,251,30,6,73,42,184,
191,61,160,146,211,178,4,136,75,111,210,91,
252,46,177,104,82,172,123,160,125,147,106,73,
101,95,10,113,226,20,215,135,120,211,98,106,
252,191,198,174,76,93,224,112,74,8,33,200,
25,42,150,38,147,51,85,134,126,68,123,24,
33,203,18,181,130,201,204,116,233,100,51,163,
199,32,63,41,202,19,237,169,200,197,152,174,
23,243,129,166,98,206,141,225,245,4,165,233,
73,62,191,184,206,228,120,76,171,250,58,200,
25,20,116,86,199,180,198,229,175,225,117,118,
208,115,89,214,223,222,210,95,163,25,37,204,
82,131,106,94,99,179,147,193,144,210,185,251,
65,58,195,80,56,250,222,214,21,10,166,74,
232,36,124,126,233,61,90,149,105,182,11,37,
106,189,61,218,86,142,114,224,51,52,11,88,
161,75,160,26,168,105,148,197,66,252,62,178,
230,16,170,6,86,232,225,152,57,10,222,128,
94,174,76,117,176,207,110,105,130,122,119,135,
56,95,166,222,221,38,82,53,234,237,45,242,
138,2,154,197,230,248,213,83,207,213,203,21,
127,238,148,196,66,35,247,66,239,57,95,183,
25,94,148,242,230,9,198,217,78,47,98,188,
250,27,236,26,59,204,27,63,100,41,184,138,
212,124,192,92,241,61,182,39,127,147,87,38,
82,2,95,37,213,182,112,226,147,103,116,179,
117,153,169,218,50,130,44,40,221,29,214,9,
162,204,99,50,100,153,171,18,168,239,222,195,
156,24,227,238,163,58,159,147,20,36,145,32,
15,92,44,235,10,213,130,159,197,64,22,215,
73,175,92,194,80,84,156,64,99,38,92,69,
74,98,182,31,194,75,99,31,179,90,186,202,
101,13,60,85,1,93,69,32,35,247,134,56,
111,223,35,244,34,242,55,35,66,57,37,137,
92,118,122,230,97,89,196,115,43,138,225,5,
43,128,235,121,227,133,41,10,128,158,23,114,
144,221,157,47,222,228,254,246,59,76,213,202,
180,62,220,194,117,99,38,47,105,236,238,166,
20,148,61,10,198,193,231,58,116,218,33,31,
111,149,121,117,126,31,228,18,123,113,141,234,
240,39,108,251,117,164,226,4,245,254,58,247,
229,219,212,49,176,71,29,58,14,206,176,149,
83,73,231,38,209,21,133,170,173,49,56,6,
33,72,82,198,192,219,113,158,108,245,246,189,
152,187,155,3,220,48,161,227,68,220,156,44,
188,80,100,33,136,210,83,214,206,103,57,58,
78,196,70,115,128,97,106,152,197,47,81,48,
191,74,32,69,152,193,128,253,213,21,156,112,
9,212,20,89,74,81,117,151,237,165,63,193,
84,230,41,140,207,161,232,37,186,219,119,144,
85,137,98,245,37,118,151,191,13,72,200,24,
148,167,94,103,97,44,135,16,130,216,119,144,
228,132,241,98,158,75,213,20,85,134,112,16,
157,171,36,0,90,171,63,65,210,99,140,254,
171,120,237,37,180,133,44,81,193,249,104,156,
106,253,235,72,129,202,15,180,159,240,70,243,
91,92,53,255,156,157,214,60,154,90,63,215,
251,83,141,35,216,195,42,79,208,94,123,159,
242,244,203,200,146,196,171,151,74,135,107,26,
37,41,107,173,199,82,199,21,13,146,232,4,
237,248,113,167,188,96,170,204,217,30,247,29,
147,239,223,248,18,179,221,125,250,165,41,250,
165,41,8,135,244,206,224,137,202,48,177,246,
169,54,152,7,54,107,183,146,193,164,195,210,
4,132,14,139,133,6,232,57,22,11,99,24,
126,31,249,28,58,243,188,239,254,220,40,137,
70,193,224,114,227,197,7,244,21,89,226,234,
88,158,187,91,253,231,242,8,119,251,62,141,
82,149,46,95,230,202,205,21,248,32,235,18,
227,185,17,134,173,16,239,120,124,125,60,193,
145,139,108,236,153,60,242,116,118,220,20,89,
78,168,22,155,72,146,192,245,11,12,189,18,
229,252,30,5,171,75,179,59,195,175,206,119,
49,63,122,4,142,135,182,125,159,151,172,101,
62,118,175,51,111,174,147,223,235,240,185,27,
125,190,179,61,198,213,214,29,164,238,16,231,
202,117,180,97,66,24,171,224,7,132,97,202,
252,238,123,44,110,95,101,225,230,71,168,27,
219,240,214,43,172,154,55,240,146,10,97,162,
82,121,37,207,140,178,199,187,73,138,210,153,
131,249,113,54,142,145,91,102,138,194,25,17,
235,121,29,224,2,51,21,121,89,143,92,85,
35,0,118,213,254,169,90,135,131,225,247,50,
199,213,76,5,178,51,60,108,80,127,246,70,
191,224,231,3,142,223,103,176,175,103,120,191,
80,232,238,171,244,223,200,83,88,104,208,252,
254,58,141,48,68,142,19,118,150,218,52,94,
41,163,235,50,105,42,177,181,227,97,4,1,
139,75,10,245,27,62,66,72,244,186,17,178,
183,78,50,113,3,179,191,195,130,180,140,129,
32,196,64,33,70,33,65,150,36,74,37,13,
105,224,192,68,141,107,69,139,183,157,38,110,
251,8,91,172,136,132,158,227,158,253,142,126,
31,149,132,84,128,59,218,141,3,7,118,82,
235,176,247,193,227,115,118,145,241,248,181,173,
97,8,142,255,220,243,251,60,215,42,126,196,
230,82,135,208,143,49,115,58,249,188,65,190,
96,97,26,13,76,74,228,205,50,57,67,193,
84,101,34,189,131,97,87,113,187,27,224,251,
24,118,102,193,239,174,124,15,171,48,65,42,
60,20,85,197,31,116,232,45,125,147,193,222,
34,170,82,102,252,234,47,161,140,2,194,49,
16,14,59,164,79,32,60,179,138,147,88,197,
44,93,80,221,31,99,248,232,3,20,93,65,
83,26,104,70,38,92,149,177,175,241,227,221,
30,95,151,223,69,222,253,223,121,144,124,141,
82,242,230,57,71,96,64,124,64,77,143,141,
140,197,254,210,15,177,203,167,11,13,39,85,
144,36,159,253,3,101,17,14,179,140,166,227,
2,63,82,113,219,206,49,15,74,96,133,29,
6,65,140,213,219,132,156,125,116,246,162,115,
132,183,223,127,106,252,227,240,30,0,81,246,
121,229,246,22,106,18,177,121,6,83,172,151,
70,71,242,225,137,22,138,147,221,239,66,149,
217,17,89,179,6,231,194,123,171,158,215,25,
87,76,220,118,239,194,103,225,137,30,192,99,
251,69,2,38,149,144,173,158,255,169,231,85,
68,30,27,235,67,102,26,55,217,40,94,97,
193,124,155,192,151,24,118,2,242,173,93,212,
126,143,222,59,155,104,115,227,212,172,42,215,
47,107,244,21,133,149,129,207,70,211,70,184,
57,188,200,100,188,177,118,228,93,42,18,115,
182,137,144,178,182,100,98,148,6,123,213,90,
102,55,108,144,87,28,118,186,21,222,154,116,
17,45,19,113,101,134,245,110,131,107,99,219,
60,108,78,32,185,30,75,143,134,72,50,244,
243,30,185,253,69,164,249,27,116,63,82,152,
211,30,161,92,154,1,9,126,178,107,241,173,
253,43,20,86,250,84,237,47,243,104,113,25,
113,204,235,201,164,123,110,180,146,101,149,0,
0,32,0,73,68,65,84,236,104,186,114,23,
8,94,133,206,137,244,88,71,213,25,171,158,
239,14,30,208,222,150,146,225,83,44,238,11,
126,62,25,137,91,100,213,41,140,50,136,90,
139,5,190,179,187,199,91,147,21,170,245,102,
22,221,148,51,107,225,206,71,93,102,103,115,
12,7,14,67,95,69,37,160,39,198,168,111,
127,130,83,190,65,28,36,36,66,70,236,44,
34,1,197,36,179,54,21,32,144,109,230,230,
12,246,183,250,248,126,130,181,211,130,84,96,
92,153,161,48,185,137,162,171,24,249,236,153,
109,192,174,38,236,246,3,156,32,6,164,140,
74,68,150,40,82,96,108,124,138,102,223,103,
173,149,29,84,93,149,105,76,21,207,140,241,
216,23,172,48,125,252,218,181,96,8,185,232,
185,231,247,121,174,53,115,48,87,159,38,77,
83,122,29,151,78,107,200,226,214,0,153,148,
106,65,34,223,16,84,234,86,86,148,100,20,
200,163,162,213,42,72,146,116,8,83,70,81,
155,238,250,93,110,252,234,63,71,146,36,34,
111,128,216,143,185,116,251,239,147,171,157,22,
104,178,162,158,9,19,157,55,198,171,95,57,
27,38,45,252,87,252,96,103,1,67,190,135,
225,127,64,191,5,19,215,254,238,25,71,192,
66,207,149,143,9,246,148,56,236,97,87,39,
73,147,228,84,123,213,133,42,184,27,253,17,
44,40,101,124,82,199,218,164,150,43,38,246,
99,181,50,133,72,103,208,245,185,127,109,142,
183,62,121,155,119,174,125,17,145,238,34,14,
185,168,164,147,173,86,71,61,187,47,116,126,
225,176,106,124,32,233,72,34,133,220,233,191,
213,85,227,152,124,120,146,89,222,207,42,173,
47,92,153,29,31,171,90,127,242,222,154,154,
190,116,102,79,248,167,157,133,39,141,179,246,
139,93,5,103,103,72,239,241,88,216,51,204,
107,87,82,185,148,111,16,132,191,200,204,237,
117,154,239,46,225,186,41,234,39,107,20,199,
101,218,205,1,254,106,23,33,171,168,183,39,
208,94,187,76,127,104,211,85,36,130,84,99,
60,239,80,208,84,194,52,69,150,36,174,213,
53,162,237,61,116,63,96,45,152,230,146,190,
153,173,139,20,17,139,108,143,121,161,202,36,
29,228,213,109,226,55,110,162,12,70,222,225,
208,67,222,220,67,81,36,174,93,47,176,219,
94,99,216,120,25,81,173,83,249,201,119,208,
126,241,6,169,4,107,142,79,95,116,200,185,
121,198,170,191,75,100,78,227,135,67,56,150,
76,249,220,208,147,44,73,23,110,69,88,177,
245,167,66,51,207,10,179,28,190,136,84,197,
86,219,152,63,126,196,218,110,64,167,29,18,
4,233,33,108,180,182,234,16,161,34,80,8,
48,176,226,61,246,90,49,202,112,9,145,64,
132,138,46,78,6,180,10,121,141,104,24,227,
80,225,250,173,76,224,147,36,164,181,50,74,
32,51,222,144,216,221,94,68,207,213,15,131,
209,246,168,32,236,180,181,227,157,130,152,194,
56,101,101,207,101,170,108,98,27,202,115,207,
71,146,10,6,254,207,38,141,242,204,189,33,
203,84,106,121,42,181,60,11,76,224,14,92,
134,173,125,58,237,33,155,27,109,76,83,103,
108,162,72,165,90,56,1,219,219,186,194,229,
241,55,152,188,241,235,39,2,210,51,55,255,
254,211,99,102,73,132,72,18,20,253,233,105,
147,59,15,190,73,117,230,11,232,246,17,181,
138,170,91,84,103,191,1,124,131,180,185,68,
52,188,88,219,30,69,213,201,213,110,101,231,
179,185,68,113,234,218,105,24,199,84,143,226,
71,73,152,89,170,7,239,44,18,130,225,81,
194,71,18,122,228,164,152,28,30,150,42,211,
122,249,101,102,219,27,172,24,181,35,79,36,
242,30,163,9,249,116,35,231,13,81,211,8,
231,12,69,145,72,202,207,116,15,141,23,141,
11,43,137,23,49,206,75,54,121,150,177,221,
243,89,152,248,5,30,186,119,153,41,111,208,
239,134,220,95,49,208,59,15,113,221,152,36,
17,148,74,2,205,243,208,115,10,247,183,138,
196,137,198,165,177,135,200,131,41,44,20,190,
58,94,66,14,34,36,111,157,7,251,13,110,
197,9,150,228,147,140,148,131,44,165,148,180,
193,72,136,199,200,187,109,68,33,199,94,84,
99,92,218,102,189,211,96,90,201,246,174,231,
37,236,247,101,118,173,91,248,141,121,190,176,
253,103,132,21,29,145,203,206,72,89,87,233,
59,130,66,244,57,236,234,220,97,253,220,241,
241,220,138,98,188,100,80,180,46,214,145,231,
89,251,208,62,213,214,56,38,92,20,202,188,
236,6,172,45,246,73,83,65,48,74,83,60,
65,121,65,76,146,111,224,72,121,172,193,10,
49,50,81,16,2,58,42,49,1,6,198,177,
184,68,181,97,208,30,250,252,120,127,138,72,
236,49,149,119,97,188,66,34,12,114,121,25,
53,117,73,54,131,67,37,177,244,253,127,205,
194,151,254,209,169,84,80,175,187,133,52,154,
234,199,49,208,182,19,210,118,66,198,139,6,
179,53,251,185,226,21,103,181,56,252,121,26,
118,78,199,214,202,140,153,89,128,52,21,2,
223,57,157,109,226,134,9,123,131,128,99,6,
251,211,141,84,127,64,103,176,206,176,247,136,
40,191,132,18,213,208,197,36,73,24,96,151,
103,169,78,159,132,144,132,72,241,228,187,52,
215,119,185,116,227,247,207,188,167,102,22,24,
31,255,229,103,126,207,179,148,196,129,2,60,
218,176,70,102,129,75,50,138,44,81,174,150,
79,132,184,67,87,70,179,100,174,23,43,104,
86,158,56,17,72,3,7,187,191,201,221,250,
141,3,237,153,197,58,194,225,73,24,235,25,
135,107,218,72,231,196,40,196,207,48,53,247,
82,213,162,148,250,159,217,231,137,103,136,185,
62,45,78,39,234,160,148,126,21,243,210,59,
244,187,107,52,252,187,252,208,191,201,21,62,
4,96,108,220,66,188,113,13,71,148,121,125,
204,227,97,63,51,26,114,213,93,90,131,60,
111,59,6,183,199,99,26,139,155,92,27,44,
211,141,75,200,178,32,61,38,91,106,106,22,
129,146,62,88,66,177,214,136,223,184,201,208,
55,152,92,185,143,61,55,228,99,111,158,55,
184,71,161,164,242,227,237,25,126,227,214,125,
238,138,58,105,146,160,42,18,73,185,128,27,
39,252,112,183,143,188,126,147,241,155,191,204,
135,235,131,51,99,156,207,173,40,246,6,1,
99,5,227,84,74,226,89,67,85,36,114,70,
86,129,252,34,198,1,205,118,232,118,209,245,
37,180,187,155,164,199,56,74,30,239,167,23,
96,162,137,24,85,196,232,196,28,213,95,74,
120,24,104,68,56,228,200,225,48,55,151,163,
239,43,104,196,88,102,130,39,20,10,70,76,
154,151,16,94,151,20,11,83,145,73,211,35,
43,112,238,243,127,239,204,122,129,86,243,187,
232,105,31,189,240,79,206,205,138,106,246,3,
20,69,98,166,242,233,172,167,131,54,166,255,
57,13,89,146,176,243,103,87,213,182,6,33,
243,23,62,225,130,230,238,191,71,189,186,135,
52,145,53,18,20,56,68,172,145,250,18,254,
182,12,156,84,20,65,127,31,74,29,226,96,
72,228,15,208,204,194,179,205,119,232,19,210,
37,116,251,104,86,145,200,27,32,41,79,54,
132,108,82,148,112,72,26,12,179,107,67,7,
97,20,40,89,218,83,123,88,168,138,4,183,
110,145,220,191,71,209,237,209,183,75,153,213,
161,104,231,199,44,0,68,138,20,251,136,39,
84,117,199,170,206,205,165,31,179,50,126,149,
203,205,69,86,198,22,152,221,91,97,179,62,
131,126,70,220,71,83,100,100,41,203,62,76,
210,159,142,101,50,85,54,153,44,155,184,237,
207,110,63,110,182,189,172,178,254,5,120,246,
61,47,162,80,152,192,81,236,145,145,26,98,
225,208,98,146,26,219,108,110,56,212,204,85,
170,95,211,209,162,41,222,40,60,96,91,86,
120,93,145,200,41,123,44,214,74,4,190,78,
247,234,117,20,199,161,252,201,3,30,120,11,
92,183,78,51,249,166,100,68,144,162,152,67,
233,103,172,197,250,157,79,40,95,29,103,152,
171,144,207,7,92,137,238,34,172,49,116,111,
72,24,164,40,69,11,36,9,69,146,8,214,
166,184,116,227,31,18,37,226,220,68,152,231,
86,20,113,34,232,186,209,133,203,232,77,237,
197,41,138,3,11,109,111,235,155,252,210,204,
14,155,63,60,77,185,113,242,255,9,19,118,
159,65,105,150,170,223,100,115,195,61,84,41,
22,62,17,42,182,228,49,61,109,163,104,50,
219,171,30,6,18,229,112,143,141,164,204,235,
245,14,242,198,46,34,103,145,206,77,210,48,
53,238,137,35,69,161,232,167,133,188,219,89,
167,81,94,229,101,179,199,15,182,222,39,205,
221,58,247,125,186,78,244,169,21,197,102,199,
63,164,185,248,155,48,204,39,24,30,145,55,
192,237,173,147,175,94,70,209,45,154,171,127,
142,125,173,195,66,213,198,82,20,134,113,76,
221,208,233,69,49,31,175,228,152,186,250,219,
167,161,64,103,27,165,156,160,20,83,122,59,
119,168,207,253,194,51,61,159,85,153,56,194,
241,115,101,68,18,97,228,159,236,2,25,192,
148,176,217,88,243,17,102,25,43,117,152,26,
203,81,178,46,238,105,111,155,21,38,215,151,
232,219,37,228,230,143,72,199,223,122,114,65,
158,36,63,81,73,0,4,70,158,15,94,202,
60,167,247,171,89,79,242,238,168,7,70,235,
152,91,87,207,235,76,150,205,195,190,39,66,
64,207,139,216,31,132,116,130,11,167,62,252,
220,193,77,153,193,27,178,213,125,113,222,203,
192,143,41,87,45,90,165,203,140,149,150,232,
247,66,166,88,164,79,149,125,166,168,187,91,
244,238,236,18,183,58,204,188,252,57,196,157,
187,220,124,117,26,165,94,68,89,222,224,70,
163,140,188,252,16,76,3,199,147,16,182,201,
165,36,99,153,93,139,103,153,144,119,208,229,
204,27,47,165,155,36,138,150,197,153,162,20,
105,212,26,90,108,181,16,174,199,246,166,135,
170,75,108,52,13,198,205,7,44,47,15,185,
246,249,28,164,130,119,59,3,20,105,10,73,
146,9,227,243,229,242,115,3,114,186,42,211,
40,62,61,70,17,196,89,31,230,174,251,98,
48,116,69,146,208,71,120,162,170,238,50,124,
119,5,215,121,154,219,40,81,42,233,88,122,
66,165,114,250,153,139,182,224,165,235,121,114,
229,28,170,101,113,251,86,145,98,189,64,67,
238,82,179,143,166,74,114,125,136,227,140,223,
77,30,60,1,14,25,178,223,252,67,174,109,
223,199,189,179,130,72,90,79,172,179,120,82,
122,231,147,134,31,37,52,251,254,223,24,37,
97,168,50,179,231,20,84,249,131,38,219,59,
255,10,81,253,35,54,151,254,53,251,43,111,
51,220,93,69,54,82,174,21,108,102,108,131,
155,197,28,21,93,101,199,11,208,146,217,51,
5,105,18,123,216,134,194,23,39,138,12,229,
239,176,179,244,167,63,213,119,234,110,220,1,
96,178,100,50,94,202,60,240,133,70,142,90,
94,127,166,117,159,174,152,236,95,189,205,151,
238,255,21,233,196,23,63,147,170,109,67,149,
185,49,153,103,97,44,119,162,57,150,36,101,
112,242,213,241,28,183,166,11,79,111,232,116,
129,81,47,232,204,213,63,91,78,171,157,158,
207,242,158,243,66,239,57,24,181,17,150,202,
127,7,190,252,5,166,102,178,56,82,145,246,
161,1,219,23,5,150,151,186,180,191,251,125,
6,253,144,205,31,173,130,162,144,220,186,130,
188,188,137,36,4,242,110,155,92,175,133,19,
107,88,138,207,254,228,13,228,178,133,174,132,
248,178,205,35,119,142,173,173,0,57,12,113,
191,179,132,250,104,249,80,99,151,6,219,224,
12,72,82,193,245,171,69,60,185,66,65,25,
146,47,106,136,158,139,114,231,33,151,108,131,
116,68,15,99,27,202,185,49,154,231,94,217,
169,99,100,125,231,121,28,247,183,7,47,180,
134,226,184,197,25,7,30,149,220,0,225,60,
189,242,91,0,65,148,144,26,50,221,238,105,
108,220,119,83,150,154,42,254,204,53,138,21,
139,180,82,162,250,229,24,251,221,31,35,185,
163,9,212,53,210,169,6,3,36,150,187,62,
34,58,173,112,226,192,165,179,245,62,78,112,
143,249,219,33,206,159,14,72,174,93,129,36,
193,15,207,47,234,81,229,79,167,183,215,90,
222,207,60,54,17,133,49,126,16,97,153,6,
170,246,233,237,15,77,145,184,49,153,7,247,
244,250,12,118,63,97,175,245,77,190,244,37,
104,24,69,126,164,111,82,210,154,248,249,144,
170,174,157,50,36,84,73,34,42,220,103,243,
97,202,248,236,55,80,141,35,28,63,73,60,
82,41,165,162,107,124,237,106,192,95,62,188,
71,18,254,202,153,94,225,139,24,189,189,143,
177,43,179,232,185,18,245,188,206,108,181,68,
48,124,66,203,215,56,34,141,163,145,229,158,
49,208,38,81,128,174,170,204,150,84,118,175,
92,167,208,219,102,96,21,94,156,41,127,206,
152,40,153,79,245,122,52,89,230,218,68,158,
181,118,120,110,247,180,167,141,74,78,227,242,
11,44,166,91,217,119,145,165,140,149,247,188,
244,253,141,142,199,86,231,197,25,89,51,85,
11,211,178,88,108,58,244,189,152,98,233,50,
205,238,239,50,188,34,49,87,184,79,115,185,
75,205,223,34,240,3,246,203,115,116,165,34,
241,254,71,20,139,5,134,195,152,220,119,239,
82,170,154,104,179,13,68,169,0,105,138,236,
5,24,190,67,207,46,179,150,191,198,245,217,
85,66,253,117,62,220,185,206,231,55,255,29,
66,19,44,6,87,176,21,151,57,255,193,209,
153,220,223,98,109,211,99,108,204,164,35,234,
212,213,14,8,152,158,182,144,37,137,120,106,
140,84,128,34,219,135,8,204,88,201,96,189,
229,189,88,69,113,145,140,167,253,97,72,95,
188,248,182,156,57,125,68,160,215,91,103,190,
18,95,200,42,211,137,232,123,54,82,94,224,
6,39,175,87,117,29,119,236,85,228,100,192,
53,239,67,210,166,142,209,85,105,231,190,136,
59,247,10,254,134,79,58,55,73,251,210,56,
32,241,189,181,30,233,226,101,230,94,249,157,
19,247,113,58,171,236,245,254,136,194,252,128,
25,93,229,70,179,195,74,55,160,47,213,169,
205,125,131,181,141,225,11,245,40,134,65,76,
215,251,217,100,58,9,33,104,237,245,233,111,
110,179,211,219,71,8,193,196,116,149,43,215,
39,62,157,151,40,75,220,152,40,96,106,202,
137,46,34,34,77,24,238,175,177,183,247,45,
102,94,235,80,55,10,40,146,196,151,199,179,
202,232,75,5,147,118,24,29,118,29,60,24,
47,151,243,180,195,46,137,120,143,245,71,219,
76,78,253,46,102,49,75,245,12,146,85,68,
42,24,70,9,213,40,226,234,100,192,230,198,
219,140,95,254,213,103,122,102,73,86,8,157,
238,33,123,172,72,5,170,105,163,60,214,67,
97,238,141,127,116,241,121,77,98,214,222,249,
35,38,95,254,117,204,66,3,205,202,35,210,
4,73,81,209,70,196,129,245,27,69,38,255,
244,223,241,221,185,47,160,42,233,147,22,41,
51,253,159,208,114,245,105,99,181,229,178,213,
245,209,148,172,69,166,34,75,168,178,132,161,
101,36,121,7,86,168,4,204,55,108,130,56,
121,102,195,176,104,169,92,25,203,241,162,252,
163,174,27,29,210,6,237,244,124,74,150,198,
120,201,200,226,65,18,164,34,83,36,47,138,
90,232,192,104,30,47,26,196,114,214,191,250,
209,238,144,219,211,69,172,242,77,226,194,191,
224,142,244,199,76,55,62,194,248,240,35,186,
29,129,190,252,231,4,213,151,217,217,217,65,
213,52,76,195,96,125,101,200,230,154,195,75,
146,132,97,42,164,243,83,72,174,143,182,182,
131,253,242,20,195,78,155,63,92,154,162,160,
39,124,253,210,58,194,156,70,249,112,137,134,
220,162,164,244,143,25,197,2,73,202,250,193,
219,182,194,193,106,44,45,15,153,156,178,176,
76,5,20,133,165,190,135,38,29,21,150,78,
150,50,195,127,187,227,159,232,108,250,92,138,
162,104,169,231,54,146,57,82,38,252,84,44,
158,188,53,226,215,143,66,212,97,138,19,169,
132,25,161,111,70,9,125,136,173,197,40,35,
86,126,1,116,251,49,179,181,29,150,204,91,
132,124,130,197,0,203,148,233,141,125,129,198,
240,3,220,129,199,98,148,82,174,232,92,190,
90,166,182,126,135,251,151,94,231,195,198,36,
179,179,58,127,213,236,146,174,204,80,176,127,
25,115,122,226,20,215,79,107,255,219,124,229,
182,79,195,75,241,239,175,177,252,206,22,118,
65,103,75,52,40,34,63,145,52,238,211,164,
230,117,156,232,5,0,136,207,174,32,246,118,
251,108,172,236,225,185,33,121,201,69,136,44,
147,169,189,223,231,202,181,9,158,245,196,203,
146,196,245,137,252,153,105,194,107,247,255,0,
119,216,102,234,205,62,175,87,74,168,143,9,
188,130,166,144,87,149,83,114,48,167,42,124,
165,81,226,237,189,30,92,221,102,187,245,191,
160,110,93,65,248,6,233,244,58,170,4,134,
34,145,54,42,92,93,221,102,49,94,126,118,
15,200,42,28,122,0,7,117,21,177,239,16,
37,17,154,249,236,153,72,105,146,208,92,250,
11,22,190,250,95,34,146,39,83,74,188,51,
118,147,237,205,54,151,102,159,64,179,113,48,
41,207,219,107,33,73,57,43,206,233,133,9,
87,199,115,199,192,93,184,58,150,231,227,205,
254,133,201,41,115,134,194,181,241,252,19,145,
137,103,29,222,99,15,219,243,34,122,94,132,
34,103,9,53,126,111,72,168,171,47,236,243,
114,134,194,120,73,59,133,166,124,188,209,103,
188,100,48,94,52,41,206,254,67,90,238,175,
18,221,252,191,152,92,254,1,227,152,4,206,
50,197,249,121,194,240,200,216,75,83,193,198,
98,143,133,175,94,70,20,115,8,219,68,84,
138,200,142,79,59,116,24,132,21,26,213,7,
228,141,42,82,39,70,146,68,166,36,14,12,
2,178,30,241,227,99,38,195,97,68,167,23,
18,89,21,174,153,15,89,238,132,184,78,194,
173,87,202,8,93,193,221,76,25,171,221,58,
21,35,26,120,241,197,250,81,92,100,92,132,
124,78,85,228,99,157,85,94,204,40,219,26,
249,17,45,135,93,157,165,213,77,169,17,162,
159,65,228,23,163,16,161,143,122,98,168,24,
105,132,33,66,94,175,44,179,102,76,225,71,
26,161,164,51,27,124,76,243,88,101,108,191,
23,177,185,214,71,183,34,174,78,45,241,190,
62,197,255,123,79,167,186,127,131,75,55,254,
65,214,123,224,88,85,246,209,73,81,176,16,
236,125,247,99,182,54,124,144,4,99,179,53,
144,175,63,53,163,226,211,96,188,97,44,64,
255,217,40,136,51,175,73,33,21,41,242,51,
176,197,74,18,92,25,207,157,217,140,40,14,
92,210,194,6,98,170,142,38,73,231,218,28,
231,201,152,162,166,242,43,19,85,254,106,183,
139,91,11,17,181,123,144,102,150,214,205,98,
14,125,4,247,89,65,68,46,191,79,232,116,
208,115,149,231,154,39,213,204,145,132,254,168,
149,106,157,103,209,154,161,211,38,136,86,241,
7,215,48,236,243,91,128,38,169,160,244,240,
30,106,126,26,82,59,75,145,61,111,4,131,
147,133,121,47,112,4,103,4,64,85,69,226,
218,68,158,187,155,131,167,114,159,153,154,194,
141,137,194,83,13,206,103,223,171,231,207,91,
223,139,33,121,177,66,233,82,213,230,241,110,
154,144,101,134,109,118,124,182,187,1,99,69,
131,233,74,13,253,218,127,203,102,144,80,216,
253,22,90,190,128,64,194,87,109,44,186,39,
12,167,212,212,25,36,6,178,42,241,200,77,
104,186,54,187,237,10,151,103,238,82,53,52,
76,69,70,140,148,93,124,107,1,198,170,200,
171,91,164,133,60,202,242,6,210,192,165,209,
48,89,90,116,209,106,41,177,44,24,27,183,
152,120,121,2,201,80,8,119,246,73,7,87,
48,198,78,238,179,254,99,74,34,51,184,159,
99,184,97,242,212,244,56,237,5,111,0,67,
149,79,20,180,233,86,145,117,103,129,137,235,
53,114,182,122,6,182,150,160,17,34,35,40,
89,33,95,120,57,71,49,39,48,212,148,107,
197,93,46,229,118,121,163,186,140,239,158,196,
229,84,85,162,215,13,209,107,121,246,43,37,
202,82,155,177,254,44,179,183,126,15,89,61,
31,175,53,148,9,54,84,5,163,108,35,16,
52,26,38,205,252,52,118,249,234,19,201,250,
114,134,138,125,12,71,77,210,44,155,108,163,
253,228,248,195,103,153,229,254,240,254,54,15,
239,110,158,169,36,36,89,162,90,207,115,253,
246,52,242,51,198,90,46,215,115,84,206,170,
177,17,130,157,165,255,200,149,5,137,151,198,
246,233,134,9,223,91,47,176,218,110,176,213,
187,184,48,87,37,8,143,117,148,67,206,98,
24,115,163,130,35,210,20,116,149,153,43,130,
237,181,63,33,77,158,31,202,83,116,19,179,
212,32,114,7,36,225,197,49,112,179,216,96,
254,214,63,33,63,162,159,63,111,56,65,194,
15,195,34,82,46,255,244,94,236,63,37,37,
145,121,26,103,111,78,91,87,184,57,149,127,
34,156,106,235,10,55,39,243,159,58,137,227,
231,101,148,109,237,132,145,151,158,33,19,83,
33,216,233,249,124,180,209,39,17,32,42,47,
83,170,26,164,200,168,132,180,229,73,214,185,
193,58,215,89,231,6,189,94,136,255,160,73,
249,195,59,228,246,155,108,136,62,129,181,141,
162,100,222,219,148,101,100,36,128,155,251,136,
156,133,152,168,33,100,137,116,188,134,188,189,
135,52,200,50,159,76,67,161,79,25,181,249,
33,81,146,146,155,44,33,93,157,38,153,155,
98,125,162,129,38,166,207,52,88,231,235,246,
9,198,136,231,230,122,90,92,234,60,49,149,
45,117,91,112,81,98,177,167,124,190,161,202,
204,23,114,196,125,247,4,87,139,174,222,98,
213,126,192,228,116,159,197,135,131,115,239,160,
171,242,41,201,90,48,98,246,123,18,206,177,
138,241,66,94,195,206,43,20,198,138,44,223,
190,196,114,223,133,213,155,148,103,222,196,237,
236,28,94,119,22,183,140,20,219,116,6,17,
87,71,245,1,149,154,206,178,168,18,13,90,
244,130,33,248,241,41,97,152,51,52,46,21,
76,220,253,30,65,44,177,189,189,141,187,217,
63,124,143,98,122,126,191,15,188,54,68,241,
11,153,223,167,93,123,125,86,101,97,178,66,
146,164,200,178,132,36,73,200,146,140,28,202,
25,197,129,4,224,128,227,156,135,171,64,228,
67,114,36,56,199,11,38,118,228,159,200,151,
247,135,109,250,173,247,72,205,14,115,55,251,
220,174,20,24,196,49,175,85,84,96,56,250,
58,207,148,4,101,99,135,100,172,6,134,54,
178,122,13,190,214,40,227,39,41,107,174,207,
166,27,144,34,8,210,20,91,81,50,247,98,
224,176,48,86,99,247,165,53,86,238,254,43,
138,214,91,168,74,225,137,252,81,39,60,129,
51,184,166,68,154,32,201,10,73,212,38,141,
35,52,43,127,98,223,100,74,228,252,218,151,
56,112,71,199,109,128,16,41,73,24,16,141,
200,8,31,173,117,184,221,251,132,225,188,5,
142,148,213,82,36,241,167,226,122,2,104,228,
13,198,138,198,17,194,45,50,107,120,152,184,
164,138,74,16,39,4,145,32,76,82,226,99,
197,121,66,145,113,219,46,145,59,64,245,135,
72,199,184,158,100,224,138,37,88,105,185,39,
60,15,41,137,168,229,20,198,45,133,184,239,
240,164,39,254,180,188,103,178,23,131,243,4,
185,115,81,94,172,167,92,43,1,99,249,60,
110,123,64,28,184,200,178,138,47,148,115,207,
192,246,238,128,154,212,32,23,238,163,217,58,
82,43,59,11,87,210,123,71,71,143,60,81,
237,6,190,187,75,94,3,223,54,73,71,80,
154,36,122,204,216,6,115,150,142,122,231,33,
82,187,71,58,123,20,19,148,91,93,228,189,
206,145,204,211,101,222,122,41,70,83,114,8,
73,176,185,220,162,162,69,136,217,73,118,252,
50,166,126,249,76,100,36,15,92,203,103,61,
128,142,20,197,115,112,61,117,129,52,80,201,
27,42,141,51,10,239,226,68,192,224,162,153,
36,231,127,190,161,201,220,154,42,28,210,114,
31,231,106,177,171,147,172,44,191,207,229,153,
125,228,229,33,73,124,182,48,212,117,153,237,
29,15,219,84,40,149,51,188,70,145,4,187,
219,3,252,72,69,37,34,103,171,92,191,81,
32,77,5,195,171,243,180,165,108,129,244,124,
229,20,143,140,16,226,196,207,68,154,176,215,
252,255,248,188,20,225,239,101,105,179,74,41,
79,32,205,97,149,26,56,29,13,148,147,207,
214,80,218,76,214,138,72,200,4,206,62,251,
76,227,170,14,226,216,60,200,133,194,9,111,
227,196,51,236,58,160,215,158,123,126,47,122,
173,122,6,94,41,57,242,137,231,61,223,106,
136,33,116,97,84,153,109,235,10,179,211,69,
36,73,208,217,248,0,63,220,34,102,143,208,
220,194,153,29,227,86,77,231,141,241,194,33,
132,116,238,147,118,250,96,234,72,125,7,121,
109,39,179,166,188,144,228,122,86,11,176,209,
173,113,169,188,79,213,136,168,26,26,215,10,
54,155,94,192,48,74,176,21,5,121,175,131,
212,115,208,146,132,133,162,133,119,115,31,101,
251,47,145,149,223,126,46,238,160,225,222,18,
187,139,63,100,238,141,223,29,201,231,33,178,
162,18,251,14,154,153,39,9,253,195,251,31,
111,92,117,240,125,228,13,71,113,144,44,152,
29,135,62,178,36,51,244,2,30,52,187,72,
229,155,92,157,185,74,34,43,35,162,193,240,
233,156,68,34,65,114,59,8,205,58,164,254,
144,36,152,159,175,156,9,223,153,134,126,234,
189,132,32,83,24,73,138,161,41,168,178,68,
168,91,168,86,30,89,57,189,78,175,142,65,
207,141,104,59,33,182,174,82,84,19,52,57,
61,84,156,79,27,159,134,247,76,141,83,214,
130,222,19,22,236,130,252,77,79,185,118,162,
108,82,30,25,202,135,235,27,43,224,158,109,
204,68,134,134,94,154,64,238,150,80,84,25,
133,132,20,25,249,24,62,111,49,68,149,124,
122,234,4,246,116,202,125,89,230,160,161,180,
144,125,154,110,132,8,187,164,121,11,49,63,
133,40,142,98,68,174,143,188,182,115,250,204,
170,114,198,83,103,41,24,186,76,122,105,146,
206,120,141,206,15,167,153,185,121,235,212,154,
109,118,188,81,242,130,140,80,237,231,143,81,
28,199,180,250,94,76,156,138,23,214,80,228,
44,120,226,73,189,27,172,218,111,241,32,220,
33,111,239,208,235,135,143,25,153,50,17,26,
173,150,71,189,174,51,49,97,225,12,99,186,
221,144,161,19,227,57,49,141,177,34,249,98,
129,106,94,128,4,105,181,204,70,46,71,63,
236,143,98,29,173,51,112,241,147,39,107,111,
237,123,204,191,188,69,109,24,209,73,82,20,
89,162,111,85,48,115,95,192,15,19,146,84,
61,133,171,23,115,57,92,47,66,205,143,209,
211,108,186,93,255,20,156,180,211,11,184,58,
118,122,169,50,74,241,244,51,139,81,188,232,
145,229,203,167,172,223,255,191,177,46,127,194,
203,117,11,93,150,136,133,66,148,238,211,28,
218,44,119,202,92,206,239,195,57,138,66,222,
104,162,60,88,59,45,15,243,71,251,240,122,
121,131,59,123,243,220,154,220,64,150,4,57,
85,225,122,225,232,247,202,221,37,72,5,97,
187,199,123,170,198,27,150,206,140,182,198,127,
104,253,152,226,196,181,79,253,126,249,198,2,
65,184,199,195,191,252,95,153,127,235,31,96,
20,50,44,56,242,135,104,118,1,33,146,51,
247,82,146,130,114,14,197,146,98,88,108,183,
98,246,154,61,126,219,28,208,139,2,92,227,
25,206,156,164,100,74,226,49,193,31,167,233,
133,123,163,72,82,230,221,95,52,249,66,26,
193,51,7,20,62,73,232,147,38,63,221,230,
90,186,42,163,42,210,211,59,14,2,150,174,
16,198,233,51,87,153,235,170,204,84,217,60,
51,46,113,174,173,36,196,8,106,139,145,100,
9,133,100,68,29,20,18,145,205,79,181,170,
131,41,8,52,155,255,39,29,35,23,52,143,
254,222,138,136,221,28,235,101,131,250,88,25,
59,63,202,56,11,35,148,123,203,16,68,103,
198,140,212,156,74,172,25,204,124,161,70,103,
102,156,239,255,200,166,210,56,155,24,179,217,
15,50,99,32,126,65,89,79,143,111,158,241,
162,113,6,134,249,252,27,162,81,48,78,5,
122,227,84,48,12,226,195,252,104,179,56,206,
202,226,2,223,120,101,153,143,190,191,115,98,
209,35,20,108,60,144,4,81,152,242,225,251,
157,19,27,200,208,21,106,227,5,86,220,41,
118,6,49,168,26,78,125,154,102,148,9,160,
207,87,11,188,187,246,116,122,140,48,106,178,
224,56,236,191,253,136,141,77,151,114,69,103,
37,169,97,85,166,216,223,220,58,51,224,182,
216,31,205,217,224,252,194,189,246,48,196,59,
86,17,155,193,41,233,185,109,65,255,115,24,
245,130,78,78,19,60,250,225,191,68,127,189,
201,68,201,98,82,83,80,222,185,79,154,183,
80,38,170,204,167,67,86,246,84,148,206,14,
201,194,204,153,158,132,178,180,121,182,64,189,
191,140,188,182,77,252,242,21,80,85,110,79,
110,240,193,230,28,175,207,172,156,188,238,209,
58,162,82,32,186,50,139,245,112,21,26,53,
12,199,67,18,130,168,190,200,214,195,127,199,
212,25,236,177,23,29,181,233,47,82,157,252,
194,161,119,240,164,8,83,24,167,108,116,124,
218,195,128,133,177,28,133,51,228,176,31,37,
44,173,181,73,147,148,238,141,219,132,154,241,
66,214,195,11,83,52,75,230,111,210,200,233,
42,189,115,82,199,21,73,162,86,52,104,20,
12,114,134,130,16,89,26,240,89,233,178,99,
69,3,97,26,167,234,67,102,107,214,153,65,
248,228,9,202,41,244,227,204,82,79,123,244,
119,157,17,68,39,112,177,176,113,209,52,153,
203,95,188,204,189,248,37,46,205,238,34,239,
234,52,151,95,99,124,254,3,144,192,212,93,
134,18,180,212,1,149,216,62,46,124,72,22,
166,144,91,125,228,245,157,19,168,177,176,77,
238,26,211,44,123,55,145,6,121,164,247,114,
140,95,254,10,145,219,63,195,232,143,206,84,
174,71,59,35,24,28,97,151,94,251,136,10,
216,107,103,216,114,236,131,215,29,105,47,7,
105,68,85,44,121,93,136,3,198,243,26,146,
215,38,77,98,226,208,195,239,103,120,161,231,
14,178,123,3,248,221,236,51,68,154,221,55,
137,50,119,217,107,103,82,51,242,144,130,236,
225,165,160,15,145,139,34,195,152,230,144,198,
33,73,28,226,247,246,16,34,101,121,171,197,
250,214,206,8,9,235,17,121,3,80,166,73,
45,131,90,221,56,227,72,10,16,208,235,159,
158,136,160,118,141,29,175,76,161,245,1,213,
193,61,138,251,31,80,243,91,120,94,141,217,
156,201,182,23,146,164,1,113,224,34,210,4,
191,183,55,42,136,58,120,30,65,228,13,81,
69,157,69,219,102,236,90,13,211,80,40,141,
231,232,37,147,136,36,98,224,184,163,247,76,
51,60,217,239,29,225,159,7,180,207,79,152,
247,229,61,151,192,233,17,185,125,118,122,1,
119,30,172,225,185,238,209,92,138,131,107,71,
248,100,48,60,127,222,211,248,228,188,135,238,
17,147,169,223,27,113,7,137,99,107,20,61,
251,179,139,244,196,158,33,24,28,102,230,40,
65,143,169,188,76,111,251,35,180,87,154,228,
12,133,40,21,124,228,248,48,86,102,235,157,
117,164,15,31,161,220,91,102,161,247,19,146,
203,211,103,27,39,94,240,196,236,149,116,122,
12,242,54,194,212,145,36,193,171,83,107,188,
183,57,70,60,202,14,144,247,58,8,77,229,
237,202,203,8,219,128,48,230,55,58,61,38,
31,173,145,234,26,74,37,36,28,251,9,187,
139,223,33,28,29,42,191,191,79,18,122,135,
251,64,164,9,73,232,17,12,179,121,15,156,
30,161,211,67,8,129,223,111,29,198,25,210,
56,36,141,35,146,208,199,27,116,216,237,7,
12,251,125,134,205,53,2,167,203,250,242,35,
238,220,123,200,254,254,46,105,119,13,167,221,
196,239,239,51,220,93,33,24,180,113,218,219,
184,237,45,186,237,22,222,246,34,211,149,152,
217,221,79,208,219,139,224,236,103,115,237,247,
70,107,228,157,156,247,96,120,20,119,138,131,
108,175,248,221,209,158,9,144,188,46,59,61,
159,208,237,19,58,189,209,123,182,136,3,15,
145,166,135,251,253,32,147,235,0,102,57,40,
24,12,134,157,236,108,136,131,107,195,195,179,
193,168,209,212,225,252,12,187,68,190,147,205,
79,111,47,243,46,226,232,112,46,227,192,37,
24,180,143,206,181,239,28,206,123,124,214,188,
15,50,79,63,116,251,135,202,248,224,217,45,
141,163,253,30,7,72,94,59,107,10,85,16,
92,45,103,232,135,26,245,71,159,145,50,161,
57,204,150,84,164,131,253,142,96,34,151,82,
83,3,46,55,108,38,205,16,41,114,179,118,
202,242,144,146,46,30,123,246,108,31,148,108,
21,45,113,30,155,119,127,116,54,90,168,146,
160,168,110,225,142,250,162,167,66,34,29,197,
137,163,40,197,93,223,103,242,74,192,31,237,
84,177,10,235,20,167,31,209,222,190,13,66,
34,111,245,24,120,89,188,45,95,82,112,250,
217,254,151,186,67,194,239,223,67,90,217,198,
117,98,186,157,8,161,106,236,206,206,242,103,
233,151,104,74,191,79,125,234,183,24,191,252,
119,40,213,94,25,237,201,8,191,191,127,56,
239,131,238,62,171,45,239,244,158,57,225,81,
24,133,236,101,224,36,30,119,240,189,172,100,
28,250,128,208,115,135,188,243,194,42,163,41,
50,211,181,28,138,156,63,196,107,213,81,133,
107,162,216,71,89,23,102,249,244,125,143,127,
175,89,8,35,195,175,15,254,157,40,89,228,
142,185,119,74,169,193,192,143,241,208,81,164,
236,243,14,90,82,26,70,141,158,98,211,235,
61,158,218,37,81,46,235,216,57,133,208,79,
217,111,29,89,6,194,170,80,46,24,88,131,
7,236,140,170,181,167,166,44,234,117,135,15,
147,42,143,186,62,134,238,32,71,121,212,145,
139,127,128,217,202,234,17,126,171,89,121,114,
149,57,122,238,95,147,220,152,227,170,27,176,
19,231,49,202,95,1,69,195,139,149,163,236,
20,205,134,131,36,31,179,120,198,92,171,167,
230,125,24,196,60,236,42,68,177,32,74,92,
48,75,35,105,41,31,93,171,30,173,17,199,
170,144,15,231,61,28,158,61,239,250,49,203,
228,224,190,161,115,206,26,157,241,236,145,123,
246,181,170,121,236,121,10,168,82,130,26,123,
204,205,76,96,90,26,157,65,139,68,146,185,
61,149,99,92,83,9,87,54,25,214,202,204,
126,126,26,121,127,100,148,12,92,180,239,253,
4,97,91,8,9,208,84,36,63,68,40,89,
11,199,83,214,28,74,134,251,86,75,153,34,
25,141,123,61,135,78,20,243,193,94,133,74,
201,103,204,212,177,234,101,22,123,10,137,39,
209,143,18,234,18,152,187,153,144,146,75,121,
110,154,6,134,13,159,4,63,194,237,85,208,
237,215,48,139,245,67,235,63,49,171,72,178,
130,162,91,89,239,111,192,24,237,197,108,122,
106,199,166,245,0,231,215,72,245,226,136,202,
121,180,70,142,0,106,112,80,142,96,148,144,
243,38,166,153,64,177,142,102,229,49,10,217,
156,186,93,159,118,92,230,242,213,113,182,180,
152,208,204,29,53,11,146,228,236,75,179,178,
175,131,115,125,184,46,149,163,88,145,89,6,
73,249,255,217,123,179,231,74,178,251,206,239,
147,123,222,204,187,47,216,151,42,20,170,186,
150,222,73,54,73,81,92,52,34,135,26,205,
200,146,173,177,60,178,60,139,60,225,136,121,
177,223,253,111,56,194,79,158,8,217,15,19,
30,143,253,224,152,69,17,214,50,163,141,20,
69,13,69,118,55,187,187,246,42,20,246,237,
238,55,247,204,227,135,11,220,2,10,219,5,
112,81,5,64,248,70,84,160,187,43,59,145,
121,206,201,243,59,191,237,251,5,85,65,168,
6,117,39,36,200,103,72,155,234,174,103,143,
188,246,174,103,223,166,111,223,153,95,48,210,
5,2,36,36,73,222,149,207,216,254,103,213,
180,81,183,204,90,156,27,0,0,32,0,73,
68,65,84,94,206,72,231,183,66,79,209,254,
215,26,86,239,59,211,237,92,207,208,110,143,
251,206,107,21,61,213,235,164,215,173,44,209,
150,81,217,126,246,17,37,33,22,163,221,16,
111,74,37,151,26,238,121,0,94,35,238,61,
251,206,251,142,0,41,203,228,241,170,78,193,
214,152,170,216,120,141,238,102,60,57,54,204,
80,148,224,6,49,121,171,184,207,179,167,144,
21,21,69,83,248,224,230,196,86,239,137,64,
136,12,113,210,245,88,98,189,128,174,169,221,
94,47,169,107,36,170,155,107,20,236,238,225,
86,150,37,252,134,75,169,179,78,217,206,176,
184,49,67,37,183,68,97,244,115,214,151,222,
193,176,170,196,177,198,245,116,170,219,56,152,
8,212,159,61,32,106,186,212,170,1,15,87,
189,46,101,199,59,37,190,152,186,193,179,165,
175,48,114,227,251,187,243,78,189,253,75,67,
79,23,186,77,163,24,60,106,4,68,113,188,
239,154,25,72,232,105,178,152,58,176,14,90,
57,69,19,141,42,75,140,236,67,54,184,186,
149,137,143,19,129,27,196,189,144,140,97,15,
179,222,49,240,95,161,201,40,148,13,102,166,
187,12,157,113,44,8,35,104,108,221,163,102,
94,103,218,123,192,226,234,203,74,156,149,101,
15,123,110,149,242,109,141,71,207,135,25,10,
103,24,191,241,107,123,66,106,47,54,93,70,
83,97,143,222,192,46,76,241,226,241,8,159,
100,55,144,167,239,176,185,60,65,190,52,75,
195,13,15,87,246,235,19,142,255,230,36,78,
7,129,91,195,105,244,68,69,223,26,175,116,
233,26,158,250,23,252,172,214,98,218,78,113,
79,128,253,55,247,247,41,255,137,145,154,237,
93,129,26,105,159,48,94,183,223,72,226,51,
229,125,236,107,195,148,50,27,116,252,16,47,
78,120,210,118,89,171,143,51,94,122,204,103,
141,24,67,145,217,244,5,127,190,88,224,219,
211,85,44,20,80,213,110,5,84,146,160,124,
241,140,217,119,110,160,62,90,96,114,168,192,
191,91,255,19,178,241,189,94,226,239,209,106,
155,142,31,115,107,36,125,108,250,124,219,80,
48,66,249,208,134,180,150,23,49,100,74,251,
254,247,40,136,200,100,83,76,191,248,132,167,
35,55,143,151,163,56,4,115,155,46,119,199,
50,72,23,187,90,117,87,14,225,36,210,169,
185,148,198,135,211,249,125,199,225,56,185,25,
77,145,217,34,184,166,237,71,140,229,77,30,
234,74,119,175,220,162,103,105,249,9,97,167,
129,81,236,134,86,23,164,155,196,229,235,20,
202,49,215,218,46,168,79,217,104,116,19,244,
149,241,79,88,93,248,136,15,103,63,166,108,
164,241,61,65,146,64,60,51,129,250,243,199,
164,109,149,184,104,160,168,18,155,111,223,228,
197,226,183,25,185,241,205,35,159,83,8,120,
178,222,57,52,159,115,234,160,100,218,80,41,
103,244,67,63,138,147,46,188,209,188,185,199,
0,5,113,66,109,71,51,200,78,254,120,35,
83,98,61,201,97,189,210,79,97,154,42,157,
96,171,49,5,149,116,49,141,143,142,75,138,
84,237,17,47,94,116,118,229,52,52,77,234,
70,125,218,42,227,226,6,19,183,126,171,231,
33,109,227,233,186,67,203,139,120,176,220,230,
217,186,67,156,8,100,85,103,106,246,119,89,
250,252,75,180,195,127,70,126,166,171,115,208,
112,194,11,243,113,229,44,141,201,98,138,140,
169,14,180,83,22,224,201,90,167,71,150,6,
224,59,85,146,117,11,91,85,184,237,251,200,
75,235,167,202,145,1,168,82,196,189,228,103,
100,234,75,252,209,147,97,126,184,222,228,167,
181,173,16,156,144,144,229,152,140,170,144,81,
85,254,98,190,200,59,67,45,202,186,198,74,
152,176,112,251,58,11,31,221,195,183,186,30,
152,250,240,5,120,62,138,227,114,107,186,198,
230,252,95,246,214,224,54,3,242,126,34,47,
253,174,237,195,208,116,35,252,87,186,139,19,
209,109,22,83,117,21,73,146,152,175,76,227,
15,40,71,1,208,241,35,230,171,14,23,29,
226,149,198,35,113,2,34,180,65,46,253,186,
19,242,96,185,141,174,202,200,178,76,24,70,
136,249,149,110,17,65,16,80,25,170,32,109,
245,30,13,199,79,201,36,235,172,135,10,195,
102,119,95,45,231,150,201,217,85,22,54,102,
81,181,14,27,126,192,159,172,214,8,2,129,
97,72,72,159,60,197,173,57,44,175,184,36,
66,144,27,178,120,82,53,40,142,127,212,215,
243,45,214,92,220,224,240,67,232,169,13,133,
169,31,126,11,67,147,25,203,31,159,104,77,
83,164,125,169,203,55,154,193,174,179,121,123,
199,198,35,73,50,126,146,165,60,213,117,255,
21,89,66,215,101,106,27,30,134,18,211,242,
85,130,16,214,23,107,24,4,164,112,73,241,
50,137,156,32,227,144,197,180,85,116,67,193,
121,145,101,104,250,59,251,158,234,118,110,254,
235,45,159,7,203,109,132,232,186,115,163,179,
223,39,149,29,222,181,80,206,51,100,73,98,
188,96,242,254,84,142,183,70,210,140,230,77,
166,74,41,238,140,101,6,106,44,252,40,225,
201,90,135,229,45,93,98,207,155,71,189,214,
226,94,198,70,118,125,146,225,226,137,239,237,
185,49,190,159,224,109,109,224,185,82,130,153,
121,200,114,245,165,108,170,216,122,149,91,89,
139,191,169,181,64,192,7,37,133,167,29,23,
39,142,249,207,213,38,107,65,196,220,236,47,
16,191,117,141,164,152,99,81,185,78,184,238,
112,167,81,39,22,243,123,60,187,32,74,78,
68,155,175,246,81,97,244,170,191,177,209,10,
168,215,58,228,11,221,48,206,248,198,60,122,
232,15,116,45,172,52,252,174,238,250,5,69,
228,187,123,170,17,165,55,236,34,173,54,252,
94,3,109,28,39,184,173,77,82,81,215,32,
103,50,105,82,169,151,251,163,38,199,76,223,
74,248,253,231,6,63,222,124,153,108,214,84,
31,77,241,25,78,55,144,58,227,68,173,49,
30,182,35,106,142,68,120,231,58,169,82,154,
233,73,27,36,112,98,9,159,4,209,103,15,
80,63,251,211,137,12,133,244,202,38,115,20,
198,242,230,129,236,141,7,255,63,169,93,247,
22,91,27,242,98,221,221,179,105,239,174,56,
184,129,249,246,36,182,165,114,237,90,26,9,
48,44,5,77,17,100,140,136,140,25,51,60,
244,242,52,183,192,77,22,184,201,10,215,88,
103,2,129,68,199,151,73,52,149,200,16,72,
251,212,133,239,39,16,212,246,35,230,54,247,
158,198,122,37,172,231,20,182,161,242,246,68,
134,241,66,170,71,219,190,211,27,60,137,235,
126,20,22,54,93,58,126,76,121,252,23,73,
54,109,158,187,30,47,202,5,208,250,8,225,
28,176,222,12,67,70,211,36,76,67,33,122,
239,38,43,170,130,151,196,220,27,94,97,163,
54,74,181,49,140,165,183,233,120,89,254,243,
26,60,90,29,227,78,217,227,71,155,13,70,
83,6,43,238,22,183,191,166,210,72,154,60,
180,198,248,56,247,109,138,99,9,43,155,50,
43,63,91,192,100,101,43,236,248,234,9,245,
248,39,214,163,14,185,150,161,144,210,118,215,
200,46,86,93,30,125,177,72,101,139,12,113,
165,48,70,168,14,190,54,250,201,90,135,39,
107,157,93,82,195,23,5,170,145,58,87,207,
227,133,73,175,242,106,173,233,51,59,59,132,
228,61,221,210,191,209,208,216,189,73,235,186,
140,90,76,163,9,125,207,170,146,165,152,200,
94,32,178,22,136,172,69,234,157,12,153,108,
130,70,12,158,79,202,82,200,101,53,172,124,
10,55,14,17,73,127,7,152,126,202,130,143,
157,163,144,37,137,187,19,89,12,93,71,214,
204,190,92,52,73,130,153,33,139,79,23,154,
125,209,97,27,154,188,163,75,180,43,241,185,
88,115,247,221,112,189,48,33,136,146,222,38,
87,158,252,14,63,120,252,136,177,153,20,173,
184,134,157,121,198,154,113,131,102,92,195,74,
154,60,125,214,161,213,10,169,83,161,198,48,
83,220,103,155,91,113,91,17,47,107,232,24,
25,25,173,54,185,239,169,248,32,102,204,253,
44,243,121,245,38,36,9,198,11,41,70,243,
230,161,52,32,165,180,206,98,205,59,80,249,
234,68,161,1,224,249,70,135,187,227,195,232,
43,183,168,24,143,153,176,12,146,76,106,255,
147,139,161,33,116,13,145,207,144,140,85,144,
87,171,72,141,118,151,48,45,159,65,122,190,
132,220,232,230,48,132,166,32,114,25,164,173,
78,224,144,144,98,35,193,243,50,52,134,76,
76,221,33,140,4,35,197,23,44,135,18,154,
44,241,172,237,210,222,186,254,65,203,33,78,
58,60,124,254,1,223,189,177,140,58,54,196,
88,171,198,139,159,68,148,197,38,126,224,177,
147,222,72,145,165,110,29,254,49,153,170,75,
105,141,68,88,60,91,223,63,212,51,85,178,
64,188,188,105,195,9,121,252,104,149,242,112,
14,107,43,241,89,105,174,225,26,22,145,162,
14,124,125,108,182,3,54,219,1,102,212,161,
16,165,201,89,42,182,161,14,156,147,233,178,
99,99,71,73,109,156,8,202,57,3,201,93,
128,48,68,32,237,249,38,83,150,138,176,76,
62,26,119,248,243,249,105,12,163,73,214,170,
17,197,26,138,242,114,223,81,219,99,144,91,
70,85,243,36,165,28,210,245,113,88,175,145,
82,13,30,148,71,176,54,191,133,102,101,222,
156,161,16,136,158,110,194,113,60,58,83,83,
40,218,122,95,110,237,68,33,213,35,66,124,
190,225,28,201,111,223,242,162,30,221,185,172,
104,20,174,253,83,156,181,255,133,137,122,196,
92,254,61,238,24,207,105,69,37,150,226,17,
90,25,7,175,53,135,67,134,235,252,124,235,
157,36,2,12,4,144,211,67,242,149,20,63,
168,77,80,25,255,59,123,63,160,214,193,207,
31,68,201,30,54,231,243,232,198,107,138,204,
173,17,27,187,79,47,111,40,171,243,98,115,
176,50,171,29,63,102,173,225,83,30,251,54,
247,159,222,103,226,109,3,41,136,16,25,11,
169,229,144,76,14,131,166,33,63,95,34,186,
125,13,81,202,247,172,76,60,243,74,185,108,
41,71,178,89,71,94,169,34,42,121,80,21,
134,84,133,41,219,100,209,241,193,8,72,181,
116,172,226,227,238,162,151,36,242,134,198,134,
23,226,199,2,127,71,137,237,102,59,67,199,
203,82,46,60,165,234,84,120,212,132,108,45,
79,54,187,78,51,82,80,116,131,168,253,114,
3,31,61,66,143,229,48,84,50,6,142,31,
179,250,74,237,126,38,165,146,53,213,94,181,
58,192,147,165,38,245,90,155,119,63,188,222,
251,111,53,187,120,38,70,98,247,65,44,102,
185,225,177,188,85,17,109,168,50,150,174,144,
50,20,44,189,251,199,212,148,43,139,112,0,
94,53,172,69,91,199,8,87,240,219,1,13,
134,200,179,142,182,69,102,122,251,43,35,200,
133,12,200,18,183,154,75,40,215,178,252,245,
154,196,102,107,152,36,86,40,229,186,30,109,
202,31,162,88,222,64,83,116,76,69,166,26,
132,44,230,50,172,27,38,178,4,237,199,83,
76,205,126,167,239,103,140,69,191,134,194,221,
34,218,241,26,47,75,237,14,57,13,182,106,
50,166,161,163,104,71,187,189,65,187,134,172,
168,8,1,110,189,3,135,157,76,189,6,150,
161,99,199,17,94,163,43,200,115,160,214,130,
247,178,61,191,182,209,198,142,119,39,7,55,
107,211,124,229,102,157,220,31,254,140,103,157,
144,155,55,124,198,109,133,166,98,240,103,107,
83,140,69,159,237,120,167,110,11,125,132,138,
157,86,89,170,220,32,49,255,1,113,224,19,
7,187,19,172,235,107,109,216,174,86,121,101,
188,44,77,193,111,190,180,250,27,237,128,246,
182,188,162,215,56,198,215,121,244,60,156,244,
90,83,215,185,150,183,80,188,0,207,59,122,
222,0,50,2,100,183,117,112,229,86,191,207,
144,196,91,253,28,221,49,90,92,170,145,30,
205,16,183,114,64,68,50,62,132,40,229,81,
62,127,74,50,57,130,48,117,226,241,74,183,
26,105,251,212,21,69,136,125,186,180,69,41,
79,92,218,45,69,250,110,62,205,132,101,240,
67,81,35,201,215,123,231,183,72,8,194,125,
78,81,141,78,153,106,171,76,33,189,129,23,
216,124,105,102,147,244,39,143,120,248,96,17,
95,135,182,100,33,53,54,104,110,116,215,177,
38,75,228,137,240,26,173,94,159,64,63,216,
57,182,21,21,26,81,123,151,183,172,72,42,
94,195,235,149,134,118,58,29,238,127,252,136,
107,227,89,36,255,229,239,201,212,23,104,202,
9,225,118,66,187,223,53,182,109,129,98,191,
191,245,181,211,163,222,250,83,219,21,191,150,
48,132,131,105,165,73,25,26,134,42,163,171,
50,186,34,239,123,152,140,195,0,145,68,135,
242,92,237,55,86,131,188,246,56,243,213,239,
181,145,239,32,201,234,174,125,49,15,212,146,
14,109,63,70,2,84,63,66,138,35,4,18,
25,170,84,25,101,152,57,20,69,66,191,54,
132,152,30,69,121,182,8,142,79,59,99,99,
25,30,45,55,143,72,100,100,41,65,120,58,
215,172,132,97,205,198,222,234,200,204,106,42,
73,10,86,230,21,54,94,36,84,42,111,247,
250,93,142,122,175,40,138,16,142,211,167,161,
216,174,123,79,226,190,248,79,60,45,38,159,
77,247,106,170,15,223,27,186,245,210,115,155,
14,142,170,30,238,195,36,49,83,147,227,152,
41,173,43,155,138,10,135,133,28,183,158,213,
81,100,204,92,110,215,95,165,179,14,34,155,
166,82,49,105,181,67,150,87,29,130,32,193,
113,98,70,197,110,18,44,153,24,89,150,184,
49,165,240,176,242,30,97,254,183,200,149,222,
218,155,135,240,34,124,77,125,217,71,240,202,
120,229,243,38,102,174,251,192,110,16,179,82,
109,65,202,218,243,188,125,109,168,103,112,173,
165,194,91,55,167,15,38,24,220,103,222,122,
39,161,208,98,227,32,111,170,223,103,72,34,
80,140,94,255,69,4,136,84,6,164,105,188,
248,17,166,34,35,173,110,34,76,3,130,16,
76,125,15,117,135,208,250,63,65,187,113,66,
51,136,16,49,132,15,115,140,124,165,131,27,
199,120,113,130,19,197,200,72,61,227,183,90,
155,38,155,106,50,61,252,144,48,210,25,205,
85,49,69,30,201,245,240,131,136,124,201,166,
165,220,36,107,149,232,108,173,227,145,98,10,
107,71,245,210,171,156,72,253,142,237,172,85,
228,243,165,86,143,117,84,75,235,152,57,187,
215,64,246,100,57,160,21,154,204,142,238,14,
133,118,114,30,137,85,130,157,121,138,126,230,
65,217,106,146,212,251,84,147,59,226,158,9,
224,122,10,110,98,81,243,213,174,37,217,10,
227,234,154,140,169,42,152,154,140,169,117,127,
106,90,136,46,245,199,245,244,234,88,13,234,
218,227,204,87,191,215,110,115,61,189,186,47,
222,202,150,249,124,177,137,23,38,152,185,2,
73,77,224,6,18,38,46,66,174,179,104,124,
137,143,138,95,32,63,94,64,88,38,242,220,
10,241,219,55,94,122,186,66,66,150,147,173,
13,59,207,204,104,2,66,166,221,76,48,83,
18,154,222,85,224,241,54,134,41,230,111,81,
154,254,168,239,247,138,245,44,164,204,62,13,
197,49,209,116,35,70,242,199,115,95,183,123,
31,14,67,218,80,201,166,52,218,126,196,124,
213,61,198,253,119,231,41,0,34,233,46,207,
229,117,102,11,58,241,35,65,189,22,30,158,
67,185,97,241,176,124,143,32,253,219,100,75,
179,251,94,183,118,68,8,44,187,213,31,144,
8,193,147,245,206,145,92,252,175,19,5,75,
99,92,179,250,50,18,251,97,56,107,28,108,
40,78,8,93,233,110,30,145,113,141,255,244,
226,199,252,226,68,158,236,88,5,33,75,136,
157,100,71,71,168,179,5,177,96,105,35,192,
141,99,102,71,82,104,178,68,59,138,249,235,
205,38,245,103,26,234,24,196,29,9,199,143,
49,116,25,47,78,200,233,42,237,32,102,181,
85,164,227,229,249,238,245,77,90,145,199,188,
147,48,147,19,188,219,10,208,190,120,70,140,
220,45,99,156,28,33,149,121,159,229,250,75,
55,172,96,107,3,25,7,75,87,184,81,182,
120,188,238,32,132,216,69,213,18,199,130,249,
229,6,19,211,123,181,41,82,190,139,148,62,
191,9,231,174,244,112,130,31,38,52,220,93,
71,111,20,98,172,180,232,134,175,12,5,75,
87,177,141,203,25,194,234,246,131,153,60,223,
232,206,111,172,166,208,21,153,24,152,40,120,
12,23,92,138,217,12,178,20,34,30,189,96,
205,156,164,42,222,34,162,43,107,42,18,137,
225,210,66,247,80,43,105,248,177,203,154,23,
32,105,80,14,117,90,94,194,95,175,117,200,
230,127,25,195,236,95,93,210,13,18,22,27,
29,250,17,43,56,145,161,104,123,17,113,34,
232,119,90,55,219,253,37,116,135,115,6,66,
192,179,117,231,216,26,208,115,155,14,55,135,
95,158,80,52,179,196,98,43,228,70,198,70,
86,55,187,140,178,18,152,166,66,161,160,163,
171,114,247,44,41,4,118,206,224,126,241,54,
65,225,159,161,198,234,1,49,245,232,200,141,
210,208,186,247,124,178,218,57,87,205,113,105,
67,229,198,176,141,87,107,157,248,30,182,161,
246,77,178,214,47,166,183,56,239,21,205,196,
175,234,196,227,162,39,196,178,203,227,91,171,
146,12,151,8,124,193,134,23,18,198,2,191,
45,88,223,84,112,130,18,178,116,29,149,34,
170,101,243,199,207,126,76,74,95,97,124,58,
164,98,104,212,59,42,196,18,250,237,58,190,
36,122,149,60,113,34,88,110,148,201,164,218,
124,127,58,34,76,18,202,166,137,46,75,204,
100,82,232,235,85,252,133,117,86,151,61,70,
38,211,252,92,190,78,108,78,178,190,254,146,
62,58,25,224,57,160,144,214,121,199,84,104,
185,17,229,29,242,194,155,142,143,231,133,76,
94,219,107,40,66,85,71,72,23,147,159,41,
78,4,45,47,218,85,181,152,183,52,174,149,
173,61,213,119,151,197,88,232,170,140,44,73,
8,186,210,205,49,16,4,9,183,114,43,60,
116,110,112,203,122,130,228,5,12,41,243,56,
114,158,207,86,134,145,165,132,182,151,71,136,
5,36,4,34,145,249,163,149,42,154,44,241,
118,62,77,168,37,124,188,26,17,190,152,101,
236,222,7,251,139,169,177,149,19,108,250,56,
65,132,23,38,36,66,32,156,54,24,217,30,
203,198,192,13,133,16,93,105,193,138,217,95,
41,154,165,31,109,82,74,105,29,83,245,88,
174,123,71,54,127,236,135,90,39,100,185,238,
245,26,153,52,61,139,37,52,72,18,108,91,
197,233,68,20,75,6,67,67,6,198,142,231,
9,239,92,231,47,230,43,232,197,223,38,149,
169,236,59,208,235,173,128,23,155,71,199,241,
158,174,117,8,226,228,92,149,21,154,154,204,
173,145,193,200,76,26,170,66,20,15,78,255,
124,251,4,233,181,255,130,219,119,21,100,137,
46,61,135,169,131,235,163,172,110,34,173,108,
34,76,157,103,142,205,243,213,175,162,152,101,
100,73,69,183,74,164,71,75,100,182,54,202,
174,154,92,133,76,229,6,34,142,248,226,139,
127,201,175,126,203,163,241,118,147,213,143,243,
92,127,199,69,24,130,165,45,90,143,229,182,
14,146,64,85,29,126,180,1,155,205,97,12,
89,163,104,228,184,93,123,140,188,176,134,124,
99,140,248,163,60,31,63,206,34,167,254,238,
158,230,186,249,77,151,91,35,233,129,53,103,
153,170,130,153,217,253,173,172,108,118,251,2,
246,235,5,144,16,151,106,51,173,59,33,159,
46,52,153,42,165,168,100,140,75,245,110,110,
24,147,53,85,64,208,142,96,234,70,145,213,
143,157,94,184,49,217,22,251,82,21,226,233,
81,158,198,14,195,249,110,31,69,37,183,196,
114,117,154,156,189,73,118,75,128,43,76,4,
143,154,14,145,16,116,26,10,229,225,131,195,
77,75,53,143,197,186,203,105,2,28,39,46,
153,104,56,33,149,62,69,198,114,86,183,203,
119,103,40,38,111,105,140,23,82,221,18,60,
77,166,146,49,104,172,215,89,236,156,188,186,
102,161,234,98,27,42,217,148,138,145,173,176,
185,106,144,140,148,152,253,80,2,215,103,67,
205,50,31,153,196,178,130,162,201,108,198,22,
181,185,183,40,140,255,125,124,161,237,113,155,
171,237,128,229,186,119,168,50,221,78,188,218,
211,241,166,161,200,18,183,70,50,3,83,16,
83,7,92,26,217,116,35,114,154,143,164,174,
51,27,4,104,65,183,153,242,193,134,96,105,
62,143,164,94,67,200,32,60,13,157,33,74,
215,191,209,215,125,37,69,165,80,250,59,252,
217,211,255,139,214,194,56,182,57,78,198,250,
25,78,172,179,82,29,162,214,30,66,83,3,
102,199,62,165,233,20,112,188,12,133,244,6,
95,25,146,41,155,26,157,168,204,23,134,206,
162,36,225,62,206,50,90,254,29,214,2,155,
68,236,206,254,55,220,144,231,27,29,174,87,
236,99,191,123,18,133,36,113,216,227,53,218,
215,115,247,35,30,63,94,99,248,218,254,164,
136,114,18,35,137,203,101,44,226,68,240,108,
221,33,73,216,183,225,246,162,162,233,70,148,
51,6,205,181,71,40,149,58,201,156,67,140,
210,51,20,37,171,197,134,62,65,113,52,225,
105,185,64,163,190,155,199,108,180,56,71,171,
154,193,121,2,202,112,17,121,164,70,91,196,
68,13,25,179,250,33,153,91,55,15,140,180,
244,19,246,63,51,67,209,116,163,163,66,199,
187,60,16,77,145,240,119,8,10,153,154,140,
109,40,216,91,13,50,94,152,240,162,234,34,
244,147,55,120,9,186,205,66,247,38,50,232,
138,10,226,23,248,143,139,63,33,137,43,36,
202,24,118,238,235,152,153,33,36,89,65,36,
49,170,110,146,13,19,238,175,181,81,229,144,
219,163,153,222,98,253,108,177,53,208,222,129,
55,129,107,101,11,83,27,140,27,239,133,9,
77,119,176,61,33,115,155,14,239,76,100,49,
172,255,146,255,248,248,143,49,83,45,218,173,
18,249,194,55,177,75,185,99,37,27,95,69,
182,114,139,185,79,167,201,151,62,100,227,249,
95,240,233,15,211,84,39,117,42,249,101,70,
138,243,4,145,193,243,213,219,100,204,58,35,
197,46,157,252,39,117,184,155,179,89,114,125,
234,137,0,33,72,37,239,160,90,37,214,54,
246,175,40,90,111,5,84,178,198,177,63,36,
73,146,120,250,87,191,199,228,251,191,65,42,
187,191,33,120,177,210,4,68,175,111,98,255,
208,211,229,236,107,152,219,116,64,130,204,37,
120,151,40,17,248,81,66,193,214,88,93,249,
24,169,20,34,101,82,200,212,136,99,137,141,
13,159,92,105,141,39,197,111,82,200,44,49,
239,236,95,142,152,41,182,72,98,200,38,223,
69,108,4,36,73,132,149,157,198,190,53,177,
239,245,43,13,111,32,70,226,84,134,98,91,
15,34,99,170,71,24,148,144,249,70,125,151,
219,163,43,50,227,133,212,46,239,228,241,90,
135,100,0,34,60,97,156,240,120,181,195,157,
209,12,197,241,175,33,196,71,56,155,203,216,
229,87,63,70,141,106,39,232,241,52,205,14,
219,59,6,216,191,240,70,162,146,49,122,189,
37,167,69,183,159,165,51,240,64,71,156,8,
158,174,119,152,29,186,65,166,52,75,18,71,
228,71,212,94,40,41,17,130,48,18,196,137,
32,74,186,63,227,68,144,8,241,146,219,73,
150,73,155,251,135,54,167,223,249,39,136,56,
66,183,10,172,60,248,67,204,118,7,165,176,
45,159,235,115,109,248,62,171,181,9,130,200,
64,87,3,64,160,74,82,175,1,41,90,51,
24,25,122,159,213,166,127,104,83,210,252,166,
203,245,99,58,21,146,162,82,152,156,166,181,
254,111,208,140,255,1,213,72,239,25,155,141,
134,199,244,141,161,131,67,85,129,139,36,18,
46,43,230,54,28,38,141,16,235,130,191,71,
219,139,152,42,165,144,147,144,64,157,231,94,
38,69,123,185,133,132,32,8,4,115,115,29,
38,53,13,237,3,155,164,152,231,171,73,194,
159,174,213,119,245,247,244,188,200,74,139,206,
179,207,152,188,249,143,15,253,157,181,78,56,
208,222,167,83,117,235,212,58,225,145,134,34,
99,106,140,164,76,54,219,1,65,148,32,1,
51,67,118,175,17,197,9,98,30,173,118,43,
132,164,1,78,204,147,181,14,55,134,108,36,
73,238,17,110,237,196,106,195,239,81,110,228,
45,141,162,253,114,83,245,47,184,145,72,233,
10,211,229,193,81,25,188,216,116,14,236,70,
31,132,75,190,29,151,86,100,9,39,240,112,
252,152,78,189,131,191,217,255,242,76,69,29,
110,152,69,172,87,42,103,36,69,197,202,143,
51,243,213,127,198,242,227,255,64,152,172,236,
34,174,25,46,44,176,217,28,38,138,117,222,
170,52,152,170,215,185,47,117,239,33,183,75,
168,99,5,86,95,52,142,12,57,110,138,144,
241,220,241,222,189,60,245,43,60,127,48,71,
242,236,199,152,185,33,242,163,111,239,50,206,
153,220,225,91,100,199,176,73,228,203,221,236,
182,84,243,40,14,39,125,51,182,158,71,100,
83,221,208,251,230,252,143,49,198,234,140,45,
184,124,90,19,180,153,160,204,34,138,12,217,
225,52,181,164,251,142,186,44,147,81,149,158,
161,144,187,105,236,222,65,45,50,86,240,59,
155,24,118,233,192,144,229,147,181,206,64,223,
225,84,134,162,238,132,76,149,82,71,184,216,
93,26,242,201,98,138,182,31,33,75,210,174,
228,246,102,59,56,147,50,210,106,39,128,53,
184,49,180,247,168,183,88,115,89,172,117,221,
59,85,145,246,240,25,37,23,56,236,43,75,
18,179,195,246,192,200,252,226,68,28,89,22,
124,90,132,113,178,119,97,71,241,203,126,149,
62,224,134,49,159,45,53,153,44,166,24,201,
237,95,23,62,124,237,123,204,205,63,68,157,
172,242,145,109,210,82,100,42,134,78,67,219,
32,211,236,160,254,116,149,133,197,38,51,191,
246,1,247,195,8,89,216,108,182,131,125,184,
157,246,119,243,203,199,220,208,36,69,165,84,
254,101,150,31,253,7,92,191,72,110,228,46,
210,86,114,190,31,170,140,172,219,164,97,23,
8,21,141,203,138,68,116,195,201,119,198,50,
72,103,180,246,250,149,128,61,205,55,9,208,
241,63,227,195,74,154,234,159,222,167,147,100,
208,240,9,242,179,232,184,168,105,147,157,212,
76,186,44,163,72,18,51,233,20,211,105,147,
143,107,45,214,189,173,208,175,20,35,31,80,
169,20,39,221,170,203,65,239,169,167,26,33,
47,140,143,21,162,73,27,234,158,10,168,179,
228,142,169,118,2,30,175,181,119,109,252,115,
27,78,207,72,0,76,151,172,61,11,229,34,
243,217,92,43,91,123,200,228,78,3,69,150,
40,88,23,67,148,91,136,110,55,255,131,229,
246,46,189,223,222,98,87,117,82,220,33,110,
43,108,68,49,149,141,58,169,181,42,242,162,
135,107,26,40,211,195,196,223,127,151,207,231,
44,164,249,25,76,109,154,229,122,127,70,50,
22,130,167,39,56,197,101,135,110,147,26,85,
152,253,112,153,165,71,191,199,252,227,127,69,
109,249,227,190,62,244,106,166,68,164,92,126,
250,140,182,23,177,218,240,6,118,191,48,238,
230,67,255,102,174,206,79,231,26,71,82,4,
13,204,232,73,14,150,128,122,45,162,34,45,
35,210,21,198,245,69,52,83,167,83,151,176,
229,38,247,171,218,150,113,129,138,169,243,86,
214,194,143,18,170,91,90,216,73,71,193,240,
239,162,165,246,119,95,23,107,238,190,107,255,
141,122,20,219,94,197,72,238,228,139,117,56,
107,176,82,247,14,21,36,63,109,120,172,237,
181,25,215,124,106,157,112,151,134,110,46,165,
237,27,199,215,148,139,105,40,202,105,253,80,
109,144,147,226,198,176,77,170,38,179,84,247,
184,8,69,54,13,55,228,147,249,38,99,5,
147,209,220,110,226,202,145,235,191,194,242,3,
120,182,210,226,145,188,134,144,35,210,234,135,
180,31,127,129,176,171,200,139,195,76,221,249,
29,100,85,103,125,117,25,175,211,255,65,168,
229,69,44,214,60,198,11,102,255,223,207,242,
199,220,157,232,48,243,224,25,111,151,214,249,
163,96,132,116,238,239,247,197,191,51,84,95,
197,209,45,162,75,236,81,108,99,173,233,31,
232,41,30,7,155,237,128,185,13,135,200,9,
33,245,122,23,179,38,85,8,233,32,43,18,
145,128,247,203,243,44,27,31,16,151,203,124,
146,46,147,77,92,254,116,46,197,139,214,26,
137,30,16,37,130,31,46,36,108,174,106,36,
126,17,77,26,34,99,220,166,48,251,165,3,
188,90,191,91,121,105,15,254,217,79,109,40,
106,157,240,84,19,168,108,117,45,46,212,220,
51,155,160,40,22,251,10,204,76,20,247,127,
238,179,118,69,207,2,166,166,48,93,62,155,
180,159,68,151,105,182,156,49,88,168,186,23,
66,175,32,17,130,133,170,203,122,203,199,214,
85,162,36,193,80,21,138,105,141,92,229,61,
172,226,40,98,43,17,44,73,50,165,248,107,
136,56,70,153,122,185,38,78,146,151,89,170,
185,100,204,110,137,246,209,30,80,66,179,250,
231,92,143,22,120,252,241,42,229,239,141,64,
235,23,208,172,60,181,206,209,21,102,103,69,
51,126,30,225,133,9,45,239,232,226,153,195,
48,95,117,119,117,214,119,191,27,153,114,250,
245,148,225,150,204,58,197,182,195,156,19,211,
97,136,41,217,71,155,45,51,34,109,96,122,
79,120,212,186,67,49,87,163,45,154,52,107,
35,16,36,104,245,219,84,134,239,97,12,21,
14,165,76,234,248,241,153,138,78,117,71,125,
75,216,187,43,164,221,135,149,13,93,144,84,
80,53,218,29,104,233,153,3,195,53,94,99,
227,104,247,59,17,168,78,155,120,75,208,188,
191,149,211,236,114,7,245,131,3,222,107,109,
169,65,54,165,98,106,47,85,248,188,198,6,
174,20,64,199,59,241,125,207,234,121,15,186,
86,145,96,178,98,227,215,15,15,125,244,51,
23,71,93,59,170,66,222,74,88,109,122,47,
213,5,251,125,222,36,134,208,131,216,27,236,
120,29,114,253,54,137,221,54,54,128,49,163,
191,67,137,83,91,3,213,57,214,239,23,192,
227,167,235,220,24,74,31,232,153,238,28,91,
219,254,58,255,222,11,249,165,223,40,243,163,
249,28,150,53,133,83,93,198,119,35,112,54,
145,0,37,114,144,100,9,169,43,179,220,101,
41,6,198,23,239,243,116,104,154,104,91,125,
177,223,49,219,38,5,12,59,131,155,7,191,
211,189,95,63,201,245,40,236,18,67,246,243,
251,119,172,45,183,214,70,57,132,241,248,176,
245,189,222,10,88,109,122,123,222,107,168,100,
225,214,14,159,227,160,93,35,233,163,209,52,
242,29,100,89,69,62,128,44,213,82,154,180,
30,172,116,53,111,180,17,230,37,155,210,143,
255,128,154,23,49,59,155,230,131,184,193,240,
237,119,248,211,133,113,108,115,5,93,234,96,
164,190,69,232,180,72,194,0,73,213,247,109,
71,72,146,46,101,144,136,18,164,227,238,73,
145,223,215,156,117,71,221,30,122,121,118,180,
251,168,93,15,58,221,182,111,181,75,205,29,
152,246,161,165,152,86,113,244,200,91,78,25,
1,207,158,211,223,239,135,238,239,239,151,56,
239,128,247,90,23,176,238,116,19,238,182,174,
98,104,50,65,232,210,86,179,96,103,79,124,
223,179,122,222,131,174,157,153,185,78,161,79,
221,230,126,230,226,168,107,45,160,52,210,173,
134,90,105,248,253,63,111,18,65,224,244,72,
1,7,55,94,253,95,47,128,21,103,157,74,
118,184,71,21,225,53,215,240,218,171,164,75,
55,80,245,20,113,224,33,27,41,252,101,183,
39,46,127,156,223,31,1,75,161,202,157,202,
193,93,0,194,255,57,168,25,10,227,31,224,
189,120,198,243,70,27,211,254,7,164,43,211,
47,199,184,216,253,221,175,146,231,197,137,224,
241,90,135,57,201,196,55,237,151,31,122,191,
99,22,28,131,20,176,223,123,42,77,208,173,
190,232,32,136,252,238,90,232,139,148,176,187,
182,84,89,162,50,146,63,178,111,107,191,53,
219,112,67,86,235,237,221,223,244,214,123,89,
133,12,214,17,222,159,172,168,167,34,5,220,
94,121,234,134,64,73,18,36,4,19,241,103,
240,52,38,123,61,77,106,72,39,73,64,206,
216,124,178,82,195,168,67,93,181,9,54,45,
38,110,93,71,36,49,138,110,30,232,81,60,
89,235,224,27,41,48,78,176,39,157,37,133,
199,158,240,147,19,158,186,102,191,156,209,217,
76,169,52,120,253,16,162,91,82,214,246,65,
242,35,132,202,133,193,80,214,232,219,72,12,
26,83,37,11,85,145,89,236,112,161,144,8,
120,81,117,153,221,170,136,243,58,107,52,204,
255,135,234,166,1,158,134,236,20,24,190,245,
187,7,211,170,247,129,150,23,177,214,244,119,
9,112,245,54,148,246,50,55,243,127,133,155,
168,44,46,60,36,118,52,150,248,128,145,27,
183,251,14,215,206,14,217,240,232,167,124,81,
158,193,49,46,122,167,193,209,40,216,250,137,
169,82,170,135,112,205,205,87,29,238,142,101,
57,235,190,197,36,10,145,18,129,254,214,36,
252,120,13,146,24,1,204,207,59,68,81,194,
91,111,101,177,23,215,40,149,115,216,249,95,
103,88,86,209,135,243,40,122,138,160,93,63,
240,190,77,55,124,45,161,224,129,4,227,27,
78,56,144,36,231,88,62,117,33,243,3,111,
10,197,180,78,37,243,102,99,212,99,121,147,
130,165,94,184,177,171,182,131,94,14,34,63,
114,15,165,94,66,29,114,81,167,154,200,22,
125,211,182,28,134,245,3,72,36,213,248,175,
40,110,46,50,211,122,142,174,9,166,222,249,
109,70,110,252,131,99,221,91,145,37,226,183,
238,224,107,198,165,95,231,138,44,29,152,79,
236,7,135,81,207,116,252,152,167,235,103,127,
210,241,26,207,169,110,10,30,124,17,162,203,
49,195,55,198,41,15,89,68,91,98,103,178,
44,129,44,161,142,203,84,87,254,130,84,110,
20,69,63,186,23,234,56,44,219,111,220,80,
108,51,65,158,126,65,192,76,229,242,159,142,
6,129,188,165,113,163,98,159,139,103,25,203,
167,72,155,23,207,88,244,136,30,37,9,41,
25,194,82,21,16,160,196,5,58,167,101,255,
21,98,223,210,113,183,250,41,211,222,3,130,
166,203,143,243,19,132,198,215,78,252,43,38,
86,159,98,132,254,165,95,235,251,149,176,31,
107,156,138,169,67,213,28,55,219,193,190,197,
46,131,68,236,111,16,169,163,200,181,26,185,
114,154,82,70,208,40,124,200,236,173,28,153,
140,138,174,203,120,215,199,89,143,98,68,216,
159,123,83,237,4,167,95,167,175,211,80,108,
63,244,32,144,179,180,125,221,245,43,236,240,
36,108,157,155,195,105,206,13,205,143,212,109,
108,188,104,253,39,78,16,247,202,165,115,133,
143,104,61,205,193,23,119,41,148,63,194,9,
78,121,240,145,164,61,205,119,110,237,1,185,
167,191,135,247,249,83,156,247,175,179,182,62,
78,166,242,214,137,127,197,231,217,201,75,239,
81,164,77,245,212,37,223,146,4,111,141,164,
119,233,124,188,138,181,166,207,163,213,246,137,
152,171,251,122,143,225,143,48,203,223,37,165,
36,108,110,122,180,171,117,110,79,45,241,120,
234,123,20,126,249,23,152,127,235,91,252,81,
124,135,23,247,71,41,95,255,187,125,221,115,
173,249,250,170,15,7,102,40,6,217,97,61,
85,74,93,233,240,30,128,114,70,231,198,176,
205,121,227,130,51,84,121,79,135,251,69,192,
54,105,90,186,52,195,228,196,191,96,242,157,
127,132,85,152,220,37,77,122,82,236,220,152,
218,43,127,192,104,248,175,153,181,235,148,191,
127,143,191,126,56,129,165,221,62,241,189,221,
48,198,126,250,224,82,123,20,154,34,31,171,
39,229,48,168,138,196,237,209,244,161,205,168,
181,78,151,230,252,225,74,123,224,148,53,146,
162,34,68,27,93,248,196,49,116,58,49,27,
159,175,114,239,157,38,76,148,72,87,20,198,
45,147,162,234,99,101,142,46,30,136,18,65,
203,11,95,219,92,12,204,80,196,137,232,171,
246,187,175,135,146,164,45,158,166,43,195,176,
19,67,89,131,153,138,205,121,29,150,162,173,
51,94,72,93,168,121,171,59,97,79,63,225,
176,88,74,0,0,32,0,73,68,65,84,68,
213,83,61,221,135,65,112,11,109,111,74,161,
87,103,172,241,199,92,75,59,152,182,202,223,
188,24,103,114,246,119,41,77,126,253,68,247,
13,99,193,195,149,54,207,43,215,47,173,71,
33,73,48,59,108,15,148,218,94,83,100,222,
26,75,31,57,183,117,39,228,254,114,139,159,
206,53,152,219,112,6,146,175,2,136,157,57,
146,150,67,77,30,161,153,185,67,16,74,168,
181,6,249,148,195,211,205,44,75,155,79,176,
181,247,251,94,183,175,179,249,117,160,153,227,
65,182,194,219,134,194,68,49,197,21,94,158,
78,167,75,231,255,196,62,94,48,185,55,158,
37,109,92,156,156,69,175,31,100,7,172,1,
200,114,154,186,76,18,71,200,115,255,43,179,
102,21,127,122,20,223,206,32,180,41,56,161,
50,93,156,8,30,174,180,240,195,132,233,245,
103,151,214,163,152,42,89,167,106,174,59,8,
186,34,115,123,52,131,38,31,61,254,97,156,
176,218,244,121,188,214,97,190,234,158,184,6,
46,137,67,68,28,161,182,30,98,73,62,197,
100,153,130,115,159,112,252,109,30,6,55,249,
124,101,130,213,185,77,50,242,239,80,153,254,
86,95,247,124,221,196,165,3,157,137,166,27,
225,135,9,198,128,52,16,70,115,38,77,39,
218,69,187,241,183,21,51,21,235,194,156,212,
45,93,225,238,120,134,181,166,207,74,195,27,
72,24,231,44,225,239,195,141,211,143,42,99,
63,227,208,169,62,33,103,8,126,122,227,58,
203,171,53,180,141,73,204,210,135,39,186,159,
16,240,120,181,211,75,96,206,93,82,143,162,
146,209,25,62,195,60,165,161,201,84,178,58,
75,199,176,177,203,117,143,142,31,159,200,203,
233,44,252,17,35,155,127,128,213,106,83,173,
250,140,12,219,88,57,133,145,239,164,65,93,
167,227,36,44,43,31,97,23,167,143,113,96,
120,189,115,50,240,90,212,65,19,108,205,12,
89,23,150,123,105,80,40,216,218,161,85,27,
231,21,67,89,131,119,39,115,220,27,207,50,
154,55,207,37,85,180,44,73,20,109,109,159,
77,94,61,245,125,77,77,33,83,185,69,60,
245,63,179,240,249,91,132,143,238,144,157,250,
239,145,78,72,13,254,116,189,179,235,208,116,
25,61,10,219,80,185,86,62,251,106,190,224,
4,135,151,166,27,242,104,165,125,236,144,79,
208,89,103,36,29,32,86,87,80,100,9,59,
173,161,32,144,130,238,92,134,106,130,43,253,
4,183,177,116,44,207,242,194,122,20,0,27,
237,96,160,33,35,77,145,153,169,216,60,88,
105,255,173,53,20,67,23,92,63,120,91,201,
112,60,167,225,118,36,48,50,4,81,87,245,
43,216,250,179,253,207,225,107,62,42,141,23,
204,125,11,39,44,93,65,130,19,135,27,236,
29,161,43,73,86,24,191,245,155,47,67,26,
238,241,235,246,87,26,222,158,198,170,197,210,
20,193,37,226,122,210,20,153,155,103,92,168,
33,4,60,219,232,176,209,9,224,4,219,84,
203,139,152,175,58,76,29,35,12,92,188,253,
59,60,121,248,140,130,181,72,218,214,88,91,
245,40,166,5,25,199,67,88,38,170,44,33,
77,172,179,244,236,95,113,45,245,63,30,202,
233,180,141,68,92,112,67,17,68,9,13,39,
36,55,192,110,225,156,165,49,146,51,89,25,
32,213,240,69,129,161,202,123,198,50,112,155,
104,186,133,164,92,60,47,67,147,101,116,67,
229,0,117,79,132,128,32,78,104,215,124,68,
202,238,25,146,48,78,186,74,119,241,110,197,
187,211,26,176,131,8,45,37,169,75,180,232,
158,226,222,131,130,19,196,251,54,86,141,212,
150,240,180,235,68,23,112,29,236,25,111,186,
201,107,253,12,189,206,68,8,158,172,118,168,
57,167,11,101,175,52,124,108,67,237,155,141,
66,146,21,54,152,161,242,43,54,238,143,62,
69,248,155,100,111,20,72,178,54,2,88,116,
124,12,69,38,20,18,138,222,223,161,112,40,
107,208,246,163,94,33,198,133,51,20,221,240,
83,48,80,67,1,93,166,215,186,19,156,251,
120,247,160,177,95,82,181,182,252,19,28,231,
33,186,152,34,63,250,206,165,122,95,73,234,
26,71,161,43,152,125,124,136,219,242,168,94,
35,68,207,102,17,162,235,5,116,127,10,90,
110,196,106,211,223,195,209,47,73,112,189,114,
248,233,53,165,203,184,39,92,110,131,12,21,
174,54,252,125,195,29,107,249,97,66,245,114,
80,140,151,51,198,153,36,175,119,174,147,135,
43,237,129,52,6,3,60,91,119,80,21,137,
92,170,191,241,79,165,60,74,247,127,78,80,
177,232,172,84,73,189,115,141,100,203,64,214,
131,16,231,121,137,74,241,239,109,253,151,163,
145,49,85,222,153,200,178,92,247,216,104,7,
103,110,48,206,196,124,215,156,128,40,30,176,
194,146,36,157,25,141,246,121,198,126,97,17,
51,61,140,174,13,49,122,247,123,39,142,119,
95,22,40,178,132,174,202,232,106,55,39,144,
210,21,44,93,193,54,20,210,134,202,104,222,
228,237,137,236,158,15,250,90,217,58,50,97,
125,154,143,111,144,155,94,253,128,19,112,185,
185,142,22,71,23,126,14,37,9,198,6,212,
47,113,144,39,113,127,121,112,70,98,251,158,
143,86,58,52,251,44,180,17,66,102,249,69,
11,51,173,49,117,187,140,212,118,64,215,112,
227,132,134,27,147,150,191,66,166,124,243,216,
123,226,120,33,197,123,219,121,192,156,121,102,
249,220,51,49,225,66,116,115,21,35,185,193,
198,214,183,133,134,46,130,30,194,160,176,31,
13,68,110,232,46,185,161,187,92,161,207,69,
46,75,220,26,77,51,183,225,176,222,242,169,
100,117,42,71,228,125,90,94,68,39,136,79,
20,199,78,155,234,64,67,40,7,197,163,235,
86,254,82,132,157,42,25,227,204,10,29,94,
86,138,13,222,160,38,66,240,112,165,195,205,
17,251,72,207,66,36,2,215,147,241,231,214,
177,127,237,43,196,106,247,144,242,179,90,11,
119,37,197,228,200,123,167,122,150,237,60,96,
73,74,163,100,114,52,189,144,150,27,209,244,
162,129,40,222,157,90,143,226,64,119,217,87,
200,196,54,254,0,52,16,118,157,162,36,193,
102,167,117,166,250,14,103,114,237,9,159,183,
209,145,232,168,233,94,35,216,73,198,236,188,
92,43,146,152,200,115,136,130,163,35,255,253,
106,0,28,231,250,33,25,74,25,65,216,106,
225,84,151,15,188,206,9,98,230,54,28,112,
27,253,207,217,142,249,205,106,38,78,181,125,
234,241,138,252,46,255,144,234,10,226,104,239,
129,193,174,47,80,151,34,216,46,145,189,128,
122,20,18,80,200,100,112,170,141,51,89,135,
139,53,143,186,19,156,238,123,60,228,218,4,
120,240,24,74,105,157,178,145,160,40,251,235,
81,200,146,0,33,72,144,144,151,214,73,166,
70,8,19,193,218,115,25,169,62,66,144,110,
16,116,246,31,131,208,105,161,104,198,129,58,
23,175,142,129,73,87,228,206,86,97,36,13,
97,36,104,7,33,29,47,166,229,69,47,213,
19,95,167,30,197,129,235,6,112,53,27,43,
55,24,13,132,222,160,197,9,180,82,103,170,
239,112,38,215,158,240,121,19,192,55,109,138,
246,233,244,62,206,195,181,73,28,17,185,109,
244,116,254,232,225,234,83,3,224,36,215,59,
178,116,224,243,174,53,125,94,52,92,18,203,
6,73,233,127,206,182,230,55,155,82,25,31,
201,156,72,55,97,239,126,222,53,54,186,16,
248,251,132,77,188,92,136,72,87,96,187,242,
233,2,234,81,100,83,26,185,74,250,76,214,
225,102,59,160,38,117,246,151,6,61,206,247,
216,199,181,155,2,218,142,203,244,80,134,124,
54,179,143,173,81,17,36,212,59,144,111,117,
199,191,29,197,72,181,113,174,189,243,79,121,
117,193,136,173,205,92,146,36,124,205,68,53,
82,125,85,67,29,52,94,185,29,94,80,181,
29,178,209,246,105,86,95,179,30,197,65,88,
172,185,220,28,96,73,180,16,48,183,241,122,
104,117,207,19,22,170,30,5,75,191,162,52,
57,67,8,1,143,215,218,167,162,161,201,89,
26,179,103,64,61,35,239,115,195,145,218,18,
35,181,101,150,138,227,23,122,220,139,233,179,
75,198,191,42,123,122,214,240,163,132,135,43,
29,70,67,149,201,210,238,152,101,168,86,184,
121,35,69,44,52,228,213,42,34,151,38,51,
54,132,102,105,236,183,96,118,70,16,90,235,
143,105,180,126,132,22,143,99,149,134,41,79,
127,227,84,107,169,156,209,41,103,116,154,134,
75,83,164,216,236,196,251,54,157,238,250,255,
206,122,224,170,157,193,117,85,63,88,105,157,
136,165,86,237,37,60,101,46,226,94,235,133,
49,107,205,203,79,39,253,166,81,63,225,90,
205,152,42,55,135,108,222,26,73,159,9,131,
238,206,28,69,202,119,248,198,231,127,198,106,
126,148,191,185,241,229,11,61,222,146,4,5,
235,108,250,64,26,110,56,48,142,166,99,27,
168,134,199,98,109,183,145,18,137,75,152,43,
96,170,49,237,64,197,121,90,101,126,33,198,
221,220,64,36,135,63,103,18,133,168,148,112,
155,139,164,178,19,3,123,78,93,145,24,203,
155,188,59,149,227,122,197,58,52,175,118,230,
153,176,245,150,199,104,34,6,242,1,101,76,
13,69,146,144,36,9,69,209,49,243,41,20,
73,34,140,19,194,173,250,250,56,17,168,138,
132,165,171,216,70,183,2,198,175,123,88,197,
174,243,21,39,130,166,27,241,162,234,188,182,
26,228,65,121,103,165,180,142,170,92,185,21,
103,181,105,233,170,124,228,201,234,85,104,138,
196,205,145,52,81,235,108,78,175,27,45,159,
150,39,35,33,248,242,163,191,226,225,248,109,
126,112,247,91,151,98,204,179,166,118,102,235,
121,163,21,188,241,239,85,87,101,42,25,157,
56,244,81,237,247,249,100,245,17,239,76,166,
248,172,145,161,169,90,180,63,174,160,106,16,
184,155,24,189,240,255,94,228,199,238,82,210,
63,4,73,66,57,3,202,22,137,110,65,65,
57,109,176,218,244,89,172,185,123,122,148,206,
220,80,68,9,60,93,235,112,115,36,125,234,
123,237,164,28,246,26,14,102,238,248,37,117,
138,44,81,176,53,156,192,96,177,118,113,194,
88,81,34,120,177,233,48,51,100,115,133,179,
193,80,214,56,182,98,88,218,84,81,101,137,
179,40,82,109,123,17,11,85,151,153,234,50,
142,97,241,241,245,15,47,85,39,246,89,133,
157,4,93,213,205,55,141,249,170,67,209,214,
80,84,157,165,207,254,28,35,51,206,95,54,
190,65,216,106,99,151,110,162,217,159,18,231,
230,168,45,125,204,200,205,239,29,189,119,157,
194,72,8,33,14,44,136,217,121,88,26,201,
25,228,45,149,135,43,157,93,21,151,175,133,
124,167,230,132,44,156,163,77,57,17,98,96,
66,75,175,19,27,237,96,224,92,90,87,120,
137,225,156,177,199,253,30,202,26,76,20,83,
251,246,92,108,215,177,159,5,194,56,97,121,
185,198,91,139,15,88,41,140,178,82,24,187,
84,70,2,32,127,70,97,167,181,134,79,244,
154,185,144,246,61,220,197,130,213,166,15,146,
196,204,55,254,9,1,77,220,230,10,237,206,
99,52,35,77,34,66,80,100,124,229,17,129,
219,60,99,143,249,112,35,225,183,215,169,46,
252,132,36,244,49,53,133,123,227,153,93,170,
149,175,141,165,109,169,230,157,155,254,135,231,
27,206,153,41,89,157,53,230,54,220,55,22,
123,189,236,144,37,137,201,98,10,89,146,80,
101,137,169,98,138,107,101,139,177,173,166,189,
27,67,187,153,67,135,178,198,64,88,102,247,
158,254,224,201,90,7,17,6,44,22,199,112,
140,203,231,69,218,166,122,38,205,97,45,175,
27,86,62,47,88,169,123,196,137,64,86,84,
38,110,254,183,88,217,41,102,63,252,23,196,
113,128,87,173,49,148,138,185,113,183,206,242,
252,191,62,22,41,224,192,215,190,106,224,180,
159,243,248,7,255,146,36,142,81,100,137,155,
59,40,85,94,107,183,206,147,181,14,45,47,
98,178,152,122,35,178,153,66,192,66,205,125,
227,241,203,211,122,67,143,87,219,220,30,203,
160,43,50,87,24,44,74,105,157,156,213,205,
133,249,205,112,207,223,101,83,42,79,214,58,
56,65,204,88,254,224,208,167,23,198,196,201,
201,56,159,158,109,116,104,186,17,101,223,69,
77,34,90,151,112,156,243,169,193,111,61,13,
55,100,169,217,70,136,243,243,158,81,34,88,
111,249,140,228,76,140,116,9,35,221,21,171,
170,173,252,21,74,201,103,101,65,97,185,170,
33,25,53,22,150,254,119,242,155,191,68,101,
230,235,175,253,57,53,51,203,196,237,223,36,
153,13,145,183,180,58,52,69,102,118,216,62,
216,80,200,146,68,74,151,207,68,184,123,173,
233,211,112,66,102,134,236,51,229,118,217,239,
195,125,178,214,121,109,98,228,103,251,46,9,
159,47,182,184,53,128,188,207,21,246,226,48,
189,1,109,75,248,38,17,98,223,178,85,232,
146,198,205,87,29,36,36,238,77,100,14,149,
223,124,21,139,53,175,119,144,241,53,131,56,
81,47,229,24,103,83,131,203,79,36,66,48,
183,225,178,81,245,16,118,230,220,189,235,106,
211,223,67,62,57,113,247,55,89,249,252,143,
16,66,144,207,191,139,179,57,135,61,114,13,
171,48,246,102,189,234,87,154,168,183,5,200,
118,175,66,9,202,105,157,137,98,10,93,149,
121,186,222,57,147,211,183,31,37,124,177,212,
98,36,103,50,81,52,15,252,224,78,178,96,
162,68,16,199,221,159,81,44,232,4,17,205,
13,23,167,209,122,237,212,188,103,137,96,107,
12,135,228,16,179,32,6,54,134,87,232,63,
76,181,231,244,24,11,158,174,119,122,220,76,
2,193,243,117,135,183,70,211,135,206,79,24,
11,156,32,98,165,225,239,74,194,106,113,136,
122,9,184,156,94,133,161,201,251,114,152,157,
4,78,16,243,100,181,131,27,198,231,182,244,
221,15,247,103,212,46,78,125,169,215,36,106,
191,210,15,19,39,130,106,39,36,138,19,10,
138,224,77,51,186,169,0,239,76,100,241,194,
4,213,245,200,236,16,13,41,167,141,51,13,
211,108,115,236,87,50,71,115,239,236,23,70,
106,186,17,53,39,160,225,132,4,145,56,208,
16,72,126,132,80,197,165,251,224,226,68,176,
210,242,216,20,77,110,142,164,7,74,109,125,
133,99,206,133,128,199,203,173,61,249,163,150,
23,241,249,98,171,183,49,10,4,113,211,37,
118,91,91,101,221,201,129,161,146,248,146,18,
62,14,34,236,36,68,183,161,110,177,238,114,
17,206,127,171,77,191,111,70,237,133,170,203,
106,179,203,24,60,150,55,207,197,33,80,133,
174,30,115,74,87,112,94,41,168,49,245,179,
143,129,135,113,194,82,221,99,169,238,145,137,
92,134,141,128,108,74,219,215,253,15,99,65,
211,13,169,117,66,26,155,45,98,83,189,218,
161,232,234,55,84,219,1,182,113,165,49,254,
166,176,222,244,113,34,245,192,83,239,78,3,
34,121,17,66,57,218,83,144,133,64,78,46,
31,173,126,238,148,213,78,110,24,243,116,205,
57,19,162,191,179,66,221,9,241,163,228,72,
242,195,197,154,199,114,221,163,146,53,24,47,
164,208,20,137,160,253,230,117,120,14,221,105,
117,69,70,150,164,215,22,178,105,251,17,173,
213,46,15,138,170,116,105,163,77,77,38,78,
4,29,63,222,205,130,40,4,87,120,137,166,
23,114,34,170,211,43,12,4,87,203,177,63,
40,178,68,218,80,17,201,201,54,249,149,134,
199,66,213,187,144,97,228,103,235,29,102,135,
211,7,230,192,154,110,196,70,219,231,222,68,
246,76,170,233,206,204,80,192,22,13,201,27,
152,147,40,22,180,227,136,182,119,245,113,245,
131,142,31,19,37,226,216,194,239,87,24,12,
10,105,141,245,198,96,239,153,72,18,146,124,
185,42,219,242,150,134,36,29,127,75,241,195,
132,167,235,157,129,106,74,188,246,195,156,27,
241,241,139,6,89,83,197,84,5,146,31,96,
42,1,138,36,225,133,49,27,237,128,219,163,
153,115,169,45,127,168,161,112,130,248,181,139,
120,95,225,52,11,49,60,148,101,246,10,103,
7,83,149,177,13,117,160,225,16,53,142,80,
147,203,213,51,99,104,199,223,4,215,154,62,
243,85,247,82,236,69,113,34,186,82,172,73,
4,190,215,165,103,223,58,144,223,25,59,159,
70,226,72,67,113,69,68,119,241,78,44,87,
134,226,13,110,130,170,76,103,128,159,76,160,
234,68,226,114,229,40,142,227,241,250,81,194,
243,13,151,182,234,92,250,181,51,81,72,245,
74,81,47,148,161,240,194,248,138,46,226,130,
161,225,132,8,193,21,29,249,27,130,174,14,
118,224,141,208,71,77,34,26,151,104,140,218,
94,76,37,117,116,252,125,181,233,51,191,233,
34,252,232,53,183,5,191,126,228,82,26,163,
121,243,92,63,227,190,126,78,148,8,30,173,
118,174,18,116,23,12,126,148,240,100,189,115,
53,16,111,8,169,1,39,32,93,35,69,231,
146,209,119,84,59,65,87,98,246,144,3,234,
23,75,45,230,54,156,75,213,247,116,16,50,
166,202,205,145,243,63,199,42,128,215,88,239,
186,186,237,26,200,42,207,214,59,135,115,33,
133,110,87,253,171,31,146,50,175,209,159,60,
226,73,174,237,23,103,249,12,231,236,121,171,
46,60,115,245,125,79,40,65,187,134,220,167,
198,242,89,92,155,36,49,177,231,144,196,71,
51,123,250,237,218,241,140,228,49,174,239,247,
121,143,115,79,191,93,67,209,99,112,59,3,
155,95,187,177,210,165,240,208,245,193,174,177,
109,41,212,216,31,220,61,253,118,247,126,125,
188,215,211,185,136,137,188,78,58,253,146,40,
52,22,176,222,242,217,104,6,36,59,83,221,
231,229,123,236,119,92,251,221,23,147,24,130,
54,19,102,134,160,121,120,197,78,232,180,144,
53,3,165,15,41,212,227,124,183,126,187,70,
146,196,200,253,74,161,110,119,7,198,81,196,
130,151,162,163,168,135,87,90,42,71,75,161,
238,26,144,126,37,7,143,115,45,156,205,125,
47,193,243,174,199,32,2,157,169,146,181,43,
38,156,196,81,223,114,161,103,113,109,18,71,
68,90,127,82,168,59,215,101,191,56,139,119,
59,206,51,104,153,50,180,234,3,155,223,118,
46,70,22,98,240,107,76,57,134,20,106,191,
247,148,212,190,165,80,131,200,103,174,19,51,
156,74,147,77,105,180,253,136,149,186,71,148,
168,144,178,207,223,247,120,156,113,237,123,95,
140,64,86,232,40,57,70,114,135,95,47,43,
26,138,110,246,37,133,26,71,225,174,53,187,
83,82,117,251,223,119,50,201,234,233,66,95,
134,69,238,186,123,9,47,54,93,30,44,183,
122,244,3,87,184,216,216,104,5,124,58,223,
188,144,116,234,23,21,138,44,145,183,6,199,
97,148,117,155,20,218,213,75,57,86,113,34,
88,170,123,220,95,110,177,80,117,207,5,45,
248,155,192,124,213,25,40,27,244,171,116,226,
73,232,225,62,249,75,98,175,189,239,223,247,
11,25,224,147,249,6,43,13,143,248,42,39,
113,161,241,234,26,8,227,132,199,171,29,30,
173,182,9,226,228,106,128,94,3,198,6,152,
148,172,217,5,54,178,229,171,65,189,196,16,
2,94,108,158,172,170,43,142,250,59,4,202,
149,155,68,225,233,10,147,174,56,48,46,1,
74,233,110,78,34,165,41,116,252,136,106,39,
100,181,233,245,138,17,106,157,144,166,219,100,
76,11,177,174,134,235,76,145,54,85,134,179,
93,73,201,211,162,208,169,161,69,33,243,233,
43,99,113,153,113,162,46,108,33,88,254,225,
255,205,196,183,126,231,112,47,87,79,161,232,
167,103,108,184,50,20,231,121,1,25,10,37,
91,239,113,5,121,97,220,219,252,37,169,75,
218,88,76,167,201,237,144,71,77,155,42,105,
83,165,156,209,121,182,254,146,86,61,78,4,
11,85,15,223,112,152,46,89,87,37,180,103,
136,233,178,69,24,159,94,69,177,102,23,144,
197,149,39,120,81,81,206,232,84,219,225,129,
213,91,186,34,51,61,156,166,96,107,59,246,
127,209,95,120,72,146,142,52,18,131,196,149,
161,56,99,108,207,249,73,162,122,166,170,236,
170,94,18,162,75,136,22,39,130,148,174,160,
202,18,78,181,113,224,41,229,238,120,150,133,
170,203,114,253,101,85,197,90,211,71,150,96,
170,116,229,91,156,37,102,135,109,106,29,157,
197,218,201,21,9,175,60,138,139,141,225,172,
201,100,209,98,185,238,177,217,246,9,183,98,
251,154,34,83,202,152,148,53,176,108,237,149,
253,98,127,35,145,196,33,10,111,174,215,226,
202,80,156,17,74,105,157,241,66,10,83,147,
73,132,96,109,197,101,61,86,142,37,193,186,
139,4,113,203,232,28,199,77,149,128,201,45,
189,231,103,235,78,207,88,85,59,33,83,165,
171,57,58,107,20,108,141,130,173,81,237,4,
44,213,61,156,99,138,102,93,121,20,23,27,
115,155,14,119,199,50,76,149,82,76,149,82,
56,65,140,166,72,104,138,76,18,71,199,98,
133,93,255,233,239,51,250,181,127,136,244,134,
168,231,175,12,197,128,81,76,235,76,20,204,
93,194,44,178,212,173,134,25,206,102,183,40,
213,143,230,208,87,100,137,161,172,49,144,103,
42,165,117,76,77,230,225,227,117,66,24,104,
101,206,21,250,88,19,182,78,209,214,105,121,
17,11,85,151,118,159,61,145,87,30,197,197,
70,219,139,88,107,250,189,239,248,52,140,176,
195,95,254,245,55,102,36,174,12,197,0,97,
106,10,179,195,246,161,139,65,146,96,188,96,
82,76,107,172,54,124,154,110,72,20,11,36,
169,235,142,218,134,138,109,40,40,182,77,113,
40,63,208,60,130,109,168,220,24,182,217,72,
244,35,235,182,175,112,54,200,152,42,119,198,
50,44,38,53,86,35,233,200,146,208,43,143,
226,226,163,214,9,7,118,224,235,186,22,15,
208,133,71,39,61,67,82,123,129,90,156,64,
73,229,174,12,197,155,134,36,117,115,5,110,
24,31,26,98,152,169,216,40,125,18,158,165,
52,133,107,229,131,115,4,78,36,159,73,178,
89,149,37,102,202,246,165,157,171,56,240,136,
215,31,161,143,191,115,174,159,179,96,105,12,
229,178,60,223,112,168,117,194,43,143,226,178,
238,29,116,19,218,39,69,228,187,72,73,72,
214,157,39,43,155,88,178,76,144,10,89,238,
40,40,186,137,156,41,35,94,147,176,213,149,
161,56,2,55,135,211,228,45,141,229,186,199,
124,213,221,243,247,150,174,48,59,148,190,170,
34,234,23,66,224,183,54,193,107,32,217,101,
116,43,219,251,171,176,190,68,134,14,25,29,
214,90,13,18,59,191,71,236,253,80,3,92,
127,68,165,100,176,120,1,134,65,83,100,110,
14,167,169,118,2,158,111,56,68,251,52,49,
93,121,20,23,23,170,44,49,59,108,147,77,
157,44,204,27,215,151,80,171,15,185,62,61,
197,247,126,227,151,105,181,90,252,229,127,254,
152,7,203,45,50,179,95,235,202,163,102,135,
95,223,251,92,77,233,193,176,116,165,23,207,
31,205,155,212,156,144,246,43,194,41,147,197,
212,161,70,66,136,4,214,31,80,148,67,108,
167,78,181,19,17,228,174,247,213,142,255,186,
224,44,63,196,26,189,117,172,255,39,9,28,
226,218,60,66,179,80,91,171,228,181,38,178,
36,227,135,49,181,196,38,137,3,108,201,195,
54,52,162,68,16,197,96,42,49,69,91,99,
246,218,4,195,149,41,94,44,46,179,176,242,
20,4,40,138,204,187,95,187,65,185,212,205,
178,39,73,194,31,255,224,175,89,216,112,137,
133,132,42,75,104,138,132,34,75,132,177,32,
140,5,25,83,69,149,187,255,220,104,183,176,
45,149,185,149,42,234,181,139,179,198,138,182,
78,214,212,152,219,116,216,108,7,87,30,197,
101,216,55,12,133,155,195,233,19,106,75,8,
148,245,207,25,73,193,127,241,143,127,139,78,
167,195,191,253,255,254,132,185,6,40,185,97,
114,250,58,222,243,31,161,95,255,250,235,53,
124,87,211,122,248,71,188,19,163,57,147,71,
91,173,240,219,176,143,208,237,214,214,62,225,
119,254,254,183,49,12,163,183,1,254,248,167,
159,242,131,159,252,156,212,236,55,80,141,55,
87,166,26,7,30,122,237,1,211,70,196,220,
243,159,96,91,22,81,34,72,20,147,116,176,
73,197,109,227,6,49,155,129,142,156,27,67,
111,60,97,36,107,34,68,130,231,121,252,87,
191,241,75,212,235,117,178,217,247,80,213,151,
227,176,180,188,66,218,182,200,102,179,135,254,
254,59,183,102,185,115,107,118,223,191,147,101,
153,239,125,243,171,199,122,159,102,179,73,173,
209,226,179,199,79,0,193,102,59,66,136,132,
68,82,112,66,16,118,5,45,93,60,119,235,
76,85,36,110,12,217,20,109,157,231,27,14,
225,86,23,125,221,206,95,74,205,236,203,12,
89,146,184,53,156,70,63,161,0,81,180,250,
144,127,242,171,95,199,52,77,62,253,226,33,
127,250,249,10,73,106,8,173,146,70,209,77,
98,235,107,40,193,235,151,253,188,50,20,7,
133,49,116,133,225,87,146,190,57,75,237,202,
56,138,151,238,229,81,66,44,57,59,213,51,
18,219,27,224,87,63,124,151,39,243,43,212,
234,139,248,128,146,29,69,77,165,207,56,226,
211,221,112,194,245,167,24,82,128,18,121,124,
105,118,140,247,190,249,75,132,97,136,16,2,
203,234,26,173,78,167,131,109,191,204,101,52,
154,77,238,63,122,202,189,143,62,34,109,239,
206,113,20,139,123,55,222,177,209,145,55,50,
103,217,108,150,108,54,203,244,228,248,222,15,
48,138,248,252,193,99,22,214,30,179,212,217,
164,211,94,67,207,13,161,228,70,207,205,154,
43,216,26,153,84,150,23,155,14,27,173,128,
92,167,129,22,135,56,153,202,213,7,121,81,
66,78,138,116,34,35,145,196,33,202,234,207,
121,107,172,136,105,154,196,113,204,143,190,88,
68,27,186,77,232,180,122,215,73,178,130,98,
190,254,60,227,149,161,216,111,195,73,169,220,
28,78,239,73,78,47,213,188,93,101,173,59,
75,96,15,194,74,71,162,217,108,238,58,93,
87,171,85,174,141,150,248,205,247,222,70,146,
36,62,187,255,136,31,60,89,1,35,139,145,
63,163,77,225,201,127,98,186,50,196,183,127,
241,75,20,11,187,217,91,119,122,3,192,46,
35,1,144,203,102,249,234,151,222,191,216,31,
176,170,242,238,189,219,188,123,239,165,215,243,
7,63,248,41,65,118,228,92,41,61,169,178,
196,76,197,102,36,103,82,75,138,52,79,32,
26,47,73,160,72,221,48,157,34,75,8,1,
137,16,221,159,8,132,44,117,255,253,234,83,
31,56,130,40,33,138,5,170,114,188,53,37,
43,26,229,140,206,47,127,227,43,0,44,44,
44,160,7,85,234,75,95,160,228,39,222,252,
186,188,154,218,87,60,128,148,202,205,145,204,
158,189,195,11,19,150,27,222,30,15,99,63,
248,245,85,140,252,48,254,252,207,240,125,159,
249,165,21,238,237,48,20,165,82,137,111,148,
94,118,188,125,248,222,219,164,237,167,164,76,
157,251,79,158,81,147,59,232,78,139,148,38,
179,214,244,105,5,18,137,98,32,103,134,80,
76,27,73,234,255,196,18,182,107,40,237,101,
190,252,254,59,124,249,189,183,175,38,120,135,
215,243,91,191,250,109,254,245,239,255,25,126,
249,30,138,182,127,9,163,72,226,55,82,191,
110,233,10,101,51,1,25,30,23,82,108,180,
3,188,173,202,59,89,146,48,117,25,75,87,
176,116,133,148,174,96,168,50,81,59,196,202,
231,187,137,206,195,214,132,219,93,183,178,97,
227,110,209,195,244,254,248,241,223,10,193,160,
179,12,61,245,91,253,40,146,152,200,109,131,
44,19,53,86,120,231,189,235,56,142,195,191,
253,195,63,99,161,234,98,100,43,164,70,110,
19,180,235,111,252,189,174,12,197,14,140,228,
12,202,210,254,201,105,83,147,185,53,146,166,
229,70,248,81,130,161,201,228,147,189,167,189,
176,190,68,174,249,136,78,42,3,118,153,89,
123,133,122,253,104,225,147,91,179,51,0,76,
78,236,127,122,112,93,151,7,143,159,177,178,
177,194,226,90,149,104,236,203,135,110,96,73,
20,80,108,222,231,131,187,51,220,184,254,45,
100,89,190,154,224,87,55,99,203,226,215,190,
243,101,158,189,88,228,15,255,244,135,228,166,
223,69,18,49,36,33,202,208,109,0,26,15,
126,64,246,214,47,244,45,6,51,72,248,166,
133,172,27,140,21,76,198,10,38,113,34,136,
133,64,87,246,159,75,79,150,142,52,18,59,
161,200,82,143,27,172,183,121,9,232,4,17,
45,55,162,233,69,180,61,137,248,106,169,244,
141,76,74,237,219,65,77,117,230,73,69,6,
43,205,128,148,183,198,236,204,215,248,183,255,
254,247,169,102,238,146,45,159,47,138,157,75,
109,40,100,73,194,210,21,116,52,178,101,139,
148,174,236,178,246,81,44,122,100,123,217,148,
74,209,214,15,228,78,234,122,27,26,185,29,
229,110,78,245,165,165,79,226,136,120,253,49,
186,95,229,191,251,71,191,201,239,253,155,127,
135,215,113,249,242,247,191,201,196,196,248,169,
223,37,149,74,241,254,59,119,1,248,223,254,
143,127,69,208,105,96,100,14,78,204,134,141,
21,190,254,225,237,3,13,207,21,186,168,148,
203,148,138,69,238,220,156,65,150,101,84,85,
101,109,99,147,249,197,21,86,157,58,234,84,
153,245,181,143,145,116,187,27,86,64,69,41,
207,188,22,195,161,5,30,74,20,18,238,216,
216,21,206,54,76,38,73,144,54,84,210,134,
202,40,224,166,92,18,51,139,187,245,157,120,
97,210,251,249,183,193,243,144,164,110,81,139,
171,42,123,56,187,50,166,186,43,196,36,233,
58,227,197,254,153,90,67,217,196,111,181,25,
75,9,102,238,188,7,192,199,15,158,50,252,
181,47,157,187,113,184,20,134,66,2,12,77,
38,181,229,134,119,93,114,21,83,147,183,54,
244,54,214,1,221,145,217,212,233,135,32,108,
87,49,27,79,248,221,95,255,110,47,113,253,
207,255,209,175,19,69,17,138,50,216,176,69,
189,209,164,221,118,208,70,246,63,85,198,161,
143,178,249,144,175,206,12,95,25,137,126,15,
20,178,76,38,147,233,253,251,228,248,24,147,
227,99,251,94,219,106,181,249,244,139,71,84,
91,14,43,237,152,164,124,251,204,158,43,210,
116,18,69,121,227,223,150,109,40,216,198,222,
231,8,162,4,47,76,240,163,24,175,25,16,
235,6,177,80,8,99,65,20,119,99,245,97,
146,240,166,236,201,182,119,117,82,131,166,171,
50,183,71,211,224,68,248,90,138,7,43,221,
138,199,180,169,50,83,177,246,228,40,67,23,
100,165,127,207,61,182,71,208,202,105,212,230,
23,188,115,231,38,0,99,35,35,68,113,140,
164,156,175,173,249,66,25,10,73,234,118,53,
155,154,130,166,153,100,138,54,166,166,144,210,
148,55,146,143,76,156,26,5,185,202,189,27,
19,220,121,235,87,247,48,63,190,154,36,30,
4,238,63,126,134,249,214,119,118,181,237,199,
129,71,180,248,49,88,69,148,230,2,255,252,
31,255,195,129,27,168,43,108,157,34,51,105,
126,225,163,15,0,248,228,231,159,243,39,115,
171,152,249,179,105,124,82,163,16,57,142,56,
175,26,133,186,42,111,85,248,168,4,146,129,
154,178,246,245,180,226,68,244,140,71,152,8,
2,79,38,10,67,132,102,130,4,50,18,72,
221,239,91,66,218,250,217,253,247,208,48,49,
114,54,130,46,5,119,34,94,254,12,227,164,
151,99,217,175,97,241,90,197,194,214,21,238,
47,183,123,37,199,199,49,50,55,135,211,152,
154,130,7,100,83,26,138,44,97,168,50,111,
141,164,251,206,67,244,99,138,151,22,23,49,
205,110,95,85,189,229,96,206,255,4,237,218,
87,207,213,92,159,91,67,161,171,221,100,157,
109,116,189,131,148,174,96,104,114,207,241,246,
26,14,166,173,159,217,239,247,86,31,99,84,
174,239,155,7,16,73,140,177,246,49,223,127,
239,6,247,110,223,124,173,227,242,181,47,189,
199,227,127,255,103,56,91,134,66,120,77,174,
101,28,190,251,223,252,10,65,16,224,121,183,
175,140,196,107,194,187,111,223,5,233,62,159,
60,249,148,78,43,34,8,218,8,89,70,203,
143,17,52,55,81,77,11,53,149,61,249,65,
68,86,224,18,132,119,182,171,175,216,242,240,
99,213,32,137,21,180,212,209,97,26,39,212,
176,210,71,127,231,65,156,176,178,208,164,38,
201,248,91,172,203,219,116,252,215,42,22,143,
86,218,199,122,230,235,21,107,151,23,37,73,
221,242,229,177,124,106,128,70,2,36,89,230,
253,183,239,244,14,149,255,211,63,255,109,254,
207,127,247,31,241,206,217,28,158,11,67,97,
106,50,150,174,162,43,6,185,161,52,182,174,
30,187,188,108,144,8,218,117,70,229,26,180,
2,214,99,27,181,56,189,59,36,224,117,248,
230,189,235,220,125,205,70,98,27,5,75,97,
91,60,49,39,187,252,189,239,252,93,0,52,
77,219,83,218,122,133,51,54,22,247,110,243,
238,189,219,8,33,16,66,208,104,52,249,236,
225,19,174,125,48,194,255,251,151,143,225,20,
134,226,10,199,56,88,42,50,229,140,206,84,
49,71,221,9,187,93,238,98,251,123,209,40,
166,117,170,237,254,124,179,209,188,73,105,31,
227,52,83,25,204,183,37,146,152,246,163,31,
162,143,222,33,172,45,49,122,119,168,247,119,
154,166,33,233,38,157,39,63,194,156,124,111,
32,234,116,131,192,107,45,133,217,214,83,40,
103,116,166,74,22,119,198,50,124,233,90,158,
119,39,115,204,14,219,148,51,58,185,148,246,
70,141,4,128,158,206,211,22,38,255,245,175,
126,135,91,249,4,111,237,9,201,242,207,137,
253,238,246,108,181,231,184,243,214,173,55,246,
124,105,83,71,196,221,174,227,129,50,83,94,
225,20,107,91,66,150,101,10,133,60,191,248,
213,47,177,81,107,34,231,198,78,119,79,145,
32,93,113,61,29,27,121,75,227,198,144,189,
171,97,118,166,98,245,69,175,159,179,52,38,
139,103,183,57,71,94,155,66,227,115,190,249,
222,12,238,250,115,126,229,237,34,183,95,97,
39,24,206,232,88,186,74,84,91,184,252,30,
197,182,81,48,182,42,142,108,163,155,104,150,
207,33,123,158,72,98,226,141,39,20,244,132,
60,62,206,70,157,196,235,118,67,126,247,155,
95,101,230,249,255,223,222,217,236,68,13,69,
1,248,235,109,103,218,233,212,153,129,128,146,
152,12,152,32,46,212,96,134,24,37,98,72,
136,11,23,18,214,186,50,190,130,79,224,43,
184,117,39,15,96,76,92,185,211,200,66,87,
70,9,2,25,35,138,38,4,73,181,210,249,
233,116,122,235,162,136,252,201,20,24,226,44,
250,173,155,211,228,156,222,211,123,239,249,251,
204,153,254,34,143,159,60,167,146,59,203,208,
169,92,188,113,133,199,196,229,75,231,153,123,
246,10,47,76,113,99,106,52,241,12,29,200,
226,151,239,104,249,163,5,186,165,154,100,175,
183,109,71,172,40,12,245,89,172,56,81,115,
207,112,207,155,13,149,193,147,199,123,34,15,
126,126,101,234,214,24,66,8,102,102,151,89,
119,119,55,26,29,191,90,226,253,244,83,244,
66,231,164,200,106,0,165,254,2,78,205,199,
94,249,65,69,19,187,38,171,197,49,130,153,
86,49,55,178,35,178,186,182,25,96,174,218,
149,127,102,28,181,77,249,190,183,89,48,37,
131,38,66,168,160,40,173,139,165,194,144,96,
117,129,98,94,48,113,115,132,204,150,59,83,
207,243,254,238,70,6,162,171,167,59,147,19,
188,152,121,205,248,232,181,255,106,52,211,52,
185,59,121,157,15,139,229,205,32,88,66,103,
113,113,240,52,47,23,150,80,186,7,14,45,
67,245,27,168,65,19,47,81,103,219,232,203,
27,228,50,41,202,75,206,182,56,192,65,71,
5,28,218,166,93,69,222,206,206,115,101,100,
152,92,86,167,252,109,141,210,142,166,7,186,
174,211,211,219,139,107,21,59,231,196,108,89,
86,120,239,193,163,200,57,174,219,232,39,186,
145,97,136,223,220,200,84,144,18,41,163,242,
255,40,243,0,2,191,134,80,84,244,140,129,
161,9,210,91,130,204,59,225,178,57,96,0,
0,0,216,73,68,65,84,249,35,51,14,113,
159,13,165,164,177,178,64,90,19,8,36,61,
57,19,183,86,71,17,26,94,189,138,153,49,
144,50,36,157,214,112,235,1,191,60,137,106,
245,146,202,22,80,54,10,207,164,107,99,120,
171,148,46,156,219,214,139,41,33,225,168,120,
94,131,249,79,203,216,97,14,177,199,80,153,
184,223,185,112,108,148,80,18,20,226,117,143,
109,84,28,210,217,214,67,108,154,94,228,34,
53,221,104,155,76,191,86,65,213,141,104,147,
214,234,164,228,55,144,82,198,122,255,113,248,
15,0,207,117,168,10,147,106,61,192,212,5,
93,251,36,198,28,68,175,66,8,68,106,255,
224,187,92,251,200,216,240,32,111,102,203,132,
34,197,216,240,238,198,152,239,230,22,177,51,
253,40,138,18,233,86,75,181,148,123,80,29,
52,42,14,90,198,106,105,179,135,247,111,71,
63,10,215,117,147,213,157,144,144,144,144,176,
39,191,1,229,210,15,128,154,237,161,94,0,
0,0,0,73,69,78,68,174,66,96,130,
};
  static imagedef_t images[]= {
 {394, 261, img0data},
};
  canvas->set_images(images, 1);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_2);
  tf_9(canvas);
}
static void tf_10(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,0);
  canvas->pbreak();
  canvas->println(1,29);
  canvas->pbreak();
  canvas->println(30,34);
  canvas->verbatim_vspace();
  canvas->verbatim_line(35,36);
  canvas->verbatim_line(37,37);
  canvas->verbatim_line(38,39);
  canvas->verbatim_vspace();
  canvas->println(40,66);
  canvas->pbreak();
  canvas->println(67,95);
  canvas->pbreak();
  canvas->println(96,103);
  canvas->bullet_println(1, 104, 107);
  canvas->bullet_println(1, 108, 110);
  canvas->pbreak();
  canvas->println(111,111);
  canvas->pbreak();
  canvas->println(112,137);
  canvas->verbatim_vspace();
  canvas->verbatim_line(138,139);
  canvas->verbatim_vspace();
  canvas->println(140,146);
  canvas->bullet_println(1, 147, 157);
  canvas->bullet_println(1, 158, 167);
  canvas->bullet_println(1, 168, 173);
  canvas->println(174,189);

 canvas->end_doc(tf_10);
}
static void cf_10(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x23d,0x0,0x5f,0x0,0x54,0x0,0x23e,0x0,0x23f,0x0,0x240,0x0,0x241,0x0,0x14f,0x0,0x242,0x0,0x16,0x0,0x238,0x0,0x243,0x0,0x64,0x0,0x13,0x0,0x244,0x0,0x14d,0x0,0xc,0x0,0x245,0x0,0x9f,0x0,0x89,0x0,0x246,0x0,0x5b,0x0,0x6,0x0,0x36,0x0,0x84,0x0,0x65,0x0,0x247,0x0,0x24,0x0,0x13,0x0,0x248,0x0,0x14f,0x0,0x242,0x0,0x16,0x0,0x249,0x0,0x91,0x4,0x24a,0x50004,0xe4,0x50004,0x24b,0x4,0x24c,0x50004,0xe4,0x0,0x10e,0x0,0x65,0x0,0x189,0x0,0x24d,0x0,0x65,0x0,0x24e,0x0,0xff,0x0,0x16,0x0,0x24f,0x0,0x1d,0x0,0x65,0x0,0x14d,0x0,0x150,0x0,0x24,0x0,0x65,0x0,0x250,0x0,0x251,0x0,0x10e,0x0,0x65,0x0,0x18c,0x0,0x24d,0x0,0x65,0x0,0x24e,0x0,0xff,0x0,0x16,0x0,0xf9,0x0,0x252,0x0,0x19c,0x0,0x253,0x0,0x5b,0x0,0x5d,0x0,0x6,0x0,0x36,0x0,0x84,0x0,0x65,0x0,0x18c,0x0,0x24d,0x0,0x254,0x0,0x6f,0x0,0x255,0x0,0x59,0x0,0x16a,0x0,0x5b,0x0,0x256,0x0,0x65,0x0,0x257,0x0,0x258,0x0,0x8e,0x0,0x16,0x0,0xaf,0x0,0x100,0x0,0x259,0x0,0x84,0x0,0x65,0x0,0x189,0x0,0x25a,0x0,0x3d,0x0,0x16,0x0,0x3e,0x0,0x7,0x0,0x257,0x0,0x88,0x0,0x85,0x0,0x236,0x0,0x0,0x40000,0x25b,0x40000,0x14d,0x40000,0x25c,0x0,0x0,0x1000200,0x14d,0x1000200,0xcd,0x9,0x25d,0x0,0x9f,0x0,0x58,0x0,0x116,0x0,0x65,0x40000,0x25e,0x0,0x25f,0x0,0x5b,0x0,0x24e,0x0,0x260,0x0,0x261,0x0,0xb2,0x0,0x58,0x0,0x116,0x0,0x1a1,0x0,0x262,0x0,0x5b,0x0,0x22,0x0,0x263,0x0,0x5f,0x0,0x1ac,0x0,0xa3,0x0,0x3e,0x0,0x264,0x0,0x16,0x0,0x223,0x0,0x265,0x4,0x266,0x50004,0xe4,0x0,0x8e,0x0,0x70,0x0,0x6,0x0,0x267,0x0,0x268,0x0,0x65,0x0,0x269,0x0,0x0,0x0,0x1e9,0x0,0x15f,0x0,0x84,0x0,0x65,0x0,0x24e,0x0,0x16,0x0,0x26a,0x0,0x1ea,0x0,0x20e,0x0,0x26b,0x0,0x0,0x0,0x8e,0x0,0x89,0x0,0x59,0x0,0xc7,0x0,0x26c,0x0,0x64,0x0,0x65,0x0,0x1ae,0x0,0xab,0x0,0x0,0x0,0x26d,0x0,0x262,0x0,0x70,0x0,0x26e,0x0,0x26f,0x0,0x5f,0x0,0x264,0x0,0x25f,0x0,0x270,0x0,0x5b,0x0,0x271,0x0,0x13,0x0,0x14d,0x0,0x14e,0x0,0x171,0x0,0x65,0x0,0x75,0x0,0x76,0x0,0x64,0x0,0x3e,0x0,0x81,};
  canvas->set_elts(seq,190);
static const unsigned int bgrects[]= {
0x69,0x6b,0x4,0x74,0x74,0x4,};
  canvas->set_bgrects(bgrects, 2);
static const unsigned int underlines[]= {
0x6d,0x6e,0x2,};
  canvas->set_underlines(underlines, 1);
static const unsigned int verboxes[]= {
0x23,0x27,0x5,0x8a,0x8b,0x5,};
  canvas->set_vboxes(verboxes, 2);
  canvas->set_images(NULL,0);
static const unsigned int links[]= {
0x6d,0x6e,0x9,};
  canvas->set_links(links, 1);
  win->set_section_menu(secmen_2);
  tf_10(canvas);
}
static void tf_12(FYGfxCanvas* canvas) {
  int *TL_ROW= canvas->get_row_lines(9);
  int *TL_COL= canvas->get_col_lines(3);
canvas->reset_doc();
  canvas->println(0,1);
  canvas->pbreak();
  canvas->println(2,30);
 /// BEGIN TABLE 
 {
  int cw0, ch0,ca0;
  canvas->get_seq_metrics(31,31, &cw0, &ch0, &ca0);
  int cw1, ch1,ca1;
  canvas->get_seq_metrics(32,32, &cw1, &ch1, &ca1);
  int cw2, ch2,ca2;
  canvas->get_seq_metrics(33,35, &cw2, &ch2, &ca2);
  int cw3, ch3,ca3;
  canvas->get_seq_metrics(36,37, &cw3, &ch3, &ca3);
  int cw4, ch4,ca4;
  canvas->get_seq_metrics(38,40, &cw4, &ch4, &ca4);
  int cw5, ch5,ca5;
  canvas->get_seq_metrics(41,42, &cw5, &ch5, &ca5);
  int cw6, ch6,ca6;
  canvas->get_seq_metrics(43,45, &cw6, &ch6, &ca6);
  int cw7, ch7,ca7;
  canvas->get_seq_metrics(46,47, &cw7, &ch7, &ca7);
  int cw8, ch8,ca8;
  canvas->get_seq_metrics(48,50, &cw8, &ch8, &ca8);
  int cw9, ch9,ca9;
  canvas->get_seq_metrics(51,52, &cw9, &ch9, &ca9);
  int cw10, ch10,ca10;
  canvas->get_seq_metrics(53,55, &cw10, &ch10, &ca10);
  int cw11, ch11,ca11;
  canvas->get_seq_metrics(56,57, &cw11, &ch11, &ca11);
  int cw12, ch12,ca12;
  canvas->get_seq_metrics(58,60, &cw12, &ch12, &ca12);
  int cw13, ch13,ca13;
  canvas->get_seq_metrics(61,62, &cw13, &ch13, &ca13);
  int cw14, ch14,ca14;
  canvas->get_seq_metrics(63,65, &cw14, &ch14, &ca14);
  int cw15, ch15,ca15;
  canvas->get_seq_metrics(66,67, &cw15, &ch15, &ca15);
 int rh0=0, ra0=0;
 int rh1=0, ra1=0;
 int rh2=0, ra2=0;
 int rh3=0, ra3=0;
 int rh4=0, ra4=0;
 int rh5=0, ra5=0;
 int rh6=0, ra6=0;
 int rh7=0, ra7=0;
  rh0= max_i(rh0, ch0); ra0= max_i(ra0, ca0);
  rh0= max_i(rh0, ch1); ra0= max_i(ra0, ca1);
  rh1= max_i(rh1, ch2); ra1= max_i(ra1, ca2);
  rh1= max_i(rh1, ch3); ra1= max_i(ra1, ca3);
  rh2= max_i(rh2, ch4); ra2= max_i(ra2, ca4);
  rh2= max_i(rh2, ch5); ra2= max_i(ra2, ca5);
  rh3= max_i(rh3, ch6); ra3= max_i(ra3, ca6);
  rh3= max_i(rh3, ch7); ra3= max_i(ra3, ca7);
  rh4= max_i(rh4, ch8); ra4= max_i(ra4, ca8);
  rh4= max_i(rh4, ch9); ra4= max_i(ra4, ca9);
  rh5= max_i(rh5, ch10); ra5= max_i(ra5, ca10);
  rh5= max_i(rh5, ch11); ra5= max_i(ra5, ca11);
  rh6= max_i(rh6, ch12); ra6= max_i(ra6, ca12);
  rh6= max_i(rh6, ch13); ra6= max_i(ra6, ca13);
  rh7= max_i(rh7, ch14); ra7= max_i(ra7, ca14);
  rh7= max_i(rh7, ch15); ra7= max_i(ra7, ca15);
  int cow0=0;
  int cow1=0;
  cow0= max_i(cow0, cw0);
  cow0= max_i(cow0, cw2);
  cow0= max_i(cow0, cw4);
  cow0= max_i(cow0, cw6);
  cow0= max_i(cow0, cw8);
  cow0= max_i(cow0, cw10);
  cow0= max_i(cow0, cw12);
  cow0= max_i(cow0, cw14);
  cow1= max_i(cow1, cw1);
  cow1= max_i(cow1, cw3);
  cow1= max_i(cow1, cw5);
  cow1= max_i(cow1, cw7);
  cow1= max_i(cow1, cw9);
  cow1= max_i(cow1, cw11);
  cow1= max_i(cow1, cw13);
  cow1= max_i(cow1, cw15);
  int D=0;
  int rp0;
  int rp1;
  int rp2;
  int rp3;
  int rp4;
  int rp5;
  int rp6;
  int rp7;
  int rp8;
  int cp0;
  int cp1;
  int cp2;
  cp0= rp0= 0;
  rp1= rp0 + 2*CELL_PAD + rh0;
  rp2= rp1 + 2*CELL_PAD + rh1;
  rp3= rp2 + 2*CELL_PAD + rh2;
  rp4= rp3 + 2*CELL_PAD + rh3;
  rp5= rp4 + 2*CELL_PAD + rh4;
  rp6= rp5 + 2*CELL_PAD + rh5;
  rp7= rp6 + 2*CELL_PAD + rh6;
  rp8= rp7 + 2*CELL_PAD + rh7;
  cp1= cp0 + 2*CELL_PAD + cow0;
  cp2= cp1 + 2*CELL_PAD + cow1;
  int off_x= canvas->get_width() - cp2;
  if (off_x<0) off_x= 0; else off_x/= 2;
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 31, 31);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 32, 32);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 33, 35);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 36, 37);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 38, 40);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 41, 42);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 43, 45);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 46, 47);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 48, 50);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 51, 52);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 53, 55);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 56, 57);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp6+CELL_PAD, rh6, ra6, 58, 60);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp6+CELL_PAD, rh6, ra6, 61, 62);
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp7+CELL_PAD, rh7, ra7, 63, 65);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp7+CELL_PAD, rh7, ra7, 66, 67);
  int off_y= canvas->get_cursor_y();
  TL_ROW[0]= off_y+rp0; TL_ROW[1]= off_x+cp0; TL_ROW[2]= off_x+cp2;
  TL_ROW[3]= off_y+rp1; TL_ROW[4]= off_x+cp0; TL_ROW[5]= off_x+cp2;
  TL_ROW[6]= off_y+rp2; TL_ROW[7]= off_x+cp0; TL_ROW[8]= off_x+cp2;
  TL_ROW[9]= off_y+rp3; TL_ROW[10]= off_x+cp0; TL_ROW[11]= off_x+cp2;
  TL_ROW[12]= off_y+rp4; TL_ROW[13]= off_x+cp0; TL_ROW[14]= off_x+cp2;
  TL_ROW[15]= off_y+rp5; TL_ROW[16]= off_x+cp0; TL_ROW[17]= off_x+cp2;
  TL_ROW[18]= off_y+rp6; TL_ROW[19]= off_x+cp0; TL_ROW[20]= off_x+cp2;
  TL_ROW[21]= off_y+rp7; TL_ROW[22]= off_x+cp0; TL_ROW[23]= off_x+cp2;
  TL_ROW[24]= off_y+rp8; TL_ROW[25]= off_x+cp0; TL_ROW[26]= off_x+cp2;
  TL_COL[0]= off_y+rp0; TL_COL[1]= off_y+rp8; TL_COL[2]= off_x+cp0;
  TL_COL[3]= off_y+rp0; TL_COL[4]= off_y+rp8; TL_COL[5]= off_x+cp1;
  TL_COL[6]= off_y+rp0; TL_COL[7]= off_y+rp8; TL_COL[8]= off_x+cp2;
  canvas->xcoord(cp2+3);
  canvas->advance_y(rp8+2);
 /// END TABLE 
 }

 canvas->end_doc(tf_12);
}
static void cf_12(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x272,0xa,0x273,0x0,0xd2,0x0,0xc2,0x0,0xf3,0x0,0x70,0x0,0x249,0x0,0x1e1,0x0,0x84,0x0,0xdd,0x0,0xde,0x0,0x5f,0x0,0x189,0x0,0xe6,0x0,0x16,0x0,0x65,0x0,0xe1,0x0,0x88,0x0,0x65,0x0,0x1ef,0x0,0x70,0x0,0x65,0x0,0x274,0x0,0x5b,0x0,0x59,0x0,0x275,0x0,0x5f,0x0,0xc1,0x0,0x276,0x0,0x70,0x0,0x157,0x0,0x277,0x0,0x278,0x0,0x279,0x0,0x27a,0x0,0x27b,0x30000,0x27a,0x30000,0xd3,0x0,0x27c,0x0,0x27a,0x0,0x27b,0x1000000,0x27a,0x1000000,0xd3,0x0,0x27d,0x0,0x27a,0x0,0x27b,0x4,0x27e,0x4,0x27f,0x0,0x280,0x0,0x27a,0x0,0x27b,0x40000,0x27a,0x40000,0xd3,0x0,0x281,0x0,0x27a,0x0,0x27b,0x1,0x282,0x1,0x283,0x0,0x284,0x0,0x27a,0x0,0x27b,0x2,0x285,0x2,0x286,0x0,0x287,0x0,0x27a,0x0,0x27b,0x3,0x288,0x3,0x289,};
  canvas->set_elts(seq,68);
static const unsigned int bgrects[]= {
0x24,0x25,0x3,0x33,0x34,0x4,};
  canvas->set_bgrects(bgrects, 2);
static const unsigned int underlines[]= {
0x29,0x2a,0x0,};
  canvas->set_underlines(underlines, 1);
  canvas->set_vboxes(NULL, 0);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_2);
  tf_12(canvas);
}
static void tf_13(FYGfxCanvas* canvas) {
  int *TL_ROW= canvas->get_row_lines(22);
  int *TL_COL= canvas->get_col_lines(9);
canvas->reset_doc();
  canvas->println(0,5);
  canvas->pbreak();
  canvas->println(6,20);
  canvas->verbatim_vspace();
  canvas->verbatim_line(21,21);
  canvas->verbatim_line(22,22);
  canvas->verbatim_line(23,23);
  canvas->verbatim_vspace();
  canvas->println(24,53);
  canvas->pbreak();
  canvas->println(54,66);
  canvas->pbreak();
  canvas->println(67,67);
  canvas->pbreak();
  canvas->println(68,98);
 /// BEGIN TABLE 
 {
  int cw0, ch0,ca0;
  canvas->get_seq_metrics(99,99, &cw0, &ch0, &ca0);
  int cw1, ch1,ca1;
  canvas->get_seq_metrics(100,100, &cw1, &ch1, &ca1);
  int cw2, ch2,ca2;
  canvas->get_seq_metrics(101,101, &cw2, &ch2, &ca2);
  int cw3, ch3,ca3;
  canvas->get_seq_metrics(102,104, &cw3, &ch3, &ca3);
  int cw4, ch4,ca4;
  canvas->get_seq_metrics(105,105, &cw4, &ch4, &ca4);
  int cw5, ch5,ca5;
  canvas->get_seq_metrics(106,108, &cw5, &ch5, &ca5);
 int rh0=0, ra0=0;
 int rh1=0, ra1=0;
 int rh2=0, ra2=0;
  rh0= max_i(rh0, ch0); ra0= max_i(ra0, ca0);
  rh0= max_i(rh0, ch1); ra0= max_i(ra0, ca1);
  rh1= max_i(rh1, ch2); ra1= max_i(ra1, ca2);
  rh1= max_i(rh1, ch3); ra1= max_i(ra1, ca3);
  rh2= max_i(rh2, ch4); ra2= max_i(ra2, ca4);
  rh2= max_i(rh2, ch5); ra2= max_i(ra2, ca5);
  int cow0=0;
  int cow1=0;
  cow0= max_i(cow0, cw0);
  cow0= max_i(cow0, cw2);
  cow0= max_i(cow0, cw4);
  cow1= max_i(cow1, cw1);
  cow1= max_i(cow1, cw3);
  cow1= max_i(cow1, cw5);
  int D=0;
  int rp0;
  int rp1;
  int rp2;
  int rp3;
  int cp0;
  int cp1;
  int cp2;
  cp0= rp0= 0;
  rp1= rp0 + 2*CELL_PAD + rh0;
  rp2= rp1 + 2*CELL_PAD + rh1;
  rp3= rp2 + 2*CELL_PAD + rh2;
  cp1= cp0 + 2*CELL_PAD + cow0;
  cp2= cp1 + 2*CELL_PAD + cow1;
  int off_x= canvas->get_width() - cp2;
  if (off_x<0) off_x= 0; else off_x/= 2;
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 99, 99);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 100, 100);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 101, 101);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 102, 104);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 105, 105);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 106, 108);
  int off_y= canvas->get_cursor_y();
  TL_ROW[0]= off_y+rp0; TL_ROW[1]= off_x+cp0; TL_ROW[2]= off_x+cp2;
  TL_ROW[3]= off_y+rp1; TL_ROW[4]= off_x+cp0; TL_ROW[5]= off_x+cp2;
  TL_ROW[6]= off_y+rp2; TL_ROW[7]= off_x+cp0; TL_ROW[8]= off_x+cp2;
  TL_ROW[9]= off_y+rp3; TL_ROW[10]= off_x+cp0; TL_ROW[11]= off_x+cp2;
  TL_COL[0]= off_y+rp0; TL_COL[1]= off_y+rp3; TL_COL[2]= off_x+cp0;
  TL_COL[3]= off_y+rp0; TL_COL[4]= off_y+rp3; TL_COL[5]= off_x+cp1;
  TL_COL[6]= off_y+rp0; TL_COL[7]= off_y+rp3; TL_COL[8]= off_x+cp2;
  canvas->xcoord(cp2+3);
  canvas->advance_y(rp3+2);
 /// END TABLE 
 }
  canvas->println(109,132);
  canvas->pbreak();
  canvas->println(133,134);
  canvas->pbreak();
  canvas->println(135,148);
 /// BEGIN TABLE 
 {
  int cw0, ch0,ca0;
  canvas->get_seq_metrics(149,149, &cw0, &ch0, &ca0);
  int cw1, ch1,ca1;
  canvas->get_seq_metrics(150,150, &cw1, &ch1, &ca1);
  int cw2, ch2,ca2;
  canvas->get_seq_metrics(151,151, &cw2, &ch2, &ca2);
  int cw3, ch3,ca3;
  canvas->get_seq_metrics(152,155, &cw3, &ch3, &ca3);
  int cw4, ch4,ca4;
  canvas->get_seq_metrics(156,156, &cw4, &ch4, &ca4);
  int cw5, ch5,ca5;
  canvas->get_seq_metrics(157,160, &cw5, &ch5, &ca5);
  int cw6, ch6,ca6;
  canvas->get_seq_metrics(161,161, &cw6, &ch6, &ca6);
  int cw7, ch7,ca7;
  canvas->get_seq_metrics(162,165, &cw7, &ch7, &ca7);
  int cw8, ch8,ca8;
  canvas->get_seq_metrics(166,166, &cw8, &ch8, &ca8);
  int cw9, ch9,ca9;
  canvas->get_seq_metrics(167,171, &cw9, &ch9, &ca9);
  int cw10, ch10,ca10;
  canvas->get_seq_metrics(172,172, &cw10, &ch10, &ca10);
  int cw11, ch11,ca11;
  canvas->get_seq_metrics(173,176, &cw11, &ch11, &ca11);
  int cw12, ch12,ca12;
  canvas->get_seq_metrics(177,177, &cw12, &ch12, &ca12);
  int cw13, ch13,ca13;
  canvas->get_seq_metrics(178,181, &cw13, &ch13, &ca13);
  int cw14, ch14,ca14;
  canvas->get_seq_metrics(182,182, &cw14, &ch14, &ca14);
  int cw15, ch15,ca15;
  canvas->get_seq_metrics(183,186, &cw15, &ch15, &ca15);
  int cw16, ch16,ca16;
  canvas->get_seq_metrics(187,187, &cw16, &ch16, &ca16);
  int cw17, ch17,ca17;
  canvas->get_seq_metrics(188,191, &cw17, &ch17, &ca17);
 int rh0=0, ra0=0;
 int rh1=0, ra1=0;
 int rh2=0, ra2=0;
 int rh3=0, ra3=0;
 int rh4=0, ra4=0;
 int rh5=0, ra5=0;
 int rh6=0, ra6=0;
 int rh7=0, ra7=0;
 int rh8=0, ra8=0;
  rh0= max_i(rh0, ch0); ra0= max_i(ra0, ca0);
  rh0= max_i(rh0, ch1); ra0= max_i(ra0, ca1);
  rh1= max_i(rh1, ch2); ra1= max_i(ra1, ca2);
  rh1= max_i(rh1, ch3); ra1= max_i(ra1, ca3);
  rh2= max_i(rh2, ch4); ra2= max_i(ra2, ca4);
  rh2= max_i(rh2, ch5); ra2= max_i(ra2, ca5);
  rh3= max_i(rh3, ch6); ra3= max_i(ra3, ca6);
  rh3= max_i(rh3, ch7); ra3= max_i(ra3, ca7);
  rh4= max_i(rh4, ch8); ra4= max_i(ra4, ca8);
  rh4= max_i(rh4, ch9); ra4= max_i(ra4, ca9);
  rh5= max_i(rh5, ch10); ra5= max_i(ra5, ca10);
  rh5= max_i(rh5, ch11); ra5= max_i(ra5, ca11);
  rh6= max_i(rh6, ch12); ra6= max_i(ra6, ca12);
  rh6= max_i(rh6, ch13); ra6= max_i(ra6, ca13);
  rh7= max_i(rh7, ch14); ra7= max_i(ra7, ca14);
  rh7= max_i(rh7, ch15); ra7= max_i(ra7, ca15);
  rh8= max_i(rh8, ch16); ra8= max_i(ra8, ca16);
  rh8= max_i(rh8, ch17); ra8= max_i(ra8, ca17);
  int cow0=0;
  int cow1=0;
  cow0= max_i(cow0, cw0);
  cow0= max_i(cow0, cw2);
  cow0= max_i(cow0, cw4);
  cow0= max_i(cow0, cw6);
  cow0= max_i(cow0, cw8);
  cow0= max_i(cow0, cw10);
  cow0= max_i(cow0, cw12);
  cow0= max_i(cow0, cw14);
  cow0= max_i(cow0, cw16);
  cow1= max_i(cow1, cw1);
  cow1= max_i(cow1, cw3);
  cow1= max_i(cow1, cw5);
  cow1= max_i(cow1, cw7);
  cow1= max_i(cow1, cw9);
  cow1= max_i(cow1, cw11);
  cow1= max_i(cow1, cw13);
  cow1= max_i(cow1, cw15);
  cow1= max_i(cow1, cw17);
  int D=0;
  int rp0;
  int rp1;
  int rp2;
  int rp3;
  int rp4;
  int rp5;
  int rp6;
  int rp7;
  int rp8;
  int rp9;
  int cp0;
  int cp1;
  int cp2;
  cp0= rp0= 0;
  rp1= rp0 + 2*CELL_PAD + rh0;
  rp2= rp1 + 2*CELL_PAD + rh1;
  rp3= rp2 + 2*CELL_PAD + rh2;
  rp4= rp3 + 2*CELL_PAD + rh3;
  rp5= rp4 + 2*CELL_PAD + rh4;
  rp6= rp5 + 2*CELL_PAD + rh5;
  rp7= rp6 + 2*CELL_PAD + rh6;
  rp8= rp7 + 2*CELL_PAD + rh7;
  rp9= rp8 + 2*CELL_PAD + rh8;
  cp1= cp0 + 2*CELL_PAD + cow0;
  cp2= cp1 + 2*CELL_PAD + cow1;
  int off_x= canvas->get_width() - cp2;
  if (off_x<0) off_x= 0; else off_x/= 2;
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 149, 149);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 150, 150);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 151, 151);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 152, 155);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 156, 156);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 157, 160);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 161, 161);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 162, 165);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 166, 166);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 167, 171);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 172, 172);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 173, 176);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp6+CELL_PAD, rh6, ra6, 177, 177);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp6+CELL_PAD, rh6, ra6, 178, 181);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp7+CELL_PAD, rh7, ra7, 182, 182);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp7+CELL_PAD, rh7, ra7, 183, 186);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp8+CELL_PAD, rh8, ra8, 187, 187);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp8+CELL_PAD, rh8, ra8, 188, 191);
  int off_y= canvas->get_cursor_y();
  TL_ROW[12]= off_y+rp0; TL_ROW[13]= off_x+cp0; TL_ROW[14]= off_x+cp2;
  TL_ROW[15]= off_y+rp1; TL_ROW[16]= off_x+cp0; TL_ROW[17]= off_x+cp2;
  TL_ROW[18]= off_y+rp2; TL_ROW[19]= off_x+cp0; TL_ROW[20]= off_x+cp2;
  TL_ROW[21]= off_y+rp3; TL_ROW[22]= off_x+cp0; TL_ROW[23]= off_x+cp2;
  TL_ROW[24]= off_y+rp4; TL_ROW[25]= off_x+cp0; TL_ROW[26]= off_x+cp2;
  TL_ROW[27]= off_y+rp5; TL_ROW[28]= off_x+cp0; TL_ROW[29]= off_x+cp2;
  TL_ROW[30]= off_y+rp6; TL_ROW[31]= off_x+cp0; TL_ROW[32]= off_x+cp2;
  TL_ROW[33]= off_y+rp7; TL_ROW[34]= off_x+cp0; TL_ROW[35]= off_x+cp2;
  TL_ROW[36]= off_y+rp8; TL_ROW[37]= off_x+cp0; TL_ROW[38]= off_x+cp2;
  TL_ROW[39]= off_y+rp9; TL_ROW[40]= off_x+cp0; TL_ROW[41]= off_x+cp2;
  TL_COL[9]= off_y+rp0; TL_COL[10]= off_y+rp9; TL_COL[11]= off_x+cp0;
  TL_COL[12]= off_y+rp0; TL_COL[13]= off_y+rp9; TL_COL[14]= off_x+cp1;
  TL_COL[15]= off_y+rp0; TL_COL[16]= off_y+rp9; TL_COL[17]= off_x+cp2;
  canvas->xcoord(cp2+3);
  canvas->advance_y(rp9+2);
 /// END TABLE 
 }
  canvas->pbreak();
  canvas->println(192,192);
  canvas->pbreak();
  canvas->println(193,217);
 /// BEGIN TABLE 
 {
  int cw0, ch0,ca0;
  canvas->get_seq_metrics(218,218, &cw0, &ch0, &ca0);
  int cw1, ch1,ca1;
  canvas->get_seq_metrics(219,219, &cw1, &ch1, &ca1);
  int cw2, ch2,ca2;
  canvas->get_seq_metrics(220,220, &cw2, &ch2, &ca2);
  int cw3, ch3,ca3;
  canvas->get_seq_metrics(221,222, &cw3, &ch3, &ca3);
  int cw4, ch4,ca4;
  canvas->get_seq_metrics(223,223, &cw4, &ch4, &ca4);
  int cw5, ch5,ca5;
  canvas->get_seq_metrics(224,225, &cw5, &ch5, &ca5);
  int cw6, ch6,ca6;
  canvas->get_seq_metrics(226,226, &cw6, &ch6, &ca6);
  int cw7, ch7,ca7;
  canvas->get_seq_metrics(227,228, &cw7, &ch7, &ca7);
  int cw8, ch8,ca8;
  canvas->get_seq_metrics(229,229, &cw8, &ch8, &ca8);
  int cw9, ch9,ca9;
  canvas->get_seq_metrics(230,231, &cw9, &ch9, &ca9);
  int cw10, ch10,ca10;
  canvas->get_seq_metrics(232,232, &cw10, &ch10, &ca10);
  int cw11, ch11,ca11;
  canvas->get_seq_metrics(233,234, &cw11, &ch11, &ca11);
  int cw12, ch12,ca12;
  canvas->get_seq_metrics(235,235, &cw12, &ch12, &ca12);
  int cw13, ch13,ca13;
  canvas->get_seq_metrics(236,237, &cw13, &ch13, &ca13);
 int rh0=0, ra0=0;
 int rh1=0, ra1=0;
 int rh2=0, ra2=0;
 int rh3=0, ra3=0;
 int rh4=0, ra4=0;
 int rh5=0, ra5=0;
 int rh6=0, ra6=0;
  rh0= max_i(rh0, ch0); ra0= max_i(ra0, ca0);
  rh0= max_i(rh0, ch1); ra0= max_i(ra0, ca1);
  rh1= max_i(rh1, ch2); ra1= max_i(ra1, ca2);
  rh1= max_i(rh1, ch3); ra1= max_i(ra1, ca3);
  rh2= max_i(rh2, ch4); ra2= max_i(ra2, ca4);
  rh2= max_i(rh2, ch5); ra2= max_i(ra2, ca5);
  rh3= max_i(rh3, ch6); ra3= max_i(ra3, ca6);
  rh3= max_i(rh3, ch7); ra3= max_i(ra3, ca7);
  rh4= max_i(rh4, ch8); ra4= max_i(ra4, ca8);
  rh4= max_i(rh4, ch9); ra4= max_i(ra4, ca9);
  rh5= max_i(rh5, ch10); ra5= max_i(ra5, ca10);
  rh5= max_i(rh5, ch11); ra5= max_i(ra5, ca11);
  rh6= max_i(rh6, ch12); ra6= max_i(ra6, ca12);
  rh6= max_i(rh6, ch13); ra6= max_i(ra6, ca13);
  int cow0=0;
  int cow1=0;
  cow0= max_i(cow0, cw0);
  cow0= max_i(cow0, cw2);
  cow0= max_i(cow0, cw4);
  cow0= max_i(cow0, cw6);
  cow0= max_i(cow0, cw8);
  cow0= max_i(cow0, cw10);
  cow0= max_i(cow0, cw12);
  cow1= max_i(cow1, cw1);
  cow1= max_i(cow1, cw3);
  cow1= max_i(cow1, cw5);
  cow1= max_i(cow1, cw7);
  cow1= max_i(cow1, cw9);
  cow1= max_i(cow1, cw11);
  cow1= max_i(cow1, cw13);
  int D=0;
  int rp0;
  int rp1;
  int rp2;
  int rp3;
  int rp4;
  int rp5;
  int rp6;
  int rp7;
  int cp0;
  int cp1;
  int cp2;
  cp0= rp0= 0;
  rp1= rp0 + 2*CELL_PAD + rh0;
  rp2= rp1 + 2*CELL_PAD + rh1;
  rp3= rp2 + 2*CELL_PAD + rh2;
  rp4= rp3 + 2*CELL_PAD + rh3;
  rp5= rp4 + 2*CELL_PAD + rh4;
  rp6= rp5 + 2*CELL_PAD + rh5;
  rp7= rp6 + 2*CELL_PAD + rh6;
  cp1= cp0 + 2*CELL_PAD + cow0;
  cp2= cp1 + 2*CELL_PAD + cow1;
  int off_x= canvas->get_width() - cp2;
  if (off_x<0) off_x= 0; else off_x/= 2;
  canvas->cell_center(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 218, 218);
  canvas->cell_center(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp0+CELL_PAD, rh0, ra0, 219, 219);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 220, 220);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp1+CELL_PAD, rh1, ra1, 221, 222);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 223, 223);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp2+CELL_PAD, rh2, ra2, 224, 225);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 226, 226);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp3+CELL_PAD, rh3, ra3, 227, 228);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 229, 229);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp4+CELL_PAD, rh4, ra4, 230, 231);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 232, 232);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp5+CELL_PAD, rh5, ra5, 233, 234);
  canvas->cell_left(off_x+cp0+CELL_PAD, off_x+cp1-CELL_PAD, rp6+CELL_PAD, rh6, ra6, 235, 235);
  canvas->cell_left(off_x+cp1+CELL_PAD, off_x+cp2-CELL_PAD, rp6+CELL_PAD, rh6, ra6, 236, 237);
  int off_y= canvas->get_cursor_y();
  TL_ROW[42]= off_y+rp0; TL_ROW[43]= off_x+cp0; TL_ROW[44]= off_x+cp2;
  TL_ROW[45]= off_y+rp1; TL_ROW[46]= off_x+cp0; TL_ROW[47]= off_x+cp2;
  TL_ROW[48]= off_y+rp2; TL_ROW[49]= off_x+cp0; TL_ROW[50]= off_x+cp2;
  TL_ROW[51]= off_y+rp3; TL_ROW[52]= off_x+cp0; TL_ROW[53]= off_x+cp2;
  TL_ROW[54]= off_y+rp4; TL_ROW[55]= off_x+cp0; TL_ROW[56]= off_x+cp2;
  TL_ROW[57]= off_y+rp5; TL_ROW[58]= off_x+cp0; TL_ROW[59]= off_x+cp2;
  TL_ROW[60]= off_y+rp6; TL_ROW[61]= off_x+cp0; TL_ROW[62]= off_x+cp2;
  TL_ROW[63]= off_y+rp7; TL_ROW[64]= off_x+cp0; TL_ROW[65]= off_x+cp2;
  TL_COL[18]= off_y+rp0; TL_COL[19]= off_y+rp7; TL_COL[20]= off_x+cp0;
  TL_COL[21]= off_y+rp0; TL_COL[22]= off_y+rp7; TL_COL[23]= off_x+cp1;
  TL_COL[24]= off_y+rp0; TL_COL[25]= off_y+rp7; TL_COL[26]= off_x+cp2;
  canvas->xcoord(cp2+3);
  canvas->advance_y(rp7+2);
 /// END TABLE 
 }

 canvas->end_doc(tf_13);
}
static void cf_13(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xb,0x28a,0xb,0x28b,0xb,0x28c,0xb,0x28d,0xb,0x4c,0xb,0x28e,0x0,0x5f,0x0,0x28f,0x0,0x290,0x0,0x16,0x0,0x223,0x0,0x291,0x0,0x5f,0x0,0xc1,0x40000,0x292,0x40000,0x293,0x0,0x294,0x0,0x12,0x0,0x1b1,0x0,0x65,0x0,0x295,0x50004,0x296,0x50004,0x297,0x50004,0x298,0x0,0x5f,0x40000,0x54,0x0,0x299,0x0,0x94,0x0,0x65,0x0,0x7c,0x0,0x29a,0x40000,0x29b,0x0,0x16,0x0,0x65,0x0,0x264,0x0,0x24e,0x0,0x24,0x0,0x65,0x0,0x14d,0x0,0xb2,0x0,0x29c,0x0,0x5b,0x0,0x29d,0x0,0x6e,0x0,0xb2,0x0,0x29e,0x40000,0x29f,0x0,0xb6,0x0,0x57,0x0,0x65,0x0,0x189,0x0,0x14b,0x0,0x16,0x0,0x2a0,0x0,0x2a1,0x0,0x2a2,0x0,0xda,0x0,0x88,0x0,0xd6,0x0,0x58,0x0,0x59,0x0,0x2a3,0x0,0x9c,0x0,0x65,0x0,0x2a4,0x0,0xc7,0x0,0x2a5,0x9,0x2a6,0x0,0x5f,0x0,0x23e,0x0,0x21e,0x0,0x107,0x0,0xc7,0x0,0x2a7,0x0,0x64,0x0,0x2a8,0x0,0xda,0x0,0x88,0x0,0x2a9,0x0,0x56,0x0,0x2a7,0x0,0x58,0x0,0x59,0x0,0x2a3,0x0,0x9c,0x0,0x65,0x40000,0x2aa,0x0,0x2ab,0x0,0x237,0x0,0x11a,0x0,0x107,0x0,0x65,0x0,0x2ac,0x0,0x24e,0x0,0x16,0x0,0x2ad,0x0,0x53,0x0,0x1a1,0x0,0x2ae,0x1,0x2af,0x1,0x2b0,0x0,0x2a2,0x0,0x2b1,0x0,0x2a2,0x0,0x2ac,0x0,0x2a2,0x0,0x2b2,0x0,0x2a2,0x0,0x2ac,0x0,0x2b2,0x0,0x2a2,0x0,0x16,0x0,0x136,0x0,0xa3,0x0,0x45,0x0,0xc6,0x0,0x100,0x0,0x91,0x0,0x2b3,0x0,0x2b4,0x0,0x88,0x0,0x2b5,0x0,0x1e9,0x0,0x2b6,0x0,0x13d,0x0,0x2b7,0x0,0x2b8,0x0,0x70,0x0,0x1aa,0x0,0x9c,0x0,0x65,0x0,0x1af,0x0,0x2b9,0x9,0x2ba,0x9,0x2bb,0x0,0x3d,0x0,0x16,0x0,0x13,0x0,0x18,0x0,0x24,0x0,0x2a2,0x0,0xda,0x0,0xb2,0x0,0x58,0x0,0x2a3,0x0,0x9c,0x0,0x65,0x40000,0x2bc,0x0,0x2bd,0x1,0x2af,0x1,0x2b0,0x0,0x2be,0x0,0x2b1,0x0,0x36,0x0,0x2a2,0x0,0x2be,0x0,0x229,0x0,0x2bf,0x0,0x36,0x0,0x2a2,0x0,0x2be,0x0,0x2c0,0x0,0x2b2,0x0,0x36,0x0,0x2a2,0x0,0x2be,0x0,0x2c1,0x0,0x2b2,0x0,0x13d,0x0,0x229,0x0,0x2a2,0x0,0x2be,0x0,0x2c2,0x0,0x2c3,0x0,0x150,0x0,0x2a2,0x0,0x2be,0x0,0x2c4,0x0,0x2c5,0x0,0x150,0x0,0x2a2,0x0,0x2be,0x0,0x2c,0x0,0x2c6,0x0,0x150,0x0,0x2a2,0x0,0x2be,0x0,0x2b,0x0,0x2c7,0x0,0x150,0x0,0x2a2,0x0,0x2be,0x9,0x2c8,0x0,0x2c9,0x0,0x88,0x0,0xd4,0x0,0xd6,0x0,0x70,0x0,0x2ca,0x0,0x5b,0x0,0x2cb,0x0,0x88,0x0,0x2cc,0x0,0x5f,0x0,0x2b6,0x0,0x2cd,0x0,0xb2,0x0,0x58,0x0,0x2ce,0x0,0xbe,0x0,0x2cf,0x0,0x16,0x0,0x3e,0x0,0x2d0,0x0,0x24,0x0,0x65,0x0,0x2d1,0x0,0x2d2,0x1,0x2af,0x1,0x2b0,0x0,0x257,0x0,0x2d3,0x0,0xd5,0x0,0x2d4,0x0,0x2d5,0x0,0xd4,0x0,0x217,0x0,0x2d6,0x0,0xd4,0x0,0x6a,0x0,0x11f,0x0,0xd4,0x0,0xd0,0x0,0x2d7,0x0,0xd4,0x0,0x2b4,0x0,0x2d8,0x0,0xd4,};
  canvas->set_elts(seq,238);
static const unsigned int bgrects[]= {
0xe,0xf,0x4,0x19,0x19,0x4,0x1f,0x1f,0x4,0x2e,0x2e,0x4,0x56,0x56,0x4,0x93,0x93,0x4,};
  canvas->set_bgrects(bgrects, 6);
  canvas->set_underlines(NULL, 0);
static const unsigned int verboxes[]= {
0x15,0x17,0x5,};
  canvas->set_vboxes(verboxes, 1);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_13);
  tf_13(canvas);
}
static void tf_17(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,0);
  canvas->pbreak();
  canvas->println(1,7);

 canvas->end_doc(tf_17);
}
static void cf_17(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xb,0x2d9,0x0,0x15,0x0,0x16,0x0,0x2da,0x0,0x2db,0x0,0x2dc,0x0,0xa3,0x0,0x2dd,};
  canvas->set_elts(seq,8);
  canvas->set_bgrects(NULL, 0);
  canvas->set_underlines(NULL, 0);
  canvas->set_vboxes(NULL, 0);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_17);
  tf_17(canvas);
}
static void tf_18(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,0);
  canvas->pbreak();
  canvas->println(1,4);
  canvas->pbreak();
  canvas->println(5,7);
  canvas->pbreak();
  canvas->println(8,28);
  canvas->bullet_println(1, 29, 47);
  canvas->bullet_println(1, 48, 76);
  canvas->println(77,192);
  canvas->pbreak();
  canvas->println(193,232);

 canvas->end_doc(tf_18);
}
static void cf_18(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x2de,0x0,0x2df,0x0,0x2e0,0x0,0x2e1,0x0,0x2dc,0x0,0x1e9,0x0,0x2e2,0x0,0x2e3,0x0,0x2e4,0x0,0x88,0x0,0x116,0x0,0x84,0x0,0x51,0x0,0x88,0x0,0x2e5,0x0,0x2e6,0x0,0x64,0x0,0xc,0x0,0x112,0x0,0x2e7,0x0,0x70,0x0,0x2e8,0x0,0x2e9,0x0,0x11a,0x0,0x65,0x0,0xc1,0x0,0x2ea,0x0,0x70,0x0,0x2eb,0x0,0x0,0x0,0x2ec,0x0,0x24,0x0,0x51,0x0,0x6a,0x0,0x15c,0x0,0x2ed,0x0,0x65,0x0,0x2ee,0x0,0x95,0x0,0x2ef,0x0,0x9a,0x0,0x18,0x0,0x24,0x0,0x2ea,0x0,0x88,0x0,0x65,0x0,0xc1,0x0,0x2f0,0x0,0x0,0x0,0x2ec,0x0,0x84,0x0,0x2e5,0x0,0x2d1,0x0,0x15c,0x0,0x2f1,0x0,0x65,0x0,0x2ee,0x0,0x95,0x0,0x2ef,0x0,0x9a,0x0,0x18,0x0,0x24,0x0,0x2ea,0x0,0x88,0x0,0x65,0x0,0xc1,0x0,0x2f2,0x0,0x84,0x0,0x65,0x0,0x2f3,0x0,0x2f4,0x0,0x2b6,0x0,0x2f5,0x0,0x2e9,0x0,0x64,0x0,0x65,0x0,0x2f6,0x0,0x2f7,0x0,0x2f8,0x0,0x2f9,0x0,0x2fa,0x0,0x2fb,0x0,0x2fc,0x0,0x2fd,0x0,0x2fe,0x0,0x2ff,0x0,0x300,0x0,0x301,0x0,0x302,0x0,0x2ff,0x0,0x303,0x0,0x304,0x0,0x305,0x0,0x306,0x0,0x307,0x0,0x308,0x0,0x309,0x0,0x30a,0x0,0x30b,0x0,0x30c,0x0,0x2fc,0x0,0x306,0x0,0x30d,0x0,0x30e,0x0,0x30f,0x0,0x2ff,0x0,0x310,0x0,0x311,0x0,0x14f,0x0,0x312,0x0,0x313,0x0,0x314,0x0,0x315,0x0,0x316,0x0,0x317,0x0,0x318,0x0,0x319,0x0,0x2fc,0x0,0x2fd,0x0,0x31a,0x0,0x305,0x0,0x300,0x0,0x31b,0x0,0x31c,0x0,0x311,0x0,0x303,0x0,0x31d,0x0,0x31e,0x0,0x31f,0x0,0x320,0x0,0x321,0x0,0x305,0x0,0x322,0x0,0x323,0x0,0x324,0x0,0x309,0x0,0x30a,0x0,0x30b,0x0,0x30c,0x0,0x325,0x0,0x30e,0x0,0x326,0x0,0x327,0x0,0x305,0x0,0x328,0x0,0x329,0x0,0x30e,0x0,0x32a,0x0,0x32b,0x0,0x305,0x0,0x32c,0x0,0x305,0x0,0x32d,0x0,0x32e,0x0,0x32f,0x0,0x330,0x0,0x2ff,0x0,0x331,0x0,0x303,0x0,0x332,0x0,0x30e,0x0,0x333,0x0,0x334,0x0,0x316,0x0,0x335,0x0,0x336,0x0,0x333,0x0,0x305,0x0,0x337,0x0,0x338,0x0,0x339,0x0,0x305,0x0,0x33a,0x0,0x33b,0x0,0x316,0x0,0x303,0x0,0x33c,0x0,0x33d,0x0,0x30e,0x0,0x2fc,0x0,0x33e,0x0,0x30e,0x0,0x2f7,0x0,0x33f,0x0,0x340,0x0,0x341,0x0,0x342,0x0,0x30e,0x0,0x2fc,0x0,0x343,0x0,0x30e,0x0,0x344,0x0,0x345,0x0,0x5f,0x0,0x9d,0x0,0x346,0x0,0x16,0x0,0x347,0x0,0x348,0x0,0x65,0x0,0x349,0x0,0x34a,0x0,0x21,0x0,0x11b,0x0,0xaf,0x0,0x34b,0x0,0x34c,0x0,0xb7,0x0,0x70,0x0,0x34d,0x0,0x34e,0x0,0xc,0x0,0x34f,0x0,0x21,0x0,0x1de,0x0,0x1b,0x0,0x350,0x0,0x59,0x0,0x351,0x0,0x168,0x0,0xb2,0x0,0x5d,0x0,0x65,0x0,0x352,0x0,0x51,0x0,0xa3,0x0,0x353,0x0,0x354,0x0,0x7e,0x0,0x355,0x0,0x19a,0x0,0x13,0x0,0x356,};
  canvas->set_elts(seq,233);
  canvas->set_bgrects(NULL, 0);
  canvas->set_underlines(NULL, 0);
  canvas->set_vboxes(NULL, 0);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_17);
  tf_18(canvas);
}
static void tf_19(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->println(0,2);
  canvas->pbreak();
  canvas->bullet_println(1, 3, 5);
  canvas->bullet_println(1, 6, 15);
  canvas->bullet_println(1, 16, 21);
  canvas->bullet_println(1, 22, 24);
  canvas->bullet_println(1, 25, 26);
  canvas->bullet_println(1, 27, 35);
  canvas->bullet_println(1, 36, 40);
  canvas->bullet_println(1, 41, 44);
  canvas->bullet_println(1, 45, 51);
  canvas->bullet_println(1, 52, 57);
  canvas->bullet_println(1, 58, 65);

 canvas->end_doc(tf_19);
}
static void cf_19(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0xa,0x357,0xa,0x358,0xa,0x359,0x0,0x0,0x0,0x35a,0x0,0x35b,0x0,0x0,0x0,0x35c,0x0,0x88,0x0,0x35d,0x0,0x35e,0x0,0xa3,0x0,0x35f,0x0,0x360,0x0,0x13,0x0,0xbb,0x0,0x0,0x0,0x361,0x0,0x88,0x0,0x362,0x0,0x363,0x0,0x364,0x0,0x0,0x0,0x2c3,0x0,0x365,0x0,0x0,0x0,0x366,0x0,0x0,0x0,0x367,0x0,0x65,0x0,0x368,0x0,0x1a9,0x0,0x23b,0x0,0x88,0x0,0x2b6,0x0,0x369,0x0,0x0,0x0,0xd3,0x0,0x36a,0x0,0x36b,0x0,0x36c,0x0,0x0,0x0,0x36d,0x0,0xf7,0x0,0x36e,0x0,0x0,0x0,0x36f,0x0,0x370,0x0,0x24,0x0,0x2a7,0x0,0x88,0x0,0x371,0x0,0x0,0x0,0x36f,0x0,0x370,0x0,0x24,0x0,0x18,0x0,0x372,0x0,0x0,0x0,0x373,0x0,0x35b,0x0,0x374,0x0,0x375,0x0,0x376,0x0,0x377,0x0,0x102,};
  canvas->set_elts(seq,66);
  canvas->set_bgrects(NULL, 0);
  canvas->set_underlines(NULL, 0);
  canvas->set_vboxes(NULL, 0);
  canvas->set_images(NULL,0);
  canvas->set_links(NULL,0);
  win->set_section_menu(secmen_17);
  tf_19(canvas);
}
static void tf_toc2(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->bullet_println(1, 0, 2);
  canvas->bullet_println(1, 3, 6);
  canvas->bullet_println(1, 7, 9);
  canvas->bullet_println(2, 10, 13);
  canvas->bullet_println(2, 14, 15);
  canvas->bullet_println(2, 16, 17);
  canvas->bullet_println(2, 18, 20);
  canvas->bullet_println(2, 21, 22);
  canvas->bullet_println(2, 23, 24);
  canvas->bullet_println(2, 25, 26);
  canvas->bullet_println(2, 27, 29);
  canvas->bullet_println(1, 30, 36);
  canvas->bullet_println(1, 37, 38);
  canvas->bullet_println(2, 39, 40);
  canvas->bullet_println(2, 41, 44);

 canvas->end_doc(tf_toc2);
}
static void cf_toc2(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0x0,0x0,0x1000200,0x378,0x1000200,0x2c3,0x0,0x0,0x1000200,0x379,0x1000200,0x37a,0x1000200,0x37b,0x0,0x0,0x1000200,0x37c,0x1000200,0x37d,0x0,0x0,0x1000200,0x37e,0x1000200,0x88,0x1000200,0x37f,0x0,0x0,0x1000200,0x16c,0x0,0x0,0x1000200,0xcb,0x0,0x0,0x1000200,0xce,0x1000200,0xd3,0x0,0x0,0x1000200,0xcc,0x0,0x0,0x1000200,0xcd,0x0,0x0,0x1000200,0xca,0x0,0x0,0x1000200,0xd2,0x1000200,0x380,0x0,0x0,0x1000200,0x28f,0x1000200,0x381,0x1000200,0x5b,0x1000200,0x65,0x1000200,0x37a,0x1000200,0x382,0x0,0x0,0x1000200,0x383,0x0,0x0,0x1000200,0x384,0x0,0x0,0x1000200,0x385,0x1000200,0x5b,0x1000200,0x386,};
  canvas->set_elts(seq,45);
  canvas->set_bgrects(NULL, 0);
static const unsigned int underlines[]= {
0x1,0x2,0x2,0x4,0x6,0x2,0x8,0x9,0x2,0xb,0xd,0x2,0xf,0xf,0x2,0x11,0x11,0x2,0x13,0x14,0x2,0x16,0x16,0x2,0x18,0x18,0x2,0x1a,0x1a,0x2,0x1c,0x1d,0x2,0x1f,0x24,0x2,0x26,0x26,0x2,0x28,0x28,0x2,0x2a,0x2c,0x2,};
  canvas->set_underlines(underlines, 15);
  canvas->set_vboxes(NULL, 0);
  canvas->set_images(NULL,0);
static const unsigned int links[]= {
0x1,0x2,0x0,0x4,0x6,0x1,0x8,0x9,0x2,0xb,0xd,0x4,0xf,0xf,0x5,0x11,0x11,0x6,0x13,0x14,0x7,0x16,0x16,0x8,0x18,0x18,0x9,0x1a,0x1a,0xa,0x1c,0x1d,0xc,0x1f,0x24,0xd,0x26,0x26,0x11,0x28,0x28,0x12,0x2a,0x2c,0x13,};
  canvas->set_links(links, 15);
  static const sipair_t empty[]= { { NULL, 0 } };
  win->set_section_menu(empty);
  tf_toc2(canvas);
}
static void tf_toc0(FYGfxCanvas* canvas) {
 canvas->get_row_lines(0); 
 canvas->get_col_lines(0); 
canvas->reset_doc();
  canvas->bullet_println(1, 0, 2);
  canvas->bullet_println(1, 3, 6);
  canvas->bullet_println(1, 7, 9);
  canvas->bullet_println(3, 10, 12);
  canvas->bullet_println(2, 13, 16);
  canvas->bullet_println(2, 17, 18);
  canvas->bullet_println(2, 19, 20);
  canvas->bullet_println(2, 21, 23);
  canvas->bullet_println(2, 24, 25);
  canvas->bullet_println(2, 26, 27);
  canvas->bullet_println(2, 28, 29);
  canvas->bullet_println(3, 30, 31);
  canvas->bullet_println(2, 32, 34);
  canvas->bullet_println(1, 35, 41);
  canvas->bullet_println(3, 42, 43);
  canvas->bullet_println(3, 44, 46);
  canvas->bullet_println(3, 47, 48);
  canvas->bullet_println(1, 49, 50);
  canvas->bullet_println(2, 51, 52);
  canvas->bullet_println(2, 53, 56);

 canvas->end_doc(tf_toc0);
}
static void cf_toc0(FYHelpWin *win) {
  FYGfxCanvas *canvas= win->canvas;
static const unsigned int seq[]={
0x0,0x0,0x1000200,0x378,0x1000200,0x2c3,0x0,0x0,0x1000200,0x379,0x1000200,0x37a,0x1000200,0x37b,0x0,0x0,0x1000200,0x37c,0x1000200,0x37d,0x0,0x0,0x1000200,0x387,0x1000200,0x388,0x0,0x0,0x1000200,0x37e,0x1000200,0x88,0x1000200,0x37f,0x0,0x0,0x1000200,0x16c,0x0,0x0,0x1000200,0xcb,0x0,0x0,0x1000200,0xce,0x1000200,0xd3,0x0,0x0,0x1000200,0xcc,0x0,0x0,0x1000200,0xcd,0x0,0x0,0x1000200,0xca,0x0,0x0,0x1000200,0x389,0x0,0x0,0x1000200,0xd2,0x1000200,0x380,0x0,0x0,0x1000200,0x28f,0x1000200,0x381,0x1000200,0x5b,0x1000200,0x65,0x1000200,0x37a,0x1000200,0x382,0x0,0x0,0x1000200,0x38a,0x0,0x0,0x1000200,0xd8,0x1000200,0x38b,0x0,0x0,0x1000200,0x38c,0x0,0x0,0x1000200,0x383,0x0,0x0,0x1000200,0x384,0x0,0x0,0x1000200,0x385,0x1000200,0x5b,0x1000200,0x386,};
  canvas->set_elts(seq,57);
  canvas->set_bgrects(NULL, 0);
static const unsigned int underlines[]= {
0x1,0x2,0x2,0x4,0x6,0x2,0x8,0x9,0x2,0xb,0xc,0x2,0xe,0x10,0x2,0x12,0x12,0x2,0x14,0x14,0x2,0x16,0x17,0x2,0x19,0x19,0x2,0x1b,0x1b,0x2,0x1d,0x1d,0x2,0x1f,0x1f,0x2,0x21,0x22,0x2,0x24,0x29,0x2,0x2b,0x2b,0x2,0x2d,0x2e,0x2,0x30,0x30,0x2,0x32,0x32,0x2,0x34,0x34,0x2,0x36,0x38,0x2,};
  canvas->set_underlines(underlines, 20);
  canvas->set_vboxes(NULL, 0);
  canvas->set_images(NULL,0);
static const unsigned int links[]= {
0x1,0x2,0x0,0x4,0x6,0x1,0x8,0x9,0x2,0xb,0xc,0x3,0xe,0x10,0x4,0x12,0x12,0x5,0x14,0x14,0x6,0x16,0x17,0x7,0x19,0x19,0x8,0x1b,0x1b,0x9,0x1d,0x1d,0xa,0x1f,0x1f,0xb,0x21,0x22,0xc,0x24,0x29,0xd,0x2b,0x2b,0xe,0x2d,0x2e,0xf,0x30,0x30,0x10,0x32,0x32,0x11,0x34,0x34,0x12,0x36,0x38,0x13,};
  canvas->set_links(links, 20);
  static const sipair_t empty[]= { { NULL, 0 } };
  win->set_section_menu(empty);
  tf_toc0(canvas);
}
static void init_words(FYGfxCanvas *canvas) {
static const char *words[]={
"",
"Demo",
"Chapter",
"Let\047s",
"try",
"making",
"some",
"example",
"code\072",
"\040\040\075\075\040section\040",
"header",
"\077",
"or",
"code\077",
"Another",
"try\072",
"now",
"we",
"do",
"a",
"list\056",
"This",
"is",
"good",
"list",
"We",
"will",
"it",
"properly",
"by",
"nested",
"And",
"unnested",
"I",
"make",
"test",
"of",
"\133paren",
"vs\056",
"\135stuff",
"\056",
"Yeah\056",
"header0",
"header1",
"header2",
"M\1370\1370",
"M\1370\1371",
"M\1370\1372",
"M\1370\1373",
"M\1371\1370",
"M\1371\1373",
"M\1372\1370",
"really",
"long",
"text",
"which",
"M\1372\1373",
"M\1373\1370",
"M\1373\1371",
"M\1373\1372",
"M\1373\1373",
"Here",
"an",
"image\072",
"don\047t",
"know",
"how",
"work\056\056",
"Variaous",
"verbatim",
"environments\072",
"\040\040\040This\040is\040code\056",
"\040\040\040Does\040it\040look\040OK\077",
"\040\040\040This\040is\040output\056",
"\040\040\040This\040is\040shell\056",
"Fox",
"Help",
"Compiler",
"program",
"generates",
"C\053\053",
"source",
"files",
"from",
"help",
"documents\056",
"These",
"then",
"can",
"be",
"linked",
"to",
"your",
"provide",
"functionality\056",
"The",
"called",
"foxhelpc",
"It",
"invoked",
"with",
"the",
"document",
"compiled\072",
"foxhelpc\040\074input\137file1\076\040\074input\137file2\076\040\056\056\040\074input\137fileN\076",
"generated",
"code",
"printed",
"standard",
"output\056",
"If",
"there",
"are",
"any",
"errors",
"during",
"compilation\054",
"compiler",
"exits",
"error",
"message\056",
"Usually",
"line",
"number",
"given",
"incorrect",
"but",
"near",
"actual",
"error\056",
"because",
"still",
"in",
"its",
"early",
"stages",
"and",
"can\047t",
"track",
"numbers",
"very",
"well\056",
"There",
"command",
"options",
"as",
"well\072",
"\055h",
"displays",
"copyright",
"information",
"short",
"help\056",
"\055H",
"this",
"document\054",
"using",
"FOX",
"system\056",
"You",
"need",
"X",
"display",
"for",
"this\054",
"naturally\056",
"\055f",
"\074file\076",
"creates",
"header",
"file",
"name\056",
"option",
"isn\047t",
"given\054",
"no",
"created\056",
"When",
"you",
"give",
"multiple",
"foxhelpc",
"\054",
"they",
"concatenated",
"processed",
"single",
"document\056",
"Document",
"Syntax",
"Each",
"simple",
"mark\055up\056",
"following",
"formatting",
"features",
"available\072",
"Chapters\054",
"sections",
"two",
"levels",
"sub\055sections",
"Links",
"Lists",
"Tables",
"Images",
"Verbatim",
"\050for",
"shell",
"sequences\051",
"Inline",
"Text",
"background",
"foreground",
"colors",
"Underlines",
"Font",
"families",
"sizes",
"Markup",
"done",
"\133\135",
"pairs\056",
"general",
"format",
"directive",
"is\072",
"\133directive\040args",
"\135",
"Here\054",
"word",
"consisting",
"ASCII",
"letters\056",
"args",
"usually",
"want",
"arguments",
"controlling",
"directive\056",
"Quote",
"character",
"table",
"directives",
"latter\056",
"kinds",
"directives\072",
"inline",
"block\056",
"simply",
"their",
"different",
"way\056",
"Block",
"define",
"part",
"such",
"list\054table",
"etc\056",
"oriented\056",
"Therefore\054",
"block",
"recognized",
"only",
"when",
"at",
"beginning",
"line\056",
"Escape",
"Sequences",
"In",
"order",
"special",
"characters",
"without",
"starting",
"ending",
"unintentionally\054",
"use",
"escape",
"sequences\056",
"sure",
"that",
"have",
"balanced",
"parenthesis",
"characters\056",
"Code",
"Rendering",
"Mnemonic",
"\133pLs\135",
"\133",
"Left",
"square",
"\133pRs\135",
"\135",
"Right",
"\133pLr\135",
"\050",
"round",
"\133pRr\135",
"\051",
"\133spc\135",
"\040",
"space",
"Normally\054",
"spaces",
"ignored",
"outside",
"environments",
"used",
"dividing",
"into",
"words\056",
"include",
"literal",
"somewhere",
"text\054",
"sequence\056",
"Chapters",
"and",
"Sections",
"considered",
"book\056",
"top",
"level",
"chapters",
"shown",
"menu\056",
"start",
"new",
"chapter",
"section\054",
"section",
"title\056",
"A",
"title",
"starts",
"\047\075\047",
"followed",
"text\056",
"Four",
"titles",
"recognized\072",
"\075\075\075\075\040Chapter\040Title",
"\075\075\075\040Section\040Title",
"\075\075\040Sub\055section\040Title",
"\075\040Sub\055sub\055section\040Title",
"must",
"appear",
"leading",
"whitespace",
"ignored\056",
"Paragraphs",
"Anything",
"another",
"paragraph\056",
"paragraph",
"ends",
"next",
"if",
"contains",
"nothing",
"whitespace\056",
"Paragraphs",
"wrapped",
"displayed\056",
"end",
"within",
"Otherwise",
"compiler\056",
"Lists",
"Only",
"bullet",
"lists",
"supported\056",
"dash",
"dashes",
"succession",
"indicate",
"more",
"indentation",
"than",
"default\056",
"example\072",
"\055\040Top\040list\040first\040item",
"\055\055\040Inner\040list\040first\040item",
"\055\055\040Inner\040list\040second\040item",
"\055\040Top\040list\040second\040item",
"\055\040Top\040list\040third\040item",
"rendered",
"as\072",
"Top",
"first",
"item",
"Inner",
"second",
"third",
"either",
"empty",
"found",
"on",
"spans",
"lines\054",
"continuation",
"lines",
"Otherwise\054",
"interpreted",
"Verbatim",
"Text",
"not",
"preserves",
"It\047s",
"useful",
"displaying",
"commands\056",
"one",
"these",
"\133code\135",
"\133shell\135",
"\133output\135",
"continues",
"until",
"\133end\135",
"seen\056",
"between",
"displayed",
"recognized\056",
"syntax",
"situation",
"same",
"normal",
"exception\072",
"all",
"started\056",
"also",
"change",
"formatting\056",
"Normally",
"\042\133\042",
"\042\135\042",
"However\054",
"pair",
"giving",
"argument",
"For",
"instance\054",
"changes",
"\042\074\042",
"\042\076\042\056",
"\133code\040\074\076",
"\040\040for\050i\0750\073\074red\040i\041\0753\076\073i\053\053\051",
"\040\040\040\040printf\050\042\045d\134n\042\054\040i\051\073",
"\133end",
"\040\040for\050i\0750\073",
"i\041\0753",
"\073i\053\053\051",
"Tables",
"perhaps",
"most",
"complicated",
"system\054",
"tables",
"defined",
"cells",
"rows\056",
"\133table\135",
"data\056",
"encountered",
"row",
"data",
"terminated",
"lone",
"\047\056\047",
"character\054",
"except",
"last",
"row\056",
"think",
"feed\056",
"cell",
"enclosed",
"\042\133\135\042",
"pairs",
"contain",
"both",
"separated",
"plus",
"sign\056",
"All",
"before",
"sign",
"\050if",
"any\051",
"information\056",
"rest",
"supported\072",
"L",
"left",
"justification",
"C",
"center",
"R",
"right",
"indicating",
"column",
"span",
"table\072",
"\133table",
"\040\040",
"\133L\053Justification",
"\135\040",
"\133L\053Left",
"\133C\053Center",
"\133R\053Right",
"\135\040\056",
"\133L\053Column\040span",
"\135\040\040\040",
"\133\053One",
"\1332C\053Two\040Centered",
"\133L\053Inline\040format",
"\133None",
"\133\053\133red\040red\040item",
"\135\135\040",
"\133\053\133ul\040underlined\040item",
"\135\135",
"it\047s",
"rendered\072",
"Justification",
"Center",
"Column",
"One",
"Two",
"Centered",
"None",
"red",
"underlined",
"Images",
"PNG",
"program\056",
"copies",
"image",
"uses",
"run",
"time\056",
"Therefore",
"get",
"quite",
"large\056",
"doesn\047t",
"decompress",
"images",
"representation",
"big",
"four",
"five",
"times",
"original",
"file\056",
"\133image",
"\074filename\076\135",
"filename",
"relative",
"directory",
"directive\072",
"\133image\040world\056png",
"rendering\072",
"Note",
"always",
"elements\054",
"them",
"paragraphs",
"cells\056",
"Links",
"system",
"supports",
"cross",
"references\056",
"reference",
"associated",
"chapter\054",
"subsection\056",
"refer",
"middle",
"section\056",
"written",
"\133link\040\074text\076\045\050name\051\074text\076",
"\040\040\040or\040\040",
"\133link\040\074text\076\045\173name\175\074text\076",
"form\054",
"name",
"replaced",
"target",
"anchor\056",
"removed\056",
"mandatory",
"otherwise",
"would",
"tether",
"link",
"to\056",
"restriction",
"form\056",
"\133link",
"\045\050fmt\056images\051\135",
"Anchors",
"\133anchor\135",
"tag",
"sections\056",
"Then",
"names",
"links\056",
"anchor",
"simple\072",
"\133anchor\040\074name\076",
"rules",
"regarding",
"name\072",
"removed",
"processed\056",
"anchors",
"Anchor",
"case",
"sensitive\056",
"has",
"precede",
"Inline",
"Formatting",
"words",
"formatted\056",
"formatters",
"Directive",
"Result",
"\133red",
"Example",
"Text\135",
"\133ul",
"\133prg",
"Example",
"Text",
"\133expr",
"\133e",
"Example",
"Text",
"\133m",
"Example",
"Text",
"\133b",
"Example",
"Text",
"C\057C\053\053",
"Interface",
"to",
"the",
"System",
"C\057C\053\053",
"interface",
"narrow\056",
"extern",
"\042C\042",
"functions",
"work\072",
"void\040help\050const\040char\040\052subject\051\073",
"void\040help\137setting\137i\050const\040char\040\052key\054\040int\040value\051\073",
"void\040help\137setting\137s\050const\040char\040\052key\054\040const\040char\040\052value\051\073",
"function",
"subject\056",
"subject",
"wish",
"display\056",
"pass",
"NULL",
"shown\056",
"Fonts\054",
"font",
"set",
"latter",
"functions\056",
"Fonts",
"fonts",
"various",
"decorations\056",
"help\137setting\137s",
"function\056",
"family",
"taken",
"settings\056",
"Key",
"Meaning",
"Normal",
"Monospace",
"code\054",
"output",
"shell\056",
"other",
"including",
"headers\054",
"font\056",
"Font",
"Sizes",
"help\137setting\137i",
"function\072",
"size",
"Big",
"mono\137size",
"mono\137big",
"header4",
"Chapter",
"header3",
"Section",
"Subsection",
"Sub\055subsection",
"Colors",
"Foreground",
"fixed",
"black",
"white\056",
"colors\054",
"set\056",
"color",
"integer",
"form",
"0xRRGGBB\056",
"Link",
"hilight",
"Hilight",
"Red",
"Shell",
"Output",
"About",
"foxhelpc\0551\056",
"Contact",
"kt8216\100sdf\056org",
"comments\056",
"Copying",
"Copyright",
"\050c\051",
"2014\054",
"rights",
"reserved\056",
"Redistribution",
"binary",
"forms\054",
"modification\054",
"permitted",
"provided",
"conditions",
"met\072",
"Redistributions",
"retain",
"above",
"notice\054",
"disclaimer\056",
"reproduce",
"disclaimer",
"documentation",
"and\057or",
"materials",
"distribution\056",
"THIS",
"SOFTWARE",
"IS",
"PROVIDED",
"BY",
"THE",
"COPYRIGHT",
"HOLDERS",
"AND",
"CONTRIBUTORS",
"\042AS",
"IS\042",
"ANY",
"EXPRESS",
"OR",
"IMPLIED",
"WARRANTIES\054",
"INCLUDING\054",
"BUT",
"NOT",
"LIMITED",
"TO\054",
"WARRANTIES",
"OF",
"MERCHANTABILITY",
"FITNESS",
"FOR",
"PARTICULAR",
"PURPOSE",
"ARE",
"DISCLAIMED\056",
"IN",
"NO",
"EVENT",
"SHALL",
"HOLDER",
"BE",
"LIABLE",
"DIRECT\054",
"INDIRECT\054",
"INCIDENTAL\054",
"SPECIAL\054",
"EXEMPLARY\054",
"CONSEQUENTIAL",
"DAMAGES",
"\050INCLUDING\054",
"PROCUREMENT",
"SUBSTITUTE",
"GOODS",
"SERVICES\073",
"LOSS",
"USE\054",
"DATA\054",
"PROFITS\073",
"BUSINESS",
"INTERRUPTION\051",
"HOWEVER",
"CAUSED",
"ON",
"THEORY",
"LIABILITY\054",
"WHETHER",
"CONTRACT\054",
"STRICT",
"TORT",
"\050INCLUDING",
"NEGLIGENCE",
"OTHERWISE\051",
"ARISING",
"WAY",
"OUT",
"USE",
"SOFTWARE\054",
"EVEN",
"IF",
"ADVISED",
"POSSIBILITY",
"SUCH",
"DAMAGE\056",
"toolkit",
"licensed",
"under",
"LGPL",
"license\056",
"idea",
"whether",
"mutually",
"compatible",
"not\056",
"should",
"fine",
"full",
"each",
"component\054",
"I\047m",
"lawyer\056",
"Things",
"to",
"Do",
"Term\055definition",
"lists\056",
"Prev",
"Next",
"buttons",
"sequentially",
"reading",
"History",
"prev\055history\054",
"next\055history",
"buttons\056",
"TOC\077\077",
"Searching\056",
"Redo",
"spacing",
"elements\056",
"selection",
"\046",
"copying\056",
"More",
"choices\077",
"User",
"configuration",
"colors\056",
"bullets\056",
"Enumerated",
"\055A\054",
"\055B\054",
"\0551\054",
"\055ii",
"Demo",
"Fox",
"Help",
"Compiler",
"Document",
"Syntax",
"Chapters",
"Sections",
"Formatting",
"Interface",
"System",
"About",
"Copying",
"Things",
"Do",
"Escape",
"Sequences",
"Anchors",
"Fonts",
"Sizes",
"Colors",
};
  canvas->set_words(words,909);
}
typedef struct { void (*func)(FYHelpWin*); int eltno; } sec_t; 
static sec_t sections[]= {
 { cf_0, 0 },
 { cf_1, 0 },
 { cf_2, 0 },
 { cf_2, 161},
 { cf_4, 0 },
 { cf_5, 0 },
 { cf_6, 0 },
 { cf_7, 0 },
 { cf_8, 0 },
 { cf_9, 0 },
 { cf_10, 0 },
 { cf_10, 111},
 { cf_12, 0 },
 { cf_13, 0 },
 { cf_13, 67},
 { cf_13, 133},
 { cf_13, 192},
 { cf_17, 0 },
 { cf_18, 0 },
 { cf_19, 0 },
};

/* This file defines the glue code between the application and the
   help system. It creates the help window, initializes it etc. 
   The contents here are supposed to go after the document functions
   are defined. */

static FYHelpWin *help_window;

static void create_help_window()
{
  if (help_window) return ;  // FIXME: raise it somehow
  help_window= new FYHelpWin( FXApp::instance(),
                              "Help", NULL, NULL, // FIXME: icons
                              DECOR_ALL , 640, 600);
  help_window->show(PLACEMENT_SCREEN);
  help_window->create();
  init_words(help_window->canvas);
  init_doc(help_window);
}

static void follow_link(int linkno)
{
  int ltsize;
  // printf("following link %d\n", linkno);
  ltsize= sizeof(sections)/sizeof(sections[0]);
  if (linkno<0 || linkno>=ltsize) return ;

  (*(sections[linkno].func))(help_window);
  help_window->update();
}

static void show_toc0(int dummy)
{
  cf_toc0(help_window);
  help_window->update();
}

static void show_toc2(int dummy)
{
  cf_toc2(help_window);
  help_window->update();
}

void help(const char *subject)
{
  int i,j,k,R;
  create_help_window(); 
  if (!subject) goto notfound;
  i=0; j=sizeof(anchors)/sizeof(anchors[0])-1;
  while(i<=j)
  {
     k= (i+j)/2;
     R= strcmp(anchors[k].name, subject);
     if (R<0) i= k+1;
     else if (R>0) j= k-1;
     else break;
  }
  if (i>j) goto notfound;
  (*sections[anchors[k].secno].func)(help_window);
  return ;
notfound:
  cf_0(help_window);
}




int main(int argc,char **argv)
{
  FXApp* app;

  app= new FXApp("YAK", "KUT");
  app->init(argc, argv);
  app->create();
  help("fmt.images");
  app->run();
  return 0;
}

